/*++
complete.mu

Author: Marcus Hausdorf, Uni Karlsruhe
        (hausdorf@ira.uka.de)
Created: 	Mon Oct  4 14:54:44 1999
Last update: 	Fri Mar 17 18:44:06 2000
++*/
// 13.11.2000

/*
All global variables are stored in a table detools::data. To enable easier 
access, their value can be obtained by the macro DAT; a new value can be stored
with SDAT. If the variable x is itself a table, the entry y can be assigned the
value z with S2DAT. CDAT is used to check whether detools::data contains an 
index.
This is the generic version of the algorithm. Each instance of it must provide
the procedures:
		prolong
		addEquations
		row2Equation
		addCompCond
*/

alias (DAT(x)=detools::data[x],
       CDAT(x)=contains(detools::data, x),
       SDAT(x,y)=sysassign(detools::data[x],y),
       S2DAT(x,y,z)= sysassign((detools::data[x][y]), z),
       N_ERR=error("wrong number of arguments"),
       A_ERR=error("wrong type of arguments")):
alias (prolong=DAT("prol"),
       addEquations=DAT("addEqs"),
       row2Equation=DAT("r2E"),
       addCompCond=DAT("addCC")):

//===============
// ordered lists 
//===============

// An ordered list is a list of lists of equations. The first element contains 
// the equations of order 0, the second of order 1 etc.
detools::ordInsert := proc(l:DOM_LIST, item, o:DOM_INT):DOM_LIST
    begin
	if nops(l) < (o + 1) then
	    return(l . [[] $ (o - nops(l))] . [[item]]);
	else
	    return(subsop(l, (o + 1) = append(l[o + 1], item)));
	end_if;
    end_proc;
        
detools::ordRemove := proc(l:DOM_LIST, item, o:DOM_INT):DOM_LIST
    local k:DOM_INT;
    begin
	if not iszero((k := contains(l[o + 1], item))) then
	    delete l[o + 1][k];
	    if nops(l) = o + 1 then
		while (o >= 0 and l[o + 1] = []) do
		    delete l[o + 1];
		    o := o - 1;
		end_while;
	    end_if;
	end_if;
	l;
    end_proc;

//=====================
// auxiliary functions 
//=====================
    
// distribute m things on n containers
detools::dmtonc := proc(m, n)
    option remember;
    local i:DOM_INT, k;
    begin
	if n = 1 then
	    return([[m]]);
	else
	    return([(([i] . k) $ k in detools::dmtonc(m - i, n - 1)) 
		$ i = 0..m]);
	end_if;
    end_proc;
    
detools::changeMIndex := proc(mu, A)
    local m2r:DOM_PROC;
    begin
	m2r := (l -> [i $ l[i] $ i = 1..nops(l)]);
	map(map(detools::dmtonc(_plus(op(mu)), nops(mu)), 
	    ll -> [ll, combinat::permute(m2r(ll))]), 
		l -> [_plus(_mult(op(zip(mu2, m2r(mu), 
		    (l1, l2) -> A[l1, l2]))) $ mu2 in l[2]), l[1]]);
    end_proc;
    
detools::transformDF := proc(sys:DOM_LIST, DF:DOM_DOMAIN, newvars:DOM_LIST, A)
    local DV:DOMAIN, DFnew:DOM_DOMAIN, DVnew:DOM_DOMAIN, d_l:DOM_LIST, alldv:DOM_SET,
	tr_t:DOM_TABLE, res_l:DOM_LIST, k:DOM_INT, subs_l:DOM_LIST, dv, A1, RDV:
	DOM_DOMAIN, RDVnew:DOM_DOMAIN;
    begin
	A1 := (domtype(A))::_invert(A);
	DV := DF::Variables;
	DFnew := Dom::LinearDifferentialFunction
	    (Vars=[newvars, map(DV::depVariables, DV::expr)], Rest=[Order < 0]);
	DVnew := DFnew::Variables;
	if not DF::constCoeff then
	    RDV := DF::coeffRing::Variables;
	    RDVnew := DFnew::coeffRing::Variables;
	end_if;
	d_l := map(sys, DF::indets, DiffVars);
	alldv := map(_union(op(d_l)), l -> (if DV::vartype(l) = "Deriv" then DV::multiIndex(l) else l end_if));
	tr_t := table();
	for dv in alldv do
	    if domtype(dv) = DV then
		if DV::vartype(dv) = "Indep" then
		    tr_t[dv] := _plus(_mult(A1[DV::index(dv), i], 
			RDVnew::newMulti("Indep", i)) $ i = 1..nops(newvars));
		else
		    tr_t[dv] := DVnew::newMulti("Dep", DV::index(dv));
		end_if;
	    else
		tr_t[dv] := select(detools::changeMIndex(dv, A), l -> not iszero(l[2]));
	    end_if;
	end_for;
	sys := map(sys, DF::convert_to, DOM_EXPR);
	res_l := [NIL $ nops(sys)];
	for k from 1 to nops(sys) do
	    subs_l := [];
	    for dv in d_l[k] do
		case DV::vartype(dv)
		of "Indep" do
		    subs_l := append(subs_l, RDV::convDV(dv) = tr_t[dv]);
		    break;
		of "Dep" do
		    subs_l := append(subs_l, dv = tr_t[dv]);
		    break;
		of "Deriv" do
		    subs_l := append(subs_l, 
			dv = _plus(op(map(tr_t[DV::multiIndex(dv)], 
			    l -> _mult(l[1], DVnew::newMulti("Deriv", 
				DV::index(dv), l[2]))))));
		end_case;
	    end_for;
	    res_l[k] := subs(sys[k], op(subs_l));
	end_for;
	[map(res_l, DFnew::convert), DFnew];
    end_proc;
	    
detools::transformLDO := proc(sys:DOM_LIST, LDO:DOM_DOMAIN, newvars:DOM_LIST, A)
    local A1, LDOnew:DOM_DOMAIN, subs_l:DOM_LIST, allind:DOM_SET, tr_t:DOM_TABLE, 
	mu:DOM_LIST, j:DOM_INT, k:DOM_INT;
    begin
	A1 := (domtype(A))::_invert(A);
	LDOnew := Dom::LinearDifferentialOperator(newvars, LDO::coeffRing);
	subs_l := zip(LDO::Variables, 
	    [_plus(_mult(newvars[i], A1[i, i2]) 
		$ i = 1..nops(newvars)) 
		    $ i2 = 1..nops(newvars)], (l1,l2) -> l1 = l2);
	sys := map(sys, l -> map(l, LDO::poly2list));
	allind := {op(map(map(map(sys, op), op), op, 2))}; //my favourite line!
	tr_t := table();
	for mu in allind do
	    tr_t[mu] := detools::changeMIndex(mu, A);
	end_for;
	for j from 1 to nops(sys) do
	    for k from 1 to nops(sys[j]) do
		sys[j][k] := map(sys[j][k], 
		    l -> op(map(tr_t[l[2]], 
			l2 -> [subs(_mult(l2[1], l[1]), op(subs_l)), l2[2]])));
	    end_for;
	end_for;
	[map(sys, l -> map(l, LDOnew)), LDOnew];
    end_proc;

// computes a row-echelon-form of sys. The resulting system is ordered.
detools::echelonSys := proc(sys:DOM_LIST):DOM_LIST
    local l1_l:DOM_LIST, l2_l:DOM_LIST, res_l:DOM_LIST, eq:DOM_TABLE, k:DOM_INT;
    begin
	if DAT("Compatibility") then
	    sys := map(sys, l -> row2Equation(l[1], l[2]));
	else
	    sys := map(sys, row2Equation)
	end_if;
	l1_l := [];
	l2_l := [];
	res_l := [];
	for eq in sys do
	    while not iszero((k := contains(l1_l, [eq["leader"], eq["pos"]]))) do
		eq := addEquations(eq, l2_l[k]);
	    end_while;
	    if not iszero(eq["co"]) then
		l1_l := append(l1_l, [eq["leader"], eq["pos"]]);
		l2_l := append(l2_l, eq);
		res_l := detools::ordInsert(res_l, eq, eq["order"]);
	    end_if;
	end_for;
	res_l;
    end_proc;
	
detools::getdatacomp := proc(test:DOM_BOOL):DOM_NULL
    local a_l:DOM_LIST, sys, opt, path, p_flag:DOM_BOOL,
	EQD:DOM_DOMAIN, rm, tmp, newvars:DOM_LIST;
    begin 
	a_l := [];
	if args(0) < 2 then
	    N_ERR;
	else
	    sys := args(2);
	end_if;
	if args(0) > 2 then
	    a_l := [args(3..args(0))];
	end_if;
	if domtype(sys) <> DOM_LIST then
	    if (domtype(sys))::hasProp(Cat::Matrix) then
		SDAT("MType", domtype(sys));
		sys := (domtype(sys))::convert_to(sys, DOM_LIST);
	    else
		A_ERR;
	    end_if;
	end_if;
	for opt in a_l do
	    sysassign(detools::data[expr2text(op(opt,1))],op(opt,2));
	end_for;
	p_flag := FALSE;
	EQD := domtype(sys[1]);
	case EQD 
	of DOM_LIST do
	    EQD := domtype(sys[1][1]);
	    if test then
		if ({op(map(sys, domtype))} <> {DOM_LIST}
		    or nops({op(map(sys, nops))}) <> 1)
		then
		    A_ERR;
		end_if;
	    end_if;
	    SDAT("nIvars", EQD::numVars);
	    SDAT("nDvars", nops(sys[1]));
	    path := pathname("DETOOLS","COMPLETE","LDO");
	    SDAT("addCC", eval(loadproc(detools::addCC,path,"complete")));
	    if CDAT("Random") then
		rm := linalg::randomMatrix(DAT("nIvars"), DAT("nIvars"), 
		    Dom::Integer, DAT("Random"), hold(Unimodular));
		SDAT("RMatrix", rm);
		if CDAT("NewVars") then
		    newvars := DAT("NewVars");
		else
		    newvars := [genident("_X_") $ i = 1..DAT("nIvars")];
		end_if;
		tmp := detools::transformLDO(sys, EQD, newvars, rm);
		sys := tmp[1];
		EQD := tmp[2];
	    end_if;
	    break;
	otherwise
	    if test then
		if {op(map(sys, domtype))} <> {EQD} then
		    A_ERR;
		end_if;
	    end_if;
	    SDAT("nIvars", EQD::Variables::numIndVar);
	    SDAT("nDvars", EQD::Variables::numDepVar);
	    case EQD::constructor
	    of Dom::LinearDifferentialFunction do
		path := pathname("DETOOLS","COMPLETE","LF");
		SDAT("addCC", eval(loadproc(detools::addCC,path,"complete")));
		break;
	    of Dom::DifferentialPolynomial do
		path := pathname("DETOOLS","COMPLETE","DP");
		p_flag := TRUE;
		break;
	    otherwise
		error("not yet (if ever) implemented");
	    end_case;
	    if CDAT("Random") then
		rm := linalg::randomMatrix(DAT("nIvars"), DAT("nIvars"), 
		    Dom::Integer, DAT("Random"), hold(Unimodular));
		SDAT("RMatrix", rm);
		if CDAT("NewVars") then
		    newvars := DAT("NewVars");
		else
		    newvars := [genident("_X_") $ i = 1..DAT("nIvars")];
		end_if;
		tmp := detools::transformDF(sys, EQD, newvars, rm);
		sys := tmp[1];
		EQD := tmp[2];
	    end_if;
	end_case;
        SDAT("r2E", eval(loadproc(detools::r2E,path,"complete")));
	SDAT("prol", eval(loadproc(detools::prol,path,"complete")));
	SDAT("addEqs", eval(loadproc(detools::addEqs,path,"complete")));
	SDAT("eqs", sys);
	SDAT("DF", EQD);
	SDAT("CR", EQD::coeffRing);
	if (DAT("CR"))::hasProp(Cat::Field) then
	    if not CDAT("FractionFree") or p_flag then
		SDAT("FractionFree", FALSE);
	    end_if;
	else
	    SDAT("FractionFree", FALSE);
	end_if;
	if not CDAT("Output") then
	    SDAT("Output", 0);
	end_if;
	if not CDAT("Compatibility") or p_flag then
	    SDAT("Compatibility", FALSE);
	end_if;
	if not CDAT("Elaborate") then
	    SDAT("Elaborate", FALSE);
	end_if;
	null();
    end_proc;
        
// row2Equation (which must be provided for each instance of the algorithm seperately) 
// converts a single row of the systems into a table with the following entries:
// 	ident  - unique identifier
// 	ops    - list of operators
// 	leader - leading operator
// 	pos    - index of the leader's column
// 	co     - coefficient of the leader
// 	class  - class of the leader
// 	order  - order of the leader
// 	lvl    - number of the iteration in which the equation was added
//	g_lvl  - number of the iteration in which the equation was removed
//		 (0 if not yet removed)
// Additionally, as second argument a right-hand-side can be given.

//============================
// prolongation and reduction 
//============================

// is eq1 a multiplicative prolongation of eq2?
// if not, the result is 0.				
detools::isMultProl := proc(eq1:DOM_TABLE, eq2:DOM_TABLE)
    local d:DOM_LIST;
    begin
	if eq1["ident"] <> eq2["ident"] then
	    if eq1["pos"] = eq2["pos"] then
		d := zip(eq1["leader"], eq2["leader"], _subtract);
		if not contains({op(map(d, l -> bool(l >= 0)))}, FALSE) then
		    if eq1["order"] = _plus((op(d, i) $ i = 1..eq2["class"]), 
			eq2["order"]) 
		    then
			return(d);
		    end_if;
		end_if;
	    end_if;
	end_if;
	return(0);
    end_proc;
        
// is eq1 a prolongation of eq2?
// if not, the result is 0.				
detools::isProl := proc(eq1:DOM_TABLE, eq2:DOM_TABLE)
    local d:DOM_LIST;
    begin
	if eq1["pos"] = eq2["pos"] then
	    d := zip(eq1["leader"], eq2["leader"], _subtract);
	    if not contains({op(map(d, l -> bool(l >= 0)))}, FALSE) then
		return(d);
	    end_if;
	end_if;
	return(0);
    end_proc;

detools::reduceMultiplicative := proc(eq:DOM_TABLE, cl:DOM_INT):DOM_TABLE
    local reduced:DOM_BOOL, found:DOM_BOOL, o:DOM_INT, k:DOM_INT, eq2:DOM_TABLE,
	d, pr_l:DOM_LIST, neq:DOM_TABLE, d2, count:DOM_INT, red_l:DOM_LIST,
	cond1:DOM_PROC, cond2:DOM_PROC, cond:DOM_PROC, c_flag:DOM_BOOL;
    begin
	cond1 := ((leq1, leq2) 
	    -> bool((leq2["order"] + cl - leq2["lvl"]) = leq1["order"]));
	cond2 := ((leq1, leq2) 
	    -> bool((leq2["order"] + cl - leq2["lvl"]) >= leq1["order"]));
	if eq["lvl"] = cl then
	    cond := cond2;
	    c_flag := FALSE;
	else
	    cond := cond1;
	    c_flag := TRUE;
	end_if;
	if DAT("sysg_l") = [] then
	    red_l := DAT("sys_l");
	else
	    red_l := zip(DAT("sys_l"), DAT("sysg_l"), _concat, []);
	end_if;
	repeat
	    reduced := FALSE;
	    for o from min(nops(red_l), eq["order"] + 1) downto
		max(eq["order"] - cl + 1, 1) do
		for eq2 in red_l[o] do
		    if (cond(eq, eq2) 
			and (iszero(eq2["g_lvl"])
			    or bool((eq2["order"] + cl - eq2["g_lvl"]) <= eq["order"])))
		    then
			d := detools::isMultProl(eq, eq2);
			if not iszero(d) then
			    SDAT("rs", DAT("rs") + 1);
			    userinfo(5, "    with " . expr2text(eq2["ops"]));   
			    // all intermediate results necessary to compute the
			    // final prolongation are also stored and thus are 
			    // available in later steps 
			    pr_l := (DAT("pr_t"))[eq2["ident"]];
			    found := FALSE;
			    for k from min(_plus(op(d), 1), nops(pr_l)) 
				downto 1 do
				for neq in pr_l[k] do
				    d2 := detools::isProl(eq, neq);
				    if not iszero(d2) then
					count := _plus(op(
					    zip(d, d2, _subtract)));
					break;
				    end_if;
				end_for;
				if not iszero(d2) then 
				    break;
				end_if;
			    end_for;
			    SDAT("lu", DAT("lu") + count);
			    d2 := [i $ d2[i] $ i = 1..nops(d2)];
			    for k in d2 do
				count := count + 1;
				neq := prolong(neq, k);
				SDAT("mp", DAT("mp") + 1);
				pr_l := detools::ordInsert(pr_l, neq, count);
			    end_for;
			    eq := addEquations(eq, neq);
			    userinfo(5, "    to   " . expr2text(eq["ops"]));
			    userinfo(5, "");
			    S2DAT("pr_t", eq2["ident"], pr_l);
			    reduced := TRUE;
			    if c_flag then
				c_flag := FALSE;
				cond := cond2;
			    end_if;
			    break;
			end_if;
		    end_if;
		end_for; 
		if reduced then
		    break;
		end_if;	    
	    end_for; 
	until (not reduced or iszero(eq["co"])) end_repeat;
	eq;
    end_proc;
          
//================
// main algorithm 
//================

/*
Options (so far):

FractionFree - If TRUE, any divisions by elements of the coefficient ring (CR) 
    are avoided. Otherwise, CR must have a method _invert (i.e. it should be 
    a field), and the leading coefficient of any equation is always 1 (if the 
    equation is <> 0, of course).

Compatibility - The Compatibility Conditions of the sytem are computed (only
    possible for the linear cases).
    
Output - Controls the amount of information on the computation given to the 
    user.
	1 - Information about the computation (e.g. number of prolongations /
	    reductions) is shown.
	2 - The development of the parameters r and s is traced and for each
	    iteration the current state of symbol and system is given.
	3 - The dimensions of the intermediate systems and symbols are dis-
	    played.
	4 - The equations of the systems being examined are displayed.
	5 - All of the above, and additionally all reductions and non-multipli-
	    cative prolongations are traced.
	
Elaborate - Instead of the completed system, a table is returned with various 
    additional information for further computations.

Random - If TRUE, a random linear change of the independent coordinates is perfor-
    med on the system. This is to handle delta-singular systems.

NewVars - If a coordinate change (Random = TRUE) is desired, the names of the new variables
    can be specified in a list
*/  
   
detools::complete := proc()
    local binom:DOM_PROC, res_t:DOM_TABLE, lRank:DOM_LIST, systemInvolutive:
	DOM_BOOL, r:DOM_INT, s:DOM_INT, q:DOM_INT, cl:DOM_INT, k:DOM_INT, 
	o:DOM_INT, eq, tmp, w_l:DOM_LIST, neq:DOM_TABLE, tmp1, tmp2, tmp3, oldRank:
	DOM_INT, dimSym:DOM_PROC, dimSys:DOM_PROC, newRank:DOM_INT, ghost_l:DOM_LIST,
	onestep:DOM_BOOL;
    begin
	// binom(k, o) = number of multiplicative prolongations of order o of
	// an equation of class k. (k = 1) -> # of derivatives of order o.    
	binom := proc(k, o)
	    option remember;
	    begin
		binomial(k + o - 1, k - 1);
	    end_proc;
	
	// maximal rank of symbol of order o
	dimSym := proc(o)
	    option remember;
	    begin
		DAT("nDvars") * binom(DAT("nIvars"), o);
	    end_proc;
	    
	// maximal rank of system containing equations up to order o
	dimSys := proc(o)
	    option remember;
	    begin
		DAT("nDvars") * binomial(DAT("nIvars") + o, o)
	    end_proc;
	    
	// Initialisation. The arguments are processed in the extra procedure
	// detools::getdatacomp (see below).
	sysassign(detools::data, table());
	detools::getdatacomp(testargs(), args());
	setuserinfo(detools::complete, DAT("Output"));
	setuserinfo(detools::reduceMultiplicative, DAT("Output"));
	SDAT("sys_l", []);        // real equations   
	SDAT("sysg_l", []);	  // ghost equations
	SDAT("pr_t", table());	  // computed mps of equations in sys_l
	SDAT("nmp", 0);		  // number of nmps
	SDAT("mp", 0);		  // number of mps
	SDAT("rs", 0);		  // number of reduction steps
	SDAT("r20", 0);		  // number of reductions to zero
	SDAT("lu", 0);		  // number of looked-up prolongations
	SDAT("Pivots", table());  // table of pivot-elements
	ghost_l := [];		  // data for ghost equations
	// r and s represent the well-known geometric parameters for the system.
	// cl is increased by one in each iteration; the difference between cl
	// and eq["lvl"] for an equation in sys_l determines the number of multi-
	// plicative prolongations of eq implicitly contained in the current
	// system (of course, it may not exceed q (order of the starting system)
	// + r).  
	r := 0;
	SDAT("R", 0);
	s := 0;
	SDAT("S", 0);
	cl := 0;    
	// system is transformed to echelon form; esp. this means that different
	// equations have different leaders". (Note: I use the terms `equation'
	// and `row' interchangeably, although in the case of LDOs, the rows
	// are not really equations. However, in view of upcoming version of
	// this algorithm for LinearDifferentialFunctions and DifferentialPoly-
	// nomials, I have been used to referring to them as equations.) 
	tmp := DAT("eqs");
	if DAT("Compatibility") then
	    SDAT("CC_l", []);
	    tmp := addCompCond(tmp);
	end_if;
	tmp := detools::echelonSys(tmp);
	tmp2 := table();
	map(map(tmp, op), l -> ((tmp2[l["ident"]] := [[l]])));
	SDAT("pr_t", tmp2);  
	// lRank[p] = # equations of order p in sys_l
	lRank := [nops(tmp[i]) $ i = 1..nops(tmp)];
	SDAT("sys_l", tmp);
	q := nops(DAT("sys_l")) - 1;
	// Loop invariant: the equations in sys_l plus their allowed multiplica-
	// tive prolongation represent the system of DFs R(r, s).
	systemInvolutive := FALSE;
	if (CDAT("OneStep") and DAT("OneStep") = TRUE) then
	    onestep := TRUE;
	else
	    onestep := FALSE;
	end_if;
	repeat
	    oldRank := _plus(op(lRank, 1..(q + r + 1)));
	    userinfo(2, 
		"-----------------------------------------------------");
	    userinfo(3, "");
	    userinfo(3, "                   (".expr2text(s).")");
	    userinfo(3, "Current system is R   :");
	    userinfo(3, "                   ".expr2text(q + r));
	    userinfo(3, "");
	    for eq in map(DAT("sys_l"), op) do
		userinfo(4, "    ".expr2text(eq["ops"]));
	    end_for;
	    userinfo(4, "");
	    userinfo(3, "Dimension: ".expr2text(dimSys(q + r) - oldRank));
	    userinfo(3, "");
	    lRank := append(lRank, 0);
	    w_l := map(DAT("sys_l"), op);
	    r := r + 1;
	    SDAT("R", r);
	    cl := r + s;
	    // Step 1: Compute all nonmultiplicative prolongations of equations
	    // not yet examined; count all new multiplicative prolongations
	    userinfo(5, 
		".....................................................");
	    userinfo(5, 
		"Nonmultiplicative prolongations:");
	    userinfo(5, "");
	    for eq in w_l do
		if cl - eq["lvl"] = 1 then
		    for k from (eq["class"] + 1) to DAT("nIvars") do
			neq := prolong(eq, k);
			neq["ident"] := genident("_lc_eq_");
			neq["lvl"] := cl;
			SDAT("sys_l", detools::ordInsert(DAT("sys_l"), neq, 
			    neq["order"]));
			S2DAT("pr_t", neq["ident"], [[neq]]);
			SDAT("nmp", DAT("nmp") + 1);
			userinfo(5, "    ".expr2text(neq["ops"]));
		    end_for;
		    lRank[eq["order"] + 2] := lRank[eq["order"] + 2] + DAT("nIvars");
		else
		    tmp := eq["order"] + cl - eq["lvl"];
		    lRank[tmp + 1] := lRank[tmp + 1] + 
			binom(eq["class"], cl - eq["lvl"]);
		end_if;
	    end_for;
	    // Compute the rank corrections due to ghost equations. ghost_l con-
	    // tains the necessary data: [eq["order"], eq["class"], eq["g_lvl"], 
	    // eq["lvl"]].
	    for eq in ghost_l do
		tmp := eq[1] + cl - eq[3];
		lRank[tmp + 1] := lRank[tmp + 1] - binom(eq[2], cl - eq[3]);
		tmp := eq[1] + cl - eq[4];
		lRank[tmp + 1] := lRank[tmp + 1] + binom(eq[2], cl - eq[4]);
	    end_for;
	    userinfo(5,"");
	    // Step 2: "Normalize" sys_l, i.e. remove (algebraic) dependencies
	    // in the new system. As only non-multiplicative prolongations are
	    // actually present (the `skeleton' of the system), this leads to
	    // reducing the equations with (allowed) multiplicative prolonga-
	    // tions. 
	    userinfo(5, 
		".....................................................");
	    userinfo(5, "Removing dependencies:");
	    userinfo(5, "");
	    tmp := DAT("sys_l");
	    for o from nops(tmp) downto 1 do
		for eq in tmp[o] do
		    userinfo(5, "Reducing " . expr2text(eq["ops"]));
		    userinfo(5, "");
		    neq := detools::reduceMultiplicative(eq, cl);
		    if eq["ident"] <> neq["ident"] then
			// if an equation has been reduced, it is eliminated 
			// from the system and the new equation is added. Fur-
			// thermore, if mps have been already computed of this
			// equation, it can produce new mps in later iterations
			// although not being present as a `real' equation in
			// the skeleton. This gives rise to the phenomenon of
			// `ghost equations', which are stored in sysg_l.
			SDAT("sys_l", detools::ordRemove(DAT("sys_l"), 
			    eq, eq["order"]));
			if eq["lvl"] < cl then
			    eq["g_lvl"] := cl;
			    SDAT("sysg_l", detools::ordInsert(DAT("sysg_l"), eq, 
				eq["order"]));
			    ghost_l := append(ghost_l, [eq["order"], eq["class"],
				eq["g_lvl"], eq["lvl"]]);
			end_if;
			// if reduction <> 0, the system cannot be involutive.
			// Reductions to 0 might create compatibility condi-
			// tions.
			if not iszero(neq["co"]) then
			    neq["lvl"] := cl;
			    SDAT("sys_l", detools::ordInsert(DAT("sys_l"), 
				neq, neq["order"]));
			    S2DAT("pr_t", neq["ident"], [[neq]]);
	       		    lRank[neq["order"] + 1] := lRank[neq["order"] + 1] + 1;
			else
			    if DAT("Compatibility") then
				SDAT("CC_l", append(DAT("CC_l"), neq["rhsops"]));
			    end_if;
			    SDAT("r20", DAT("r20") + 1);
			end_if;
			lRank[eq["order"] + 1] := lRank[eq["order"] + 1] - 1;
		    end_if;
		end_for;
	    end_for;
	    userinfo(5, 
		".....................................................");
	    // Step 3: Analyse the result of the iteration. 
	    // If there exist equations of maximal order, the symbol is not 
	    // involutive.
	    if nops(DAT("sys_l")) = (q + r + 1) then
		userinfo(2, " (".expr2text(s).")");
		userinfo(2, "M    not involutive");
		userinfo(2, " ".expr2text(q + r - 1));
		userinfo(2, "");
		userinfo(3, "Dimension: ".expr2text(dimSym(q + r - 1) 
		    - lRank[q + r]));
		userinfo(3, "");
	    else
		userinfo(2, " (".expr2text(s).")");
		userinfo(2, "M    involutive");
		userinfo(2, " ".expr2text(q + r - 1));
		userinfo(2, "");
		userinfo(3, "Dimension: ".expr2text(dimSym(q + r) 
		    - lRank[q + r + 1]));
		userinfo(3, "");	
		// If some equation was not reduced to zero, the system is
		// not involutive.
		newRank := _plus(op(lRank, 1..(q + r)));
		if newRank > oldRank then
		    // For a non-involutive system with involutive symbol, pro-
		    // jections take place. 
		    userinfo(2, " (".expr2text(s).")");
		    userinfo(2, "R    not involutive");
		    userinfo(2, " ".expr2text(q + r - 1));
		    userinfo(2, "");
		    tmp := nops(DAT("sys_l")) - 1;
		    s := s + q + r - tmp;
		    SDAT("S", s);
		    r := tmp - q;
		    SDAT("R", r);
		else
		    userinfo(2, " (".expr2text(s).")");
		    userinfo(2, "R    involutive");
		    userinfo(2, " ".expr2text(q + r - 1));
		    userinfo(2, "");
		    systemInvolutive := TRUE;
		end_if;
	    end_if;
	until (systemInvolutive or onestep) end_repeat;
	userinfo(1, 
	    "-----------------------------------------------------");
	userinfo(2, "");
        userinfo(2, "                 (".expr2text(s).")");
	userinfo(2, "Final system is R   :");
	userinfo(2, "                 ".expr2text(q + r - 1));
	userinfo(2, "");
	for eq in map(DAT("sys_l"), op) do
	    userinfo(4, "    ".expr2text(eq["ops"]));
	end_for;
	userinfo(4, "");
	userinfo(3, "Dimension: ".expr2text(dimSys(q + r - 1) - newRank));
	userinfo(3, "");
	userinfo(3, "");
	userinfo(1, "Nonmultiplicative prolongations: ".expr2text(DAT("nmp")));
	userinfo(1, "Multiplicative prolongations:    ".expr2text(DAT("mp")));
	userinfo(1, "Looked-up prolongations:         ".expr2text(DAT("lu")));
	userinfo(1, "Reduction steps:                 ".expr2text(DAT("rs")));
	userinfo(1, "Reductions to 0:                 ".expr2text(DAT("r20")));
	tmp1 := map(map(DAT("sys_l"), op), l -> l["ops"]);
	if CDAT("MType") then
	    tmp1 := (DAT("MType"))(tmp1);
	end_if;
	if DAT("Compatibility") then
	    tmp2 := map(map(DAT("sys_l"), op), l -> l["rhsops"]);
	    tmp3 := DAT("CC_l");
	    if CDAT("MType") then
		tmp2 := (DAT("MType"))(tmp2);
		if tmp3 <> [] then
		    tmp3 := (DAT("MType"))(tmp3);
		end_if;
	    end_if;
	end_if;
	if DAT("Elaborate") then
	    // The returned table res_t contains the following infor-
	    // mation:
	    // 	 Eqs - equations in the resulting system
	    //	 R, S - obvious
	    // 	 DF, CR - used domains
	    //	 CC - compatibility conditions
	    //	 Pivots - pivot elements
	    //	 RankList - # equations for each order
	    res_t := table();
	    res_t["eqs"] := tmp1;
	    if DAT("Compatibility") then
		res_t["Rhseqs"] := tmp2;
		res_t["CC"] := tmp3;
		res_t["CCDOM"] := DAT("CCDOM");
	    end_if;
	    res_t["R"] := q + r - 1;
	    res_t["S"] := s;
	    res_t["DF"] := DAT("DF");
	    res_t["CR"] := DAT("CR");
	    res_t["Pivots"] := DAT("Pivots");
	    res_t["RankList"] := [op(lRank, 1..(q + r))];
	    if CDAT("Random") then
		res_t["RMatrix"] := DAT("RMatrix");
	    end_if;
	    sysassign(detools::data, null());
	    return(res_t);
	else
	    if DAT("Compatibility") then
		sysassign(detools::data, null());
		return([tmp1, tmp2, tmp3]);
	    else
        	sysassign(detools::data, null());
		return(tmp1);
	    end_if;
	end_if;
    end_proc;

unalias (DAT, CDAT, SDAT, S2DAT, N_ERR, A_ERR, prolong, row2Equation, addEquations, addCompCond):




