/*++
ncdetsys.mu
Author : Marcus Hausdorf
Created : November 17th, 1998
Last Update: 	Fri Aug 25 12:30:58 2000

detools::ncDetSys(sys) : Compute the determining system for a non-
    classical symmetry group of a system of partial differential 
    equations. This is done by augmenting sys by the invariant surface
    conditions determined by the ansatz. detools::ncDetSys is called 
    in the same way as detools::detSys, with the following additional
    optional argument:
    
    - Steps : DOM_INT
	Computes the higher-order integrability conditions 
        of order n, i.e. the ansatz is mapped n times to the equations
        in sys with the resulting equations being added to sys in the
        next step.
++*/

alias (m_print(x) = print(Unquoted, x),
       DAT(x)=detools::data[x],
       CDAT(x)=contains(detools::data,x),
       SDAT(x,y)=sysassign(detools::data[x],y),
       T_ERR = error("wrong type of argument"),
       N_ERR = error("wrong number of arguments"),
       R_ERR = error("argument out of range"),
       D_ERR = error("unable to create DV/DF"),
       C_ERR = error("unable to convert equations"),
       S_ERR = error("unable to solve equations"),
       A_ERR = error("unable to create ansatz")):
       
detools::ncDetSys := proc()
local sys, ans, s_l, n:DOM_INT, DV:DOM_DOMAIN, DVout:DOM_DOMAIN, DF:DOM_DOMAIN, 
    JDF:DOM_DOMAIN, DFout:DOM_DOMAIN, JVF:DOM_DOMAIN, v_l:DOM_LIST, pr_ans, m:
    DOM_INT, dl, ds, eq, k:DOM_INT;
begin
    // create the necessary domains and convert equations
    detools::getdatasymm(testargs(), args());
    ans := detools::getAnsatz();
    s_l := detools::simpSList(DAT("S_eqs"));
    sys := DAT("Eqs");
    DV := DAT("DVsymm");
    DF := DAT("DFsymm");
    JDF := DAT("JDFsymm");
    JVF := DAT("JVFsymm");
    if CDAT("Selected") then
	if (not testtype(DAT("Selected"), Type::ListOf(Type::PosInt))
	    or max(op(DAT("Selected"))) > nops(sys))
	then
	    R_ERR;
	end_if;
	sys := [sys[i] $ i in DAT("Selected")];
    end_if;
    // determine domains for output
    if DAT("Expr") <> TRUE then
	DVout := Dom::DifferentialVariable(map(DV::indVariables . 
	    DV::depVariables, DV::expr), DAT("Params"));
	if DAT("Linear") or DAT("Polynomial") then
	    DFout := Dom::DifferentialPolynomial(DVout, 
		Dom::DifferentialExpression(Dom::RestrictedDifferentialVariable
		(DVout, Order < 0)));
	else
	    DFout := Dom::DifferentialExpression(DVout);
	end_if;
	SDAT("DVsymmout", DVout);
	SDAT("DFsymmout", DFout);
    end_if;	    
    // add ISCs
    s_l := s_l . JVF::characteristic(ans);
    // number of steps
    if not CDAT("Steps") then
        n := 1
    else
        n := DAT("Steps");
        if domtype(n) <> DOM_INT or n < 1 then T_ERR end_if;
    end_if;
    // compute determining system
    if n = 1 then
	v_l := [op(select(_union(DF::indets(sys[i], DiffVars) $ 
	    i = 1..nops(sys)), l -> (DV::vartype(l) = "Deriv")))];
	pr_ans := JVF::prolong(ans, v_l);
    else
	pr_ans := JVF::prolong(ans, max(op(map(sys, DF::order))));
    end_if;
    for k from 1 to n do
        sys := map(sys, pr_ans);
	for m from 1 to nops(s_l) do
            if domtype(s_l[m]) <> DOM_LIST then
		if DAT("Interactive") then
                    m_print("Solve ".expr2text(s_l[m]));
                    dl := input("for ?", dl);
                    if dl = hold(leader) then 
                        dl := JDF::leader(s_l[m]);
                    else
                        if (dl := DV(dl)) = FAIL 
                            or iszero(contains(
				JDF::indets(s_l[m], DiffVars), dl))
                        then 
			    error ("unable to convert input") 
                        end_if;
                    end_if;
                else
                    dl := JDF::leader(s_l[m]);
                end_if;
                ds := JDF::solve(s_l[m], dl);
                if ds = FAIL then S_ERR end_if;
                s_l[m] := [dl, ds];
            end_if;
	end_for;
	s_l := detools::simpSList(s_l);
	for eq in s_l do 
	    sys := map(sys, JDF::subs, eq[1] = eq[2]);
	end_for; 
        s_l := s_l . sys;
    end_for;
    if DAT("Linear") or DAT("Polynomial") then
	sys := map(map(map(sys, JDF::coeff), (JDF::coeffRing)::expr), numer);
    else
	sys := map(sys, ((l) -> (JDF::numer(l))));
	sys := map(map(sys, ((l) -> (JDF::poly(l, select(JDF::diffVars(l), 
	    DV::vartype = "Deriv"), hold(Expr))))), coeff);
    end_if; 
    if DAT("Expr") = TRUE then
	if CDAT("Sorted") and DAT("Sorted") = TRUE then
	    sys := sort(sys, (l1, l2) -> (bool(length(l1) < length(l2))));
	end_if;
	return(sys);
    end_if;
    sys := map(sys, DFout::coerce);
    if DAT("Autoreduced") then
	sys := DFout::autoreduce(sys);
    end_if;
    if DAT("Expr") = hold(NoDiff) then
	sys := map(sys, DFout::expr);
    end_if;
    if CDAT("Sorted") and DAT("Sorted") = TRUE then
	sys := sort(sys, (l1, l2) -> (bool(length(l1) < length(l2))));
    end_if;
    sys;
end_proc:

unalias(m_print, DAT, CDAT, SDAT, T_ERR, N_ERR, R_ERR, D_ERR, C_ERR, S_ERR,
    A_ERR):
	    
	
	
	

