/*++
Euler operator of variational calculus
        Lagrangians in any number of independent and dependent variables
        and of any order can be handled. The signs are chosen such that
        the highest order terms get a positive sign.

euler(L,t,z) - applies Euler operator to the Lagrangian L
                t represents the independent and z the dependent variables

There are three possibilities to denote the coordinates t,z. One can enter
an identifier (DOM_IDENT) for them, one can enter a list of identifiers,
or one can enter an indexed identifier. In the latter case the index
represents the number of coordinates.

Examples:
1.) one-dimensional particle in potential V
        >> L := 1/2*diff(q(t),t)^2 - V(q(t));
        >> euler(L,t,q);
2.) two-dimensional particle in potential V
        >> L := 1/2*(diff(x(t),t)^2+diff(y(t),t)^2) - V(x(t),y(t));
        >> euler(L,t,[x,y]);
    or with indexed variables
        >> L := 1/2*(diff(q[1](t),t)^2+diff(q[2](t),t)^2) -
                V(q[1](t),q[2](t));
        >> euler(L,t,q[2]);
3.) field theory
        >> L := 1/2*(diff(psi(t,x),t)^2+diff(psi(t,x),x)^2) - psi(t,x)^2;
        >> euler(L,[t,x],psi);

Internally the domains DV := Dom::DifferentialVariable(t,z) and
DE := Dom::DifferentialExpression(DV) are used. Thus effectively t and z
can be anything that is valid as input for this domain and L must be
written in such a way that it can be converted to DE.

It is also possible to call euler directly with domain elements in order to
save time consuming conversions. If L belongs to a domain from the category
Cat::DifferentialFunction(DV), then this domain is used instead of DE
and the output will also be of this domain. It is also possible to enter
DV directly. In this case euler takes only two arguments: euler(L,DV).


Author: Werner M. Seiler, Uni Karlsruhe/Mannheim
	(wms@ira.uka.de, seiler@euler.math.uni-mannheim.de)
Created: October 15th, 1997
Last update: 	Tue Feb 15 16:05:43 2000
++*/

alias(DAT(x)=detools::data[x]):


detools::euler := 
proc()
local ele:DOM_LIST,i:DOM_INT,m:DOM_INT,q:DOM_INT,vars:DOM_LIST,dvars:DOM_LIST,
      dv,mu:DOM_LIST,dL;
begin
    detools::getdata(testargs(),args());
    m := (DAT("DVin"))::numDepVar;
    sysassign(detools::data["System"],bool(m>1));
    q := (DAT("DFin"))::order(DAT("Eqs")[1]) mod 2;
    vars := (DAT("DFin"))::indets(DAT("Eqs")[1],DiffVars);
    dvars := select(vars, ((v) -> bool((DAT("DVin"))::vartype(v)="Deriv")));

    ele := [ (DAT("DFin"))::zero $ m ];
    for i from 1 to m do
	dv := (DAT("DVin"))::new("Dep",i);
	if not(iszero(contains(vars,dv))) then
	    ele[i] := (DAT("DFin"))::diff(DAT("Eqs")[1],dv);
	    if q=1 then ele[i] := (DAT("DFin"))::_negate(ele[i]) end_if;
	end_if;
	for dv in dvars do
	    if (DAT("DVin"))::index(dv)<>i then next end_if;
	    mu := (DAT("DVin"))::repeatedIndex(dv);
	    dL := (DAT("DFin"))::totalDiff(
			    (DAT("DFin"))::diff(DAT("Eqs")[1],dv),
			    mu);
	    if ((DAT("DVin"))::order(dv) mod 2)=q then
		ele[i] := (DAT("DFin"))::_plus(ele[i],dL);
	    else
		ele[i] := (DAT("DFin"))::_subtract(ele[i],dL);
	    end_if;
	end_for;
    end_for;

    detools::output(ele);
end_proc:


unalias(DAT):
