/*++
The method of the modified equation for the analysis of one-step methods
for the numerical solution of ordinary differential equations.

This is an improved MuPAD version of the Maple code presented by Ahmed
and Corless at the IMACS ACA conference 1997 (Maui, Hawaii). It still
follows the brute-force Taylor series approach, but care is taken that
no unnecessary terms are computed. The interface is a bit more user
friendly and we use lists instead of vectors.

Usage: Write your numerical scheme in the form of a one-step method
	    y_{k+1} = y_k + h*Psi(h,y_k)
then call
	    modode(F,z,t,h,N);
where F is a list containing the components of Psi, z a list with the
*names* of the unknown functions, t the name of the independent
variable, h the name of the step size and N the order of the modified
equation.  The output is the right hand side of the modified equation.

Example: In order to compute the modified equation of order 3 for the Euler
method applied to the system y'=z, z'=-y enter
	    modode([z,-y],[y,z],t,h,3);

For methods implemented in the library procedure numeric::odesolve, it is
also possible to use another form of modode. Define F as a procedure in the
same way as required by numeric::odesolve and call
            modode(F,method,z,t,h,N)
where method is the name of the numerical scheme. In the example above one
could write
            F := proc(t,y) begin [y[2],-y[1]] end_proc;
            modode(F,EULER1,[y,z],t,h,3);


Author: Werner M. Seiler, Uni Karlsruhe/Mannheim
	(wms@ira.uka.de, seiler@euler.math.uni-mannheim.de)
Created: November 13th, 1997
Last update: 	Mon Mar  1 20:31:29 1999
++*/


detools::modode := 
proc()
local Psi, y, x, h, hp, N:DOM_INT, method, t:DOM_IDENT, dim:DOM_INT, 
      lte:DOM_LIST, deriv:DOM_LIST, dy:DOM_LIST, i:DOM_INT, j:DOM_INT, 
      k:DOM_INT, m:DOM_INT, cut:DOM_PROC, cutsubs:DOM_PROC;
begin
    if args(0)=5 then
        Psi := args(1);
        y := args(2);
        x := args(3);
        h := args(4);
        N := args(5);
    elif args(0)=6 then
        Psi := args(1);
        method := args(2);
        y := args(3);
        x := args(4);
        h := args(5);
        N := args(6);
    else error("wrong no of args") end_if;
    if testargs() then
        if (domtype(y)<>DOM_LIST) or (domtype(x)<>DOM_IDENT) or
           (domtype(h)<>DOM_IDENT) or (domtype(N)<>DOM_INT) or
           ((args(0)=5) and (domtype(Psi)<>DOM_LIST)) or
           ((args(0)=6) and ((domtype(method)<>DOM_IDENT) or
                             (domtype(Psi)<>DOM_PROC))) then
            error("wrong type of argument");
        end_if;
    end_if;
    t := genident("t");
    dim := nops(y);
    if args(0)=6 then
        Psi := numeric::odesolve(x..x+h,Psi,y,method,hold(Symbolic));
	Psi := map(map(zip(Psi,y,_subtract),_divide,h),normal);
    end_if;
    if testargs() and nops(Psi)<>dim then
        error("incompatible dimensions");
    end_if;

    // two auxiliary functions for handling polynomials 
    cut := (p,h,m) -> ((if m>=degree(p) then p
	         else poly([[coeff(p,h,i),i] $ i=0..m],[h]) end_if));
    cutsubs := (p,l,r,h,m) -> (poly(_plus(subs(coeff(p,h,i),
				    l=expr(cut(r,h,m-i)))*h^i
			       $ i=0..m),[h]));

    /* compute the local truncation error
      and its derivatives up to the needed order */
    dy := [ [diff(y[j](t),t$k) $ j=1..dim ] $ k=1..N+1 ];
    lte := [ 0 $ dim ];
    hp := 1;
    for i from 2 to N+1 do
	hp := h*hp/i;
	lte := zip(lte,[hp*dy[i][j] $ j=1..dim],_plus);
    end_for;
    lte := map(lte,poly,[h]);
    if has(Psi,h) then
	Psi := map(map(Psi,series,h,N+1),expr);
    end_if;
    Psi := map(map(Psi,subs,map(y,((f,t)->f=f(t)),t)),poly,[h]);
    deriv := [ 0 $ N+1 ];
    deriv[1] := zip(Psi,lte,_subtract);
    for i from 2 to N+1 do
	deriv[i] := map(map(deriv[i-1],cut,h,N+1-i),diff,t);
    end_for;

    /* eliminate step by step higher order derivatives of y
      using the prolonged equations */
    for i from N downto 1 do
        for j from i+1 to N+1 do
	    for k from 1 to dim do
	        deriv[i] := map(deriv[i],cutsubs,
			        dy[j][k],deriv[j][k],h,N+1-i);
	    end_for;
	end_for;
	/* deriv[i] contains now only derivatives up to order i;
	  those are eliminated recursively */
	m := 0;
	while 2^m <= N+1-i do
	    for k from 1 to dim do
	        deriv[i] := map(deriv[i],cutsubs,
			        dy[i][k],deriv[i][k],h,N+1-i);
	    end_for;
	    m := m+1;
	end_while;
	if i>1 then
	    for j from i+1 to N+1 do
		for k from 1 to dim do
		    deriv[j] := map(deriv[j],cutsubs,
			            dy[i][k],deriv[i][k],h,N+1-i);
		end_for;
	    end_for;
	end_if;
    end_for;

    // return right hand side of modified equation 
    map(map(deriv[1],subs,map(y,((f,t)->f(t)=f),t)),expr);
end_proc:

