/*++
Method of characteristics for (quasi-)linear equations of first order

Three methods: detools::charODESystem for finding the system of odes which
                                   will determine the characteristics 
	       detools::characteristics for computing characteristics
	       detools::charSolve for solving (quasi-)linear PDEs	       

detools::charODESystem(ldf,var,<initcond>)
    where ldf is a linear differential function (charSystem will
    consider ldf to be equivalent to the equation ldf=0), var is
    a variable to be used as the independent variable in the
    characteristics, initcond is a set of  optional initial conditions
    (an initial curve for the characteristics), for example,
    if x,t are the independent variables, u the dependent variable,
    an initial condition could be {x=T^2,t=2*T,u=T-1}.

detools::characteristics(ldf,var,<initcond>)
    where ldf, var and initcond are as above.

detools::charSolve(ldf,init,par)
    where ldf is an element of this domain (first order, assumed to
    represent the equation ldf = 0), init is a set parameterizing the
    initial "curve" (an n-1 dimensional surface if there are n independent
    variables) and par is the set of parameters used in init (so there
    should be n-1 of them).
    This method will try to return the solution of ldf = 0 passing through
    the initial curve.  It will call characteristics(ldf,init).  If
    characteristics does not return a solution, this method will return
    the original information.  If characteristics does return a solution,
    this method will attempt to eliminate the parameters and return the
    function.  If this is not possible, this method will return the
    characteristics.


Author: Jay Belanger (belanger@truman.edu)
	(adapted to detools library by Werner M. Seiler)
Created: 	Thu Apr 29 19:50:48 1999
Last update: 	Fri Apr 30 11:16:15 1999
++*/

alias(DAT(x)=detools::data[x],
      CDAT(x)=contains(detools::data,x)):


detools::charODESystem :=
    proc()
    local sys:DOM_SET, fn:DOM_SET, u:DOM_EXPR, i:DOM_INT, j:DOM_INT,
	  DVI:DOM_LIST, iv:DOM_EXPR, c, cff, dv, vars_to_functions:DOM_PROC, 
	  ldf, var:DOM_IDENT;
    begin
	// Analysis of input
	detools::getdata(testargs(),args());
	if DAT("System") then error("cannot handle systems") end_if;
	if not DAT("Linear") then error("non-linear equation") end_if;
	ldf := DAT("Eqs")[1];
	if testargs() then
	    if args(0)<2 or args(0)>3 then 
		error("wrong number of args") 
	    end_if;
	    if (DAT("DFin"))::order(ldf)<>1 then
		error("Differential equation must be first order");
	    end_if;
	end_if;
	if DAT("Parms")=[] then error("no variable name given") end_if;
	var := DAT("Parms")[1];
	
	// Now, we need to make sure that there is only one dependent 
	// variable.  "dv" will be a the set of all dependent
	// variables in the equation.
	dv := (DAT("DFin"))::dVars(ldf);
	// If "dv" has more than one element, we need to signal an error,
	// otherwise make "u" the dependent variable
	if nops(dv) <> 1 then
	    error("There must be exactly one dependent variable");
	end_if;
	dv := op(dv);
	DVI := (DAT("DVin"))::indVariables;
	u := (DAT("DVin"))::expr(dv);
	// "vars_to_functions" will turn independent variables (e.g. "x")
	// into functions (e.g. "x(S)").
	vars_to_functions:= 
	    proc (ex:DAT("DFin"))
	    local k:(DAT("DVin"));
	    begin
		ex := expr(ex); 
		if DAT("Quasi") then
		    ex := subs(ex,u=u(var));
		end_if;
		for k in DVI do
		    k := (DAT("DVin"))::expr(k);
		    ex := subs(ex,k=k(var));
		end_for;
	    end_proc;
	// "sys" will be the set of odes
	if DAT("Quasi") then
	  sys := {hold(diff)(u(var),var) = (-1)*vars_to_functions(op(ldf,2))};
        else
	  cff := (DAT("DFin"))::coeff(ldf,dv);
	  cff := expr(cff);
  	  sys := {hold(diff)(u(var),var) = (-1)*cff*u(var)};
	end_if;
       	for i from 1 to (DAT("DVin"))::numIndVar do
	    iv := (DAT("DVin"))::expr(DVI[i])(var); /* the ith variable */
	    c := (DAT("DFin"))::coeff(ldf,
		          (DAT("DVin"))::newRepeated("Deriv",
			           (DAT("DVin"))::index(dv),[i]));
	                       /* the coefficient of the derivative
                                   w.r.t. the ith variable */
	    sys := sys union {hold(diff)(iv,var)=
				    vars_to_functions(c,var)};
	end_for;
	// Next, check for optional initial conditions.  If given,
	// they are contained in detools::data["Parms"] (we assume
	// that no other parameters are given).
	if nops(DAT("Parms"))>1 then
	    for j in DAT("Parms")[2] do
		// Add the intial conditions to sys, replacing
		// an equation like x=T by x(0)=T
		sys := sys union {op(j,1)(0)=op(j,2)};
	    end_for;
	end_if;
      	// Next, create a list of functions in the system of odes
	fn := DVI . [dv];
	fn := {op(map(fn, ((ldv,lv) -> (DAT("DVin"))::expr(ldv)(lv)), var))};
	// Now, either return the solved system of odes or the unsolved
	// system of odes.
	return(ode(sys,fn));
    end_proc:



detools::characteristics :=
    proc()
    local sys:DOM_SET, ldf;
    begin
	// Analysis of input
	detools::getdata(testargs(),args());
	ldf := DAT("Eqs")[1];
	if testargs() then
	    if args(0)<2 or args(0)>3 then error("wrong no of args") end_if;
	    if (DAT("DFin"))::order(ldf)<>1 then
		error("Differential equation must be first order");
	    end_if;
	end_if;
       	sys := detools::charODESystem(ldf,op(DAT("Parms")));
	return(solve(sys));
    end_proc:



detools::charSolve :=
    proc()
    local soln, fn, var:DOM_IDENT, char, k, dv, functions_to_vars:DOM_PROC,
	  init:DOM_SET, par:DOM_SET, ldf;
    begin
	// Analysis of input
	detools::getdata(testargs(),args());
	if DAT("System") then error("cannot handle systems") end_if;
	ldf := DAT("Eqs")[1];
	if testargs() then
	    if nops(DAT("Parms"))<>2 then 
		error("wrong number of arguments");
	    end_if;
	    if (DAT("DFin"))::order(ldf)<>1 then
		error("Differential equation must be first order");
	    end_if;
	end_if;
	init := DAT("Parms")[1];
	par := DAT("Parms")[2];
	
	// This function should replace expressions of the form "x(S) = ..."
	// by "x = ..."
	functions_to_vars :=
	    proc(l)
	        begin op(l,[1,0]) = op(l,2) 
	    end_proc;
	// First of all, we need a new variable for "characteristics"
	var := genident();
	// Now we need to try to get the characteristics.
	// If the system of odes cannot be solved by MuPAD, we'll return
	// the original information
	char := detools::characteristics(ldf,var,init);
	if type(char)="solve" then return(ldf,init,par) end_if;
	// For the time being, we'll assume that all "solve"s return only
	// one solution set (or rather, we'll use the first solution set).
	// If necessary, this should be easy enough to change.
	soln := op(char,1);
	// After "characteristics" goes to the trouble of changing the
	// independent variables into functions, we need to change them
	// back into variables.
	soln := map(soln,functions_to_vars);
	// Next, seperate the actual function that we are trying to solve
	// for from the other variables.
	// We need to find out which dependent variable appears in the
	// equation.  ("characteristics" already does this, but I'm
	// not sure that we want another argument passed.)
	dv:=(DAT("DFin"))::dVars(ldf);
	dv := op(dv);
	soln := split(soln,has,(DAT("DVin"))::expr(dv));
	fn := op(soln[1],1);
	soln := soln[2];
	// Next, we'll solve for var and parameters in terms of the 
	// independent variables (if this cannot be done, we'll return the 
	// characteristics)
	soln := solve(soln, par union {var});
	if type(soln)="solve" then return(char) end_if;
	soln := op(soln,1);
	// If there are no solutions, then the characteristics lie in the
	// initial "curve" (or somesuch)  I'm not sure how this should be
	// handled, for the time being we'll just return the latest result
	if domtype(soln)=DOM_FAIL then return(char) end_if;
	// Finally, in the expression for the function we are solving for,
	// replace var and the parameters with the independent variables.
	for k in soln do
	    fn := subs(fn,k);
	end_for;
	// tadaa
	return(fn);
    end_proc:


unalias(DAT,CDAT):



