/*++
getdata --- analyses input of procedures of the detools library

getdata checks whether the given differential equations belong to some basic
domain or to some of the domains for differential equations. In the first case
it tries to convert it to domains of rising complexity: linear equations with
constant coefficients, linear equations with variable coefficients, polynomial
equations, quasi-polynomial equations and finally general differential
expressions. getdata also extracts all passed options and generates newly the
table detools::data.

The first argument of getdata is a boolean value specifying the value of
testargs() in the method calling getdata.


Author: Werner M. Seiler, Uni Karlsruhe/Mannheim
	(wms@ira.uka.de, seiler@euler.math.uni-mannheim.de)
Created: Feb 28th, 1999
Last update: 	Tue Feb 15 16:05:47 2000
++*/

alias(SDAT(x,y)=sysassign(detools::data[x],y)):


// auxiliary method; recognizes all differential variables
detools::subsDV :=
proc(de,DV)
local tmp;
begin
    tmp := DV::coerce(de);
    if tmp<>FAIL then return(tmp) end_if;
    case domtype(de)
    of DOM_INT do
    of DOM_RAT do
    of DOM_COMPLEX do
    of DOM_FLOAT do
    of DOM_BOOL do return(de);
    otherwise return(map(de,detools::subsDV,DV)); 
    end_case;
end_proc;


detools::getdata :=
proc(test:DOM_BOOL)
local tmp:DOM_LIST, sys:DOM_LIST, opt:DOM_EXPR, n:DOM_INT,
      DV:DOM_DOMAIN, DF:DOM_DOMAIN, 
      RDV1:DOM_DOMAIN, RDE1:DOM_DOMAIN,
      RDV2:DOM_DOMAIN, RDE2:DOM_DOMAIN;
begin
    sysassign(detools::data,table());		// initialize table

    // analyse differential equations
    
    if domtype(args(2))<>DOM_LIST then		// check whether system
	SDAT("System",FALSE);
	tmp := [args(2)];
    else
	SDAT("System",TRUE);
	tmp := args(2);
    end_if;
    DF := domtype(tmp[1]);	// check domains
    DV := DF::Variables;
    if DV=FAIL or DV::hasProp=FAIL or		 // no differential domain
       not(DV::hasProp(Cat::DifferentialVariable)) then
        if test and args(0)<4 then
	    error("wrong number of arguments");
	end_if;
        DV := Dom::DifferentialVariable(args(3),args(4));
	// in order to speed up subsequent conversion attempts
	// all differential variables are recognized already now
	tmp := map(tmp,detools::subsDV,DV);
	SDAT("Linear", FALSE);
	SDAT("ConstCoeff", FALSE);
	SDAT("Quasi", FALSE);
	SDAT("Polynomial", FALSE);
	// try linear, constant coefficients
	DF := Dom::LinearDifferentialFunction(DV);
	sys := map(tmp,DF::coerce);
	if iszero(contains(sys,FAIL)) then
	    SDAT("Linear", TRUE);
	    SDAT("ConstCoeff", TRUE);
	else				// try linear, variable coefficients
	    RDV1 := Dom::RestrictedDifferentialVariable(DV,
					    hold(Types) = {"Indep"});
	    RDE1 := Dom::DifferentialExpression(RDV1);
	    DF := Dom::LinearDifferentialFunction(DV, RDE1);
	    sys := map(tmp,DF::coerce);
	    if iszero(contains(sys,FAIL)) then SDAT("Linear",TRUE);
	    else    			// try "quasi-linear"
		RDV2 := Dom::RestrictedDifferentialVariable(DV,
					    hold(Types) = {"Indep", "Dep"});
		RDE2 := Dom::DifferentialExpression(RDV2);
		DF := Dom::LinearDifferentialFunction(DV, RDE2);
		sys := map(tmp,DF::coerce);
		if iszero(contains(sys,FAIL)) then
		    SDAT("Linear", TRUE);
		    SDAT("Quasi", TRUE);
		end_if;
	    end_if;
	end_if;
	if not detools::data["Linear"] then
	    // try polynomial, constant coefficients
	    DF := Dom::DifferentialPolynomial(DV);
	    sys := map(tmp,DF::coerce);
	    if iszero(contains(sys,FAIL)) then
		SDAT("Polynomial", TRUE);
		SDAT("ConstCoeff", TRUE);
	    else		// try polynomial, variable coefficients
		DF := Dom::DifferentialPolynomial(DV, RDE1);
		sys := map(tmp,DF::coerce);
		if iszero(contains(sys,FAIL)) then
		    SDAT("Polynomial", TRUE);
		else		// try "quasi-polynomial"
		    DF := Dom::DifferentialPolynomial(DV, RDE2);
		    sys := map(tmp,DF::coerce);
		    if iszero(contains(sys,FAIL)) then
			SDAT("Polynomial", TRUE);
			SDAT("Quasi", TRUE);
		    else
			DF := Dom::DifferentialExpression(DV);
			sys := map(sys, DF::coerce);
			if not iszero(contains(sys, FAIL)) then
			    error("unable to create appropriate domains");
			end_if;
		    end_if;
		end_if;
	    end_if;
	end_if;
	SDAT("DVin",DV);
	SDAT("DFin",DF);
	SDAT("Expr","NoDiff");		// default is simple output
	n := 5;
    else					// differential domain
	sys := tmp;
	SDAT("DVin",DV);
	SDAT("DFin",DF);
	SDAT("Expr",FALSE);			// default is domain output
	if (DF::linear <> FAIL) and DF::linear then
	    SDAT("Linear", TRUE);
	    SDAT("ConstCoeff", DF::constCoeff);
	    SDAT("Quasi", DF::quasi);
	elif (DF::polynomial <> FAIL) and DF::polynomial then
	    SDAT("Polynomial", TRUE);
	    SDAT("ConstCoeff", DF::constCoeff);
	    SDAT("Quasi", DF::quasi);
	end_if;	    
	n := 3;
	if test then
	    if not(DF::hasProp(Cat::DifferentialFunction(DV))) or
	       {op(map(tmp,domtype))}<>{DF} then
		error("inconsistent domains for differential equations");
	    end_if;
	end_if;
    end_if;
    SDAT("Eqs",sys);
    SDAT("DVout",detools::data["DVin"]);
    SDAT("DFout",detools::data["DFin"]);
       
    // search for options and further parameters
    
    tmp := split([args(n..args(0))],((a)->bool(type(a)="_equal")));
    for opt in tmp[1] do
	sysassign(detools::data[expr2text(op(opt,1))],op(opt,2));
    end_for;
    SDAT("Parms",tmp[2]):

    NIL;    
end_proc:


unalias(SDAT):
