/*++
Cat::DifferentialFunction -- category of functions of differential variables


Domains in Cat::DifferentialFunction(DV) are a Cat::AbelianSemiGroup and
contain functions that depend on differential variables from a domain DV, an
element of Cat::DifferentialVariable.  (We cannot use Cat::Rng etc., as we
must allow for linear functions where multiplication is not possible.
Nonlinear domains should also be in Cat::PartialDifferentialRing).

The following methods must be provided by domains in this category:
	"diff" :     differentiate with respect to an element of DV or
		     with respect to a parameter
	"indets" :   return indeterminates of function as a set;
	             with option "DiffVars" only those in DV
	"subs" :     substitute an expression for an element of DV
	"solve" :    solve for an element of DV;
	             if not possible return FAIL
	"has" :      checks whether function depends on differential variable
	"Variables" :  the domain DV
and the methods of a Cat::AbelianSemiGroup.

For these methods default implementations are provided:
	"order" :      order of a function
	"class" :      class of a function
	"leader" :     leader of a function (with respect to order in DV)
	"diffSubs" :   substitute for a differential variable and all its 
		       derivatives
	"isConst" :    checks whether a function is a constant (wrt to DV)
	"multiDiff" :  computes efficiently several derivatives of the same
		       functions
	"jacobian" :   Jacobian of a list of functions
	"totalDiff" :  total derivative with respect to an independent
	               variable (can be given by its index)
	"autosimplify" :   purely algebraic simplification
	"autoreduce" : simplification including differential operations
	"eval" :     eval the differential function for a given list
		     of expression ("check a solution")
	"linear":    TRUE, if the domain represents linear functions
	"polynomial": TRUE, if the domain represents polynomial functions
Some of these are not very efficient/powerful; some are valid only under
additional assumptions.

The entry "jacobianType" returns the domain for Jacobian matrices. By default 
this is the domain Dom::SparseMatrix(DV,dom,DVgreater). However, some domains 
(e.g. for linear functions) may use other domains.


Authors: Eduard Niess, Werner M. Seiler, Uni Karlsruhe/Mannheim
         (wms@ira.uka.de, seiler@euler.math.uni-mannheim.de)
Created: ??? (1996)
Last update: 	Fri Nov 17 11:31:40 2000
++*/


alias(TYPE_ERR=error("wrong type of argument"),
      ARG_ERR=error("wrong number of arguments")):


category Cat::DifferentialFunction(DV)

    local DVcoerce:DOM_PROC, DVdiff:DOM_PROC, DVorder:DOM_PROC, 
	  DVclass:DOM_PROC, DVderiv:DOM_PROC, DVtype:DOM_PROC, 
	  DVindex:DOM_PROC, DVless:DOM_PROC, DVgreater:DOM_PROC, 
	  DVmulti:DOM_PROC, DVrepeated:DOM_PROC, DVr2m:DOM_PROC, 
	  DVnotation:DOM_PROC,
	  mm:DOM_INT, nn:DOM_INT, remDup:DOM_PROC, smerge:DOM_PROC, 
	  merge:DOM_PROC, sortinsert:DOM_PROC;

    category Cat::AbelianSemiGroup;


    /* ================= */
    /* undefined methods */
    /* ================= */

    diff;
    indets;
    solve;
    has;

    /* ============== */
    /* simple methods */
    /* ============== */

    Variables := DV;
    
    linear := FALSE;
    
    polynomial := FALSE;
        
    jacobianType := Dom::SparseMatrix(DV,dom,DVgreater);


    isConst :=
        proc(df:dom):DOM_BOOL
	begin iszero(nops(dom::indets(df,hold(DiffVars)))) end_proc;


    order :=
	proc(df:dom):DOM_INT
	local tmp:DOM_SET;
	begin
	    tmp := dom::indets(df,hold(DiffVars));
	    if tmp={} then -2 else max(op(map(tmp,DVorder))) end_if
	end_proc;


    class :=
	proc(df:dom):DOM_INT
	local tmp:DOM_SET,q:DOM_INT;
	begin
	    q := dom::order(df);
	    tmp := dom::indets(df,hold(DiffVars));
	    if tmp={} then return(0);
	    else
	        tmp := select(tmp, ((jv,q) -> bool(DVorder(jv)=q)), q);
	        max(op(map(tmp,DVclass)));
	    end_if;
	end_proc;


    leader :=
        proc(df:dom)
        local tmp:DOM_LIST, res:DV, dv:DV;
        begin
            tmp := [op(dom::indets(df,hold(DiffVars)))];
            if tmp=[] then return(FAIL)
	    else
		res := tmp[1];
		delete tmp[1];
		for dv in tmp do
		    if DVless(res,dv) then res := dv end_if;
		end_for;
		return(res);
	    end_if;
        end_proc;


    /* ============= */
    /* substitutions */
    /* ============= */
    
    // In order to facilitate the handling of multiple substitutions, here
    // in the category a method multiSubs is implemented which requires the
    // existence of a method subsOne in the domain. It is recommend that
    // in domains the following approach is taken:
    //		subs := dom::multiSubs;
    //		subsOne := ....
    // where subsOne must only handle single substitutions without any
    // argument checks. This somehwat roundabout construction is necessary,
    // as a default implementation of subs is already provide in BaseDomain
    // and this would overwrite any categorially definition of subs.
    
    // takes care of argument checks for a single equation
    // returns FAIL, if something is not correct
    // otherwise some coercions are performed
    checkEq :=
	proc(eq)
	local y, ex;
	begin
	    if type(eq)<>"_equal" then return(FAIL) end_if;
	    y := DVcoerce(op(eq,1));
	    if y=FAIL then y := op(eq,1) end_if;
	    ex := op(eq,2);
	    if iszero(ex) then ex := dom::zero
	    elif ex=1 then ex := dom::one
	    else
		ex := DVcoerce(ex);
		if ex=FAIL then
		    ex := dom::coerce(op(eq,2));
		    if ex=FAIL then
			ex := op(eq,2);
		    end_if;
		end_if;
	    end_if;
	    return(y=ex);
	end_proc;
	
	
    // this method translates multiple substitutions into a sequence of
    // single substitutions.
    multiSubs :=
	proc(de):dom
	local res, eq, eql:DOM_LIST;
	begin
	    case args(0)
	    of 0 do ARG_ERR;
	    of 1 do
		res := dom::coerce(de);
		if res=FAIL then TYPE_ERR
		else return(res) end_if;
	    of 2 do 
		case type(args(2))
		of "_equal" do 
		    eq := args(2);
		    de := dom::coerce(de);
		    if de=FAIL then TYPE_ERR end_if;
		    eq := dom::checkEq(eq);
		    if eq=FAIL then TYPE_ERR end_if;
		    return(dom::subsOne(de,eq));
		of DOM_LIST do 
		of DOM_SET do 
		of DOM_TABLE do 
		    eql := [op(args(2))];
		    res := dom::coerce(de);
		    if res=FAIL then TYPE_ERR end_if;
		    eql := map(eql,dom::checkEq);
		    if not(iszero(contains(eql,FAIL))) then 
			TYPE_ERR;
		    end_if;
		    for eq in eql do
			res := dom::subsOne(res,eq);
		    end_for;
		    return(res);
		otherwise TYPE_ERR;
		end_case;
	    otherwise 
		eql := [args(2..args(0))];
		res := dom::coerce(de);
		if res=FAIL then TYPE_ERR end_if;
		eql := map(eql,dom::checkEq);
		if not(iszero(contains(eql,FAIL))) then TYPE_ERR end_if;
		for eq in eql do
		    res := dom::subsOne(res,eq);
		end_for;
		return(res);
	    end_case;
	end_proc;


    /* ================================ */
    /* Methods based on derivative tree */
    /* ================================ */

    // computation of several derivatives of the same expression
    // input: domain element and list of multi (or repeated) indices
    // output: list of derivatives
    multiDiff :=
	proc(ex:dom,der:DOM_LIST):DOM_LIST
	local derivTree:DOM_PROC, tree:DOM_LIST;
	begin
	    // compute derivative tree
	    if DVnotation="repeated" then
		tree := detools::derList2Tree(map(der,DVr2m));
	    else
		tree := detools::derList2Tree(der);
	    end_if;
	    // recursive procedure for working down the tree
	    derivTree :=
		proc(f,T:DOM_LIST,res:DOM_LIST,DOM:DOM_DOMAIN):DOM_LIST
		local df, l, i:DOM_INT, vars:DOM_LIST;
		begin
		    df := f;
		    for l in T do
			if testtype(l,Type::ListOf(Type::NonNegInt)) then
			    // node with derivative -> compute it
			    vars := [];
			    for i from 1 to nn do
				vars := vars . [ i $ l[i] ];
			    end_for;
			    if vars<>[] then
				df := DOM::totalDiff(df,op(vars)); 
			    end_if;
			elif testtype(l,Type::NonNegInt) then
			    // leaf -> return result
			    res[l] := df;
			else
			    // branch -> recursive call
			    res := derivTree(df,l,res,DOM);
			end_if;
		    end_for;
		    return(res);
		end_proc;
	    // finally, call this procedure with the given data
	    derivTree(ex,tree,[ 0 $ nops(der) ],dom);
	end_proc;
	

    // eval enters a "solution" into a given differential expression
    // the user must take care that some reasonable type is entered
    eval :=
        proc(sys,sol)
	local DOM:DOM_DOMAIN, SDOM:DOM_DOMAIN, sysflag:DOM_BOOL,
	      j:DOM_INT, allvars:DOM_SET, vars:DOM_LIST, dersol:DOM_LIST, 
	      res, slist:DOM_LIST, mm:DOM_INT;
        begin
	    if domtype(sys)<>DOM_LIST then
		sysflag := FALSE;
		sys := [sys];
	    else sysflag := TRUE end_if;
	    mm := DV::numDepVar;
	    DOM := dom;
	    if mm=1 and domtype(sol)<>DOM_LIST then
		sol := [sol];
	    end_if;
	    SDOM := domtype(sol[1]);		// domain of "solution"
	    sys := map(sys,dom::coerce);
	    if testargs() then
		if not(iszero(contains(sys,FAIL))) then TYPE_ERR end_if;
		if nops(sol)<>mm then
		    error("wrong number of functions given");
		end_if;		
	    end_if;
	    if SDOM::hasProp=FAIL or 
	       not(SDOM::hasProp(Cat::DifferentialFunction(DV))) then
		SDOM := Dom::DifferentialExpression(
			    Dom::RestrictedDifferentialVariable(DV,
					Types="Indep"));
		sol := map(sol,SDOM::convert);
		if not(iszero(contains(sol,FAIL))) then TYPE_ERR end_if;
	    end_if;
	    
	    res := [ 0 $ nops(sys) ];
	    // get all contained derivatives
	    allvars := _union(op(map(sys,dom::indets,hold(DiffVars))));
	    allvars := select(allvars,
			      ((jv) -> contains({"Dep","Deriv"},DVtype(jv))));
	    // compute derivatives via derivative tree
	    vars := [ 0 $ mm ];
	    dersol := [ 0 $ mm ];
	    slist := [];
	    for j from 1 to mm do
		vars[j] := [op(select(allvars, 
				      ((jv,k) -> bool(DVindex(jv)=k)),
				      j))];
		if DVnotation="repeated" then
		    dersol[j] := SDOM::multiDiff(sol[j],
					     map(vars[j],DVrepeated));
		else
		    dersol[j] := SDOM::multiDiff(sol[j],
					     map(vars[j],DVmulti));
		end_if;
		slist := slist . zip(vars[j],dersol[j],_equal);
	    end_for;
	    // enter derivatives in expressions
	    res :=map(sys,dom::convert_to,DOM_EXPR);
	    res := map(res,subs,op(slist));
	    res := map(res,SDOM::convert);
	    if not(iszero(contains(res,FAIL))) then
		error("internal error");
	    end_if;
      	    if sysflag then res else res[1] end_if;
	end_proc;


    diffSubs :=
        proc(f,eq):dom
	local nof:dom, nf:dom, JVar:DOM_LIST, DerList:DOM_LIST,
	      jv, ex, i:DOM_INT, n:DOM_INT, DOM:DOM_DOMAIN;
	begin
	    // extract parts of eq
	    if args(0)<>2 then ARG_ERR end_if;
	    if domtype(f)<>dom then f := dom::coerce(f) end_if;
	    if f=FAIL or type(eq)<>"_equal" then TYPE_ERR end_if;
	    jv := DV::coerce(op(eq,1));
	    if jv=FAIL then TYPE_ERR end_if;
	    ex := dom::coerce(op(eq,2));
	    if ex=FAIL then TYPE_ERR end_if;

	    DOM := dom;
	    nof := dom::zero;
	    nf := dom::subs(f,jv=ex);
	    while nf<>nof do
		nof := nf;
		JVar := [op(select(dom::indets(nf,hold(DiffVars)),
				   ((dv,i) -> 
				    (DVtype(dv)="Deriv" and DVindex(dv)=i)), 
				   DVindex(jv)))];
		i := 0;
		n := nops(JVar);
		DerList := [ [] $ n ];
		while i<n do
		    i := i+1;
		    DerList[i] := DV::derivativeOf(JVar[i],jv);
		    if DerList[i]=[] then
			delete JVar[i];
			delete DerList[i];
			i := i-1;
			n := n-1;
		    end_if;
		end_while;
		if iszero(n) then break	end_if;
		DerList := dom::multiDiff(ex,DerList);
		nf := dom::subs(nf,op(zip(JVar,DerList,_equal)));
	    end_while;
	    
	    nf;
	end_proc;


    /* ================ */
    /* differentiations */
    /* ================ */

    // we use D as an abbreviation for totalDiff !
    D := 
	proc(l,df:dom):dom 
	begin dom::totalDiff(df,l) end_proc;
    
    
    totalDiff :=
        if dom::hasProp(Cat::CommutativeRing) then
	    proc(sys)
            local mu:DOM_LIST, i:DOM_INT, tmp, sysflag:DOM_BOOL;
            begin
		sysflag := bool(domtype(sys)=DOM_LIST);
                if not(sysflag) then sys := [sys] end_if;
                if domtype(args(2))=DOM_LIST then mu := args(2)
                elif args(0)>2 then mu := [args(2..args(0))]
                else mu := [args(2)] end_if;

                sys := map(sys,dom::coerce);
		if testargs() then
                    if not(iszero(contains(sys,FAIL))) then TYPE_ERR end_if;
                    for i from 1 to nops(mu) do
                        case domtype(mu[i])
                        of DOM_INT do
                            if mu[i]<1 or mu[i]>DV::numIndVar then 
				TYPE_ERR; 
			    end_if;
                            break;
                        of DV do
                            if DV::vartype(mu[i])<>"Indep" then 
				TYPE_ERR; 
			    end_if;
                            mu[i] := DV::index(mu[i]);
                            break;
                        otherwise
                            tmp := DVcoerce(mu[i]);
                            if tmp=FAIL or DVtype(tmp)<>"Indep" then
                                TYPE_ERR;
                            end_if;
                            mu[i] := DV::index(tmp);
                        end_case;
                    end_for;
                end_if;

                tmp := (sys,map(map(sys,dom::indets,hold(DiffVars)),
				(l -> sort([op(l)],DVgreater))));
                for i in mu do
		    if _and(op(map(op(tmp,2),((l) -> bool(l=[]))))) then
			if sysflag then return([dom::zero $ nops(sys)])
			else return(dom::zero) end_if;
		    end_if;
                    tmp := dom::totalDiffJM(op(tmp,1),i,dom::jacobian(tmp));
                end_for;
                if sysflag then op(tmp,1) else op(tmp,[1,1]) end_if;
            end_proc
	else toBeDefined end_if;


    totalDiffJM :=
        if dom::hasProp(Cat::CommutativeRing) then
            proc(sys,mu,JM)
            local n,vars,nvars,dvars,tdvars,rvars,rows,i,j,k,pos,res,t,DOM;
            begin
		DOM := dom;
                n := (JM::dom)::nrows(JM);
                vars := (JM::dom)::allIndices(JM);
                if vars=[] then
                    return(([dom::zero $ n], [[] $ n]));
                end_if;
                rows := (JM::dom)::rows(JM);
                nvars := select(vars,
                             ((jv) -> contains({"Deriv","Dep"},DVtype(jv))));
                dvars := map(nvars,DVdiff,mu);
		tdvars := map(dvars,dom::convert);
                vars := [ [] $ n ];

                for i from 1 to n do         // for each equation
		    if rows[i]=[] then
			sys[i] := dom::zero;
			next;
		    end_if;
                    rvars := map(rows[i],op,1);
                    j := nops(rvars);
                    vars[i] := [ 0 $ j ];
                    res := [ dom::zero $ j ];
                    j := 0;
		    k := 0;
                    for t in rows[i] do      // compute derivative
                        case DV::vartype(t[1])
                        of "Dep" do
                        of "Deriv" do
                            pos := contains(nvars,t[1]);
			    k := k+1;
                            res[k] := dom::_mult(t[2],tdvars[pos]);
                            j := j+1;
                            vars[i][j] := dvars[pos];
                            break;
                        of "Indep" do
                            if DV::index(t[1])=mu then
				k := k+1;
                                res[k] := t[2];
                            end_if;
                            break;
                        end_case;
                    end_for;
		    if iszero(k) then
			sys[i] := dom::zero;
			vars[i] := [];
			next;
		    end_if;
		    sys[i] := dom::_plus(op(res,1..k));
                    vars[i] := [op(vars[i],1..j)];

                    k := 1;
//                     while k<=nops(rvars) do   // check variables
//                         if iszero(contains(vars[i],rvars[k])) or
//                            not(dom::has(sys[i],rvars[k])) then k := k+1
//                         else delete rvars[k] end_if;
//                     end_while;
// this is in general too expensive!
                    vars[i] := merge(vars[i],rvars,DVgreater);
                end_for;
                (sys,vars);
            end_proc
        else NIL end_if;


    /* ======== */
    /* jacobian */
    /* ======== */

    jacobian :=
        if dom::hasProp(Cat::CommutativeRing) then
            proc(sys)
            local v, vars, varlist, i, j, f, r, rows;
            begin
		if args(0)>1 then vars := args(2) end_if;
		if domtype(sys)<>DOM_LIST then
		    sys := [sys];
		    if args(0)>1 then vars := [vars] end_if;
		end_if;
		sys := map(sys,dom::coerce);
                if testargs() then
		    if args(0)<1 or args(0)>2 then ARG_ERR end_if;
		    if not(iszero(contains(sys,FAIL))) then TYPE_ERR end_if;
		    if args(0)=2 then
			if {op(map(vars,domtype))}<>{DOM_LIST} then
			    TYPE_ERR
			end_if;
			vars := map(vars,map,DV::coerce);
			if has(vars,FAIL) then TYPE_ERR end_if;
		    end_if;
                end_if;

                if args(0)=1 then 
		    vars := map(map(sys,dom::indets,hold(DiffVars)),
				(v -> sort([op(v)],DVgreater)));
		end_if;
		varlist := [];
		for v in vars do
		    varlist := smerge(v,varlist,DVgreater);
		end_for;
                if varlist=[] then error("empty matrix") end_if;
                // should return zero matrix - but of what size?

		rows := [ [] $ nops(sys) ];
		i := 0;
                for f in sys do
		    i := i+1;
		    if vars[i]=[] then next end_if;
		    r := [ 0 $ nops(vars[i]) ];
		    j := 0;
		    for v in vars[i] do
			j := j+1;
			r[j] := [v,dom::diff(f,v)];
		    end_for;
		    rows[i] := r;
		end_for;
		(dom::jacobianType)::convert(nops(sys),varlist,rows);
            end_proc
        else toBeDefined end_if;


    /* ======================== */
    /* simplification/reduction */
    /* ======================== */

    autoreduce := 
	proc(sys)
        local new_l:DOM_LIST, notSolved_l:DOM_LIST, solved_l:DOM_LIST, eq:dom, 
	    ld:DV, k:DOM_INT, oldSolved_l:DOM_LIST, sub:DOM_LIST, neq:DOM_LIST,
	    ic:dom, tmp:DOM_LIST, d:DOM_LIST, old_l:DOM_LIST, eq2:dom,
	    DOM:DOM_DOMAIN; 
       begin
	    if args(0) <> 1 then ARG_ERR end_if;
	    if domtype(sys) <> DOM_LIST then sys := [sys] end_if;
	    sys := map(sys, dom::coerce);
	    if testargs() then
		if not iszero(contains(sys, FAIL)) then TYPE_ERR end_if;
	    end_if;
	    sys := map(sys, dom::simplify); 
	    DOM := dom;
	    if nops(sys) = 1 then return(sys) end_if; 
	    new_l := sys;
	    notSolved_l := [];
	    solved_l := [];   
	    while new_l <> [] do
		for eq in new_l do
		    ld := dom::leader(eq);
       		    if (k := dom::solve(eq, ld)) = FAIL then 
			notSolved_l := append(notSolved_l, eq)
		    else 
			solved_l := sortinsert(solved_l, [ld, k],
				((l1, l2) -> DVless(l1[1], l2[1])));
		    end_if;
		end_for;
		new_l := [];
		oldSolved_l := [];
		while solved_l <> [] do
		    sub := solved_l[1];
		    oldSolved_l := append(oldSolved_l, sub);
		    delete solved_l[1];
		    while solved_l <> [] do
			neq := solved_l[1];
			if sub[1] <> neq[1] then break end_if;
			delete solved_l[1];
			ic := neq[2] - sub[2];
			if not(dom::iszero(ic)) then
			    new_l := append(new_l, ic)
			end_if;
		    end_while;
		    tmp := [];
		    for neq in solved_l do
			d := DV::derivativeOf(neq[1], sub[1]);
			if d = [] then
			    tmp := append(tmp, [neq[1], dom::diffSubs(neq[2], 
			                  sub[1] = sub[2])]);
			else
			    ic := dom::diffSubs(
				    dom::totalDiff(sub[2], d) - neq[2], 
				    sub[1] = sub[2]);
			    if not(dom::iszero(ic)) then
				new_l := append(new_l, ic)
			    end_if;
			end_if;
		    end_for;
		    solved_l := tmp;
		end_while;
		solved_l := oldSolved_l;
		old_l := [];
		for eq in notSolved_l do
		    eq2 := eq;
		    for sub in solved_l do
			eq2 := dom::diffSubs(eq2, sub[1] = sub[2]);
		    end_for;
		    if eq2 = eq then
			old_l := append(old_l, eq)
		    else
			if not(dom::iszero(eq2)) then
			    new_l := append(new_l, eq2)
			end_if;
		    end_if;
		end_for;
		notSolved_l := old_l;
		old_l := [];
		for eq in new_l do
	       	    for sub in solved_l do
			eq := dom::diffSubs(eq, sub[1] = sub[2]);
		    end_for;
		    if not(dom::iszero(eq)) then
			old_l := append(old_l, eq);
		    end_if;
		end_for;
		new_l := old_l;
	    end_while;
	    notSolved_l . map(solved_l, ((l) -> (DOM::coerce(l[1]) - l[2])));
	end_proc;


    autosimplify := 
	proc(sys)
        local new_l:DOM_LIST, notSolved_l:DOM_LIST, solved_l:DOM_LIST, eq:dom, 
	    ld:DV, k:DOM_INT, oldSolved_l:DOM_LIST, sub:DOM_LIST, neq:DOM_LIST,
	    ic:dom, tmp:DOM_LIST, old_l:DOM_LIST, eq2:dom, DOM:DOM_DOMAIN;
        begin
	    if args(0) <> 1 then ARG_ERR end_if;
	    if domtype(sys) <> DOM_LIST then sys := [sys] end_if;
	    sys := map(sys, dom::coerce);
	    if testargs() then
		if not iszero(contains(sys, FAIL)) then TYPE_ERR end_if;
	    end_if;
	    sys := map(sys, dom::simplify); 
	    DOM := dom;
	    if nops(sys) = 1 then return(sys) end_if; 
	    new_l := sys;
	    notSolved_l := [];
	    solved_l := [];   
	    while new_l <> [] do
		for eq in new_l do
		    ld := dom::leader(eq);
		    if (k := dom::solve(eq, ld)) = FAIL then
			notSolved_l := append(notSolved_l, eq)
		    else
			solved_l := sortinsert([ld, k], solved_l,  
				((l1, l2) -> DVless(l1[1], l2[1])));
		    end_if;
		end_for;
		new_l := [];
		oldSolved_l := [];
		while solved_l <> [] do
		    sub := solved_l[1];
		    oldSolved_l := append(oldSolved_l, sub);
		    delete solved_l[1];
		    while solved_l <> [] do
			neq := solved_l[1];
			if sub[1] <> neq[1] then break end_if;
			delete solved_l[1];
			ic := neq[2] - sub[2];
			if not(dom::iszero(ic)) then
			    new_l := append(new_l, ic)
			end_if;
		    end_while;
		    tmp := [];
		    for neq in solved_l do
			tmp := append(tmp, [neq[1], dom::subs(neq[2], 
			                  sub[1] = sub[2])]);
		    end_for;
		    solved_l := tmp;
		end_while;
		solved_l := oldSolved_l;
		old_l := [];
		for eq in notSolved_l do
		    eq2 := eq;
		    for sub in solved_l do
			eq2 := dom::subs(eq2, sub[1] = sub[1]);
		    end_for;
		    if eq2 = eq then
			old_l := append(old_l, eq)
		    else
			if not(dom::iszero(eq2)) then
			    new_l := append(new_l, eq2)
			end_if;
		    end_if;
		end_for;
		notSolved_l := old_l;
		old_l := [];
		for eq in new_l do
		    for sub in solved_l do
			eq := dom::subs(eq, sub[1] = sub[2]);
		    end_for;
		    if not(dom::iszero(eq)) then
			old_l := append(old_l, eq2);
		    end_if;
		end_for;
		new_l := old_l;
	    end_while;
	    notSolved_l . map(solved_l, ((l) -> (DOM::coerce(l[1])) - l[2]));
	end_proc;

        
begin

    if args(0)<>1 then ARG_ERR end_if;
    if (domtype(DV)<>DOM_DOMAIN) or (DV::hasProp=FAIL) or
       not(DV::hasProp(Cat::DifferentialVariable)) then
	error("arg must be in Cat::DifferentialVariable");
    end_if;
  
    DVcoerce := DV::coerce;
    DVdiff := DV::totalDiff;
    DVorder := DV::order;
    DVclass := DV::class;
    DVindex := DV::index;
    DVderiv := DV::derivativeOf;
    DVtype := DV::vartype;
    DVless := DV::_less;
    DVgreater := (dv1,dv2) -> DVless(dv2,dv1);
    DVmulti := DV::multiIndex;
    DVrepeated := DV::repeatedIndex;
    DVr2m := DV::r2m;
    DVnotation := DV::notation;
    
    nn := DV::numIndVar;
    mm := DV::numDepVar;
    smerge := listlib::singleMerge;
    merge := listlib::merge;
    sortinsert := listlib::insert;
    remDup := listlib::removeDuplicates;
    
end_category:

unalias(TYPE_ERR,ARG_ERR,DV):

