/*++
Cat::DifferentialVariable -- category of differential variables

Its purpose is to provide basic data structures and procedures for
differential variables. In geometric approaches one may consider it as an
implementation of adapted charts in jet bundles.  Every instance of this
category has a fixed number of independent and dependent variables which can
be obtained with the methods "numIndVar" and "numDepVar", respectively.

Differential variables are classified into three types: "Indep" are the
independent, "Dep" the dependent variables, "Deriv" derivatives.  The method
"varype" returns this type as a string.

For derivatives two kinds of notations are supported: multi index and repeated
index notation. The method "notation" returns a string "multi" or "repeated"
indicating the used notation for a specific domain. The methods "multiIndex"
and "repeatedIndex" return the indices in the respective notation. In the case
of repeated indices, it is always assumed that a sorted list (lowest index
first) is returned. For conversion, two auxiliary methods "m2r" and "r2m" are
provided. Internally, the independent and dependent variables are always
indexed. The method "index" returns this index. For derivatives it returns the
index of the corresponding dependent variable.

Each domain should provide two methods "newMulti" and "newRepeated" to
generate differential variables.  They take as first argument the type of the
variable, as second argument its index and, for derivatives, as third argument
its multi or repeated index, respectively.  These two methods are mainly
thought for the use in programs and not for interactive use.  Therefore they
should be fast!  The category provides a default implementation which assumes
that convert can handle such input.

Cat::DifferentialVariable is a subcategory of Cat::OrderedSet. Two rankings
are already implemented here:
        "DegreeOrder" -- derivatives are first ordered according to their
                         order then lexicographically. 
        "LexOrder"    -- derivatives are ordered purely lexicographically.
They are available via the methods orderLex and pureLex, resp. By default
_less is orderLex. The entry orderly return TRUE or FALSE depending on 
whether or not the domain uses an orderly ranking.

The following methods have to been defined in domains of this category
	vartype :   type of a differential variable
	index :     index of a differential variable
	numIndVar : number of independent variables
	numDepVar : number of dependent variables
	notation :  used notation (string "multi" or "repeated")
The last three are not functions but entries and thus should yield directly
the corresponding constants. In addition one of following two methods must be
defined (depending on the used notation):
	repeatedIndex -- repeated index of a derivative
			 (if dom::notation="repeated")
	multiIndex    -- multi index of a derivative
			 (if dom::notation="multi")

Default implementations are provided for:
        newRepeated :   generation of differential variables given
                        the repeated index
        newMulti :      generation of differential variables given
                        the multi index
	m2r :           converts a multi index into a repeated index
	r2m :           converts a repeated index into a multi index
	multiRaise :    increase a multi index
	multiRepeated : increase a repeated index
	allRepeated :   list of all repeated indices corresponding to a
	                given multi index
	class :         class of a derivative
	order :         order of a derivative
	dim :           number of differential variables up to a given order
	dimq :          number of differential variables of a given order
	diff :          differentiates a differential variable wrt to some
			variable (treats all variables as ordinary variables)
	totalDiff :     differentiates a differential variable wrt to an
			independent variables ("true" differentiation)
	derivativeOf :  checks whether a differential variable is the
	                the derivative of another one
	int :     	integrates a differential variable
	indVariables :	list of all independent variables
	depVariables :	list of all dependent variables
	derivatives :   list all derivatives up to a given order
	has :		overloads system function `has'
	length :        overloads system function `length'


Authors: Eduard Niess, Werner M. Seiler, Uni Karlsruhe/Mannheim
	 (wms@ira.uka.de, seiler@euler.math.uni-mannheim.de)
Created: ??? (1996)
Last update: 	Fri Nov 17 11:22:05 2000
++*/


alias(TYPE_ERR=error("wrong type of argument"),
      ARG_ERR=error("wrong number of arguments"),
      IND_ERR=error("illegal index"),
      VAR_ERR=error("unknown type of differential variable (internal bug?)")):


category Cat::DifferentialVariable

    category Cat::OrderedSet;


    /* ================= */
    /* undefined methods */
    /* ================= */

    vartype;
    index;
    numIndVar;
    numDepVar;
    notation;


    repeatedIndex :=
        if dom::notation="multi" then
            proc(jv:dom):DOM_LIST
            begin dom::m2r(dom::multiIndex(jv)) end_proc;
        else toBeDefined end_if;


    multiIndex :=
        if dom::notation="repeated" then
            proc(jv:dom):DOM_LIST
            begin dom::r2m(dom::repeatedIndex(jv)) end_proc;
        else toBeDefined end_if;


    /* ================= */
    /* auxiliary methods */
    /* ================= */

    multiRaise :=
        proc(mu:DOM_LIST,l:(Type::Union(DOM_INT,DOM_LIST))):DOM_LIST
        local i:DOM_INT;
	begin
            case domtype(l)
            of DOM_INT do
	        if mu=[] then mu := [ 0 $ dom::numIndVar ] end_if;
	        return(subsop(mu, l = mu[l]+1));
   	    of DOM_LIST do
	        if mu=[] then mu := [ 0 $ dom::numIndVar ] end_if;
		for i in l do
		    mu[i] := mu[i]+1;
		end_for;
		return(mu);
	    otherwise TYPE_ERR;
	    end_case;
        end_proc;
	
	
    repeatedRaise :=
	proc(ri:DOM_LIST,l:(Type::Union(DOM_INT,DOM_LIST))):DOM_LIST
	begin
	    case domtype(l)
	    of DOM_INT do return(listlib::insert(ri,l));
	    of DOM_LIST do 
		l := sort(l);
		return(listlib::merge(ri,l));
	    otherwise TYPE_ERR;
	    end_case;
	end_proc;


    m2r :=
        proc(l:DOM_LIST):DOM_LIST
        local i:DOM_INT;
        begin
            [ (i $ l[i]) $ i=1..dom::numIndVar ];
        end_proc;


    r2m :=
        case dom::numIndVar
        of 1 do
            proc(l:DOM_LIST):DOM_LIST
            begin [nops(l)] end_proc;
            break;
        of 2 do
            proc(l:DOM_LIST):DOM_LIST
	    local n:DOM_INT,pos:DOM_INT;
            begin
                n := nops(l);
                pos := contains(l,2);
                case pos
                of 0 do return([n,0]);
                of 1 do return([0,n]);
                otherwise
                    pos := pos-1;
                    return([pos,n-pos]);
		end_case;
            end_proc;
            break;
        otherwise
            proc(l:DOM_LIST):DOM_LIST
            local i:DOM_INT,res:DOM_LIST;
            begin
                res := [ 0 $ dom::numIndVar ];
	        for i in l do
	           res[i] := res[i]+1;
	        end_for;
	        res;
            end_proc;
        end_case;


    // returns a list of all possible realizations of a given multi-index
    // as repeated index.
    allRepeated :=
        proc(l:DOM_LIST):DOM_LIST
        local i:DOM_INT,ll:DOM_LIST,res:DOM_LIST;
        begin
	    res := [];
	    for i from 1 to dom::numIndVar do
	        if not(iszero(l[i])) then
                    ll := l;
		    ll[i] := ll[i]-1;
		    res := res . map(dom::allRepeated(ll),
		                     ((l,j) -> ([j].l)), i);
	        end_if;
	    end_for;
	    if res=[] then [[]] else res end_if;
        end_proc;


    newMulti :=
        if dom::notation="multi" then dom::convert
        else
            proc(t:DOM_STRING,i:DOM_INT,mu:DOM_LIST)
            begin dom::convert(t,i,dom::m2r(mu)) end_proc;
        end_if;


    newRepeated :=
        if dom::notation="repeated" then dom::convert
        else
            proc(t:DOM_STRING,i:DOM_INT,mu:DOM_LIST)
            begin dom::convert(t,i,dom::r2m(mu)) end_proc;
        end_if;


    /* ================ */
    /* "simple" methods */
    /* ================ */

    // we define the class of a dependent variables as n !!!
    class :=
	if dom::notation="multi" then
	    proc(jv:dom):DOM_INT
	    local j:DOM_INT, mind:DOM_LIST;
	    begin
		case dom::vartype(jv)
		of "Indep" do return(0);
		of "Dep"   do return(dom::numIndVar);
		of "Deriv" do
		    mind := dom::multiIndex(jv);
		    j := 1;
		    while (j<dom::numIndVar) and iszero(mind[j]) do
			j := j+1;
		    end_while;
		    return(j);
		otherwise VAR_ERR;
		end_case;
	    end_proc;
	else
	    proc(jv:dom):DOM_INT
	    begin
		case dom::vartype(jv)
		of "Indep" do return(0);
		of "Dep"   do return(dom::numIndVar);
		of "Deriv" do return(dom::repeatedIndex(jv)[1]);
      		otherwise VAR_ERR;
		end_case;
	    end_proc;
	end_if;


    // we define the order not only for derivatives but use it to
    // distinguish between the different type of differential variables:
    // independent variables have order -1
    order :=
        if dom::notation="multi" then
	    proc(jv:dom):DOM_INT
	    begin
		case dom::vartype(jv)
		of "Indep" do return(-1);
		of "Dep"   do return(0);
		of "Deriv" do return(_plus(op(dom::multiIndex(jv))));
		otherwise VAR_ERR;
		end_case;
	    end_proc
	else
	    proc(jv:dom):DOM_INT
	    begin
		case dom::vartype(jv)
		of "Indep" do return(-1);
		of "Dep"   do return(0);
		of "Deriv" do return(nops(dom::repeatedIndex(jv)));
		otherwise VAR_ERR;
		end_case;
	    end_proc
	end_if;


    length :=
	proc(jv:dom):DOM_INT
	begin 1 end_proc;


    dim :=
        proc(q:DOM_INT):DOM_INT
	option remember;
	begin
	    if q=0 then dom::numDepVar
	    elif q<0 then 0
	    else dom::dim(q-1)+dom::dimq(q) end_if;
	end_proc;


    dimq :=
        proc(q:DOM_INT):DOM_INT
	option remember;
	begin dom::numDepVar * binomial(q-1+dom::numIndVar, q) end_proc;


    /* =============== */
    /* differentiation */
    /* =============== */

    // this is more or less a hack needed for domains like
    // Dom::DifferentialExpression in order to provide an
    // interface for the system function "diff"; differential
    // variables are treated like ordinary variables
    diff :=
        proc()
        local tmp, dt:DOM_DOMAIN, dummy:DOM_IDENT;
        begin
            case args(0)
	    of 0 do ARG_ERR;
	    of 1 do return(args(1));
	    otherwise
		dt := domtype(args(1));
		if dt=dom then
		    if (args(0)>2) or (args(1)<>args(2)) then return(0)
		    else return(1) end_if;
		elif (dt::hasProp=FAIL) or 
		     not(dt::hasProp(Cat::DifferentialFunction(dom))) then
		    dummy := genident();
		    tmp := subs(args(1),args(2)=dummy);
		    tmp := subs(diff(tmp,dummy),dummy=args(2));
		    if args(0)=2 then return(tmp)
		    else return(diff(tmp,args(3..args(0)))) end_if;
		else
		    return(dt::diff(args()));
		end_if;
            end_case;
        end_proc;


    // the "real" differentiation
    totalDiff :=
        proc(jv):dom
	local mu,tmp,i;
	begin
	    if args(0)>2 then mu := [args(2..args(0))]
	    else mu := args(2) end_if;
	    jv := dom::coerce(jv);
	    if testargs() then
		if jv=FAIL then TYPE_ERR end_if;
		if dom::vartype(jv)="Indep" then
		    error("not defined for independent variables");
		end_if;
	    end_if;
	    if domtype(mu)=DOM_LIST then
		for i from 1 to nops(mu) do
		    case domtype(mu[i])
		    of DOM_INT do
		        if testargs() and
		           (mu[i]<1 or mu[i]>dom::numIndVar) then TYPE_ERR
		        end_if;
		        break;
		    of dom do
			if testargs() and
			   dom::vartype(mu[i])<>"Indep" then TYPE_ERR end_if;
			mu[i] := dom::index(mu[i]);
		        break;
		    otherwise
			tmp := dom::coerce(mu[i]);
			if testargs() then
		            if tmp=FAIL or
			       dom::vartype(tmp)<>"Indep" then TYPE_ERR end_if;
		        end_if;
		        mu[i] := dom::index(tmp);
		    end_case;
		end_for;
	    else
		case domtype(mu)
		of DOM_INT do
		    if testargs() and
		       (mu<1 or mu>dom::numIndVar) then TYPE_ERR end_if;
		    break;
		of dom do
		    if testargs() and
		       dom::vartype(mu)<>"Indep" then TYPE_ERR end_if;
		    mu := dom::index(mu);
		    break;
		otherwise
		    tmp := dom::coerce(mu);
		    if testargs() then		        
		        if tmp=FAIL or
		           dom::vartype(tmp)<>"Indep" then TYPE_ERR end_if;
		    end_if;
		    mu := dom::index(tmp);
		end_case;
	    end_if;

            if domtype(mu)=DOM_LIST then dom::totalDiffN(jv,mu)
	    else dom::totalDiff2(jv,mu) end_if;
	end_proc;


    totalDiff2 :=
        if dom::notation="multi" then
	    proc(jv:dom,i:DOM_INT):dom
	    local tmp:DOM_LIST;
	    begin
	        case dom::vartype(jv)
	    	of "Dep" do
	            return(dom::newRepeated("Deriv",dom::index(jv),[i]));
	    	of "Deriv" do
	            tmp := dom::multiRaise(dom::multiIndex(jv),i);
	            return(dom::newMulti("Deriv",dom::index(jv),tmp));
		otherwise error("not defined");
	    	end_case;
	    end_proc
	else
	    proc(jv:dom,i:DOM_INT):dom
	    local tmp:DOM_LIST;
	    begin
	        case dom::vartype(jv)
	    	of "Dep" do
	            return(dom::newRepeated("Deriv",dom::index(jv),[i]));
	    	of "Deriv" do
	            tmp := listlib::insert(dom::repeatedIndex(jv),i);
	            return(dom::newRepeated("Deriv",dom::index(jv),tmp));
		otherwise error("not defined");
	    	end_case;
	    end_proc
	end_if;


    totalDiffN :=
        if dom::notation="multi" then
	    proc(jv:dom,l:DOM_LIST):dom
	    local tmp:DOM_LIST;
	    begin
	        case dom::vartype(jv)
	        of "Dep" do
	            return(dom::newMulti("Deriv",dom::index(jv),dom::r2m(l)));
		of "Deriv" do
		    tmp := dom::multiRaise(dom::multiIndex(jv),l);
		    return(dom::newMulti("Deriv",dom::index(jv),tmp));
		otherwise error("not defined");
	        end_case;
	    end_proc
	else
	    proc(jv:dom,l:DOM_LIST):dom
	    local tmp:DOM_LIST;
	    begin
	        case dom::vartype(jv)
	        of "Dep" do
	            return(dom::newRepeated("Deriv",dom::index(jv),sort(l)));
	        of "Deriv" do
	            tmp := listlib::merge(sort(l),dom::repeatedIndex(jv));
	            return(dom::newRepeated("Deriv",dom::index(jv),tmp));
		otherwise error("not defined");
	        end_case;
	    end_proc
	end_if;


    derivativeOf :=
        proc(j1,j2):DOM_LIST
	local jt:DOM_STRING,res:DOM_LIST,mi1:DOM_LIST,mi2:DOM_LIST,i:DOM_INT;
	begin
	    j1 := dom::coerce(j1);
	    j2 := dom::coerce(j2);
	    if testargs() then
		if (j1 = FAIL or j2 = FAIL) then
		    TYPE_ERR;
		end_if;
	    end_if;

	    if dom::vartype(j1)<>"Deriv" then return([]) end_if;
	    jt := dom::vartype(j2);
	    if jt<>"Deriv" and jt<>"Dep" then return([]) end_if;
	    if dom::index(j1)<>dom::index(j2) then return([]) end_if;

	    res := [ 0 $ dom::numIndVar ];
	    mi1 := dom::multiIndex(j1);
	    mi2 := dom::multiIndex(j2);
	    for i from 1 to dom::numIndVar do
		res[i] := mi1[i]-mi2[i];
		if res[i]<0 then return([]) end_if;
	    end_for;

	    if dom::notation="repeated" then return(dom::m2r(res)) end_if;
	    res;
	end_proc;


    int :=
        proc(jv,i)
	local nind:DOM_LIST,pos:DOM_INT,isderiv:DOM_BOOL,j:DOM_INT;
	begin
	    if domtype(i)=dom and dom::vartype(i)="Indep" then
	        i := dom::index(i);
	    end_if;
	    jv := dom::coerce(jv);
	    if testargs() then
		if jv=FAIL then TYPE_ERR end_if;
	    end_if;
	    if domtype(i)<>DOM_INT then
		i := dom::coerce(i);
		if i=FAIL or dom::vartype(i)<>"Indep" then TYPE_ERR end_if;
		i := dom::index(i);
	    end_if;
	    if (i<1) or (i>dom::numIndVar) then IND_ERR end_if;
	    if dom::vartype(jv)<>"Deriv" then return(FAIL) end_if;

	    if dom::notation="multi" then
		nind := dom::multiIndex(jv);
		if iszero(nind[i]) then	return(FAIL) end_if;
		nind[i] := nind[i]-1;
		isderiv := FALSE;
		for j in nind do
		    if not(iszero(j)) then
			isderiv := TRUE;
			break;
		    end_if;
		end_for;
		if isderiv then
		    return(dom::newMulti("Deriv",dom::index(jv),nind));
		else
		    return(dom::newMulti("Dep",dom::index(jv),[]));
		end_if;
	    else
		nind := dom::repeatedIndex(jv);
		pos := contains(nind,i);
		if iszero(pos) then return(FAIL) end_if;
		delete nind[pos];
		if nind = [] then
		    return(dom::newRepeated("Dep",dom::index(jv),[]));
		else
		    return(dom::newRepeated("Deriv",dom::index(jv),nind));
		end_if;
	    end_if;
	end_proc;


    /* ========= */
    /* variables */
    /* ========= */

    indVariables := [ dom::newRepeated("Indep",j) $ j=1..dom::numIndVar ];


    depVariables := [ dom::newRepeated("Dep",j) $ j=1..dom::numDepVar ];


    // generates all possible multi indices up to order q
    makeIndexList :=
	proc(q:DOM_INT):DOM_LIST
	local ilist:DOM_LIST,nlist:DOM_LIST,j:DOM_INT,k:DOM_INT;
	begin
	    nlist := [];
	    for j from 1 to q do
		ilist := [[]];
		for k from 1 to (dom::numIndVar-1) do
		    ilist := map(ilist,
				 ((l,j) -> (l.[i] $ i=0..(j-_plus(op(l))))),j);
		end_for;
		ilist := map(ilist, ((l,j) -> (l.[j-_plus(op(l))])), j);
		nlist := nlist.ilist;
	    end_for;
	    return(nlist);
	end_proc;
	
	
    // if cl=0 all derivatives UP to order q are computed;
    // if cl>0 all derivatives OF order q and class>=cl are computed
    derivatives := proc(q:DOM_INT, cl = 0:DOM_INT):DOM_LIST
	local dmtoncR:DOM_PROC, DOM:dom, a;
	begin
	    dmtoncR := proc(m, n)
		option remember;
		local i:DOM_INT, k;
		begin
		if n = 1 then
		    return([[m]]);
		else
		    return([(([i] . k) $ k in dmtoncR(m - i, n - 1)) 
			$ i = 0..m]);
		end_if;
	    end_proc;
	    DOM := dom;
	    if cl = 0 then
		return([op(map(a, (l -> DOM::newMulti("Deriv", i, l) 
		    $ i = 1..DOM::numDepVar))) 
			$ a in [dmtoncR(q - j, dom::numIndVar) 
				$ j = 0..(q - 1)]]);
	    else
		return([op(map(map(dmtoncR(q, dom::numIndVar - cl + 1), 
		    (r -> [0 $ cl - 1] . r)), 
			(l -> DOM::newMulti("Deriv", i, l)
			    $ i = 1..DOM::numDepVar)))]);
	    end_if;
	end_proc;
	
	
    /* ======== */
    /* ordering */
    /* ======== */

    multiLess :=
        if dom::notation="multi" then
            proc(j1:dom,j2:dom):DOM_INT
            local l1:DOM_LIST,l2:DOM_LIST;
            begin
                l1 := dom::multiIndex(j1);
                l2 := dom::multiIndex(j2);
                while (l1<>[]) and (l1[1]=l2[1]) do
                    delete l1[1];
                    delete l2[1];
                end_while;
	        if l1=[] then 0
	        elif l1[1]>l2[1] then 1 else -1 end_if;
	    end_proc;
	else
            proc(j1:dom,j2:dom):DOM_INT
            local l1:DOM_LIST,l2:DOM_LIST;
            begin
                l1 := dom::repeatedIndex(j1);
                l2 := dom::repeatedIndex(j2);
                while (l1<>[]) and (l2<>[]) and (l1[1]=l2[1]) do
                    delete l1[1];
                    delete l2[1];
                end_while;
	        if l1=[] then
	            if l2=[] then 0 else 1 end_if;
		elif l2=[] then -1
	        elif l1[1]<l2[1] then 1 else -1 end_if;
	    end_proc;
	end_if;


    orderLex :=
        proc(j1:dom,j2:dom):DOM_BOOL
        local o1:DOM_INT,o2:DOM_INT;
        begin
	    case dom::vartype(j2)
	    of "Indep" do
		if dom::vartype(j1)="Indep" then 
		    return(bool(dom::index(j1)<dom::index(j2)));
		else return(FALSE) end_if;
	    of "Dep" do
		case dom::vartype(j1)
		of "Indep" do return(TRUE);
		of "Dep"   do return(bool(dom::index(j1)<dom::index(j2)));
		otherwise return(FALSE);
		end_case;
	    of "Deriv" do
		if dom::vartype(j1)<>"Deriv" then return(TRUE) end_if;
		o1 := dom::order(j1);
		o2 := dom::order(j2);
		if o1=o2 then
		    case dom::multiLess(j1,j2)
		    of 1  do return(TRUE);
		    of -1 do return(FALSE);
		    of 0  do return(bool(dom::index(j1)<dom::index(j2)));
		    end_case;
		else return(bool(o1<o2)) end_if;
	    otherwise VAR_ERR;
	    end_case
	end_proc;


    pureLex :=
        proc(j1:dom,j2:dom):DOM_BOOL
        begin
	    case dom::vartype(j2)
	    of "Indep" do
		if dom::vartype(j1)="Indep" then
		    return(bool(dom::index(j1)<dom::index(j2)));
		else return(FALSE) end_if;
	    of "Dep" do
		case dom::vartype(j1)
		of "Indep" do return(TRUE);
		of "Dep"   do return(bool(dom::index(j1)<dom::index(j2)));
		otherwise return(FALSE);
		end_case;
	    of "Deriv" do
		case dom::multiLess(j1,j2)
		of 1  do return(TRUE);
		of -1 do return(FALSE);
		of 0  do return(bool(dom::index(j1)<dom::index(j2)));
		end_case;
	    otherwise VAR_ERR;
	    end_case;
	end_proc;


    _less := dom::orderLex;


    orderly := TRUE;
    

    /* ============= */
    /* miscellaneous */
    /* ============= */

    has := 
	proc(x:dom,y):DOM_BOOL
	begin bool(x=y) end_proc;


    indets :=
	proc(x:dom):DOM_SET
	begin {x} end_proc;
	
	
    random :=
        proc(q=6:DOM_INT):dom
	local r:DOM_PROC, ri:DOM_PROC, rd:DOM_PROC, i:DOM_INT;
	begin
	    r := random(1..q);
	    case r()
	    of 1 do
		ri := random(1..dom::numIndVar);
		return(dom::newRepeated("Indep",ri()));
	    of 2 do
	    of 3 do
		rd := random(1..dom::numDepVar);
		return(dom::newRepeated("Dep",rd()));
	    of 4 do
	    of 5 do
	    of 6 do
		ri := random(1..dom::numIndVar);
		rd := random(1..dom::numDepVar);
		return(dom::newRepeated("Deriv", rd(), 
					sort([ri() $ i=1..r()])));
	    end_case;
	end_proc;


begin

    if args(0)<>0 then ARG_ERR end_if;
    
end_category:


unalias(TYPE_ERR,ARG_ERR,IND_ERR,VAR_ERR):
