/*++
Cat::FiniteMonoidRing -- category of free algebras over an ordered monoid
	with finite support

A Cat::FiniteMonoidRing(R,S) is a Cat::MonoidRing(R,S) with the additional
property that each element is only of finite support. It defines the
following additional methods:
   mapcoeffs(x,f) :   applies the function f to every coefficient in x,
                      thus f must map elements of R to elements of R
   terms(x) :         a (sorted) list of all terms of x with non-vanishing
                      coefficient
   coeff(x) :         a list of all coefficients (order matching that of
		      terms(x), i.e. the first entry is the coefficient of
		      the first entry of support(x) etc.)
   coeff(x,s) :       the coefficient of the generator s
   nterms(x) :        the number of monomials in x
   nthmonomial(x,n) : the n-th monomial of x
   nthterm(x,n) :     the n-th term of x
   nthcoeff(x,n) :    the n-th coefficient of x

Default implementations are provided for lcoeff, lterm, isMonomial, nthmonomial,
nthterm, nthcoeff, mapcoeffs, degree and nterms. Some of these are not very
efficient!

Author: Werner M. Seiler, Uni Karlsruhe/Mannheim
        (wms@ira.uka.de, seiler@euler.math.uni-mannheim.de)
Created: July 3rd, 1997
Last updated: August 5th, 1998
++*/

alias(IND_ERR=error("invalid index")):


category Cat::FiniteMonoidRing(R,S)

    local Rzero:R, Riszero:DOM_PROC, Sone, Sisone:DOM_PROC, Sdegree;
    category Cat::MonoidRing(R,S);


    // ====================================================== 
    // implementations for methods defined in Cat::MonoidRing 
    // ====================================================== 

    lcoeff :=
        proc(x:dom):R
        begin
            if dom::iszero(x) then Rzero
            else dom::coeffs(x)[1] end_if;
        end_proc;


    lterm :=
        proc(x:dom):S
        begin
            if dom::iszero(x) then Sone
            else dom::support(x)[1] end_if;
        end_proc;


    isMonomial :=
        proc(x:dom):DOM_BOOL
        begin bool(dom::nterms(x)=1) end_proc;


    degree :=
        if Sdegree<>FAIL then
	    proc(x:dom):DOM_INT
	    begin max(op(map(dom::terms(x),Sdegree))) end_proc;
        else NIL end_if;


    _divide :=
        if R::hasProp(Cat::Field) then
	    proc(x:dom,r:R):dom
	    begin dom::mapcoeffs(x,R::_divide,r) end_proc;
	else NIL end_if;


    // ================== 
    // additional methods 
    // ================== 

    terms;

    nterms :=
        proc(x:dom):DOM_INT
        begin nops(dom::terms(x)) end_proc;


    nthterm :=
        proc(x:dom,n:DOM_INT):S
        begin
            if testargs() then
                if (n<1) or (dom::nterms(x)<n) then IND_ERR end_if;
            end_if;
            dom::terms(x)[n];
        end_proc;


    nthcoeff :=
        proc(x:dom,n:DOM_INT):R
        begin
            if testargs() then
                if (n<1) or (dom::nterms(x)<n) then IND_ERR end_if;
            end_if;
            dom::coeff(x)[n];
        end_proc;


    nthmonomial :=
        proc(x:dom,n:DOM_INT):dom
        begin
            if testargs() then
                if (n<1) or (dom::nterms(x)<n) then IND_ERR end_if;
            end_if;
            dom::monomial(dom::nthcoeff(x,n),dom::nthterm(x,n));
        end_proc;


    mapcoeffs :=
        proc(x:dom,f:DOM_PROC):dom
	local rargs,r;
	begin
            rargs := args(3..args(0));
	    r := f(dom::lcoeff(x),rargs);
	    if Riszero(r) then
		 return(dom::mapcoeffs(dom::reductum(x),f,rargs));
	    end_if;
	    dom::_plus(dom::monomial(r,dom::lterm(x)),
		       dom::mapcoeffs(dom::reductum(x),f,rargs));
	end_proc;


begin

    if args(0)<>2 then error("wrong no of args") end_if;
    if (domtype(R)<>DOM_DOMAIN) or (domtype(S)<>DOM_DOMAIN) or
       (R::hasProp=FAIL) or (S::hasProp=FAIL) then
        error("arguments must be domains");
    end_if;
    if R::hasProp(Cat::Ring) then
	Rzero := R::zero;
	Riszero := R::iszero;
    else error("first argument must belong to Cat::Ring") end_if;
    if S::hasProp(Cat::OrderedMonoid) then
	Sone := S::one;
	Sisone := S::isone;
    elif S::hasProp(Cat::OrderedAbelianMonoid) then
	Sone := S::zero;
	Sisone := S::iszero;
    else error("second argument must be an ordered monoid") end_if;
    Sdegree := S::degree;

end_category:


unalias(IND_ERR);
