/*++
Dom::DifferentialExpression(DV) -- Domain of general expressions in 
				   differential variables

Dom::DifferentialExpression(DV) is an instance of Cat::DifferentialFunction.
It implements general expressions in the differential variables from the domain
DV (an element of Cat::DifferentialVariable). The implementation is based
on DOM_EXPR.


Author: Werner M. Seiler, Uni Karlsruhe/Mannheim
        (wms@ira.uka.de, seiler@euler.math.uni-mannheim.de)
Created: September 30th, 1997
Last update: 	Fri Nov 17 13:26:16 2000
++*/

alias(REP(x)=extop(x,1),
      VARS(x)=extop(x,2),
      TYPE_ERR=error("wrong type of argument"),
      ARG_ERR=error("wrong number of arguments")):


domain Dom::DifferentialExpression()

    local DVcoerce:DOM_PROC, DVconvto:DOM_PROC, DVexpr:DOM_PROC,
          DVdiff:DOM_PROC, DVindex:DOM_PROC, DVtype:DOM_PROC,
          DVgreater:DOM_PROC, DVrandom:DOM_PROC, 
	  DVnewRep:DOM_PROC, DVrepIndex:DOM_PROC, DVmulti:DOM_PROC, 
	  nn:DOM_INT, mm:DOM_INT, insertord:DOM_PROC, singleMerge:DOM_PROC, 
	  merge:DOM_PROC, remDup:DOM_PROC, numTypes:DOM_SET, opts:DOM_LIST,
	  dummyVar:DOM_IDENT, dummyList:DOM_LIST, dummyN:DOM_INT,
	  DV:DOM_DOMAIN, DF:DOM_DOMAIN, DFvars:DOM_PROC, DFexpr:DOM_PROC,
	  DFdiff:DOM_PROC, DFdiffN:DOM_PROC, DFconvto:DOM_PROC, 
	  DFsubs:DOM_PROC, DFeval:DOM_PROC, DV2:DOM_DOMAIN, DF2:DOM_DOMAIN,
	  DF2convert:DOM_PROC, DVtest:DOM_PROC, restrictedDV:DOM_BOOL,
	  DVconv:DOM_PROC;
	  
    inherits Dom::BaseDomain;
    category Cat::DifferentialFunction(DV), Cat::PartialDifferentialRing,
	     Cat::Field;


    /* ========================= */
    /* representation, in/output */
    /* ========================= */

    // Most elements are represented by an element of DOM_EXPR and a set
    // of the differential variables appearing in it. The main exceptions
    // are differential variables and arbitrary functions which are represented
    // by elements of DV and DF, resp., in the first slot. The set in the
    // second slot may contain differential variables which do not explicitly
    // appear in the expression; however, no variable not contained in the
    // set can appear.
      
    generate :=
	proc(x,vars:DOM_SET):dom
	local v:DV;
	begin
	    case domtype(x)
	    of DV do return(new(dom,x,{x}));
	    of DF do return(new(dom,x,DFvars(x)));
	    of DOM_EXPR do
		for v in vars do
		    if not(has(x,v)) then
			vars := vars minus {v};
		    end_if;
		end_for;
		return(new(dom,x,vars));
	    otherwise return(new(dom,x,{}));
	    end_case;
	end_proc;
	

    convert := if restrictedDV then
	proc(x)
	begin
	    x := DF2convert(x);
	    if (x=FAIL) or contains(map(VARS(x), DVtest), FALSE) then
		return(FAIL);
	    else
		return(new(dom, misc::maprec(REP(x), {DV2} = DVconv), 
			   map(VARS(x), DVconv)));
	    end_if;
	end_proc;
    else
        proc(x)
        local tmp, T:DOM_DOMAIN;
        begin
	    T := domtype(x);
	    case T
            of dom do return(x);
            of DV do return(new(dom,x,{x}));
            of DF do return(new(dom,x,DFvars(x)));
            of DOM_EXPR do return(dom::convertExpr(x));
	    of DOM_POLY do return(dom::convert(expr(x)));
	    of DOM_IDENT do
		tmp := DVcoerce(x);
		if tmp=FAIL then return(new(dom,x,{}))
		else return(new(dom,tmp,{tmp})) end_if;
	    otherwise
		if contains(numTypes,T) then
		    return(new(dom,x,{}));
		elif T::hasProp<>FAIL then
		    if T::hasProp(Cat::DifferentialFunction(DV)) then
			return(dom::convert(T::convert_to(x,DOM_EXPR)));
		    else
			return(dom::convert(expr(x)));
		    end_if;
		else
		    tmp := DVcoerce(x);
		    if tmp=FAIL then return(FAIL)
		    else return(new(dom,tmp,{tmp})) end_if;
		end_if;
            end_case;
        end_proc;
    end_if;


    convertExpr :=
        proc(x:DOM_EXPR)
	local convOps:DOM_PROC, dvars:DOM_SET, res;
	begin
	    // auxiliary method for recursively recognizing differential
	    // variables and collecting them in `dvars'
	    // eliminates all unwanted types
	    convOps := proc(x)
		local tmp;
		begin
		    tmp := DVcoerce(x);
		    if tmp<>FAIL then
			dvars := dvars union {tmp};
			return(tmp);
		    end_if;
		    case domtype(x)
		    of DOM_EXPR do 
			if not(testtype(x,Type::Arithmetical)) then
			    return(FAIL);
			end_if;
			return(map(x,convOps));
		    of DV do
			dvars := dvars union {x};
			return(x);
		    of DF do
			dvars := dvars union DFvars(x);
			return(x);
		    of dom do
			dvars := dvars union VARS(x);
			return(REP(x));
		    otherwise 
			if testtype(x,Type::Arithmetical) then return(x)
			else return(FAIL) end_if;
		    end_case;
		end_proc;
	    dvars := {};
	    res := convOps(x);
	    if has(res,FAIL) then FAIL else new(dom,convOps(x),dvars) end_if;
	end_proc;


    convert_to :=
        proc(x:dom,T)
        begin
            case T
            of dom do return(x);
            of DOM_EXPR do
		return(misc::maprec(REP(x),
				    {DF}=((f) -> DFconvto(f,DOM_EXPR))));
	    of DOM_POLY do
	    	if args(0)=2 then
		    return(dom::polyDE(x,[op(VARS(x))]))
		else return(dom::polyDE(x,args(3..args(0)))) end_if;
            of DV do
                if domtype(REP(x))=DV then return(REP(x))
                else return(FAIL) end_if;
            of DF do
                if domtype(REP(x))=DF then return(REP(x))
                else return(FAIL) end_if;
            of "diff" do
                return(misc::maprec(REP(x),{DF}=((f) -> DFconvto(f,"diff")),
                                           {DV}=((v) -> DVconvto(v,"diff"))));
            of "D" do
                return(misc::maprec(REP(x),{DF}=((f) -> DFconvto(f,"D")),
                                           {DV}=((v) -> DVconvto(v,"D"))));
            otherwise return(FAIL);
            end_case;
        end_proc;


    recExpr :=
	proc(x):Type::Arithmetical
        begin misc::maprec(x,{DF}=DFexpr,{DV}=DVexpr,PostMap) end_proc;


    expr :=
        proc(x:dom):Type::Arithmetical
        begin dom::recExpr(REP(x)) end_proc;


    polyDE :=
        proc(de:dom):DOM_POLY
	local res:DOM_POLY, vars:DOM_LIST, l:DOM_LIST, DOM:DOM_DOMAIN;
	begin
            if args(0)=1 then return(poly(REP(de),[op(VARS(de))])) end_if;
	    if domtype(args(2))=DOM_LIST then
	        vars := map(args(2),DVcoerce);
		if not(iszero(contains(vars,FAIL))) then TYPE_ERR end_if;
	        res := poly(REP(de),vars,args(3..args(0)));
	    else
		vars := [op(VARS(de))];
	        res := poly(REP(de),vars,args(2..args(0)));
	    end_if;
	    if contains({args()},hold(Expr)) then
		DOM := dom;
	        l := poly2list(res);
	        l := map(l, ((c,v) -> (subsop(c,1=DOM::recExpr(c[1]),
					      Unsimplified))), vars);
	        res := poly(l,op(res,2));
	    end_if;
	    res;
	end_proc;


    equal :=
        proc(x1:dom,x2:dom):DOM_BOOL
	local r1, r2;
        begin 
	    r1 := REP(x1);
	    r2 := REP(x2);
	    if domtype(r1)<>domtype(r2) then FALSE
	    elif domtype(r1)<>DOM_EXPR then bool(r1=r2)
	    else iszero(simplify(r1-r2)) end_if;
	end_proc;


    arbFunction :=
	proc(f,vars):dom
	local res:DF;
	begin
	    if domtype(vars)<>DOM_LIST then vars := [vars] end_if;
	    vars := map(vars, ((v) -> (DVcoerce(v);
					   if %=FAIL then v else % end_if)));
	    res := DF(f,vars);
	    new(dom,res,DFvars(res));
	end_proc;


    /* ====================== */
    /* differential variables */
    /* ====================== */

    // In all of the following methods we must take into account that
    // VARS(x) contains in general also variables which do not appear
    // in REP(x) because of cancellations in arithmetic operations or in
    // differentiations. Thus we must explicitly test all elements of 
    // VARS(x) with has.

    indets :=
        proc(x:dom):DOM_SET
        local res:DOM_SET, opts:DOM_LIST, flag:DOM_BOOL, pos:DOM_INT, dv:DV;
        begin
	    if args(0)>1 then
		opts := [args(2..args(0))];
		pos := contains(opts,hold(DiffVars));
		flag := iszero(pos);
		if not(flag) then delete opts[pos] end_if;
	    else 
		flag := TRUE;
		opts := [];
	    end_if;

	    if flag or opts<>[] then return(indets(REP(x),op(opts)))
	    else
		res := VARS(x);
		for dv in res do
		    if not(has(REP(x),dv)) then
			res := res minus {dv};
		    end_if;
		end_for;
		return(res);
	    end_if;
        end_proc;


    leader :=
        proc(de:dom)
        local vars:DOM_LIST;
        begin
            vars := sort([op(VARS(de))],DVgreater);
	    while vars<>[] do
		if has(REP(de),vars[1]) then return(vars[1]);
		else delete vars[1] end_if;
	    end_while;
            return(FAIL);
        end_proc;


    has :=
	proc(de:dom,x):DOM_BOOL
	begin
	    if domtype(x)=DV then 
		if not(contains(VARS(de),x)) then FALSE
		else has(REP(de),x) end_if;
	    else has(REP(de),x) end_if;
	end_proc;


    /* =========== */
    /* arithmetics */
    /* =========== */

    zero := new(dom,0,{});


    one:= new(dom,1,{});


    iszero :=
        proc(x:dom):DOM_BOOL
        begin iszero(REP(x)) end_proc;


    isone :=
        proc(x:dom):DOM_BOOL
        begin bool(REP(x)=1) end_proc;


    _plus :=
        proc():dom
	local DOM:DOM_DOMAIN, summ:DOM_LIST, i:DOM_INT, tmp;
        begin
	    summ := [args()];
	    DOM := dom;
	    for i from 1 to args(0) do
		case domtype(args(i))
		of dom do
		of DV do
		of DF do
		of DOM_INT do
		of DOM_RAT do
		of DOM_FLOAT do
		of DOM_COMPLEX do break;
		otherwise
		    tmp := dom::coerce(args(i));
		    if tmp=FAIL then break end_if;
		    summ[i] := tmp;
		end_case;
	    end_for;

	    tmp := _plus(map(op(summ),
			     ((x) -> (if domtype(x)=DOM then REP(x)
				      else x end_if))));
	    if contains(numTypes,domtype(tmp)) then new(dom,tmp,{})
	    else
		new(dom, tmp,
		    _union(map(op(summ),
			       ((x) -> (case domtype(x)
					of DOM do return(VARS(x));
					of DV do return({x});
					of DF do return(DFvars(x));
					otherwise return({}) end_case)))));
	    end_if;
	end_proc;


    _mult :=
        proc():dom
	local DOM:DOM_DOMAIN, facts:DOM_LIST, i:DOM_INT, tmp;
        begin
	    facts := [args()];
	    DOM := dom;
	    for i from 1 to args(0) do
		case domtype(args(i))
		of dom do
		of DV do
		of DF do
		of DOM_INT do
		of DOM_RAT do
		of DOM_FLOAT do
		of DOM_COMPLEX do break;
		otherwise
		    tmp := dom::coerce(args(i));
		    if tmp=FAIL then break end_if;
		    facts[i] := tmp;
		end_case;
	    end_for;

            tmp := _mult(map(op(facts),
			     ((x) -> (if domtype(x)=DOM then REP(x)
				      else x end_if))));
	    if contains(numTypes,domtype(tmp)) then new(dom,tmp,{})
	    else
		new(dom, tmp,
		    _union(map(op(facts),
			       ((x) -> (case domtype(x)
					of DOM do return(VARS(x));
					of DV do return({x});
					of DF do return(DFvars(x));
					otherwise return({}) end_case)))));
	    end_if;
	end_proc;


    _negate :=
        proc(x:dom):dom
        begin new(dom,_negate(REP(x)),VARS(x)) end_proc;


    _invert :=
        proc(x:dom):dom
        begin new(dom,_invert(REP(x)),VARS(x)) end_proc;


    /* ==================== */
    /* differential methods */
    /* ==================== */

    coerceArgs :=
        proc(x)
        local tmp;
        begin
            tmp := DVcoerce(x);
	    if tmp=FAIL then x else tmp end_if;
	end_proc;


    diff :=
        proc(x:dom):dom
	local v:DOM_LIST, e, vv, dv:DOM_LIST, parm:DOM_LIST, vars:DOM_SET;
        begin
	    case args(0)
	    of 1 do return(x);
	    of 2 do
		v := DVcoerce(args(2));
		if testargs() then
                    if v=FAIL then v := args(2) end_if; 
                end_if;
		vars := VARS(x);
		if domtype(v)=DV then
		    if not(contains(vars,v)) then return(dom::zero) end_if;
		    e := REP(x);
		    case domtype(e)
		    of DV do return(dom::one);
		    of DF do return(new(dom,DFdiff(e,v),vars));
		    of DOM_EXPR do
			e := dom::diffE(e,[v]);
			if contains(numTypes,domtype(e)) then
			    return(new(dom,e,{}));
			else return(new(dom,e,vars)) end_if;
		    otherwise return(dom::zero);
		    end_case;
		else
		    e := REP(x);
		    case domtype(e)
		    of DV do return(dom::zero);
		    of DF do
		        e := DFdiff(e,v);
		        if iszero(e) then return(dom::zero)
		        else return(new(dom,e,vars)) end_if;
		    of DOM_EXPR do
			e := diff(e,v);
			if contains(numTypes,domtype(e)) then
			    return(new(dom,e,{}));
			else return(new(dom,e,vars)) end_if;
		    otherwise return(dom::zero);
		    end_case;
		end_if;
	    otherwise
		v := map([args(2..args(0))],dom::coerceArgs);
		dv := split(v, (i) -> bool(domtype(i)=DV));
		parm := dv[2];
		dv := dv[1];
		vars := VARS(x);
		for vv in dv do
		    if not(contains(vars,vv)) then return(dom::zero) end_if;
		end_for;
		e := REP(x);
		case domtype(e)
		of DV do return(dom::zero);
		of DF do
		    e := DFdiffN(e,v);
		    if iszero(e) then return(dom::zero)
		    else return(new(dom,e,vars)) end_if;
		of DOM_EXPR do
		    if parm<>[] then e := diff(e,op(parm)) end_if;
		    e := dom::diffE(e,dv);
		    if contains(numTypes,domtype(e)) then
			return(new(dom,e,{}));
		    else return(new(dom,e,vars)) end_if;
		otherwise return(dom::zero);
		end_case;
	    end_case;
	end_proc;


    // differentiation of an expression (DOM_EXPR) with respect to a
    // differential variable requires substitution of the variable by
    // a dummy variable in order to avoid problems with the kernel
    // function diff.
    diffE :=
	proc(x:DOM_EXPR,v:DOM_LIST):Type::Arithmetical
	local tmp:DOM_EXPR, vv:DV;
	begin
	    tmp := x;
	    for vv in v do
		if iszero(tmp) then break end_if;
		tmp := subs(tmp,vv=dummyVar,Unsimplified);
		tmp := diff(tmp,dummyVar);
		tmp := subs(tmp,dummyVar=vv,Unsimplified);
	    end_for;
	    return(tmp);
	end_proc;


    totalDiffJM :=
        proc(sys,mu,JM)
        local n,vars,nvars,dvars,rvars,rows,i,j,k,pos,res,t,tmp;
        begin
            n := (dom::jacobianType)::nrows(JM);
            vars := (dom::jacobianType)::allIndices(JM);
            if vars=[] then
                return(([dom::zero $ n], [[] $ n]));
            end_if;
            rows := (dom::jacobianType)::rows(JM);
            nvars := select(vars,
                             ((jv) -> contains({"Deriv","Dep"},DVtype(jv))));
            dvars := map(nvars,DVdiff,mu);
            vars := [ [] $ n ];

            for i from 1 to n do         // for each equation
		if rows[i]=[] then
		    sys[i] := dom::zero;
		    next;
		end_if;
                rvars := map(rows[i],op,1);
                j := nops(rvars);
                vars[i] := [ 0 $ j ];
                res := [ 0 $ j ];
                j := 0;
		k := 0;
                for t in rows[i] do      // compute derivative
                    case DVtype(t[1])
                    of "Dep" do
                    of "Deriv" do
                        pos := contains(nvars,t[1]);
			k := k+1;
                        res[k] := REP(t[2])*dvars[pos];
                        j := j+1;
                        vars[i][j] := dvars[pos];
                        break;
                    of "Indep" do
                        if DVindex(t[1])=mu then
			    k := k+1;
                            res[k] := REP(t[2]);
                        end_if;
                        break;
                    end_case;
                end_for;
		if iszero(k) then
		    sys[i] := dom::zero;
		    vars[i] := [];
		    next;
		end_if;

                vars[i] := [op(vars[i],1..j)];
                tmp := _plus(op(res,1..k));
                k := 1;
                vars[i] := vars[i] . rvars;
		sys[i] := new(dom,tmp,{op(vars[i])});
            end_for;
            (sys,vars);
        end_proc;


    /* ============== */
    /* simplification */
    /* ============== */

    normal :=
        proc(x:dom):dom
	local tmp;
        begin
	    if domtype(REP(x))=DOM_EXPR then
		tmp := normal(REP(x));
		if tmp=REP(x) then return(x)
		elif contains(numTypes,domtype(tmp)) then
		    return(new(dom,tmp,{}));
		else return(dom::generate(tmp,VARS(x))) end_if;
	    else return(x) end_if;
	end_proc;


    simplify :=
        proc(x:dom):dom
	local tmp, i:DOM_INT, n:DOM_INT, vars:DOM_LIST, dummy:DOM_LIST;
        begin
	    tmp := REP(x);
	    if domtype(tmp)=DOM_EXPR then
		vars := [op(VARS(x))];
		n := nops(vars);
		if n<=dummyN then	// sufficiently many dummies
		    dummy := [ op(dummyList,1..n) ];
		else			// need more dummy variables
		    dummyList := dummyList . 
				    [ genident("_dummy_") $ i=(dummyN+1)..n ];
		    dummyN := n;
		    dummy := dummyList;
		end_if;
		tmp := subs(tmp, (vars[i]=dummy[i] $ i=1..n));
		tmp := simplify(tmp,args(2..args(0)));
		tmp := subs(tmp, (dummy[i]=vars[i] $ i=1..n));
		if contains(numTypes,domtype(tmp)) then
		    return(new(dom,tmp,{}));
		else return(dom::generate(tmp,VARS(x))) end_if;
	    else return(x) end_if;
	end_proc;


    combine :=
	proc(x:dom):dom
	local tmp;
	begin
	    if domtype(REP(x))=DOM_EXPR then
		tmp := combine(REP(x),args(2..args(0)));
		if tmp=REP(x) then return(x)
		else new(dom,tmp,VARS(x)) end_if;
	    else return(x) end_if;
	end_proc;
	
	
    radsimp :=
	proc(x:dom):dom
	local tmp;
	begin
	    if domtype(REP(x))=DOM_EXPR then
		tmp := radsimp(REP(x),args(2..args(0)));
		if tmp=REP(x) then return(x)
		else new(dom,tmp,VARS(x)) end_if;
	    else return(x) end_if;
	end_proc;
	
	
    expand :=
	proc(x:dom):dom
	local tmp;
	begin
	    if domtype(REP(x))=DOM_EXPR then
		tmp := expand(REP(x),args(2..args(0)));
		if tmp=REP(x) then return(x)
		elif contains(numTypes,domtype(tmp)) then
		    return(new(dom,tmp,{}));
		else dom::generate(tmp,VARS(x)) end_if;
	    else return(x) end_if;
	end_proc;
	
	
    rewrite :=
	proc(x:dom):dom
	local tmp;
	begin
	    if domtype(REP(x))=DOM_EXPR then
		tmp := rewrite(REP(x),args(2..args(0)));
		if tmp=REP(x) then return(x)
		else new(dom,tmp,VARS(x)) end_if;
	    else return(x) end_if;
	end_proc;
	
	
    eval :=
        proc(sys)
	local tmp:DOM_EXPR, sol:DOM_LIST, rep, sysflag:DOM_BOOL, i:DOM_INT, 
	      j:DOM_INT, allvars:DOM_SET, vars:DOM_LIST, dersol:DOM_LIST, 
	      pos:DOM_INT, res, slist:DOM_LIST, SDOM:DOM_DOMAIN, f:dom;
        begin
	    if domtype(sys)<>DOM_LIST then
		sysflag := FALSE;
		sys := [sys];
	    else sysflag := TRUE end_if;
	    if args(0)>1 then 
		sol := args(2);
		if mm=1 and domtype(sol)<>DOM_LIST then sol := [sol] end_if;
		SDOM := domtype(sol[1]);
	    end_if;
	    sys := map(sys,dom::coerce);
	    if testargs() then
	        if not(iszero(contains(sys,FAIL))) then TYPE_ERR end_if;
		if args(0)>1 then
		    if args(0)>2 then ARG_ERR end_if;
		    if nops(sol)<>mm then
			error("wrong number of functions given");
		    end_if;
		    if SDOM::hasProp=FAIL or 
		       not(SDOM::hasProp(Cat::DifferentialFunction(DV)))
		    then
			SDOM := Dom::DifferentialExpression(
				    Dom::RestrictedDifferentialVariable(DV,
					    hold(Types)="Indep"));
		        sol := map(sol,SDOM::convert);
			if not(iszero(contains(sol,FAIL))) then 
			    TYPE_ERR 
			end_if;
		    end_if;
		end_if;
	    end_if;
	    
	    if args(0)=1 then		// lift of kernel function eval
		res := [ 0 $ nops(sys) ];
		i := 0;
		for f in sys do
		    i := i+1;
		    rep := REP(f);
		    if domtype(rep)=DOM_EXPR then
			tmp := eval(rep);
			if tmp=rep then res[i] := f
			elif contains(numTypes,domtype(tmp)) then
			    res[i] := new(dom,tmp,{});
			else
			    res[i] := dom::generate(tmp,VARS(f));
			end_if;
		    else res[i] := f end_if;
		end_for;
	    elif args(0)=2 then		// evaluation at solution
		res := [ 0 $ nops(sys) ];
		// get all contained derivatives
		allvars := _union(op(map(sys,dom::indets,hold(DiffVars))));
		allvars := select(allvars,
			      ((jv) -> not(iszero(contains(["Dep","Deriv"],
							   DVtype(jv))))));
		// compute derivatives via "derivative tree"
		vars := [ 0 $ mm ];
		dersol := [ 0 $ mm ];
		slist := [];
		for j from 1 to mm do
		    vars[j] := [op(select(allvars, 
					  ((jv) -> bool(DVindex(jv)=j))))];
		    if DV::notation="repeated" then
			dersol[j] := SDOM::multiDiff(sol[j],
						     map(vars[j],DVrepIndex));
		    else
			dersol[j] := SDOM::multiDiff(sol[j],
						     map(vars[j],DVmulti));
		    end_if;
		    slist := slist . zip(vars[j],dersol[j],_equal);
		end_for;
		// enter derivatives in expressions
		i := 0;
		for f in sys do
		    i := i+1;
		    rep := REP(f);
		    case domtype(rep)
		    of DV do
			case DVtype(rep)
			of "Indep" do 
			    res[i] := SDOM(rep);
			    break;
			of "Dep" do 
			    res[i] := sol[DVindex(rep)];
			    break;
			otherwise
			    j := DVindex(rep);
			    pos := contains(vars[j],rep);
			    res[i] := dersol[j][pos];
			end_case;
			break;
		    of DF do 
			res[i] := DFeval(rep,sol);
			break;
		    of DOM_EXPR do
			res[i] := SDOM(subs(rep,slist));
			break;
		    otherwise res[i] := SDOM(rep);
		    end_case;
		end_for;
	    else ARG_ERR end_if;
	    if sysflag then res else res[1] end_if;
	end_proc;


    /* ============= */
    /* miscellaneous */
    /* ============= */

    subs := dom::multiSubs;
    
    
    subsOne :=
        proc(de:dom,eq:DOM_EXPR):dom
	local y, v, vars:DOM_SET, res;
	begin
	    v := op(eq,1);
	    y := op(eq,2);
     	    if domtype(v)=dom then v := REP(v) end_if;
	    vars := VARS(de);
	    if domtype(v)=DV then
		if not(contains(vars,v)) then return(de) end_if;
		vars := vars minus {v};
	    end_if;
	    res := REP(de);

	    case domtype(res)
	    of DV do
		if v<>res then return(de)
		else
		    case domtype(y)
		    of DV do return(new(dom,y,{y}));
		    of DF do return(new(dom,y,DFvars(y)));
		    of dom do return(y);
		    end_case;
		end_if;
		break;
	    of DF do 
		res := DFsubs(res,v=y);
		return(new(dom,res,DFvars(res)));
		break;
	    of DOM_EXPR do
		case domtype(y)
		of dom do
		    vars := VARS(y) union vars;
		    res := subs(REP(de),v=REP(y));
		    break;
		of DV do
		    vars := vars union {y};
		    res := subs(REP(de),v=y);
		    break;
		of DF do
		    vars := DFvars(y) union vars;
		    res := subs(REP(de),v=y);
		    break;
		otherwise
		    res := subs(REP(de),v=y);
		end_case;
		break;
	    otherwise return(de);
	    end_case;
	    new(dom,res,vars);
	end_proc;


    solve :=
        proc(de,dv)
	local sol, dex, ex:DOM_EXPR, vars:DOM_SET;
	begin
	    de := dom::coerce(de);
	    dv := DVcoerce(dv);
	    if testargs() then
		if args(0)<>2 then ARG_ERR end_if;
		if (de=FAIL) or (dv=FAIL) then TYPE_ERR end_if;
	    end_if;

	    vars := VARS(de);
	    if not(contains(vars,dv)) then return(FAIL) end_if;
	    vars := vars minus {dv};
	    if domtype(REP(de))=DV then return(dom::zero) end_if;
	    ex := subs(REP(de),dv=dummyVar);
	    // Due to problems with the properties mechanism we cannot
	    // use `solve'. So currently only linear equations can be
	    // solved.
	    dex := diff(ex,dummyVar);
	    if has(dex,dummyVar) then return(FAIL)	// nonlinear in dv!
	    else
		sol := (dex*dummyVar-ex)/dex;
	    end_if;
// 	    sol := solve(ex,dummyVar);
// 	    if domtype(sol)=piecewise then
// 		sol := op(piecewise::disregardPoints(sol),1);
// 	    elif nops(sol)<>1 then return(FAIL)
// 	    else sol := op(sol,1) end_if;
	    if contains(numTypes,domtype(sol)) then new(dom,sol,{})
	    else new(dom,sol,vars) end_if;
	end_proc;


    numer :=
	proc(x:dom):dom
	local tmp;
	begin
	    if domtype(REP(x))=DOM_EXPR then
		tmp := numer(REP(x));
		if tmp=REP(x) then return(x)
		else new(dom,tmp,VARS(x)) end_if;
	    else return(x) end_if;
	end_proc;
	
	
    denom :=
	proc(x:dom):dom
	local tmp;
	begin
	    if domtype(REP(x))=DOM_EXPR then
		tmp := denom(REP(x));
		if tmp=REP(x) then return(x)
		else new(dom,tmp,VARS(x)) end_if;
	    else return(x) end_if;
	end_proc;
	
	
    random :=
        proc(N=10:DOM_INT):dom
        local i:DOM_INT, r:DOM_PROC, vars:DOM_LIST;
        begin
            r := random(1..N);
	    vars := remDup([DVrandom() $ i=1..r()]);
	    dom::convert(expr(polylib::randpoly(vars,hold(Expr))));
	end_proc;
	
	
    TeX :=
	proc(x:dom,p:DOM_INT):DOM_STRING
	begin generate::tex(REP(x),p) end_proc;


    length := 
	proc(x:dom):DOM_INT
	begin length(REP(x)) end_proc;
    
    characteristic := 0;


    // changeIndVars will take as input an element of this domain, 
    // a list of new independent variables, and a list of equations 
    // equating the new variables with the old.
    // (If, for example, the old independent variables are 
    // x and y, the new dependent variables are X and Y, and the change
    // is given by X=a*x + b*y, Y=c*x + d*y, then the third argument
    // would be [X=a*x + b*y, Y=c*x + d*y].)  The old
    // variables in terms of the new will be computed, unless given by
    // the optional argument OldFromNew = ...  
    // The output domain can be given by adding the argument 
    // OutputDomain = ... (although there is a default domain).

    changeIndVars :=
	proc(de, newVars:DOM_LIST, newFromOld:DOM_LIST)
	local ex, dvs, var, newVarTable:DOM_TABLE,
	      newDV:DOM_DOMAIN, i:DOM_INT, outputDomain:DOM_DOMAIN,
	      oldVarTable:DOM_TABLE, ivars:DOM_LIST, dvars:DOM_LIST, 
	      changeVars:DOM_PROC, oldFromNew:DOM_LIST, 
	      optionalArgTable:DOM_TABLE, eqn;
	begin
	    ivars:= map(DV::indVariables, DV::expr);
	    dvars:= map(DV::depVariables, DV::expr);
	    // Read in the optional arguments
	    optionalArgTable := table();
	    for i from 4 to args(0) do
		optionalArgTable[op(args(i),1)] := op(args(i),2);
	    end_for;
	    // Put the newFromOld in a useful form
	    newVarTable := table();
	    for eqn in newFromOld do
		newVarTable[op(eqn,1)] := op(eqn,2);
	    end_for;
	    // We will probably need oldFromNew.
	    // If it's not given to us, then we will invert newFromOld
	    // and take the first solution.
	    if contains(optionalArgTable, hold(OldFromNew)) then
		oldFromNew := optionalArgTable[hold(OldFromNew)];
	    else
		oldFromNew:= op(solve({op(newFromOld)},{op(ivars)}),1);	    
	    end_if;
	    oldVarTable := table();
	    for eqn in oldFromNew do
		oldVarTable[op(eqn,1)] := op(eqn,2);
	    end_for;
	    // Next, the workhorse function
	    changeVars :=
		proc(v)
		local fn:DOM_EXPR;
		begin
		    case DVtype(v)
		    of "Dep" do
		    of "Deriv" do
			fn:= dvars[DVindex(v)](op(newVars));
			for i in DVrepIndex(v) do
			    fn:= _plus(diff(fn,newVars[j])*
				diff(newVarTable[newVars[j]],ivars[i]) $
				j=1..nn);
			end_for;
			break;
		    of "Indep" do
			fn := oldVarTable[DVexpr(v)];
			break;
		    end_case;
		    return(fn);
		end_proc; /* changeVars */
	    ex:= REP(de);  // The expression
	    dvs:= VARS(de); // The differential variables
	    for var in dvs do
		ex:= subs(ex, var=changeVars(var));
	    end_for;
	    // Now prepare the output domain
	    if contains(optionalArgTable, hold(OutputDomain)) then
		outputDomain := optionalArgTable[OutputDomain]
	    else
		newDV:= Dom::DifferentialVariable(newVars, dvars);	    
		outputDomain := Dom::DifferentialExpression(newDV);
	    end_if;
	    return(outputDomain(ex));
	 end_proc; /* changeIndVars */


    // changeDepVars will take as input an element of this domain, 
    // a list of new dependent variables, and a list of
    // equations equating the old variables with the new,
    // (If, for example, the old dependent variables are 
    // u and v, the new dependent variables are U and V, and the change
    // is given by u=a*U + b*V, v=c*U + d*V, then the third argument
    // would be [u=a*U + b*V, v=c*U + d*V].
    // The output domain can be given by adding the argument 
    // OutputDomain = ... (although there is a default domain).

    changeDepVars :=
	proc(de, newVars:DOM_LIST, oldFromNew:DOM_LIST)
	local ex, dvs, var, oldVarTable:DOM_TABLE, ivars:DOM_LIST,
	      dvars:DOM_LIST, optionalArgTable:DOM_TABLE, eqn, 
	      newDV:DOM_DOMAIN, outputDomain:DOM_DOMAIN, i:DOM_INT, 
	      changeVars:DOM_PROC;
	begin
	    ivars:= map(DV::indVariables, DV::expr);
	    dvars:= map(DV::depVariables, DV::expr);
	    // Read in the optional arguments
	    optionalArgTable := table();
	    for i from 4 to args(0) do
		optionalArgTable[op(args(i),1)] := op(args(i),2);
	    end_for;
	    // Put the oldFromNew in a useful form
	    oldVarTable := table();
	    for eqn in oldFromNew do
		oldVarTable[op(eqn,1)] := op(eqn,2);
	    end_for;
	    // Next, the workhorse function
	    changeVars :=
		proc(v)
		local fn:DOM_EXPR;
		begin
		    case DVtype(v)
		    of "Dep" do
			fn:= oldVarTable[DVexpr(v)];
			break;
		    of "Indep" do
			fn:= DVexpr(v);
			break;
		    of "Deriv" do
			fn:= 
			 oldVarTable[dvars[DVindex(v)]](map(DVrepIndex(v),
						  x->ivars[x]));
			break;
		    end_case;
		    return(fn);
		end_proc; /* changeVars */
	    ex:= REP(de);  // The expression
	    dvs:= VARS(de); // The differential variables
	    for var in dvs do
		ex:= subs(ex, var=changeVars(var));
	    end_for;
	    // Now prepare the output domain
	    if contains(optionalArgTable, hold(OutputDomain)) then
		outputDomain := optionalArgTable[OutputDomain]
	    else
		newDV:= Dom::DifferentialVariable(ivars,newVars);	    
		outputDomain := Dom::DifferentialExpression(newDV);
	    end_if;
	    return(outputDomain(ex));
	end_proc; /* changeIndVars */


begin

    // analysis of passed arguments
    opts := select([args()], (a -> bool(type(a)="_equal")));
    if iszero(nops(opts)) then		// DV directly given
	if args(0)<>1 then ARG_ERR end_if;
	DV := args(1);
	if domtype(DV)<>DOM_DOMAIN or DV::hasProp=FAIL or
	   not(DV::hasProp(Cat::DifferentialVariable)) then
	    error("argument must be Cat::DifferentialVariable");
	end_if;
	restrictedDV := bool(DV::constructor=
			     Dom::RestrictedDifferentialVariable);
    else				// options passed
	if nops(opts)<>args(0) then error("inconsistent arguments") end_if;
	if args(0)>2 then ARG_ERR end_if;
	if op(opts[1],1)<>hold(Vars) then TYPE_ERR end_if;
	if args(0)=2 and op(opts[2],1)<>hold(Rest) then TYPE_ERR end_if;
	opts := map(opts, 
		    (a -> if domtype(op(a,2))<> DOM_LIST then 
			      subsop(a,2=[op(a,2)]) 
			  else a end_if));
	DV := Dom::DifferentialVariable(op(op(opts[1],2)));
	if args(0)=1 then
	    restrictedDV := FALSE;
	else
	    DV := Dom::RestrictedDifferentialVariable(DV,op(op(opts[2],2)));
	    restrictedDV := TRUE;
	end_if;
    end_if;
    
    // restricted differential variables
    if restrictedDV then
	DV2 := DV::Variables;
	DF2 := Dom::DifferentialExpression(DV2);
	DVconv := DV::convDV;
	DVtest := DV::test;
	DF2convert := DF2::convert;
    end_if;

    // methods from DV
    DVcoerce := DV::coerce;
    DVconvto := DV::convert_to;
    DVexpr := DV::expr;
    DVdiff := DV::totalDiff;
    DVindex := DV::index;
    DVtype := DV::vartype;
    DVgreater := ((j1,j2) -> DV::_less(j2,j1));
    DVrandom := DV::random;
    DVrepIndex := DV::repeatedIndex;
    DVmulti := DV::multiIndex;
    DVnewRep := DV::newRepeated;

    // other global variables
    nn := DV::numIndVar;
    mm := DV::numDepVar;
    merge := listlib::merge;
    singleMerge := listlib::singleMerge;
    remDup := listlib::removeDuplicates;
    dummyVar := genident("_dummy_");
    dummyList := [ genident("_dummy_") $ dummyN=1..20 ];
    dummyN := 20;
    numTypes := {DOM_IDENT,DOM_INT,DOM_COMPLEX,DOM_FLOAT,DOM_RAT};

    // differential functions
    DF := Dom::DifferentialFunction(DV);
    DFvars := ((x) -> DF::indets(x,hold(DiffVars)));
    DFexpr := DF::expr;
    DFdiff := DF::diff;
    DFdiffN := DF::diffN;
    DFconvto := DF::convert_to;
    DFsubs := DF::subs;
    DFeval := DF::eval;
	
end_domain:


unalias(REP,VARS,TYPE_ERR,ARG_ERR):



