/*++
Dom::DifferentialFunction(DV) -- Domain of arbitrary functions of differential
	variables from the domain DV of Cat::DifferentialVariable()

This domain is a hack around problems with arbitrary functions of differential
variables in Dom::DifferentialExpression(DV).  Its elements represent such
functions.  The main functionality is that they can be efficiently
differentiated.  But no arithmetics is implemented, as the elements of this
domain should always be used from within Dom::DifferentialExpression(DV).  For
this reason Dom::DifferentialFunction(DV) belongs only to Cat::BaseCategory 
and not to Cat::DifferentialFunction(DV)!

An element is represented by its operator which can be either an (possibly
indexed) identifier or an expression of type "D" and by its two variable
lists.  The first list contains all differential variables from DV (sorted by
DVgreater); the second list contains all other variables (without any special
order).


Author:	 Werner M. Seiler, Uni Karlsruhe/Mannheim
	 (wms@ira.uka.de, seiler@euler.math.uni-mannheim.de)
Created:	Mon Oct  5 09:54:13 1998
Last update: 	Fri Sep  3 16:36:17 1999
++*/


alias(OP(x)=extop(x,1),
      VARS(x)=extop(x,2),
      ARG_ERR=error("wrong number of arguments"),
      TYPE_ERR=error("wrong type of argument"),
      SUBS_ERR=error("only substitution of identifiers or differential
variables allowed")):


domain Dom::DifferentialFunction(DV)
    local DVcoerce:DOM_PROC, DVexpr:DOM_PROC, DVgreater:DOM_PROC,
          DVconvto:DOM_PROC, insertord:DOM_PROC, remDup:DOM_PROC,
          merge:DOM_PROC;
    inherits Dom::BaseDomain;
    category Cat::BaseCategory;


    DVcoerce2 :=
	proc(v)
	local tmp;
	begin
	    tmp := DVcoerce(v);
	    if tmp=FAIL then
	        case type(v)
		of DOM_IDENT do
		of "_index" do return(v);
	        otherwise TYPE_ERR
	        end_case;
	    else tmp end_if;
	end_proc;


    /* ========================= */
    /* representation, in/output */
    /* ========================= */

    convert :=
	proc(x,v)
	begin
	    if testargs() then
		if args(0)<>2 then ARG_ERR end_if;
		if iszero(contains([DOM_IDENT,"D","_index"],type(x))) or
		   domtype(v)<>DOM_LIST then return(FAIL) end_if;
	        v := remDup(v);
	        v := map(v,dom::DVcoerce2);
	    end_if;
	    new(dom,x,v);
	end_proc;


    convert_to :=
        proc(x:dom,T)
        begin
            case T
            of dom do return(x);
            of DOM_EXPR do return(OP(x)(op(VARS(x))));
            of "diff" do
                return(OP(x)(op(map(VARS(x),
				    (v -> if domtype(v)=DV then 
					      DVconvto(v,"diff") 
					  else v end_if)))));
            of "D" do
                return(OP(x)(op(map(VARS(x),
				    (v -> if domtype(v)=DV then 
					      DVconvto(v,"diff") 
					  else v end_if)))));
            otherwise return(FAIL);
            end_case;
        end_proc;


    expr :=
	proc(x:dom):DOM_EXPR
	begin 
	    OP(x)(op(map(VARS(x),
		         (v -> if domtype(v)=DV then DVexpr(v) 
			       else v end_if)))); 
	end_proc;


    TeX :=
	proc (x:dom,p:DOM_INT):DOM_STRING
	begin
	    generate::TeXfun(generate::tex(OP(x),p),p,op(VARS(x)));
	end_proc;


    /* ===================== */
    /* miscellaneous methods */
    /* ===================== */

    has :=
	proc(x:dom,v):DOM_BOOL
	begin
	    has(OP(x),v) or not(iszero(contains(VARS(x),v)));
	end_proc;
	
	
    equal :=
	proc(x:dom,y:dom):DOM_BOOL
	begin 
	    bool(OP(x)=OP(y)) and bool({op(VARS(x))}={op(VARS(y))});
        end_proc;


    indets :=
	proc(x:dom):DOM_SET
	begin 
	    if args(0)>1 then
		if testargs() and (args(2)<>hold(DiffVars)) then
		    error("unknown option");
		end_if;
		{op(select(VARS(x), (v -> bool(domtype(v)=DV))))};
	    else {op(VARS(x))} end_if;
	end_proc;


    subs :=
	proc(x:dom,eq):dom
	local v1, v2, pos1:DOM_INT, pos2:DOM_INT, ind:DOM_LIST,
	      vars:DOM_LIST, xop;
	begin
	    if testargs() and type(eq)<>"_equal" then TYPE_ERR end_if;
	    if testargs() then
		v1 := dom::DVcoerce2(op(eq,1));
		v2 := dom::DVcoerce2(op(eq,2));
	    else
		v1 := op(eq,1);
		v2 := op(eq,2);
	    end_if;
	    vars := VARS(x);
	    pos1 := contains(vars,v1);
	    if iszero(pos1) then return(x) end_if;
	    pos2 := contains(vars,v2);
	    if iszero(pos2) then
		return(new(dom,OP(x),subsop(vars,pos1=v2)));
	    else
		delete vars[pos1];
		xop := OP(x);		     // correction of derivatives
		if type(xop)<>"D" then return(new(dom,xop,vars)) end_if;
		if pos2>pos1 then pos2 := pos2-1 end_if;
		ind := map(op(xop,1), 
			   ((j,p1,p2) -> if j=p1 then p2
					 elif j>p1 then j-1
					 else j end_if),
			   pos1,pos2);
		return(new(dom,subsop(xop,1=ind,Unsimplified),vars));
	    end_if;
	end_proc;


    /* ==================== */
    /* differential methods */
    /* ==================== */

    diff :=
    	proc(x:dom)
	local pos:DOM_INT, y, dop, l:DOM_LIST, vars:DOM_LIST;
	begin
	    if args(0)>2 then
		if testargs() then y := map([args(2..args(0))],dom::DVcoerce2);
		else y := [args(2..args(0))] end_if;
		return(dom::diffN(x,y));
	    else
		if testargs() then
		    y := dom::DVcoerce2(args(2));
		else y := args(2) end_if;
	    end_if;

	    vars := VARS(x);
	    pos := contains(vars,y);
	    if iszero(pos) then return(0) end_if;
	    dop := OP(x);
	    if type(dop)="D" then
		l := insertord(op(dop,1), pos);
		new(dom,subsop(dop,1=l,Unsimplified),vars);
	    else new(dom,D([pos],dop),vars) end_if;
	end_proc;


    diffN :=
    	proc(x:dom,l:DOM_LIST)
	local pos:DOM_INT, i:DOM_INT, dop, vars:DOM_LIST;
	begin
	    vars := VARS(x);
	    for i from 1 to nops(l) do
	        pos := contains(vars,l[i]);
	        if iszero(pos) then return(0) end_if;
	        l[i] := pos;
	    end_for;

	    l := sort(l);
	    dop := OP(x);
	    if type(dop)="D" then
		l := merge(l,op(dop,1));
		new(dom,subsop(dop,1=l,Unsimplified),vars);
	    else new(dom,D(l,dop),vars) end_if;
	end_proc;


begin

    if (domtype(DV)<>DOM_DOMAIN) or (DV::hasProp=FAIL) or
       not(DV::hasProp(Cat::DifferentialVariable)) then
        error("argument must be Cat::DifferentialVariable");
    end_if;

    DVcoerce := DV::coerce;
    DVexpr := DV::expr;
    DVgreater := ((j1,j2) -> DV::_less(j2,j1));
    DVconvto := DV::convert_to;

    insertord := listlib::insert;
    remDup := listlib::removeDuplicates;
    merge := listlib::merge;

end_domain:


unalias(OP,VARS,PARMS,ARG_ERR,TYPE_ERR,SUBS_ERR):
