/*++
Dom::DifferentialVariables -- domain of differential variables
                              (element of Cat::DifferentialVariable)

Depending on the arguments this domain provides two fairly different
implementations of differential variables. The first possibility is to give
as arguments two list of identifiers (DOM_IDENT). The first list will be
considered as the names of the independent variables and the second list as
the names of the dependent ones. Dom::DifferentVariables uses then repeated
index notation.

The second possibility is to give three arguments, namely identifiers used
to generate the names of the independent and dependent variables and
derivatives, respectively. Numbers given behind the two first identifiers
are interpreted as the numbers of variables. Dom::DifferentialVariables
uses now multi index notation.

Examples for the generation of domains and differential variables.
    repeated index:
	JB := Dom::DifferentialVariable([x, y, z], [u, v]);
	j1 := JB(y);                  independent variable y
	j2 := JB(u);                  dependent variable u
	j3 := JB(v([x,x,z]));         derivative v_{xxz}
    multi index:
	IJB := Dom::DifferentialVariable(x[3], u[2], p);
	ij1 := IJB(x[2]);             independent variable x_2
	ij2 := IJB(u[1]);             dependent variable u_1
	ij3 := IJB(p[2]([2,0,1]));    derivative of u_2 with respect to
				      x_1 (twice) and x_3 (once)
Other possibilities are the diff or D notation of MuPAD or the direct
entering of the representation. Output in these notation is also possible
with convert_to(jv,"diff") or convert_to(jv,"D").

The representation consists of five slots: (i) the type of the differential
variable, (2) the upper index numbering the independent or dependent
variables, (3) the lower index for derivatives, (4) the order and (5) the
class of derivatives. The form of the lower index depends on the chosen
notation; it is either a multi or a repeated index.  Order and class are
stored to accelerate some operations (especially _less).


Authors: Eduard Niess, Werner M. Seiler, Uni Karlsruhe/Mannheim
         (wms@ira.uka.de, seiler@euler.math.uni-mannheim.de)
Created: ??? (1996)
Last update: 	Fri Jun  2 16:31:43 2000
++*/
// 13.10.2000

alias(TYPE(x)=extop(x,1),
      INDEX(x)=extop(x,2),
      DERIV_INDEX(x)=extop(x,3),
      ORDER(x)=extop(x,4),
      CLASS(x)=extop(x,5),
      ARG_ERR=error("wrong number of arguments"),
      TYPE_ERR=error("wrong type of argument"),
      VAR_ERR=error("unknown type of differential variable (internal bug?)")):


domain Dom::DifferentialVariable(ivar, dvar)

    local nn:DOM_INT, mm:DOM_INT, IndexedFlag:DOM_BOOL, i, pvar,
          IVar:DOM_LIST, DVar:DOM_LIST, PVar:DOM_LIST,
	  TeXindex:DOM_PROC, TeXexp:DOM_PROC;
    inherits Dom::BaseDomain;
    category Cat::DifferentialVariable();
    axiom Ax::canonicalRep, Ax::normalRep;


    /* ========================= */
    /* representation, in/output */
    /* ========================= */

    newMulti :=
        if IndexedFlag then
            proc(t:DOM_STRING,i:DOM_INT,l=[]:DOM_LIST):dom
	    local c:DOM_INT, o:DOM_INT;
            begin 
		case t
		of "Indep" do 
		    o := -1;
		    c := 0;
		    l := [ 0 $ nn ];
		    break;
		of "Dep" do 
		    o := 0;
		    c := nn;
		    l := [ 0 $ nn ];
		    break;
		of "Deriv" do 
		    c := 1;
		    while (c<=nn) and iszero(l[c]) do c := c+1 end_while;
		    if c>nn then
			t := "Dep";
			o := 0;
			c := nn;
		    else o := _plus(op(l)) end_if;
		    break;
		otherwise VAR_ERR;
		end_case;
		new(dom,t,i,l,o,c);
	    end_proc
	else
	    proc(t:DOM_STRING,i:DOM_INT,l=[]:DOM_LIST):dom
	    local rind:DOM_LIST, c:DOM_INT, o:DOM_INT;
	    begin
		case t
		of "Indep" do 
		    o := -1;
		    c := 0;
		    rind := [];
		    break;
		of "Dep" do 
		    o := 0;
		    c := nn;
		    rind := [];
		    break;
		of "Deriv" do 
		    rind := dom::m2r(l);
		    if rind=[] then
			t := "Dep";
			o := 0;
			c := nn;
		    else 
			o := nops(rind);
			c := rind[1];
		    end_if;
		    break;
		otherwise VAR_ERR;
		end_case;
		new(dom,t,i,rind,o,c);
		end_proc
	end_if;


    newRepeated :=
        if IndexedFlag then
	    proc(t:DOM_STRING,i:DOM_INT,l=[]:DOM_LIST):dom
	    local mind:DOM_LIST, o:DOM_INT, c:DOM_INT;
	    begin
		if l<>[] then l := sort(l) end_if;
		case t
		of "Indep" do 
		    o := -1;
		    c := 0;
		    mind := [ 0 $ nn ];
		    break;
		of "Dep" do 
		    o := 0;
		    c := nn;
		    mind := [ 0 $ nn ];
		    break;
		of "Deriv" do 
		    if l=[] then
			t := "Dep";
			mind := [ 0 $ nn ];
			o := 0;
			c := nn;
		    else 
			mind := dom::r2m(l);
			o := nops(l);
			c := l[1];
		    end_if;
		    break;
		otherwise VAR_ERR;
		end_case;
		new(dom,t,i,mind,o,c);
	    end_proc
	else
            proc(t:DOM_STRING,i:DOM_INT,l=[]:DOM_LIST):dom
	    local o:DOM_INT, c:DOM_INT;
            begin 
		if l<>[] then l := sort(l) end_if;
		case t
		of "Indep" do 
		    o := -1;
		    c := 0;
		    break;
		of "Dep" do 
		    o := 0;
		    c := nn;
		    break;
		of "Deriv" do 
		    if l=[] then
			t := "Dep";
			o := 0;
			c := nn;
		    else 
			o := nops(l);
			c := l[1];
		    end_if;
		    break;
		otherwise VAR_ERR;
		end_case;
		new(dom,t,i,l,o,c);
	    end_proc
	end_if;


    convert :=
        proc(e)
	local ind:DOM_INT,derind:DOM_LIST,ord:DOM_INT,cl:DOM_INT,dep;
	begin
            if args(0)>1 then          // direct entering of representation
		return(dom::convertRep(testargs(),args()));
	    end_if;

	    case type(e)                            // conversion
	    of dom do return(e);
	    of DOM_INT do return(FAIL);
	    of DOM_IDENT do
	    of "_index" do
	    of "function" do 
		ind := contains(IVar,e);
		if not(iszero(ind)) then 
		    return(new(dom,"Indep",ind,[],-1,0));
	        end_if;
		ind := contains(DVar,e);
		if not(iszero(ind)) then 
		    return(new(dom,"Dep",ind,[],0,nn)); 
		end_if;
		if domtype(e)=DOM_IDENT then return(FAIL) 
		else return(dom::convertFun(e)) end_if;
	    of "diff" do return(dom::convertDiff(e));
	    of "D" do
		dep := dom::convert(op(e,2));
		if dep=FAIL or TYPE(dep)<>"Dep" then return(FAIL) end_if;
		derind := op(e,1);
		derind := sort(derind);
		ord := nops(derind);
		cl := derind[1];
		if IndexedFlag then
		    derind := dom::r2m(derind);
		end_if;
		return(new(dom,"Deriv",INDEX(dep),derind,ord,cl));
	    otherwise return(FAIL);
	    end_case;
	end_proc;
	
	
    convertRep :=
	proc(test:DOM_BOOL,t:DOM_STRING,ind:DOM_INT)
	local derind:DOM_LIST, ord:DOM_INT, cl:DOM_INT, i:DOM_INT;
	begin
	    if test then
		case t
		of "Indep" do
		    if ind<0 or ind>nn then return(FAIL) end_if;
		    break;
		of "Dep" do
		    if ind<0 or ind>mm then return(FAIL) end_if;
		    break;
		of "Deriv" do break;
		otherwise return(FAIL);
		end_case;
	    end_if;

	    if args(0)>3 then
                if t<>"Deriv" then return(FAIL) end_if;
                if test and not(testtype(args(4),Type::ListOf(DOM_INT))) then
		    return(FAIL);
		end_if;
            end_if;
	    if test and (args(0)>4) and ((args(0)<>6) or
	       (domtype(args(5))<>DOM_INT) or
	       (domtype(args(6))<>DOM_INT) or
	       (args(5)<0) or (args(6)<0) or (args(6)>nn)) then 
		return(FAIL);
	    end_if;

	    if args(0)=4 then
		if IndexedFlag then
		    if test and nops(args(4))<>nn then return(FAIL) end_if;
		    derind := args(4);
		    ord := 0;
		    cl := 0;
		    for i from 1 to nn do
			if test and derind[i]<0 then return(FAIL) end_if;
			if not(iszero(derind[i])) then
			    if iszero(cl) then cl := i end_if;
			    ord := ord+derind[i];
			end_if;
		    end_for;
		else
		    if args(4)=[] then
			derind := [];
			if t="Dep" then
			    ord := 0;
			    cl := nn;
			else
			    ord := -1;
			    cl := 0;
			end_if;
		    else
			if test then
			    for i in args(4) do
				if i<0 or i>nn then return(FAIL) end_if;
			    end_for;
			end_if;
			derind := sort(args(4));
			ord := nops(derind);
			cl := derind[1];
		    end_if;
		end_if;
	    elif args(0)=6 then
		derind := args(4);
		if IndexedFlag then
		    if test then
			for i in derind do
			    if i<0 then return(FAIL) end_if;
			end_for;
		    end_if;
		else
		    if test then
			for i in derind do
			    if i<0 or i>nn then return(FAIL) end_if;
			end_for;
		    end_if;
		    derind := sort(derind);
		end_if;
		ord := args(5);
		cl := args(6);
	    else 
		derind := [];
		case t
		of "Indep" do 
		    ord := -1;
		    cl := 0;
		    break;
		of "Dep" do 
		    ord := 0;
		    cl := nn;
		    break;
		otherwise return(FAIL);
		end_case;
	    end_if;
	    return(new(dom,t,ind,derind,ord,cl));
	end_proc;


    convertFun :=
	proc(e:DOM_EXPR)
	local derind:DOM_LIST, ord:DOM_INT, cl:DOM_INT, i:DOM_INT, 
	      pos:DOM_INT, eop, vars:DOM_LIST, dep, ind:DOM_INT;
	begin
	    eop := op(e,0);
	    case type(eop)
	    of DOM_IDENT do
	    of "_index" do
	    of "function" do
		if IndexedFlag then ind := contains(PVar,eop)
		else ind := contains(DVar,eop) end_if;
		if iszero(ind) then return(FAIL) end_if;
		derind := op(e,1);		// check for derivative
		if nops(e)=1 and domtype(derind)=DOM_LIST then
		    if IndexedFlag then
			if nops(derind)>nn then return(FAIL) end_if;
			ord := 0;
			cl := 0;
			for i from 1 to nn do
			    if not(testtype(derind[i],Type::NonNegInt)) then
				return(FAIL)
			    end_if;
			    if not iszero(derind[i]) then
				if iszero(cl) then cl := i end_if;
				ord := ord+derind[i];
			    end_if;
			end_for;
		    else
			derind := [ 0 $ nops(op(e,1)) ];
			ord := 0;
			for i in op(e,1) do
			    ord := ord+1;
			    pos := contains(IVar,i);
			    if iszero(pos) then return(FAIL) end_if;
			    derind[ord] := pos;
			end_for;
			derind := sort(derind);
			cl := derind[1];
		    end_if;
		    return(new(dom,"Deriv",ind,derind,ord,cl));
		else 	// check for dependent variable with arguments
		    ind := contains(DVar,eop);
		    derind := {op(e)};
		    if iszero(ind) or derind<>{op(IVar)} then 
			return(FAIL); 
		    end_if;
		    return(new(dom,"Dep",ind,[],0,nn));
		end_if;
	    of "D" do
		vars := [op(e)];
		dep := dom::convert(op(eop,2));
		if dep=FAIL or TYPE(dep)<>"Dep" then 
		    return(FAIL); 
		end_if;
		ind := INDEX(dep);
		derind := op(eop,1);
		ord := 0;
		for i in derind do
		    pos := contains(IVar,vars[i]);
		    if iszero(pos) then return(FAIL) end_if;
		    ord := ord+1;
		    derind[ord] := pos;
		end_for;
		derind := sort(derind);
		cl := derind[1];
		if IndexedFlag then
		    derind := dom::r2m(derind);
		end_if;
		return(new(dom,"Deriv",ind,derind,ord,cl));
	    otherwise return(FAIL);
	    end_case;
	end_proc;


    convertDiff :=
	if IndexedFlag then
	    proc(e:DOM_EXPR)
	    local funname, ind:DOM_INT, derind:DOM_LIST, ord:DOM_INT,
		  cl:DOM_INT, i:DOM_INT, pos:DOM_INT;
	    begin
		funname := dom::convert(op(e,[1,0]));
		if funname=FAIL or TYPE(funname)<>"Dep" then 
		    return(FAIL) 
		end_if;
		ind := INDEX(funname);
		derind := [ 0 $ nn ];
		ord := nops(e)-1;
		cl := nn;
		for i in [op(e,2..nops(e))] do
		    pos := contains(IVar,i);
		    if iszero(pos) then return(FAIL) end_if;
		    derind[pos] := derind[pos]+1;
		    if pos<cl then cl := pos end_if;
		end_for;
		return(new(dom,"Deriv",ind,derind,ord,cl));
	    end_proc;
	else
	    proc(e:DOM_EXPR)
	    local funname, ind:DOM_INT, derind:DOM_LIST, ord:DOM_INT,
		  cl:DOM_INT, i:DOM_INT, pos:DOM_INT;
	    begin
		funname := dom::convert(op(e,[1,0]));
		if funname=FAIL or TYPE(funname)<>"Dep" then 
		    return(FAIL) 
		end_if;
		ind := INDEX(funname);
		derind := [ 0 $ nops(e)-1 ];
		ord := 0;
		for i in [op(e,2..nops(e))] do
		    ord := ord+1;
		    pos := contains(IVar,i);
		    if iszero(pos) then return(FAIL) end_if;
		    derind[ord] := pos;
		end_for;
		derind := sort(derind);
		cl := derind[1];
		return(new(dom,"Deriv",ind,derind,ord,cl));
	    end_proc;
	end_if;
	
	
    convert_to :=
        proc(jv:dom,T)
	local f,vars;
	begin
	    case T
	    of dom do return(jv);
	    of DOM_EXPR do return(dom::expr(jv));
	    of "diff" do
		case TYPE(jv)
		of "Indep" do return(IVar[INDEX(jv)]);
		of "Dep" do return(DVar[INDEX(jv)](op(IVar)));
		of "Deriv" do
		    f := DVar[INDEX(jv)](op(IVar));
		    vars := map(dom::repeatedIndex(jv),
				(()->(IVar[args(1)])));
		    return(diff(f,op(vars)));
		otherwise VAR_ERR;
		end_case;
	    of "D" do
		case TYPE(jv)
		of "Indep" do return(IVar[INDEX(jv)]);
		of "Dep" do return(DVar[INDEX(jv)]);
		of "Deriv" do
		    f := DVar[INDEX(jv)];
		    vars := dom::repeatedIndex(jv);
		    return(D(vars,f));
		otherwise VAR_ERR;
		end_case;
	    otherwise return(FAIL);
	    end_case;
	end_proc;


    expr :=
        proc(jv:dom):Type::Arithmetical
        local tmp:DOM_LIST;
        begin
            case TYPE(jv)
	    of "Indep" do return(IVar[INDEX(jv)]);
	    of "Dep" do return(DVar[INDEX(jv)]);
	    of "Deriv" do
		if IndexedFlag then
		    return(PVar[INDEX(jv)](DERIV_INDEX(jv)));
		else
		    tmp := map(DERIV_INDEX(jv),((i)->IVar[i]));
		    return(DVar[INDEX(jv)](tmp));
		end_if;
	    otherwise VAR_ERR;
	    end_case;
	end_proc;


    TeX :=
	if IndexedFlag then
	    proc(jv:dom,p:DOM_INT):DOM_STRING
	    local tmp;
	    begin;
		case TYPE(jv)
		of "Indep"do return(TeXexp(op(IVar[1],1),INDEX(jv),p));
		of "Dep"  do return(TeXexp(op(DVar[1],1),INDEX(jv),p));
		of "Deriv"do 
		    tmp := op(PVar[1],1)[DERIV_INDEX(jv)];
		    return(TeXexp(tmp,INDEX(jv),p));
		otherwise VAR_ERR;
		end_case;
	    end_proc;
	else
	    proc(jv:dom,p:DOM_INT):DOM_STRING
	    local sub:DOM_LIST;
	    begin
		case TYPE(jv)
		of "Indep"do return(generate::tex(IVar[INDEX(jv)],p));
		of "Dep"  do return(generate::tex(DVar[INDEX(jv)],p));
		of "Deriv"do 
		    sub := map(DERIV_INDEX(jv),((i)->IVar[i]));
		    return(TeXindex(DVar[INDEX(jv)],op(sub),p));
		otherwise VAR_ERR;
		end_case;
	    end_proc;
	end_if;
	

    /* ============== */
    /* miscelleaneous */
    /* ============== */

    index :=
        proc(jv:dom):DOM_INT
        begin INDEX(jv) end_proc;


    repeatedIndex :=
        if IndexedFlag then
            proc(jv:dom):DOM_LIST
            begin dom::m2r(DERIV_INDEX(jv)) end_proc;
	else
	    proc(jv:dom):DOM_LIST
	    begin DERIV_INDEX(jv) end_proc;
	end_if;


    multiIndex :=
        if IndexedFlag then
            proc(jv:dom):DOM_LIST
            begin DERIV_INDEX(jv) end_proc;
	else
	    proc(jv:dom):DOM_LIST
	    begin dom::r2m(DERIV_INDEX(jv)) end_proc;
	end_if;


    vartype :=
        proc(jv:dom):DOM_STRING
        begin TYPE(jv) end_proc;
	
	
    order :=
	proc(jv:dom):DOM_INT
	begin ORDER(jv) end_proc;


    class :=
	proc(jv:dom):DOM_INT
	begin CLASS(jv) end_proc;


    notation := if IndexedFlag then "multi" else "repeated" end_if;


    numIndVar := nn;


    numDepVar := mm;


    /* =============== */
    /* differentiation */
    /* =============== */
    
    // we overwrite the default implementations in order to gain a bit
    // of efficiency, as we omit to recompute the order and the class
    
    
    totalDiff2 :=
        if dom::notation="multi" then
	    proc(jv:dom,i:DOM_INT):dom
	    local tmp:DOM_LIST;
	    begin
	        case dom::vartype(jv)
	    	of "Dep" do
		    tmp := [ 0 $ nn ];
		    tmp[i] := 1;
		    return(new(dom,"Deriv",INDEX(jv),tmp,1,i));
	            //return(dom::newRepeated("Deriv",dom::index(jv),[i]));
	    	of "Deriv" do
		    tmp := DERIV_INDEX(jv);
		    tmp[i] := tmp[i]+1;
		    return(new(dom,"Deriv",INDEX(jv),tmp,
		               ORDER(jv)+1,min(i,CLASS(jv))));
	            //tmp := dom::multiRaise(dom::multiIndex(jv),i);
	            //return(dom::newMulti("Deriv",dom::index(jv),tmp));
		otherwise error("not defined");
	    	end_case;
	    end_proc
	else
	    proc(jv:dom,i:DOM_INT):dom
	    local tmp:DOM_LIST;
	    begin
	        case dom::vartype(jv)
	    	of "Dep" do
		    return(new(dom,"Deriv",INDEX(jv),[i],1,i));
	            //return(dom::newRepeated("Deriv",dom::index(jv),[i]));
	    	of "Deriv" do
	            tmp := listlib::insert(dom::repeatedIndex(jv),i);
		    return(new(dom,"Deriv",INDEX(jv),tmp,ORDER(jv)+1,tmp[1]));
	            //return(dom::newRepeated("Deriv",dom::index(jv),tmp));
		otherwise error("not defined");
	    	end_case;
	    end_proc
	end_if;


    totalDiffN :=
        if dom::notation="multi" then
	    proc(jv:dom,l:DOM_LIST):dom
	    local tmp:DOM_LIST;
	    begin
	        case dom::vartype(jv)
	        of "Dep" do
		    return(new(dom,"Deriv",INDEX(jv),dom::r2m(l),
			       nops(l),min(op(l))));
	            //return(dom::newMulti("Deriv",dom::index(jv),dom::r2m(l)));
		of "Deriv" do
		    tmp := dom::multiRaise(dom::multiIndex(jv),l);
		    return(new(dom,"Deriv",INDEX(jv),tmp,
			       ORDER(jv)+nops(l),min(CLASS(jv),min(op(l)))));
		    //return(dom::newMulti("Deriv",dom::index(jv),tmp));
		otherwise error("not defined");
	        end_case;
	    end_proc
	else
	    proc(jv:dom,l:DOM_LIST):dom
	    local tmp:DOM_LIST;
	    begin
	        case dom::vartype(jv)
	        of "Dep" do
		    tmp := sort(l);
		    return(new(dom,"Deriv",INDEX(jv),tmp,nops(tmp),tmp[1]));
	            //return(dom::newRepeated("Deriv",dom::index(jv),sort(l)));
	        of "Deriv" do
	            tmp := listlib::merge(sort(l),dom::repeatedIndex(jv));
		    return(new(dom,"Deriv",INDEX(jv),tmp,
			       ORDER(jv)+nops(l),tmp[1]));
	            //return(dom::newRepeated("Deriv",dom::index(jv),tmp));
		otherwise error("not defined");
	        end_case;
	    end_proc
	end_if;

    
    /* ======== */
    /* ordering */
    /* ======== */
    
    // essentially identical with the versions in Cat::DifferentialVariable
    // main difference: use of class
    // macros are used instead of methods for vartype etc
      
    orderLex :=
        proc(j1:dom,j2:dom):DOM_BOOL
        local o1:DOM_INT,o2:DOM_INT;
        begin
	    case TYPE(j2)
	    of "Indep" do
		if TYPE(j1)="Indep" then 
		    return(bool(INDEX(j1)<INDEX(j2)));
		else return(FALSE) end_if;
	    of "Dep" do
		case TYPE(j1)
		of "Indep" do return(TRUE);
		of "Dep"   do return(bool(INDEX(j1)<INDEX(j2)));
		otherwise return(FALSE);
		end_case;
	    of "Deriv" do
		if TYPE(j1)<>"Deriv" then return(TRUE) end_if;
		o1 := ORDER(j1);
		o2 := ORDER(j2);
		if o1=o2 then
		    o1 := CLASS(j1);
		    o2 := CLASS(j2);
		    if o1=o2 then
			case dom::multiLess(j1,j2)
			of 1  do return(TRUE);
			of -1 do return(FALSE);
			of 0  do return(bool(INDEX(j1)<INDEX(j2)));
			end_case;
		    else return(bool(o1<o2)) end_if;
		else return(bool(o1<o2)) end_if;
	    otherwise VAR_ERR;
	    end_case
	end_proc;


    pureLex :=
        proc(j1:dom,j2:dom):DOM_BOOL
	local o1:DOM_INT, o2:DOM_INT;
        begin
	    case TYPE(j2)
	    of "Indep" do
		if TYPE(j1)="Indep" then 
		    return(bool(INDEX(j1)<INDEX(j2)));
		else return(FALSE) end_if;
	    of "Dep" do
		case TYPE(j1)
		of "Indep" do return(TRUE);
		of "Dep"   do return(bool(INDEX(j1)<INDEX(j2)));
		otherwise return(FALSE);
		end_case;
	    of "Deriv" do
		o1 := CLASS(j1);
		o2 := CLASS(j2);
		if o1=o2 then
		    case dom::multiLess(j1,j2)
		    of 1  do return(TRUE);
		    of -1 do return(FALSE);
		    of 0  do return(bool(INDEX(j1)<INDEX(j2)));
		    end_case;
		else return(bool(o1<o2)) end_if;
	    otherwise VAR_ERR;
	    end_case;
	end_proc;
    

begin
    // if IndexedFlag=TRUE then we use multi index notation
    // otherwise repeated index notation
    IndexedFlag := bool((type(ivar)="_index") or (type(dvar)="_index"));

    IVar := ivar;               // list of indep. variables
    DVar := dvar;               // list of dep. variables
    if IndexedFlag then
        if type(ivar)<>"_index" then IVar := ivar[1] end_if;
        if type(dvar)<>"_index" then DVar := dvar[1] end_if;
        pvar := if args(0)=2 then op(DVar,1) else args(3) end_if;
        if domtype(op(IVar,1))<>DOM_IDENT or
	   domtype(op(DVar,1))<>DOM_IDENT or
	   domtype(pvar)<>DOM_IDENT then
	    error("names of variables must be identifiers");
	end_if;
	nn := op(IVar,2);
	mm := op(DVar,2);
	if nn=1 then IVar := [op(IVar,1)]
	else IVar := [op(IVar,1)[i] $ i=1..nn] end_if;
	if mm=1 then
	    DVar := [op(DVar,1)];
	    PVar := [pvar];
	else
	    DVar := [op(DVar,1)[i] $ i=1..mm];
	    PVar := [pvar[i] $ i=1..mm];
	end_if;
    else
	if args(0)<>2 then ARG_ERR end_if;
	if domtype(ivar)<>DOM_LIST then IVar := [ivar] end_if;
	if domtype(dvar)<>DOM_LIST then DVar := [dvar] end_if;
	for i in ivar do
	    if not(contains({DOM_IDENT,"_index","function"},type(i))) then
		error("invalid variable");
	    end_if;
	end_for;
        for i in dvar do
	    if not(contains({DOM_IDENT,"_index","function"},type(i))) then
		error("invalid variable");
	    end_if;
	end_for;
	nn := nops(IVar);
	mm := nops(DVar);
    end_if;

    if iszero(nn) or iszero(mm) then error("empty list of variables") end_if;
    
    TeXindex := () -> slot(_index,"TeX")(_index,
					 args(1)[args(2..args(0)-1)],
				         args(args(0)));
    TeXexp := (b,e,p) -> slot(_power,"TeX")(_power,hold(_power)(b,e),p);
end_domain:

unalias(TYPE, INDEX, DERIV_INDEX, ARG_ERR, TYPE_ERR, VAR_ERR):
