/*++
Dom::ExteriorAlgebra(R,S) -- the exterior algebra over a vector space with
        a basis given by the ordered set S (can also be a list; the ordering
        is then given by the position in the list) and coefficients in the
        ring R.

The implementation is based on Dom::FreeModule(R,Dom::ExteriorProduct(S)) from
which all arithmetic functions and the representation are inherited.  The
representation is extended by a flag signalling whether or not an exterior form
is normalized.

Additional operations are:
    normal :       normalizes all base forms and coefficients
    isnormalized : checks whether an exterior form is normalized
    wedge :        alternative to _mult (used in output)
    variables :    all variables from S apparing in wedge products


Author: Werner M. Seiler, Uni Karlsruhe/Mannheim 
	(wms@ira.uka.de, seiler@euler.math.uni-mannheim.de)
Created: August 4th, 1997
Last update: 	Fri Nov 17 11:39:01 2000
++*/

alias(REP(x)=extop(x,1),
      NORMAL(x)=extop(x,2),
      TYPE_ERR=error("wrong type of argument")):


domain Dom::ExteriorAlgebra(R, S)

    local Riszero:DOM_PROC, Risone:DOM_PROC, Rone:R, Rplus:DOM_PROC, 
	  Rmult:DOM_PROC, Rnegate:DOM_PROC, Rexpr:DOM_PROC, Rcoerce:DOM_PROC,
          EPS:DOM_DOMAIN, EPSisnormalized:DOM_PROC, EPSnormal:DOM_PROC, 
	  EPSsign:DOM_PROC, EPSiszero:DOM_PROC, EPSwedge:DOM_PROC,
          EPSnegate:DOM_PROC, EPSequal:DOM_PROC, EPSexpr:DOM_PROC, 
	  EPSone:DOM_PROC, EPSless:DOM_PROC, EPSvars:DOM_PROC,
	  EPSisone:DOM_PROC,
          remDup:DOM_PROC, merge:DOM_PROC, dim:DOM_INT;
    inherits Dom::FreeModule(R,EPS);
    category Cat::LeftModule(R), Cat::Ring;


    /* ========================= */
    /* representation, in/output */
    /* ========================= */
    
    REPSconv :=
	proc(a)
	local tmp;
	begin
	    tmp := Rcoerce(a);
	    if tmp=FAIL then
		tmp := dom::coerce(a);
		if tmp=FAIL then TYPE_ERR end_if;
	    end_if;
	    tmp;
	end_proc;
	
	
    swapSign := 
	proc(l:DOM_LIST):DOM_LIST
	begin
	    if EPSsign(l[2])=-1 then [ Rnegate(l[1]), EPSnegate(l[2]) ]
	    else l end_if;
	end_proc;


    generate1 := 
	proc(r:R,s:EPS):dom
	begin
	    if EPSisnormalized(s) then new(dom, [dom::swapSign([r,s])], TRUE)
	    else new(dom, [dom::swapSign([r,s])], FALSE) end_if;
	end_proc;


    generate := 
	proc(l:DOM_LIST):dom
	begin new(dom, map(l,dom::swapSign), FALSE) end_proc;


    normalGenerate := 
	proc(l:DOM_LIST):dom
	begin new(dom,l,TRUE) end_proc;


    isnormalized := 
	proc(x:dom):DOM_BOOL
	begin NORMAL(x) end_proc;


    normal := 
	proc(x:dom):dom
        option escape;
	local li:DOM_LIST, res:DOM_LIST, i:DOM_INT, i1:DOM_INT, k:DOM_INT;
	begin
	    if NORMAL(x) then return(x) end_if;
	    res := [ [] $ nops(REP(x)) ];
	    k := 0;
	    for li in REP(x) do
		li[2] := EPSnormal(li[2]);
		if not(EPSiszero(li[2])) then
		    k := k+1;
		    if EPSsign(li[2])=-1 then
			res[k] := [ Rnegate(li[1]), EPSnegate(li[2]) ];
		    else res[k] := li end_if;
		end_if;
	    end_for;
	    res := sort([op(res,1..k)], ((l1,l2) -> EPSless(l2[2],l1[2])));
	    i := 1;
	    while i<k do
		i1 := i+1;
		if EPSequal(res[i][2],res[i1][2]) then
		    res[i][1] := Rplus(res[i][1],res[i1][1]);
		    delete res[i1];
		    k := k-1;
		else i := i1 end_if
	    end_while;
	    dom::normalGenerate(select(res, ((l) -> not(Riszero(l[1])))));
	end_proc;


    exprMonomial := 
	proc(m:DOM_LIST):Type::Arithmetical
	local r:R, s:EPS;
	begin
	    r := m[1];
	    s := m[2];
	    if EPSsign(s)=-1 then
		r := Rnegate(r);
		s := EPSnegate(s);
	    end_if;
	    if Risone(r) then EPSexpr(s)
	    elif EPSisone(s) then Rexpr(r)
	    else _mult(Rexpr(r),EPSexpr(s)) end_if;
	end_proc;


    variables := 
	if iszero(dim) then
	    proc(x:dom):DOM_LIST
	    local l:DOM_LIST, ls:DOM_LIST, res:DOM_LIST;
	    begin
		l := dom::terms(x);
		if l=[] then return([]) end_if;
		l := map(l,EPSvars);
		res := [];
		for ls in l do
		    res := remDup(merge(res,ls,EPSless));
		end_for;
		res;
	    end_proc;
	else
	    proc(x:dom):DOM_LIST
	    local l:DOM_LIST, ls:DOM_LIST, res:DOM_LIST;
	    begin
		l := dom::terms(x);
		if l=[] then return([]) end_if;
		l := map(l,EPSvars);
		res := [];
		for ls in l do
		    res := remDup(merge(res,ls,EPSless));
		    if nops(res)=dim then return(res) end_if;
		end_for;
		res;
	    end_proc;
	end_if;


    /* =========== */
    /* arithmetics */
    /* =========== */

    one := dom::generate1(Rone,EPSone);


    // Opposed to Dom::ExteriorProduct we must distinguish here between wedge
    // and _mult, as _mult must also handle the scalar multiplication.
    _mult := 
	proc():dom
	local al:DOM_LIST, tmp:DOM_LIST, w:dom;
	begin
	    if iszero(args(0)) then dom::one
	    elif args(0)=1 then args(1)
	    else 
		al := map([args()],dom::REPSconv);
	    end_if;
	    tmp := split(al, ((a,EA) -> bool(domtype(a)=EA)), dom);
	    if nops(tmp[1])=1 then w := tmp[1][1]
	    else w := dom::wedge(op(tmp[1])) end_if;
	    case nops(tmp[2])
	    of 0 do return(w);
	    of 1 do return(dom::smult(tmp[2][1],w));
	    otherwise return(dom::smult(Rmult(op(tmp[2])),w));
	    end_case;
	end_proc;


    wedge := 
	proc():dom
	local n:DOM_INT, m:DOM_INT;
	begin
	    n := args(0);
	    case n
	    of 0 do return(dom::one);
	    of 1 do return(args(1));
	    of 2 do return(dom::wedge2(args(1),args(2)));
	    otherwise
		m := n div 2;
		dom::wedge2(dom::wedge(args(1..m)),dom::wedge(args(m+1..n)));
	    end_case;
	end_proc;


    wedge2 := 
	proc(w1:dom,w2:dom):dom
	local res:DOM_LIST;
	begin
	    res := map(REP(w2),
		       ((li,l) -> op(map(l,
				         ((l1,l2) -> [Rmult(l1[1],l2[1]),
						      EPSwedge(l1[2],l2[2])]),
				         li))),
		       REP(w1));
	    res := select(res, ((li) -> not(EPSiszero(li[2]))));
	    if not(R::hasProp(Ax::noZeroDivisors)) then
		res := select(res, ((li) -> not(Riszero(li[1]))));
	    end_if;
	    return(dom::generate(res));
	end_proc;


    // This is essentially the same hack as in Dom::ExteriorProduct
    // using the fact that ^ is translated by MuPAD into _power. But
    // now we must be more careful, as here it could really happen
    // that a power is meant!
    _power :=
	proc(x:dom,e):dom
	local tmp;
	begin
	    case domtype(e)
	    of DOM_INT do 
		if e>1 then return(dom::power(x,e))
		elif e=1 then return(x)
		elif iszero(e) then return(dom::one)
		else error("inverse not defined") end_if;
	    of dom do return(dom::wedge2(x,e));
	    of R do return(dom::smult(e,x));
	    otherwise 
		tmp := Rcoerce(e);
		if tmp=FAIL then
		    tmp := dom::coerce(e);
		    if tmp=FAIL then TYPE_ERR end_if;
		    return(dom::wedge2(x,e));
		else return(dom::smult(e,x)) end_if;
	    end_case;	    
	end_proc;


    power := domains::repeatedSquaring(dom,"wedge2");


    equal := 
	proc(w1:dom,w2:dom):DOM_BOOL
	begin
	    if not(NORMAL(w1)) then dom::equal(dom::normal(w1),w2)
	    elif not(NORMAL(w2)) then dom::equal(w1,dom::normal(w2))
	    else bool(REP(w1)=REP(w2)) end_if;
	end_proc;


    /* ============= */
    /* miscellaneous */
    /* ============= */

    // Most of the following methods are copied from Dom::FreeModule. Special
    // versions are needed in order to keep information on the normalization
    // of the argument.

    _plus := 
	proc():dom
	local m:DOM_INT, n:DOM_INT;
	begin
	    n := args(0);
	    case n
	    of 0 do return(dom::zero);
	    of 1 do return(args(1));
	    of 2 do return(dom::plus2(args(1),args(2)));
	    otherwise
		m := n div 2;
		dom::plus2(dom::_plus(args(1..m)),dom::_plus(args((m+1)..n)));
	    end_case;
	end_proc;


    plus2 := 
	proc(x1:dom,x2:dom):dom
	local i:DOM_INT, l1:DOM_LIST, l2:DOM_LIST, tmp:R, res:DOM_LIST;
	begin
	    l1 := REP(x1);
	    l2 := REP(x2);
            res := [ 0 $ nops(l1)+nops(l2) ];
	    i := 0;
            while (l1<>[]) and (l2<>[]) do
		if EPSless(l2[1][2],l1[1][2]) then
		    i := i+1;
                    res[i] := l1[1];
                    delete l1[1];
                elif l1[1][2]=l2[1][2] then
                    tmp := Rplus(l1[1][1],l2[1][1]);
                    if not Riszero(tmp) then
			i := i+1;
                        res[i] := [tmp,l1[1][2]];
                    end_if;
                    delete l1[1];
                    delete l2[1];
                else
		    i := i+1;
                    res[i] := l2[1];
                    delete l2[1];
                end_if;
            end_while;
            if (l1<>[]) then res := [op(res,1..i)] . l1
            elif (l2<>[]) then res := [op(res,1..i)] . l2
	    else res := [op(res,1..i)]
            end_if;
	    dom::generate(res);
	end_proc;


    reductum := 
	proc(x:dom):dom
	local l:DOM_LIST;
	begin
	    case dom::nterms(x)
	    of 0 do error("not defined for zero");
	    of 1 do return(dom::zero);
	    otherwise
		l := REP(x);
		delete l[1];
		return(dom::generate(l));
	    end_case;
	end_proc;


    mapcoeffs := 
	proc(x:dom,f:DOM_PROC):dom
	local res:DOM_LIST;
	begin
	    if dom::iszero(x) then return(dom::zero) end_if;
	    res := select(map(REP(x),
			      ((li,f) -> subsop(li,
						1=f(li[1],args(3..args(0))))),
			      f,args(3..args(0))),
			  ((li) -> not(Riszero(li[1]))));
	    dom::generate(res);
	end_proc;


    characteristic := R::characteristic;


begin

    if domtype(R)=DOM_DOMAIN and (R::hasProp<>FAIL) and
       R::hasProp(Cat::Ring) then
	Riszero := R::iszero;
	Risone := R::isone;
	Rone := R::one;
	Rplus := R::_plus;
	Rmult := R::_mult;
	Rnegate := R::_negate;
	Rcoerce := R::coerce;
	Rexpr := R::expr;
    else error("first argument must be Cat::Ring") end_if;
    if (domtype(S)=DOM_LIST) or
       ((domtype(S)=DOM_DOMAIN) and (S::hasProp<>FAIL) and
       (S::hasProp(Cat::OrderedSet))) then
	EPS := Dom::ExteriorProduct(S);
	EPSisnormalized := EPS::isnormalized;
	EPSnormal := EPS::normal;
	EPSsign := EPS::sign;
	EPSiszero := EPS::iszero;
	EPSwedge := EPS::wedge;
	EPSnegate := EPS::_negate;
	EPSequal := EPS::equal;
	EPSexpr := EPS::expr;
	EPSone := EPS::one;
	EPSisone := EPS::isone;
	EPSless := EPS::_less;
	EPSvars := EPS::variables;
    else error("second argument must be Cat::OrderedSet or list") end_if;

    remDup := listlib::removeDuplicates;
    merge := listlib::merge;
    if domtype(S)=DOM_LIST then dim := nops(S)
    else dim := 0 end_if;

end_domain:


unalias(REP,NORMAL,TYPE_ERR):
