/*++
Dom::ExteriorProduct(S) -- monoid of exterior products over the ordered
	set S of variables (S can also be a list; the ordering of the variables
	is then determined by the position in the list).

Dom::ExteriorProduct(S,order) defines a Cat::OrderedMonoid.  The
representation consists of a list of elements of S, a sign coefficient which
is either -1, 0 or 1 and a flag signalling whether the list is normalized.  If
the list is normalized, the entries are sorted according to the ordering in S.
Equality etc. can be tested only after normalization. 

Besides the monoid operations there are the following methods implemented:
    normal :       normalizes an element
    isnormalized : checks whether an element is normalized
    degree :       form degree (number of factors in product)
    variables :    list of all variables
    sign :         sign of an element
    zero :         zero element
    iszero :       checks whether an element is zero
    wedge :        alternative to _mult (used in output)


Author: Werner M. Seiler, Uni Karlsruhe/Mannheim
	(wms@ira.uka.de,seiler@euler.math.uni-mannheim.de)
Created: July 31st, 1997
Last update: 	Tue Feb 15 14:28:32 2000
++*/


alias(REP(x)=extop(x,1),
      SIGN(x)=extop(x,2),
      NORM(x)=extop(x,3)):


domain Dom::ExteriorProduct(S)
    local dim:DOM_INT,Sexpr:DOM_DOMAIN,Scoerce:DOM_DOMAIN,Srandom:DOM_DOMAIN;
    inherits Dom::BaseDomain;
    category Cat::OrderedMonoid;


    /* ================= */
    /* "local" functions */
    /* ================= */

    Sless := 
	if iszero(dim) then S::_less
	else
	   proc(a,b):DOM_BOOL
	   begin bool(contains(S,a)<contains(S,b)) end_proc;
	end_if;


    // total degree, lexicographic ordering
    less := 
	proc(x1:dom,x2:dom):DOM_BOOL
	local l1:DOM_LIST,l2:DOM_LIST,n1:DOM_INT,n2:DOM_INT;
	begin
	    l1 := REP(x1);
	    l2 := REP(x2);
	    n1 := nops(l1);
	    n2 := nops(l2);
	    if n1<n2 then return(TRUE)
	    elif n1>n2 then return(FALSE)
	    else
		while l1<>[] do
		    if dom::Sless(l1[1],l2[1]) then return(TRUE)
		    elif l1[1]=l2[1] then
			delete l1[1];
			delete l2[1];
		    else return(FALSE) end_if;
		end_while;
	    end_if;
	    bool(SIGN(x1)<SIGN(x2));
	end_proc;


    // checks for non-empty intersection between two lists
    overlap := 
	proc(l1:DOM_LIST,l2:DOM_LIST):DOM_BOOL
	local i;
	begin
	    for i in l1 do
		if not(iszero(contains(l2,i))) then return(TRUE) end_if;
	    end_for;
	    FALSE;
	end_proc;


    // checks whether all elements of a list are elements of another one
    subset := 
	proc(l1:DOM_LIST,l2:DOM_LIST):DOM_BOOL
	local i;
	begin
	    for i in l1 do
		if iszero(contains(l2,i)) then return(FALSE) end_if;
	    end_for;
	    TRUE;
	end_proc;
	
	
    // checks whether some element of a list occurs more than once
    hasDuplicates :=
	proc(l:DOM_LIST):DOM_BOOL
	begin bool(nops({op(l)})<nops(l)) end_proc;


    /* ========================= */
    /* representation, in/output */
    /* ========================= */

    convert := 
	if iszero(dim) then
	    proc(x)
	    local tmp,e;
	    begin
		case domtype(x)
		of dom do return(x);
		of S do return(new(dom,[x],1,FALSE));
		of DOM_LIST do
		    if testtype(x,Type::ListOf(S)) then 
			if dom::hasDuplicates(x) then return(dom::zero)
			else return(new(dom,x,1,FALSE)) end_if;
		    else return(FAIL) end_if;
		of DOM_EXPR do
		    x := expand(x);
		    case type(x)
		    of "_power" do
			tmp := dom::coerce(op(x,1));
			if tmp=FAIL then return(FAIL) end_if;
			e := op(x,2);
			if testtype(e,Type::PosInt) then return(dom::zero)
			else
			    e := dom::coerce(e);
			    if e=FAIL then return(FAIL)
			    else dom::wedge(tmp,e) end_if;
			end_if;
		    of "_mult" do
			tmp := map([op(x)],Scoerce);
			if iszero(contains(tmp,FAIL)) then
			    if dom::hasDuplicates(tmp) then return(dom::zero)
			    else return(new(dom,tmp,1,FALSE)) end_if;
			elif (nops(x)=2) and (op(x,2)=-1) then
			    tmp := dom::coerce(op(x,1));
			    if tmp=FAIL then return(FAIL)
			    else return(dom::_negate(tmp)) end_if;
			else return(FAIL) end_if;
		    of "_negate" do
			tmp := dom::convert(op(x,1));
			if tmp<>FAIL then return(dom::_negate(tmp))
			else return(FAIL) end_if;
		    of "function" do
			if op(x,0)=hold(wedge) then
			    tmp := map([op(x)],Scoerce);
			    if iszero(contains(tmp,FAIL)) then
				if dom::hasDuplicates(tmp) then 
				    return(dom::zero);
				else return(new(dom,tmp,1,FALSE)) end_if;
			    else return(FAIL) end_if;
			else return(FAIL) end_if;
		    otherwise
			if iszero(x) then return(dom::zero)
			elif x=1 then return(dom::one)
			else return(FAIL) end_if;
		    end_case;
		otherwise
		    tmp := Scoerce(x);
		    if tmp=FAIL then return(FAIL)
		    else return(new(dom,[tmp],1,TRUE));
		    end_if;
		end_case;
	    end_proc;
	else
	    proc(x)
	    local tmp,e;
	    begin
		case domtype(x)
		of dom do return(x);
		of DOM_LIST do
		    if dom::subset(x,S) then
			if dom::hasDuplicates(x) then return(dom::zero)
			else return(new(dom,x,1,FALSE)) end_if;
		    else return(FAIL) end_if;
		of DOM_EXPR do
		    x := expand(x);
		    case type(x)
		    of "_power" do
			tmp := dom::coerce(op(x,1));
			if tmp=FAIL then return(FAIL) end_if;
			e := op(x,2);
			if testtype(e,Type::PosInt) then return(dom::zero)
			else
			    e := dom::coerce(e);
			    if e=FAIL then return(FAIL)
			    else return(dom::wedge(tmp,e)) end_if;
			end_if;
		    of "_mult" do
			tmp := [op(x)];
			if dom::subset(tmp,S) then
			    if dom::hasDuplicates(tmp) then return(dom::zero)
			    else return(new(dom,tmp,1,FALSE)) end_if;
			elif (nops(tmp)=2) and (tmp[2]=-1) then
			    tmp := dom::convert(tmp[1]);
			    if tmp<>FAIL then return(dom::_negate(tmp))
			    else return(FAIL) end_if;
			else return(FAIL) end_if;
		    of "_negate" do
			tmp := dom::convert(op(x,1));
			if tmp<>FAIL then return(dom::_negate(tmp))
			else return(FAIL) end_if;
		    of "function" do
			if op(x,0)=hold(wedge) then
			    tmp := [op(x)];
			    if dom::subset(tmp,S) then
				if dom::hasDuplicates(tmp) then 
				    return(dom::zero);
				else return(new(dom,tmp,1,FALSE)) end_if;
			    else return(FAIL) end_if;
			else return(FAIL) end_if;
		    otherwise
			if iszero(contains(S,x)) then return(FAIL)
			else return(new(dom,[x],1,FALSE)) end_if;
		    end_case;
		otherwise
		    if iszero(x) then return(dom::zero)
		    elif x=1 then return(dom::one)
		    elif iszero(contains(S,x)) then return(FAIL)
		    else return(new(dom,[x],1,FALSE)) end_if;
		end_case;
		FAIL;
	    end_proc;
	end_if;


    convert_to :=
	proc(x:dom,T)
	begin
	    case T
	    of DOM_LIST do return(REP(x))
	    of dom do return(x);
	    of DOM_EXPR do return(dom::expr(x));
	    otherwise return(FAIL);
	    end_case;
	end_proc;


    wedgeexpr := 
	if iszero(dim) then
	    proc(l:DOM_LIST):Type::Arithmetical
	    begin
		if nops(l)=1 then Sexpr(l[1])
		else hold(wedge)(op(map(l,Sexpr))) end_if;
	    end_proc;
	else
	    proc(l:DOM_LIST):Type::Arithmetical
	    begin
		if nops(l)=1 then l[1]
		else hold(wedge)(op(l)) end_if;
	    end_proc;
	end_if;


    expr := 
	proc(x:dom):Type::Arithmetical
	begin
	    if REP(x)=[] then SIGN(x)
	    elif SIGN(x)=1 then dom::wedgeexpr(REP(x))
	    else _negate(dom::wedgeexpr(REP(x))) end_if;
	end_proc;


    isnormalized := 
	proc(x:dom):DOM_BOOL
	begin NORM(x) end_proc;


    normal := 
	proc(x:dom):dom
	local c:DOM_INT,l:DOM_LIST,n:DOM_INT,i:DOM_INT,tmp,flag:DOM_BOOL;
	begin
            if NORM(x) then return(x) end_if;
            c := SIGN(x);
            if iszero(c) then return(dom::zero) end_if;
            l := REP(x);
            n := nops(l)-1;
            flag := TRUE;

            while flag do		// bubble sort ?!?
                flag := FALSE;
                for i from 1 to n do
                    if dom::Sless(l[i+1],l[i]) then
                        flag := TRUE;
                        c := -c;
                        tmp := l[i];
                        l[i] := l[i+1];
                        l[i+1] := tmp;
                    end_if;
                end_for;
            end_while;

            new(dom,l,c,TRUE);
	end_proc;


    sign := 
	proc(x:dom):DOM_INT
	begin SIGN(x) end_proc;


    _less :=
	proc(x1:dom,x2:dom):DOM_BOOL
	begin
	    if not(NORM(x1)) then
		return(dom::_less(dom::normal(x1),x2))
	    elif not(NORM(x2)) then
		return(dom::_less(x1,dom::normal(x2)))
	    else dom::less(x1,x2) end_if;
	end_proc;


    equal := 
	proc(x1:dom,x2:dom):DOM_BOOL
	begin
	    if not(NORM(x1)) then dom::equal(dom::normal(x1),x2)
	    elif not(NORM(x2)) then dom::equal(x1,dom::normal(x2))
	    else bool((REP(x1)=REP(x2)) and (SIGN(x1)=SIGN(x2))) end_if;
	end_proc;
	
	
    has :=
	if iszero(dim) then
	    proc(x:dom,s):DOM_BOOL
	    begin
		if domtype(s)=S then not(iszero(contains(REP(x),s)))
		else FALSE end_if;
	    end_proc;
	else
	    proc(x:dom,s):DOM_BOOL
	    begin
		if iszero(contains(S,s)) then FALSE
		else not(iszero(contains(REP(x),s))) end_if;
	    end_proc;
	end_if;


    /* =========== */
    /* arithmetics */
    /* =========== */

    zero := new(dom,[],0,TRUE);


    iszero := 
	proc(x:dom):DOM_BOOL
	begin
	    if NORM(x) then iszero(SIGN(x))
	    else dom::iszero(dom::normal(x)) end_if;
	end_proc;


    one := new(dom,[],1,TRUE);


    isone := 
	proc(x:dom):DOM_BOOL
	begin
	    if NORM(x) then bool(REP(x)=[] and SIGN(x)=1)
            else dom::isone(dom::normal(x)) end_if;
	end_proc;


    _invert := proc() begin FAIL end_proc;


    _negate := 
	proc(x:dom):dom
	begin new(dom,REP(x),-SIGN(x),NORM(x)) end_proc;


    wedge := 
	proc():dom
	local n:DOM_INT,m:DOM_INT;
	begin
	    n := args(0);
	    case n
	    of 0 do return(dom::one);
	    of 1 do return(args(1));
	    of 2 do
		if dom::iszero(args(1)) or dom::iszero(args(2)) then
		    return(dom::zero);
		elif dom::overlap(REP(args(1)),REP(args(2))) then
		    return(dom::zero);
		else
		    return(new(dom, REP(args(1)).REP(args(2)),
                               SIGN(args(1))*SIGN(args(2)), FALSE))
		end_if;
	    otherwise
		m := n div 2;
		dom::wedge(dom::wedge(args(1..m)),dom::wedge(args(m+1..n)));
	    end_case;
	end_proc;


    _mult := dom::wedge;


    // The following is really a hack exploiting the fact that in MuPAD
    // the symbol for _power is a kind of a wedge, namely ^. 
    _power := dom::wedge;


    // fast left multiplication with element of S
    lmult := 
	proc(s,x:dom):dom
	begin
	    if iszero(contains(REP(x),s)) then 
		new(dom, [s].REP(x), SIGN(x), FALSE);
	    else dom::zero end_if;
	end_proc;


    /* ============= */
    /* miscellaneous */
    /* ============= */

    degree := 
	proc(x:dom):DOM_INT
	begin nops(REP(x)) end_proc;


    variables := 
	proc(x:dom):DOM_LIST
	begin 
	    if NORM(x) then REP(x) else sort(REP(x),dom::Sless) end_if;
	end_proc;


    TeX := 
	proc(x:dom,p:DOM_INT):DOM_STRING
	local res:DOM_STRING;
	begin 
	    if REP(x)=[] then expr2text(SIGN(x))
	    else
		res :=generate::TeXoperator("\\wedge ",p,150,op(REP(x)));
		if SIGN(x)=1 then res
		else generate::TeXquote("- ".res,p,120)	end_if;
	    end_if;
	end_proc;


    random := 
	if not(iszero(dim)) then
	    proc():dom
	    local i:DOM_INT,j:DOM_INT,r:DOM_PROC,res:DOM_LIST;
	    begin
		r := random(1..dim);
		res := [];
		for i from 1 to r() do
		    j := r();
		    if iszero(contains(res,S[j])) then
			res := append(res,S[j])
		    end_if;
		end_for;
		if nops(res)>1 then new(dom,res,1,FALSE)
		else new(dom,[],1,TRUE) end_if;
	    end_proc;
	elif Srandom<>FAIL then
	    proc(N=10:DOM_INT):dom
	    local i:DOM_INT,r:DOM_PROC,s:S,res:DOM_LIST;
	    begin
		r := random(1..N);
		res := [];
		for i from 1 to r() do
		    s := Srandom();
		    if iszero(contains(res,s)) then 
			res := append(res,s); 
		    end_if;
		end_for;
		if nops(res)>1 then new(dom,res,1,FALSE)
		else new(dom,[],1,TRUE) end_if;
	    end_proc;
	else FAIL end_if;

begin

    if args(0)<>1 then error("wrong number of arguments") end_if;
    if domtype(S)=DOM_LIST then
	dim := nops(S);
	if iszero(dim) then error("empty list") end_if;
    elif domtype(S)=DOM_DOMAIN and (S::hasProp<>FAIL) and
	 S::hasProp(Cat::OrderedSet) then
	dim := 0;
	Sexpr := S::expr;
	Scoerce := S::coerce;
	Srandom := S::random;
    else error("argument must be ordered set or list") end_if;
     
end_domain:


unalias(REP,NORM,SIGN):
