/*++
Dom::FreeModule(R,S)  -- free module over the ring R with (finite) support
        in the ordered set S (S can also be a list; the ordering is then
	defined by the position in the list)

An element is represented as a domain element with operands:

0 - domain Dom::FreeModule(R,S)
1 - a list [[r1,s1], ...,[rn,sn]] of lists of one element of R and one
    element of S, with s1 > s2 > ... > sn

Elements are generated exclusively with the two methods generate and generate1.
The first one expects as argument a list as described above; generate1 takes as
arguments one element of R and one of S. Domains based on Dom::FreeModule which
want to extend the representation by further operands (computable from the
list) can thus simply overwrite these two methods and then still inherit all
other methods implemented here.

The provided methods are essentially those of a Cat::FiniteMonoidRing. Some
methods like _mult, _divide, convert, convert_to or expr are implemented under
different names, so that the implementation can be more easily inherited by
other domains build on Dom::FreeModule. For example, scalar multiplication is
called smult and _mult is identical to smult. If now an algebra is constructed
upon Dom::FreeModule, it will have to redefine _mult. But it can still make use
of the here implemented scalar multiplication by calling smult.

If S has a method degree or degreevec, then they are lifted to Dom::FreeModule.
If there exists methods has and subs in S (or if S is a list), then a subs
method is implemented for Dom::FreeModule. S::has must return TRUE whenever
S::subs can perform a substitution. The provided implementation of subs
supports all calls of the standard MuPAD function subs. However, the option
Unsimplified has no effect and only substitutions in/for elements of S are
possible.


Authors: Stefan Wehmeier, Uni Paderborn
         Werner M. Seiler, Uni Karlsruhe/Mannheim
         (wms@ira.uka.de, seiler@euler.math.uni-mannheim.de)
Created: June 30th, 1997
Last update: 	Tue Feb 15 14:04:09 2000
++*/
// 13.10.2000


alias(REP(x)=extop(x,1),
      TYPE_ERR=error("wrong type of argument"),
      INT_ERR=error("invalid index")):


domain Dom::FreeModule(R,S)

    local Rzero:R, Riszero:DOM_PROC, Rconvto:DOM_PROC, Rone:R, Risone:DOM_PROC,
	  Rcoerce:DOM_PROC, Rexpr:DOM_PROC, Rplus:DOM_PROC, Rmult:DOM_PROC, 
	  Rnegate:DOM_PROC, Rdivide:DOM_PROC, Rlength:DOM_PROC,
	  Sone, Sisone, Sexpr, Sconvto, Sdegree, Sdegreevec, Smult, Slength,
	  maxr:DOM_INT, finite:DOM_BOOL;
    inherits Dom::BaseDomain;
    category Cat::LeftModule(R), 
	     (if R::hasProp(Cat::Field) then Cat::VectorSpace(R) end_if);
    axiom (if R::hasProp(Ax::canonicalRep) then Ax::canonicalRep end_if),
          (if R::hasProp(Ax::normalRep) then Ax::normalRep end_if);


    /* =============== */
    /* local functions */
    /* =============== */

    Srandom :=
        if finite then
	    proc()
	    local r:DOM_PROC;
	    begin
	        r := random(1..nops(S));
	        S[r()];
	    end_proc;
	else 
	    proc():S
	    begin S::random() end_proc;
	end_if;


    Scoerce :=
        if finite then
	    proc(x)
	    begin if iszero(contains(S,x)) then FAIL else x end_if end_proc;
	else 
	    proc()
	    begin S::coerce(args()) end_proc; 
	end_if;


    Shas :=
        if finite then _equal 
	else 
	    proc():DOM_BOOL
	    begin S::has(args()) end_proc;
	end_if;


    Ssubs :=
        if finite then
            proc(x,y)
            begin if x=op(y,1) then op(y,2) else x end_if end_proc;
        else 
	    proc():S
	    begin S::subs(args()) end_proc;
	end_if;


    greater2 :=
        if finite then
	    proc(x:DOM_LIST,y:DOM_LIST):DOM_BOOL
	    begin bool(contains(S,x[2])<contains(S,y[2])) end_proc;
	else
	    proc(x:DOM_LIST,y:DOM_LIST):DOM_BOOL
	    begin S::_less(y[2],x[2]) end_proc;
	end_if;


    // removes multiple occurrence of generators
    //    x - list of lists of the form [r,e]
    normalizeRep :=
        proc(x:DOM_LIST):DOM_LIST
        local i:DOM_INT;
        begin
            if x=[] then return([]) end_if;
            x := sort(x,dom::greater2);
            i := 1;
            while i<nops(x) do
	        if x[i][2]=x[i+1][2] then
		    x[i][1] := Rplus(x[i][1],x[i+1][1]);
		    delete x[i+1];
	        else i:=i+1 end_if;
            end_while;
            select(x, ((l) -> not(Riszero(l[1]))));
        end_proc;


    coerceSummand :=
        proc(l:DOM_LIST):DOM_LIST
        begin [Rcoerce(l[1]),dom::Scoerce(l[2])] end_proc;


    /* ========================= */
    /* representation, in/output */
    /* ========================= */

    generate :=
        proc(x:DOM_LIST):dom
        begin new(dom,x) end_proc;


    generate1 :=
        proc(x:R,y):dom
        begin new(dom,[[x,y]]) end_proc;


    moduleConvert :=
        proc(x)
        local r,s,f,res,rres,sres,ures;
        begin
	    if Rone<>FAIL then
	        s := dom::Scoerce(x);
	        if s<>FAIL then return(dom::generate1(Rone,s)) end_if;
	    end_if;
	    case domtype(x)
	    of dom do return(x);
	    of DOM_LIST do
	        if nops(x)=2 and (r:=Rcoerce(x[1]))<>FAIL and
			         (s:=dom::Scoerce(x[2]))<>FAIL then
		    if Riszero(r) then return(dom::zero)
		    else return(dom::generate1(r,s)) end_if;
	        elif {op(map(x,domtype))}<>{DOM_LIST} then return(FAIL)
	        elif {op(map(x,nops))}<>{2} then return(FAIL)
	        else
		    x := map(x,dom::coerceSummand);
		    if has(x,FAIL) then return(FAIL);
		    else return(dom::generate(dom::normalizeRep(x))) end_if;
	        end_if;
	    of DOM_EXPR do
	        x := expand(x);
	        case type(x)
	        of "_plus" do
		    res := dom::zero;
		    for r in op(x) do
		        s := dom::convert(r);
		        if s=FAIL then return(FAIL) end_if;
		        res := dom::_plus(res,s);
		    end_for;
		    return(res);
	        of "_mult" do
		    rres := [];
		    sres := [];
		    ures := [];
		    for f in op(x) do
		        s := dom::Scoerce(f);
		        if s<>FAIL then
			    sres := append(sres,s);
			    next;
		        end_if;
		        r := Rcoerce(f);
		        if r<>FAIL then
			    rres := append(rres,r);
			    next;
		        end_if;
		        ures := append(ures,f);
		    end_for;
		    if ures<>[] then
		        f := hold(_mult)(op(ures));
		        s := dom::Scoerce(f);
		        if s<>FAIL then sres := append(sres,s)
		        else
			    r := Rcoerce(f);
			    if r<>FAIL then rres := append(rres,r)
			    else return(FAIL) end_if;
		        end_if;
		    end_if;
		    if rres=[] then
		        if Rone=FAIL then return(FAIL)
		        else rres := Rone end_if;
		    else rres := Rmult(op(rres)) end_if;
		    if sres=[] then
		         if Sone=FAIL then return(FAIL)
		         else sres := Sone end_if;
		    elif nops(sres)=1 then sres := sres[1]
		    else
		        if Smult<>FAIL then sres := Smult(op(sres))
		        else return(FAIL) end_if;
		    end_if;
		    return(dom::generate1(rres,sres));
		otherwise
		    if Sone<>FAIL then
			r := Rcoerce(x);
			if r<>FAIL then
			    if Riszero(r) then return(dom::zero)
			    else return(dom::generate1(r,Sone)) end_if;
			end_if;
		    else return(FAIL) end_if;
	        end_case;
	    of DOM_POLY do return(dom::convert(expr(x)));
	    otherwise
	        if x=0 then return(dom::zero) end_if;
	        if x=1 and dom::one<>FAIL then return(dom::one) end_if;
	        if Sone<>FAIL then
	            r := Rcoerce(x);
	            if r<>FAIL then
		        if Riszero(r) then return(dom::zero)
		        else return(dom::generate1(r,Sone)) end_if;
	            end_if;
	        else return(FAIL) end_if;
	    end_case;
        end_proc;


    convert := dom::moduleConvert;


    moduleConvertTo :=
        proc(x:dom,T)
        begin
            case T
	    of dom do return(x);
	    of DOM_LIST do return(REP(x));
	    of R do
	        if (domtype(S)<>DOM_LIST) and
	           (S::hasProp(Cat::Monoid) or 
		    S::hasProp(Cat::AbelianMonoid)) then
		    if dom::nterms(x)>1 then return(FAIL)
		    elif dom::iszero(x) then return(Rzero)
		    elif Sisone(REP(x)[1][2]) then return(REP(x)[1][1])
		    else return(FAIL) end_if;
	         else return(FAIL) end_if;
	         break;
	    of S do
	        if Risone=FAIL then return(FAIL) end_if;
	        if dom::nterms(x)=1 and Risone(REP(x)[1][1]) then
		    return(REP(x)[1][2])
	        else return(FAIL) end_if;
	    of DOM_EXPR do return(dom::expr(x,DOM_EXPR));
	    otherwise FAIL;
	    end_case;
	    FAIL;
	end_proc;


    convert_to := dom::moduleConvertTo;


    exprMonomial :=
        if Risone=FAIL then
	    if Sisone=FAIL then
	        proc(x:DOM_LIST,T):Type::Arithmetical
	        begin 
		    if T=FAIL then _mult(Rexpr(x[1]),Sexpr(x[2]))
		    else _mult(Rconvto(x[1],T),Sconvto(x[2],T)) end_if;
		end_proc;
	    else
		proc(x:DOM_LIST,T):Type::Arithmetical
		begin
		    if T=FAIL then
			if Sisone(x[2]) then Rexpr(x[1])
			else _mult(Rexpr(x[1]),Sexpr(x[2])) end_if;
		    else
			if Sisone(x[2]) then Rconvto(x[1],T)
			else _mult(Rconvto(x[1],T),Sconvto(x[2],T)) end_if;
		    end_if;
		end_proc;
	    end_if
	elif Sisone=FAIL then
	    proc(x:DOM_LIST,T):Type::Arithmetical
	    begin
		if T=FAIL then
		    if Risone(x[1]) then Sexpr(x[2])
		    else _mult(Rexpr(x[1]),Sexpr(x[2])) end_if;
		else
		    if Risone(x[1]) then Sconvto(x[2],T)
		    else _mult(Rconvto(x[1],T),Sconvto(x[2],T)) end_if;
		end_if;
            end_proc;
	else
	    proc(x:DOM_LIST,T):Type::Arithmetical
	    begin
		if T=FAIL then
		    if Risone(x[1]) then Sexpr(x[2])
		    elif Sisone(x[2]) then Rexpr(x[1])
		    else _mult(Rexpr(x[1]),Sexpr(x[2])) end_if;
		else
		    if Risone(x[1]) then Sconvto(x[2],T)
		    elif Sisone(x[2]) then Rconvto(x[1],T)
		    else _mult(Rconvto(x[1],T),Sconvto(x[2],T)) end_if;
		end_if;
            end_proc;
	end_if;


    expr :=
        proc(x:dom,T=FAIL):Type::Arithmetical
        local i:DOM_LIST;
        begin
            case dom::nterms(x)
            of 0 do return(0);
            of 1 do return(dom::exprMonomial(REP(x)[1],T));
            otherwise
		_plus(dom::exprMonomial(REP(x)[i],T) $ i=1..nops(REP(x)));
		//hold(_plus)(dom::exprMonomial(i) $ i in REP(x) );
            end_case;
        end_proc;


    /* ====================== */
    /* "polynomial" functions */
    /* ====================== */

    terms :=
        proc(x:dom):DOM_LIST
        begin if dom::iszero(x) then [] else map(REP(x),op,2) end_if end_proc;


    coeff :=
        proc(x:dom)
        local pos:DOM_INT,s;
        begin
            if args(0)=1 then map(REP(x),op,1)
 	    else
		s := dom::Scoerce(args(2));
		if testargs() then
		    if s=FAIL then TYPE_ERR end_if;
		else s := args(2) end_if;
		pos := contains(dom::terms(x),s);
	        if iszero(pos) then Rzero else op(op(REP(x),pos),1) end_if;
	    end_if;
	end_proc;


    const :=
        if not(finite) and (S::hasProp(Cat::Monoid) or
                            S::hasProp(Cat::AbelianMonoid)) then
            proc(x:dom):R
            local tmp;
            begin
                tmp := select(REP(x),Sisone);
                if tmp=[] then Rzero else tmp[1][1] end_if;
            end_proc;
        else NIL end_if;


    lcoeff :=
        proc(x:dom):R
        begin
            if dom::iszero(x) then Rzero else op(REP(x),[1,1]) end_if;
        end_proc;


    lterm :=
        proc(x:dom)
        begin
            if dom::iszero(x) then Sone else op(REP(x),[1,2]) end_if;
        end_proc;


    lmonomial :=
        proc(x:dom):dom
        begin
            if dom::iszero(x) then x
            else dom::generate1(op(REP(x)[1])) end_if;
        end_proc;


    tcoeff :=
        proc(x:dom):R
        local n:DOM_INT;
        begin
            n := dom::nterms(x);
            if iszero(n) then Rzero else op(REP(x),[n,1]) end_if;
        end_proc;


    tterm :=
        proc(x:dom)
        local n:DOM_INT;
        begin
            n := dom::nterms(x);
	    if iszero(n) then Sone else op(REP(x),[n,2]) end_if;
	end_proc;


    tmonomial :=
        proc(x:dom):dom
        local n:DOM_INT;
        begin
            n := dom::nterms(x);
            if iszero(n) then x else dom::generate1(op(REP(x)[n])) end_if;
        end_proc;


    nthmonomial :=
        proc(x:dom,i:DOM_INT):dom
        begin
            if testargs() then
		if (dom::nterms(x)<i) or (i<1) then INT_ERR end_if;
	    end_if;
            dom::generate1(op(REP(x)[i]));
        end_proc;


    nthterm :=
        proc(x:dom,i:DOM_INT)
        begin
            if testargs() then
		if (dom::nterms(x)<i) or (i<1) then INT_ERR end_if;
            end_if;
            REP(x)[i][2];
        end_proc;


    nthcoeff :=
        proc(x:dom,i:DOM_INT):R
        begin
            if testargs() then
		if (dom::nterms(x)<i) or (i<1) then INT_ERR end_if;
	    end_if;
            REP(x)[i][1];
        end_proc;


    reductum :=
        proc(x:dom):dom
	local l:DOM_LIST;
	begin
	    case dom::nterms(x)
	    of 0 do error("not defined for zero");
	    of 1 do return(dom::zero);
	    otherwise
		l := REP(x);
		delete l[1];
		return(dom::generate(l));
	    end_case;
	end_proc;


    monomial :=
        proc(r:R,s):dom
        begin dom::convert([r,s]) end_proc;


    isMonomial :=
        proc(x:dom):DOM_BOOL
        begin bool(dom::nterms(x)=1) end_proc;


    nterms :=
        proc(x:dom):DOM_INT
        begin nops(REP(x)) end_proc;


    degree :=
        if Sdegree<>FAIL then
            proc(x:dom):DOM_INT
            begin
                max(op(map(dom::terms(x),Sdegree,args(2..args(0)))));
            end_proc;
        else NIL end_if;


    degreevec :=
        if Sdegreevec<>FAIL then
	    proc(x:dom):DOM_LIST
	    begin Sdegreevec(dom::lterm(x)) end_proc;
	else NIL end_if;


    /* =========== */
    /* arithmetics */
    /* =========== */

    zero := dom::generate([]);


    iszero :=
        proc(x:dom):DOM_BOOL
        begin bool(REP(x)=[]) end_proc;


    _mult := dom::smult;


    intmult :=
        proc(x:dom,i:DOM_INT):dom
        begin dom::smult(i,x) end_proc;


    smult :=
        proc(lambda,x):dom
        begin
            if testargs() then
	        if args(0)<>2 then error("wrong number of arguments") end_if;
	    end_if;
            lambda := Rcoerce(lambda);
	    x := dom::convert(x);
	    if testargs() then
                if lambda=FAIL then TYPE_ERR end_if;
                if x=FAIL then TYPE_ERR end_if;
            end_if;
	    if Riszero(lambda) or dom::iszero(x) then return(dom::zero) end_if;
	    dom::mapcoeffs(x, ((t,r) -> Rmult(r,t)), lambda);
    end_proc;


    _negate :=
        proc(x:dom):dom
        begin
            if dom::iszero(x) then x else mapcoeffs(x,Rnegate) end_if;
        end_proc;


    plus2 :=
        proc(l1:DOM_LIST,l2:DOM_LIST):DOM_LIST
	local i:DOM_INT, tmp:R, res:DOM_LIST;
	begin
            res := [ 0 $ nops(l1)+nops(l2) ];
	    i := 0;
            while (l1<>[]) and (l2<>[]) do
		if dom::greater2(l1[1],l2[1]) then
		    i := i+1;
                    res[i] := l1[1];
                    delete l1[1];
                elif l1[1][2]=l2[1][2] then
                    tmp := Rplus(l1[1][1],l2[1][1]);
                    if not Riszero(tmp) then
			i := i+1;
                        res[i] := [tmp,l1[1][2]];
                    end_if;
                    delete l1[1];
                    delete l2[1];
                else
		    i := i+1;
                    res[i] := l2[1];
                    delete l2[1];
                end_if;
            end_while;
            if (l1<>[]) then [op(res,1..i)] . l1
            elif (l2<>[]) then [op(res,1..i)] . l2
	    else [op(res,1..i)] end_if;
	end_proc;


    _plus :=
        proc():dom
        local n:DOM_INT, s:DOM_LIST;
        begin
	    s := map([args()],dom::coerce);
	    if testargs() then
		if not(iszero(contains(s,FAIL))) then TYPE_ERR end_if;
	    end_if;
	    
            case args(0)
	    of 0 do return(dom::zero);
	    of 1 do return(s[1]);
	    of 2 do return(dom::generate(dom::plus2(REP(s[1]),REP(s[2]))));
	    otherwise
	        n := args(0) div 2;
	        dom::_plus(dom::_plus(op(s,1..n)),
			   dom::_plus(op(s,(n+1)..args(0))));
	    end_case;
	end_proc;


    sdivide :=
        if Rdivide<>FAIL then
	    proc(x,lambda)
	    begin
		x := dom::convert(x);
		lambda := Rcoerce(lambda);
                if testargs() then
                    if x=FAIL then TYPE_ERR end_if;
                    if lambda=FAIL then TYPE_ERR end_if;
                end_if;
	        if dom::iszero(x) or Risone(lambda) then return(x) end_if;
 	        dom::mapcoeffs(x,Rdivide,lambda);
	    end_proc;
        else NIL end_if;


    _divide := dom::sdivide;


    _invert := proc() begin FAIL end_proc;
    
    
    /* ============= */
    /* miscellaneous */
    /* ============= */

    mapcoeffs :=
        proc(x:dom,f):dom
        begin
            if dom::iszero(x) then return(dom::zero) end_if;
	    dom::generate(
		select(map(REP(x),
			   ((t,f) -> subsop(t,1=f(t[1],args(3..args(0))))),
			   args(2..args(0))),
		       ((t) -> not(Riszero(t[1])))));
	end_proc;


    // monomialSubs(m,s,fm) performs the substitution s=fm in a
    // monomial m=[r,s] where r is an element of R and s is an element
    // of S and fm of the free module; returns a sequence of tuples [ri,si]
    monomialSubs :=
        proc(m:DOM_LIST,s,fm:dom)
        begin
            if m[2]=s then
                op(map(REP(fm), 
		       ((l,m) -> subsop(l,1=Rmult(m[1],l[1])), m)))
            else m end_if;
        end_proc;


    // allows substitutions only for elements of S
    subs :=
        proc(x):dom
        local i:DOM_INT, eqs:DOM_LIST, eq, res:DOM_LIST, DOM:DOM_DOMAIN;
        begin
            eqs := [];
            for i from 2 to args(0) do
                case type(args(i))
                of "_equal" do
                    eqs := append(eqs,args(i));
                    break;
                of DOM_LIST do
                    eqs := eqs . args(i);
                    break;
                of DOM_SET do
                of DOM_TABLE do
                    eqs := eqs . [op(args(i))];
                    break;
                of DOM_IDENT do
                    if args(i)=hold(Unsimplified) then break end_if;
                otherwise TYPE_ERR;
                end_case;
            end_for;

	    DOM := dom;
	    x := dom::coerce(x);
            if testargs() then
                if x=FAIL then TYPE_ERR end_if;
                for i from 1 to nops(eqs) do
                    if type(eqs[i])<>"_equal" then TYPE_ERR end_if;
                    eqs[i] := subsop(eqs[i],1=DOM::Scoerce(op(eqs[i],1)),
                                            2=DOM::coerce(op(eqs[i],2)));
                    if contains([op(eqs[i])],FAIL)>0 then TYPE_ERR end_if;
                end_for;
            end_if;

            res := REP(x);
            for eq in eqs do
                res := map(res,DOM::monomialSubs,op(eq,1),op(eq,2));
            end_for;
            dom::generate(dom::normalizeRep(res));
        end_proc;
	
	
    length :=
	proc(x:dom):DOM_INT
	local l:DOM_LIST;
	begin
	    l := REP(x);
	    if l=[] then return(0) end_if;
	    1+_plus(op(map(l,(ll->Rlength(ll[1])+Slength(ll[2])+1))));
	end_proc;


    TeXMonomial :=
        if Risone=FAIL then
	    if Sisone=FAIL then
	        proc(x:DOM_LIST,p:DOM_INT):DOM_STRING
	        begin generate::TeXoperator(" ",p,130,op(x)) end_proc;
	    else
		proc(x:DOM_LIST,p:DOM_INT):DOM_STRING
		begin
		    if Sisone(x[2]) then generate::tex(x[1],p)
		    else generate::TeXoperator(" ",p,130,op(x)) end_if;
		end_proc;
	    end_if
	elif Sisone=FAIL then
	    proc(x:DOM_LIST,p:DOM_INT):DOM_STRING
	    begin
	        if Risone(x[1]) then generate::tex(x[2],p)
		else generate::TeXoperator(" ",p,130,op(x)) end_if;
            end_proc;
	else
	    proc(x:DOM_LIST,p:DOM_INT):DOM_STRING
	    begin
	        if Risone(x[1]) then generate::tex(x[2],p)
		elif Sisone(x[2]) then generate::tex(x[1],p)
		else generate::TeXoperator(" ",p,130,op(x)) end_if;
            end_proc;
	end_if;


    TeX :=
        proc(x:dom,p:DOM_INT):DOM_STRING
	local tmp:DOM_LIST,res:DOM_LIST,i;
        begin 
	    tmp := map(REP(x),dom::TeXMonomial,120);
	    res := tmp[1];
	    for i in [op(tmp,2..nops(tmp))] do
		if i[0]="-" then res := res." ".i
		else res := res." + ".i	end_if;
	    end_for;
	    generate::TeXquote(res,p,120);
	end_proc;


    random :=
        if R::random<>FAIL and dom::Srandom<>FAIL then
	    proc(N=maxr:DOM_INT):dom
	    local r:DOM_PROC, n:DOM_INT, i:DOM_INT;
	    begin
	        r := random(1..N);
	        n := r();
	        dom::generate(
	             dom::normalizeRep([[R::random(),
					 dom::Srandom()] $ i=1..n]));
	    end_proc; 
	else NIL end_if;


    coeffRing := R;


    generatorSet := S;
    

    characteristic := R::characteristic;


begin

   if args(0)<>2 then error("wrong number of arguments") end_if;
     if (domtype(R)=DOM_DOMAIN) and (R::hasProp<>FAIL) and
	R::hasProp(Cat::Rng) then
     	 Rzero := R::zero;
     	 Riszero := R::iszero;
	 Rone := R::one;
	 Risone := R::isone;
    	 Rcoerce := R::coerce;
	 Rconvto := R::convert_to;
     	 Rexpr := R::expr;
	 Rplus := R::_plus;
	 Rmult := R::_mult;
	 Rnegate := R::_negate;
	 Rdivide := R::_divide;
	 Rlength := R::length;
	 if Rlength=FAIL then Rlength := length end_if;
     else error("first argument must be Cat::Rng") end_if;
     if domtype(S)=DOM_LIST then
         finite := TRUE;
	 maxr := nops(S);
	 Sexpr := id;
	 Sconvto := ((s,T) -> s);
	 Sone := FAIL;
	 Sisone := FAIL;
         Sdegree := ((s) -> (1));
         Sdegreevec := FAIL;
	 Smult := FAIL;
	 Slength := length;
     elif domtype(S)=DOM_DOMAIN and (S::hasProp<>FAIL) and
	  S::hasProp(Cat::OrderedSet) then
         finite := FALSE;
	 maxr := 10;        // one should check for a finite set!
     	 Sexpr := S::expr;
	 Sconvto := S::convert_to;
     	 Sone := (if S::hasProp(Cat::Monoid) then S::one
	          elif S::hasProp(Cat::AbelianMonoid) then S::zero
	          else S::one;           // will probably be FAIL !
	          end_if);
     	 Sisone := (if S::hasProp(Cat::Monoid) then S::isone
	           elif S::hasProp(Cat::AbelianMonoid) then S::iszero
	           else S::isone;        // will probably be FAIL !
	           end_if);
	 Smult := (if S::hasProp(Cat::Monoid) then S::_mult
		   else FAIL end_if);
         Sdegree := S::degree;
         Sdegreevec := S::degreevec;
	 Slength := S::length;
	 if Slength=FAIL then Slength := length end_if;
     else error("second argument must be list or Cat::OrderedSet") end_if;

end_domain:


unalias(REP,TYPE_ERR,INT_ERR):
