/*++
JetVecFld.mu
Author: Marcus Hausdorf
created: November 6th, 1997
last update: 	Fri Aug 25 12:38:13 2000

Dom::JetVectorField(DF) represents vector fields over manifolds in jet 
spaces of arbitrary dimension. It inherits all methods of Dom::VectorField(DF, 
DV). DF must be in Cat::DifferentialFunction, DV is the underlying domain of
jet variables. Additional methods are:

prolong(jvf, ord) - prolong jvf to order ord: if no order is given, prolong by 1
prolong(jvf, dv_list) - prolong jvf only in the directions given by the 
    differential variables in dv_list
project(jvf, ord) - project jvf to order ord: if no order is given, project by 1
order(jvf) - order of jvf
characteristic(jvf) - characteristic of jvf
evolutionaryForm(jvf) - compute evolutionary form of jvf
ansatz(jvf, varlist) - compute general ansatz for symmetry analysis; varlist
		       contains the names of the coefficient functions
evolProlong(jvf, ord) - prolong jvf in evolutionary form to order ord: if no 
    order is given, prolong by 1
++*/
// 13.11.2000

alias(REP(x) = extop(x, 1)):

domain Dom::JetVectorField(DF)

    local DVorder:DOM_PROC, DVnIV:DOM_INT, DVnDV:DOM_INT, DVindex:DOM_PROC,
	  DVnewMulti:DOM_PROC, DVnewRepeated:DOM_PROC, DVvartype:DOM_PROC,
	  DVmultiIndex:DOM_PROC, DVtotalDiff:DOM_PROC, DVclass:DOM_PROC,
	  DVdimq:DOM_PROC, DVindVariables:DOM_LIST, DVdepVariables:DOM_LIST,
	  DVderivatives:DOM_PROC, DVzero:DV, DVmultiRaise:DOM_PROC, 
	  DFtotalDiff:DOM_PROC, DFmult:DOM_PROC, DFsubtract:DOM_PROC, 
	  DFplus:DOM_PROC, DFzero:DOM_PROC, DFconvert:DOM_PROC, DFarbFunction:
	  DOM_PROC;
    inherits (if args(0) = 1 then Dom::VectorField(DF, DV)
	      else Dom::VectorField(DF, DV, args(2)) end_if);
    category Cat::LeftModule(DF);
    axiom (if DF::hasProp(Ax::canonicalRep) then Ax::canonicalRep end_if),
          (if DF::hasProp(Ax::normalRep) then Ax::normalRep end_if);


    //===================
    // auxiliary methods 
    //===================
    
    Variables := DV;

    genMultInd :=
	proc(n:DOM_INT):DOM_LIST
	local i_l:DOM_LIST, i:DOM_INT, k:DOM_INT;
	begin
	    i_l := [[]];
	    for k from 1 to (DVnIV - 1) do
		i_l := map(i_l, ((l) -> (l . [i] $ i = 0..(n - _plus(op(l))))));
	    end_for;
	    map(i_l, ((l, j) -> (l . [n - _plus(op(l))])));
	end_proc;


    //=========================
    // prolongation/projection 
    //=========================

    prolong := 
	proc(jvf:dom):dom
	begin
	    jvf := dom::coerce(jvf);
	    if testargs() then
		if jvf = FAIL then
		    error("wrong type of argument");
		end_if;
	    end_if;
	    if args(0) = 1 then
		return(dom::prolongSlow(jvf));
	    elif args(0) = 2 then
		if domtype(args(2)) = DOM_LIST then
		    return(dom::prolongFast(jvf, args(2)))
		else
		    return(dom::prolongSlow(jvf, args(2)))
		end_if;
	    else
		error("wrong number of arguments");
	    end_if;
	end_proc;

    prolongSlow :=
	proc(jvf:dom):dom
	local newco:DF, result_l:DOM_LIST, indep_l:DOM_LIST, old_l:DOM_LIST,
	    new_l:DOM_LIST, v:DOM_LIST, dmi:DOM_LIST, ddi:DOM_INT,
	    xi_a:DOM_ARRAY, D_Ta:DOM_TABLE, do1:DOM_INT, do2:DOM_INT,
	    count:DOM_INT, bound:DOM_INT, j:DOM_INT, k:DOM_INT, DOM:DOM_DOMAIN;
	begin
	    do1 := dom::order(jvf);
	    if do1 < 0 then
		do1 := 0;
	    end_if;
	    if args(0) = 2 then
		if domtype(args(2)) <> DOM_INT or args(2) < 0 then
		    error("wrong type of argument")
		end_if;
		if args(2) > do1 then
		    do2 := args(2)
		else
		    return(jvf)
		end_if;
	    else
		do2 := do1 + 1
	    end_if;
	    DOM := dom;
	    result_l := REP(jvf);
	    indep_l := select(REP(jvf), ((l) -> (DVvartype(l[2]) = "Indep")));
	    xi_a := array(1..DVnIV, 1..DVnIV, [[DFzero $ DVnIV] $ DVnIV]);
	    for v in indep_l do
		ddi := DVindex(v[2]);
		for j from 1 to DVnIV do
		    xi_a[ddi, j] := DFtotalDiff(v[1], j);
		end_for;
	    end_for;
	    if do1 = 0 then
		old_l := map(DVdepVariables,
			 ((l) -> ([DOM::coeff(jvf, l), l])))
	    else
		old_l := map(DVderivatives(do1, 1),
			 ((l) -> ([DOM::coeff(jvf, l), l])))
	    end_if;
	    repeat
		do1 := do1 + 1;
		count := 1;
		new_l := [NIL $ DVdimq(do1)];
		for j in dom::genMultInd(do1) do
		    for k from 1 to DVnDV do
			D_Ta[j, k] := DVnewMulti("Deriv", k, j);
		    end_for;
		end_for;
		for v in old_l do
		    if do1 = 1 then
			bound := DVnIV;
		    else
			bound := DVclass(v[2]);
		    end_if;
		    dmi := DVmultiIndex(v[2]);
		    ddi := DVindex(v[2]);
		    for j from 1 to bound do
			newco := DFtotalDiff(v[1], j);
			for k from 1 to DVnIV do
			    newco := DFsubtract(newco, DFmult(xi_a[k, j],
				     DFconvert(D_Ta[DVmultiRaise(dmi, k), ddi])));
			end_for;
			new_l[count] := [newco, D_Ta[DVmultiRaise(dmi, j), ddi]];
			count := count + 1;
		    end_for;
		end_for;
		result_l := result_l . new_l;
		old_l := new_l;
	    until (do1 = do2) end_repeat;
	    dom::convert(result_l);
	end_proc;

    prolongFast := 
	proc(jvf:dom, c_l:DOM_LIST):dom
	local tmp, indep_l:DOM_LIST, dep_l:DOM_LIST, xi_a:DOM_ARRAY, v, d1_l:
	    DOM_LIST, d2_l:DOM_LIST, ddi:DOM_INT, res_l:DOM_LIST, alpha:
	    DOM_INT, prolong2:DOM_PROC;
	begin
	    prolong2 := proc(d_l:DOM_LIST, cphi:DF, ind:DOM_LIST)
		local e, k:DOM_INT; 
		begin
		    if (domtype(d_l) = DOM_LIST 
			and nops(d_l) = 2 
			and domtype(d_l[1]) = DOM_LIST 
			and domtype(d_l[2]) = DOM_LIST) 
		    then
			map(d_l, l -> prolong2(l, cphi, ind));
		    else
			while d_l <> [] do
			    e := d_l[1];
			    delete d_l[1];
			    if domtype(e) = DOM_LIST then
				if (nops(e) = 2 
				    and domtype(e[1]) = DOM_LIST
				    and domtype(e[2]) = DOM_LIST) 
				then
				    map(e, l -> prolong2(l, cphi, ind));
				elif testtype(e, 
				    Type::ListOf(Type::NonNegInt)) 
				then
				    e := [i $ e[i] $ i = 1..nops(e)];
				    for k in e do
					cphi := DFsubtract(DFtotalDiff(cphi, k), 
					    DFplus(DFmult(xi_a[i, k], 
					    DFconvert(DVnewMulti("Deriv", alpha, 
					    DVmultiRaise(ind, i)))) $ i = 1..DVnIV));
					ind := DVmultiRaise(ind, k);
				    end_for;
				else
				    d_l := e;
				end_if;
			    else
				res_l[e] := [cphi, 
				    DVnewMulti("Deriv", alpha, ind)];
			    end_if;
			end_while
		    end_if;
		end_proc;
	    
	    tmp := split(REP(jvf), l -> DVvartype(l[2]) = "Indep");
	    indep_l := tmp[1];
	    dep_l := tmp[2];
	    xi_a := array(1..DVnIV, 1..DVnIV, [[DFzero $ DVnIV] $ DVnIV]);
	    for v in indep_l do
		ddi := DVindex(v[2]);
		for j from 1 to DVnIV do
		    xi_a[ddi, j] := DFtotalDiff(v[1], j);
		end_for;
	    end_for;
	    d1_l := [[] $ DVnDV];
	    d2_l := [DFzero $ DVnDV];
	    for v in c_l do
		tmp := DVindex(v);
		d1_l[tmp] := append(d1_l[tmp], DVmultiIndex(v));
	    end_for;
	    for v in dep_l do
		d2_l[DVindex(v[2])] := v[1];
	    end_for;
	    res_l := [NIL $ nops(c_l)];
	    for alpha from 1 to DVnDV do
		if d1_l[alpha] <> [] then
		    prolong2(detools::derList2Tree(d1_l[alpha]), d2_l[alpha], 
			[0 $ DVnIV]);
		end_if;
	    end_for;
	    dom::convert(REP(jvf) . res_l);
	end_proc;	

    evolProlong :=
	proc(jvf:dom)
	local do1:DOM_INT, do2:DOM_INT, DOM:DOM_DOMAIN, result_l:DOM_LIST,
	    old_l:DOM_LIST, new_l:DOM_LIST, bound:DOM_INT, count:DOM_INT,
	    j:DOM_INT;
	begin
	    do1 := dom::order(jvf);
	    if do1 < 0 then
		do1 := 0;
	    end_if;
	    if args(0) = 2 then
		if domtype(args(2)) <> DOM_INT or args(2) < 0 then
		    error("wrong type of argument")
		end_if;
		if args(2) > do1 then
		    do2 := args(2)
		else
		    return(jvf)
		end_if;
	    else
		do2 := do1 + 1
	    end_if;
	    DOM := dom;
	    result_l := REP(jvf);
	    if do1 = 0 then
		old_l := map(DVdepVariables,
			 ((l) -> ([DOM::coeff(jvf, l), l])))
	    else
		old_l := map(DVderivatives(do1, 1),
			 ((l) -> ([DOM::coeff(jvf, l), l])))
	    end_if;
	    repeat
		do1 := do1 + 1;
		count := 1;
		new_l := [NIL $ DVdimq(do1)];
		for v in old_l do
		    if do1 = 1 then
			bound := DVnIV;
		    else
			bound := DVclass(v[2]);
		    end_if;
		    for j from 1 to bound do
			new_l[count] := [DFtotalDiff(v[1], j),
				         DVtotalDiff(v[2], j)];
			count := count + 1;
		    end_for;
		end_for;
		result_l := result_l . new_l;
		old_l := new_l;
            until (do1 = do2) end_repeat;
	    dom::convert(result_l);
	end_proc;

    project :=
	proc(jvf:dom):dom
	local do2:DOM_INT;
	begin
	    if args(0) = 2 then
		if domtype(args(2)) <> DOM_INT or args(2) < 0 then
		    error("wrong type of argument")
		end_if;
		do2 := args(2);
	    else
		do2 := dom::order(jvf) - 1;
	    end_if;
	    dom::convert(select(REP(jvf), ((l) -> (DVorder(l[2]) <=  do2))));
	end_proc;


    //===============
    // miscellaneous 
    //===============

    order :=
	proc(jvf:dom):DOM_INT
	begin
	    max(op(map(dom::terms(jvf), DVorder)))
	end_proc;

    characteristic :=
	proc(jvf:dom):DOM_LIST
	local i:DOM_INT, a:DOM_INT, DOM:DOM_DOMAIN;
	begin
	    DOM := dom;
	    [DFsubtract(
		DOM::coeff(jvf, DVnewMulti("Dep", a)),
		DFplus(
		    DFmult(
			DOM::coeff(jvf, DVnewMulti("Indep", i)),
			DFconvert(DVnewRepeated("Deriv", a, [i])))
		$ i = 1..DVnIV))
	    $ a = 1..DVnDV];
	end_proc;

    ansatz := if DFarbFunction <> FAIL then
	proc(n_l:DOM_LIST):dom
	local var_l;
	begin
	    if nops(n_l) <> (DVnIV + DVnDV) then
		error("wrong number of identifiers")
	    end_if;
	    var_l := DVindVariables . DVdepVariables:
	    dom::generate(zip(n_l, var_l,
	        ((l1, l2) -> ([DFarbFunction(l1, var_l), l2]))));
	end_proc;
    else
	proc(n_l:DOM_LIST):dom
	begin
	    if nops(n_l) <> (DVnIV + DVnDV) then
		error("wrong number of identifiers")
	    end_if;
	    dom::generate(zip(n_l, DVindVariables . DVdepVariables, 
	        ((l1, l2) -> 
		    ([DFconvert(l1(op(DVindVariables . DVdepVariables))), l2]))));
	end_proc;
    end_if;

    evolutionaryForm :=
	proc(jvf:dom):dom
	begin
	    dom::generate(zip(dom::characteristic(jvf), DVdepVariables,
		((l1, l2) -> ([l1, l2]))));
	end_proc;

    begin
	if args(0) < 1 or args(0) > 2 then 
	    error("wrong number of arguments") 
	end_if;
	if (domtype(DF) <> DOM_DOMAIN
	    or DF::hasProp = FAIL
	    or not DF::hasProp(Cat::DifferentialFunction)
	    or not DF::hasProp(Cat::PartialDifferentialRing))
	then error
	    ("argument must be in Cat::DifferentialFunction")
	end_if;
	DV := DF::Variables;
	DVorder := DV::order;
	DVnIV := DV::numIndVar;
	DVnDV := DV::numDepVar;
	DVindex := DV::index;
	DVnewMulti := DV::newMulti;
	DVnewRepeated := DV::newRepeated;
	DVvartype := DV::vartype;
	DVmultiIndex := DV::multiIndex;
	DVtotalDiff := DV::totalDiff;
	DVclass := DV::class;
	DVdimq := DV::dimq;
	DVzero := DV::zero;
	DVindVariables := DV::indVariables;
	DVdepVariables := DV::depVariables;
	DVderivatives := DV::derivatives;
	DVmultiRaise := DV::multiRaise;
	DFtotalDiff := DF::totalDiff;
	DFmult := DF::_mult;
	DFsubtract := DF::_subtract;
	DFplus := DF::_plus;
	DFzero := DF::zero;
	DFconvert := DF::convert;
	DFarbFunction := DF::arbFunction;

end_domain:

unalias(REP);






