/*++
Dom::LinearDifferentialFunction(DV,R) -- linear functions in the
	differential variables from DV with coefficients from the ring R


If the optional argument R is omitted, Dom::ExpressionField() is used;
this corresponds to linear differential functions with constant
coefficients. For variable coefficients, R should be a ring of differential
functions over Dom::RestrictedDifferentialVariable(DV,Types="Indep").

The implemented methods are those of a Cat::DifferentialFunction.  The
implementation is based on Dom::FreeModule. However, an additional
second slot is introduced in order to represent inhomogeneous terms.
This requires modifications of many methods.

The simplification is done by Gaussian elimination. Additional methods are
    makeMatrix :    extracts the coefficient matrix of a system
		    (list of functions)
    makeSystem :    takes a coefficient matrix and transforms it back 
		    into a system
    isHomogeneous : checks whether the linear function is homogeneous
    inhomogeneity :  returns the inhomogeneous part of the function
Furthermore, all the "polynomial" methods of Dom::FreeModule are available.

In which differential variables the domain elements are linear depends on the
coefficient ring R. If it is an instance of Cat::DifferentialFunction, then
it must be over a differential variable domain of the form
Dom::RestrictedDifferentialVariable(DV,...). The linearity is now with respect
to all differential variables *not* contained in this domain. This affects in
particular the method jacobian: it takes only these variables into account and
ignores those variables contained in the coefficients.  This is necessary to
avoid algebraic problems with the matrices.  Linear functions do not form a
ring and thus we cannot build matrices upon them.  Jacobians are always
matrices over the coefficient ring.

This domain can also be used to represent something similar to quasi-linear
functions, if the coordinate ring consists of differential functions over an
instance of Dom::RestrictedDifferentialVariable. In this case the linearity
concerns only the differential variables not contained in this domain and the
method "jacobian" considers only the dependency on these.


Authors: Eduard Niess, Werner M. Seiler, Uni Karlsruhe/Mannheim
         (wms@ira.uka.de, seiler@euler.math.uni-mannheim.de)
         (with contributions by Jay Belanger (belanger@truman.edu))
Created: ??? (1996)
Last update: 	Thu Apr 20 17:40:19 2000
++*/
// 13.10.2000

alias(REP(x)=extop(x,1),
      INH(x)=extop(x,2),
      ARG_ERR=error("wrong number of arguments"),
      TYPE_ERR=error("wrong type of argument")):


domain Dom::LinearDifferentialFunction()

    local DVcoerce:DOM_PROC, DVdiff:DOM_PROC, DVindex:DOM_PROC, 
	  DVtype:DOM_PROC, DVorder:DOM_PROC, DVconvto:DOM_PROC, 
	  DVexpr:DOM_PROC, DVgreater:DOM_PROC, DVmulti:DOM_PROC,
	  DVrepeated:DOM_PROC,
	  Rlength:DOM_PROC, Riszero:DOM_PROC, Risone:DOM_PROC, Rexpr:DOM_PROC, 
	  Rcoerce:DOM_PROC, Rplus:DOM_PROC, Rmult:DOM_PROC, Rdiff:DOM_PROC, 
	  Rinvert:DOM_PROC, Rindets:DOM_PROC, Rnegate:DOM_PROC, Rone:R,
	  Rdiffvars:DOM_PROC, Rconvto:DOM_PROC, Rhas:DOM_PROC, Rzero:R,
	  DV:DOM_PROC, R:DOM_PROC, opts:DOM_LIST,
	  SMR:DOM_DOMAIN, ConstCoeff:DOM_BOOL, RDV:DOM_DOMAIN, 
	  testDV:DOM_PROC, convDV:DOM_PROC, convRDV:DOM_PROC,
	  remdup:DOM_PROC, swap:DOM_PROC, mode:DOM_STRING,
	  TyLO:DOM_PROC, TyAT:DOM_PROC, nn:DOM_INT, mm:DOM_INT,
	  indvars:DOM_LIST, depvars:DOM_LIST, evars:DOM_LIST, pvars:DOM_SET;
    inherits Dom::FreeModule(R,DV);
    category Cat::DifferentialFunction(DV), Cat::LeftModule(R);
    axiom (if R::hasProp(Ax::normalRep) then Ax::normalRep end_if),
          (if R::hasProp(Ax::canonicalRep) then Ax::canonicalRep end_if);


    /* ========================= */
    /* representation, in/output */
    /* ========================= */

    // overwrites the corresponding methods in FreeMod.mu
    // takes possible right hand side into account
    generate :=
	proc(x:DOM_LIST,inh=Rzero):dom
	begin new(dom,x,inh) end_proc;
	
	
    generate1 :=
	proc(x:R):dom
	begin
	    if args(0)=2 then new(dom,[[x,args(2)]],Rzero)
	    else new(dom,[],x) end_if;
	end_proc;
    
    // recognize all differential variables in a DOM_EXPR
    subsDV :=
	proc(de)
	local tmp;
	begin
	    tmp := DVcoerce(de);
	    if tmp<>FAIL then return(tmp) end_if;
	    if domtype(de)=DOM_EXPR then return(map(de,dom::subsDV))
	    else return(de) end_if;
	end_proc;


    convert :=
        proc(x)
        local n, r, s, f, res, rres, sres;
        begin
	    if iszero(x) then return(dom::zero) end_if;
	    if x=1 then return(dom::one) end_if;
	    s := DVcoerce(x);
	    if s<>FAIL then
		if testDV(s) then 
		    r := Rcoerce(convRDV(s));
		    if r=FAIL then return(FAIL)
		    else return(dom::generate1(r)) end_if;
		else return(dom::generate1(Rone,s)) end_if;
	    end_if;

	    case domtype(x)
	    of dom do return(x);
	    of DOM_LIST do
	        if (nops(x)=2) and ((r:=Rcoerce(x[1]))<>FAIL) and
	           ((s:=DVcoerce(x[2]))<>FAIL) then
		    if testDV(s) then return(FAIL)
		    elif Riszero(r) then return(dom::zero)
		    else return(dom::generate1(r,s));
		    end_if;
	        else
		    n := nops(x);
		    r := Rcoerce(x[n]);
		    if r<>FAIL then x := [op(x,1..(n-1))] 
		    else r := Rzero end_if;
		    if {op(map(x,domtype))}<>{DOM_LIST} or
		       {op(map(x,nops))}<>{2} then return(FAIL) end_if;
		    x := map(x,dom::coerceSummand);
		    if has(x,FAIL) then return(FAIL);
		    else
		        for s in x do
		            if testDV(s[2]) then return(FAIL) end_if;
		        end_for;
	 	        return(dom::generate(dom::normalizeRep(x),r));
		    end_if;
	         end_if;
	    of DOM_EXPR do
	        x := dom::subsDV(expand(x));
	        case type(x)
	        of "_plus" do
		    res := dom::zero;
		    for r in op(x) do
		        s := dom::convert(r);
		        if s=FAIL then return(FAIL) end_if;
		        res := dom::_plus(res,s);
		    end_for;
		    return(res);
	        of "_mult" do
		    rres := [];
		    sres := [];
		    for f in op(x) do
			if domtype(f)=DV then
			    if testDV(f) then
				if ConstCoeff then return(FAIL) end_if;
				r := Rcoerce(convRDV(f));
				if r=FAIL then return(FAIL)
				else rres := append(rres,r) end_if;
			    else sres := append(sres,f) end_if;
		        else
			    r := Rcoerce(f);
			    if r<>FAIL then rres := append(rres,r);
			    else return(FAIL) end_if;
		        end_if;
		    end_for;
		    if rres=[] then rres := Rone
		    else rres := Rmult(op(rres)) end_if;
		    case nops(sres)
		    of 0 do return(dom::generate1(rres));
		    of 1 do return(dom::generate1(rres,sres[1]));
		    otherwise return(FAIL);
		    end_case;
		otherwise
		    x := misc::maprec(x,
			    {DV}=(dv -> if testDV(dv) then convRDV(dv) 
					else FAIL end_if));
		    if has(x,FAIL) then return(FAIL) end_if;
		    r := Rcoerce(x);
		    if r<>FAIL then
			if Riszero(r) then return(dom::zero)
			else return(dom::generate1(r)) end_if;
		    end_if;
	        end_case;
	    otherwise
	        r := Rcoerce(x);
	        if r<>FAIL then
		    if Riszero(r) then return(dom::zero)
		    else return(dom::generate1(r)) end_if;
	        end_if;
	        return(FAIL);
	    end_case;
        end_proc;


    convDiff :=
        proc(x:DOM_LIST,T:DOM_STRING):DOM_EXPR
        begin
            if Risone(x[1]) then DVconvto(x[2],T)
            elif ConstCoeff then _mult(Rexpr(x[1]),DVconvto(x[2],T)) 
	    else _mult(Rconvto(x[1],T),DVconvto(x[2],T)) end_if;
        end_proc;


    convert_to :=
        proc(x:dom,T)
        local i:DOM_INT, res, inh;
        begin
            case T
	    of dom do return(x);
	    of DOM_EXPR do
		res := REP(x);
		inh := INH(x);
		if ConstCoeff then
		    res := map(res,((l) -> [Rexpr(l[1]),l[2]]));
		    if Riszero(inh) then inh := null();
		    else inh := Rexpr(inh) end_if;
		else
		    res := map(res,((l) -> [Rconvto(l[1],DOM_EXPR),l[2]]));
		    if Riszero(inh) then inh := null();
		    else inh := Rconvto(inh,DOM_EXPR) end_if;
		end_if;
		res := _plus(op(map(res,_mult@op)),inh);
		return(res);
            of "diff" do
            of "D" do
		inh := INH(x);
		if Riszero(inh) then inh := 0
		else inh := Rconvto(inh,T) end_if;
                case nops(REP(x))
                of 0 do return(inh);
                of 1 do return(inh+dom::convDiff(REP(x)[1],T));
                otherwise
		    inh+_plus(dom::convDiff(REP(x)[i],T) $ i=1..nops(REP(x)));
                end_case;
                break;
	    of R do
		if REP(x)=[] then return(INH(x))
		else return(FAIL) end_if;
	    of DV do
		if REP(x)=[] then return(Rconvto(INH(x),DV))
		elif nops(REP(x))=1 and Riszero(INH(x)) and 
		     Risone(REP(x)[1][1]) then return(REP(x)[1][2])
		else return(FAIL) end_if;
            otherwise return(FAIL);
            end_case;
        end_proc;
	
	
    expr :=
	proc(l:dom):Type::Arithmetical
	local res:DOM_LIST;
	begin
	    res := map(REP(l), 
		       (ll -> if Risone(ll[1]) then DVexpr(ll[2])
			      else _mult(Rexpr(ll[1]),DVexpr(ll[2])) end_if));
	    if Riszero(INH(l)) then _plus(op(res))
	    else _plus(op(res),Rexpr(INH(l))) end_if;
	end_proc;


    /* ============== */
    /* simple methods */
    /* ============== */

    // Most of the following methods are modifications of methods already
    // defined in FreeMod.mu. They need new implementations due to the
    // existence of the "right hand side".
	    
    zero := dom::generate1(Rzero);
    
    
    one := dom::generate1(Rone);


    iszero :=
	proc(l:dom):DOM_BOOL
	begin bool(REP(l)=[]) and Riszero(INH(l)) end_proc;


    isone :=
	proc(l:dom):DOM_BOOL
	begin bool(REP(l)=[]) and Risone(INH(l)) end_proc;
	
	
    _plus :=
        proc():dom
        local n:DOM_INT, s:DOM_LIST;
        begin
	    s := map([ args() ], dom::coerce);
	    if testargs() then
		if not(iszero(contains(s,FAIL))) then TYPE_ERR end_if;
	    end_if;
	    
            case args(0)
	    of 0 do return(dom::zero);
	    of 1 do return(s[1]);
	    of 2 do 
		return(dom::generate(
			    dom::plus2(REP(s[1]),REP(s[2])),
			    Rplus(INH(s[1]),INH(s[2]))));
	    otherwise
	        n := args(0) div 2;
	        dom::_plus(dom::_plus(op(s,1..n)),
			   dom::_plus(op(s,(n+1)..args(0))));
	    end_case;
	end_proc;
	
	
    length :=
	proc(l:dom):DOM_INT
	begin
	    if REP(l)=[] then return(Rlength(INH(l))) end_if;
	    Rlength(INH(l))+1+2*nops(REP(l))+
		_plus(op(map(REP(l),(ll->Rlength(ll[1])))));
	end_proc;
	
	
    indets :=
        proc(l):DOM_SET
        local res:DOM_LIST, c:R, cres:DOM_SET, tres:DOM_SET, opts:DOM_SET;
        begin
	    l := dom::convert(l);
	    if testargs() then
	        if (args(0)<1) or (args(0)>3) then ARG_ERR end_if;
	        if l=FAIL then TYPE_ERR end_if;
	    end_if;

            if dom::iszero(l) then return({}) end_if;
	    opts := {args(2..args(0))};

            res := dom::terms(l);
	    tres := {op(res)};
	    if Rindets=FAIL then return(tres) end_if;

	    if ConstCoeff then
		if contains(opts,hold(DiffVars)) then return(tres)
		else
		    cres := Rindets(INH(l),op(opts));
		    for c in dom::coeff(l) do
			cres := cres union Rindets(c,op(opts));
		    end_for;
		end_if;
		return(tres union cres);
	    else
		cres := Rindets(INH(l),op(opts));
		for c in dom::coeff(l) do
		    cres := cres union Rindets(c,op(opts));
		end_for;
		return(tres union
		       map(cres, 
			   (v -> if domtype(v)=RDV then convDV(v) 
				 else v end_if)));
	    end_if;
        end_proc;
	
	
    isHomogeneous :=
	proc(l:dom):DOM_BOOL
	begin Riszero(INH(l)) end_proc;
	
	
    inhomogeneity :=
	proc(l:dom):R
	begin INH(l) end_proc;


    /* ================== */
    /* polynomial methods */
    /* ================== */
    
    coeff :=
	proc(l:dom)
	local res:DOM_LIST, s, pos:DOM_INT;
	begin
	    if args(0)=1 then
		res := if REP(l)=[] then [] else map(REP(l),op,1) end_if;
		if Riszero(INH(l)) then return(res)
		else return([op(res),INH(l)]) end_if;
	    else
		s := DVcoerce(args(2));
		if testargs() then
		    if s=FAIL then TYPE_ERR end_if;
		else s := args(2) end_if;
		pos := contains(map(REP(l),op,2),s);
	        if iszero(pos) then Rzero else op(op(REP(l),pos),1) end_if;
	    end_if;
	end_proc;
	
	
    mapcoeffs :=
        proc(x:dom,f):dom
        begin
            if dom::iszero(x) then return(dom::zero) end_if;
	    dom::generate(
		select(map(REP(x),
			   ((t,f) -> subsop(t,1=f(t[1],args(3..args(0))))),
			   args(2..args(0))),
		       ((t) -> not(Riszero(t[1])))),
		f(INH(x),args(3..args(0))));
	end_proc;
	
    
    lcoeff :=
	proc(l:dom):R
	begin
	    if REP(l)=[] then INH(l) else op(REP(l),[1,1]) end_if;
	end_proc;
	
	
    lterm :=
	proc(l:dom)
	begin
	    if REP(l)=[] then FAIL else op(REP(l),[1,2]) end_if;
	end_proc;

        
    lmonomial :=
	proc(l:dom):dom
	begin
	    if REP(l)=[] then dom::generate1(INH(l))
	    else dom::generate1(op(REP(l)[1])) end_if;
	end_proc;
	
	
    tcoeff :=
	proc(l:dom):R
	local n:DOM_INT;
	begin
	    if Riszero(INH(l)) then
		n := nops(REP(l));
		op(REP(l),[n,1]);
	    else INH(l) end_if;
	end_proc;
	
	
    tterm :=
	proc(l:dom)
	local n:DOM_INT;
	begin
	    if Riszero(INH(l)) then
		n := nops(INH(l));
		if iszero(n) then FAIL
		else op(REP(l),[n,2]) end_if;
	    else FAIL end_if;
	end_proc;
	
	
    tmonomial :=
	proc(l:dom):dom
	local n:DOM_INT;
	begin
	    if Riszero(INH(l)) then
		n := nops(REP(l));
		dom::generate1(op(REP(l)[n]));
	    else dom::generate1(INH(l)) end_if;
	end_proc;
	
	
    nthmonomial :=
	proc(l:dom,i:DOM_INT):dom
	begin
	    if testargs() then
		if (dom::nterms(l)<1) or (i<1) then
		    error("index out of range");
		end_if;
	    end_if;
	    if i>nops(REP(l)) then dom::generate1(INH(l))
	    else dom::generate1(op(REP(l)[i])) end_if;
	end_proc;
	
	
    nthterm :=
	proc(l:dom,i:DOM_INT)
	begin
	    if testargs() then
		if (dom::nterms(l)<1) or (i<1) then
		    error("index out of range");
		end_if;
	    end_if;
	    if i>nops(REP(l)) then FAIL
	    else REP(l)[i][2] end_if;
	end_proc;
	
	
    nthcoeff :=
	proc(l:dom,i:DOM_INT)
	begin
	    if testargs() then
		if (dom::nterms(l)<1) or (i<1) then
		    error("index out of range");
		end_if;
	    end_if;
	    if i>nops(REP(l)) then INH(l)
	    else REP(l)[i][1] end_if;
	end_proc;
	
	
    reductum :=
	proc(l:dom)
	local newl:DOM_LIST;
	begin
	    case nops(REP(l))
	    of 0 do
		if Riszero(INH(l)) then error("not defined for zero")
		else return(dom::zero) end_if;
	    of 1 do return(dom::generate1(INH(l)));
	    otherwise
		newl := REP(l);
		delete newl[1];
		return(dom::generate(newl,INH(l)));
	    end_case;	    
	end_proc;
	
	
    nterms :=
	proc(l:dom):DOM_INT
	begin
	    if Riszero(INH(l)) then nops(REP(l)) else nops(REP(l))+1 end_if;
	end_proc;
	
	
    /* ==================== */
    /* differential methods */
    /* ==================== */

    diff :=
        proc(l:dom):dom
        local dv, tmp:DOM_LIST;
        begin
            case args(0)
            of 0 do ARG_ERR;
            of 1 do return(l);
            of 2 do
		dv := DVcoerce(args(2));
                if testargs() then
                    if dv<>FAIL then return(dom::diff2(l,dv))
                    else return(dom::diff2(args())) end_if;
                else return(dom::diff2(args())) end_if;
            otherwise
                if testargs() then
		    tmp := map([args(2..args(0))],
                               ((k) -> (DVcoerce(k);
                                        if %=FAIL then k else % end_if)));
                    return(dom::diffn(l,op(tmp)));
                else return(dom::diffn(args())) end_if;
            end_case;
        end_proc;


    diff2 :=
        proc(l:dom,x):dom
        local pos:DOM_INT, tmp:DOM_LIST;
        begin
	    if domtype(x)<>DV then     // differentiation wrt parameter
	        if Rdiff=FAIL then return(dom::zero) end_if;
	        dom::mapcoeffs(l,Rdiff,x);
	    elif not(ConstCoeff) and testDV(x) then
	        dom::mapcoeffs(l,Rdiff,convRDV(x));
	    else
	        tmp := REP(l);
	        pos := contains(dom::terms(l),x);
	        if iszero(pos) then dom::zero
	        else dom::generate1(tmp[pos][1]) end_if;
	    end_if;
        end_proc;


    diffn :=
        proc(l:dom):dom
        local x, avars:DOM_LIST, ivars:DOM_LIST, dvars:DOM_LIST;
        begin
            avars := [];
            ivars := [];
            dvars := [];
            for x in [args(2..args(0))] do
                if domtype(x)<>DV  then avars := append(avars,x)
                elif testDV(x) then ivars := append(ivars,convRDV(x))
                else dvars := append(dvars,x) end_if;
            end_for;

            if ConstCoeff then
                if ivars<>[] then return(dom::zero) end_if;
            else
                avars := avars .ivars;
            end_if;
            if Rdiff=FAIL and avars<>[] then return(dom::zero) end_if;

            case nops(dvars)
            of 0 do return(dom::mapcoeffs(l,Rdiff,op(avars)));
            of 1 do return(dom::mapcoeffs(dom::diff2(l,dvars[1]),
				          Rdiff,op(avars)));
            otherwise return(dom::zero);
            end_case;
        end_proc;


    // totalDiff is implemented for all possible cases of restrictions.
    // However, as soon as quasi-linear functions belong to the domain,
    // totalDiff will lead in general out of the domain and thus yield
    // an error!
    totalDiff :=
	if ConstCoeff then	// ***** constant coefficients *****
	    proc(sys)
	    local mu:DOM_LIST, i:DOM_INT, tmp, l:DOM_LIST, nl:DOM_INT, 
		  sysflag:DOM_BOOL;
	    begin
		// analyse arguments
                if args(0)=1 then return(sys) end_if;
	        sysflag := bool(domtype(sys)=DOM_LIST);
                if not(sysflag) then sys := [sys] end_if;
	        if domtype(args(2))<>DOM_LIST then mu := [args(2..args(0))]
	        else mu := args(2) end_if;

	        sys := map(sys,dom::coerce);
		if testargs() then
                    if not(iszero(contains(sys,FAIL))) then TYPE_ERR end_if;
                    for i from 1 to nops(mu) do
                        if domtype(mu[i])=DOM_INT then
                            if mu[i]<1 or mu[i]>nn then TYPE_ERR end_if;
                        else
                            tmp := DVcoerce(mu[i]);
                            if tmp=FAIL or DVtype(tmp)<>"Indep" then
                                TYPE_ERR
                            end_if;
                            mu[i] := DVindex(tmp);
                        end_if;
                    end_for;
	        end_if;

		// differentiate all generators
                for i from 1 to nops(sys) do	// loop over equations
                    if dom::iszero(sys[i]) then next end_if;
	            l := REP(sys[i]);
	            nl := nops(l);
	            if l=[] then sys[i] := dom::zero;
		    else
			sys[i] := dom::generate(
				    map(l, 
					((k,v) -> subsop(k,2=DVdiff(k[2],v))),
					mu));
		    end_if;
                end_for;
		// return result
                if sysflag then sys else sys[1] end_if;
            end_proc
	else
	    case mode
	    of "Indep" do 	// ***** variable coefficients *****
		proc(sys,mu)
		local i:DOM_INT, j:DOM_INT, tmp, l:DOM_LIST, 
		      nl:DOM_INT, res:dom, sysflag:DOM_BOOL;
		begin
		    // analyse arguments
		    if args(0)=1 then return(sys) end_if;
		    sysflag := bool(domtype(sys)=DOM_LIST);
		    if not(sysflag) then sys := [sys] end_if;
		    if domtype(args(2))<>DOM_LIST then mu := [args(2..args(0))]
		    else mu := args(2) end_if;

		    sys := map(sys,dom::coerce);
		    if testargs() then
			if not(iszero(contains(sys,FAIL))) then 
			    TYPE_ERR;
		        end_if;
			for i from 1 to nops(mu) do
			    if domtype(mu[i])=DOM_INT then
				if mu[i]<1 or mu[i]>nn then TYPE_ERR end_if;
			    else
				tmp := DVcoerce(mu[i]);
				if tmp=FAIL or DVtype(tmp)<>"Indep" then
				    TYPE_ERR
				end_if;
				mu[i] := DVindex(tmp);
			    end_if;
			end_for;
		    end_if;

		    // perform differentiations
		    for j in mu do	// loop over differentiations
			for i from 1 to nops(sys) do	// loop over system
			    if dom::iszero(sys[i]) then next end_if;
			    // differentiate coefficients
			    res := dom::mapcoeffs(sys[i],Rdiff,
						  RDV::new("Indep",j));
			    // differentiate generators
			    l := REP(sys[i]);
			    nl := nops(l);
			    if l=[] then sys[i] := res end_if;
			    sys[i] := dom::_plus(res,
				dom::generate(
				    map(l,((k,v)->subsop(k,2=DVdiff(k[2],v))),
					j)));
			end_for;
		    end_for;
		    if sysflag then sys else sys[1] end_if;
		end_proc;
		break;
	    of "Less" do	// ***** "quasi-linear" case *****
	    of "Greater" do 	// ***** degenerate case *****
	    of "General" do 	// ***** general case *****
		proc() begin error("not a linear function") end_proc;
		break;
	    otherwise error("unknown restriction mode");
	    end_case;
	end_if;
		

    // changeDepVars will take as input an element of this domain, 
    // a list of new dependent variables, and either a list of
    // equations equating the old variables with the new, or
    // a list of lists, where
    // each sublist will be the coefficients of a linear change of
    // variables.  (If, for example, the old dependent variables are 
    // u and v, the new dependent variables are U and V, and the change
    // is given by u=a*U + b*V, v=c*U + d*V, then the third argument
    // would be either [u=a*U + b*V, v=c*U + d*V] or [[a,b],[c,d]].)  
    // The output domain can be given by adding the argument 
    // OutputDomain = ... (although there is a default domain).
    changeDepVars :=
	proc(ldf, newVars:DOM_LIST, oldFromNew)
	local changeVars:DOM_PROC, ivars:DOM_LIST, optionalArgTable:DOM_TABLE, 
	      tmpTable:DOM_TABLE, dvars:DOM_LIST, i:DOM_INT, j:DOM_INT, eqn,
	      newR:DOM_DOMAIN, newRD:DOM_DOMAIN, changeCoeffs:DOM_PROC, 
	      newDV:DOM_DOMAIN, outputDomain:DOM_DOMAIN, onlyType:DOM_PROC,
	      rHSide;
	begin
	    // First of all, a useful procedure
	    onlyType :=
		proc(list,dtype):DOM_BOOL
		begin
		    list := map(list, (x) -> (bool(type(x)=dtype)));
		    return(iszero(contains(list,FALSE)));
		end_proc; /* onlyType */
	    ivars := map(indvars,DV::expr);
	    // Next, read in the optional arguments
	    optionalArgTable := table();
	    for i from 4 to args(0) do
		optionalArgTable[op(args(i),1)] := op(args(i),2);
	    end_for;
	    // Next, make sure that oldFromNew is in the proper form.
	    if onlyType(oldFromNew,"_equal") then
		tmpTable := table();
		for eqn in oldFromNew do
		    tmpTable[op(eqn,1)] := op(eqn,2);
		end_for;
		dvars := map(depvars,DV::expr);	    
		oldFromNew :=
		   [[coeff(tmpTable[dvars[i]],newVars[j],1) $ j=1..mm] 
		                                            $ i=1..mm];
	    end_if;
	    // Next, make sure that outputDomain is in the proper form
	    if contains(optionalArgTable,hold(OutputDomain)) then
		outputDomain := optionalArgTable[hold(OutputDomain)];
	    else
                newDV:=Dom::DifferentialVariable(ivars,newVars);
     		if ConstCoeff then
		    newR := Dom::ExpressionField();
		    outputDomain:=Dom::LinearDifferentialFunction(newDV,newR);
		else
		  case mode
		  of "Indep" do 
		     newRD := 
		     Dom::RestrictedDifferentialVariable(newDV,
			                                 hold(Types)="Indep");
		     newR := Dom::DifferentialExpression(newRD);
		     outputDomain := 
		                  Dom::LinearDifferentialFunction(newDV,newR);
		     break;
		  of "Less"  do 
		     newRD := 
		     Dom::RestrictedDifferentialVariable(newDV,
						hold(Types)={"Indep","Dep"});
		     newR := Dom::DifferentialExpression(newRD);
		     outputDomain := 
		                 Dom::LinearDifferentialFunction(newDV,newR);
		     break;
		  otherwise TYPE_ERR;
		  end_case;
		end_if;
	    end_if;
	    // Now, the crucial procedures
	    changeCoeffs :=
		if mode = "Less" then
		    proc(r)
		    begin
			for i from 1 to mm do
			    r := subs(r, dvars[i]=
				_plus(oldFromNew[i][j]*newVars[j] $ j=1..mm));
			end_for;
		    end_proc; 
		else
		    id
		end_if;
	    changeVars :=
		proc(lf)
		local dvindex, ivs, lfone;
		begin
		    lfone := changeCoeffs(R::expr(lf[1]));
		    dvindex := DV::index(lf[2]);
		    ivs := map(DV::repeatedIndex(lf[2]),(x)->(ivars[x]));
			if nops(ivs) = 0 then
			    return(
			    [oldFromNew[dvindex][i]*lfone,
			    newVars[i]]
			    $ i=1..mm);
			else
			    return(
			    [oldFromNew[dvindex][i]*lfone,
			    newVars[i](ivs)]
			    $ i=1..mm);
			end_if;
		end_proc; /* changeVars */
	    rHSide := changeCoeffs(Rexpr(INH(ldf)));
	    ldf := REP(ldf);
	    ldf := map(ldf, changeVars); 
	    // return(new(outputDomain,ldf,rHSide)); isn't working...
	    return(outputDomain(ldf)+outputDomain(rHSide));
	end_proc; /* changeDepVars */


    // changeIndVars will take as input an element of this domain, 
    // a list of new independent variables, and a list of equations 
    // equating the new variables with the old, or a list of lists, where
    // each sublist will be the coefficients of a linear change of
    // variables.  (If, for example, the old independent variables are 
    // x and y, the new dependent variables are X and Y, and the change
    // is given by X=a*x + b*y, Y=c*x + d*y, then the third argument
    // would be [X=a*x + b*y, Y=c*x + d*y] or  [[a,b],[c,d]].)  The old
    // variables in terms of the new will be computed by assuming that the
    // change of variables is linear, or the old variables in terms of the
    // new can be specified by the argument OldFromNew = ...
    // The output domain can be given by adding the argument 
    // OutputDomain = ... (although there is a default domain).

    changeIndVars :=
	proc(ldf, newVars:DOM_LIST, newFromOld)
	local optionalArgTable:DOM_TABLE, onlyType:DOM_PROC, ivars:DOM_LIST,
	      dvars:DOM_LIST, tmpTable:DOM_TABLE, oldFromNew:DOM_LIST, 
	      DM:DOM_DOMAIN, newDV:DOM_DOMAIN, newR:DOM_DOMAIN, 
	      newRD:DOM_DOMAIN, i:DOM_INT, j:DOM_INT, needOldFromNew:DOM_BOOL,
	      outputDomain:DOM_DOMAIN, changeRep:DOM_PROC, changeVars:DOM_PROC,
	      changeCoeffs:DOM_PROC, changeRepBack:DOM_PROC, eqn, rHSide;
	begin
	    // First of all, a useful procedure
	    onlyType :=
		proc(list,dtype):DOM_BOOL
		begin
		    list := map(list, (x) -> (bool(type(x)=dtype)));
		    return(iszero(contains(list,FALSE)));
		end_proc; /* onlyType */
	    ivars := map(indvars,DV::expr);
	    dvars := map(depvars,DV::expr);	
	    // Next, read in the optional arguments
	    optionalArgTable := table();
	    for i from 4 to args(0) do
		optionalArgTable[op(args(i),1)] := op(args(i),2);
	    end_for;
	    // Now, see if there is a user supplied OldFromNew
	    needOldFromNew := TRUE;
	    if contains(optionalArgTable,hold(OldFromNew)) then
		needOldFromNew := FALSE;
		oldFromNew := optionalArgTable[hold(OldFromNew)];
		if onlyType(oldFromNew,"_equal") then
		    tmpTable := table();
		    for eqn in oldFromNew do
			tmpTable[op(eqn,1)] := op(eqn,2);
		    end_for;
		    oldFromNew := [tmpTable[ivars[j]] $ j=1..nn];
		elif onlyType(oldFromNew,DOM_LIST) then
		    oldFromNew := 
			[_plus(oldFromNew[i][j]*newVars[j] $ j=1..nn) $
			 i=1..nn];
		end_if;
	    end_if;
	    // Next, make sure that newFromOld is in the proper form.
	    if onlyType(newFromOld,"_equal") then
		// Take care of oldFromNew
		if needOldFromNew then
		    needOldFromNew := FALSE;
		    oldFromNew := solve(newFromOld, ivars);
		    if nops(oldFromNew) <> 1 then
			error("OldFromNew needs to be supplied");
		    end_if;
		    oldFromNew := op(oldFromNew);
		    tmpTable := table();
		    for eqn in oldFromNew do
			tmpTable[op(eqn,1)] := op(eqn,2);
		    end_for;
		    oldFromNew := [tmpTable[ivars[j]]$j=1..nn];
		end_if;
		tmpTable := table();
		for eqn in newFromOld do
		    tmpTable[op(eqn,1)] := op(eqn,2);
		end_for;
		newFromOld :=
		     [[diff(tmpTable[newVars[i]],ivars[j]) $ j=1..nn] 
		       $ i=1..nn];
	    end_if;
	    // Finally, make sure that we have oldFromNew.
            // If OldFromNew isn't given to us, we'll assume that
	    // the change of variables is linear
	    if needOldFromNew then
		DM := Dom::Matrix();
		oldFromNew := DM::convert_to(1/DM(newFromOld),DOM_LIST);
		oldFromNew := 
			[_plus(oldFromNew[i][j]*newVars[j] $ j=1..nn) 
			 $ i=1..nn];
	    end_if;
	    // Now, make sure that outputDomain is in the proper form
	    if contains(optionalArgTable,hold(OutputDomain)) then
		outputDomain := optionalArgTable[hold(OutputDomain)];
	    else
                newDV:=Dom::DifferentialVariable(newVars,dvars);
     		if ConstCoeff then
		    // A non-linear change of variables could take an ldf
		    // with constant coefficients and return an ldf
		    // with variable coeffients.
		    // I'll have to add a check for this, but for right
		    // now, the output domain should be able to take
		    // non-constant coefficients
		     newRD := Dom::RestrictedDifferentialVariable(newDV,
							hold(Types)="Indep");
		     newR := Dom::DifferentialExpression(newRD);
		     outputDomain := 
		                  Dom::LinearDifferentialFunction(newDV,newR);
		// newR := Dom::ExpressionField();
	        //outputDomain:=Dom::LinearDifferentialFunction(newDV,newR);
		else
		  case mode
		  of "Indep" do 
		     newRD := Dom::RestrictedDifferentialVariable(newDV,
							hold(Types)="Indep");
		     newR := Dom::DifferentialExpression(newRD);
		     outputDomain := 
		                  Dom::LinearDifferentialFunction(newDV,newR);
		     break;
		  of "Less"  do 
		     newRD := 
		     Dom::RestrictedDifferentialVariable(
		                         newDV,hold(Types)={"Indep","Dep"});
		     newR := Dom::DifferentialExpression(newRD);
		     outputDomain := 
		                 Dom::LinearDifferentialFunction(newDV,newR);
		     break;
		  otherwise TYPE_ERR;
		  end_case;
		end_if;
	    end_if;
    	    // Now, the crucial procedures
	    changeRep :=
		proc(lf:DOM_LIST):DOM_LIST
		begin
		    return([lf[1],DV::index(lf[2]),[],
			    DV::repeatedIndex(lf[2])]);
		end_proc; /* changeRep */
	    changeVars :=
		proc(rep:DOM_LIST)
		local oldVarIndex:DOM_INT, remainingVars:DOM_LIST;
		begin
		    if nops(rep[4]) = 0 then
			return(rep);
		    end_if;
		    oldVarIndex := rep[4][1];
		    rep[4][1] := null();
		    remainingVars := map(rep[4],eval);
		    return( map (
			    [newFromOld[i][oldVarIndex]*rep[1],rep[2],
			     rep[3].[i],remainingVars] $ i=1..nn,
				changeVars));
		end_proc; /* changeVars */
	    changeCoeffs :=
		proc(r)
		begin
		    r := R::expr(r);
		    for i from 1 to nn do
			r := subs(r, 
				  ivars[i] = oldFromNew[i])
		    end_for;
		    return(r);
		end_proc; /* changeCoeffs */
	    changeRepBack :=
		proc(rep:DOM_LIST):DOM_LIST
		begin
		    if nops(rep[3]) = 0 then
			return( [changeCoeffs(rep[1]),dvars[rep[2]]] );
		    else
			return(
			    [changeCoeffs(rep[1]), 
			    dvars[rep[2]](map(rep[3],(x)->(newVars[x])))]
			    );
		    end_if;
		end_proc; /* changeRepBack */
	    // Finally, the actual work
	    rHSide := INH(ldf); 
	    rHSide := changeCoeffs(rHSide);
	    ldf := REP(ldf);
	    ldf := map(ldf, changeRep);
	    ldf := map(ldf, changeVars); 
	    ldf := map(ldf, changeRepBack);
	    return(outputDomain(ldf)+outputDomain(rHSide));
	    end_proc; /* changeIndVars */


    /* ================= */
    /* matrix operations */
    /* ================= */

    jacobianType := SMR;
    

    jacobian :=
        proc():SMR
        begin dom::makeMatrix(args()) end_proc;


    // makeMatrix simply ignores the inhomogeneous part!
    makeMatrix :=
        proc(sys):SMR
	local varlist:DOM_LIST,j:DOM_INT;
        begin
	    if domtype(sys)<>DOM_LIST then
	        return(dom::makeMatrix(map(args(), ((l) -> [l]))));
	    end_if;
	    sys := map(sys,dom::coerce);
	    if testargs() then
		if has(sys,FAIL) then TYPE_ERR end_if;
		if args(0)>1 then
		    varlist := map(args(2),DVcoerce);
		    if not(iszero(contains(varlist,FAIL))) then 
			TYPE_ERR
		    end_if;
		end_if;
	    end_if;
	    if args(0)=1 then
		varlist := sort([op(dom::indets(sys[1],hold(DiffVars)))],
				DVgreater);
		for j from 2 to nops(sys) do
		    varlist := listlib::singleMerge(
			    sort([op(dom::indets(sys[j],hold(DiffVars)))],
				 DVgreater),
			    varlist);
		end_for;
	    else varlist := args(2) end_if;
	    
	    sys := map(map(sys, extop, 1), map, swap);
	    if args(0)=1 then SMR::convert(sys) 
	    else SMR::convert(varlist,sys) end_if;
        end_proc;


    makeSystem :=
        proc(jm:SMR):DOM_LIST
        local tmp:DOM_LIST;
        begin
            if testargs() then
                if args(0)<>1 then ARG_ERR end_if;
                if domtype(jm)<>SMR then TYPE_ERR end_if;
            end_if;
            tmp := map(SMR::rows(jm), map, swap);
            tmp := map(tmp, select, ((l) -> not(Riszero(l[1]))));
            map(tmp, dom::generate);
        end_proc;


    // nthOrderTerms will take an ldf and and an integer n and
    // return those terms with order n
    nthOrderTerms :=
	proc(ldf:dom, n:DOM_INT):dom
	begin
	    if testargs() and (args(0)<>2) then ARG_ERR end_if;
	    ldf := REP(ldf);
	    ldf := select(ldf, ((x,q)-> bool(DVorder(x[2])=q)), n);
	    return(dom::generate(ldf));
	end_proc; /* nthOrderTerms */


    //  nthDepVarTerms will take an ldf and an index n and return that part 
    //  which contains the only the nth dependent variable
    nthDepVarTerms :=
	proc(ldf:dom, n:DOM_INT):dom
	begin
	    if testargs() and (args(0)<>2) then ARG_ERR end_if;
	    ldf := REP(ldf);
	    ldf := select(ldf, ((x,i) -> bool(DVindex(x[2])=i)), n);
	    return(dom::generate(ldf));
	end_proc; /* nthDepVarTerms */


    // principalPart will return the principal part of a linfun
    principalPart :=
	proc(ldf:dom):dom
	begin
	    if testargs() then
		if (args(0)<>1) then ARG_ERR end_if;
	    end_if;
	    return(dom::nthOrderTerms(ldf,dom::order(ldf)));
	end_proc; /* principalPart */


    // symbol will take a list of ldfs and a list of variables (one for each
    // independent variable of DV) and return the symbol
    symbol :=
	proc(sys:DOM_LIST, newvars:DOM_LIST)
	local DOM:DOM_DOMAIN, symbolentry:DOM_PROC, symb:DOM_LIST, ord:DOM_INT,
	      DP:DOM_DOMAIN;
	begin
            sys := map(sys,dom::coerce);
	    if testargs() then
		if nops(newvars) <> nn then 
		    error("Wrong number of new variables");
		end_if;
		if not(iszero(contains(sys,FAIL))) then
		    TYPE_ERR;
		end_if;
	    end_if;
	    DOM := dom;
	    if R::hasProp(Ax::normalRep) then
		DP := Dom::DistributedPolynomial(newvars,R);
	    else
		DP := Dom::Polynomial(R);
	    end_if;
	    symbolentry :=
	        proc(ldf:dom)
		begin
		    ldf := REP(ldf);
		    ldf := map(ldf,(x -> ([x[1],DV::multiIndex(x[2])])));
		    ldf := DP(poly(ldf,newvars,R));
	            return(ldf); 
		end_proc: /* symbolentry */
	    ord := max(op(map(sys,dom::order)));
	    sys := map(sys,dom::nthOrderTerms,ord);
	    symb := [[symbolentry(dom::nthDepVarTerms(sys[i],j))$j=1..mm]
	                                                  $i=1..nops(sys)];
	    return(Dom::Matrix(DP)(symb));
       	end_proc; /* symbol*/



    /* ============= */
    /* miscellaneous */
    /* ============= */

    linear := TRUE;

    quasi := if mode="Less" then TRUE else FALSE end_if;


    constCoeff := ConstCoeff;

	        
    leader := dom::lterm;


    order :=
        if DV::orderly then
	    proc(l:dom):DOM_INT
	    local tmp:DOM_LIST;
	    begin
	        tmp := dom::terms(l);
	        if tmp=[] then 0 else DVorder(tmp[1]) end_if;
	    end_proc
        else
	    proc(l:dom):DOM_INT
	    local tmp:DOM_LIST;
	    begin
	        tmp := dom::terms(l);
	        if tmp=[] then 0 else max(op(map(tmp,DVorder))) end_if;
	    end_proc
        end_if;


    has :=
	if ConstCoeff then
	    proc(l:dom,x):DOM_BOOL
	    local res:DOM_BOOL, cl:DOM_LIST;
	    begin
		if domtype(x)=DV then
		    return(not(iszero(contains(dom::terms(l),x))));
		else
		    res := FALSE;
		    cl := dom::coeff(l);
		    while not(res) and (cl<>[]) do
			res := Rhas(cl[1],x);
			delete cl[1];
		    end_while;
		    return(res);
		end_if;
	    end_proc;
	else
	    proc(l:dom,x):DOM_BOOL
	    local res:DOM_BOOL, cl:DOM_LIST;
	    begin
		if domtype(x)<>DV or testDV(x) then
		    if domtype(x)=DV then x := convRDV(x) end_if;
		    res := FALSE;
		    cl := dom::coeff(l); 
		    while not(res) and (cl<>[]) do
			res := Rhas(cl[1],x);
			delete cl[1];
		    end_while;
		    return(res);
		else 
		    return(not(iszero(contains(dom::terms(l),x))));
		end_if;
	    end_proc;
	end_if;
	
	
    // dVars will take a linfun and return exactly those dependent
    // variables which appear in it.
    dVars :=
	proc(ldf):DOM_SET
	local dv:DOM_SET, v:DV;
	begin
	    ldf := dom::convert(ldf);
	    if testargs() then
		if (args(0)<>1) then ARG_ERR end_if;
		if ldf=FAIL then TYPE_ERR end_if;
	    end_if;
	    dv := {};
	    for v in [op(dom::indets(ldf,hold(DiffVars)))] do
		if contains({"Dep","Deriv"},DVtype(v)) then
		    dv := dv union {depvars[DVindex(v)]};
		end_if;
	    end_for;
	    return(dv);
	end_proc; /* dVars */


    TeX :=
        proc(x:dom,p:DOM_INT):DOM_STRING
	local tmp:DOM_LIST,res:DOM_LIST,i;
        begin
	    if REP(x)=[] then
		return(generate::tex(INH(x),p));
	    end_if;
	    
	    tmp := map(REP(x),dom::TeXMonomial,120);
	    res := tmp[1];
	    for i in [op(tmp,2..nops(tmp))] do
		if i[0]="-" then res := res." ".i
		else res := res." + ".i	end_if;
	    end_for;
	    
	    if not(Riszero(INH(x))) then
		tmp := generate::tex(INH(x),120);
		if tmp[0]="-" then res := res." ".i
		else res := res." + ".i end_if;
	    end_if;
	    
	    generate::TeXquote(res,p,120);
	end_proc;


    // little hack so that the random method from FreeMod does not
    // produce invalid differential variables
    Srandom :=
	proc():DV
	local res:DV;
	begin
	    repeat
		res := DV::random();
	    until not(testDV(res)) end_repeat;
	    res;	    
	end_proc;
	
	
    /* ======================= */
    /* further DFunCat methods */
    /* ======================= */

    // can solve only for those variables in which the element is linear
    solve :=
        proc(l,jv)
        local rl:DOM_LIST, jl:DOM_LIST, c:R, pos:DOM_INT;
        begin
	    if domtype(l)<>dom then l := dom::coerce(l) end_if;
	    if domtype(jv)<>DV then jv := DVcoerce(jv) end_if;
	    if testargs() then
	        if args(0)<>2 then ARG_ERR end_if;
	        if l=FAIL then TYPE_ERR end_if;
	        if jv=FAIL then TYPE_ERR end_if;
	    end_if;

	    rl := REP(l);
	    jl := dom::terms(l);
	    pos := contains(jl,jv);
	    if iszero(pos) then return(FAIL) end_if;
	    if nops(jl)=1 then return(dom::zero) end_if;
	    c := Rinvert(rl[pos][1]);
	    if c=FAIL then return(FAIL) end_if;
	    delete rl[pos];
	    dom::generate(map(rl, 
			      ((k,r) -> [Rnegate(Rmult(k(1)[1],r)),k[2]]), 
			      c),
			  Rnegate(Rmult(INH(l),c)));
        end_proc;


    subs := dom::multiSubs;
    
    
    subsOne :=
        proc(l:dom,eq:DOM_EXPR):dom
        local jl:DOM_LIST, pos:DOM_INT, c:R, jv:DV, ex:R, inh:R;
        begin
	    jv := op(eq,1);
	    ex := op(eq,2);

	    if testDV(jv) then       // substitution in coeffients
	        if ConstCoeff then return(l) end_if;
	        return(dom::mapcoeffs(l,R::subs,convRDV(jv)=ex));
	    end_if;

	    jl := dom::terms(l);
	    pos := contains(jl,jv);
	    if iszero(pos) then return(l) end_if;
	    inh := INH(l);
	    l := REP(l);
	    c := l[pos][1];
	    delete l[pos];
	    delete jl[pos];

	    case domtype(ex)
	    of R do                 // substitution by ring element
	        if not(Riszero(ex)) then
		    inh := Rplus(inh,Rmult(c,ex));
	        end_if;
	        return(dom::generate(l,inh));
	    of DV do                // substitution by differential variable
		pos := contains(jl,ex);
		if iszero(pos) then
		    l := listlib::insert(l,[c,ex],
				   ((l1,l2) -> bool(l1[2]>l2[2])));
		else
		    c := Rplus(l[pos][1],c);
		    if Riszero(c) then delete l[pos]
		    else l[pos][1] := c end_if;
		end_if;
	        return(dom::generate(l,inh));
            of dom do              // substitution by linear function
	        if not(dom::iszero(ex)) then
		    if not(Riszero(INH(ex))) then
			inh := Rplus(inh,Rmult(c,INH(ex)));
		    end_if;
		    ex := map(REP(ex), ((a,b) -> [Rmult(b,a[1]),a[2]]), c);
		    return(dom::generate(dom::plus2(l,ex),inh));
	        end_if;
	        return(dom::generate(l,inh));
	    otherwise TYPE_ERR;
	    end_case;
        end_proc;
	
	
    evalTerm :=
	if mode="Indep" then
	    proc(t:DOM_LIST,vars:DOM_LIST,dersol:DOM_LIST,SDOM:DOM_DOMAIN)
	    local j:DOM_INT,pos:DOM_INT;
	    begin
		j := DVindex(t[2]);
		pos := contains(vars[j],t[2]);
		SDOM::_mult(SDOM::coerce(t[1]),dersol[j][pos]);
	    end_proc;
	else 
	    proc() begin error("not yet implemented") end_proc;
	end_if;
	

    eval :=
	proc(sys,sol)
	local sysflag:DOM_BOOL,res:DOM_LIST,allvars:DOM_LIST,vars:DOM_LIST,
	      dersol:DOM_LIST,DOM:DOM_DOMAIN,i:DOM_INT,SDOM:DOM_DOMAIN;
	begin
	    if domtype(sys)<>DOM_LIST then
		sysflag := FALSE;
		sys := [sys];
	    else sysflag := TRUE end_if;
	    if mm=1 and domtype(sol)<>DOM_LIST then sol := [sol] end_if;
	    SDOM := Dom::DifferentialExpression(
			Dom::RestrictedDifferentialVariable(DV,
						    hold(Types)="Indep"));
	    sys := map(sys,dom::coerce);
	    sol := map(sol,SDOM::coerce);
	    if testargs() then
	        if not(iszero(contains(sys,FAIL))) then TYPE_ERR end_if;
		if domtype(sol)<>DOM_LIST or nops(sol)<>mm then
		    error("wrong number of functions given");
		end_if;
		if not(iszero(contains(sol,FAIL))) then TYPE_ERR end_if;
	    end_if;
	    
	    // get all contained derivatives
	    // get all contained derivatives
	    allvars := _union(op(map(sys,dom::indets,hold(DiffVars))));
	    allvars := select(allvars,
			      ((jv) -> contains({"Dep","Deriv"},DVtype(jv))));
	    // compute derivatives via "derivative tree"
	    vars := [ 0 $ mm ];
	    dersol := [ 0 $ mm ];
	    for i from 1 to mm do
		vars[i] := [op(select(allvars, 
				      ((jv) -> bool(DVindex(jv)=i))))];
		if DV::notation="repeated" then
		    dersol[i] := SDOM::multiDiff(sol[i],
						 map(vars[i],DVrepeated));
		else
		    dersol[i] := SDOM::multiDiff(sol[i],
						 map(vars[i],DVmulti));
		end_if;
	    end_for;
	    // enter derivatives in equations
	    DOM := dom;
	    res := map(sys, 
		       ((eq,v,d) -> SDOM::_plus(
					op(map(REP(eq),DOM::evalTerm,
						v,d,SDOM)))),
		       vars,dersol);
	    if sysflag then res else res[1] end_if;
	end_proc;


    /* ============== */
    /* simplification */
    /* ============== */

    // what about inhomogeneous functions ???
    simplify :=
        proc(sys)
        local M, GE, sysflag:DOM_BOOL;
        begin
	    if domtype(sys)<>DOM_LIST then
		sysflag := FALSE;
		sys := [sys];
	    else sysflag := TRUE end_if;
	    sys := map(sys,dom::coerce);
	    if testargs() then
	        if not(iszero(contains(sys,FAIL))) then TYPE_ERR end_if;
	    end_if;

	    M := dom::makeMatrix(sys);
	    GE := SMR::gaussElim(M,hold(Primitive),hold(All));
	    if sysflag then dom::makeSystem(op(GE,1))
	    else dom::makeSystem(op(GE,1))[1] end_if;
        end_proc;
	
	
    autosimplify :=
	proc() begin dom::simplify(args()) end_proc;


    /* ==================== */
    /* polynomial solutions */
    /* ==================== */
    
    // computes polynomial solutions of constant coefficient system;
    // assumes that system is involutive (i.e. complete);
    // very primitive implementation using brute force;
    // should work similar to eval for derivatives!
    
    makeident :=
	proc(a:DOM_STRING):DOM_IDENT
	local res:DOM_IDENT;
	begin
	    res := genident(a);
	    pvars := pvars union {res};
	    res;
	end_proc;

        
    // generates a polynomial of order N with generic coefficients
    // in the independent variables of DV
    genericpoly :=
	proc(a:DOM_STRING,N:DOM_INT):DOM_POLY
	local ilist:DOM_LIST,DOM:DOM_DOMAIN;
	begin
	    DOM := dom;
	    ilist := [[ 0 $ nn]] . DV::makeIndexList(N);
	    ilist := map(ilist,
		       ((l) -> [DOM::makeident(a)/_mult(op(map(l,fact))),l]));
	    return(poly(ilist,indvars));
	end_proc;

		
    // substitutes polynomial for differential variable
    subspoly :=
	proc(p:DOM_POLY,jv:DV):DOM_POLY
	local l:DOM_LIST, res:DOM_POLY, i:DOM_INT;
	begin
	    l := DV::multiIndex(jv);
	    res := p;
	    for i from 1 to nn do
		res := diff(res,indvars[i] $ l[i]);
		if iszero(res) then return(res) end_if;
	    end_for;
	    return(res);
	end_proc;
	
	
    // substitutes polynomials in linear differential function
    genconds := 
	proc(ldf:dom,vars:DOM_LIST,subsl:DOM_LIST):DOM_POLY
	local t:DOM_LIST,res:DOM_POLY;
	begin
	    res := poly(expr(INH(ldf)),indvars);
	    for t in REP(ldf) do
		res := res+multcoeffs(subsl[contains(vars,t[2])],expr(t[1]));
	    end_for;
	    res;
	end_proc;
	
    
    // the main routine
    polysolve :=
	if ConstCoeff then
	    proc(sys,N:DOM_INT)
	    local a:DOM_STRING,j:DOM_INT,DOM:DOM_DOMAIN,psol,
		  ansatz:DOM_LIST,vars:DOM_LIST,subsl:DOM_LIST,ssys:DOM_LIST;
	    begin
		if domtype(sys)<>DOM_LIST then sys := [sys] end_if;
		sys := map(sys,dom::convert);
		if testargs() then
		    if args(0)<2 or args(0)>3 then ARG_ERR end_if;
		    if not(iszero(contains(sys,FAIL))) then TYPE_ERR end_if;
		    if N<0 then error("degree must be nonnegative") end_if;
		    if args(0)=3 and 
		       domtype(args(3))<>DOM_STRING then TYPE_ERR end_if;
		end_if;
	    
		pvars := {};
		a := (if args(0)=3 then args(3) else "a" end_if);
		ansatz := [dom::genericpoly(a,N) $ j=1..mm];
		vars := [op(_union(map(sys,dom::indets,hold(DiffVars))))];
		DOM := dom;
		subsl := map(vars,((v)->DOM::subspoly(ansatz[DVindex(v)],v)));
		ssys := map(sys,extop,1);
		ssys := {op(map(map(ssys,DOM::genconds,vars,subsl),coeff))};
		psol := linsolve(ssys,pvars);
		if psol=FAIL then FAIL
		else
		    psol := subs(ansatz,op(psol),op(evars));
		    if mm=1 then psol[1] else psol end_if;
		end_if;
	    end_proc;
	else NIL end_if;


begin

    // analysis of passed parameters
    opts := select([args()], (a -> bool(type(a)="_equal")));
    if iszero(nops(opts)) then
	if (args(0)<1) or (args(0)>2) then ARG_ERR end_if;
	DV := args(1);
	if domtype(DV)<>DOM_DOMAIN or DV::hasProp=FAIL or
	   not(DV::hasProp(Cat::DifferentialVariable)) then
	    error("first argument must be Cat::DifferentialVariable");
	end_if;
	if args(0)=1 then 
	    R := Dom::ExpressionField();
	    ConstCoeff := TRUE; 
	else
	    R := args(2);
	    if domtype(R)<>DOM_DOMAIN or R::hasProp=FAIL or
	       not(R::hasProp(Cat::Rng)) then
		error("second argument must be a Cat::Rng");
	    end_if;
	    ConstCoeff := not(R::hasProp(Cat::DifferentialFunction));
	    if not(ConstCoeff) then
		RDV := R::Variables;
		if domtype(RDV)<>DOM_DOMAIN or RDV::hasProp=FAIL or
		   RDV::constructor<>Dom::RestrictedDifferentialVariable or
		   RDV::Variables<>DV then
		    error("invalid coefficient domain");
		end_if;
	    end_if;
	end_if;
    else
	if nops(opts)<>args(0) then error("inconsistent arguments") end_if;
	if args(0)>2 then ARG_ERR end_if;
	if op(opts[1],1)<>hold(Vars) then TYPE_ERR end_if;
	if args(0)=2 and op(opts[2],1)<>hold(Rest) then TYPE_ERR end_if;
	opts := map(opts, 
		    (a -> if domtype(op(a,2))<> DOM_LIST then 
			      subsop(a,2=[op(a,2)]) 
			  else a end_if));
	DV := Dom::DifferentialVariable(op(op(opts[1],2)));
	if args(0)=1 then
	    R := Dom::ExpressionField();
	    ConstCoeff := TRUE;
	else
	    RDV := Dom::RestrictedDifferentialVariable(DV,op(op(opts[2],2)));
	    R := Dom::DifferentialExpression(RDV);
	    ConstCoeff := FALSE;
	end_if;
    end_if;

    // constant or variables coefficients
    if not(ConstCoeff) then
	testDV := RDV::test;
	convDV := RDV::convtoDV;
	convRDV := RDV::convDV;
	mode := RDV::mode[1];
    else
	mode := "Indep";
	testDV := proc(dv:DV):DOM_BOOL
	          begin bool(DVtype(dv)="Indep") end_proc;
	convDV := id;
	convRDV := id;
    end_if;
    
    // methods from DV
    DVcoerce := DV::coerce;
    DVdiff := DV::totalDiff;
    DVindex := DV::index;
    DVmulti := DV::multiIndex;
    DVrepeated := DV::repeatedIndex;
    DVtype := DV::vartype;
    DVorder := DV::order;
    DVconvto := DV::convert_to;
    DVexpr := DV::expr;
    DVgreater := ((j1,j2) -> DV::_less(j2,j1));

    // methods from R
    Rone := R::one;
    Rzero := R::zero;
    Riszero := R::iszero;
    Risone := R::isone;
    Rexpr := R::expr;
    Rcoerce := R::coerce;
    Rmult := R::_mult;
    Rplus := R::_plus;
    Rinvert := R::_invert;
    Rindets := R::indets;
    if Rindets=FAIL then Rindets := indets end_if;
    Rhas := R::has;
    if Rhas=FAIL then Rhas := has end_if;
    Rdiffvars := R::diffVars;
    Rnegate := R::_negate;
    Rdiff := R::diff;
    Rconvto := R::convert_to;
    Rlength := R::length;
    if Rlength=FAIL then Rlength := length end_if;

    // further global methods
    swap := ((l) -> ([l[2],l[1]]));
    remdup := listlib::removeDuplicates;
    SMR := Dom::SparseMatrix(DV,R,DVgreater);
    TyLO := Type::ListOf;
    TyAT := Type::AnyType;
    nn := DV::numIndVar;
    mm := DV::numDepVar;
    indvars := DV::indVariables;
    depvars := DV::depVariables;
    evars := map(indvars, ((v) -> (v=DVexpr(v))));
    pvars := {};

end_domain:

unalias(REP,ARG_ERR,TYPE_ERR):

