/*++
Dom::MonoidOperatorAlgebra(R,S,f) -- free algebra over the ring R with support
                in the ordered monoid S. The function f defines the
                commutator of an operator in S and an element of R (if they
                commute, use Dom::MonoidAlgebra(R,S)).

This domain is a specialization of Dom::FreeModule(R,S). Especially, it
inherits its representation. There are two additional operations: the
multiplication of elements of the domain and the application of an element
of the operator algebra to an element of the coefficient ring.


Author: Werner M. Seiler, Uni Karlsruhe/Mannheim
        (wms@ira.uka.de, seiler@euler.math.uni-mannheim.de)
Created: July 6th, 1997
Last update: 	Thu Mar  4 08:50:40 1999
++*/

alias(REP(x)=extop(x,1)):


domain Dom::MonoidOperatorAlgebra(R,S)

    local Rmult:DOM_PROC, Rplus:DOM_PROC, Riszero:DOM_PROC, Rone:R,
          Smult:DOM_PROC, Sone:S, Scall:DOM_PROC, sigma:DOM_PROC;
    inherits Dom::FreeModule(R,S);
    category Cat::FiniteMonoidRing(R,S), Cat::LeftModule(R);
    axiom
	(if R::hasProp(Ax::canonicalRep) and
	    S::hasProp(Ax::canonicalRep) then Ax::canonicalRep end_if),
        (if R::hasProp(Ax::normalRep) and
	    S::hasProp(Ax::normalRep) then Ax::normalRep end_if);


    one := dom::generate1(Rone,Sone);

    _mult :=
        proc():dom
        local res,x1,x2,i;
        begin
	    case args(0)
	    of 0 do error("wrong number of arguments");
	    of 1 do return(args(1));
	    of 2 do
	        x1 := args(1);
	        x2 := args(2);
	        if contains({R,DOM_INT},domtype(x1)) then
                    if Riszero(x1) then return(dom::zero) end_if;
   		    return(dom::mapcoeffs(x2, ((r1,r2) -> Rmult(r2,r1)), x1));
	        elif contains({R,DOM_INT},domtype(x2)) then
                    if Riszero(x2) then return(dom::zero) end_if;
		    return(dom::mapcoeffs(x1,Rmult,x2));
	        else
		    res := [];
		    for i in REP(x1) do
		        res := res .
		              map(REP(x2),
                                  ((a,b) ->
                                      ([Rmult(b[1],a[1]),Smult(b[2],a[2])],
                                       [Rmult(b[1],sigma(b[2],a[1])),a[2]])),
                                  i);
                    end_for;
		    return(dom::generate(dom::normalizeRep(res)));
	        end_if;
	    otherwise
	        i := args(0) div 2;
	        dom::_mult(dom::_mult(args(1..i)),
			   dom::_mult(args(i+1..args(0))));
	    end_case;
        end_proc;


    applyMonomial := 
	if Scall<>FAIL then
	    proc(x:DOM_LIST,r:R):R
	    begin Rmult(x[1],Scall(x[2],r)) end_proc;
	else NIL end_if;


    func_call := 
	if Scall<>FAIL then
	    proc(x:dom,r:R):R
	    begin Rplus(op(map(REP(x),dom::applyMonomial,r))) end_proc;
	else NIL end_if;


begin

    if args(0)<2 or args(0)>3 then error("wrong number of arguments") end_if;
    if (domtype(R)<>DOM_DOMAIN) or (domtype(S)<>DOM_DOMAIN) or
       (R::hasProp=FAIL) or (S::hasProp=FAIL) then
        error("first two arguments must be domains");
    end_if;
    if R::hasProp(Cat::Ring) then
        Rmult := R::_mult;
        Rplus := R::_plus;
        Riszero := R::iszero;
        Rone := R::one;
    else error("first argument must be a Cat::Ring") end_if;
    if S::hasProp(Cat::OrderedMonoid) then
        Smult := S::_mult;
        Sone := S::one;
    elif S::hasProp(Cat::OrderedAbelianMonoid) then
        Smult := S::_plus;
        Sone := S::zero;
    else error("second argument must be ordered monoid") end_if;
    Scall := S::func_call;

    if args(0)=2 then sigma := Scall
    else sigma := args(3) end_if;

end_domain:

unalias(REP):
