/*++
Dom::PowerProduct(S,order) -- abelian monoid of power products over the ordered
	set S of variables (S can also be a list; the ordering of the variables
	is then determined by the position in the list). The optional parameter
	order is a string naming the ordering used for the products.

Dom::PowerProduct(S,order) defines a Cat::OrderedAbelianMonoid. The
representation depends on the type of S. If S is a list and thus finite,
exponents vectors are used.  Otherwise the representation is a list of pairs
[s,i] with s an element of S and i a positive integer. This list is ordered
according to the ordering defined in S.

Six different orderings are provided:
    LexOrder :       pure lexicographic
    RevLexOrder :    reverse lexicographic
    InvLexOrder :    inverse lexicographic
    DegreeOrder :    total degree, lexicographic
    DegRevLexOrder : total degree, reverse lexicographic
    DegInvLexOrder : total degree, inverse lexicographic

Besides the monoid operations the following methods are implemented:
    degree :    degree(x) gives the total degree,
                degree(x,s1,...,sk) the degree in the variables s1,...,sk
    exponents : list of all non-vanishing exponents
    variables : list of all variables with non-vanishing exponents
    diff :      differentiation with respect to an element of S
                diff(x,s) yields the tuple (n,dx) where n is the
                exponent of s in x and dx is x with the exponent of s
                reduced by 1
    poly :	converts a power product to DOM_POLY

As usually the operation in Dom::PowerProduct(S,order) is understood as a
multiplication, but the operation in a Cat::AbelianMonoid is an addition,
Dom::PowerProduct(S,order) is also a Cat::OrderedMonoid. _mult is defined
as _plus and one as zero.  This is a hack around the basic problem that
traditionally in computer algebra an abelian monoid is *not* a monoid,
as the operation in the former one is addition and in the latter one
multiplication. Thus it is not possible to define an abelian multiplicative
monoid.


Author: Werner M. Seiler, Uni Karlsruhe/Mannheim
        (wms@ira.uka.de, seiler@euler.math.uni-mannheim.de)
Created: July 3rd, 1997
Last update: 	Fri Nov 17 11:46:33 2000
++*/


alias(REP(x)=extop(x,1),
      ARG_ERR=error("wrong no of args"),
      TYPE_ERR=error("wrong type of argument")):


domain Dom::PowerProduct(S, order)

    local dim:DOM_INT, Scoerce:DOM_PROC, Srandom:DOM_PROC, Sexpr:DOM_PROC,
	  insertord:DOM_PROC, TeXexp:DOM_PROC;
    inherits Dom::BaseDomain;
    category Cat::OrderedAbelianMonoid, Cat::OrderedMonoid;
    axiom Ax::canonicalRep;


    // =============== 
    // "local" methods 
    // =============== 

    Sless :=
        if iszero(dim) then 
	    proc(a:S,b:S):DOM_BOOL
	    begin S::_less(a,b) end_proc;
        else
            proc(a,b):DOM_BOOL
            begin bool(contains(S,a)<contains(S,b)) end_proc;
        end_if;


    // sorts and remove duplicates from list 
    // needed only if S is an ordered set 
    normalize :=
        if iszero(dim) then
            proc(x:DOM_LIST):DOM_LIST
	    local i:DOM_INT;
	    begin
	        if x=[] then return(x) end_if;
	        x := sort(x, ((a,b) -> dom::Sless(a[1],b[1])));
	        i := 1;
	        while i<nops(x) do
		    if x[i][1]=x[i+1][1] then
		        x[i][2] := x[i][2]+x[i+1][2];
		        delete x[i+1];
		    else i := i+1 end_if;
	        end_while;
	        return(x);
	    end_proc;
        else NIL end_if;


    // also only needed if S is an ordered set 
    coerceFactor :=
        if iszero(dim) then
            proc(l:DOM_LIST):DOM_LIST
            begin [Scoerce(l[1]),l[2]] end_proc;
        else NIL end_if;


    // ========================= 
    // representation, in/output 
    // ========================= 

    convert :=
        if iszero(dim) then
        proc(x)
            local s,res,e,f,facts;
            begin
		if iszero(x) or x=1 then return(dom::one) end_if;
                case domtype(x)
                of dom do return(x);
                of DOM_LIST do
                    if testtype(x,Type::ListOf(S)) then
                        return(dom::convert(hold(_mult)(op(x))))
                    end_if;
		    if nops(x)=2 and (s:=Scoerce(x[1]))<>FAIL and
		       testtype(x[2],Type::PosInt) then
		        return(new(dom,[[s,x[2]]]));
		    elif {op(map(x,domtype))}<>{DOM_LIST} then return(FAIL)
		    elif {op(map(x,nops))}<>{2} then return(FAIL)
		    else
		        x := map(x,dom::coerceFactor);
		        if has(x,FAIL) then return(FAIL)
		        else return(new(dom,dom::normalize(x))) end_if;
		    end_if;
                of DOM_EXPR do
                    case type(x)
                    of "_mult" do
                        res := [];
                        facts := op(x);
                        for f in facts do
                            case type(f)
                            of "_power" do
                                s := Scoerce(op(f,1));
                                if s=FAIL then return(FAIL) end_if;
                                e := op(f,2);
                                if testtype(e,Type::PosInt) then
                                    res := append(res,[s,e]);
                                elif not(iszero(e)) then return(FAIL) end_if;
                                break;
                            otherwise
                                s := Scoerce(f);
                                if s=FAIL then return(FAIL) end_if;
			        res := append(res,[s,1]);
                                break;
                            end_case;
                        end_for;
                        return(new(dom,dom::normalize(res)));
                    of "_power" do
                        s := Scoerce(op(x,1));
                        if s=FAIL then return(FAIL) end_if;
                        e := op(x,2);
                        if testtype(e,Type::PosInt) then
                            return(new(dom,[[s,e]]));
                        else return(FAIL) end_if;
		    otherwise
                        s := Scoerce(x);
                        if s=FAIL then return(FAIL) end_if;
                        return(new(dom,[[s,1]]));
                    end_case;
                    return(FAIL);
                otherwise
                    s := Scoerce(x);
                    if s=FAIL then return(FAIL) end_if;
                    return(new(dom,[[s,1]]));
                end_case;
                FAIL;
            end_proc;
        else
            proc(x)
            local i,res,e,f,facts;
            begin
		if iszero(x) or x=1 then return(dom::one) end_if;
                case domtype(x)
                of dom do return(x);
                of DOM_LIST do
                    if testtype(x,Type::ListOf(Type::NonNegInt)) and
                       (nops(x)=dim) then return(new(dom,x))
                    else return(dom::convert(hold(_mult)(op(x)))) end_if;
                of DOM_EXPR do
                    case type(x)
                    of "_mult" do
                        res := [ 0 $ dim ];
                        facts := op(x);
                        for f in facts do
                            case type(f)
                            of "_power" do
                                i := contains(S,op(f,1));
                                if iszero(i) then return(FAIL) end_if;
                                e := op(f,2);
                                if testtype(e,Type::PosInt) then
                                    res[i] := res[i]+e;
                                elif not(iszero(e)) then return(FAIL) end_if;
                                break;
                            of DOM_IDENT do
                                i := contains(S,f);
                                if iszero(i) then return(FAIL)
                                else res[i] := res[i]+1 end_if;
                                break;
                            otherwise return(FAIL);
                            end_case;
                        end_for;
                        return(new(dom,res));
                    of "_power" do
                        i := contains(S,op(x,1));
                        if iszero(i) then return(FAIL) end_if;
                        e := op(x,2);
                        if testtype(e,Type::PosInt) or iszero(e) then
                            res := [ 0 $ dim ];
                            res[i] := e;
                            return(new(dom,res));
                        else return(FAIL) end_if;
                    end_case;
                    return(FAIL);
                of DOM_IDENT do
                    i := contains(S,x);
                    if iszero(i) then return(FAIL) end_if;
                    res := [ 0 $ dim ];
                    res[i] := 1;
                    return(new(dom,res));
                end_case;
                FAIL;
            end_proc;
        end_if;


    convert_to :=
        if iszero(dim) then
            proc(x:dom,t)
            begin
                case t
                of DOM_LIST do return(REP(x));
                of DOM_EXPR do return(dom::expr(x));
                of DOM_POLY do return(dom::poly(x))
                of dom do return(x);
	        otherwise return(FAIL);
                end_case;
            end_proc;
        else
	    proc(x:dom,t)
	    local i:DOM_INT, j:DOM_INT, k:DOM_INT, res:DOM_LIST;
	    begin
	        case t
                of DOM_LIST do
                    res := [ 0 $ dim ];
                    i := 1;
                    j := 0;
                    for k in REP(x) do
                        if not(iszero(k)) then
                            j := j+1;
                            res[j] := [S[i],k]
                        end_if;
                        i := i+1;
                    end_for;
                    return([op(res,1..j)])
                of dom do return(x);
                of DOM_EXPR do return(dom::expr(x));
                of DOM_POLY do return(dom::poly(x));
	        otherwise return(FAIL);
                end_case;
            end_proc;
        end_if;


    poly := 
        if iszero(dim) then
	    proc(x:dom):DOM_POLY
	    begin
                if args(0)=1 then 
		    poly([[1,map(REP(x),op,2)]],map(REP(x),op,1))
	        else 
		    poly([[(args(2))::one,map(REP(x),op,2)]],
		          map(REP(x),op,1),args(2))
	        end_if;
	    end_proc;
        else
	    proc(x:dom):DOM_POLY
	    begin
		if args(0)=1 then poly([[1,REP(x)]],S)
	        else poly([[(args(2))::one,REP(x)]],S,args(2))
	        end_if
	    end_proc;
        end_if;


    expr := 
	if iszero(dim) then
	    proc(x:dom):DOM_EXPR
	    local l:DOM_LIST;
	    begin
		if dom::isone(x) then 1
	        elif nops(REP(x))=1 then
		    l := REP(x)[1];
		    if l[2]=1 then l[1] else hold(_power)(op(l)) end_if;
		    else hold(_mult)(op(map(REP(x),
				     ((a) -> (if a[2]=1 then a[1]
					      else hold(_power)(op(a)) 
					      end_if)))));
                end_if;
	    end_proc;
	else
            proc(x:dom):DOM_EXPR
            local i:DOM_INT, j:DOM_INT, k:DOM_INT, res:DOM_LIST;
            begin
                res := [ 0 $ dim ];
                i := 1;
	        j := 0;
                for k in REP(x) do
                    if k>1 then
                        j := j+1;
                        res[j] := hold(_power)(S[i],k);
                    elif k=1 then
                        j := j+1;
                        res[j] := S[i];
                    end_if;
                    i := i+1;
                end_for;
                case j
                of 0 do return(1);
                of 1 do return(res[1]);
                otherwise return(hold(_mult)(op(res,1..j)));
                end_case;
            end_proc;
	end_if;


    // ========= 
    // orderings 
    // ========= 

    _less := 
	case order
	of hold(LexOrder) do       dom::lexless; break;
	of hold(RevLexOrder) do    dom::revlexless; break;
	of hold(InvLexOrder) do    dom::invlexless; break;
	of hold(DegreeOrder) do    dom::deglexless; break;
	of hold(DegRevLexOrder) do dom::degrevlexless; break;
	of hold(DegInvLexOrder) do dom::deginvlexless; break;
	end_case;


    lexless := 
	if iszero(dim) then
	    proc(x1:dom,x2:dom):DOM_BOOL
	    local l1:DOM_LIST, l2:DOM_LIST;
	    begin
		l1 := REP(x1);
		l2 := REP(x2);
		while (l1<>[]) and (l2<>[]) do
		    if dom::Sless(l1[1][1],l2[1][1]) then return(FALSE)
		    elif l1[1][1]=l2[1][1] then
			if l1[1][2]<l2[1][2] then return(TRUE)
			elif l1[1][2]=l2[1][2] then
			    delete l1[1];
			    delete l2[1];
			else return(FALSE) end_if;
		    else return(TRUE) end_if;
		end_while;
		FALSE
	    end_proc;
	else
	    proc(x1:dom,x2:dom):DOM_BOOL
	    local i:DOM_INT, l1:DOM_LIST, l2:DOM_LIST;
	    begin
		l1 := REP(x1);
		l2 := REP(x2);
		i := 1;
		while (i<=dim) and (l1[i]=l2[i]) do i := i+1 end_while;
		if (i>dim) then FALSE else bool(l1[i]<l2[i]) end_if;
	    end_proc;
	end_if;


    revlexless := 
	if iszero(dim) then
	    proc(x1:dom,x2:dom):DOM_BOOL
	    local l1:DOM_LIST, l2:DOM_LIST;
	    begin
		l1 := revert(REP(x1));
		l2 := revert(REP(x2));
		while (l1<>[]) and (l2<>[]) do
		    if dom::Sless(l2[1][1],l1[1][1]) then return(TRUE)
		    elif l1[1][1]=l2[1][1] then
			if l1[1][2]>l2[1][2] then return(TRUE)
			elif l1[1][2]=l2[1][2] then
			    delete l1[1];
			    delete l2[1];
			else return(FALSE) end_if;
		    else return(FALSE) end_if;
		end_while;
		FALSE
	    end_proc;
	else
	    proc(x1:dom,x2:dom):DOM_BOOL
	    local i:DOM_INT, l1:DOM_LIST, l2:DOM_LIST;
	    begin
		l1 := REP(x1);
		l2 := REP(x2);
		i := dim;
		while not(iszero(i)) and (l1[i]=l2[i]) do i := i-1 end_while;
		if iszero(i) then FALSE else bool(l1[i]>l2[i]) end_if;
	    end_proc;
	end_if;


    invlexless := 
	if iszero(dim) then
	    proc(x1:dom,x2:dom):DOM_BOOL
	    local l1:DOM_LIST, l2:DOM_LIST;
	    begin
		l1 := revert(REP(x1));
		l2 := revert(REP(x2));
		while (l1<>[]) and (l2<>[]) do
		    if dom::Sless(l1[1][1],l2[1][1]) then return(TRUE)
		    elif l1[1][1]=l2[1][1] then
			if l1[1][2]<l2[1][2] then return(TRUE)
			elif l1[1][2]=l2[1][2] then
			    delete l1[1];
			    delete l2[1];
			else return(FALSE) end_if;
		    else return(FALSE) end_if;
		end_while;
		FALSE
	    end_proc;
	else
	    proc(x1:dom,x2:dom):DOM_BOOL
	    local i:DOM_INT, l1:DOM_LIST, l2:DOM_LIST;
	    begin
		l1 := REP(x1);
		l2 := REP(x2);
		i := dim;
		while not(iszero(i)) and (l1[i]=l2[i]) do i := i-1 end_while;
		if iszero(i) then FALSE else bool(l1[i]<l2[i]) end_if;
	    end_proc;
	end_if;


    deglexless := 
	proc(x1:dom,x2:dom):DOM_BOOL
	local o1:DOM_INT, o2:DOM_INT;
	begin
	    o1 := dom::degree(x1); 
	    o2 := dom::degree(x2);
	    if o1<o2 then TRUE 
	    elif o1>o2 then FALSE
	    else dom::lexless(x1,x2) 
	    end_if;
	end_proc;


    degrevlexless :=
	proc(x1:dom,x2:dom):DOM_BOOL
	local o1:DOM_INT, o2:DOM_INT;
	begin
	    o1 := dom::degree(x1); 
	    o2 := dom::degree(x2);
	    if o1<o2 then TRUE 
	    elif o1>o2 then FALSE
	    else dom::revlexless(x1,x2) 
	    end_if;
	end_proc;


    deginvlexless :=
	proc(x1:dom,x2:dom):DOM_BOOL
	local o1:DOM_INT, o2:DOM_INT;
	begin
	    o1 := dom::degree(x1); 
	    o2 := dom::degree(x2);
	    if o1<o2 then TRUE 
	    elif o1>o2 then FALSE
	    else dom::invlexless(x1,x2) 
	    end_if;
	end_proc;


    // =========== 
    // arithmetics 
    // =========== 

    zero :=
	if iszero(dim) then new(dom,[])
	else new(dom,[ 0 $ dim ]) end_if;
	
	
    iszero :=
	if iszero(dim) then
	    proc(x:dom):DOM_BOOL
	    begin bool(REP(x)=[]) end_proc;
	else
	    proc(x:dom):DOM_BOOL
	    local i:DOM_INT;
	    begin
		for i in REP(x) do
		    if not(iszero(i)) then return(FALSE) end_if;
		end_for;
		TRUE;
	    end_proc;
	end_if;


    _negate := FAIL;


    _plus :=
	proc():dom
	local l:DOM_LIST, n:DOM_INT;
	begin
             l := map([args()],dom::convert);
             if not(iszero(contains(l,FAIL))) then TYPE_ERR end_if;
             case args(0)
             of 0 do return(dom::zero);
             of 1 do return(l[1]);
             of 2 do return(dom::plus2(l[1],l[2]));
             otherwise
                 n := args(0) div 2;
                 dom::plus2(dom::_plus(op(l,1..n)),
                            dom::_plus(op(l,(n+1)..args(0))));
             end_case;
	 end_proc;


    plus2 := 
	if iszero(dim) then
	    proc(x1:dom,x2:dom):dom
	    local i:DOM_INT, l1:DOM_LIST, l2:DOM_LIST, res:DOM_LIST;
	    begin
		l1 := REP(x1);
		l2 := REP(x2);
		res := [ 0 $ nops(l1)+nops(l2) ];
		i := 0;
		while (l1<>[]) and (l2<>[]) do
		    i := i+1;
		    if dom::Sless(l1[1][1],l2[1][1]) then
			res[i] := l1[1];
			delete l1[1];
		    elif (l1[1][1]=l2[1][1]) then
			res[i] := [l1[1][1],l1[1][2]+l2[1][2]];
			delete l1[1];
			delete l2[1];
		    else
			res[i] := l2[1];
			delete l2[1];
		    end_if;
		end_while;
		if l1<>[] then res := [op(res,1..i)] . l1
		elif l2<>[] then res := [op(res,1..i)] . l2
		else res := [op(res,1..i)] end_if;
		new(dom,res);
	    end_proc;
	else
	    proc(x1:dom,x2:dom):dom
	    begin new(dom,zip(REP(x1),REP(x2),_plus)) end_proc;
	end_if;


    _subtract := 
	if iszero(dim) then
	    proc(x,y):dom
	    local i:DOM_INT, l1:DOM_LIST, l2:DOM_LIST, res:DOM_LIST;
	    begin
		if args(0)<>2 then ARG_ERR end_if;
		x := dom::convert(x);
		if x=FAIL then TYPE_ERR end_if;
		y := dom::convert(y);
		if y=FAIL then TYPE_ERR end_if;
		l1 := REP(x);
		l2 := REP(y);
		res := [ 0 $ nops(l1)+nops(l2) ];
		i := 0;
		while (l1<>[]) and (l2<>[]) do
		    if dom::Sless(l1[1][1],l2[1][1]) then
			i := i+1;
			res[i] := l1[1];
			delete l1[1];
		    elif (l1[1][1]=l2[1][1]) then
			tmp := l1[1][2]-l2[1][2];
			if tmp<0 then return(FAIL) end_if;
			i := i+1;
			res[i] := [l1[1][1],tmp];
			delete l1[1];
			delete l2[1];
		    else return(FAIL) end_if;
		end_while;
		if l2<>[] then return(FAIL) end_if;
		if l1<>[] then res := [op(res,1..i)] . l1
		else res := [op(res,1..i)] end_if;
		return(new(dom,res));
	    end_proc;
	else
	    proc(x,y):dom
	    local res:DOM_LIST, i:DOM_INT, j:DOM_INT;
	    begin
		if args(0)<>2 then ARG_ERR end_if;
		x := dom::convert(x);
		if x=FAIL then TYPE_ERR end_if;
		y := dom::convert(y);
		if y=FAIL then TYPE_ERR end_if;
		res := REP(x);
		j := 0;
		for i in REP(y) do
		    j := j+1;
		    res[j] := res[j]-i;
		    if res[j]<0 then return(FAIL) end_if;
		end_for;
		new(dom,res);
	    end_proc;
	end_if;


    _mult := dom::_plus;


    _invert := dom::_negate;


    _divide := dom::_subtract;


    one := dom::zero;


    isone := dom::iszero;


    // ============= 
    // miscellaneous 
    // ============= 

    degree :=
	if iszero(dim) then
	    proc(x:dom):DOM_INT
	    local l:DOM_LIST, vars:DOM_LIST;
	    begin
		if args(0)=1 then return(_plus(op(dom::exponents(x)))) end_if;
                 vars := map([args(2..args(0))],Scoerce);
                 if not(iszero(contains(vars,FAIL))) then TYPE_ERR end_if;
                 l := dom::variables(args(1));
                 l := map(vars, ((a,b) -> contains(b,a)), l);
                 l := select(l, ((a) -> not(iszero(a))));
                 l := map(l, ((a,b) -> op(b,a)), dom::exponents(x));
                 _plus(op(l));
	    end_proc;
	else
	    proc(x:dom):DOM_INT
	    local l:DOM_LIST;
	    begin
		if args(0)=1 then return(_plus(op(REP(x)))) end_if;
                l :=  map([args(2..args(0))], ((a,b) -> contains(b,a)), S);
		if testargs() and not(iszero(contains(l,0))) then
		    error("illegal variable"); 
		end_if;
                l :=  map(l, ((a,b) -> op(b,a)), REP(x));
                _plus(op(l));
	    end_proc;
	end_if;


    degreevec :=
	if iszero(dim) then NIL
	else
	    proc(x:dom):DOM_LIST
	    begin REP(x) end_proc;
	end_if;
	
	
    variables := 
	if iszero(dim) then
	    proc(x:dom):DOM_LIST
	    begin map(REP(x),op,1) end_proc;
	else
	    proc(x:dom):DOM_LIST
	    local i:DOM_INT, j:DOM_INT, k:DOM_INT, res:DOM_LIST;
	    begin
		res := [ 0 $ dim ];
		i := 1;
		j := 0;
		for k in REP(args(1)) do
		    if not(iszero(k)) then
			j := j+1;
			res[j] := S[i];
		    end_if;
		    i := i+1;
		end_for;
		[op(res,1..j)];
	    end_proc;
	end_if;


    exponents := 
	if iszero(dim) then
	    proc(x:dom):DOM_LIST
	    begin map(REP(x),op,2) end_proc;
	else
	    proc(x:dom):DOM_LIST
	    begin select(REP(x), (i) -> not(iszero(i))) end_proc;
	end_if;


    diff :=
	if iszero(dim) then
	    proc(x:dom)
	    local i:DOM_INT, j:DOM_INT, l:DOM_LIST, lvars:DOM_LIST, 
		  pos:DOM_INT, n:DOM_INT, expv:DOM_LIST, vars:DOM_LIST;
	    begin
		if testargs() then
		    if iszero(args(0)) then ARG_ERR end_if;
		    for i in [args(2..args(0))] do
			if domtype(i)<>S then TYPE_ERR end_if;
		    end_for;
		end_if;

		case args(0)
		of 1 do return(x);
		of 2 do
		    l := REP(x);
		    pos := contains(map(l,op,1),args(2));
		    if iszero(pos) then return((0,dom::zero)) end_if;
		    n := l[pos][2];
		    if n=1 then delete l[pos]
		    else l[pos][2] := n-1 end_if;
		    return((n,new(dom,l)));
		otherwise
		    l := REP(x);
		    lvars := map(l,op,1);
		    expv := [];
		    vars := [];
		    for v in [args(2..args(0))] do
			pos := contains(lvars,v);
			if iszero(pos) then return((0,dom::zero)) end_if;
			n := l[pos][2];
			pos := contains(vars,v);
			if iszero(pos) then
			    vars := insertord(vars,v,dom::Sless);
			    expv := insertord(expv,[v,1],
					    ((a,b) -> dom::Sless(a[1],b[1])));
			else
			    expv[pos][2] := expv[pos][2]+1;
			    if expv[pos][2]>n then return((0,dom::zero)) end_if;
			end_if;
		    end_for;
		    n := 1;
		    i := 1;
		    while i<=nops(l) do
			if l[i][1]<>expv[1][1] then i := i+1;
			else
			    for j from l[i][2] downto l[i][2]-expv[1][2]+1 do
				n := j*n;
			    end_for;
			    l[i][2] := l[1][2]-expv[1][2];
			    delete expv[1];
			    if iszero(l[i][2]) then delete l[i]
			    else i := i+1 end_if;
			end_if;
		    end_while;
		    return((n,new(dom,l)));
		end_case;
	    end_proc;
	else
	    proc(x:dom)
	    local i:DOM_INT, j:DOM_INT, l:DOM_LIST, pos:DOM_INT, n:DOM_INT, 
		  expv:DOM_LIST;
	    begin
		if testargs() then
		    if iszero(args(0)) then ARG_ERR end_if;
		    for i in [args(2..args(0))] do
			if iszero(contains(S,i)) then TYPE_ERR end_if;
		    end_for;
		end_if;

		case args(0)
		of 1 do return(x);
		of 2 do
		    l := REP(x);
		    pos := contains(S,args(2));
		    if iszero(l[pos]) then return((0,dom::zero)) end_if;
		    n := l[pos];
		    l[pos] := n-1;
		    return((n,new(dom,l)));
		otherwise
		    l := REP(x);
		    expv := [ 0 $ dim ];
		    for i in [args(2..args(0))] do
			pos := contains(S,i);
			expv[pos] := expv[pos]+1;
			if expv[pos]>l[pos] then return((0,dom::zero)) end_if;
		    end_for;
		    n := 1;
		    for i from 1 to dim do
			if iszero(expv[i]) then next end_if;
			for j from l[i] downto l[i]-expv[i]+1 do
			    n := j*n;
			end_for;
			l[i] := l[i]-expv[i];
		    end_for;
		    return((n,new(dom,l)));
		end_case;
	    end_proc;
	end_if;


    TeX := 
	if iszero(dim) then
	    proc(x:dom,p:DOM_INT):DOM_STRING
	    local l:DOM_LIST;
	    begin
		l := REP(x);
		case nops(l)
		of 0 do return("1");
		of 1 do return(TeXexp(l[1],l[2],p));
		otherwise 
		    generate::TeXoperator(" ",p,130,
			  op(map(l,((i) -> TeXexp(i[1],i[2],130)))));
		end_case;
 	    end_proc;
 	else
 	    proc(x:dom,p:DOM_INT):DOM_STRING
	    local l:DOM_LIST, j:DOM_INT, n:DOM_INT, res:DOM_STRING;
 	    begin 
		l := REP(x);
		n := nops(select(l,iszero));
		case n
		of dim do return("1");
		of dim-1 do
		    j := 1;
		    while iszero(l[j]) do j := j+1 end_while;
		    return(TeXexp(S[j],l[j],p));
		otherwise
		    res := "";
		    for j from 1 to dim do
			if not(iszero(l[j])) then
			    res := res . " " . TeXexp(S[j],l[j],130);
			end_if;
		    end_for;
		    return(res);
		end_case;
	    end_proc;
	end_if;


    random := 
	if not(iszero(dim)) then
	    proc():dom
	    local i:DOM_INT, j:DOM_INT, r1:DOM_PROC, r2:DOM_PROC, res:DOM_LIST;
	    begin
		r1 := random(1..dim);
		r2 := random(1..5);
		res := [0 $ dim];
		for i from 1 to dim do
		    j := r1();
		    res[j] := res[j]+r2();
		end_for;
		new(dom,res);
	    end_proc;
	elif Srandom<>FAIL then
	    proc():dom
	    local i:DOM_INT, r:DOM_PROC, res:DOM_LIST;
	    begin
		r := random(1..10);
		res := [ 0 $ r()+r() ];
		for i from 1 to nops(res) do
		    res[i] := [Srandom(),r()];
		end_for;
		new(dom,res);
	    end_proc;
	else FAIL end_if;


begin

    if args(0)<1 or args(0)>2 then error("wrong no of args") end_if;
    if domtype(S)=DOM_LIST then
        dim := nops(S);
        if iszero(dim) then error("empty list") end_if;
    elif domtype(S)=DOM_DOMAIN and (S::hasProp<>FAIL) and
         S::hasProp(Cat::OrderedSet) then
        dim := 0;
        Scoerce := S::coerce;
        Srandom := S::random;
        Sexpr := S::expr;
    else error("first argument must be Cat::OrderedSet or list") end_if;

    if args(0)<2 then order := hold(DegreeOrder);
    elif contains({ hold(LexOrder), hold(RevLexOrder), hold(InvLexOrder),
                    hold(DegreeOrder), hold(DegRevLexOrder),
                    hold(DegInvLexOrder) }, order)=0 then
        error("undefined ordering");
    end_if;
    insertord := listlib::insert;
    TeXexp := proc(b:S,e:DOM_INT,p:DOM_INT):DOM_STRING
	      local tmp:DOM_STRING;
	      begin
		  if e=1 then
		      generate::tex(b,p);
		  else
		      tmp := expr2text(e);
		      if length(tmp)<>1 then tmp := "{" . tmp . "}" end_if;
		      generate::TeXquote(generate::tex(b,150)."^".tmp, p, 150);
		  end_if;
	      end_proc;

end_domain:


unalias(REP,TYPE_ERR):

