// $Date: 2000/11/22 18:55:13 $  $Author: stefanw $   $Revision: 1.21 $
// frankp, 2/04/95

/*
    Dom::Real  --  the field of real numbers

    The elements of Real are numbers of type DOM_INT, DOM_RAT, 
    DOM_FLOAT or DOM_EXPR. The test whether an expression is real is 
    performed by calling 'rectform'.

    The domain has category Cat::Field, which is very optimistic 
    (indeed false, because for example bool(1.0 = float(3) / float(3))
    FALSE. The category is assumed only because of pragmatism.
    Note that elements may not have an unic representation, for example
    bool( 0 = sin(2)^2 + cos(2)^2 - 1 ) returns FALSE.
*/

domain Dom::Real
  local selectreals;
  inherits Dom::Complex;
  category Cat::OrderedSet, Cat::DifferentialRing, Cat::Field;
  axiom Ax::systemRep, Ax::canonicalOrder,
        Ax::efficientOperation("_divide"), Ax::efficientOperation("_mult"),
        Ax::efficientOperation("_invert");

// ----------------
// --- Methods: ---
// ----------------

convert:= proc(x)
    local tmp;
begin
    if args(0)<>1 then return(FAIL) end_if;

    case domtype(x)
    of DOM_INT     do
    of DOM_RAT     do
    of DOM_FLOAT   do
        return( x )
    of DOM_COMPLEX do
        return( FAIL )
    of DOM_IDENT do
        if contains( Type::ConstantIdents,x ) then
            return( x )
        else  
            return( FAIL )
        end_if
    of DOM_EXPR do
        if testtype(x, Type::Arithmetical)
           and (tmp:= indets(x)) = {} or tmp minus Type::ConstantIdents = {}
        then
            if iszero(Im(x)) then
                return( Re(x) )
            else
               tmp:= rectform( x );
               if iszero(op(tmp,2)) and op(tmp,3) = 0 and not has(tmp,{hold(Re),hold(Im)}) then
                   return( x )
               end_if
            end_if
        end_if
    otherwise
        return( FAIL )
    end_case
end_proc;

convert_to:= proc(x,T)
begin
    if domtype(x) = T then
        return( x )
    elif T::dom <> DOM_DOMAIN then
        T:= T::dom
    end_if;

    case T
    of DOM_FLOAT do
    of Dom::Float do
        return( float(x) )
    of DOM_COMPLEX do
        return( FAIL )
    of dom do
    of Dom::Complex do
    of Dom::Numerical  do
    of Dom::ArithmeticalExpression do
        return( x )
    of Dom::Integer do
        if domtype(x) = DOM_INT then return( x ) else return( FAIL ) end_if
    of Dom::Rational do
        if domtype(x) = DOM_INT or domtype(x) = DOM_RAT then return( x ) else return( FAIL ) end_if
    otherwise
        if T::constructor = Dom::ExpressionField then
            return(T::coerce( x ))
        else
            return( FAIL )
        end_if
    end_case
end_proc;

_power:= proc(z,n)
begin
    if contains( {DOM_INT,DOM_FLOAT,DOM_RAT}, domtype(n) ) then
        return( z^n )
    elif (n:= dom::coerce(n)) = FAIL then
        error("invalid exponent")
    else
        return( z^n )
    end_if
end_proc;

Im:= 0;

Re:= id;

conjugate:= id;

random:= proc()
    option escape;
    local randgen;
begin
    if args(0) = 0 then 
        return( random() )
    else 
        randgen:= random(args());
        return( () -> randgen() )
    end_if
end_proc;

_less:= _less;

_leequal:= _leequal;

domsolve:=
proc(eqs, vars, options)
begin
  // convert polys to expressions
  eqs:= misc::maprec(eqs, {DOM_POLY}=expr);
  if nops(vars)>1 or contains({DOM_SET, DOM_LIST}, type(eqs)) then
    options["Domain"]:= hold(Expr);
    solve(eqs, vars, options);
    selectreals(%, vars, options);
    if %=FAIL then
      hold(solve)(eqs, vars, solvelib::nonDefaultOptions(options))
    else
      %
    end_if
  elif contains({DOM_SET, DOM_LIST}, type(vars)) then
    // only one variable - remove brackets or braces, respectively
    hold(_in)(op(vars), solve(eqs, op(vars), options))
  else
    // vars and eqs are identifiers
    options["Domain"]:= hold(Expr);
    solve(eqs, vars, options) intersect solvelib::BasicSet(Dom::Real)
  end_if
end_proc;


// ---------------------------
// --- Body of the domain: ---
// ---------------------------
begin

  // local method selectreals to extract the real solutions
  // out of a system
  // solutions may be of type "_in" or in the set-of lists format
  // may fail in some cases !
  
    selectreals:=
    proc(solutions, vars, options)
      local
      checkRhs:DOM_PROC,
      backSubs: DOM_PROC;
    begin

      // local method backSubs
      // backSubs(list1, list2)
      // returns list1.list2, with BackSubstitutions performed in list1

      backSubs:=
      if options(["BackSubstitution"])=FALSE then
        _concat
      else
        proc(l1:DOM_LIST, l2:DOM_LIST): DOM_LIST
        begin
          subs(l1, l2).l2
        end_proc
      end_if;

      
      // local method checkRhs
      // returns a set (piecewise defined or not)
      // equal to {l} or {} depending whether or not
      // is a list of real solutions
      checkRhs:=
      proc(l)
        local i, cond, remainingVars, sol, rhsin;
      begin
        cond:= [];
        for i from 1 to nops(l) do
          case is(rhs(l[i]), Type::Real)
            of FALSE do
              // not a real solution
              return({})
            of UNKNOWN do
              // warning("Could not decide whether ".expr2text(rhs(l[i])).
              //        " is real")
              cond:= cond. [rhs(l[i])];
          end_case;
        end_for;
        if (remainingVars:= indets(cond) intersect {op(vars)})={} then
          // this is, in particular, true for cond=[]
          cond:= _and(op(map(cond, _in, R_)));
          piecewise([cond, {l}], [not cond, {}])
        else
          // we have to solve cond for the remaining variables;
          // this gives further restrictions that have to be appended
          // to l
          cond:= map(cond, Im);
          sol:= solve(cond, remainingVars, Domain=R_);
          case type(sol)
            of "_in" do
              case type((rhsin:=op(sol,2)))
                of solvelib::BasicSet do
                of Dom::ImageSet do
                  if simplify(rhsin)=R_ then
                    return({l})
                  else
              // cannot handle this
                    return(FAIL)
                  end_if;
                of DOM_SET do
                  return({backSubs
                          (l, [op(sol,1)=op(rhsin, i)]) $i=1..nops(rhsin)})
                otherwise
                  return(FAIL)
              end_case
                  
            of DOM_SET do
              if map(sol, type) minus {DOM_LIST} <>{} then
            // cannot handle this at the moment
                FAIL
              else
                {backSubs(l, op(sol,i)) $i=1..nops(sol)}
              end_if
          end_case
        end_if;
          
      end_proc;

      case type(solutions)
        of "solve" do
          return(FAIL)
        of piecewise do
          return(piecewise::map(solutions, selectreals))
        of "_in" do
          return(hold(_in)(op(solutions,1), R_ intersect op(solutions,2)))
      end_case;

      // set-of-lists format

      map(solutions, checkRhs);
      if has(%, FAIL) then
        FAIL
      else
        _union(op(%))
      end_if;
    end_proc;



    // initialization part of domain
  
    if args(0) <> 0 then
        error("no arguments expected")
    end_if
end_domain:

