/*++
Dom::RestrictedDifferentialVariable(DV,crit) - poses restrictions on the
	    differential variables contained in DV
	    
This domain allows to impose restrictions on differential variables. It
takes two arguments: DV is a domain in Cat::DifferentialVariable and
crit defines the restrictions.

There exist five different ways to impose restrictions.
    Types = cset
    Order < q, Order > q, Order = q
    InSet = cset
    OutSet = cset
    Crit = cproc
In the first case only differential variables of the types given in the set 
`cset' are admitted. In the second case the order of the variables must 
satisfy the given inequality. In the next two cases one prescribes either
the complete set of admitted variables or the complete set of all variables
not admitted. The last case is the most general one; here a procedure is
given that takes a differential variables from DV as argument and returns
TRUE or FALSE depending on whether or not the variable is admitted.
 

Author: Werner M. Seiler, Uni Karlsruhe/Mannheim
        (wms@ira.uka.de, seiler@euler.math.uni-mannheim.de)
Created: 	Fri Aug 13 11:31:42 1999
Last update: 	Sat Nov 27 12:35:34 1999
++*/



alias(AD_ERR=error("not admitted differential variable")):


domain Dom::RestrictedDifferentialVariable(DV,crit)

    local q:DOM_INT, mode, cset:DOM_SET, cproc:DOM_PROC;
    inherits DV;
    category Cat::DifferentialVariable;
    axiom (if DV::hasProp(Ax::normalRep) then Ax::normalRep end_if),
	  (if DV::hasProp(Ax::canonicalRep) then Ax::canonicalRep end_if);


    /* ========================= */
    /* conversion to and from DV */
    /* ========================= */
    
    convDV :=
	proc(dv:DV):dom
	begin new(dom,extop(dv)) end_proc;
	
	
    convtoDV :=
	proc(x:dom):DV
	begin new(DV,extop(x)) end_proc;
	

    test :=
	case mode
	of hold(Types) do 
	    proc(dv:DV):DOM_BOOL
	    begin contains(cset,DV::vartype(dv)) end_proc;
	    break;
	of "less" do 
	    proc(dv:DV):DOM_BOOL
	    begin bool(DV::order(dv)<q) end_proc;
	    break;
	of "greater" do 
	    proc(dv:DV):DOM_BOOL
	    begin bool(DV::order(dv)>q) end_proc;
	    break;
	of "equal" do 
	    proc(dv:DV):DOM_BOOL
	    begin bool(DV::order(dv)=q) end_proc;
	    break;
	of hold(InSet) do 
	    proc(dv:DV):DOM_BOOL
	    begin contains(cset,dv) end_proc;
	    break;
	of hold(OutSet) do 
	    proc(dv:DV):DOM_BOOL
	    begin not(contains(cset,dv)) end_proc;
	    break;
	of hold(Crit) do 
	    cproc;
	    break;
	otherwise error("unknown type of restriction");
	end_case;
	

    Variables := DV;
    
    
    mode :=
	case mode
	of hold(Types) do 
	    case cset
	    of {"Indep"} do 
		["Indep"];
		break;
	    of {"Indep","Dep"} do 
		["Less",1];
		break;
	    otherwise 
		["General"];
	    end_case;
	    break;	    
	of "greater" do 
	    ["Greater",q];
	    break;
	of "less" do 
	    ["Less",q];
	    break;
	of "equal" do 
	of hold(InSet) do 
	of hold(OutSet) do 
	of hold(Crit) do 
	    ["General"];
	    break;
	otherwise error("unknown type of restriction");
	end_case;
	
			
    /* ======================= */
    /* generation of variables */
    /* ======================= */
    
    convert :=
	proc()
	local res;
	begin
	    if domtype(args(1))=DV then
		if dom::test(args(1)) then dom::convDV(args(1))
		else FAIL end_if;
	    else
		res := DV::convert(args());
		if res=FAIL then FAIL
		elif dom::test(res) then dom::convDV(res)
		else FAIL end_if;
	    end_if;
	end_proc;
    
    
    convert_to :=
	proc(x:dom,T)
	local dv:DV;
	begin
	    dv := dom::convtoDV(x);
	    if T=DV then dv else DV::convert_to(dv,T) end_if;
	end_proc;
	
	
    new :=
	proc():dom
	local res:DV;
	begin
	    res := DV::new(args());
	    if dom::test(res) then dom::convDV(res) else AD_ERR end_if;
	end_proc;
	
	
    newMulti :=
	proc():dom
	local res;
	begin
	    res := DV::newMulti(args());
	    if dom::test(res) then dom::convDV(res) else AD_ERR end_if;
	end_proc;
	
	
    newRepeated :=
	proc():dom
	local res;
	begin
	    res := DV::newRepeated(args());
	    if dom::test(res) then dom::convDV(res) else AD_ERR end_if;
	end_proc;
	
	
    /* ============= */
    /* miscellaneous */
    /* ============= */
    
    totalDiff :=
	proc(x:dom):dom
	local res:DV, argl:DOM_LIST, DOM:DOM_DOMAIN;
	begin
	    DOM := dom;
	    argl := map([args(2..args(0))],
			(i -> (if domtype(i)=DOM then DOM::convDV(i)
			       else i end_if)));
	    res := DV::totalDiff(dom::convtoDV(x),op(argl));
	    if dom::test(res) then dom::convDV(res) else AD_ERR end_if;
	end_proc;
	
	
    int :=
	proc(x:dom,i):dom
	local res:DV;
	begin
	    if domtype(i)=dom then
		if dom::vartype(i)="Indep" then
		    i := dom::index(i);
		else error("illegal second argument") end_if;
	    end_if;
	    res := DV::int(dom::convtoDV(x),i);
	    if dom::test(res) then dom::convDV(res) else AD_ERR end_if;
	end_proc;
    
    
    indVariables := 
	map(select(DV::indVariables,dom::test),dom::convDV);
    
    
    depVariables := 
	map(select(DV::depVariables,dom::test),dom::convDV);
	
	
    derivatives :=
	proc()
	begin select(DV::derivatives(args()),dom::test) end_proc;
	

    random :=
	proc()
	local res:DV;
	begin
	    repeat
		res := DV::random(args());
	    until dom::test(res) end_repeat;
	    dom::convDV(res);
	end_proc;
	
	
begin

    if args(0)<>2 then error("wrong number of arguments") end_if;

    if domtype(DV)<>DOM_DOMAIN or DV::hasProp=FAIL or
       not(DV::hasProp(Cat::DifferentialVariable)) then
        error("argument must be in Cat::DifferentialVariable");
    end_if;
    
    if not(contains({"_equal","_less","_leequal"},type(crit))) then
	error("illegal second argument");
    end_if;
    
    if contains({"_less","_leequal"},type(crit)) then
	if op(crit,1)=hold(Order) then
	    q := op(crit,2);
	    if domtype(q)<>DOM_INT then error("illegal criterion") end_if;
	    if type(crit)="_leequal" then q := q+1 end_if;
	    if iszero(q) then 
		mode := hold(Types);
		cset := {"Indep"};
	    elif q=1 then
		mode := hold(Types);
		cset := {"Indep","Dep"};
	    else mode := "less" end_if;
	elif op(crit,2)=hold(Order) then
	    q := op(crit,1);
	    if domtype(q)<>DOM_INT then error("illegal criterion") end_if;
	    if type(crit)="_leequal" then q := q-1 end_if;
	    mode := "greater";
	else error("unknown criterion") end_if;
    else
	mode := op(crit,1);
	case mode
	of hold(Order) do
	    q := op(crit,2);
	    if domtype(q)<>DOM_INT then error("illegal criterion") end_if;
	    mode="equal";
	    break;
	of hold(Types) do
	    cset := op(crit,2);
	    if domtype(cset)<>DOM_SET then cset := {cset} end_if;
	    if cset minus {"Indep","Dep","Deriv"} <> {} then
		error("unknown type for differential variables");
	    end_if;
	    break;
	of hold(InSet) do 
	of hold(OutSet) do 
	    cset := op(crit,2);
	    if not(testtype(cset,Type::SetOf(DV))) then
		error("not a set of differential variables");
	    end_if;
	    break;
	of hold(Crit) do
	    cproc := op(crit,2);
	    if domtype(cproc)<>DOM_PROC then
		error("criterion must be procedure");
	    end_if;
	    break;
	otherwise error("unknown criterion");
	end_case;
    end_if;
    
end_domain:


unalias(AD_ERR):
