/*++
Dom::TensorProduct(S,order) -- nonabelian monoid of tensor products over the
        ordered set S of variables (S can also be a list; the ordering of the
        variables is then determined by the position in the list). The
        optional parameter order is a string naming the ordering used for the
        products.

Dom::TensorProduct(S,order) defines a Cat::OrderedMonoid. The representation
is a list of elements of S, if S is a domain, and a list of integers denoting
the position in the list S otherwise.

Two different orderings are provided:
    DegreeOrder :    total degree, lexicographic
    DegRevLexOrder : total degree, reverse lexicographic
Purely lexicographic orderings are not admissable, as they do not respect the
monoid structure! InvLex and RevLex are identical here.

Besides the monoid operations the following methods are implemented:
    degree :    degree(x) gives the total degree,
                degree(x,s1,...,sk) the degree in the variables s1,...,sk
    variables : list of all variables occuring in the product
    revert :    revert order of factors
    diff :	differentiation wrt to elements of S
    		returns a list of tensor products!
    has :       checks whether one product is contained in another one
    subs :      overloads the standard MuPAD subs


Author: Werner M. Seiler, Uni Karlsruhe/Mannheim
        (wms@ira.uka.de, seiler@euler.math.uni-mannheim.de)
Created: March 14th, 1998
Last update: 	Fri Nov 17 11:54:33 2000
++*/

alias(REP(x)=extop(x,1),
      ARG_ERR=error("wrong number of arguments"),
      TYPE_ERR=error("wrong type of argument")):


domain Dom::TensorProduct(S,order)

    local dim:DOM_INT, remDup:DOM_PROC, sublist:DOM_PROC,
          Scoerce:DOM_PROC, Srandom:DOM_PROC;
    inherits Dom::BaseDomain;
    category Cat::OrderedMonoid;
    axiom Ax::canonicalRep;


    // =============== 
    // "local" methods 
    // =============== 

    Sless :=
        if iszero(dim) then S::_less else _less end_if;


    Sexpr :=
        if iszero(dim) then S::expr
        else
            proc(a):Type::Arithmetical
            begin S[a] end_proc;
        end_if;


    // only needed if S is an ordered set 
    coerceFactor :=
        if iszero(dim) then
            proc(a:DOM_LIST):DOM_LIST
            begin [Scoerce(a[1]),a[2]] end_proc;
        else NIL end_if;


    // ========================= 
    // representation, in/output 
    // ========================= 

    convert :=
        if iszero(dim) then
            proc(x)
            local s,res,e,f,facts;
            begin
                case domtype(x)
                of dom do return(x);
                of DOM_LIST do
                    x := map(x,Scoerce);
                    if iszero(contains(x,FAIL)) then return(new(dom,x))
                    else return(FAIL) end_if;
                of DOM_EXPR do
                    case type(x)
                    of "_mult" do
                        res := [];
                        facts := op(x);
                        for f in facts do
                            if type(f)="_power" then
                                e := Scoerce(op(f,1));
                                if e=FAIL then return(FAIL) end_if;
                                i := op(f,2);
                                if testtype(i,Type::PosInt) then
                                    res := res . [ e $ i ];
                                else return(FAIL) end_if;
                            else
                                e := Scoerce(f);
                                if e=FAIL then return(FAIL) end_if;
                                res := append(res,e);
                            end_if;
                        end_for;
                        return(new(dom,res));
                    of "_power" do
                        f := Scoerce(op(x,1));
                        if f=FAIL then return(FAIL) end_if;
                        i := op(x,2);
                        if testtype(i,Type::PosInt) then
                            return(new(dom,[ f $ i ]));
                        else return(FAIL) end_if;
		    of "function" do
		        if op(x,0)=hold(tensor) then
			    res := map([op(x)],Scoerce);
			    if iszero(contains(res,FAIL)) then 
				return(new(dom,res));
			    else return(FAIL) end_if;
		        else return(FAIL) end_if;
                    end_case;
                    return(FAIL);
                otherwise
                    s := Scoerce(x);
                    if s=FAIL then return(FAIL) end_if;
                     return(new(dom,[s]));
                end_case;
                FAIL;
            end_proc;
        else
            proc(x)
            local i,res,e,f,facts;
            begin
                case domtype(x)
                of dom do return(x);
                of DOM_LIST do
                    res := map(x, ((a) -> contains(S,a)));
                    if not(iszero(contains(res,0))) then return(FAIL)
                    else return(new(dom,res)) end_if;
                of DOM_EXPR do
                    case type(x)
                    of "_mult" do
                        res := [];
                        facts := op(x);
                        for f in facts do
                            case type(f)
                            of "_power" do
                                e := contains(S,op(f,1));
                                if iszero(e) then return(FAIL) end_if;
                                i := op(f,2);
                                if testtype(i,Type::PosInt) then
                                    res := res . [ e $ i ];
                                else return(FAIL) end_if;
                                break;
                            of DOM_IDENT do
                                i := contains(S,f);
                                if iszero(i) then return(FAIL)
                                else res := append(res,i) end_if;
                                break;
                            otherwise return(FAIL);
                            end_case;
                        end_for;
                        return(new(dom,res));
                    of "_power" do
                        f := contains(S,op(x,1));
                        if iszero(f) then return(FAIL) end_if;
                        i := op(x,2);
                        if testtype(i,Type::PosInt) then
                            return(new(dom,[ f $ i ]));
                        else return(FAIL) end_if;
		    of "function" do
		        if op(x,0)=hold(tensor) then
			    res := map([op(x)], ((a,b) -> contains(b,a)), S);
			    if iszero(contains(res,0)) then return(new(dom,res))
			    else return(FAIL) end_if;
		        else return(FAIL) end_if;
                    end_case;
                    return(FAIL);
                of DOM_IDENT do
                    i := contains(S,x);
                    if iszero(i) then return(FAIL) end_if;
                    return(new(dom,[i]));
                end_case;
                FAIL;
            end_proc;
        end_if;


    convert_to :=
        if iszero(dim) then
            proc(x:dom,T)
            begin
                case T
                of DOM_LIST do return(REP(x));
                of DOM_EXPR do return(dom::expr(x));
                of DOM_POLY do return(dom::poly(x));
                of dom do return(args(1));
	        otherwise return(FAIL);
                end_case;
            end_proc;
        else
            proc(x:dom,T)
            begin
                case T
                of DOM_LIST do return(map(REP(x),dom::Sexpr));
                of DOM_EXPR do return(dom::expr(x));
                of DOM_POLY do return(dom::poly(x));
                of dom do return(args(1));
	        otherwise return(FAIL);
                end_case;
            end_proc;
        end_if;


    poly := 
	if iszero(dim) then
	    proc(x:dom):DOM_POLY
	    begin poly(_mult(op(REP(x))),dom::variables(x)) end_proc;
	else
	    proc(x:dom):DOM_POLY
	    begin poly(_mult(op(map(REP(x),((i)->S[i])))),S) end_proc;
	end_if;


    expr :=
        proc(x:dom):Type::Arithmetical
        begin
            if dom::isone(x) then 1
            elif nops(REP(x))=1 then dom::Sexpr(REP(x)[1])
            else hold(tensor)(op(map(REP(x),dom::Sexpr))) end_if;
        end_proc;


    // ========= 
    // orderings 
    // ========= 

    _less :=
        case order
	of hold(DegreeOrder) do    dom::deglexless; break;
	of hold(DegRevLexOrder) do dom::degrevlexless; break;
        end_case;


    lexless :=
        proc(x1:dom,x2:dom):DOM_BOOL
        local l1:DOM_LIST, l2:DOM_LIST;
        begin
            l1 := REP(x1);
            l2 := REP(x2);
            while (l1<>[]) and (l2<>[]) do
                if dom::Sless(l1[1],l2[1]) then return(FALSE)
                elif l1[1]=l2[1] then
                    delete l1[1];
                    delete l2[1];
                else return(TRUE) end_if;
            end_while;
            FALSE
        end_proc;


    revlexless :=
        proc(x1:dom,x2:dom):DOM_BOOL
        local l1:DOM_LIST, l2:DOM_LIST;
        begin
            l1 := revert(REP(x1));
            l2 := revert(REP(x2));
            while (l1<>[]) and (l2<>[]) do
                if dom::Sless(l2[1],l1[1]) then return(TRUE)
                elif l1[1]=l2[1] then
                    delete l1[1];
                    delete l2[1];
                else return(FALSE) end_if;
            end_while;
            FALSE
        end_proc;


    deglexless :=
        proc(x1:dom,x2:dom):DOM_BOOL
        local o1:DOM_INT, o2:DOM_INT;
        begin
            o1 := dom::degree(x1);
            o2 := dom::degree(x2);
	    if o1<o2 then TRUE
	    elif o1>o2 then FALSE
	    else dom::lexless(args()) end_if;
	end_proc;


    degrevlexless :=
        proc(x1:dom,x2:dom):DOM_BOOL
        local o1:DOM_INT, o2:DOM_INT;
        begin
            o1 := dom::degree(x1);
            o2 := dom::degree(x2);
	    if o1<o2 then TRUE
	    elif o1>o2 then FALSE
	    else dom::revlexless(args()) end_if;
	end_proc;


    // =========== 
    // arithmetics 
    // =========== 

    _mult :=
        proc():dom
        local l:DOM_LIST;
        begin
            l := map([args()],dom::convert);
	    if testargs() then
                if not(iszero(contains(l,FAIL))) then TYPE_ERR end_if;
            end_if;
            case args(0)
            of 0 do return(dom::one);
            of 1 do return(l[1]);
            otherwise return(new(dom,map(map(l,extop,1),op)));
            end_case;
        end_proc;


    tensor := dom::_mult;


    /*_divide := (()->((
        sublist(REP(args(1)),REP(args(2)),hold(Consecutive));
        if iszero(%) then return(FAIL)
        else
            [REP(args(1))[i] $ i=1..(%-1)] .
            [REP(args(1))[i] $ i=(%+nops(REP(args(2))))..
                                       nops(REP(args(1)))];
            new(dom,%);
        end_if)));*/
    _divide := proc() begin FAIL end_proc;
    
    
    _invert := proc() begin FAIL end_proc;


    one := new(dom,[]);


    isone :=
        proc(x:dom):DOM_BOOL
        begin bool(REP(x)=[]) end_proc;


    revert :=
        proc(x:dom):dom
        begin new(dom,revert(REP(x))) end_proc;


    // ============= 
    // miscellaneous 
    // ============= 

    degree :=
        if iszero(dim) then
            proc(x:dom):DOM_INT
            local l:DOM_LIST;
            begin
                if args(0)=1 then return(nops(REP(x)))
                else
		    l := map([args(2..args(0))],Scoerce);
                    if testargs() then
                        if not(iszero(contains(l,FAIL))) then TYPE_ERR end_if;
                    end_if;
                    l := select(REP(x),
                                ((a,b) -> not(iszero(contains(b,a)))), l);
                    nops(l);
                end_if;
            end_proc;
        else
            proc(x:dom):DOM_INT
            local l:DOM_LIST;
            begin
                if args(0)=1 then return(nops(REP(x)))
                else
                    l := map([args(2..args(0))], ((a,b) -> contains(b,a)), S);
                    if iszero(contains(l,0)) then
                        l := select(REP(x),
                                    ((a,b) -> not(iszero(contains(b,a)))), l);
                        nops(l);
                    else error("illegal variable") end_if;
                end_if;
            end_proc;
        end_if;


    variables :=
	if iszero(dim) then
	    proc(x:dom):DOM_LIST
	    begin sort(remDup(REP(x)),dom::Sless) end_proc;
	else
	    proc(x:dom):DOM_LIST
	    begin 
		map(sort(remDup(REP(x)),dom::Sless), ((i)->S[i]));
	    end_proc;
	end_if;


    has :=
        if iszero(dim) then
            proc(x:dom,s):DOM_BOOL
            local l:DOM_INT;
            begin
                case domtype(s)
                of S do return(not(iszero(contains(dom::variables(x),s))));
                of dom do
                    l := sublist(REP(x),REP(s),hold(Consecutive));
                    return(not(iszero(l)));
                otherwise return(FALSE);
                end_case;
            end_proc;
        else
            proc(x:dom,s):DOM_BOOL
            local l:DOM_INT;
            begin
                if not(iszero(contains(S,s))) then
                    not(iszero(contains(dom::variables(x),s)));
                elif domtype(s)<>dom then FALSE;
                else
                    l := sublist(REP(x),REP(s),hold(Consecutive));
                    not(iszero(l));
                end_if;
            end_proc;
        end_if;


    subs :=
        proc(x):dom
        local i, eq, eqs:DOM_LIST, l:DOM_LIST,
              n1:DOM_INT, n2:DOM_INT, nl:DOM_INT, pos:DOM_INT;
        begin
            eqs := [];
            for i from 2 to args(0) do
                case type(args(i))
                of "_equal" do
                    eqs := append(eqs,args(i));
                    break;
                of DOM_LIST do
                    eqs := eqs . args(i);
                    break;
                of DOM_SET do
                of DOM_TABLE do
                    eqs := eqs . [op(args(i))];
                    break;
                of DOM_IDENT do
                    if args(i)=hold(Unsimplified) then break end_if;
                otherwise TYPE_ERR;
                end_case;
            end_for;

            x := dom::coerce(x);
            if x=FAIL then TYPE_ERR end_if;
            for i from 1 to nops(eqs) do
                if type(eqs[i])<>"_equal" then TYPE_ERR end_if;
                eqs[i] := subsop(eqs[i],1=dom::coerce(op(eqs[i],1)),
                                        2=dom::coerce(op(eqs[i],2)));
                if has(eqs[i],FAIL) then TYPE_ERR end_if;
            end_for;

            l := REP(x);
            nl := nops(l);
            eqs := map(eqs,(()->(REP(op(args(1),1))=REP(op(args(1),2)))));
            for eq in eqs do
                n1 := nops(op(eq,1));
                n2 := nops(op(eq,2));
                pos := sublist(l,op(eq,1),hold(Consecutive));
                while not(iszero(pos)) do
                    l := [op(l,1..(pos-1))] . op(eq,2) . [op(l,(pos+n1)..nl)];
                    nl := nl-n1+n2;
                    pos := sublist(l,op(eq,1),pos+n2,hold(Consecutive));
                end_while;
            end_for;
            new(dom,l);
        end_proc;


    // diff returns a list!
    // An empty list means zero; otherwise the real result is given by the
    // sum of the list members.
    diff :=
        if iszero(dim) then
            proc(x):DOM_LIST
            local i:DOM_INT, j:DOM_INT, l:DOM_LIST, vars:DOM_LIST,
                  res:DOM_LIST;
            begin
       	    	x := dom::convert(x);
       	    	if x=FAIL then TYPE_ERR end_if;
       	    	vars := map([args(2..args(0))],Scoerce);
       	    	if not(iszero(contains(vars,FAIL))) then TYPE_ERR end_if;
       	        
       	        case args(0)
       	        of 1 do return(x);
       	        of 2 do
       	    	    l := REP(x);
       	    	    i := dom::degree(x,vars[1]);
       	    	    if iszero(i) then return([]) end_if;
       	    	    res := [l $ i];
       	    	    i := contains(l,vars[1]);
       	    	    j := 0;
       	    	    while not(iszero(i)) do
       	    	        j := j+1;
       	    	        delete res[j][i];
       	    	        i := contains(l,vars[1],i+1);
       	    	    end_while;
       	    	    return(map(res, ((a,TP) -> new(TP,a)), dom));
       	        otherwise
       	    	    res := [x];
       	    	    for i in vars do
       	    	        res := map(res, ((a,b,TP) -> op(TP::diff(a,b))), 
				   i, dom);
       	    	    end_for;
       	    	    return(res);
       	        end_case;
            end_proc;
        else
            proc(x:dom):DOM_LIST
            local i:DOM_INT, j:DOM_INT, k:DOM_INT, l:DOM_LIST, res:DOM_LIST;
            begin
       	        if testargs() then
       	    	    if args(0)<1 then ARG_ERR end_if;
       	    	    for i in [args(2..args(0))] do
       	    	        if iszero(contains(S,i)) then TYPE_ERR end_if;
       	    	    end_for;
       	        end_if;

       	        case args(0)
       	        of 1 do return(x);
       	        of 2 do
       	    	    l := REP(x);
       	    	    i := dom::degree(x,args(2));
       	    	    if iszero(i) then return([]) end_if;
         	    k := contains(S,args(2));
       	    	    res := [l $ i];
       	    	    i := contains(l,k);
       	    	    j := 0;
       	    	    while not(iszero(i)) do
       	    	        j := j+1;
       	    	        delete res[j][i];
       	    	        i := contains(l,k,i+1);
       	    	    end_while;
       	    	    return(map(res, ((a,TP) -> new(TP,a)), dom));
       	        otherwise
       	    	    res := [x];
       	    	    for i in [args(2..args(0))] do
       	    	        res := map(res, ((a,b,TP) -> op(TP::diff(a,b))), 
				   i, dom);
       	    	    end_for;
       	    	    return(res);
       	        end_case;
            end_proc;
        end_if;


    TeX :=
	proc(x:dom,p:DOM_INT):DOM_STRING
	begin 
	    if REP(x)=[] then "1"
	    else generate::TeXoperator("\\otimes ",p,150,op(REP(x))) end_if;
	end_proc;


    random :=
        if not(iszero(dim)) then
	    proc(N=10:DOM_INT):dom
	    local i:DOM_INT, r1:DOM_PROC, r2:DOM_PROC, res:DOM_LIST;
	    begin
	        r1 := random(1..dim);
	        r2 := random(1..N);
	        res := [0 $ r2()];
	        for i from 1 to nops(res) do
		    res[i] := r1();
	        end_for;
	        new(dom,res);
	    end_proc;
        elif Srandom<>FAIL then
	    proc(N=10:DOM_INT):dom
	    local i:DOM_INT, r:DOM_PROC, res:DOM_LIST;
	    begin
	        r := random(1..N);
	        res := [ 0 $ r()];
	        for i from 1 to nops(res) do
		    res[i] := Srandom();
	        end_for;
	        new(dom,res);
	    end_proc;
        else FAIL end_if;


begin

    if args(0)<1 or args(0)>2 then ARG_ERR end_if;
    if domtype(S)=DOM_LIST then
        dim := nops(S);
        if iszero(dim) then error("empty list") end_if;
    elif domtype(S)=DOM_DOMAIN and (S::hasProp<>FAIL) and
         S::hasProp(Cat::OrderedSet) then
        dim := 0;
        Scoerce := S::coerce;
        Srandom := S::random;
    else error("first argument must be Cat::OrderedSet or list") end_if;

    if args(0)<2 then
        order := hold(DegreeOrder);
    elif iszero(contains({ hold(DegreeOrder),
                          hold(DegRevLexOrder) }, order)) then
        error("undefined ordering")
    end_if;
    remDup := listlib::removeDuplicates;
    sublist := listlib::sublist;

end_domain:

unalias(REP,ARG_ERR,TYPE_ERR):
