/*++
Dom::VectorField(R,S,OR) -- vector fields on a manifold with coefficients in a
        ring R; S contains the coordinates of the manifold; the optional
	argument OR denotes a ring on which the vector fields operates - the
	default is OR=R

Dom::VectorFields(R,S,OR) provides a representation for vector fields on a
manifold.  The coefficient functions come from a partial differential ring R. S
is either an ordered set or a list with the ordering defined by the position in
the list.  Its elements are the coordinates of the manifold. Thus it should be
possible to differentiate elements of R with respect to elements of S.

If OR is different from R, then the following operations must be contained in 
these domains: it must be possible to differentiate elements of R with
respect to elements of S and a multiplication must be defined for elements
from R and OR. Otherwise, func_call will not return a meaningful results.

The implementation is based on Dom::FreeModule(R,S) from which all arithmetic
functions and the representation are inherited. In addition the following
geometric operations are defined:
    commutator(v1,v2) : the commutator of two vector fields
    comTable(l) :       commutator table for a list of vector fields
    lie(v1,v2) :        the Lie derivative of v2 with respect to v1
                        (is just commutator(v1,v2))
    func_call(v,r) :    application of a vector field to an element of O
Components of vector fields can be accessed with usual index notations.

Dom::VectorField(R,S) is a Cat::Rng with multiplication defined by the
commutator.

Author: Werner M. Seiler, Uni Karlsruhe/Mannheim
        (wms@ira.uka.de, seiler@euler.math.uni-mannheim.de)
Created: July 18th, 1997
Last update: 	Fri Aug 25 15:35:17 2000
++*/
// 13.10.2000

alias(REP(x)=extop(x,1),
      TYPE_ERR=error("wrong type of argument")):


domain Dom::VectorField(R,S)

    local Rcoerce:DOM_PROC, Rplus:DOM_PROC, Rnegate:DOM_PROC, Rdiff:DOM_PROC,
	  Rmult:DOM_PROC, Rzero:R, Risone:DOM_PROC, Riszero:DOM_PROC, 
	  Rexpr:DOM_PROC, Rone:DOM_PROC, Sexpr:DOM_PROC, finite:DOM_BOOL,
	  OR:DOM_DOMAIN, Ocoerce:DOM_PROC, Odiff:DOM_PROC, 
	  Oplus:DOM_PROC, Omult:DOM_PROC;
    inherits Dom::FreeModule(R,S);
    category Cat::LeftModule(R), Cat::Rng;
    axiom (if R::hasProp(Ax::canonicalRep) then Ax::canonicalRep end_if),
          (if R::hasProp(Ax::normalRep) then Ax::normalRep end_if);


    /* =============== */
    /* "local" methods */
    /* =============== */

    Sless :=
	if finite then
	    proc(a,b):DOM_BOOL
	    begin bool(contains(S,a)<contains(S,b)) end_proc;
	else 
	    proc(a:S,b:S):DOM_BOOL
	    begin S::_less(args()) end_proc;
	end_if;


    /* ========================= */
    /* representation, in/output */
    /* ========================= */

    convert :=
        proc(x)
        local r, s, f, flag:DOM_BOOL, res:dom, rres:DOM_LIST;
        begin
	    case domtype(x)
	    of dom do return(x);
	    of DOM_LIST do
	        if nops(x)=2 and (r:=Rcoerce(x[1]))<>FAIL and
			         (s:=dom::Scoerce(x[2]))<>FAIL then
		    if Riszero(r) then return(dom::zero)
		    else return(dom::generate1(r,s)) end_if;
	        elif {op(map(x,domtype))}<>{DOM_LIST} then return(FAIL)
	        elif {op(map(x,nops))}<>{2} then return(FAIL)
	        else
		    x := map(x,dom::coerceSummand);
		    if has(x,FAIL) then return(FAIL)
		    else return(dom::generate(dom::normalizeRep(x))) end_if;
	        end_if;
	    of DOM_EXPR do
	        x := expand(x);
	        case type(x)
	        of "_plus" do
		    res := dom::zero;
		    for r in op(x) do
		        s := dom::convert(r);
		        if s=FAIL then return(FAIL) end_if;
		        res := dom::_plus(res,s);
		    end_for;
		    return(res);
	        of "_mult" do
		    rres := [];
		    flag := TRUE;
		    for f in op(x) do
		        if type(f)="D" and nops(f)=1 then
		            s := dom::Scoerce(op(f,1));
		            if s=FAIL then return(FAIL) end_if;
		            if flag then flag := FALSE
		            else return(FAIL) end_if;
		            next;
		        end_if;
		        r := Rcoerce(f);
		        if r<>FAIL then
			    rres := append(rres,r);
			    next;
		        end_if;
		        return(FAIL);
		    end_for;
		    r := if rres=[] then Rone else Rmult(op(rres)) end_if;
		    return(dom::generate1(r,s));
		of "D" do
		    if nops(x)<>1 then return(FAIL) end_if;
		    s := dom::Scoerce(op(x,1));
		    if s=FAIL then return(FAIL) end_if;
		    return(dom::generate1(Rone,s));
		otherwise return(FAIL);
	        end_case;
	    of DOM_POLY do return(dom::convert(expr(x)));
	    of DOM_ARRAY do return(dom::convert([op(x)]));
	    otherwise
	        if x=0 then return(dom::zero) end_if;
	        if x=1 and dom::one<>FAIL then return(dom::one) end_if;
		if ((x::dom)::constructor=dom::constructor) and 
		   ((x::dom)::Variables=S) then
		    if (x::dom)::coeffRing=R then 
			return(dom::generate(REP(x)));
		    else
			return(dom::convert(REP(x)));
		    end_if;
		end_if;
	        return(FAIL);
	    end_case;
        end_proc;


    exprMonomial :=
        proc(l:DOM_LIST):DOM_EXPR
        begin
            if Risone(l[1]) then hold(D)(Sexpr(l[2]))
            else hold(_mult)(Rexpr(l[1]),hold(D)(Sexpr(l[2]))) end_if;
        end_proc;


    TeXMonomial :=
	proc(l:DOM_LIST,p:DOM_INT):DOM_STRING
	begin 
	    if Risone(l[1]) then "\\partial_{".generate::tex(l[2],200)."}"
	    else 
		generate::TeXquote(
		    _concat(generate::tex(l[1],130)," ",
			"\\partial_{".generate::tex(l[2],200)."}"),
		    p,130);
	    end_if;
	end_proc;


    _index :=
        proc(v:dom,s)
        begin dom::coeff(v,s) end_proc;


    set_index :=
        proc(v:dom,s,r:R):dom
        local l:DOM_LIST, n:DOM_INT, i:DOM_INT;
        begin
	    s := dom::Scoerce(s);
	    if testargs() then
	        if s=FAIL then TYPE_ERR end_if;
	    end_if;

	    l := REP(v);
	    n := contains(dom::terms(v),s);
	    if iszero(n) then
	        i := 1;
	        n := nops(l);
	        while (i<=n) and dom::Sless(s,l[i][2]) do i := i+1 end_while;
	        if i>n then l := append(l,[r,s])
	        else l[i] := ([r,s],l[i]) end_if;
	    else
	        if Riszero(r) then delete l[n] else l[n][1] := r end_if;
	    end_if;
	    dom::generate(l);
        end_proc;


    /* ====================== */
    /* "geometric" operations */
    /* ====================== */

    commutator :=
        proc(v1:dom,v2:dom):dom
        local l1:DOM_LIST, l2:DOM_LIST, n1:DOM_INT, n2:DOM_INT,
              i:DOM_INT, j:DOM_INT, s:DOM_LIST, c:R, res:DOM_LIST;
        begin
            l1 := REP(v1);
            l2 := REP(v2);
            n1 := nops(l1);
            n2 := nops(l2);

            res := [ 0 $ n1+n2 ];
	    j := 0;
            for s in l2 do
                c := Rplus(Rmult(l1[i][1],Rdiff(s[1],l1[i][2])) $ i=1..n1);
                if not(Riszero(c)) then
		    j := j+1;
		    res[j] := [c,s[2]];
		end_if;
            end_for;

            for s in l1 do
                c := Rplus(Rmult(l2[i][1],Rdiff(s[1],l2[i][2])) $ i=1..n2);
                if not(Riszero(c)) then
		    j := j+1;
                    res[j] := [Rnegate(c),s[2]];
                end_if;
            end_for;

            dom::generate(dom::normalizeRep([op(res,1..j)]));
        end_proc;


    comTable :=
        proc(l:DOM_LIST):DOM_ARRAY
        local res:DOM_ARRAY, n:DOM_INT, i:DOM_INT, j:DOM_INT, v1:dom, v2:dom;
        begin
            if testargs() then
                if not(testtype(l,Type::ListOf(dom))) then TYPE_ERR end_if;
                if nops(l)<2 then
                    error("list must have at least two entries");
                end_if;
            end_if;

            n := nops(l);
            res := array(1..n,1..n);
            for i from 1 to n-1 do
                res[i,i] := Rzero;
                v1 := l[1];
                delete l[1];
                j := i;
                for v2 in l do
                    j := j+1;
                    res[i,j] := dom::commutator(v1,v2);
                    res[j,i] := dom::_negate(res[i,j]);
                end_for;
            end_for;
            res[n,n] := Rzero;
            res;
        end_proc;


    // Lie derivative with respect to vector field v
    // can handle f a vector field or f an element of R
    // otherwise the domain of f must provide a method lie
    lie :=
        proc(v:dom,f)
        local ff;
        begin
	    case domtype(f)
	    of dom do return(dom::commutator(v,f));
	    of R do
	    of OR do return(dom::func_call(v,f));
	    otherwise
		if testargs() then
		    ff := Rcoerce(f);
	            if ff<>FAIL then return(dom::func_call(v,ff)) end_if;
		    ff := Ocoerce(f);
	            if ff<>FAIL then return(dom::func_call(v,ff)) end_if;
		end_if;
		ff := f::lie;
		if ff<>FAIL then return(ff(v,f)) end_if;
		TYPE_ERR;
	    end_case;
	end_proc;


    func_call :=
        proc(v:dom,r)
        local l:DOM_LIST, tmp;
        begin
            r := context(r);
            if testargs() then
                tmp := Ocoerce(r);
                if tmp=FAIL then
		    tmp := Rcoerce(r);
		    if tmp=FAIL then TYPE_ERR else r := tmp end_if;
		else r := tmp end_if;
            end_if;
            l := REP(v);
	    if domtype(r)=R then		// v applied to R
		Rplus(op(map(l,(ll->Rmult(ll[1],Rdiff(r,ll[2]))))));
	    else				// v applied to OR
		l := map(l,(ll->[ll[1],Odiff(r,ll[2])]));
		if Omult=FAIL then
		    tmp := map(l,(ll->[ll[1],Rcoerce(ll[2])]));
		    if contains({op(map(tmp,op,2))},FAIL) then 
			error("cannot coerce");
		    else
			Rplus(op(map(tmp,(ll->Rmult(ll[1],ll[2])))));
		    end_if;
		else	    
		    tmp := map(l,(ll->[Ocoerce(ll[1]),ll[2]]));
		    if contains({op(map(tmp,op,1))},FAIL) then
			tmp := map(l,(ll->[ll[1],Rcoerce(ll[2])]));
			if contains({op(map(tmp,op,2))},FAIL) then 
			    error("cannot coerce");
			else
			    Rplus(op(map(tmp,(ll->Rmult(ll[1],ll[2])))));
			end_if;
		    else
			Oplus(op(map(tmp,(ll->Omult(ll[1],ll[2])))));
		    end_if;
		end_if;
	    end_if;
        end_proc;


    /* ============= */
    /* miscellaneous */
    /* ============= */

    _mult :=
        proc():dom
        local tmp, n:DOM_INT, r:DOM_LIST, s:DOM_LIST, res:dom, DOM:DOM_DOMAIN;
        begin
	    DOM := dom;
	    s := [ args() ];
	    if testargs() then
		s := map(s, proc(x)
			    begin
				tmp := DOM::coerce(x);
				if tmp<>FAIL then return(tmp)
				else
				    tmp := Rcoerce(x);
				    if tmp<>FAIL then return(tmp)
				    else TYPE_ERR end_if;
				end_if;
			    end_proc);
	    end_if;
	    r := select(s, ((x,d) -> bool(domtype(x)<>d)), dom);
	    s := listlib::setDifference(s,r);

            case nops(s)
	    of 1 do res := s[1]; break;
	    of 2 do res := dom::commutator(s[1],s[2]); break;
	    otherwise
	        n:= args(0) div 2;
	        res := dom::_mult(dom::_mult(op(s,1..n)),
			          dom::_mult(op(s,(n+1)..args(0))));
	    end_case;
	    if r<>[] then dom::smult(Rmult(op(r)),res)
	    else res end_if;
	end_proc;


    opDomain := OR;
    
    
    Variables := dom::generatorSet;


begin

    if domtype(R)=DOM_DOMAIN and R::hasProp<>FAIL and
       R::hasProp(Cat::PartialDifferentialRing) then
        Rcoerce := R::coerce;
        Rplus := R::_plus;
	Rmult := R::_mult;
	Rnegate := R::_negate;
        Rdiff := R::diff;
        Rzero := R::zero;
        Riszero := R::iszero;
	Rone := R::one;
        Risone := R::isone;
	Rexpr := R::expr;
    else error("first argument must be Cat::PartialDifferentialRing") end_if;

    if domtype(S)=DOM_LIST then
        finite := TRUE;
	Sexpr := id;
    elif domtype(S)=DOM_DOMAIN and (S::hasProp<>FAIL) and
	 S::hasProp(Cat::OrderedSet) then
        finite := FALSE;
	Sexpr := S::expr;
    else error("second argument must be list or Cat::OrderedSet") end_if;
    
    case args(0)
    of 2 do
	OR := R;
	Ocoerce := Rcoerce;
	Odiff := Rdiff;
	Omult := Rmult;
	Oplus := Rplus;
	break;
    of 3 do
	OR := args(3);
	if domtype(OR)=DOM_DOMAIN and OR::hasProp<>FAIL and
	   OR::hasProp(Cat::AbelianSemiGroup) then
	    Ocoerce := OR::coerce;
	    Odiff := OR::diff;
	    if Odiff=FAIL then 
		error("third argument must possess diff method");
	    end_if;
	    Oplus := OR::_plus;
	    if OR::hasProp(Cat::Rng) then
		Omult := OR::_mult;
	    else Omult := FAIL end_if;
        else error("third argument must be of 'Cat::BaseCategory'") end_if;
	break;
    otherwise error("wrong number of arguments");
    end_case;

end_domain:


unalias(REP,TYPE_ERR):
