// $Date: 2000/05/21 14:54:22 $  $Author: bij $  $Revision: 1.2 $ 

// Moved from prog::optimize to generate::optimize, W. Oevel, 15.5.00

/*++
generate::optimize -- optimize expression or list of equations

generate::optimize(e)

e - expression, equation or list of equations

'optimize' optimizes expressions or equations representing assignments
such that common subexpressions are computed only once.

- A single expression 'e' is interpreted as the lhs of an assignment.
- A single equation a=b represents an assignment a:=b.
- A list of equations represents a list of assignments.

The function returns a list of equations representing assignments. New
'temporary variables' t.i may be generated which are used as lhs of newly
generated auxiliary expressions.

Examples:

>> export(generate,optimize):
>> optimize(x = f(a+b)-(a+b)*c^(a+b));

                   [t14 = a + b, x = f(t14) - t14*c^t14 ]

>> optimize([x=a+b, y=f(a+b) + f(x)^2]);

                  [x = a + b, t15 = f(x), y = t15 + t15^2 ]

>> optimize([x=a*b, y=-a*b]);

                             [x = a b, y = -x]

>> optimize([x=sqrt(t), y=1/sqrt(t)]);

                           --      1/2      1 --
                           |  x = t   , y = -  |
                           --               x --

>> optimize([x=1/15-a/360,y=a/360]);

                   --       a                         --
                   |  t9 = ---, x = 1/15 - t9, y = t9  |
                   --      360                        --
++*/

generate::optimize:= proc(e) local p, f, r; begin
    if args(0) <> 1 then error("wrong no of args") end_if;
    if domtype(e) = DOM_LIST then
        if {op(map(e, type))} <> {"_equal"} then
            error("not an equation")
        end_if
    elif type(e) = "_equal" then
        e:= [e]
    else
        e:= [genident("t")=e]
    end_if;
    
    // optimize each rhs of the assignments in turn 
    p:= [];
    for f in e do
        r:= generate::opt_expr(op(f,2), p, []);
        p:= append(r[2], op(f,1)=r[1])
    end_for;
    p
end_proc:

/*--
generate::opt_expr -- optimize subexpression

generate::opt_expr(e,p,s)

e - subexpression to optimize
p - list of already optimized assignments
s - list of already optimized siblings of e and of the parents of e

Returns a list [e1,p1,s1,c] with:-
e1 - optimized expression
p1 - list of optimized assignments
s1 - list of optimized siblings of e1 and of the parents of e1
c  - boolean indicating that e actually has changed or not
--*/

generate::opt_expr:= proc(e, p, s) 
    local i, r, g, c, l, m, f, j;
begin
    if contains({DOM_EXPR,DOM_LIST,DOM_SET,DOM_TABLE,DOM_ARRAY}, 
		domtype(e)) then
        
        // optimize operands of e 
        c:= FALSE;
        l:= s;
        m:= nops(s);
        f:= [op(e)];
        for i from 1 to nops(f) do
            r:= generate::opt_expr(f[i], p, l);
            if r[4] then
		// replace operands by optimized expressions 
		f:= subsop(f, i=r[1], (j= r[3][j+m]) $ j=1..(i-1));
		p:= r[2];
		l:= append(r[3], r[1]);
		c:= TRUE
            else
                l:= append(l, f[i])
            end_if
        end_for;
        e:= subsop(e, (j=f[j]) $ j=1..nops(f));
        s:= [op(l, 1..m)];

        // don't optimize both x and -x 
        if type(e)="_mult" and (i:=select(e,testtype,
		Type::Union(DOM_INT,DOM_RAT)))<0 then
            g:=generate::opt_expr(-e,p,s);
            return(subsop(g,1=-g[1],4=TRUE))
        // don't optimize both x and 1/x 
        elif type(e)="_power" and contains({DOM_INT,DOM_RAT},type(op(e,2)))
	and op(e,2)<0 then
           g:=generate::opt_expr(1/e,p,s);
           return(subsop(g,1=1/g[1],4=TRUE))
        end_if;

        // is e subexpression of one of the optimized expressions? 
        for i from 1 to nops(p) do
	    // has e already a variable with its value? 
	    if op(p, [i,2]) = e then
                return([op(p, [i,1]), p, s, TRUE]);
            end_if;
            // is e a subexpression to be assigned to a new variable? 
            if has(op(p, [i,2]), e) then
                g:= genident("t");
                p:= append([op(p,1..(i-1))], g=e, subs(p[i], e=g)).
                    [op(p,(i+1)..nops(p))];
                return([g, p, s, TRUE]);
	    elif type(e)="_mult" and has(op(p, [i,2]), -e) then
                g:= genident("t");
                p:= append([op(p,1..(i-1))], g=e, subs(p[i], -e=-g)).
		    [op(p,(i+1)..nops(p))];
                return([g, p, s, TRUE]);
            end_if
        end_for;
        // is e equal to a former sibling? 
        if has(s, e) then
            // create new assignment (e is already optimized before) 
            g:= genident("t");
            return([g, append(p, g=e), subs(s, e=g), TRUE])
        end_if;
	
	// optimize integer powers further by repeated squaring 
	if type(e) = "_power" then
	    l:= op(e,2);
	    if domtype(l) = DOM_INT then
	        if l > 1 and l <= 1000 then
		    e:= op(e,1);
		    if domtype(e) = DOM_EXPR then
			g:= genident("t");
			p:= append(p, g=e);
			e:= g;
			c:= TRUE;
		    end_if;
		    if l = 2 then 
			return([e^2, p, s, c])
		    end_if;
		    l:= numlib::g_adic(l, 2);
		    f:= (if l[1] = 1 then e else 1 end_if);
		    c:= e;
		    for i from 2 to nops(l) do
		        r:= generate::opt_expr(c^2, p, s);
		        c:= r[1];
			p:= r[2];
			s:= r[3];
			if domtype(c) = DOM_EXPR then
			    if i < nops(l) or f <> 1 then
				g:= genident("t");
				p:= append(p, g=c);
				c:= g;
			    end_if
			end_if;
		        if l[i] = 1 then f:= f*c end_if
		    end_for;		    
		    return([f, p, s, TRUE])
	        end_if
	    end_if
	end_if;
        [e, p, s, c]
    else
        [e, p, s, FALSE]
    end_if;
end_proc:

// end of file 
