/*-
 kamo, 15.2.95
 Network 
 Library package for working with Networks (directed graphs)
-*/


Network := newDomain("Network"):
Network::info := "Library 'Network': package for handling directed graphs":
Network::interface := {
hold(addEdge), hold(addVertex), hold(admissibleFlow), hold(allShortPath),
hold(blockflow), hold(cycle), hold(changeEdge), hold(changeVertex),
hold(complete), hold(convertSSQ), hold(eCapacity), hold(eWeight), hold(delEdge),
hold(delVertex), hold(edge), hold(epost), hold(epre), hold(inDegree),
hold(isEdge), hold(isVertex), hold(longPath), hold(maxFlow), hold(minCost),
hold(minCut), hold(outDegree), hold(printGraph), hold(random),
hold(residualNetwork), hold(shortPath), hold(shortPathTo), hold(showGraph),   
hold(topSort), hold(vertex), hold(vWeight) 	}:
  
  
/*-
 new - generates a new Network
 new(V,Ed)
 new(V,Ed, <EWeight=[c1,..,cm]>,<Capacity=[t1,..,tm]>,<Vweight=[v1,..,vn]>)
 Parameters:
   V - list of nodes 
   Ed - list of edges
 If Ed = [e1,..,em] then the cost (weight) of arc ei is ci and the capacity
 is ti. The node weights are specified by the Vweight list. If the lists 
 Eweight, Capacity and Vweight contains fewer elements than Ed and V respectively,
 then the cost and capacity lists are filled up with 1 and 0 resp., and the 
 vertex weigth list is filled up with 0.
-*/
Network::new :=
  proc(V,Edges)
    local Epr, Epo, e, i, j, Lhs, Rhs, Vw, Ecap, Ew, arg;
  begin
/*--
 Creates a new Network.
 Parameters:
   1 - V, list of nodes
   2 - Edges, list of edges
   3,... - equations 
	Vweight = [...] - weights for the nodes
	Capacity = [...] - capacities for the edges
	Eweight   = [...] - costs for the edges

 Data structure:
	new(Network, V, Edges, Vw, Ecap, Ew, Epo, Epr);
  (1)  V - list of nodes
  (2)  Edges - list of edges
  (3)  Vw - weights of nodes
  (4)  Ecap - edge capacities
  (5)  Ew - edge costs
  (6)  Epo - adjacency lists (outgoing)
  (7)  Epr - adjacency lists (incoming)
--*/

    if testargs() then
      if args(0) < 2 then
        error("Wrong number of arguments!");
      end_if;
      if not testtype(V,DOM_LIST) then
        error("Wrong argument type: list expected!");
      end_if;
      if not testtype(Edges, Type::ListOf(DOM_LIST)) then
        error("Wrong argument type: list of lists expected!");
      end_if;
    end_if;

    // init
    Vw:= FAIL; Ecap:= FAIL; Ew:= FAIL;
    
    for i from 3 to args(0) do
      arg := args(i);
      if not testtype(arg, "_equal") then
        error("Wrong argument type: equation expected!");
      end_if;
      Lhs := op(arg,1); Rhs := op(arg,2);
      case Lhs
        of hold(Vweight) do
          case type(Rhs)
            of DOM_TABLE do
              Vw := {op(Rhs,[1..nops(Rhs),1])};
              if {op(V)} <> Vw then
                error("Illegal nodes specified!");
              end_if;
              Vw := Rhs;
              break;
            of DOM_LIST do
              if nops(Rhs) < nops(V) then
                Rhs := Rhs.[0 $ j=1..(nops(V)-nops(Rhs))];
              end_if;
              Vw := table(V[j] = Rhs[j] $ j=1..nops(V));
              break;
            otherwise
              error("Wrong argument type: list or table expected!");
          end_case;
          
          break;
        of hold(Capacity) do
          case type(Rhs)
            of DOM_LIST do
              if nops(Rhs) < nops(Edges) then
                Rhs := Rhs.[0 $ (nops(Edges)-nops(Rhs))];
              end_if;
              Ecap := table(Edges[j] = Rhs[j] $ j=1..nops(Edges));
              break;
            of DOM_TABLE do
              Ecap := {op(Rhs,[1..nops(Rhs),1])};
              if {op(Edges)} <> Ecap then
                error("Illegal edges specified!");
              end_if;
              Ecap := Rhs;
              break;
            otherwise
              error("Wrong argument type: list or table expected!");
          end_case;
          break;
        of hold(Eweight) do
          case type(Rhs)
            of DOM_LIST do
              if nops(Rhs) < nops(Edges) then
                Rhs := Rhs.[0 $ (nops(Edges)-nops(Rhs))];
              end_if;
              Ew := table(Edges[j] = Rhs[j] $ j=1..nops(Edges));
              break;
            of DOM_TABLE do
              Ew := {op(Rhs,[1..nops(Rhs),1])};
              if {op(Edges)} <> Ew then
                error("Illegal edges specified!");
              end_if;
              Ew := Rhs;
              break;
            otherwise
              error("Wrong argument type: list or table expected!");
          end_case;
          break;
        otherwise
          error("Unknown type!");
      end_case;
    end_for;
    
    delete i;
    /*
      If there are no node weights specified, assign the 
      weight 0 to every node.
     */
    if Vw = FAIL then
      Vw := table(V[i] = 0 $ i=1..nops(V));
    end_if;
    
   /*
    If there are no edge capacities specified, assign the 
    capacity 1 to every edge.
    */
   if Ecap = FAIL then
	Ecap := table(Edges[i] = 1 $ i=1..nops(Edges));
   end_if;

   /*-
    If there are no edge costs specified, assign the 
    cost 1 to every node.
   -*/
   if Ew = FAIL then
	Ew := table(Edges[i] = 1 $ i=1..nops(Edges));
   end_if;

   /*-
    Generate the adjacency lists
   -*/
   Epr := table((V[i]=[] $ i=1..nops(V))):
   Epo := Epr;
   for e in Edges do
        Epr[e[2]] := append(Epr[e[2]], e[1]);
        Epo[e[1]] := append(Epo[e[1]], e[2]);
   end_for;

   new(Network, V, Edges, Vw, Ecap, Ew, Epo, Epr);
end_proc:

/*-
 Output of Networks
-*/
Network::print := proc(G)
begin 
   hold(Network())
end_proc:

/*- 
 testtype  - test if expression is a Network 
-*/
Network::testtype:= proc(e,T) begin
    if T = Network then
        if domtype(e) = Network then
            return(TRUE)
        end_if
    end_if;
    FAIL
end_proc:  
  
 
alias(path = pathname("NETWORK")):


Network::addEdge := loadproc(Network::addEdge, path, "addedge"):
Network::addVertex := loadproc(Network::addVertex, path, "addvert"):
Network::admissibleFlow := loadproc(Network::admissibleFlow, path, "admissib"):
Network::allShortPath := loadproc(Network::allShortPath, path, "allspath"):   
Network::blockflow := loadproc(Network::blockflow, path, "blckflow"):
Network::changeVertex := loadproc(Network::changeVertex, path, "chvertex"):
Network::changeEdge := loadproc(Network::changeEdge, path, "chedge"):
Network::complete := loadproc(Network::complete, path, "complete"):
Network::convertSSQ := loadproc(Network::convertSSQ, path, "convssq"):
Network::cycle := loadproc(Network::cycle,  path, "cycle"):
Network::delEdge := loadproc(Network::delEdge, path, "deledge"):
Network::delVertex := loadproc(Network::delVertex, path, "delvert"):
Network::eCapacity := loadproc(Network::eCapacity, path, "ecap"):
Network::edge := loadproc(Network::edge, path, "edge"):   
Network::epost := loadproc(Network::epost, path, "epost"):
Network::epre := loadproc(Network::epre, path, "epre"):
Network::eWeight := loadproc(Network::eWeight, path, "eweight"):
Network::inDegree := loadproc(Network::inDegree, path, "indegre"):
Network::isEdge := loadproc(Network::isEdge, path, "isedge"):
Network::isVertex := loadproc(Network::isVertex, path, "isvertex"):
Network::longPath := loadproc(Network::longPath, path, "lpath"):
Network::maxFlow := loadproc(Network::maxFlow, path, "maxflow"):
Network::minCost := loadproc(Network::minCost, path, "mincost"):
Network::minCut := loadproc(Network::minCut, path, "mincut"):
Network::outDegree := loadproc(Network::outDegree, path, "outdegre"):
Network::printGraph := loadproc(Network::printGraph, path, "printnet"):
Network::random := loadproc(Network::random, path, "random"):
Network::residualNetwork := loadproc(Network::residualNetwork, path, "residual"):
Network::shortPath := loadproc(Network::shortPath, path, "spath"):
Network::shortPathTo := loadproc(Network::shortPathTo, path, "sspath"):
Network::showGraph := loadproc(Network::showGraph, path, "showgr"):
Network::topSort := loadproc(Network::topSort, path, "topsort"):
Network::vertex := loadproc(Network::vertex, path, "vertex"):
Network::vWeight := loadproc(Network::vWeight, path, "vweight"):
 
// functions without interface: 
Network::graph_check := loadproc(Network::graph_check, path, "gc"):

// for AllShortPath: 
Network::floyd := loadproc(Network::floyd, path, "floyd"):

// for LongPath: 
Network::lbellman := loadproc(Network::lbellman, path, "lbellman"):

// for ShortPath: 
Network::bellman := loadproc(Network::bellman, path, "bellman"):
Network::dijkstra := loadproc(Network::dijkstra, path, "dijkstra"):
Network::rbellman := loadproc(Network::rbellman, path, "rbellman"):
Network::rdijkstra := loadproc(Network::rdijkstra, path, "rdijkstr"):
Network::spath_arg_check := loadproc(Network::spath_arg_check, path, "spargc"):
Network::spath_eval := loadproc(Network::spath_eval, path, "speval"):

// for MinCost: 
Network::relaxation := loadproc(Network::relaxation, path, "relax"):

// for MinCost: 
Network::pp_fifo := loadproc(Network::pp_fifo, path, "pp_fifo"):

null():


