// $Date: 2000/12/04 08:59:03 $  $Author: frankp $   $Revision: 1.16 $
// igel, frankp, 16.02.00

/* ------------------------------------------------------------- */
/* --- plot                                                  --- */
/* ---                                                       --- */
/* --- The library domain for 2- and 3-dimensional graphics. --- */
/* ------------------------------------------------------------- */

// plot:= newDomain("plot"): // the library domain is already defined in 'sysinit.mu'!
plot::info:= "Library 'plot': graphical primitives and functions for two- and three-dimensional plots":

/*--
    plot::new -- function to put precreated graphics on the screen

    Syntax:

    plot(g1 <, g2, ...>  <, op1, op2, ...>)

    g1, g2, ...   : graphical objects
    op1, op2, ... : scene options, see plot2d and plot3d for details
   
    show either creates a scene of type plot::Scene out of the given
    graphical objects and puts it on the screen or if a complete scene 
    of type plot::Scene is given, it is shown on the screen. 

    Example:

    >> plot( plot::Function2d( sin(t), t = 1..3 ) )

    >> plot( plot::Function3d( sin(t + u), t = 0..2*PI, u = -1..1, Color = RGB::Blue ), Axes = None )

    >> plot( plot::Scene(plot::Point(1,2)) )
--*/
plot::new := proc(scene)
    local dim;
begin    
    if args(0) = 0 then
        error("no argument given")
    elif (dim:= scene::dom::plot) <> FAIL then
        return(dim( args() ))
    elif domtype(scene) = plot::Scene then
        if args(0) <> 1 then
            error("expecting one argument")
        end_if
    else
        scene:= plot::Scene( args() )
    end_if;

    dim:= scene::dom::dimension(scene);
    if dim = 2 then
        plot2d( op(plot::Scene::getPlotdata(scene)) );
    else
        plot3d( op(plot::Scene::getPlotdata(scene)) );
    end_if
end_proc:

plot::interface :={
        // Primitives
        hold(Scene), hold(Curve2d), hold(Curve3d), hold(Function2d),
        hold(Function3d), hold(Surface3d), hold(Polygon), hold(Point),
        hold(Pointlist), hold(Rectangle2d), hold(Ellipse2d), hold(Group),
        hold(Lsys), hold(Turtle),

        // Structs:
        hold(line), hold(vector),

        // Functions: 
        hold(xrotate), hold(yrotate), hold(modify), 
        hold(copy), hold(contour), 
        hold(cylindrical), hold(spherical), hold(polar), 
        hold(density), hold(vectorfield), hold(implicit), 
        hold(data), hold(inequality), hold(ode)

        // Not Documented:
        // , hold(area), hold(Lsys), hold(Turtle)
}:

alias (path = pathname( "PLOT", "PRIMITIV" )):
    plot::Curve2d     := loadproc( plot::Curve2d,     path, "Curve2d"):
    plot::Curve3d     := loadproc( plot::Curve3d,     path, "Curve3d"):
    plot::Ellipse2d   := loadproc( plot::Ellipse2d,   path, "Ellipse2d"):
    plot::Function2d  := loadproc( plot::Function2d,  path, "Function2d"):
    plot::Function3d  := loadproc( plot::Function3d,  path, "Function3d"):
    plot::Group       := loadproc( plot::Group,       path, "Group"):
    plot::Point       := loadproc( plot::Point,       path, "Point"):
    plot::Pointlist   := loadproc( plot::Pointlist,   path, "Pointlist"):
    plot::Polygon     := loadproc( plot::Polygon,     path, "Polygon"):
    plot::Rectangle2d := loadproc( plot::Rectangle2d, path, "Rectangle2d"):
    plot::Scene       := loadproc( plot::Scene,       path, "Scene"):
    plot::Surface3d   := loadproc( plot::Surface3d,   path, "Surface3d"):
    plot::NullObject  := loadproc( plot::NullObject,  path, "NullObject"):
    plot::NullObject2d:= loadproc( plot::NullObject2d,path, "NullObject"):
    plot::NullObject3d:= loadproc( plot::NullObject3d,path, "NullObject"):
unalias( path ):

alias (path = pathname( "PLOT", "STRUCTS" )):
    plot::line          := loadproc( plot::line,          path, "line"):
    plot::vector        := loadproc( plot::vector,        path, "vector"):

    // not documented: (used in simpson.mu)
    plot::area          := loadproc( plot::area,          path, "area"):
unalias( path ):

alias (path = pathname( "PLOT" )):
    plot::modify        := loadproc( plot::modify,        path, "copy"):
    plot::copy          := loadproc( plot::copy,          path, "copy"):
    plot::implicit      := loadproc( plot::implicit,      path, "implicit"):
    plot::xrotate       := loadproc( plot::xrotate,       path, "xrotate"):
    plot::yrotate       := loadproc( plot::yrotate,       path, "yrotate"):
    plot::contour       := loadproc( plot::contour,       path, "contour"):
    plot::cylindrical   := loadproc( plot::cylindrical,   path, "cylind"):
    plot::spherical     := loadproc( plot::spherical,     path, "spherical"):
    plot::polar         := loadproc( plot::polar,         path, "polar"):
    plot::density       := loadproc( plot::density,       path, "density"):
    plot::vectorfield   := loadproc( plot::vectorfield,   path, "vfield"):
    plot::data          := loadproc( plot::data,          path, "data"):
    plot::inequality    := loadproc( plot::inequality,    path, "inequal"):
    plot::ode           := loadproc( plot::ode,           path, "ode"):

    plot::Turtle        := loadproc( plot::Turtle,        path, "Turtle"):
    plot::Lsys          := loadproc( plot::Lsys,          path, "Lsys"):
    plot::HOrbital      := loadproc( plot::HOrbital,      path, "HOrbital"):

    // internal functions:
    plot::range2list      := loadproc( plot::range2list,       path, "tools"):
    plot::conv2range      := loadproc( plot::conv2range,       path, "tools"):
    plot::list2range      := loadproc( plot::list2range,       path, "tools"):
    plot::getColorFormat  := loadproc( plot::getColorFormat,   path, "tools"):
unalias( path ):

plot::exported:= {hold(plot),hold(plot2d),hold(plot3d),hold(plotfunc2d),hold(plotfunc3d)}:

stdlib::LoadlibState["plot"]:= TRUE:

