// $Date: 2000/10/10 06:50:30 $  $Author: frankp $   $Revision: 1.18 $
// frankp, 21/02/94

/*
        linalg::det( A )

        A: square matrix

        Computes the determinant of a matrix over a commutative ring.
*/

linalg::det := proc(A)
    local R, Mat, Rnormal, Rmult, Rnegate;
begin
    if args(0) <> 1 then
        error("expecting one argument, a matrix of 'Cat::Matrix'")
    end_if;

    Mat:= A::dom;
    if testargs() then
        if Mat::hasProp( Cat::Matrix ) <> TRUE then
            error("expecting a matrix of 'Cat::Matrix'")
        elif not (Mat::coeffRing)::hasProp( Cat::CommutativeRing ) then
            error("expecting matrix over 'Cat::CommutativeRing'")
        end_if;
        R:= Mat::matdim(A);
        if R[1] <> R[2] then
            error("not a square matrix")
        end_if
    end_if;

    R:= Mat::coeffRing;
    case op(Mat::matdim(A),1)
    of 1 do 
        return( A[1,1] );
    of 2 do
        userinfo(1,"use formula for the determinant of 2x2 matrices");

        Rmult:= R::_mult;
        if (Rnormal:= R::normal) = FAIL then
            if R::hasProp(Ax::systemRep) then Rnormal:= normal else Rnormal:= id end_if
        end_if;
 
        return( Rnormal(R::_plus( 
           Rmult( A[1,1],A[2,2] ), R::_negate( Rmult( A[1,2],A[2,1] ) )
        )) )
    of 3 do
        userinfo(1,"use formula for the determinant of 3x3 matrices");

        Rmult:= R::_mult;
        if (Rnormal:= R::normal) = FAIL then
            if R::hasProp(Ax::systemRep) then Rnormal:= normal else Rnormal:= id end_if
        end_if;
 
        Rmult:= R::_mult;
        Rnegate:= R::_negate;
        return( Rnormal(R::_plus( 
            Rmult( A[1,1],A[2,2],A[3,3] ),
            Rnegate( Rmult( A[1,1],A[2,3],A[3,2] ) ),
            Rnegate( Rmult( A[2,1],A[1,2],A[3,3] ) ),
            Rmult( A[2,1],A[1,3],A[3,2] ),
            Rmult( A[3,1],A[1,2],A[2,3] ),
            Rnegate( Rmult( A[3,1],A[1,3],A[2,2] ) )
        )) )
    otherwise
        case R
        of Dom::Float do
            userinfo(1,"call 'numeric::det'");
            return(numeric::det(  Mat::convert_to(A,DOM_ARRAY) ))
        otherwise
            if R::hasProp(Cat::Field) and R::hasProp( Ax::systemRep ) then
                userinfo(1,"call 'numeric::det' with option 'Symbolic'");
                return( numeric::det( Mat::convert_to(A,DOM_ARRAY),hold(Symbolic) ) )
            elif R::hasProp( Cat::IntegralDomain ) then
                userinfo(1,"use Gaussian elimination");
                return( op(Mat::gaussElim(A),3) )
            else
                userinfo(1,"use Berkowitz algorithm");
                return( linalg::detBerkowitz(A) )
            end_if
        end_case
    end_case
end_proc:

linalg::det := funcenv(linalg::det):
linalg::det::Content := (Out, data) -> Out::Capply(Out::Cdeterminant, 
                                                   Out(op(data))):

/*--
K. F. Gehrs, 2000/09/26

    linalg::detBerkowitz -- returns the determinant of the matrix A

    linalg::detBerkowitz( A )

    A: square matrix over a commutative ring
 
    Computes the the determinant of the matrix A
    The coefficient domain of A has to be a commutative ring.

    The algorithm is based on the "Berkowitz-Algorithm" to compute the 
    characterisitic polynomial. This algorithm was modified and is now
    implemented to replace the method of Laplace for computing the 
    determinant of a matrix over an ordinary commutative ring. 
            
    Reference: The Berkowitz Algorithm, Maple and Computing the 
               Characteristic Polynomial in an Arbitrary Commutative Ring
               by Jounaidi Abdeljaoued, May 1996
--*/
linalg::detBerkowitz:= proc(A)
    local Rzero, Rmult, Rplus, Rnegate, Rone, Vect, i, j, r, C, S, Q, R, n;
begin
    R:= A::dom::coeffRing;
    n:= A::dom::matdim(A)[1];

    Rone:= R::one;
    Rzero:= R::zero;
    Rmult:= R::_mult;
    Rplus:= R::_plus;
    Rnegate:= R::_negate;

    Vect:= (S:= (Q:= array( 1..n+1 )));
    Vect[1]:= Rnegate(Rone); 
    Vect[2]:= A[1,1];
    C[1]:= Vect[1];

    for r from 2 to n do
        S:= assignElements(S, [i] = A[i,r] $ i=1..r-1);
        C[2]:= A[r,r];
        for i from 1 to r-2 do
            C[i+2]:= Rzero;
            ( C[i+2]:= Rplus( C[i+2], Rmult( A[r,j], S[j])) ) $ j=1..r-1;
            for j from 1 to r-1 do
                Q[j]:= Rzero;
                ( Q[j]:= Rplus( Q[j], Rmult( A[j,k], S[k])) ) $ k=1..r-1
            end_for;
            S:= assignElements(S, [j] = Q[j] $ j=1..r-1);
        end_for;
        C[r+1]:= Rzero;
        ( C[r+1]:= Rplus( C[r+1], Rmult( A[r,j], S[j])) ) $ j=1..r-1;
        for i from 1 to r+1 do
            Q[i]:= Rzero;
            ( Q[i]:= Rplus( Q[i], Rmult( C[i+1-j], Vect[j])) ) $ j=1..min(r,i)
        end_for;
        Vect:= assignElements(Vect, [i] = Q[i] $ i=1..r+1);
    end_for;

    if Vect[1] = Rone then 
        return( Vect[n+1] )
    elif n mod 2 = 1 then // (-1)^(n+1) = 1
        return( Vect[n+1] )
    else 
        return(Rnegate( Vect[n+1] ))
    end_if
end_proc:

