// $Date: 2000/05/03 13:03:28 $  $Author: frankp $   $Revision: 1.25 $
// frankp, 14.06.1995

/*
        linalg::eigenvectors(A)  --  computes the eigenvectors of matrix A

        A: square matrix

        eigenvectors(A) computes the eigenvalues and eigenvectors 
        of the matrix A.

        Method:
        1. eigenvalues of the matrix are computed using the function
           linalg::eigenvalues. 

        2. For each eigenvalue e the matrix equation (I*e - A)*X = 0, 
           where I denotes the identity matrix, is solved using the 
           Gauss-Jordan algorithm (linalg::gaussJordan).

        Result:
        A list of sublists, where each sublist consists of an eigenvalue
        l of A, its algebraic multiplicity and a basis for the eigenspace of l.

        Comments:
        An error message is given if A is not of category Cat::Field.

        If the coefficient ring of A is Dom::Float then numeric::eigenvalues
        is used. Note that in this case the list e of eigenvalues is sorted 
        which means that e[i]<=e[j] for i<j. The eigenvectors are normalized 
        (see numeric::eigenvectors).
*/

linalg::eigenvectors := proc(A)
    local n, s, e, i, k, r, R, Aplus, ev, En, Mat;
begin
    if testargs() then
	if args(0) <> 1 then
	    error("expecting a matrix of 'Cat::Matrix'")
	elif A::dom::hasProp( Cat::Matrix ) <> TRUE then
	    error("expecting a matrix of 'Cat::Matrix'")
	elif not (A::dom::coeffRing)::hasProp( Cat::Field ) then
	    error("expecting a matrix over a 'Cat::Field'")
	end_if
    end_if;

    Mat:= A::dom;
    R:= Mat::coeffRing;
    if R = Dom::Float then
        userinfo(1,"call 'numeric::eigenvectors'");
        s:= numeric::eigenvectors( A,hold(NoErrors) );

        // collect multiple eigenvalues:
        ev:= []; k:= 1;
        Mat:= Dom::Matrix(R);
        s[2]:= Mat::coerce( s[2] );
        while s[1] <> [] do
            n:= split( s[1],_equal,s[1][1] );
            r:= nops(n[1]);

            if (e:= R::coerce(n[1][1])) <> FAIL then
                ev:= append( ev,[ e,r,[Mat::col(s[2],i) $ i=k..k+r-1] ] )
            end_if;
            k:= k + r;
            s[1]:= n[2]
        end_while;
        return( ev )
    else
        n:= Mat::matdim(A);
        if n[1] <> n[2] then 
            error("not a square matrix")
        end_if;
        n:= n[1];
 
        userinfo(1,"compute the eigenvalues of the given matrix");
        s:= linalg::eigenvalues(A,hold(Multiple));

        A:= Mat::_negate(A);
        Aplus:= Mat::_plus;
        En:= Mat::identity( n );

        ev:= [];
        for k from 1 to nops(s) do
            e:= En;
            (e[i,i]:= op(s,[k,1])) $ i=1..n;
            userinfo(2,"compute kernel of the characteristic manifold");
            userinfo(3,"for the eigenvalue ",op(s,[k,1]));
            r:= linalg::nullspace( Aplus(e,A) );	
            if r = [] then
                warning("basis of eigenspace for eigenvalue ".expr2text(op(s,[k,1])).
                           " not found (zero test failed?)");
                return( FAIL )
            else
                ev:= append(ev,[op(s,[k,1]),op(s,[k,2]),r])
            end_if
        end_for;

        return( ev )
    end_if
end_proc:

