// $Date: 2000/07/07 15:56:17 $  $Author: frankp $   $Revision: 1.21 $
// frankp, 24/02/94

/*++
    linalg::matlinsolve  --  solve a linear system of equations

    matlinsolve( A [,B] )
    matlinsolve( A {,v,l} )
    matlinsolve( A {,v,Special,Unique} )

    A, B            - matrices
    v               - vector
    l               - list with exactly ncols(A) elements
    Special, Unique - identifier (options)

    matlinsolve(A,B) solves the systems A*x=col(B,i), i=1..ncols(B) 
    and returns a matrix representing the special solutions of the 
    ncols(B) systems.
    If B is missing the the system A[1..r,1..c-1]*c=col(A,c)
    will be solved.

    matlinsolve(A,v,l) solves the system A*x=B. If the system
    is undetermined, then the general solution of the system 
    will be expressed by the sum of a special solution and a
    linear combination of the general solution of the corresponding
    homogeneus system with scalars in l. Only those scalars are
    used which correspond to a kernel vector of A.
    If v is missing then the system A[1..r,1..c-1]*c=col(A,c)
    will be solved.


    matlinsolve(A,v,Special) returns a special solution of the system.
    matlinsolve(A,v,Unique) returns an unique solution if one exits,
    otherwise FAIL is returned if more than one solution exists.
    If v is missing then the system A[1..r,1..c-1]*c=col(A,c)
    will be solved.


    matlinsolve returns [] if the system has no solution,
    and FAIL if the solution could not be computed over the 
    coefficient domain of A.

    Solutions are always elements of the domain Matrix(A::coeffRing).
++*/

linalg::matlinsolve := proc(A,B)
    local r, c, t, R, i, j, k, ns, rs, cB, s, v, l, Flag,
          RisField, create, Rdivide, Rone, Rnegate, Riszero, Rmult, 
          Rcoerce, Rzero;
begin
    if A::dom::hasProp( Cat::Matrix ) <> TRUE then
        error("first argument is not of 'Cat::Matrix'")
    end_if;

    R := A::dom::coeffRing;
    if R::hasProp( Cat::IntegralDomain ) <> TRUE then
        error("expecting matrix over Cat::IntegralDomain")
    end_if;

    t:= A::dom::matdim(A);
    r:= t[1]; c:= t[2];

    if A::dom::constructor <> Dom::Matrix then
        t:= Dom::Matrix(R);
        A:= t::coerce(A);
        if A = FAIL then
           error("can't convert matrix to the domain 'Dom::Matrix(R)'")
        end_if
    else
        t:= A::dom
    end_if;
    create:= t::create;

    if args(0) = 3 then Flag:= args(3) else Flag:= NIL end_if;

    if testargs() and args(0) > 1 then
        if not contains( {hold(Unique),hold(Special)},B )
             and domtype(B) <> DOM_LIST 
        then
            if args(0) > 3 then
                error("wrong no of args")
            end_if;

            B:= t::coerce( B );
            if B = FAIL then 
                error("types of matrices don't match")
            end_if;
            s:= B::dom::matdim(B);
            if s[1] <> r then error("invalid operands") end_if;
            if args(0) = 3 and not contains( {hold(Unique),hold(Special)},Flag ) and domtype(Flag) <> DOM_LIST then
                error("invalid 3rd argument")
            end_if
        elif args(0) > 2 then 
            error("wrong no of args")
        end_if
    end_if;

    if args(0) > 1 then
        if not contains( {hold(Unique),hold(Special)},B )  and domtype(B) <> DOM_LIST then
            B:= t::coerce(B);
            cB:= B::dom::matdim(B)[2];
            A:= A::dom::concatMatrix( A,B );
        else
            Flag:= B;
            c:= c - 1; cB:= 1
        end_if
    else
        c:= c - 1; cB:= 1;
        if c = 0 then error("invalid matrix dimension") end_if;
    end_if;

    userinfo(1,"perform Gaussian elimination");
    Riszero:= R::iszero;

    s:= A::dom::gaussElim( A );
    rs:= s[2]; // rank of (A,B)
    s:= s[1];

    // rank(A,B) = rank(A)?
    if rs > 0 then
        for t from rs to c do
            if not Riszero( s[rs,t] ) then break end_if
        end_for;
        if t > c then return( [] ) end_if
    end_if;

    RisField:= R::hasProp( Cat::Field );

    userinfo(1,"compute row reduced form");
    s:= linalg::gaussJordan( A,hold(All) );

    A:= s[1];

    // each col-no > c does not belong to ker(A)
    ns:= { $ 1..c } minus s[4];

    // does an unique solution exist?
    if nops(ns) = 0 then  // rang(A) = c
        if RisField then
            // the right side of A is the solution
            return( A[1..c,c+1..cB+c] )
        else
            Rdivide:= R::_divide;
            s:= A[1..c,c+1..cB+c]; // right-hande side
            for i from 1 to c do
                for j from 1 to cB do
                    if (t:= Rdivide( s[i,j],A[i,i] )) = FAIL then
                        // no solution
                        return( [] )
                    else
                        s[i,j]:= t
                    end_if
                end_for
            end_for;

            return( s )
        end_if
    elif cB > 1 then
        return( [] )
    elif Flag = hold(Unique) then
        return( NIL )
    end_if;


    userinfo(1,"search for special solution");

    if RisField then
        // compute special solution over fields
        s:= create(c,1);

        j:= 1;
        for i from 1 to c do
            if not contains( ns,i ) then
                s[i,1]:= A[j,c+1];
                j:= j + 1;
                if j > r then break /* i */ end_if
            end_if
        end_for;

        if Flag = hold(Special) then
            return( s )
        end_if

    elif Flag <> hold(Special) then
        error("sorry, can determine a basis for the null space only over fields")
    else
        // compute special solution over integral domains
        Rdivide:= R::_divide;

        j:= 1;
        s:= create(c,1); // the zero column
        for i from 1 to c do
            if not contains( ns,i ) then
                if (t:= Rdivide( A[j,c+1],A[j,i] )) = FAIL then
                    // no solution:
                    return( [] )
                else
                    s[i,1]:= t
                end_if;

                j:= j + 1;
                if j > r then break /* i */ end_if
            end_if
        end_for;

        return( s )
    end_if;


    userinfo(1,"solve the homogeneous system");

    Rnegate:= R::_negate;
    Rone:= R::one;
    Rzero:= R::zero;

    rs:= [];
    ns:= sort( [op(ns)] );
    l:= min( r,c );
    if domtype(Flag) <> DOM_LIST then
        for i from 1 to nops(ns) do
            k:= ns[i];
            v:= map( [op(A[1..l,k..k])], Rnegate );
            for j from 1 to i do
                if j = i then
                    v:= [op(v,1..k-1),Rone,op(v,k..l+j-1)];
                else
                    t:= ns[j];
                    v:= [op(v,1..t-1),Rzero,op(v,t..l+j-1)]
                end_if
            end_for;

            rs:= rs . [create( c,1,[op(v,1..min(c,l+i))] )]
        end_for;

        [s, rs]
    else
        if nops(Flag) < c-1 then
            Flag:= Flag . [genident() $ i = nops(Flag)+1..c-1];
        end_if;
        Rmult:= R::_mult;
        Rcoerce:= R::coerce;

        for i from 1 to nops(ns) do
            k:= ns[i];
            if (R:= Rcoerce(op(Flag,k))) = FAIL then
                return( FAIL )
            end_if;
            v:= map( [op(A[1..l,k..k])], x-> Rmult( Rnegate(x),R ) );
            for j from 1 to i do
                if j = i then
                    v:= [op(v,1..k-1),R,op(v,k..l+j-1)]
                else
                    t:= ns[j];
                    v:= [op(v,1..t-1),Rzero,op(v,t..l+j-1)]
                end_if
            end_for;

            rs:= rs . [create( c,1,[op(v,1..min(c,l+i))] )]
        end_for;

        A::dom::_plus( s,op(rs) )
    end_if

end_proc:

