
/*

linopt::dual - Convert the linear program to its dual linear program 
	

Calling sequence:

linopt::dual( [ LinCon, ObjFun <, opt1> ] )

Parameter:

LinCon - a set or list of linear constraints. A linear constraint 
         can be an equation or a non-strict inequality.
ObjFun - The opjective function. Must be a linear function.
opt1   - (optional) name NonNegative or a set of variables.
         NonNegative: All variables are assumed to be non-negative.
	 The variables of the set are assumed to be non-negative.

Return Value:

A list of linear inequalities in standard form. 

Examples:

>> c:=[[4*x+3*y<=6, 3*x+4*y<=4, x>=0, -y<=0], x+y]:
>> linopt::dual(c)

   [[u7 - 4*u6 - 3*u5 <= -1, u8 - 3*u6 - 4*u5 <= -1], 6*u5 + 4*u6]

>> c:=[[4*x+3*y<=6, 3*x+4*y<=4], x+y, NonNegative]:
>> linopt::dual(c)

   [[u7 - 4*u6 - 3*u5 <= -1, u8 - 3*u6 - 4*u5 <= -1], 6*u5 + 4*u6]

>> c:=[[4*x+3*y<=6, 3*x+4*y<=4], x+y, {x,y}]:
>> linopt::dual(c)

   [[u7 - 4*u6 - 3*u5 <= -1, u8 - 3*u6 - 4*u5 <= -1], 6*u5 + 4*u6]
   

Background:

The Dual of a Linear Program:
Given C Matrix(m,n), d Matrix(m), p Matrix(n), p0 real

Original problem:
x Matrix(n): C*x+d >= 0, maximize transpose(p)*x+p0

Dual problem:
u Matrix(m): transpose(C)*u+p <= 0, minimize transpose(d)*u+p0

*/

linopt::dual := proc(l)
   local C, LinCon, ObjFun, p, p0, ind, m, n, nid, r, s;
begin
   if args(0) <> 1 then
      error("Wrong number of arguments. Only one argument expected.")
   end_if;
   if domtype(l) <> DOM_LIST or nops(l) < 2 then
      error("List [LinearConstraints, ObjectiveFunction] expected.")
   end_if;
   LinCon := l[1];
   ObjFun := l[2];
   ind := [op(indets(LinCon) union indets(ObjFun))];
   if Type::Linear(ObjFun, ind) = FALSE then
      error("Objective function must be linear.")
   end_if;

   s := linopt::standardForm(LinCon, op(l, 3..nops(l)));
   // s in standard form: C*x <= d
   C := linalg::expr2Matrix(map(s, op, 1), ind);
   C := linalg::delCol(C, linalg::ncols(C));
   // C*x <= d -> (-C)*x + d >= 0
   C := linalg::transpose(-C);
   p0:= linopt::get_constant(ObjFun);
   ObjFun := ObjFun-p0;
   p := linalg::expr2Matrix([ObjFun], ind);
   // transpose(-C)*u + p <= 0
   // -> transpose(-C)*u <= -p
   p := linalg::delCol(-p, linalg::ncols(p));
   [m, n] := linalg::matdim(C);
   nid := [genident("u") $ i=1..n];
   r := [_plus(C[i,j]*nid[j] $ j=1..n) <= p[i] $ i=1..m];
   [r, _plus(op(s[j], 2)*nid[j] $ j=1..n) + p0]
           
end_proc:

linopt::get_constant := proc(f)
   local c;
begin 
   case type(f)
     of "_mult" do
        return(0);
     of "_plus" do
        // if the sum contains a constant, then it is the last op
        c := op(f, nops(f));
        return((if testtype(c, Type::Numeric) then c else 0 end_if))
     otherwise
        if testtype(f, Type::Numeric) then f else 0 end_if
   end_case
end_proc:
