// $Date: 2000/11/16 15:56:24 $ $Author: frankp $ $Revision: 1.8 $ 

// kg, 12/04/95 

/*++
misc::maprec -- recursively maps operands of expression

maprec(e [, q=f...] [, PreMap] [, PostMap])

e       - expression to map
q       - qualifier (set of types or predicate function)
f       - matching function
PreMap  - option
PostMap - option

maprec recursively maps matching subexpressions of e. For each subexpression
s of e it is tested wether one of the matching equations apply. The first
matching equation defines the new value f(s) of the subexpression, which is
inserted into e instead of s. If no matching equation applies then s is not
altered.

Qualifiers used in matching equations may have one of two forms:

- {t1,...,tn} matches any subexpression s of type t1,...,tn.

- A predicate function p matches any subexpression s for which p(s) returns
  TRUE (p must return a boolean value).

If the option PreMap is given, then for each subexpression s of e first the
operands of s are mapped recursively; only after mapping the operands it ist
tested wether a matching equations applies for the result.

If the option PostMap is given, then for each subexpression s of e it is
tested wether any of the matching equations applies. If no match
applies then the operands of s are mapped recursively.

Both options may be given. If no option is given then only PreMap is used.

Important notes:

- Only operands of expressions (i.e. DOM_EXPRs), sets, lists, arrays and
  tables are mapped recusively. Operands of other types are not mapped in
  order to avoid unwanted substitutions. If one wants to map the operands of
  other types of subexpressions one has to do that explicitly by adding
  matching equations, like for example {DOM_PROC}=handle_procs.

- The operators of expressions (op 0) are also mapped recursively like the
  other operands.

One may overload maprec: If the domain of a subexpression has the method
"maprec" then that method is called with the subexpression and the other
arguments of the call. The result replaces the subexpression (no pre- or
post-mapping of operands is applied).

If one wants to stop the mapping then one should call misc::breakmap(). After
calling this no further mappings are applied.
++*/

misc::maprec:= proc(e)
    local bm, pre, post, mapeq;
begin
    case args(0)
    of 0 do error("no args");
    of 1 do return(e);
    end_case;
    
    mapeq:= [args(2..args(0))];
    pre:= contains(mapeq, hold(PreMap));
    if pre = 0 then
	pre:= FALSE
    else
	delete mapeq[pre]; pre:= TRUE
    end_if;
    post:= contains(mapeq, hold(PostMap));
    if post = 0 then
	post:= FALSE
    else
	delete mapeq[post]; post:= TRUE
    end_if;
    if not pre and not post then pre:= TRUE end_if;

    if mapeq = [] then return(e) end_if;

    if testargs() then
	if {op(map(mapeq, type))} <> {"_equal"} then
	    error("illegal equation")
	end_if
    end_if;
    
    bm:= misc::do_breakmap;
    sysassign(misc::do_breakmap, FALSE);
    e:= misc::maprec_main(e, mapeq, pre, post);
    sysassign(misc::do_breakmap, bm);
    e
end_proc:

// misc::maprec_main does maprec's work 
misc::maprec_main:= proc(ex, mapeq, pre, post)
    local t, eq, f;
begin
    if misc::do_breakmap then return(ex) end_if;
    t:= domtype(ex);
    
    if t::maprec <> FAIL then 
        return(t::maprec(ex, op(mapeq),
            (if pre then hold(PreMap) else null() end_if),
            (if post then hold(PostMap) else null() end_if)))
    end_if;

    if pre then
        if contains(misc::DomainsMapOp, t) then
	    if t = DOM_EXPR then
		f:= misc::maprec_main(op(ex,0), args(2..4));
		if f = op(ex,0) then
		    ex:= map(ex, misc::maprec_main, args(2..4))
		else
		    ex:= f(op(map([op(ex)], misc::maprec_main, args(2..4))))
		end_if
            else
		ex:= map(ex, misc::maprec_main, args(2..4))
            end_if;
            if misc::do_breakmap then return(ex) end_if
        end_if
    end_if;

    for eq in mapeq do
        if domtype(op(eq,1)) = DOM_SET then
            if contains(op(eq,1), type(ex)) then
                return(op(eq,2)(ex));
            end_if
        elif op(eq,1)(ex) = TRUE then
            return(op(eq,2)(ex))
        end_if;
    end_for;

    if post then
        t:= domtype(ex);
        if contains(misc::DomainsMapOp, t) then
	    if t = DOM_EXPR then
		f:= misc::maprec_main(op(ex,0), args(2..4));
		if f = op(ex,0) then
		    ex:= map(ex, misc::maprec_main, args(2..4))
		else
		    ex:= f(op(map([op(ex)], misc::maprec_main, args(2..4))))
		end_if
            else
		ex:= map(ex, misc::maprec_main, args(2..4))
            end_if
        end_if
    end_if;
    ex
end_proc:

// domain types which map their operands 
misc::DomainsMapOp:= {
    DOM_ARRAY, DOM_SET, DOM_LIST, DOM_TABLE, DOM_EXPR
}:


/*++
misc::breakmap -- stop mapping with maprec

breakmap()

-no args-
++*/

misc::breakmap:= proc() begin
    if args(0) <> 0 then error("no args allowed") end_if;
    sysassign(misc::do_breakmap, TRUE)
end_proc:

// end of file 
