// 1.te Version  9.9.97 
// 2.te Version  1.7.98 
// 3.te Version 24.8.98 
// 4.te Version  1.9.99 
//------------------------------------------------------------------------
// The following quadrature is the control routine for numerical          
// quadrature. It should be called in the form                            
//                                                                        
//      quadrature(f(x),x=x0..x1, <MaxCalls = m> )                        
// or                                                                     
//      quadrature(f(x),x=x0..x1,method=n)                                
// or                                                                     
//      quadrature(f(x),x=x0..x1,method=n,Adaptive=TRUE/FALSE)            
//                                                                        
// Parameters:  f(x) -- univariate expression in indeterminate x          
//              m    -- positive integer or infinity                      
//              n    -- positive integer                                  
//                                                                        
// <method> is either GaussLegendre or NewtonCotes or GaussTschebyscheff  
// Following alternative names for these methods are accepted:            
//                                                                        
//    GaussLegendre,gausslegendre,Gauss,gauss,GL,gl                       
//    GaussTschebyscheff,gausstschebyscheff,GT,gt                         
//    NewtonCotes,newtoncotes,NC,nc                                       
//                                                                        
// The adaptive mechanism is based on a fixed quadrature rule specified   
// by one of these names with n nodes, where <n> may take any finite      
// value from {1,2,3,...}.                                                
// If no method is specified, then, depending on DIGITS, an adaptive      
// Gauss-Legendre quadrature with n=20,40,80 or 160 nodes (with           
// respective orders 40,80,160 or 320) is used as default.                
//                                                                        
// MaxCalls = m (with a positive integer m) is a stopping criterion:      
// after m evaluations of the integrand quadrature gives up and returns   
// an inaccurate result with a warning. m may be infinity. In this case   
// quadrature continues until it finds the result or runs into internal   
// problems (typically: evaluation of the integrand at a singularity)     
// The default value is m = MAXDEPTH*n .                                  
//                                                                        
// It may also happen, that quadrature returns inaccurate results,        
// when the internal recursive depth exceeds MAXDEPTH. In this case       
// a warning is given.                                                    
//                                                                        
// Warnings occur only when quadrature is called interactively.           
// Otherwise FAIL is returned instead of warnings!                        
//                                                                        
// Examples:                                                              
// >>  export(numeric):                                                   
// >>  quadrature( exp(-x^2), x=0..1 );                                   
// >>  quadrature( exp(-x^2), x=0..infinity);                             
// >>  quadrature( exp(-x^2), x=0..1, MaxCalls=infinity );                
// >>  quadrature( exp(-x^2), x=0..1, GaussLegendre=40 );                 
// >>  quadrature( exp(-x^2), x=0..1, GL=160, Adaptive=FALSE );           
// >>  quadrature( exp(-x^2), x=0..1, NewtonCotes=10, Adaptive=TRUE );    
//                                                                        
// Technical remark: the data of Gaussian quadrature GL=20, GL=40, GL=80  
//      and GL=160 are pre-computed and stored with 200 DIGITS, whence    
//      there is no delay when using these methods with at most 200       
//      DIGITS. For different values of <n> or higher DIGITS these data   
//      will be computed before the actual quadrature. This may take some 
//      time. Therefore it is recommended to use these values of <n> and  
//      to restrict ones ambitions to DIGITS<=200.                        
//      The algorithm will converge eventually for arbitrarily high       
//      DIGITS and arbitrary integrands (provided the integral exists).   
//      However, convergence will be slow if very high precision is       
//      required. Also, convergence will be slow if the integrand is not  
//      smooth.                                                           
//------------------------------------------------------------------------

numeric::quadrature:=proc()
local alg_expr,eq,adaptive,method,opt,n,b,c,freeStyle,
      Quadsum,adaptiveQuad,Is,x,x0,x1,macheps, lazyAbs,
      fcalls, maxcalls, interactive, problem;
begin
  if args(0)<2 then error("expecting at least 2 arguments"); end_if:
  alg_expr:= args(1); eq:= args(2);

  if (not testtype(eq,"_equal")) or
     (not testtype(op(eq,2), "_range"))
     then error("expecting range x=a..b");
  end_if;

  // ---------------------------------------------------

  x:=op(eq,1): x0:=op(eq,[2,1]): x1:=op(eq,[2,2]):
  if numeric::indets(x)={} then error("no integration variable") end_if:

  // ---------------------------------------------------
  // The following is to allow multiple integrations: return the call 
  // unevaluated, if there are symbols. Note that numeric::indets does
  // not count the integration variable of unevalated calls to 
  // numeric::quadrature inside alg_expr;

  if numeric::indets([alg_expr, x0, x1]) minus {x} <> {} 
     then return(procname(args()))
  end_if;

  if has(alg_expr, hold(int)) then 
     alg_expr:= subs(alg_expr, hold(int) = numeric::quadrature);
  end_if;

  // the symbolic abs is much too expensive to carry around!
  if has(alg_expr, hold(abs)) then 
     lazyAbs:= proc(x)
               name lazyAbs; 
               begin
                   x:= float(x);
                   case domtype(x)
                   of DOM_FLOAT do
                   of DOM_COMPLEX do
                             return(specfunc::abs(x));
                   otherwise return(procname(args()))
                   end_case;
               end_proc:
     alg_expr:= subs(alg_expr, hold(abs) = lazyAbs);
  end_if;

  // ---------------------------------------------------
  // transformation for infinite intervals. Do not (!!!) split integral
  // into 2 parts, otherwise the adaptive strategy based on a first rough
  // estimate of the (entire!) integral may be unrealistic!

  if has([x0,x1], infinity) then
     // here we need x0<x1:
     if x0 =  infinity then x0:=x1; x1:= infinity; alg_expr:= -alg_expr; end_if;
     if x1 = -infinity then x1:=x0; x0:=-infinity; alg_expr:= -alg_expr; end_if;
     if x0 = -infinity and (not has(x1,infinity))
     then // transformation y=1/(x1+1-x), i.e., x=x1+1-1/y: 
           alg_expr:= subs(alg_expr,x=x1+1-1/x)/x^2; x0:=0: x1:=1: 
     end_if;
     if (not has(x0, infinity)) and x1 = infinity
     then // transformation y=1/(x-x0+1), i.e., x=x0-1+1/y: 
           alg_expr:= subs(alg_expr, x=x0-1+1/x)/x^2; x0:=0: x1:=1: 
     end_if;
     if x0 = -infinity and x1 = infinity
     then // split into int(..,-infinity..0)+int(..,0..infinity) 
          // transform as above and combine:
          alg_expr:= (subs(alg_expr, x=(x-1)/x)+ 
                      subs(alg_expr, x=(1-x)/x) )/x^2;
          x0:=0: x1:=1: 
     end_if;
     if has([x0,x1], infinity) then // if x0:= I + infinity etc.
        error("cannot handle this infinite interval");
     end_if;
  end_if;

  // ---------------------------------------------------
  // Check options:

  adaptive:= TRUE;   // default
  freeStyle:= TRUE:  // default: use GL=20,40,80,160
  maxcalls:= 0: // initialize maxcalls, will be set below

  for opt in [args(3..args(0))] do
    if has(opt, Gauss) or has(opt, GaussLegendre) or has(opt, GL)
    then method:= GL: freeStyle:= FALSE;
         if type(opt)<>"_equal" then
            error("specify method by 'method name = number of nodes'"); 
         end_if:
         if not testtype((n:= op(opt,2)), Type::PosInt) then 
            error("number of nodes must be a positive integer");
         end_if;
         next;
    end_if;
    if has(opt, GaussTschebyscheff) or has(opt, GaussChebyshev) 
    or has(opt, GT) or has(opt, GC)
    then method:= GT: freeStyle:= FALSE; 
         if type(opt)<>"_equal" then
            error("specify method by 'method name = number of nodes'"); 
         end_if:
         if not testtype((n:= op(opt,2)), Type::PosInt) then 
            error("number of nodes must be a positive integer");
         end_if;
         next;
    end_if;
    if has(opt, NewtonCotes) or has(opt, NC) 
    then method:= NC: freeStyle:= FALSE;
         if type(opt)<>"_equal" then
            error("specify method by 'method name = number of nodes'"); 
         end_if:
         if not testtype((n:= op(opt,2)), Type::PosInt) then 
            error("number of nodes must be a positive integer");
         end_if;
         next;
    end_if;
    if opt = (Adaptive=FALSE) then adaptive:= FALSE; next; end_if:
    if opt = (Adaptive=TRUE) then next; end_if:
    if opt = Adaptive then next; end_if:
    if has(opt, MaxCalls) then
         if type(opt)<>"_equal"
           then error("use MaxCalls = positive integer or infinity"); 
         end_if:
         maxcalls:= op(opt,2):
         if maxcalls = infinity then 
            // replace infinity by RD_INF: note that
            // bool(fcalls < RD_INF) is much faster than
            // bool(fcalls < infinity) !!!
            maxcalls:= RD_INF; 
            next;
         end_if;
         if not testtype(maxcalls, Type::PosInt) 
           then error("use MaxCalls = positive integer or infinity"); 
         end_if:
         next;
    end_if;
    error("unknown option");
  end_for:

  if not adaptive and freeStyle then
     error("need to specify non-adaptive quadrature method");
  end_if;
  if not freeStyle and method = GT and adaptive then
     warning("can use Gauss-Tschebyscheff method in ".
             "non-adaptive mode only. Switching to ".
             "GaussTschebyscheff=".expr2text(n).", Adaptive=FALSE ; ".
             "requested precision will not be achieved");
     adaptive:= FALSE;
  end_if:

  // ---------------------------------------------------
  // set default method, if not specified by user

  if freeStyle then
     method:=GL: adaptive:=TRUE:
     if   DIGITS<=10 then  n:=20; 
     elif DIGITS<=50 then  n:=40; 
     elif DIGITS<=100 then n:=80; 
     else n:=160;
     end_if;
  end_if;

  if maxcalls=0 then maxcalls:= MAXDEPTH*n: end_if;

  // ---------------------------------------------------

 //----------------------------------------------------------------
 // The data (weights b[i] and abscissae c[i]) are provided for    
 // Gauss-Legendre quadrature, for Gauss-Tschebyscheff-Quadrature  
 // and for Newton-Cotes quadrature:                               
 // See GLdata.mu , GTdata.mu and NCdata.mu                        
 //----------------------------------------------------------------

 if method=GL then ([b,c]):=numeric::gldata(n,DIGITS) end_if;
 if method=GT then ([b,c]):=numeric::gtdata(n) end_if;
 if method=NC then ([b,c]):=numeric::ncdata(n) end_if;
 ([b,c]):=float([b,c]);

 //-----------------define local subroutine ------------------------------
 // Quadsum computes                                                      
 //  Quadsum( f(x),x=x0..x1):=(x1-x0)*sum(b[i]*f(x0+c[i]*(x1-x0)),i=1..n) 
 // based on the quadrature formula                                       
 //    int(f(c),c=0..1)==sum( b[i]*f(c[i]),i=1..n)                        
 // with abscissae  c[i]  and weights  b[i]                               
 //-----------------------------------------------------------------------

 Quadsum:=proc(alg_expr,x0,x1,c)
 local L,i;
 begin // important: x0, x1 may be exact. Compute L:= float(x1-x0), not
       // L:= float(x1)-float(x0). The latter may become zero!
       L:=float(x1-x0);
       if iszero(L) then return(0);end_if;
       if fcalls>=maxcalls then return(0) end_if;
       fcalls:= fcalls + n;
       x0:= float(x0):

/* This is an attempt to avoid the costs of the symbolic
   evaluation of alg_expr. Unfortunately, it does not
   work, if alg_expr=hold(some_function)(x)
       L*_plus(b[i]*float(eval(subsop(hold(subs(ALG_EXPR,X=FLOATX)),
                                      1 = alg_expr,
                                      [2, 1]= x,
                                      [2, 2]= x0 + c[i]
                                      )
                               ))$ i=1..n);
*/

       L*_plus(b[i]*float(eval(subs(alg_expr,x=x0+c[i])))$ i=1..n);
 end_proc:

/*------------- define local subroutine ------------------------------------
 adaptiveQuad is a general adaptive algorithm based on a non-adaptive 
 quadrature rule Quadsum:
   int(f(x),x=0..1)==sum(b[i]*f(c[i],i=1..n)=:Quadsum(f,x=0..1)
 given by the quadrature data b[i] (weights) and c[i] (abscissae).
 First,  an estimate  Is:= Quadsum(f,x=x0..x1) is computed.
 Then, by recursive bisectioning of the interval, the method Quadsum is used 
 to compute the integrals over each half of the interval.
 Bisectioning is stopped, when  
    Quadsum(first half)+ Quadsum(second half)-Quadsum(complete interval)
 has become so small that this difference cannot be added to the estimate
 Is within float precision.
 Note:  let  p be the order of the quadrature rule, i.e.
    Error( Quadsum( f,  x=x0..x1 ) )  =  constant( f ) * (x1-x0)^p
 Then bisectioning will occur down to subintervals of length L, until  
       constant(f) * L^p == Is * 10^(-DIGITS)
 holds,  i.e.,  L = O( 10^(-DIGITS/p ) :  the costs grow
 exponentially with DIGITS !
 For very low precision low order methods Quadsum suffice.  
 FOR HIGH PRECISION A HIGH ORDER METHOD Quadsum IS NECESSARY !
 A reasonable (heuristic) choice seems Gauss-Legendre-Quadrature
 with n nodes which has order (2n+1). 
 Example: n=20, for which the abscissae c[j] and weights b[j]
 can be computed in a fraction of a second, if DIGITS is not too
 high. Each bisection step then produces quadrature values with errors
 whose sum is approximately the error of the quadrature on the double 
 interval multplied by a factor 
     2^(-40)==10^(-12)
 Thus, for slowly varying integrands, approximately DIGITS/12
 bisection steps are needed to reduce the error of Quadsum(whole interval)
 by DIGITS decimal places. This leads to
   2^(DIGITS/12)
 subintervals, each requiring n=20 evaluations of the integrand.
---------------------------------------------------------------------*/
  // Bisect interval x0..x1. Compare sum of quadratures
  // I21 = int(..,x=x0..xmid) and I22 = int(..,x=xmid..x1) with
  // I1 = int(..,x=x0..x1). Stop bisectioning, if no
  // significant change between I2=I21+I22 and I1.

  adaptiveQuad:=proc(alg_expr,x0,x1,c,bisectionlevel,I1)
     local mid,I21,I22,I2;
  begin
     // The following line passes the information, that another
     // call to adaptiveQuad has reached MAXDEPTH. Stop integrating
     // in all calls of adaptiveQuad:
     if problem then return(I1); end_if;
     if bisectionlevel>MAXDEPTH-5 then
        if interactive then
          warning("Precision goal not achieved after MAXDEPTH=".
                   expr2text(MAXDEPTH)." recursive calls!\n".
                  "There may be a singularity of ".expr2text(alg_expr).
                  " close to ".expr2text(x)."=".expr2text(float((x0+x1)/2)).".\n".
                  "Increase MAXDEPTH and try again for a more accurate result!");
        end_if;
        problem:= TRUE:
        return(I1);
     end_if;
     // use exact arguments x0, x1 for Quadsum and adaptiveQuad:
     mid:=(x0+x1)/2; 
     c:=map(c,_mult,0.5);
     I21:=Quadsum(alg_expr,x0,mid,c); 
     I22:=Quadsum(alg_expr,mid,x1,c);
     I2:=I21+I22;

     // use global estimate Is of integral for stopping criterion
     if specfunc::abs(I2-I1) <= (specfunc::abs(Is)
                                +specfunc::abs(I1)
                                +specfunc::abs(I2))*macheps
     or fcalls>=maxcalls
     then return(I2) 
     else //-------------------------------------------------------------
          // The bisectionlevel gives the length of the present bisected 
          // interval = original interval / 2^(bisectionlevel)           
          //-------------------------------------------------------------
          // Call right half of interval first !! Note that infinite
          // intervals are mapped to int(f(1/x)/x^2,x=0..1). In this 
          // case the right half converges faster than the left half.
          // Make sure that right half produces good results, before
          // the left half produces fcalls>=maxcalls and stops the 
          // This way the inexact results returned with warnings are
          // still reasonably good!!
          return(adaptiveQuad(alg_expr,mid,x1,c,bisectionlevel+1,I22)
                +adaptiveQuad(alg_expr,x0,mid,c,bisectionlevel+1,I21));
     end_if;
  end_proc:
  //----------- end of adaptiveQuad ----------------------------

  // now do the quadrature: 
  interactive:= testargs():
  problem:= FALSE:
  // fcalls is the number of calls to the integrand
  fcalls:= 0; // each call to Quadsum increases fcalls by n
  alg_expr:=float(alg_expr);
  macheps:=float(10^(-DIGITS));
  c:=map(c,_mult,float(x1-x0));
  Is:= Quadsum(alg_expr,x0,x1,c); // non-adaptive rough estimate
  userinfo(2, "rough estimate = ".expr2text(Is));
  if adaptive then // improve first estimate
     userinfo(2, "starting adaptive quadrature on the interval ".expr2text(x0..x1));
     Is:= adaptiveQuad(alg_expr,x0,x1,c,0,Is);
  end_if;
  delete adaptiveQuad, Quadsum;
  if fcalls>=maxcalls then
     if interactive then
        warning("Precision goal not achieved after ".
                expr2text(fcalls). " function calls!\n".
                "Increase MaxCalls and try again for a more accurate result.");
     end_if;
     problem:= TRUE;
  end_if:
  userinfo(1, "result after ".expr2text(fcalls)." function calls:");
  if not interactive and problem then return(FAIL) end_if;
  return(Is);
end_proc:
//----------- end of Quadrature ----------------------------

