// 
// $Date: 2000/10/30 15:02:01 $  $Author: walter $  $Revision: 1.14 $ 
// 
// walter, 06/03/98 
/*--------------------------------------------------------------------------
  cubicSpline( pair0 , pair1, .. <, Symbolic > <, BoundaryCondition > )

  Parameter:

    pair0, pair1, ..   -- pairs of numbers [x[0],y[0]], [x[1],y[1]], ..
    Symbolic           -- option
    BoundaryCondition  -- option Natural or Periodic or NotAKnot or
                          Complete=[a,b], where a,b are numerical values

  Synopsis:

    S:=cubicSpline([x[0],y[0]],..,[x[n],y[n]],Option) is the cubic spline
    function interpolating the table [x[0],y[0]],..,[x[n],y[n]], i.e, 
    S(x[0])=y[0],..,S(x[n])=y[n]. The spline function is a piecewise 
    polynomial of degree <=3 on the intervals 
         (-infinity,x[1]], [x[1],x[2]], .. ,[x[n-1],infinity).
    S and its first two derivatives S',S'' are continuous at the points
    x[1],..,x[n-1].
    
    The 4 types of boundary conditions that may be specified are:
    *  Natural :  S''(x[0])=S''(x[n])=0.
    *  Periodic:  S(x[0])=S(x[n]), S'(x[0])=S'(x[n]), S''(x[0])=S''(x[n]).
       If y[0]<>y[n], then cubicSpline returns an error message.
    *  NotAKnot:  S''' is continuous at the points x[1] and x[n-1].
    *  Complete=[a,b] : S'(x[0])=a, S'(x[n])=b.
    The default is the NotAKnot boundary condition.

    The abscissae x[0] < .. < x[n] should be numerical real values 
    in ascending order. Otherwise cubicSpline will reorder the input
    pairs to achieve such an ordering. With the option Symbolic
    the ordering is not checked!!!!

    For y[0],..,y[n] both numerical as well as symbolic values are
    admissible.

    By default all input data are converted to floats. With the option 
    Symbolic this conversion may be suppressed. In this case symbolic 
    abscissae x[i] are admissible, which are assumed to be ordered.

    The spline function S returned by cubicSpline may be called in the 
    form S(z), where z is a numerical value. If S was generated with
    symbolic nodes (using the option Symbolic), then the spline function
    cannot be evaluated.

    Alternatively, S(z,i) returns the representation of the spline assuming
    x[i] <= z < x[i+1]. In this case S may contain symbolic nodes and
    z may be symbolic.

  Examples:

  >> S:=numeric::cubicSpline([i, sin(i*PI/20)/(1+i^2/1000)] $ i=0..100):
  >> S(11/5);
                                  0.3371060912
    
  >> plotfunc(hold(S)(x),x=0..100); (plot the spline function)

  >> S:=numeric::cubicSpline([i, y.i] $ i=0..2, Symbolic, Complete=[a,5]):
  >> collect( S(x,0),x);

                 3 / 3 a   5 y0          3 y2       \
     y0 + a x + x  | --- + ---- - 2 y1 + ---- - 5/4 | +
                   \  4     4             4         /

         2 /        7 a   9 y0   3 y2       \
        x  | 3 y1 - --- - ---- - ---- + 5/4 |
           \         4     4      4         /

------------------------------------------------------------------------------*/

numeric::cubicSpline:=proc()
option escape;
local boundaryCondition,symbolic,yp0,ypn,
      n,nm1,x,X,y,Y,h,hSet,b,B,c,C,d,DD,i,a,A,tmp;
begin
   if args(0)<1 then error("not enough arguments"); end_if;
   boundaryCondition:=NotAKnot; symbolic:=FALSE;
   for i from 1 to args(0) do
     if domtype(args(i))=DOM_LIST then
        if nops(args(i))<>2 then 
          error("input data must be of the form [x0,y0], [x1,y1], .. ");
        end_if;
        x[i-1]:=args(i)[1];
        y[i-1]:=args(i)[2];
     else if args(i)=Symbolic then symbolic:=TRUE; 
     else if args(i)=Natural  then boundaryCondition:=Natural; 
     else if args(i)=Periodic then boundaryCondition:=Periodic;
     else if args(i)=NotAKnot then boundaryCondition:=NotAKnot
     else if has(args(i),Complete)
             then boundaryCondition:=Complete; 
                  if testtype(args(i),"_equal")
                     and domtype(op(args(i),2))=DOM_LIST
                     and nops(op(args(i),2))=2 
                  then yp0:=op(args(i),[2,1]); ypn:=op(args(i),[2,2]);
                  else error("Use Complete=[a,b] to specify complete boundary ".
                             "data S'(x[0])=a, S'(x[n])=b of spline S(x)")
                  end_if
     else error("wrong type of input data or unknown option")
     end_if;end_if;end_if;end_if;end_if;end_if;
   end_for;
   n:=nops(x)-1:

   if n=0                    /* just to avoid warning about  */
      then h[0]:= float(0);  /* unitialized variable. This h */
   end_if;                   /* will never be used */
   if not symbolic
   then x:=map(x,float); 
        y:=map(y,float);
        (h[i]:= x[i+1]-x[i];) $ i=0..n-1;
        if boundaryCondition=Complete then
          yp0:=float(yp0); yp0:=subs(yp0,float(0)=0);
          ypn:=float(ypn); ypn:=subs(ypn,float(0)=0);
        end_if;
        hSet:= {h[i] $ i=0..n-1};
        if n>0 then
           if map(hSet,domtype)<>{DOM_FLOAT} 
              then error("use option Symbolic for symbolic nodes");
           end_if;
           if map(hSet, x-> bool(x>0) )<> {TRUE} then
              warning("Abscissae are not in ascending order. Input data ".
                      "will be reordered automatically.");
              tmp:= sort([[x[i],y[i]] $ i=0..n], (a,b)-> bool(a[1]<b[1]));
              (x[i]:= tmp[i+1][1];) $ i=0..n;
              (y[i]:= tmp[i+1][2];) $ i=0..n;
              (h[i]:= x[i+1]-x[i];) $ i=0..n-1;
           end_if;
           if map({op(h)}, _not@iszero@op,2)<> {TRUE} then
              error("Abscissae must be distinct");
           end_if;
        end_if;
   else (h[i]:= x[i+1]-x[i];) $ i=0..n-1;
   end_if;
   x:=subs(x,float(0)=0);
   y:=subs(y,float(0)=0);
   h:=subs(h,float(0)=0);

if boundaryCondition=Natural then
   if n=0 then b[0]:=0;c[0]:=0;d[0]:=0; end_if;
   if n=1 then b[0]:=(y[1]-y[0])/h[0];c[0]:=0;d[0]:=0; end_if;
   if n>1 then
      (b[i]:= (y[i+1]-y[i])/h[i];) $ i=0..n-1;
      (c[i]:= b[i]-b[i-1];) $ i=1.. n-1;
       d[1]:=2*(h[0]+h[1]);
      (d[i]:=2*(h[i-1]+h[i])-h[i-1]^2/d[i-1];) $ i=2..n-1;
      (c[i]:=c[i]-c[i-1]*h[i-1]/d[i-1];) $ i=2..n-1:
       c[n-1]:=c[n-1]/d[n-1]; c[n]:=0;
      for i from n-2 downto 1 do c[i]:=(c[i]-c[i+1]*h[i])/d[i]; end_for;
       c[0]:=0; 
      (b[i]:=b[i]-(2*c[i]+c[i+1])*h[i];) $ i=0..n-1;
      (d[i]:=(c[i+1]-c[i])/h[i];) $ i=0..n-1;
      (c[i]:=3*c[i];) $ i=0..n-1;
   end_if;
end_if;

if boundaryCondition=Periodic then
   if symbolic
   then if y[n]<>y[0] 
           then error("input data not periodic: y[0]<>y[n] ") end_if;
   else if abs(y[n]-y[0])>abs(y[0])*10^(1-DIGITS)
           then error("input data not periodic: y[0]<>y[n] ") end_if;
   end_if;
   if n=0 then b[0]:=0;c[0]:=0;d[0]:=0; end_if;
   if n=1 then b[0]:=0;c[0]:=0;d[0]:=0; end_if;
   if n=2 then b[0]:=(y[1]-y[0])*(h[1]-h[0])/h[0]/h[1]; b[1]:=b[0];
               c[0]:=3*(y[1]-y[0])/h[0]/h[1]; c[1]:=-c[0];
               d[0]:=2*c[1]/3/h[0]; d[1]:=2*c[0]/3/h[1];
   end_if;
   if n>2 then
      (b[i]:= (y[i+1]-y[i])/h[i];) $ i=0..n-1;
       c[0]:= b[0]-b[n-1];   (c[i]:= b[i]-b[i-1];) $ i=1.. n-1;
       d[0]:=2*(h[n-1]+h[0]);
      (d[i]:=2*(h[i-1]+h[i])-h[i-1]^2/d[i-1];) $ i=1..n-2;
       a[0]:=h[n-1]/d[0]; (a[i]:=-h[i-1]*a[i-1]/d[i];) $ i=1..n-3; 
     a[n-2]:=(h[n-2]-h[n-3]*a[n-3])/d[n-2];
       A[0]:=h[n-1]: (A[i]:=-h[i-1]*A[i-1]/d[i-1];) $ i=1..n-3; 
     A[n-2]:=(h[n-2]-h[n-3]*A[n-3])/d[n-3];
     d[n-1]:=2*(h[n-2]+h[n-1])-_plus( a[i]*A[i] $ i=0..n-2 );
      (c[i]:=c[i]-c[i-1]*h[i-1]/d[i-1];) $ i=1..n-2;
     c[n-1]:=c[n-1]-_plus(c[i]*a[i] $ i=0..n-2);
     c[n-1]:=c[n-1]/d[n-1]; c[n-2]:=(c[n-2]-A[n-2]*c[n-1])/d[n-2];
     for i from n-3 downto 0 do 
       c[i]:=(c[i]-c[i+1]*h[i]-c[n-1]*A[i])/d[i]; 
     end_for;
       c[n]:=c[0]; 
      (b[i]:=b[i]-(2*c[i]+c[i+1])*h[i];) $ i=0..n-1;
      (d[i]:=(c[i+1]-c[i])/h[i];) $ i=0..n-1;
      (c[i]:=3*c[i];) $ i=0..n-1;
   end_if;
end_if;

if boundaryCondition=Complete then
   if n=0 then if yp0<>ypn then 
                  error("boundary data ".
                         expr2text(Complete=[yp0,ypn]).
                        " make no sense for single point input");
               end_if;
               b[0]:=yp0;c[0]:=0;d[0]:=0;end_if;
   if n=1 then b[0]:=yp0;
               c[0]:= 3*(y[1]-y[0])/h[0]^2-(2*yp0+ypn)/h[0];
               d[0]:=-2*(y[1]-y[0])/h[0]^3+(  yp0+ypn)/h[0]^2;
   end_if;
   if n>1 then
      (b[i]:= (y[i+1]-y[i])/h[i];) $ i=0..n-1;
       c[0]:= b[0]-yp0; (c[i]:= b[i]-b[i-1];) $ i=1.. n-1;
       c[n]:= ypn-b[n-1];
       d[0]:=2*h[0]; (d[i]:=2*(h[i-1]+h[i])-h[i-1]^2/d[i-1];) $ i=1..n-1;
       d[n]:=2*h[n-1]-h[n-1]^2/d[n-1];
      (c[i]:=c[i]-c[i-1]*h[i-1]/d[i-1];) $ i=1..n: c[n]:=c[n]/d[n];
      for i from n-1 downto 0 do c[i]:=(c[i]-c[i+1]*h[i])/d[i]; end_for;
      (b[i]:=b[i]-(2*c[i]+c[i+1])*h[i];) $ i=0..n-1;
      (d[i]:=(c[i+1]-c[i])/h[i];) $ i=0..n-1;
      (c[i]:=3*c[i];) $ i=0..n-1;
   end_if;
end_if;

if boundaryCondition=NotAKnot then
   if n=0 then b[0]:=0;c[0]:=0;d[0]:=0; end_if;
   if n=1 then b[0]:=(y[1]-y[0])/h[0];c[0]:=0;d[0]:=0; end_if;
   if n=2 then c[0]:=((y[1]-y[0])/h[0])/(h[0]+h[1]):
               c[1]:=((y[2]-y[1])/h[1])/(h[0]+h[1]):
               b[0]:=(2*h[0]+h[1])*c[0]-h[0]*c[1];
               b[1]:=        h[1] *c[0]+h[0]*c[1];
               c[0]:= c[1]-c[0]; c[1]:= c[0];
               d[0]:= 0; d[1]:=0:
   end_if;
   if n=3 then
       (b[i]:= (y[i+1]-y[i])/h[i];) $ i=0..n-1;
       (c[i]:= b[i]-b[i-1];) $ i=1.. n-1;
        d[0]:=3*(x[3]-x[1])*(x[2]-x[0])*(x[3]-x[0]); d[1]:=c[1]:
        c[1]:=((h[1]+h[2])*(2*h[1]+h[2])*d[1]+(h[0]^2-h[1]^2)*c[2])/d[0];
        c[2]:=((h[0]+h[1])*(2*h[1]+h[0])*c[2]+(h[2]^2-h[1]^2)*d[1])/d[0];
        c[0]:=c[1]*(1+h[0]/h[1])-c[2]*h[0]/h[1]:
        c[3]:=c[2]*(1+h[2]/h[1])-c[1]*h[2]/h[1];
       (b[i]:=b[i]-(2*c[i]+c[i+1])*h[i];) $ i=0..2;
       (d[i]:=(c[i+1]-c[i])/h[i];) $ i=0..2;
       (c[i]:=3*c[i];) $ i=0..2;
   end_if;
   if n>3 then
      (b[i]:= (y[i+1]-y[i])/h[i];) $ i=0..n-1;
      (c[i]:= b[i]-b[i-1];) $ i=1.. n-1;
       d[1]:=(h[0]+h[1])*(2+h[0]/h[1]);
       d[2]:=2*(h[1]+h[2])-(h[1]^2-h[0]^2)/d[1];
      (d[i]:=2*(h[i-1]+h[i])-h[i-1]^2/d[i-1];) $ i=3..n-2;
     d[n-1]:=(h[n-2]+h[n-1])*(2+h[n-1]/h[n-2])-(h[n-2]^2-h[n-1]^2)/d[n-2];
      (c[i]:=c[i]-c[i-1]*h[i-1]/d[i-1];) $ i=2..n-2:
     c[n-1]:=(c[n-1]-c[n-2]*( h[n-2]-h[n-1]^2/h[n-2])/d[n-2])/d[n-1];
     for i from n-2 downto 2 do c[i]:=(c[i]-c[i+1]*h[i])/d[i]; end_for;
       c[1]:=(c[1]-c[2]*(h[1]-h[0]^2/h[1]))/d[1]:
       c[0]:=c[1]*(1+h[0]/h[1])-c[2]*h[0]/h[1]:
       c[n]:=c[n-1]*(1+h[n-1]/h[n-2])-c[n-2]*h[n-1]/h[n-2];
      (b[i]:=b[i]-(2*c[i]+c[i+1])*h[i];) $ i=0..n-1;
      (d[i]:=(c[i+1]-c[i])/h[i];) $ i=0..n-1;
      (c[i]:=3*c[i];) $ i=0..n-1;
   end_if;
end_if;

   if n=0 then n:=1 end_if;
   nm1:=n-1; X:=x; Y:=y; B:=b; C:=c; DD:=d;

   proc(z) local i, h, x, a, b, c, d, indexsearch;
   begin
      x:=X; a:=Y; b:=B; c:=C; d:=DD;
      if args(0)=0 then error("expecting at least one argument") end_if;
      if args(0)=2
      then i:=args(2);
           if domtype(i)<>DOM_INT then
              error("second argument must be integer");
           end_if;
           i:=min(max(i, 0), nm1);
      else if domtype(float(z))<>DOM_FLOAT then
// return the procedure call unevaluated.  The attempts which have
// been commented out all fail, because they all replace the call
// with a DOM_IDENT.  The last version returns the DOM_IDENT iff
// the current procedure has been assigned to that ident.
              // return(procname(args()));
              // do not use procname, but 
              // (op(op(_act_proc_env(),1),6) (= procname)
//              return(op(op(_act_proc_env(),1),6)(args()));
              if eval( op(op(_act_proc_env(),1),6)) = op(_act_proc_env(),1) then
                return(procname(args()));
              else
                return(subsop(hold(ret)(args()),
                       0=op(_act_proc_env(),1), Unsimplified));
              end_if;
              //error("numerical real value expected");
           end_if;
           indexsearch:=proc(k,j) local m;
           begin
              if j=k+1 then return(k) end_if;
              m:=trunc((k+j)/2);
              if domtype(float(z-x[m]))<>DOM_FLOAT then
                 error("spline function contains symbolic nodes");
              end_if;
              if float(z-x[m])<0 then indexsearch(k,m)
                                 else indexsearch(m,j)
              end_if;
           end_proc;
           if domtype(float(z-x[0]))<>DOM_FLOAT then
              error("spline function contains symbolic nodes");
           end_if;
           if float(z-x[0])<0 then i:=0
             else if float(z-x[nm1])>=0 then i:=nm1
                                        else i:=indexsearch(0,nm1)
                  end_if;
           end_if;
      end_if;
      delete indexsearch;
      h:=(z-x[i]): eval(((d[i]*h+c[i])*h+b[i])*h+a[i]):
    end_proc;
end_proc:
