/*-------------------------------------------------------------------------------*/
/* Based on diploma thesis                                                       */
/*   "Numerik dynamischer Systeme auf Lie-Gruppen, Renate Wagner, Dez.1998.      */
/* Code rewritten and optimized by W. Oevel, 1.1.1999                            */
/*-------------------------------------------------------------------------------*/
/* numeric::expMatrix - the exponential of a matrix                              */
/*                                                                               */
/* Aufruf:                                                                       */
/*                                                                               */
/* 1) numeric::expMatrix(A, <method>)                                            */
/* 2) numeric::expMatrix(A, v, <method>)                                         */
/* 3) numeric::expMatrix(A, V, <method>)                                         */
/*                                                                               */
/* Parameter:                                                                    */
/*   A      : nxn Matrix (array(1..n,1..n,..)                                    */
/*   v      : n Vektor (list or array(1..n,..)                                   */
/*   V      : nxp Matrix (array(1..n,1..p,..)                                    */
/* method   : TaylorExpansion or Krylov or Interpolation or Diagonalization      */
/*                                                                               */
/* Rueckgabewert:                                                                */
/*  1) exp(A)   = array(1..n,1..n,..)                                            */
/*  2) exp(A)*v = Liste oder array(1..n,..). Return type = input type            */
/*  3) exp(A)*V = array(1..n,1..p,..)                                            */
/*                                                                               */
/* Beschreibung:                                                                 */
/*                                                                               */
/* Die n x n Matrix A kann vom Typ DOM_ARRAY oder von der                        */
/* Kategorie Cat::Matrix sein.                                                   */
/* Der Vektor v kann eine Liste [v1,..,v.n] oder vom Typ DOM_ARRAY               */
/* (d.h., v=array(1..n,[v1,..,v.n])) sein.                                       */
/* Die n x p Matrix V kann DOM_ARRAY oder Cat::Matrix sein.                      */
/*                                                                               */
/* Die zur Verfuegung stehenden Methoden sind:                                   */
/*                                                                               */
/* - Default:                                                                    */
/*   ohne Angabe von <method> wird mittels einer Heuristik zwischen 2 Varianten  */
/*   von TaylorExpansion gewaehlt.                                               */
/* - TaylorExpansion:                                                            */
/*   Variante 1:                                                                 */
/*   Geeignetes Aufsummieren der Taylor-Reihe exp(A)*V = V+A*V+A^2*V/2!+...      */
/*   Falls kein V angegeben ist, so wird exp(A) = 1+A+A^2/2!+.. berechnet.       */
/*   Der Aufwand fuer exp(A) ist O(norm(A)*n^3), fuer exp(A)*V mit V= nxp-Matrix */
/*   ist er O(norm(A)*n^2*p). Damit unguenstig bei norm(A)>>1, guenstig bei      */
/*   1<<n, p<<n.                                                                 */ 
/*   Variante 2:                                                                 */
/*   Die Matrix A wird auf kleine Norm skaliert, dann wird die Taylor-Reihe      */
/*   aufsummiert und das Ergebnis wird durch repeated squaring zu exp(A) hoch-   */
/*   multipliziert. Gegebenenfalls wird zuletzt daraus exp(A)*v gebildet.        */
/*   Diese Version ist wenig geeignet, wenn die Dimension n sehr gross ist       */
/*   und nicht exp(A), sondern nur exp(A)*v benoetigt wird.                      */
/*   Der Aufwand ist O( ln(norm(A)*n^3 + n^2*p). Guenstig bei norm(A)>>1,        */
/*   unguenstig bei n>>1.                                                        */
/* - Krylov:                                                                     */
/*   Kann nur in der Version expMatrix(A,v,Krylov) oder expMatrix(A,V,Krylov)    */
/*   mit p=1 benutzt werden. Es wird das Arnoldi-Algorithmus verwendet.          */
/*   Geeignet fuer grosse Matrizen in folgenden Faellen:                         */
/*    * v ist von wenigen Eigenvektoren von A aufgespannt,                       */
/*    * A hat stark entartete Eigenwerte oder Cluster mit aehnlichen Eigenwerten.*/
/*   Achtung: der Arnoldi-Algorithmus neigt in einigen Faellen zur Instabilitaet.*/
/*   Achtung: Fehler ist duch norm(Delta exp(A)*v) <= norm(exp(A)*v)*10^(-DIGITS)*/
/*            gegeben. TaylorExpansion berechnet jede Komponente **einzeln** bis */
/*            auf relative Genauigkeit 10^(-DIGITS).                             */
/* - Interpolation:                                                              */
/*   Berechnet die Eigenwerte lambda von A und bestimmt ein Hermite-             */
/*   interpolierendes Polynom P mit P(lambda)=exp(lambda). Es folgt exp(A)=P(A). */
/*   Fuer exp(A)*V werden etwa n^3*p skalare Multiplikation benoetigt.           */
/*   Die Interpolation ist fuer n>>1 oft schlecht konditioniert und fuehrt zu    */
/*   starkem Anwachsen von Rundungsfehlern. Dies wird intern automatisch         */
/*   ueberprueft, expMatrix bricht bei schlechter Konditionierung mit einer      */
/*   Fehlermeldung ab.                                                           */
/* - Diagonalization:                                                            */
/*   Die Matrix A wird diagonalisiert: A = X diag(lambda1,lambda2,..) X^(-1).    */
/*   Das Ergebnis wird per exp(A) = X diag(exp(lambda1),exp(lambda2),..) X^(-1)  */
/*   berechnet, gegebenenfalls wird hieraus exp(A)*v, exp(A)*V ermittelt.        */
/*   Achtung: dies funktioniert nicht fuer beliebiges A! Nichtdiagonalisierbare  */
/*   Matrizen fuehren zu einer Fehlermeldung, bzw. es kommt zu einer grossen     */
/*   Verstaerkung von Rundungsfehlern, wenn die Nichtdiagonalisierbarkeit        */
/*   numerisch nicht ermittelt werden kann.                                      */
/*   Diese Methode ist aber FAIL safe fuer hermitesche oder schief-hermitesche   */
/*   Matrizen A.                                                                 */
/*                                                                               */
/* Die Algorithmen benutzen intern Polynomarithmetik.                            */
/*                                                                               */
/* Examples:                                                                     */
/*  >> A:= array(1..2,1..2,[[0,1],[-1,0]]):                                      */
/*  >> numeric::expMatrix(A);                                                    */
/*                                                                               */
/*                    +-                             -+                          */
/*                    |   0.5403023058, 0.8414709848  |                          */
/*                    |                               |                          */
/*                    |  -0.8414709848, 0.5403023058  |                          */
/*                    +-                             -+                          */
/*                                                                               */
/*  >> v1:= [1, 1]: v2:= array(1..2, [1, 1]):                                    */
/*  >> numeric::expMatrix(A, v1), numeric::expMatrix(A, v2, Krylov);             */
/*                                                                               */
/*                                    +-                         -+              */
/*       [1.38177329, -0.3011686789], | 1.38177329, -0.3011686789 |              */
/*                                    +-                         -+              */
/*-------------------------------------------------------------------------------*/

numeric::expMatrix:=proc(A)
local n, Arg, expAv, expA, v, p, erg, return_type, i, j, k, s,
      norm_A, method, sollNorm, NN, NN1, NN2, kk, x, APoly, vPoly,
      MatrixMult, tol, ergPoly, ergPoly_v, Apower, Apower_v,
      vNorm, KrylovBase, H, arnoldi, last_m, m, Hmm, expHe1, Err, 
      norm_Erg, eigenvals, absolute_tol, c, cc, pp, tPoly, A_is_real, 
      v_is_real, d, d2, X, dummy;
save DIGITS;
begin
  if args(0)<1 then error("not enough arguments") end_if;
  if args(0)>3 then error("too many arguments") end_if;

  /*  convert A to DOM_ARRAY */
  if A::dom::hasProp(Cat::Matrix)=TRUE then A:= expr(A) end_if;
  if domtype(A)<>DOM_ARRAY then 
     error("matrix must be an array or of category 'Cat::Matrix'");
  end_if;

  n:= op(A,[0,2,2]); /*  n ist die Dimension der Eingabematrix */
  if n<>op(A,[0,3,2]) then error("not a square matrix") end_if;

  expAv:= FALSE;   /*  no vector specified, yet */
  method:= 0;      /*  no method specified, yet */
  p:= n;           /*  no vector v specified, yet*/
  return_type:= 0; /*  default: return array.    */

  if args(0)>1 then
     for Arg in [args(2..args(0))] do
       /*  convert DOM_LIST and Cat::Matrix objects to DOM_ARRAY: */
       if domtype(Arg) = DOM_LIST
          then expAv:= TRUE;
               if nops(Arg)<>n then
                  error("dimensions of matrix and vector do not match");
               end_if;
               return_type:= 1;
               v:=array(1..n,1..1);
               (v[i,1]:= Arg[i];) $ i=1..n;
               p:= 1;  /*  v = n x p - matrix */
               next;
       end_if;
       if domtype(Arg)=DOM_ARRAY 
          then expAv:= TRUE;
               if n<>op(Arg,[0,2,2]) then
                  error("dimensions do not match");
               end_if;
               case op(Arg,[0,1])
               of 1 do return_type:= 2; /* return "Vector" = array(1..n,[..]) */
                       v:= array(1..n,1..1);
                       (v[i,1]:= Arg[i];) $ i=1..n;
                       p:= 1;  /*  v = n x p - matrix */
                       break;
               of 2 do v:= Arg;
                       p:=op(v,[0,3,2]); /* v = n x p - matrix */
                       break;
               otherwise error("wrong format of second argument");
               end_case;
               next;
       end_if;
       if Arg::dom::hasProp(Cat::Matrix)=TRUE
          then expAv:= TRUE;
               v:= expr(Arg);
               if domtype(v)<>DOM_ARRAY then
                  error("2nd argument must be an array, a list, or of category 'Cat::Matrix'");
               end_if;
               if n<>op(v,[0,2,2]) then
                  error("dimensions do not match");
               end_if;
               p:=op(v,[0,3,2]); /* v = n x p - matrix */
               next;
       end_if;
       case Arg 
          of TaylorExpansion do method:=0; break; /*  method0 splits up into method=1 */
          of Krylov          do method:=3; break; /*  or method=2, see below */
          of Interpolation   do method:=4; break;
          of Diagonalization do method:=5; break;
          otherwise error("unknown option")
       end_case;
     end_for;
  end_if;

  A:= map(A,float); if expAv then v:= map(v,float); end_if;

  if map({op(A)}, domtype) minus {DOM_FLOAT,DOM_COMPLEX}<>{}
     then error("non-numerical entry in matrix")
  end_if;

  if n=1 then 
     if expAv
     then case return_type
          of 1 do return([exp(A[1,1])*v[1,1]]);
          of 2 do erg:=array(1..1);
                  erg[1]:= exp(A[1,1])*v[1,1];
                  return(erg);
          otherwise erg:= array(1..1,1..p);
                    (erg[1,j]:= exp(A[1,1])*v[1,j];) $ j=1..p;
                    return(erg);
          end_case;
     else erg:= array(1..1,1..1);
          erg[1,1]:= exp(A[1,1]);
          return(erg);
     end_if;
  end_if;

  if n=2 and A[2,2]=-A[1,1] then /* special formula for traceless 2x2 matrix */
     userinfo(1, "using special formula for traceless 2 x 2 matrices");
     d:= A[1,1]^2+A[1,2]*A[2,1];
     d:= float(specfunc::sqrt(d));
     if iszero(d) then s:=1 else s:= sinh(d)/d; end_if;
     c:= cosh(d);
     expA:= array(1..2,1..2):
     expA[1,1]:= c+s*A[1,1]: expA[1,2]:=   s*A[1,2]:  
     expA[2,1]:=   s*A[2,1]: expA[2,2]:= c+s*A[2,2]:  
     /* stabilize the smaller of the diagonal elements using
        det( exp(traceless) ) = 1 : */
     if specfunc::abs(expA[1,1]) > specfunc::abs(expA[2,2])
       then expA[2,2]:= (1+expA[2,1]*expA[1,2])/expA[1,1];
       else expA[1,1]:= (1+expA[2,1]*expA[1,2])/expA[2,2];
     end_if;
     expA:= subs(expA, float(0)=0);
     if expAv
     then case return_type
          of 1 do return([expA[1,1]*v[1,1]+expA[1,2]*v[2,1],
                          expA[2,1]*v[1,1]+expA[2,2]*v[2,1]]);
          of 2 do return(array(1..2,
                         [expA[1,1]*v[1,1]+expA[1,2]*v[2,1],
                          expA[2,1]*v[1,1]+expA[2,2]*v[2,1]]));
          otherwise return(array(1..2,1..p,
                         [[(expA[1,1]*v[1,j]+expA[1,2]*v[2,j]) $ j=1..p ],
                          [(expA[2,1]*v[1,j]+expA[2,2]*v[2,j]) $ j=1..p ]
                         ]));
          end_case;
     else return(expA);
     end_if;
  end_if;

  if n=3 and             /* special formula for A in so(3)      */
     iszero(A[1,1]) and  /* For     [  0 -x3  x2]      [x1]     */
     iszero(A[2,2]) and  /*     A = [ x3  0  -x1], x = [x2]     */
     iszero(A[3,3]) and  /*         [-x2  x1  0 ]      [x3]     */
     A[1,2]=-A[2,1] and  /*                                     */  
     A[3,1]=-A[1,3] and  /* exp(A) = cos(d)*Eins + sin(d)/d *A  */
     A[3,2]=-A[2,3]      /*          + (1-cos(d))/d^2  x*x^T    */
     then                /* with d^2=x1^2+x^2+x^3               */
       userinfo(1,"using special formula for exp on so(3)");
       d2:= A[1,2]^2 + A[1,3]^2 + A[2,3]^2;
       d:= float(specfunc::sqrt(d2));
       c:= cos(d);
       if specfunc::abs(d2)< float(10^(-DIGITS))
        then s:=1-d2/6;  cc:= 1/2-d2/24 
        else s:=sin(d)/d; cc:= (1-cos(d))/d2;
       end_if;
       expA:= array(1..3,1..3,[
           [c + cc* A[3,2]^2 ,s*A[1,2]+cc*A[3,2]*A[1,3],s*A[1,3]+cc*A[3,2]*A[2,1]],
           [s*A[2,1]+cc*A[1,3]*A[3,2],c +  cc*A[1,3]^2,s*A[2,3]+ cc*A[1,3]*A[2,1]],
           [s*A[3,1]+cc*A[2,1]*A[3,2],s*A[3,2]+cc*A[2,1]*A[1,3],c +  cc*A[2,1]^2 ]
                              ]);
       expA:= subs(expA, float(0)=0);
       if expAv
       then case return_type
            of 1 do return([expA[1,1]*v[1,1]+expA[1,2]*v[2,1]+expA[1,3]*v[3,1],
                            expA[2,1]*v[1,1]+expA[2,2]*v[2,1]+expA[2,3]*v[3,1],
                            expA[3,1]*v[1,1]+expA[3,2]*v[2,1]+expA[3,3]*v[3,1]]);
            of 2 do return(array(1..3, 
                           [expA[1,1]*v[1,1]+expA[1,2]*v[2,1]+expA[1,3]*v[3,1],
                            expA[2,1]*v[1,1]+expA[2,2]*v[2,1]+expA[2,3]*v[3,1],
                            expA[3,1]*v[1,1]+expA[3,2]*v[2,1]+expA[3,3]*v[3,1]]));
            otherwise return(array(1..3,1..p,[
                  [(expA[1,1]*v[1,j]+expA[1,2]*v[2,j]+expA[1,3]*v[3,j]) $ j=1..p],
                  [(expA[2,1]*v[1,j]+expA[2,2]*v[2,j]+expA[2,3]*v[3,j]) $ j=1..p],
                  [(expA[3,1]*v[1,j]+expA[3,2]*v[2,j]+expA[3,3]*v[3,j]) $ j=1..p] 
                                             ]));
            end_case;
       else return(expA);
       end_if;
  end_if;

  if expAv and iszero(max(_plus(specfunc::abs(v[i,j]) $ j=1..p) $ i=1..n)) /*  norm(v)=0 */
  then case return_type
       of 1 do return([float(0) $ p]);
       of 2 do return(array(1..n, [float(0) $ n]));
       otherwise return(array(1..n,1..p,[[float(0)$p]$n]));
       end_case;
  end_if;

  norm_A:= max(_plus(specfunc::abs(A[i,j]) $ j=1..n) $ i=1..n); /*  norm(A) */

  if iszero(norm_A) then
     if expAv
     then case return_type
            of 1 do return([v[i,1] $ i=1..n]); 
            of 2 do return(array(1..n, [v[i,1] $ i=1..n]));
            otherwise return(array(1..n,1..p,[[v[i,j] $ j=1..p] $ i=1..n]));
          end_case;
     else erg:= array(1..n,1..n,[[float(0)$n]$n]);
          (erg[i,i]:= float(1.0);) $ i=1..n;
          return(erg);
     end_if;
  end_if;

  DIGITS:= DIGITS+2; /*  increase working precision. There */
                     /*  can be further increases below    */

  /* --------------------------------------------------------------------------*/
  /*  Heuristics to estimate complexity of the two implemented Taylor methods. */
  /*  The number of terms in the Taylor expansion of exp(A) with               */
  /*  norm(A)=sollNorm is approximated by                                      */
  /*                                                                           */
  /*        (sollNorm)^i/i! <= exp(sollNorm)*10^(-DIGITS)                      */
  /*                                                                           */
  /*         -->   i >=  c(sollNorm)* DIGITS/ln(DIGITS)                        */
  /*                                                                           */
  /*  with  c(0.01)=1.15, c(1.0)=3.0, c(3.0)=4.6, c(100.0)=38.9  (we use       */
  /*  sollNorm=3.0 in Version 1 and sollNorm=0.01 in Version 2).               */
  /*  This formula works is exact for DIGITS=10 and works reasonably well for  */
  /*  DIGITS = 5 .. 1000. In the following we assume that A and v are dense.   */
  /*  The estimate will not be realistic for norm(A)<<sollNorm, because        */
  /*  c(sollNorm) should be replaced by c(norm(A)), which is unknown. Also for */
  /*  sparse A and v the estimate can be far from realistic.                   */
  /* --------------------------------------------------------------------------*/

  if method=0 then  /*  no method specified. Heuristics to choose between */
                    /*  Version 1 and Version 2 of TaylorExpansion        */

     /* ------------------------------------------------------------------------*/
     /*  complexity (Version 1 of TaylorExpansion)                              */
     /*    n^2                                  <- scale matrix                 */
     /*  + NN1*4.6*DIGITS/ln(1.0*DIGITS)*n^2*p; <- Taylor series v+A*v+A^2*v/2+ */
     /* ------------------------------------------------------------------------*/
     sollNorm:= 3.0;
     NN1:= max(1,norm_A/sollNorm);          /*  number of steps to undo scaling */

     /* ------------------------------------------------------------------------*/
     /*  complexity (Version 2 of Taylor Expansion)                             */
     /*     n^2                               <- scale matrix                   */
     /*   +(1.15*DIGITS/ln(1.0*DIGITS)-1)*n^3 <- Taylor series for scaled matrix*/
     /*   + NN2*n^3                           <- repeated squaring              */
     /*   + n^2*p;                            <- multiplication exp(A)*v        */
     /* ------------------------------------------------------------------------*/
     sollNorm:= 0.01;
     NN2:= max(0,ln(norm_A/sollNorm)/ln(2.0));  /*  steps in repeated squaring  */

     kk:= DIGITS/ln(1.0*DIGITS);
     userinfo(2, "estimating complexity of Taylor versions with/without repeated squaring:");
     userinfo(2, "  with/without = ".expr2text((NN2*n+p+(1.15*kk-1)*n)/(NN1*4.6*kk*p)));
     if NN1*4.6*kk*p < NN2*n + p + (1.15*kk-1)*n 
        then method:=1
        else method:=2;
     end_if;
  end_if;

  /* ---------------------------------------------------------------*/
  /*  Convert columns of A and v to polynomials. Important: get rid */
  /*  of float(0), otherwise A and v will always be dense!          */
  /* ---------------------------------------------------------------*/
  A:= subs(A, float(0)=0);
  x:=genident(): /*  unknown for polynomial representation */
  /*  APoly = list of polynomial columns */
  APoly:= [poly([[A[i,j],i] $ i=1..n],[x]) $ j=1..n];
  if expAv then  /*  convert array v to list of polynomials */
     v:= subs(v, float(0)=0);
     vPoly:=[poly([[v[i,j],i] $ i=1..n],[x]) $ j=1..p];
  end_if;

  /*  first version of MatrixMult only useful if B is sparse */
//MatrixMult:= proc(A,B)   /*  A and B = list of polynomials */
//local j,k,L,LL;  /*  The polynomials represent the columns */
//begin (L := [degree(nthterm(B[j],k)) $ k=1..nterms(B[j])];
//       LL:= [coeff(B[j])];
//       B[j]:= _plus(mapcoeffs(A[L[k]],_mult,LL[k]) $ k=1..nops(L));
//      ) $ j=1..nops(B); B;
//end_proc;

  /*  next version of MatrixMult is 50% faster, if B is dense */
  MatrixMult:= proc(A,B)    /*  A and B = list of polynomials */
  local j,k;        /*  The polynomials represent the columns */
  begin (B[j]:= _plus(mapcoeffs(A[k],_mult,coeff(B[j],k)) $ k=1..n);
        ) $ j=1..nops(B); B;
  end_proc;

  /* ---- now the real work starts ---- */
  /* -----------------------------------------------------------------*/
  /*  method1: Berechnung von exp(A)*v basierend auf der Taylor-Reihe */
  /*               exp(A)*v = v + A*v + A^2*v/2! + ...                */
  /* -----------------------------------------------------------------*/
  if method=1 then
     userinfo(1, "using Taylor expansion without repeated squaring");

     /*  put v= unit-matrix, if no v is given. This */
     /*  allows to compute exp(A) with this version */
     if not expAv then p:=n; vPoly:=[poly([[1.0,j]],[x]) $ j=1..n ]; end_if;

     /* -------------------------------------------------------------------*/
     /*  scale matrix down to norm(A)<=sollNorm (=small). Large sollNorm   */
     /*  decreases the costs, but makes the summation of the Taylor series */
     /*  numerically unstable, if the matrix has large eigenvalues which   */
     /*  do not lie on the positive real axis. sollNorm:=3.0 seems to be a */
     /*  reasonable compromise between speed and stability.                */
     /* -------------------------------------------------------------------*/
     sollNorm:=3.0;
     if norm_A>sollNorm
        then NN:=ceil(norm_A/sollNorm);
             /*  scale A -> A/NN, i.e. norm((new)A)<= sollNorm */
             APoly:=map(APoly,mapcoeffs,_mult,float(1/NN));
             norm_A:= norm_A/NN;
        else NN:=1; 
     end_if;

     tol:= float(10^(-DIGITS));
     tol:= exp(norm_A)*tol;
     /*  How many terms of the Taylor series do we need   */
     /*  to achieve a relative precision <= 10^(-DIGITS)? */
     kk:=0; repeat kk:=kk+1; tol:=tol*kk until norm_A^kk<=tol end_repeat;

     DIGITS:= DIGITS + max(0,ceil(ln(float(kk*(0.01+NN)))/ln(10.0)));

     /*  Compute (1+A+A^2/2+..)*(1+A+A^2/2+..)*...*(1+A+A^2/2+..)*v */
     ergPoly:= vPoly;
     (Apower_v:= ergPoly;
      (/*  compute next Apower_v = A^i*v/i! and add to ergPoly */
       Apower_v:=map(MatrixMult(APoly,Apower_v),mapcoeffs,_mult,1/i);
       ergPoly:= zip(ergPoly,Apower_v,_plus);
       ) $ i=1..kk;
      ) $ j=1..NN;
     delete MatrixMult;

     case return_type 
       of 1 do return([coeff(ergPoly[1],i) $ i=1..n]);
       of 2 do return(array(1..n,[coeff(ergPoly[1],i) $ i=1..n]));
       otherwise
         return(array(1..n,1..p,[[coeff(ergPoly[j],i) $ j=1..p] $i=1..n]));
     end_case;
  end_if;

  /* ------------------------------------------------------------------------*/
  /*  method 2: Berechnung von exp(A) basierend auf der Taylor-Reihe von     */
  /*  exp(A/2^NN) mit 2^NN=ceil(norm(A)/sollNorm) und repeated squaring      */
  /*  dieser Matrix. There is no need for stabilizing measures in computing  */
  /*  exp(A/2^NN), since norm(A/2^NN)<= small sollNorm                     . */
  /* ------------------------------------------------------------------------*/
  if method=2 then
     userinfo(1, "using Taylor expansion with repeated squaring");

     tol:=float(10^(-DIGITS));
     sollNorm:=0.01;
     if norm_A>sollNorm
        then NN:=ceil(ln(norm_A/sollNorm)/ln(2.0));
             /*  scale A -> A/2^NN, i.e. norm((new)A)<= sollNorm */
             APoly:=map(APoly,mapcoeffs,_mult,float(1/2^NN));
             norm_A:= norm_A/2^NN;
        else NN:=0; 
     end_if;

     /*  How many terms of the Taylor series do we need */
     /*  to achieve a relative precision <= tol?        */
     tol:= exp(norm_A)*tol;
     kk:=0; repeat kk:=kk+1; tol:=tol*kk until norm_A^kk<=tol end_repeat;

     DIGITS:= DIGITS + max(0,ceil(ln(float(kk*(0.01+NN)))/ln(10.0)));
     /*  start to compute Taylor series ergPoly = 1 + A + A^2/2! + ... */
     Apower:= APoly;
     ergPoly:= [poly([[1.0,j]],[x])+Apower[j] $ j=1..n];  /*  1 + A */
     (/*  compute next Apower = A^k/k! and add to ergPoly */
       Apower:= map(MatrixMult(APoly,Apower),mapcoeffs,_mult,1/k);
       ergPoly:= zip(ergPoly,Apower,_plus);
     ) $ k=2..kk;

     /*  repeated squaring on ergPoly=exp(A/2^NN) to compute (ergPoly)^(2^NN) */
     (ergPoly:= MatrixMult(ergPoly,ergPoly);) $ i=1..NN;

     if expAv 
     then ergPoly_v:=MatrixMult(ergPoly,vPoly);
          delete MatrixMult;
          case return_type
          of 1 do return([coeff(ergPoly_v[1],i)$ i=1..n]);
          of 2 do return(array(1..n,[coeff(ergPoly_v[1],i) $ i=1..n]));
          otherwise  
          return(array(1..n,1..p,[[coeff(ergPoly_v[j],i)$ j=1..p] $ i=1..n]));
          end_case;
     else delete MatrixMult;
          return(array(1..n,1..n,[[coeff(ergPoly[j],i)$ j=1..n] $ i=1..n]));
     end_if;
  end_if;

  /*  ---------------------------------------------------------------*/
  /*  method 3: Berechnung von exp(A)*v basierend auf Krylovraeumen. */
  /*  ---------------------------------------------------------------*/
  if method=3 then
    userinfo(1, "using Krylov method");

    if p<>1 then 
       error("Krylov method only implemented for exp(Matrix)*Vector");
    end_if;

    if map({op(v)}, domtype) minus {DOM_FLOAT,DOM_COMPLEX}<> {}
     then error("non-numerical entry in vector")
    end_if;

    /*  ----------------------------------------------------------------*/
    /*  For an n x n input matrix A the Arnoldi-algorithm computes an   */
    /*  upper Hessenberg matrix Hmm of dimension m x m and approximates */
    /*  exp(A)*v by                                                     */
    /*     exp(A)*v = norm(v)* V_m * exp(Hmm) * e_1                     */
    /*  with the orthonormal n x m matrix V_m spanned by the Krylov     */
    /*  vectors v,A*v,A^2v,..,A^(m-1)*v and the first unit m-vector     */
    /*  e1:=[1,0$m-1].                                                  */
    /*  In principle one can try m = 1,2,3,...,n until the approximation*/
    /*  is good enough. This is expensive, if the process does not stop */
    /*  with m<<n. Therefore we do not use all m in [1,2,3,..,n],  but  */
    /*  start with m=1 and then increase m -> sqrt(2)*m. With costs     */
    /*  O(m^3) this ensures that unsuccesful attempts do not increase   */
    /*  the total costs by more than a factor of 2, if m runs to n.     */
    /*  ----------------------------------------------------------------*/

    tol:= float(10^(-DIGITS));
    vNorm:=float(specfunc::sqrt(_plus(specfunc::abs(v[i,1])^2 $ i=1..n))):
    KrylovBase[1]:=mapcoeffs(vPoly[1],_mult,1/vNorm); /*  vv = v/norm(v) */

    /* --------------------------------------------------------*/
    /*  compute Krylov basis v[j]=KrylovBase[j] and upper mxm  */
    /*  Hessenberg matrix H by the Arnoldi algorithm:          */
    /*                v[1] := v/norm(v);                       */
    /*                for j=1..m do                            */
    /*                    w = A*v[j];                          */
    /*                    for k=1..j do                        */
    /*                        H[k,j]:= <w,v[k]>;               */
    /*                        w:= w - H[k,j]*v[k];             */
    /*                    end_for;                             */
    /*                    if norm(w)<eps then Abbruch; end_if; */
    /*                    H[j+1,j]:= norm(w);                  */
    /*                    v[j+1]:= w/H[j+1,j];                 */
    /*                end_for;                                 */
    /*  Result: with orthonormal KrylovBase V=[v[1],..,v[m]]   */
    /*                                                         */
    /*   A*V = V*H + H[m+1,m]*v[m+1]*e_m^T,  H=V^T*A*V         */
    /*                                                         */
    /*      exp(A)*v =(approx)= norm(v)*V*exp(H)*e1            */
    /* --------------------------------------------------------*/

    /*  H = storage for Hessenberg matrix used in arnoldi. Its upper left m+1 x m */
    /*  part is used in arnoldi(..,m). Row m+1 is used for estimating the error.  */
    H:= array(1..n+1,1..n,[[0$n]$n+1]); 

    /* subroutine arnoldi: transforms matrix A (=APoly) to upper MxM Hessenberg    */
    /* matrix H and computes orthonormal nxM transformation matrix V = KrylovBase: */

    arnoldi:=proc(last_m,M) 
    local i, j, k, wPoly, wNorm;
    begin
      for j from last_m+1 to M do
        wPoly:= MatrixMult(APoly,[KrylovBase[j]])[1];
        wNorm:= float(specfunc::sqrt(_plus(specfunc::abs(coeff(wPoly,k))^2$ k=1..n)));
        (H[k,j]:= /* scalar_product(wPoly,KrylovBase[k])*/
            _plus(_mult(conjugate(coeff(KrylovBase[k],i)),coeff(wPoly,i)) $ i=1..n);
         wPoly:= (wPoly-mapcoeffs(KrylovBase[k],_mult,H[k,j])); /* w:=w-vv*H[k,j] */
        ) $ k=1..j;
        /* H[j+1,j] = norm(w) */
        H[j+1,j]:=float(specfunc::sqrt(_plus(specfunc::abs(coeff(wPoly,k))^2$ k=1..n)));
        /* if new norm(w) < norm(old w)*tol, then new w is */
        /* dominated by roundoff and it is time to stop!   */
        if specfunc::abs(H[j+1,j])<=wNorm*tol then M:=j; m:=j; break; end_if; /* Abbruch */
        KrylovBase[j+1]:= mapcoeffs(wPoly,_mult,1/H[j+1,j]);
      end_for;
      /* return upper M x M part of Hessenberg matrix H */
      array(1..M,1..M,[[H[i,k] $ k=1..M] $ i=1..M]);
    end_proc;

    last_m:=0; m:=1;

    while m<=n do 
      /* Compute the Krylov base vectors spanned by A^(last_m+1)*v,..,A^m*v    */
      /* and the corresponding upper Hessenberg matrix Hmm:= arnoldi(last_m,m).*/
      /* Then compute exp(Hmm)*e1 as a list: */
      Hmm:=arnoldi(last_m,m); 
        /* Warning: arnoldi can change m as a side effect. Need to call arnoldi */
        /* first, before passing it to expMatrix. Otherwise vector [1,0$m-1]    */
        /* might have wrong dimension.*/
      userinfo(1, "calling expMatrix with ".expr2text(m)." x ".
                                            expr2text(m). " Hessenberg matrix");
      expHe1:= numeric::expMatrix(Hmm,[1,0$m-1]);
      /* Compute exp(A)*v:=Krylovbase*e^Hmm*e_1*Norm(v)*/
      ergPoly:= _plus(mapcoeffs(KrylovBase[j],_mult,expHe1[j]) $ j=1..m);
      /* Bestimme die Norm des Ergebnisses.                                       */
      norm_Erg:=float(specfunc::sqrt(_plus(specfunc::abs(coeff(ergPoly,i))^2 $ i=1..n)));
      if m<n then
         /* estimate error norm(exp(A)*v - KrylovBase*exp(Hmm)*e1) */
         Err:= specfunc::abs(expHe1[m])*H[m+1,m];
         userinfo(2, "estimate of absolute error: ".expr2text(Err*vNorm));
         if Err<=norm_Erg*tol then break end_if; /* Abbruchkriterium */
      end_if;
      if m=n then break end_if; /* ultimatives Abbruchkriterium */
      last_m:=m;
      /* increase m by a factor of sqrt(2) */
      m:= min(ceil(1.41*m),n);
    end_while;
    delete MatrixMult, arnoldi;

    userinfo(1, expr2text(n)." x ".expr2text(n)." matrix reduced to ".
                expr2text(m)." x ".expr2text(m). " Hessenberg matrix");
    ergPoly:= mapcoeffs(ergPoly,_mult,vNorm);
    case return_type
      of 1 do return(           [coeff(ergPoly,i) $ i=1..n]);
      of 2 do return(array(1..n,[coeff(ergPoly,i) $ i=1..n]));
      otherwise return(array(1..n,1..1,[[coeff(ergPoly,i)]$ i=1..n]));
    end_case;
  end_if;

  /*-------------------------------------------------------------------*/
  /* method 4: compute exp(A), exp(A)*v by polynomial interpolation.   */
  /* Idea:  if    p(lambda) = exp(lambda) for all eigenvalues          */
  /*        and  p'(lambda) = exp(lambda) for all double eigenvalues   */
  /*        and p''(lambda) = exp(lambda) for all triple eigenvalues   */
  /*        etc.                                                       */
  /*        then p(A) = exp(A).                                        */
  /*        Hence, compute interpolating polynomial p through          */
  /*        [lambda1,exp(lambda1)], [lambda2,exp(lambda2)], ...        */
  /*-------------------------------------------------------------------*/
  if method=4 then
      userinfo(1, "interpolating exp(x) by a polynomial:");

      tol:= float(10^(-DIGITS));
      eigenvals:= map(numeric::eigenvalues(A),float):
      DIGITS:= 2*DIGITS+2+round(n/10);
      absolute_tol:=max(op(map(eigenvals,specfunc::abs)))*tol;
      (if specfunc::abs(eigenvals[j])<=absolute_tol then
          eigenvals[j]:=float(0) end_if; ) $ j=1..n;
      userinfo(2, "interpolation nodes = eigenvalues = ".expr2text(eigenvals));
      /* compute coefficients c[i] of interpolating polynomial */
      /*  pp(A) = c[0]+c[1]*A+..+c[m]*A^m                      */
      m:=n-1;
      /* converting list to array(0..m) */
      eigenvals:= array(0..m,eigenvals);
      c:=map(eigenvals,exp);
      (for j from m downto k do
          if specfunc::abs(eigenvals[j]-eigenvals[j-k])<=absolute_tol
             then c[j]:=c[j]/k; /*Vorsicht!*/
             else c[j]:=(c[j]-c[j-1])/(eigenvals[j]-eigenvals[j-k]);
          end_if;
       end_for:) $ k=1..m;
      (for j from m-1 downto k-1 do
          c[j]:=c[j]-eigenvals[j+1-k]*c[j+1] end_for: ) $ k=1..m;
      userinfo(2, "coefficients of the interpolating polynomial: ");
      userinfo(2, expr2text(op(c)));

      /* heuristic check, whether interpolation polynomial is ok */
      /* within numerical precision. We should have              */
      /*              pp(eigenval)=exp(eigenval)                 */
      /* Evaluation via Horner scheme:                           */
      (pp:= c[m]: (pp:= eigenvals[i]*pp+c[m-j];) $ j=1..m;
       if specfunc::abs(pp-exp(eigenvals[i]))>absolute_tol
          then error("interpolation is numerically unstable");
      end_if; ) $ i=0..m;

      /* Horner schemes:                                     */
      /* evaluate (c[0]+c[1]*A+c[2]*A^2+..+c[m]*A^m)*v by    */
      /*     t:= c[m]*v;                                     */
      /*     for i from m downto 0 do t:=A*t+c[i]*v end_for; */
      /* evaluate c[0]+c[1]*A+c[2]*A^2+..+c[m]*A^m by        */
      /*     t:= c[m]*A;                                     */
      /*     for i from m downto 1 do t:=A*t+c[i]*A end_for; */
      /*     t:= t+ c[0];                                    */
      if expAv 
         then  tPoly:= map(vPoly,mapcoeffs,_mult,c[m]);
              (tPoly:= zip(MatrixMult(APoly,tPoly),
                          map(vPoly,mapcoeffs,_mult,c[m-i]),_plus);
               ) $ i=1..m;
         else tPoly:= map(APoly,mapcoeffs,_mult,c[m]);
              (tPoly:= zip(MatrixMult(APoly, tPoly),
                          map(APoly,mapcoeffs,_mult,c[m-i]),_plus)
              ) $ i=1..m-1;
              (tPoly[k]:= tPoly[k] + poly([[c[0],k]],[x]);) $ k=1..n;
      end_if;
      delete MatrixMult;

    /* If matrix A is real, then exp(A) is real, too. Get rid   */
    /* of imaginary numerical junk by applying Re to the result */
    if has(map({coeff(APoly[j]) $ j=1..n},domtype),DOM_COMPLEX)
       then A_is_real:= FALSE else A_is_real:= TRUE;
    end_if;
    if expAv and has(map({coeff(vPoly[j]) $ j=1..p},domtype),DOM_COMPLEX)
       then v_is_real:= FALSE else v_is_real:= TRUE;
    end_if;

    if expAv then 
       case return_type
       of 1 do [coeff(tPoly[1],i) $ i=1..n];
               if A_is_real and v_is_real 
                  then return(map(%,Re));
                  else return(%);
               end_if;
       of 2 do array(1..n,[coeff(tPoly[1],i) $  i=1..n]);
               if A_is_real and v_is_real 
                  then return(map(%,Re));
                  else return(%);
               end_if;
       otherwise array(1..n,1..p,[[coeff(tPoly[j],i) $ j=1..p] $ i=1..n]);
                 if A_is_real and v_is_real 
                    then return(map(%,Re));
                    else return(%);
                 end_if;
       end_case;
    end_if;
    array(1..n,1..n,[[coeff(tPoly[j],i) $ j=1..n] $ i=1..n]);
    if A_is_real then return(map(%,Re)) else return(%); end_if;
  end_if;

  /*------------------------------------------------------------*/
  /* method 5: compute exp(A), exp(A)*v  by diagonalization     */
  /*          A  = X * diag(l1,l2,..)           * X^(-1)        */
  /*       exp(A)= X * diag(exp(l1),exp(l2),..) * X^(-1)        */
  /* Warning: if no diagonalization exists, then X will not be  */
  /*          invertible and one gets an error when computing   */
  /*          X^(-1) or X^(-1) will be wrong because of         */
  /*          numerical instability;                            */
  /* This method is fail safe for hermitean or skew-hermitean A */
  /*------------------------------------------------------------*/
  if method=5 then
    userinfo(1, "diagonalizing matrix");
    delete MatrixMult;
    DIGITS:= DIGITS + round(ln(n)/ln(10.0));
    /* If matrix A is real, then exp(A) is real, too. Get rid   */
    /* of imaginary numerical junk by applying Re to the result */
    if has(map({coeff(APoly[j]) $ j=1..n},domtype),DOM_COMPLEX)
       then A_is_real:= FALSE else A_is_real:= TRUE;
    end_if;
    if expAv and has(map({coeff(vPoly[j]) $ j=1..p},domtype),DOM_COMPLEX)
       then v_is_real:= FALSE else v_is_real:= TRUE;
    end_if;
    ([d,X,dummy]):= numeric::eigenvectors(A,NoErrors);
    d:=map(d,float@exp):

    if expAv 
       then A:=numeric::matlinsolve(X,v)[1];
            ((A[i,j]:=d[i]*A[i,j]) $ i=1..n) $ j=1..p ;
            A:= array(1..n,1..p,[[_plus(X[i,k]*A[k,j]$ k =1..n) $ j=1..p] $ i=1..n]);
            if A_is_real and v_is_real then A:= map(A,Re) end_if;
            case return_type
               of 1 do return([A[i,1] $ i=1..n]);
               of 2 do return(array(1..n,[A[i,1] $ i=1..n]))
               otherwise return(A);
            end_case;
       else A:=numeric::inverse(X):
            ((A[i,j]:=d[i]*A[i,j]) $ j=1..n) $ i=1..n;
            A:= array(1..n,1..n,[[_plus(X[i,k]*A[k,j]$ k=1..n) $ j=1..n] $ i=1..n]);
            if A_is_real then return(map(A,Re)) else return(A); end_if;
    end_if;
  end_if;

end_proc:
