/* W. Oevel, 17.2.98  */
/*-------------------------------------------------------------------------------*/
/* numeric::fMatrix - functional calculus for matrices                           */
/*                                                                               */
/* Aufruf:                                                                       */
/*                                                                               */
/* 1) numeric::fMatrix(f, A, param1, param2, ..)                                 */
/*                                                                               */
/* Parameter:                                                                    */
/*   f                  : procedure, f(a, param1, param2, ...) with a in C       */
/*                                   must produce scalar values                  */
/*   A                  : diagonalizable numerical square matrix                 */
/*                        (array or Cat::Matrix)                                 */
/*   param1, param2, .. : numerical or symbolical parameters                     */
/*                                                                               */
/* Rueckgabewert:                                                                */
/*   f(A, param1, param2,..) = array                                             */
/*                                                                               */
/* Beschreibung:                                                                 */
/*                                                                               */
/*   Die Matrix A wird diagonalisiert: A = X diag(lambda1,lambda2,..) X^(-1).    */
/*   Das Ergebnis wird per                                                       */
/*   f(A,p1,p2,..) = X diag(f(lambda1,p1,p2,..),f(lambda2,p1,p2,..),..) X^(-1)   */
/*   berechnet.                                                                  */
/*   Achtung: dies funktioniert nicht fuer beliebiges A! Nichtdiagonalisierbare  */
/*   Matrizen fuehren zu einer Fehlermeldung, bzw. es kommt zu einer grossen     */
/*   Verstaerkung von Rundungsfehlern, wenn die Nichtdiagonalisierbarkeit        */
/*   numerisch nicht ermittelt werden kann.                                      */
/*   Diese Methode ist aber FAIL safe fuer hermitesche oder schief-hermitesche   */
/*   Matrizen A.                                                                 */
/*                                                                               */
/* Examples:                                                                     */
/*  >> A:= array(1..2,1..2,[[0,1],[-1,0]]):                                      */
/*  >> B:= numeric::fMatrix(sqrt,A);                                             */
/*                                                                               */
/*                   +-                             -+                           */
/*                   |   0.7071067811, 0.7071067811  |                           */
/*                   |                               |                           */
/*                   |  -0.7071067812, 0.7071067811  |                           */
/*                   +-                             -+                           */
/*                                                                               */
/*  >> numeric::fMatrix(_power, B, 2);                                           */
/*                                                                               */
/*    +-                                                                  -+     */
/*    |  3.833233541e-20 - 2.439454887e-19 I, 10.0e-1 - 5.421010862e-20 I  |     */
/*    |                                                                    |     */
/*    |     - 10.0e-1 + 2.710505431e-20 I,          4.065758146e-20        |     */
/*    +-                                                                  -+     */
/*                                                                               */
/*  >> numeric::fMatrix(exp@_mult, A, t);                                        */
/*                                                                               */
/*    +-                                                                    -+   */
/*    |    0.5 exp(-I t) + 0.5 exp(I t),   0.5 I exp(-I t) - 0.5 I exp(I t)  |   */
/*    |                                                                      |   */
/*    |  0.5 I exp(I t) - 0.5 I exp(-I t),   0.5 exp(-I t) + 0.5 exp(I t)    |   */
/*    +-                                                                    -+   */
/*                                                                               */
/*-------------------------------------------------------------------------------*/

numeric::fMatrix:=proc(f,A)
local n, d, X, dummy, i, j, k;
save DIGITS;
begin
  if args(0)<2 then error("not enough arguments") end_if;
  /*  convert A to DOM_ARRAY */
  if A::dom::hasProp(Cat::Matrix)=TRUE then A:= expr(A) end_if;
  if domtype(A)<>DOM_ARRAY then
     error("matrix must be an array or of category 'Cat::Matrix'");
  end_if;
  n:= op(A,[0,2,2]); /*  n ist die Dimension der Eingabematrix */
  if n<>op(A,[0,3,2]) then error("not a square matrix") end_if;
  A:= map(A,float);
  if map({op(A)}, domtype) minus {DOM_FLOAT,DOM_COMPLEX}<>{}
     then error("non-numerical entry in matrix")
  end_if;
  if n=1 then 
     return(array(1..1,1..1,[[f(A[1,1], args(3..args(0)))]]));
  end_if;
  DIGITS:= DIGITS + 2+ round(ln(n)/ln(10.0));
  userinfo(1, "setting working precision to DIGITS = ".expr2text(DIGITS));
  userinfo(1, "diagonalizing matrix");
  ([d,X,dummy]):= numeric::eigenvectors(A,NoErrors);
  userinfo(1, "eigenvalues are ".expr2text(d));
  userinfo(1, "applying function to eigenvalues:");
  d:=map(d, f, args(3..args(0))):
  userinfo(1, expr2text(d));
  userinfo(1, "inverting eigenvector matrix");
  A:=numeric::inverse(X):
  ((A[i,j]:=d[i]*A[i,j]) $ j=1..n) $ i=1..n;
  userinfo(1, "building result");
  subs(array(1..n,1..n,[[_plus(X[i,k]*A[k,j]$k=1..n)$j=1..n]$i=1..n]),
       float(0)=0);
end_proc:
