
//------   file  factorCh.mu   29.9.97  ---------------------
/*  Help file:

 factorCholesky - numerical Cholesky factorization of a matrix

 Call:

 numeric::factorCholesky( A <,Symmetric> <, Symbolic> <, NoCheck> )
                                                                   
 Parameter:

  A  --  a square real or complex matrix of domtype DOM_ARRAY
         or Cat::Matrix
                                                                   
Synopsis:

  numeric::factorCholesky(A); returns the lower triangular
  Cholesky factor L of the factorization A=LL^H, where L^H 
  denotes the Hermitean transpose of L. The result is of 
  domtype DOM_ARRAY. It is normalized such that the diagonal
  of L is real and positive.

  numeric::factorCholesky(A, Symmetric); returns the lower triangular
  Cholesky factor L of the factorization A=LL^T, where L^T 
  denotes the usual transpose of L. The result is of 
  domtype DOM_ARRAY. 

                                                                  
  Non-float numerical entries are admissible, they will be 
  evaluated to floats. Conversion to floats may be suppressed
  by using the option Symbolic.
 
  If the matrix is not Hermitean or not positive definite 
  and the option <Symmetric> is not used, then
  the computation will terminate with an error message.
  Checking of these properties can be suppressed with the option
  NoCheck. This option should be used, whenever the matrix contains 
  indeterminates. If numeric symbolic entries such as PI, sqrt(2) etc.
  are present, then these properties are checked numerically.

  Examples: 

>> A := array( 1..2 , 1..2 , [ [1,2*I],[-2*I,5] ] ):

>> numeric::factorCholesky(A);                                

                            +-              -+
                            |    1.0,    0   |
                            |                |
                            |  - 2.0 I, 1.0  |
                            +-              -+

>> numeric::factorCholesky(A,Symbolic);

                              +-          -+
                              |    1,   0  |
                              |            |
                              |  - 2 I, 1  |
                              +-          -+

>> A := array( 1..2 , 1..2 , [ [-1 ,1],[1,1] ] ):

>> numeric::factorCholesky(A);                                

Error: matrix is not positive definite within working precision [nume\
ric::factorCholesky]

>> numeric::factorCholesky(A, Symmetric);                                

                     +-                        -+
                     |   I,          0          |
                     |                          |
                     |  - I, 1.414213562 + 0 I  |
                     +-                        -+

See also: linalg::factorCholesky, numeric::factorLU , numeric::factorQR 
                                                                  */
//-----------------------------------------------------------------
numeric::factorCholesky:=proc(A)
local n,real,i,j,k,B,symbolic_mode,symmetricDecomp,warning_given,opt,do_check;
begin
  if args(0)<1 then error("expecting at least one argument") end_if;
  // convert to domtype DOM_ARRAY for speed 
  if A::dom::hasProp(Cat::Matrix)=TRUE then A:= expr(A) end_if; 
  if domtype(A) <> DOM_ARRAY then
     error("matrix must be an array or of category 'Cat::Matrix'");
  end_if:
  if op(A,[0,1])<>2 then error("not a matrix");end_if;
  n:=op(A,[0,2,2]):
  if n <> op(A,[0,3,2]) then error("not a square matrix"); end_if;

  do_check:=TRUE; 
  symbolic_mode:=FALSE; 
  symmetricDecomp:= FALSE;
  if args(0)>1 then
     for opt in args(2..args(0)) do
         case opt
         of Symbolic  do symbolic_mode  := TRUE;  break;
         of NoCheck   do do_check       := FALSE; break;
         of Symmetric do symmetricDecomp:= TRUE;  break;
         otherwise error("wrong option");
         end_case;
     end_for;
  end_if;
  if symbolic_mode and do_check and
       (  (indets(A)<>{}) or has({op(map(A,domtype))},DOM_NIL) )
  then do_check:=FALSE; 
       warning("indeterminate found, switching to option NoCheck");
  end_if;

  if not symbolic_mode then
    B:=map(A,float);
    if ({op(map(B,domtype))} minus {DOM_FLOAT,DOM_COMPLEX}) <> {}
      then symbolic_mode:=TRUE;
           do_check:=FALSE;
           warning("symbolic entry found, switching to options Symbolic, NoCheck");
      else A:=B;
    end_if:
  end_if;
  delete B;
  A:=subs(A,float(0)=0);

if symmetricDecomp
then 
     for i from 1 to n do
       for j from 1 to i-1 do
          if A[i,j]<>A[j,i] then error("matrix is not symmetric") end_if;
       end_for;
     end_for;
     for i from 1 to n do
       (A[i,j]:=(A[i,j]-_plus(A[i,k]*A[j,k] $ k=1..j-1))/A[j,j]) $ j=1..i-1;
        A[i,i]:=A[i,i]-_plus(A[i,k]^2 $ k=1..i-1);
        A[i,i]:=(A[i,i])^(1/2); (A[i,j]:=0) $ j=i+1..n;
        if i<n and iszero(A[i,i]) then 
            error("regular symmetric factorization does not exist") 
        end_if;
     end_for;
else 
     real:=TRUE;
     if do_check then
       for i from 1 to n do
         for j from 1 to i-1 do
           if iszero(Im(A[i,j]))
              then if A[i,j]<>A[j,i] then error("matrix is not hermitean") end_if;
              else if A[i,j]<>conjugate(A[j,i]) then error("matrix is not hermitean") end_if:
                   real:= FALSE;
           end_if;
         end_for;
         if not iszero(Im(A[i,i]))then error("matrix is not hermitean"); end_if;
       end_for;
     end_if;
     warning_given:= FALSE;

     for i from 1 to n do
       if real
       then (A[i,j]:=(A[i,j]-_plus(A[i,k]*A[j,k] $ k=1..j-1))/A[j,j])
                          $ j=1..i-1;
       else (A[i,j]:=(A[i,j]-_plus(A[i,k]*conjugate(A[j,k]) $ k=1..j-1)
                     )/A[j,j]) $ j=1..i-1;
       end_if;
       if symbolic_mode
         then A[i,i]:=A[i,i]-_plus(abs(A[i,k])^2 $ k=1..i-1);
         else A[i,i]:=A[i,i]-_plus(specfunc::abs(A[i,k])^2 $ k=1..i-1);
       end_if;
       if do_check then
         if traperror(bool(A[i,i]<=0))<>0 
         then if not warning_given then
                warning("positiveness checked numerically");
                warning_given:=TRUE;
              end_if;
              if float(A[i,i])<=0 then
                 error("matrix is not positive definite within working precision");
              end_if;
         else if A[i,i]<=0 then
                 if symbolic_mode
                  then error("matrix is not positive definite");
                  else error("matrix is not positive definite within working precision");
                 end_if;
              end_if;
         end_if;
       end_if;
       A[i,i]:=(A[i,i])^(1/2); (A[i,j]:=0) $ j=i+1..n;
     end_for;
end_if;
  subs(A,float(0)=0);
end_proc:

