
//------   file  factorLU.mu  ---------------------
/*  Help file:

 factorLU - LU factorization of a matrix

 Call:

 numeric::factorLU( A <, Symbolic> )
                                                                   
 Parameter:

  A  --  an m x n matrix (DOM_ARRAY or category Cat::Matrix)
                                                                   
Synopsis:

  The LU factorization of a real/complex matrix is PA=LU, where
  L is lower triangular, normalized to 1 along the diagonal,
  and R is upper triangular.  P is a permutation matrix 
  corresponding to row exchanges.

  numeric::factorLU(A) returns the list [L,U,p], where L and R are
  of domtype DOM_ARRAY. The list of integers p=[ p[1],p[2], .. ]
  represents the row exchanges in pivoting steps, i.e., B = PA = LU, 
  where B[i,j] = A[p[i],j]. Multiplication y=Px of the permutation
  matrix P with a vector x is realized by y[i]:=x[p[i]].
                                                                  
  Non-float numerical entries are allowed in the matrix, they
  will be evaluated to floats in the course of the computation.
  Non-numeric symbolic entries will lead to an error, unless
  the option Symbolic is specified.

  The option Symbolic prevents conversion of the matrix entries
  to floats. No row exchanges will be performed unless necessary.

  Examples: 

>> A:=array(1..3,1..3,[[1,2,3],[2,4,6],[4,8,9]]):
>> ([L,U,p]):=numeric::factorLU(A);

          -- +-            -+  +-                -+            --
          |  |    1,  0, 0  |  |  4.0, 8.0,  9.0  |             |
          |  |              |  |                  |             |
          |  |   0.5, 1, 0  |, |   0,   0,   1.5  |, [3, 2, 1]  |
          |  |              |  |                  |             |
          |  |  0.25, 0, 1  |  |   0,   0,  0.75  |             |
          -- +-            -+  +-                -+            --

>> A:= array(1..3,1..2,[[1,2],[2,3],[3,4]]):
>> ([L,U,p]):=numeric::factorLU(A);

 -- +-                      -+  +-                   -+            --
 |  |        1,       0,  0  |  |  3.0,      4.0      |             |
 |  |                        |  |                     |             |
 |  |  0.3333333333,  1,  0  |, |   0,  0.6666666666  |, [3, 1, 2]  |
 |  |                        |  |                     |             |
 |  |  0.6666666666, 0.5, 1  |  |   0,        0       |             |
 -- +-                      -+  +-                   -+            --

See also:  linalg::factorLU, numeric::factorQR, numeric::factorCholesky
                                                                  */
//-----------------------------------------------------------------
numeric::factorLU:=proc(A)
local m,n,method,B,i,j,k,tmp,pivot,pivindex,rownorm,L,p;
begin
   if args(0)<1 then error("expecting at least one argument") end_if;
   // convert to domtype DOM_ARRAY for speed 
   if A::dom::hasProp(Cat::Matrix)=TRUE then A:= expr(A) end_if; 
   if domtype(A) <> DOM_ARRAY then
      error("matrix must be an array or of category 'Cat::Matrix'");
   end_if:
   if op(A,[0,1])<>2 then error("argument is not a matrix");end_if;
   m:=op(A,[0,2,2]):  // number of rows
   n:=op(A,[0,3,2]):  // number of columns
   if args(0)>1 then if args(2)=Symbolic then method:=1;
                     else error("unknown option") end_if;
                else method:=2;
   end_if;

   if method=2 then            
      B:=map(A,float):
      if {op(map(B,domtype))} minus {DOM_FLOAT, DOM_COMPLEX} <> {}
          then warning("symbolic entry found, switching to option Symbolic");
               method:=1;
          else A:=B;
      end_if:
   end_if;
   A:=subs(A,float(0)=0):
   p:=[ i $ i=1..m ]: // initialize list to store row permutations 

if method=1 then
   userinfo(1,"symbolic mode, avoiding pivoting if possible ");
end_if;
if method=2 then            
   userinfo(1,"using column pivoting");
end_if;

   for j from 1 to min(m-1,n) do

     if method=1 then
        for pivindex from j to m do
           if not iszero(A[pivindex,j]) then break; end_if;
        end_for;
        if pivindex>m then next end_if; // no nontrivial pivot element 
     end_if;
     if method=2 then
        pivot:=0; pivindex:=j;
        for i from j to m do 
            if iszero(A[i,j]) then next end_if;
            if j=n
            then if (tmp:= specfunc::abs(A[i,j]))>pivot
                    then pivot:= tmp; pivindex:= i;
                 end_if;
            else rownorm:=max(specfunc::abs(A[i,k]) $ k=j+1..n );
                 if not iszero(rownorm)
                 then specfunc::abs(A[i,j])/rownorm;
                      if pivot<last(1) then pivot:=last(1); pivindex:=i; end_if;
                 else pivindex:= i; break;
                 end_if;
            end_if;
        end_for;
     end_if;

     if pivindex<>j then
       (tmp:=A[j,k]; A[j,k]:=A[pivindex,k];A[pivindex,k]:=tmp;) $ k=1..n;
       tmp:=p[j];p[j]:=p[pivindex];p[pivindex]:=tmp;
     end_if;
     if not iszero(A[j,j]) then
        for i from j+1 to m do
            A[i,j]:=A[i,j]/A[j,j]:
            if not iszero(A[i,j]) then
              (A[i,k]:=A[i,k]-A[i,j]*A[j,k]) $ k=j+1..n;
            end_if;
        end_for;
     end_if;
   end_for;

   A:=subs(A,float(0)=0);
   /* LU data are stored in lower and upper part of A. Extract L and U: */
   L:=array(1..m,1..m,[[0$m]$m]); (L[i,i]:=1) $ i=1..m;
   ((L[i,j]:=A[i,j]) $ i=j+1..m ) $ j=1..min(m-1,n);
   ((A[i,j]:=0) $ i=j+1..m ) $ j=1..n;
   return([L,A,p]):
end_proc:
