
//------   file  factorQR.mu  19.9.97         ---------------------
//  auf nichtquadratische Matrizen erweitert 8.2.99 W. Oevel  
//  option QDet eingefuegt: 9.7.99 W. Oevel 
/*  Help file:

 factorQR - QR factorization of a matrix

 Call:

 numeric::factorQR( A <, Symbolic>, <, QDet> )
                                                                   
 Parameter:

  A  --  a matrix (DOM_ARRAY or category Cat::Matrix)
                                                                   
Synopsis:

  The QR factorization of a real/complex m x n matrix is A=QR, where
  Q is a unitary/orthogonal m x m and R is an upper triangular 
  m x n matrix.

  numeric::factorQR(A); returns the list [Q,R] with Q and R of
  domtype DOM_ARRAY.
                                                                  
  Non-float numerical entries are allowed in the matrix, they
  will be evaluated to floats in the course of the computation.
  Non-numeric symbolic entries will lead to a warning. Householder
  transformations are used to compute the factorization. The
  R factor is normalized such that its diagonal entries are
  real and non-negative.

  The option Symbolic prevents conversion of the matrix entries
  to floats. In this case symbolic entries are allowed, Gram-Schmidt
  orthonormalization is used to compute the factorization.

-----    New !!!!  -----
  With the option QDet the output changes to [Q,R,d] with d = det(Q).
  Presently this option is only available for the numerical
  factorization (i.e., there is no effect when using option Symbolic).
-----------------------

  For invertible n x n matrices A the result without option Symbolic
  is the float version of the result with option Symbolic.
  For non-invertible or non-square A the factorization is not
  unique and option Symbolic may return a different result
  than the numerical factorization without option Symbolic.

  Examples: 

>> A:=array(1..3,1..3,[ [1,1,I] , [1,exp(1),I] , [1,2,I] ]):

>> ([Q,R]):=numeric::factorQR(A);

    -- +-                                              -+
    |  | 0.5773502691, -0.7424322034,  - 0.339795659  I |
    |  |                                                |
    |  | 0.5773502691,  0.6654877745,  - 0.4730673192 I |,
    |  |                                                |
    |  | 0.5773502691,  0.07694442889,   0.8128629783 I |
    -- +-                                              -+
           +-                                           -+ --
           | 1.732050807, 3.301451553, 1.732050807 I     |  |
           |                                             |  |
           |      0,      1.220439979, 1.394208422e-19 I |  |
           |                                             |  |
           |      0,           0,      6.38099974e-20    |  |
           +-                                           -+ --
   
>> Q*R;
                       +-                       -+
                       | 1.0,     1.0,     1.0 I |
                       |                         |
                       | 1.0, 2.718281828, 1.0 I |
                       |                         |
                       | 1.0,     2.0,     1.0 I |
                       +-                       -+


>> A:=array(1..3,1..3,[ [1,1,I] , [1,1,I] , [1,2,I] ]):

>> ([Q,R]):=numeric::factorQR(A,Symbolic);

   -- +-                         -+                              --
   |  |  1/2     1/2  1/2    1/2  |                               |
   |  | 3       2    3      2     |  +-                       -+  |
   |  | ----, - ---------,  ----  |  |            1/2          |  |
   |  |  3          6        2    |  |  1/2    4 3         1/2 |  |
   |  |                           |  | 3   ,   ------,  I 3    |  |
   |  |  1/2     1/2  1/2     1/2 |  |           3             |  |
   |  | 3       2    3       2    |  |                         |  |
   |  | ----, - ---------, - ---- |, |        1/2  1/2         |  |
   |  |  3          6         2   |  |       2    3            |  |
   |  |                           |  |   0,  ---------,    0   |  |
   |  |  1/2    1/2  1/2          |  |           3             |  |
   |  | 3      2    3             |  |                         |  |
   |  | ----,  ---------,     0   |  |   0,      0,        0   |  |
   |  |  3         3              |  +-                       -+  |
   -- +-                         -+                              --

>> Q*R;
                               +-         -+
                               |  1, 1, I  |
                               |           |
                               |  1, 1, I  |
                               |           |
                               |  1, 2, I  |
                               +-         -+

>> ([Q,R,d]):= numeric::factorQR(A, QDet);

-- +-                                                  -+
|  |  0.5773502691, -0.4082482904,   - 0.7071067811 I   |
|  |                                                    |
|  |  0.5773502691, -0.4082482904,    0.7071067811 I    |,
|  |                                                    |
|  |  0.5773502691,  0.8164965809, - 5.421010862e-20 I  |
-- +-                                                  -+

   +-                                              -+      --
   |  1.732050807,  2.309401076,   1.732050807 I    |       |
   |                                                |       |
   |       0,      0.8164965809, 7.666467083e-20 I  |, - I  |
   |                                                |       |
   |       0,            0,        1.32787105e-19   |       |
   +-                                              -+      --



See also:  linalg::factorQR, numeric::factorLU, numeric::factorCholesky
                                                                  */
//-----------------------------------------------------------------
numeric::factorQR:=proc(A)
local n,m,mn,method,opt,showQdet,Qdet,use_normal,B,rreal,i,j,k,l,d,h,s,ss,q,r,Q,R;
begin
   if args(0)<1 then error("expecting at least one argument") end_if;
   // convert to domtype DOM_ARRAY for speed 
   if A::dom::hasProp(Cat::Matrix)=TRUE then A:= expr(A) end_if; 
   if domtype(A) <> DOM_ARRAY then
      error("matrix must be an array or of category 'Cat::Matrix'");
   end_if:
   if op(A,[0,1])<>2 then error("argument is not a matrix");end_if;
   n:=op(A,[0,2,2]):
   m:=op(A,[0,3,2]):
   method:= 2;  // default is numeric mode = 2 
   showQdet:= FALSE;
   if args(0)>1 then 
     for opt in args(2..args(0)) do
         case opt 
         of Symbolic do method:=1; break;
         of QDet do showQdet:= TRUE; break;
         otherwise error("unknown option")
         end_case;
     end_for;
   end_if;

   use_normal:= FALSE;

   if method=1 then
    if (indets(A)<>{}) 
       or has({op(map(A,domtype))},DOM_NIL)
       or has({op(map(A,domtype))},DOM_EXPR)
      then use_normal:=TRUE end_if;
   end_if;

if method=2 then            
   B:=map(A,float):
   if {op(map(B,domtype))} minus {DOM_FLOAT, DOM_COMPLEX} <> {}
       then warning("symbolic entry found, switching to option Symbolic");
            method:=1; use_normal:= TRUE;
       else A:=B;
   end_if:
end_if;
   A:=subs(A,float(0)=0):

   if {op(map(A,(iszero@Im)))} = {TRUE}     // is matrix real ? 
       then rreal:=TRUE else rreal:=FALSE
   end_if:

if method=1 then
   userinfo(1,"using Gram-Schmidt orthonormalization");
   R:=array(1..n,1..m,[[0$m]$n]); // initialize upper factor 
   if m>=n                   // initialize orthogonal factor 
      then Q:=array(1..n,1..n,[[A[i,j] $j=1..n]$i=1..n]);
      else Q:=array(1..n,1..n,[[0$n]$n]):
           ((Q[i,j]:= A[i,j];) $ j=1..m) $ i=1..n;
   end_if;
   for j from 1 to n do
      if j<=m then
       (Q[i,j]:= Q[i,j]-_plus(Q[i,k]*R[k,j]/r[k] $ k=1..j-1 );) $ i=1..n;
       if use_normal then (Q[i,j]:= normal(Q[i,j]);) $ i=1..n; end_if;
      end_if;
      
      r[j]:=_plus(conjugate(Q[i,j])*Q[i,j] $ i=1..n );
      if use_normal then r[j]:= normal(r[j]); end_if;
      if iszero(r[j]) then       // insert an arbitrary column. Try the 
         for k from 1 to n do    // standard columns q=(0,..,1,..,0)^T  
            q:=[0$n]; q[k]:=1;   // until q is independent of the first 
            for i from 1 to n do // j-1 orthonormalized columns of Q    
               q[i]:=q[i]-_plus(Q[i,l]*conjugate(Q[k,l])/r[l] $ l=1..j-1);
               if use_normal then q[i]:= normal(q[i]); end_if;
            end_for;
            r[j]:=_plus( conjugate(q[i])*q[i] $ i=1..n );
            if use_normal then r[j]:= normal(r[j]); end_if;
            if not iszero(r[j]) then
               (Q[i,j]:=q[i]) $ i=1..n; break;
            end_if;
         end_for;
         if iszero(r[j]) then error("could not orthonormalize the columns") end_if;
         if j<=m then R[j,j]:=0; end_if;
      else 
         if j<=m then R[j,j]:= sqrt(r[j]) end_if;
      end_if;
      (R[j,k]:=_plus(conjugate(Q[i,j])*A[i,k] $ i=1..n );) $ k=j+1..m;
      if use_normal then
        (R[j,k]:=normal(R[j,k]);) $ k=j+1..m;
      end_if;
   end_for;
   r:=[sqrt(r[i]) $ i=1..n];
   for j from 1 to n do
       (Q[i,j]:=Q[i,j]/r[j]) $ i=1..n;
       (R[j,i]:=R[j,i]/r[j]) $ i=j+1..m;
   end_for;

   return([Q,R]);

end_if;

if method=2 then            
   userinfo(1,"using Householder transformations");
   mn:= min(m,n);

   d:=[0$n]:   // initialize list to store diagonal 
   h:=[0$mn]:  // initialize list to store factors  

   for j from 1 to mn do
        // Householder step from the left to obtain upper   
        // triangular form. Householder data are stored in  
        // lower triangular part of A.                      
        s:= _plus(specfunc::abs(A[i,j])^2 $ i=j+1..n);
        if iszero(s) then d[j]:=A[j,j];
        else ss:=s+specfunc::abs(A[j,j])^2;
             s:=specfunc::sqrt(ss):
             h[j] := 1 / (ss + specfunc::abs(A[j,j]) * s);
             if not iszero(A[j,j]) then s:=sign(A[j,j])*s;end_if:
             A[j,j]:=s+A[j,j]; d[j]:=-s;
             for i from j+1 to m do
                 if rreal then  s:=_plus(A[k,j] *A[k,i]$ k=j..n);
                 else s:=_plus(conjugate(A[k,j])*A[k,i]$ k=j..n);
                 end_if;
                 s:=s*h[j];
                 (A[k,i]:=A[k,i]-s*A[k,j]) $ k=j..n ;
             end_for;
        end_if;
   end_for;

   // construct Q from the Householder data stored in the lower part of A
   Q:=array(1..n,1..n,[[0$n]$n]); (Q[i,i]:=1) $ i=1..n;
   for j from mn downto 1 do
     if not iszero(h[j]) then
        for i from j+1 to n do
           if rreal then  s:=_plus(A[k,j]*Q[k,i] $ k=j..n) ;
           else s:=_plus(conjugate(A[k,j])*Q[k,i] $ k=j..n) ;
           end_if;
           s:=s*h[j];
           (Q[k,i]:=Q[k,i]-s*A[k,j]) $ k=j..n ;
        end_for;
       (Q[k,j]:=A[k,j]/d[j]) $ k=j..n;
       Q[j,j]:=Q[j,j]+1;
     end_if;
   end_for;
   if showQdet then Qdet:=(-1)^(mn-nops(select(h,iszero))); end_if;

   // construct R from the upper part of A and diagonal d 
   for j from 1 to mn do
     (A[i,j]:=0) $ i=j+1..n;
      A[j,j]:=specfunc::abs(d[j]);
      if not iszero(A[j,j]) then     
         s:=sign(d[j]);                 // normalize diagonal of 
         (A[j,i]:=A[j,i]/s) $ i=j+1..m; // R to be real positive 
         (Q[i,j]:=Q[i,j]*s) $ i=1..n;
         if s<>1 and showQdet then Qdet:= s*Qdet; end_if;
      end_if;
   end_for;

   Q:=subs(Q,float(0)=0); A:=subs(A,float(0)=0);

   if showQdet
      then return([Q,A,Qdet]);
      else return([Q,A]):
   end_if;

end_if;

end_proc:
