/*++ kamo, 06.2.94
  Nonrecursive version of the discrete Fast Fourier Transformation.
  Parameters:
        A - list containing n elements
        m - integer with n = 2^m
  The elements of the transformation are computed in-place.
  Complex arithmetic is assumed and w = exp(2*PI*i/j) is
  expressed in terms of sines and cosines.
  The discrete fourier transform is defined by
        A[j] := sum(a[k]*exp(-2*PI*i*j*k/N), k=0..N-1), for j=0..N-1
  so N sample points will be transformed into N points. 
  The FFT-algorithm needs O(n logn) time.
++*/
/*++ W. Oevel, 11.3.99
   argument m in fft(A,m) now obsolete (deduced from nops(A))
++*/
/*++ W. Oevel, 12.3.99
   Option Symbolic eingebaut.
   w=exp(I*..) statt w= cos(..)+I*sin(..)
++*/
/*++ W. Oevel, 23.3.00
   auf multi-dim FFT erweitert

   z.B.

   >>  numeric::fft([L[1],L[2],..,L[2^m]]); // 1-dim FFT via list

   >>  A:= array(1..n[1],1..n[2], .. ,1..n[d], [..data..]) 

   >>  numeric::fft(A); // d-dim FFT via array



++*/

numeric::fft := proc(A)
local Atype, n, m, fPI, l, le, le1, 
      u, w, j, i, ip, t, nv2, nm1, k,
      M, N, p, r, kk, ln2, sMi, sMj, sMip,
      shift, start, d, conv;
begin
    if args(0)<1 then error("expecting at least one argument"); end_if;
    Atype:= domtype(A);
    if Atype<>DOM_LIST and
       Atype<>DOM_ARRAY then
          error("expecting a list or an array") 
    end_if;
    if has({args()}, Symbolic) 
       then fPI:= PI;
       else fPI:= float(PI); 
            A:= map(A, float):
    end_if;
    ln2:= 0.69314718056: // = ln(2.0)

    // 1-dimensional FFT using a list
    if domtype(A)=DOM_LIST then
        n:= nops(A):
        if n=0 then error("empty list") end_if;
        DIGITS:= DIGITS+10:
        m:= round(ln::float(float(n))/ln2);
        DIGITS:= DIGITS-10:
        if n<>2^m then 
           error("the list must have 2^m elements, m integer");
        end_if;
        le1 := n; 
        (le := le1; le1 := le/2; u := 1; 
         w := exp(-I*fPI/le1);
         (for i from j to n step le do
              ip := i+le1;
              t := A[i]+A[ip];
              A[ip]:= (A[i]-A[ip])*u; 
              A[i] := t;
          end_for;
          u :=  u * w;
         ) $ j=1..le1
        ) $ l=1..m;
        nv2 := n/2; nm1 := n-1; j := 1;
        (if i < j then
             t := A[j];  A[j] := A[i];  A[i] := t;
         end_if;
         k := nv2;
         while k < j do  j := j-k; k := k/2 end_while;
         j := j+k
        ) $ i=1..nm1;
        return(A);
    end_if:

    // d-dimensional FFT of an array
    if domtype(A)=DOM_ARRAY then
        // expecting A = array(1..n[1],1..n[2],...,1..,n[d])
        // d = dimension
        d:= op(A, [0,1]):
    
        // check array input
        DIGITS:= DIGITS+10:
        for k from 1 to d do
           // n[k] = k-th grid size
           n[k]:= op(A, [0, k+1, 2]);
           m[k]:= round(ln::float(float(n[k]))/ln2);
           if op(A, [0, k+1, 1]) <> 1 // indices must start from 1
           or n[k]<>2^(m[k])
           then 
              error("array indices must range from 1 to 2^m, m integer");
           end_if;
        end_for;
        DIGITS:= DIGITS-10:
        //-------------------------------------------------------
        // special case: the 1-dim list version is faster than
        // the array version, because of the index handling.
        // So, use the list version:
        if d=1 then 
          if args(0)=1 then
             return(array(1..n[1], numeric::fft([op(A)])));
          else
             return(array(1..n[1], numeric::fft([op(A)],args(2..args(0)))));
          end_if:
        end_if:
        //-------------------------------------------------------
  
        // convert d-dimensional array to a linear list
        A:= [op(A)];

        // Now the work starts:
        N:= 1;
        M:= _mult(n[k] $ k=1..d);
      
        for k from 1 to d do // transform the k-th index
      
          M:= M/n[k];  // M = n[d]*n[d-1]*..*n[k+1]
                       // N = n[k-1]*n[k-2]*..*n[1]
          start:= 0;
          for p from 1 to N do
             for r from 1 to M do
               // start = r + (p-1)*M*n[k]
               start:= start + 1;
      
               //------------------------------------------------------
               // Either do the 1-dim FFT of the k-th index via numeric::fft
               //------------------------------------------------------
               // B:= [0$n[k]]; // initialize
               // (B[j]:= A[start + (j-1)*M];) $ j=1..n[k];
               // B:= numeric::fft(B);
               // (A[start + (j-1)*M]:= B[j];) $j=1..n[k];
               //------------------------------------------------------
      
               //---------------------------------------------------
               // or do the direct version of the 1-dim FFT as above
               //---------------------------------------------------
               shift:= start - M;
               le1 := n[k];
               for l from 1 to m[k] do
                 le := le1;
                 le1 := le/2;
                 u := 1;
                 w := exp(-I*fPI/le1);
                 for j from 1 to le1 do
                   for i from j to n[k] step le do
                      sMi:= shift + M*i;
                      sMip := sMi + le1*M;
                      t := A[sMi] + A[sMip];
                      A[sMip]:= (A[sMi] - A[sMip])*u;
                      A[sMi] := t;
                   end_for;
                   u :=  u * w;
                 end_for;
               end_for;
               nv2 := n[k]/2;
               nm1 := n[k]-1;
               j := 1;
               (if i < j then
                    sMi:= shift + M*i;
                    sMj:= shift + M*j;
                    [A[sMj], A[sMi]]:= [A[sMi], A[sMj]];
                end_if;
                kk := nv2;
                while kk < j do  j := j-kk; kk := kk/2 end_while;
                j := j+kk
               ) $ i=1..nm1;
               //------------------------------------------------------
               // end of 1-dim FFT of k-th index
               //------------------------------------------------------
      
             end_for: // for r
             start:= start + M*(n[k]-1);
          end_for: // for p
          N:= N*n[k];
        end_for; // for k
      
        //-------------------------------------------------------
        // The multi-dim FFT is done and stored in the linear
        // list A. Still need to reconvert to an array structure:
        //-------------------------------------------------------
        // First,  convert linear list A to a nested list:
        //---------------------------------------------------
    
        conv:= proc(A, k)
        local N, i, j;
        begin
           if k=d then return(A) end_if:
           N:= nops(A)/n[k];
           [conv([A[j] $ j=1+(i-1)*N..i*N], k+1) $ i=1..n[k]];
        end_proc:

        //---------------------------------------------------
        // convert the nested list conv(A, 1) to an array
        //---------------------------------------------------

        return(
           array((1..n[k]) $ k=1..d, conv(A, 1))
        );
    end_if:
end_proc:
