/*
   numeric::sum=sum::float - numerical evaluation of infinite sum(f(k),k=a..b)
                             using Euler-Maclaurin formula.

   Author: Paul Zimmermann, Paul.Zimmermann@loria.fr, February 1996.
   Reference: Computational Recreations in Mathematica, Ilan Vardi, Addison Wesley,
	      1991, page 160, formula (8.16).

   /* Notduerftig ueberarbeitet am 23. 9.99, Walter */
   /* Notduerftig ueberarbeitet am 23.10.99, Walter */
   /* Erweiterung auf
      numeric::sum(f, k = RootOf(..))  am 21.11.00, Walter */

   Examples:

   >> freeze(sum)(1/k^2+1/k^3,k=1..infinity);

                         / 1    1                   \
                         | -- + --, k = 1..infinity |
                      sum|  2    3                  |
                         \ k    k                   /

   >> float(%);

                                2.84699097

   >> float(sum(ln(n)/n^(1.5-14.1*I),n=1..infinity));

                     - 0.3541793679 + 0.01416426969 I

   >>  numeric::sum(i[1], i[1]=0..100);

                                5050.0

   >>  numeric::sum(1/(1+i^2), i = -infinity..infinity);

                             3.153348094

   >> DIGITS:=40: numeric::maclaurin(ln(x)/x^2,x,2);

                0.9375482543158437537025740945678649778978

   >> numeric::sum(1/(sin(k[1]) + 2), k[1] = RootOf(X^10 - X - PI, X))

                          5.008218222

*/

numeric::sum:= proc(f,x) 
   local k, r, a, b, i;
begin
   if args(0)<>2 then error("expecting two arguments") end_if;
   if type(x)<>"_equal" then 
       error("range expected as second argument")
   end_if;
   k:= op(x,1);
   if domtype(k)<>DOM_IDENT and type(k)<>"_index" then 
       error("summation index must be an identifier or an indexed identifier")
   end_if;
   r:= op(x,2);
   if type(r)<>"_range" and type(r) <> "RootOf" then 
       error("2nd argument must be of the form i=a..b")
   end_if;
   a:=op(r,1); 
   b:=op(r,2);
   if type(r) = "_range" then
      if map({a,b},domtype) minus {DOM_INT,domtype(infinity)}<>{} then
         error("integers or +/-infinity expected as bounds")
      end_if;
      if b<a then 
         error("right bound is smaller than left bound") 
      end_if;
   end_if;
   if numeric::indets(f) <> {k} then
      error("first argument may only contain ".expr2text(k). 
            " as symbolic parameter");
   end_if;

   // !! stuff like k! etc is popular in sums.
   // Note that k! -> fact(k), but fact(k) cannot be
   // evaluated for k <> DOM_INT. So use fact(k)=gamma(k+1):
   // Also for finite sums, so do it here:

   if has(f, fact) then f:= rewrite(f, gamma); end_if;

   // Now the work starts:

   if type(r) = "RootOf" then
      r:= float(r):
      // RootOf::float either returns
      // * symbolic RootOf (if there are symbolic parameters)
      // * a set with all roots (if this is a univariate polynomial)
      // * a set with a single root (for a non-polynomial univariate equation)
      if domtype(r) <> DOM_SET then
         return(hold(sum)(f, x));
      end_if;
      // r = RootOf(h(x), x); a = op(r, 1) = h(x), k = b = op(r, 2) = x
      // Now we know that a is a univariate expression in b, because
      // that is the only case when RootOf::float yields an explicit 
      // result (a set).
      // Is the RootOf equation a polynomial?
      if indets(float(a), PolyExpr) = {b} then
         a:= eval(_plus( subs(f, k = i) $ i in r ));
         return(numeric::complexRound(a));
      end_if;
      // Now we know that a is univariate non-polynomial
      // expression. Beware, r consists of at most one root!
      return(hold(sum)(f, x));
   end_if;

   // finite sum:
   if domtype(a)=DOM_INT and domtype(b) = DOM_INT then 
      return(_plus(float(subs(f, k=i)) $ i=a..b));
   end_if;
   // infinite sum:
   if a=-infinity and b=infinity then
      return(float(eval(subs(f,k=0)))
             + numeric::maclaurin(f + subs(f,k=-k), k, 1)
             );
   end_if; 
   if domtype(a) = DOM_INT and b=infinity
      then return(numeric::maclaurin(f, k, a))
   end_if;
   if a = -infinity and domtype(b)=DOM_INT 
      then return(numeric::maclaurin(subs(f,k=-k),k,-b))
   end_if;
   // all cases should be covered above.
   hold(sum)(f, x)
end_proc:

/* 
   Euler-MacLaurin formula:
  
    sum(f(k), k=a..b) = (f(a)+f(b))/2 + int(f(k),k=a..b)
    + sum(bernoulli(2*m)/(2*m)!* ((D@@(2m-1))(f)(b)-(D@@(2m-1))(f)(a)], m=1..M)
    + Remainder(M)

*/

//numeric::maclaurin(f,k,a) = sum(f(k),k=a..infinity) 

numeric::maclaurin:= proc(f,k,a) 
   local n,m,t,S,S0,macheps;
begin
   f:= float(f);
   if args(0)<>3 then error("expecting three arguments") end_if;
   if domtype(a)<>DOM_INT then return(FAIL) end_if;

   // !! stuff like k! etc is popular in sums.
   // Note that k! -> fact(k), but fact(k) cannot be
   // evaluated for k <> DOM_INT. So use fact(k)=gamma(k+1):
   if has(f, fact) then f:= rewrite(f, gamma); end_if;

   // Heuristics: split off a finite sum to improve convergence
   // of Euler-MacLaurin for the infinite tail :
   //    sum(f(k),k=a..infinity) = 
   //              sum(f(k),k=a..n-1)       <- use _plus
   //            + sum(f(k),k=n..infinity)  <- use MacLaurin

   n:= a + 10*DIGITS;

   // S = first approximation via MacLaurin 
   //   = sum(f(k), k=a..n-1) + sum(f(k), k=n..infinity)
   //   = sum(f(k), k=a..n-1) + f(n)/2 + int(f(k), k=n..infinity)

   t:=float(eval(subs(f,k=n))); 
   if not contains({DOM_COMPLEX,DOM_FLOAT},domtype(t)) 
      then return(FAIL)
   end_if;

   // Try symbolic or numerical integration.
   // Use float(int(..)) instead of numeric::quadrature,
   // because int::float (=numeric::int) traps errors from
   // numeric::quadrature.

   S0:= float(int(f, k=n..infinity));

   if has(S0, limit) then 
      // int gave a symbolic result, but limit was not 
      // strong enough and survived as symbolic limit.
      // Switch to purely numerical integration:
      S0:= int::float(f,k=n..infinity); 
   end_if;

   // give up, if integration fails:
   if not testtype(S0, Type::Numeric) then return(FAIL) end_if; 
   
   S:=float(eval(
         _plus(subs(f,k=float(m))$m=a..n-1)//   sum(f(k), k=a..n-1)
         + t/2                             // + f(n)/2
         + S0                              // + int(f(k), k=n..infinity)
            ));

   // give up, if something went wrong:
   if not testtype(S, Type::Numeric) then return(FAIL) end_if; 

   //Now add corrections - sum(bernoulli(2*m)/(2*m)!* (D@@(2m-1))(f)(a), m)
   //from Euler-MacLaurin formula until S becomes stationary:
   macheps:= float(10^(-DIGITS)):
   f:=diff(f,k); 
   m:=1; 
   repeat
      t:=float(eval(bernoulli(2*m)/fact(2*m)*subs(f,k=float(n))));
      if not testtype(t, Type::Numeric) then return(FAIL) end_if; 
      S:=S-t; 
      // stopping criterion: last step did not change S significantly:
      if abs(t) < macheps * (macheps +abs(S)) then break end_if;
      m:=m+1; 
      f:=diff(f,k,k);
   until FALSE end_repeat;

   // We should check in addition that f is (2m+4)-times differentiable with
   // diff(f,k$2*m+2) and diff(f,k$2*m+4) of one sign on 1..infinity !!!

   // Note that sum((-1)^k*f(k), k=a..infinity) 
   // leads to  int((cos(k*PI)+I*sin(k*PI))*f(k), k=a..infinity)
   // with imaginary round-off trash. Hence apply numeric::complexRound:

   return(numeric::complexRound(S, macheps))
end_proc:

// end of file
