/*--
Ist keine Interface-Funktion und wird von plot2d- und plotfunc2d aufgerufen.

get_intervals( f, ... )
f: DOM_PROC oder DOM_EXEC (d.h. eine Funktion!)

Diese Prozedur untersucht, ob die Funktion f(x) im endlichen
Intervall [a,b] Singularitaeten besitzt und teilt ggf. die Intervalle
auf.

Enthaelt f an der Stelle d eine Singularitaet, so wird ein
maximaler bzw. minimaler Wert f"ur die Funktion gesetzt, ein sogenannter
threshold. 

Ergebnis hat folgende Form:

         [ [[a_1,b_1], ... , [a_n, b_n]] , miny, maxy ]

n ganze Zahl >= 1 und miny, maxy reelle Zahlen oder +/- infinity.
Ist miny bzw. maxy ungleich +/- infinity, so geben diese beiden
Werte einen Anhaltspunkt ueber den kleinsten und groessten
Funktionswert f(x) innerhalb der Intervalle [a_1,b_1], ... ,
[a_n, b_n].

Bsp:

>> get_intervals(x->1/cos(x),x,0,PI,10,100);

  [[[0.0, 1.507091269], [1.634501385, 3.141592654]], -15.70796327, 15.70796326]


>> get_intervals(x->10*x/(x^2-2*x-8),x,-5,5,10,100);

  [ [ [-5.0, -2.072061831], [-1.938005504, 3.860888647],
      [4.127955253, 5.0]], -47.35449739, 52.64550263]

Ohne Verfeinerung mittels Bisektion w"urden die folgenden Intervalle geliefert:

                [[-5.0, -2.09], [-1.88, 3.82], [4.13, 5.0]]

		Sind thrmin und/oder thrmax 
1) reell, dann gilt f.a. x aus [a, b]: thrmin <= f(x) <= thrmax, oder
2) thrmax = infinity, dann wird ratio fuer das Verhaeltnis zwischen 
   x- und y-Achse herangezogen.
--*/

numeric::get_intervals := proc(ff, xx, aa, bb, ratio, grid, 
                              thrminy=-infinity, thrmaxy=infinity)
   local eps, ds, i, l, ll, lr, n, t, tl, bl, br,
         max_y, min_y, nb, new_bound, bisection,
         f, x, a, b, retry, thres_min_y, thres_max_y;
begin
   f:=ff; x:=xx; a:=aa; b:=bb; // to avoid warnings 
   thres_min_y:=thrminy; thres_max_y:=thrmaxy;

   bisection := proc(aa, bb, bound)
      name numeric::get_intervals::bisection;
      local a, b, c, eps;
   begin
      a:=aa; b:=bb; // to avoid warnings 
      eps := 10^(-DIGITS);
      while abs(b-a) > eps do
         c := (a+b)/2; 
         if  float(f(c)) <= bound then
            a := c;
         else
            b := c;
         end_if
      end_while;
      a;
   end_proc;

   // Berechne a, sodass thres_min_y <= f(a) <= thres_max_y 

   new_bound := proc(aa, epss)
      name numeric::get_intervals::new_bound;
      local min_eps, y, a, eps;
   begin
      a:=aa; eps:=epss; // to avoid warnings 

      min_eps := 10^(-DIGITS+1);
     
      if traperror((
           while abs(eps) > min_eps and thres_min_y <= (y := float(f(a+eps))) and y <= thres_max_y do
             eps := eps/2;
           end_while;
         )) <> 0 or abs(eps) <= min_eps then return(a+eps*2) end_if;
/*
      if thres_min_y > y or y > thres_max_y then
         return(a+eps)
      end_if;
*/

      repeat
        a := a + eps; y := float(f(a));
      until thres_min_y <= y and y <= thres_max_y end_repeat;

      // Der folgende Abschnitt dient nur zur Verfeinerung des Wertes 'a' 
      // Diese verfeinert insbes. die graphische Ausgabe.                 
      // Vor Ausf"uhrung des Abschnitts gilt entweder                     
      // 1) f(a) <= thres_max_y und f(a-eps) > thres_max_y , oder         
      // 2) f(a) >= thres_min_y und f(a-eps) < thres_min_y                
      // Vor Ausf"uhrung des Abschnitts gilt entweder                     
      // 1) f(a) = thres_max_y oder 2) f(a) = thres_min_y                 

      y := float(f(a-eps));
      if y > thres_max_y then
         a := bisection(a, a-eps, thres_max_y)
      else
         a := bisection(a-eps, a, thres_min_y)
      end_if;

      a 

   end_proc;

   //------------- Hauptprogramm 'get_intervals' ----------------------

   t := traperror((
     if type(f(x)) = piecewise then
        ds := discont(f(x), x, Dom::Real);
     else
        ds := discont(f(x), x=a..b, Dom::Real);
     end_if
   ));

   // Enthaelt f keine Singularitaeten oder konnten keine 
   // Singularitaeten ermittelt werden, so gebe Funktion  
   // f und das Intervall [a, b] unveraendert zurueck.    

   if t <> 0 or ((type(ds) <> DOM_SET or ds = {}) and 
                  type(ds) <> "_union" and type(ds) <> "_intersect" ) then 
      if domtype(float(thres_min_y)) = DOM_FLOAT then
         return([ [[a,b]], thres_min_y, thres_max_y ])
      else
         return([ [[a,b]], infinity, -infinity ])
      end_if
   end_if:

   if type(ds) = "_union" then
      t := [];
      for i from 1 to nops(ds) do
          if type(op(ds, i)) = DOM_SET then
             if traperror(
	         (l:=select(map([op(op(ds,i))], float), testtype, DOM_FLOAT))
	        ) = 0 then
                t := t . l
	     end_if
	  else 
	     t := t . numeric::check_solution(op(ds, i))
	  end_if
      end_for;
      // t := t. (select(map([op(op(ds,i))], float), testtype, DOM_FLOAT)) 
      //          $ i=1..nops(ds);
      ds := sort(t)
   elif type(ds) = "_intersect" then
      t := numeric::check_solution(ds);
      if t = [] then
         return([ [[a,b]], infinity, -infinity ])
      end_if;
      ds := sort(t)
   else
      ds := sort(select(map([op(ds)], float), testtype, DOM_FLOAT))
   end_if;

   t := [];
   a := float(a); b := float(b);
   for i from 1 to nops(ds) do
       if ds[i] >= a and ds[i] <= b then t := append(t, ds[i]) end_if
   end_for;

   if t = [] then return([ [[a,b]], infinity, -infinity ]) end_if;
   ds := t; n := nops(ds); 
   l := []; tl := []; bl :=[]; br := [];

   if abs(ds[1] - a) > 10^(-DIGITS+1) then
      // ds[1] <> a
      t := float(f(a));
      tl := append(tl, t);
      bl := [a, ds[1]];
   end_if;

   if abs(ds[n] - b) > 10^(-DIGITS+1) then       
      // ds[n] <> b 
      t := float(f(b));
      tl := append(tl, t); 
      br := [ds[n], b]; 
   end_if; 
   
   for i from 2 to n do
      t := float(f(ds[i-1]+(ds[i]-ds[i-1])/2));
      tl := append(tl, t);  
      l  := append(l, [ds[i-1], ds[i]]);  
   end_for;

   min_y := min(op(tl)); max_y := max(op(tl));

   nb := 1; tl := l;

   repeat 

   retry := FALSE;

   t := max(0, ( ratio * (b-a) - (max_y - min_y) ) / 2);
   if thres_max_y = infinity then
      // start value
      thres_max_y := max_y + t; thres_min_y := min_y - t;
   else
      if max_y + t > thres_max_y then
         thres_max_y := max_y + t;
      end_if;
      if min_y - t < thres_min_y then
         thres_min_y := min_y - t;
      end_if;     
   end_if;

/* Alternative Version - nicht loeschen

   if thres_max_y = infinity then
      t := (max_y - min_y) / (ratio * (b-a));
      thres_max_y := max_y / t; thres_min_y := min_y / t;
   end_if;

*/

   eps := (b-a)/grid;

   for i from 1 to nops(l) do     
       if (l[i][2]-l[i][1]) < 10^(-DIGITS+1) then
          next
       end_if;
       while eps > (l[i][2]-l[i][1]) and eps > 10^(-DIGITS+1) do 
          eps := eps/2
       end_while;            
       ll := new_bound(l[i][1], +eps);
       lr := new_bound(l[i][2], -eps);
       if ll > lr then
          ll := l[i][1] + (l[i][2]-l[i][1])/4;
          lr := l[i][2] - (l[i][2]-l[i][1])/4;
       end_if;
       l[i][1] := ll; l[i][2] := lr;
   end_for;                                                   
   if bl <> [] then   
      if (bl[2]-bl[1]) < 10^(-DIGITS+1) then
         next
      end_if;
      while eps > (bl[2]-bl[1]) and eps > 10^(-DIGITS+1) do
          eps := eps/2
      end_while;
      lr := new_bound(bl[2], -eps);
      if bl[1] > lr then
         lr := bl[2] - (bl[2]-bl[1])/4
      end_if;
      l := [[bl[1], lr]].l;     
   end_if;
   if br <> [] then   
      if (br[2]-br[1]) < 10^(-DIGITS+1) then
         next
      end_if;
      while eps > (br[2]-br[1]) and eps > 10^(-DIGITS+1) do
          eps := eps/2
      end_while;
      ll := new_bound(br[1], +eps);
      if ll > br[2] then
         ll := br[1] + (br[2]-br[1])/4
      end_if;                           
      l := append(l, [ll, br[2]]);               
   end_if;

   for i in l do
       if abs(i[1]-i[2]) < 10^(-DIGITS) then
          // Linke und rechte Intervallgrenze
          // (x-Bereich) sind (fast) identisch.
          // Erhoehe das Verhaeltnis zwischen
          // x- und y-Bereich um einen groesseren
          // Bildausschnitt zu bekommen. Damit
          // wird implizit der y-Bereich festgelegt.
          ratio := ratio*2; l := tl;
          retry := TRUE; nb := nb + 1;
          break; 
       end_if
   end_for;

   // Nach 2 fehlgeschlagenen Versuchen den Bildbereich, 
   // d.h. automatisch den y-Bereich festzulegen wird abgebrochen.

   until not retry or nb > 3 end_repeat;

   if retry then
      return( [ FAIL, infinity, -infinity ] )
   else
      // Recompute min. and max value of f
      // because thres_min_y and thres_max_y
      // are not always valid(see function new_bound)
      min_y := infinity: max_y := -infinity:
      for t in l do  
          //for x in t do
          // 10 evaluations only
          for x from t[1] to t[2] step (t[2]-t[1])/10 do
              nb := float(f(x)); 
              if nb > max_y then max_y := nb end_if;
              if nb < min_y then min_y := nb end_if
          end_for 
      end_for;

      return( [ l, min_y, max_y ] )
   end_if

end_proc:

numeric::get_singval := proc(d, c : Type::Even)
   local f, im, iv, n, s, y;
begin
   s := {};

   // Compute numerical values of the expression
   // { f(x) | x in Z} intersect [a, b]
   // Compute at most c values: f(x), x in [-c/2, c/2]
   // Used especially to compute values for 
   // f(x) = 1/(n*x)

   if testtype(d,"_intersect") then
      if (type(op(d,1)) = "RootOf" and
          type(op(d,2)) = Dom::Interval) or
         (type(op(d,1)) = Dom::Interval  and
          type(op(d,2)) = "RootOf"    )  then

	 // [a,b] intersect RootOf(...)

	 s := simplify(float(d));
	 if domtype(s) <> DOM_SET then
	    return({})
	 else
	    return(s)
	 end_if
      elif type(op(d,1)) = Dom::ImageSet and
         type(op(d,2)) = Dom::Interval then
         im := op(d,1); iv := op(d,2)
      elif type(op(d,1)) = Dom::Interval and
           type(op(d,2)) = Dom::ImageSet then
         im := op(d,2); iv := op(d,1)
      else
         return( s )
      end_if;

      // Compute numerical values of the expression
      // { f(x) | x in Z} intersect [a, b]
      // Compute at most c values: f(x), x in [-c/2, c/2]
      // Used especially to compute values for
      // f(x) = 1/(n*x)

      if nops(op(im, 2)) = 1 and 
         nops(op(im, 3)) = 1 and
         testtype(op(im,[3,1]),
                  solvelib::BasicSet(Dom::Integer)) then
         f := fp::unapply(op(im, 1), op(im, [2,1]));
         n := c/2; 
         while n >= -c/2 do
	    if traperror((y:=float(f(n)))) = 0 and contains(iv, y) then
                  s := s union {y}
	    end_if;
	    n:=n-1 
         end_while
      end_if;
   end_if;

   return( s )
end_proc:

numeric::check_solution := proc(d)
   local m, n, s;
begin
   s := 20;
   n := numeric::get_singval(d, s);
   if nops(n) >= s then
      m := "\n More than " . expr2text(s) . " discontinuities detected.\n";
      m := m . " Interval [" . expr2text(min(op(n))) . ", ";
      m := m . expr2text(max(op(n))) . "]\n";
      error(m);
   else
      [op(n)]
   end_if
end_proc:
