/*
    numeric::indets(anyMuPADObject)  

returns a set of all indeterminates. 

Indeterminates are:
   identifiers
   indexed identifiers
   symbolic entries in DOM_POLY 
      + indeterminates of DOM_POLY, e.g.,
         numeric::indets( poly(a, [x])); -->  {x, a}


Not considered as indeterminates:
   Type::ConstantIdents (i.e., PI, ..)
   zero-Operands of objects
   integration variables in numeric::quadrature and hold(int)
   Beware: in hold(int)(f(x),x) this is not appropriate, but
           numeric::indets is only used in numeric::quadrature,
           which has to deal with hold(int)(f(x),x=a..b).
*/

numeric::indets:= proc(a)
   local indetsSet, search;
   begin
     if args(0)<>1 then error("expecting one argument") end_if;
     indetsSet:= {}: //initialize container for indets

     // recursive search. Adds indets to global indetsSet.
     search:= proc(a)
       name numeric::indets@search;
       local i; begin
       if domtype(a) = DOM_IDENT then
          indetsSet:= indetsSet union {a};
          return(a);
       end_if;
       if type(a) = "_index" then
          indetsSet:= indetsSet union {a};
          return(a);
       end_if;
       if domtype(a) = DOM_POLY then
          indetsSet:= indetsSet union {op(op(a,2))};
          search(map(poly2list(a),op,1));
          return(a);
       end_if;
       // mathematical interpretation of indets: ignore
       // integration variable in unevaluated numeric::quadrature:
       if op(a, 0) = hold(numeric::quadrature) 
       or op(a, 0) = hold(int) 
       then
          indetsSet:= indetsSet union (
                      numeric::indets([op(a,1),op(a,2)]) minus {op(a,[2,1])}
                                       );
          return(a);
       end_if;
       // ignore summation index in unevaluated sum:
       if op(a, 0) = hold(sum) 
       then
          indetsSet:= indetsSet union (
                      numeric::indets([op(a,1),op(a,2)]) minus {op(a,[2,1])}
                                       );
          return(a);
       end_if;

       // do not enter objects of library domains, there is danger
       // of infinite recursion!
       // Warning!!! Following characterization of kernel objects
       // works in MuPAD 1.5, but not in MuPAD 1.4 !!!!!!
       if domtype(extop(a,0))=DOM_DOMAIN then return(a) end_if;

       // make sure there is no infinite recursion:
       if a=op(a,1) then return(a) end_if;

       // handle expressions and other stuff:
       for i from 1 to nops(a) do search(op(a,i)): end_for;
       return(a);
     end_proc:

     //------ main body of numeric::indets ------------
     // treat most common objects first to avoid costs of an
     // unnecessary recursive call to search:
     case domtype(a)
     of DOM_LIST    do 
     of DOM_SET     do
     of DOM_ARRAY   do
     of DOM_TABLE   do map(a, search);
                       return(indetsSet minus Type::ConstantIdents);
     of DOM_FLOAT   do 
     of DOM_COMPLEX do 
     of DOM_INT     do 
     of DOM_RAT     do return(indetsSet);
     end_case;

     // do the recursive search, which fills up indetsSet.
     search(a);
     indetsSet minus Type::ConstantIdents;
   end_proc:
