/*--------------- Author: W. Oevel, 17.3.00 ------------------

numeric::isnonzero -- heuristic numerical test whether a (constant)
                      algebraic expression is non-zero

Call(s)     : numeric::isnonzero(x)
Parameters  : x -- an algebraic expression
Return Value: TRUE, UNKNOWN

Details: TRUE is returned, if float(x) indicates that
              x is unlikely to coincide with float(0),
              even if DIGITS is increased
         UNKNOWN is returned, if x contains symbolic objects,
              or if the numerical test is not decisive, 
              because float evaluation is subject to severe 
              round-off problems.

Roughly speaking: 
         The answer TRUE is reliable (with very high probability).
         If you get TRUE, then you do not need to increase DIGITS
         to doublecheck.
  
         If you get UNKNOWN, then it is up to you to try again 
         with increased DIGITS, which then might yield a 
         decisive TRUE.

Examples:

>> numeric::isnonzero(sqrt(24) - 3*sqrt(6))

                 TRUE     // this is a decisive answer !

>> numeric::isnonzero(sqrt(24) - 2*sqrt(6))

                 UNKNOWN

>> DIGITS:= 50: pi:= numeric::rationalize(float(PI)):
>> DIGITS:= 10:
>> float(PI - pi), numeric::isnonzero(PI - pi);

                         0.0, UNKNOWN

>> DIGITS:= 30:
>> float(PI - pi), numeric::isnonzero(PI - pi);

                         0.0, UNKNOWN

>> DIGITS:= 50:
>> float(PI - pi), numeric::isnonzero(PI - pi);

           -4.179026102...e-51, TRUE   // you can rely on TRUE !

-----------------------------------------------------------*/

numeric::isnonzero:= proc(x)
local x1, x2;
begin
   // float test
   x1:= float(x):
   if not has({DOM_FLOAT, DOM_COMPLEX}, domtype(x1)) then
      return(UNKNOWN);
   end_if:
   if iszero(x1) then return(UNKNOWN) end_if:
   // Heuristics: increase DIGITS to verify
   // that t is not complete round-off trash
   // We need at least DIGITS + 10 to make sure,
   // that PARI adds a further block:
   x1:= specfunc::abs(x1):
   DIGITS:= DIGITS + 10;
   x2:= specfunc::abs(float(x)):
   DIGITS:= DIGITS - 10;
   // the leading digit of x1 is reliable, if 1/2 < x1/x2 < 2:
   if (x1<2*x2 and x2<2*x1)
      then return(TRUE)
      else // no chance to decide the magnitude of
           // x via float with the present DIGITS.
           // Severe cancellation! 
           return(UNKNOWN)
   end_if;
end_proc:
