// 
// $Date: 2000/05/21 14:56:32 $  $Author: bij $   $Revision: 1.32 $ 
// 
// W.Oevel, 06/04/98 
//------   file  linsolve.mu   6.4.98     ------------------------
//-------------------  W. Oevel  ---------------------------------
/*  Help file:

 linsolve - solve a system of linear equations numerically

 Call:

 numeric::linsolve( eqs  <, vars> <, Symbolic> <, ShowAssumptions>  )
                                                                   
 Parameter:

  eqs         --  a list or set of linear equations or linear expressions
                  (interpreted as equation expr = 0)
                  Also lists/sets of linear DOM_POLY polynomials are accepted.
  vars        --  a list or set of unknowns
  Symbolic    --  option, prevents conversion of coefficients to floats
  ShowAssumptions
              --  results in more detailed information about the
                  internal assumptions on symbolic parameters in the
                  equations (see below)
                                                                   
Synopsis:

  Let vars=[x[1],x[2],..] or vars={x[1],x[2],..} be the unknowns.

  numeric::linsolve(eqs,vars, <Symbolic>); returns  Solvedeqs

  numeric::linsolve(eqs,vars, <Symbolic>, ShowAssumptions); returns 
            [Solvedeqs, Constraints, PivotAssumptions]

  (*) Solvedeqs is a list of equations of the form x[1]=..,x[2]=..,..
      representing the solution of eqs. Those unknowns not subject
      to any constraint (free parameters) will not turn up
      on the left hand sides in Solvedeqs. E.g., if vars=[x[1],x[2]]
      and Solvedeqs=[x[2]=..], then x[1] is a free parameter spanning
      the kernel of eqs. Solvedeqs=[] corresponds to a trivial system
      eqs, where all unknowns are arbitrary.

      With assign(Solvedeqs) the solution may be assigned to the unknowns.

  (*) Without option ShowAssumptions:

     (*) Warning: if the equations contain symbolic parameters, then only 
         those solutions are returned which hold for arbitrary values of 
         these parameters.

         Symbolic denominators occuring in the final result or in
         intermediate results are implicitly assumed to be non-zero. 
         The fact that they may vanish for certain values of the 
         symbolic parameters is ignored.

  (*) With option ShowAssumptions:

     (*) PivotAssumptions is a list of inequalities involving the
         symbolic parameters of the coefficient matrix. The
         corresponding expressions appeared as symbolic pivot elements
         in the internal Gauss elimination. They are assumed to be
         nonzero and appear as denominators in the final result or in 
         itermediate results.

         If no symbolic pivot elements were used, then PivotAssumptions
         is the empty list. This will always be the case when
         there are no symbolic parameters in the coefficient
         matrix.

         The inequalities arising in PivotAssumptions are 
         determined by the algorithm during run time. The user
         has no control over these assumptions. In particular,
         changing the ordering of the equations in eqs or the
         ordering of the variables in vars may result in a
         different solution returned by linsolve.

     (*) Constraints is a list of equations involving the symbolic parameters
         in the equations.

         eqs is solvable if and only if the
         equations given by Constraints are satisfied.

         Such constraints arise only when the coefficient matrix 
         does not have full rank.

         If no such contraints arise, then 
         the returned value of Constraints is the empty list.

     (*) Under the assumptions PivotAssumptions and assuming solvability
         of eqs (i.e., assuming the equations given by Constraints),
         Solvedeqs represents the general solution of eqs.

  If the unknowns vars are not specified, then the symbolic parameters
  in eqs are assumed as unknowns, i.e., vars=indets(eqs, PolyExpr).

  Warning: linsolve does not check that the equations eqs are indeed
  linear in the unknowns vars. For non-linear systems linsolve may
  return wrong results!

  By default all coefficients in eqs are converted to floats. No 
  non-numeric symbolic parameters are allowed as coefficients of 
  the unknowns, unless the option Symbolic is used. In the symbolic 
  mode no conversion to floats will occur. We note that the presence 
  of non-numeric symbolic coefficients will slow down the computation
  drastically.

  Gaussian elimination with partial pivoting is used by linsolve.
  If no solution exits, then linsolve returns FAIL.

  Examples: 

>> numeric::linsolve({x=y-1,x+y=0},{x,y});

                            [x = -0.5, y = 0.5]

>> numeric::linsolve({x=y-1,x+y=0},{x,y}, ShowAssumptions);

                         [[x = -0.5, y = 0.5], [], []]

>> numeric::linsolve([x+y=1,y=z],[x,y,z],Symbolic);

                            [x = 1 - z, y = z]

>> numeric::linsolve([x+y=1,y=z],[x,y,z],Symbolic, ShowAssumptions);

                         [[x = 1 - z, y = z], [], []]

>> numeric::linsolve({x+y=1,x+y=a},{x,y},Symbolic);

                                   FAIL

>> numeric::linsolve({x+y=1,x+y=a},{x,y},Symbolic, ShowAssumptions);

                        [[x = 1 - y], [a - 1 = 0], []]

>> numeric::linsolve({x+y=1,x+y=a},{x,y,a});

                          [a = 1.0, x = 1.0 - y]

>> numeric::linsolve({x+y=1,x+y=a},{x,y,a}, ShowAssumptions);

                       [[a = 1.0, x = 1.0 - y], [], []]

>> numeric::linsolve({x=x-y+1,y=PI},Symbolic);

                             FAIL

>> numeric::linsolve({x=x-y+1,y=PI},Symbolic, ShowAssumptions);

                     [[y = 1], [PI - 1 = 0], []]

>> numeric::linsolve({x=x-y+1,y=PI},{x,y},Symbolic);

                                   FAIL

>> numeric::linsolve({x=x-y+1,y=PI},{x,y},Symbolic, ShowAssumptions);

                          [[y = 1], [PI - 1 = 0], []]

>> numeric::linsolve({x=x-y+1,y=PI},{x,y});         

                                   FAIL

>> numeric::linsolve({f(a)=sin(x),f(a)=1-sin(x)},{f(a),sin(x)});

                        [sin(x) = 0.5, f(a) = 0.5]


>> numeric::linsolve({f(a)=sin(x),f(a)=1-sin(x)},{f(a),sin(x)}, ShowAssumptions);

                   [[sin(x) = 0.5, f(a) = 0.5], [], []]


>> numeric::linsolve([a*x+b*y= A, c*x + d*y = B], [x]);

                                   FAIL

>> numeric::linsolve([a*x+b*y= A, c*x + d*y = B], [x], Symbolic, ShowAssumptions);
Warning: symbolic coefficient found, switching to symbolic mode

    -- --     A - b y --                                            --
    |  |  x = -------  |, [B a - A c - a d y + b c y = 0], [a <> 0]  |
    -- --        a    --                                            --

// Do it again with a=0: 

>> numeric::linsolve([b*y= A, c*x + d*y = B], [x], ShowAssumptions);


             -- --     B - d y --                          --
             |  |  x = -------  |, [A - b y = 0], [c <> 0]  |
             -- --        c    --                          --

// Do it again with a=0, c=0: 

>> numeric::linsolve([b*y= A, d*y = B], [x], Symbolic, ShowAssumptions);

                   [[], [b y - A = 0, d y - B = 0], []]

>> numeric::linsolve([a*x+b*y= A, c*x + d*y = B], [x, y]);

                     --     A d - B b      B a - A c --
                    |  x = ---------, y = ---------  |
                    --     a d - b c      a d - b c --

>> numeric::linsolve([a*x+b*y= A, c*x + d*y = B], [x, y], Symbolic, ShowAssumptions);

  -- --     A d - B b      B a - A c --                               --
  |  |  x = ---------, y = ---------  |, [], [a <> 0, a d - b c <> 0]  |
  -- --     a d - b c      a d - b c --                               --

>> numeric::linsolve([a*x+b*y= A, c*x + d*y = B, x = y], [x, y], Symbolic);
 
                         FAIL

>> numeric::linsolve([a*x+b*y= A, c*x + d*y = B, x = y], [x, y], Symbolic, ShowAssumptions);

-- --       B          B   --                                            --
|  |  x = -----, y = -----  |, [A c - B a - B b + A d = 0], [c + d <> 0]  |
-- --     c + d      c + d --                                            --

>> eqs:= [x[1]+eps*x[2]=1+eps, eps*x[1]+x[2]=1+eps, 2*x[1]+3*x[2]=5]:
>> numeric::linsolve(eqs, [x[1], x[2]], Symbolic);

                           [x[1] = 1, x[2] = 1]

>> numeric::linsolve(eqs, [x[1], x[2]], Symbolic, ShowAssumptions);

                                                   2
                [[x[1] = 1, x[2] = 1], [], [1 - eps  <> 0]]
                 
>> eqs:=[x+y=1,x+y=a]:
   ([Solution,Constraints, Assumptions]):=
   numeric::linsolve(eqs,{x,y}, ShowAssumptions);

                  [[x = 1.0 - y], [1.0 a - 1.0 = 0], []]

>> subs(eqs,Solution);

                            [1.0 = 1, 1.0 = a]

>> assign(Solution): eqs;

                            [1.0 = 1, 1.0 = a]


>> numeric::linsolve([a*x + b*y=c, b*x -a*y , a*b*x],[x,y,z],Symbolic,ShowAssumptions);

-- --       a c          b c   --      2                       2    2
|  |  x = -------, y = -------  |, [- a  b c = 0], [a <> 0, - a  - b  <> 0]
|  |       2    2       2    2  |
-- --     a  + b       a  + b  --

   --
    |
    |
   --

>> numeric::linsolve([b*x -a*y , a*x + b*y=c, a*b*x],[x,y,z],Symbolic,ShowAssumptions);

-- --       a c          b c   --      2                     2    2
|  |  x = -------, y = -------  |, [- a  b c = 0], [b <> 0, a  + b  <> 0]
|  |       2    2       2    2  |
-- --     a  + b       a  + b  --

   --
    |
    |
   --

>> numeric::linsolve([a*b*x, a*x + b*y=c, b*x -a*y ],[x,y,z],Symbolic,ShowAssumptions);

          -- --            c --                                --
          |  |  x = 0, y = -  |, [a c = 0], [a b <> 0, b <> 0]  |
          -- --            b --                                --

// the user may conclude from the last output: a*b<>0 -> a<>0,b<>0 -> c=0 
// These simplifications were not done automatically by linsolve

>> numeric::linsolve([a*b*x, b*x -a*y, a*x + b*y=c ],[x,y,z],Symbolic,ShowAssumptions);

              [[x = 0, y = 0], [c = 0], [a b <> 0, -a <> 0]]


See also:  linsolve, linalg::matlinsolve, numeric::matlinsolve
                                                                  */
//-----------------------------------------------------------------
numeric::linsolve:=proc(eqs, vars)
local neqs, n, method, use_normal, i, j, tmp, B, b, triv, check,
      showassumptions, Arg, eqsSymbolicParameters, varsSet,
      x, X, Xvars, nTerms, koeff, np2, np2mj, p,
      piv, pivot, pivindex, rownorm, constraints,
      PivotAssumptions, lowerbands, jpl, minlength;
begin
  if args(0)<1 then error("not enough arguments"); end_if;
  if args(0)>4 then error("expecting at most four arguments");end_if;

  method:= 2;  // default: numeric mode
  showassumptions:= FALSE; // default
  if args(0)>1 then
     for Arg in [args(2..args(0))] do
       if Arg = Symbolic then method:=1; next; end_if:
       if Arg = ShowAssumptions then showassumptions:= TRUE; next; end_if:
       if domtype(Arg)=DOM_LIST or domtype(Arg)=DOM_SET then next; end_if;
       error("unknown option");
     end_for:
  end_if;
     
  if not contains({DOM_SET, DOM_LIST}, domtype(eqs))
     then error("expecting a set or a list of equations or expressions")
  end_if;
  if domtype(eqs)=DOM_SET then eqs:=[op(eqs)] end_if; // convert to list 
  
  // convert polynomials 
  (if testtype(eqs[i],DOM_POLY) 
      then eqs[i]:= expr(eqs[i]);
   end_if;) $ i=1.. nops(eqs);

  // convert equations a=b to expressions a-b 
  (if testtype(eqs[i],"_equal") 
   then eqs[i]:= op(eqs[i],1)-op(eqs[i],2);
  end_if;) $ i=1.. nops(eqs);
   // force full evaluation of the equations: 
// (eqs[i]:=eval(eqs[i]);) $ i=1..nops(eqs);

  if nops(eqs)=0 then
     if showassumptions then return([[],[],[]])
                        else return([])
     end_if
  end_if;
 
  if args(0)=1 or ({args(2..args(0))} minus {Symbolic,ShowAssumptions}={})
  then // get rid of Type::ConstantIdents (i.e., PI etc.) via float.
       // Note that indets(..,PolyExpr) would enter 1/PI into varsSet:
       varsSet:= indets(map(eqs,float), PolyExpr);
       vars:=[op(varsSet)];
  else // linsolve is called with vars 
       if not contains({DOM_SET, DOM_LIST}, domtype(vars))
       then error("expecting a set or a list of unknowns")
       end_if;
       // force evaluation of vars ?
//     vars:= map(vars, eval);
       // convert vars to List 
       if domtype(vars)<>DOM_LIST
          then //Variables are also needed as set 
               varsSet:= vars;
               vars:=[op(vars)];
          else varsSet:= {op(vars)};
       end_if; 
  end_if;

  n:=nops(vars);

  if n=0 then
     // eliminate trivial equations
     eqs:= map(eqs, proc(eq) begin 
                      if iszero(eq)
                         then null()
                      else eq end_if
                    end_proc);
     if eqs =[] and showassumptions=FALSE  then return([]) end_if;
     if eqs =[] and showassumptions=TRUE   then return([[], [], []]) end_if;
     if eqs<>[] and showassumptions=FALSE  then return(FAIL) end_if;
     if eqs<>[] and showassumptions=TRUE   then
                if indets(eqs,PolyExpr)={} then return([FAIL,[], []])
                                           else return([[], eqs, []])
                end_if;
     end_if;
  end_if;

  neqs:=nops(eqs);
  np2:=n+2;
  x:=genident("x");   // generate free identifier for polynomials 
  // convert eqs to polynomial expressions 
  eqs:=subs(eqs,[(vars[j]=x^(np2-j)) $ j=1..n]);
  // pick out right hand side of equations 
  b:=subs(eqs,x=0,float(0)=0);
  eqs:=subs(zip(eqs,b,proc() begin args(1)-args(2) end_proc ),float(0)=0);
  eqsSymbolicParameters:=(indets(eqs,PolyExpr) minus {x}) minus varsSet;


  // Now convert polynomial expressions eqs to polynomials in x. Need 
  // storage p of type table, since neqs may be increased in the following. 
  // Similarly, store right hand side b of equations in table B. 

  (p[i]:=poly(eqs[i],[x]);) $ i=1..neqs; // p must be a table 
  (B[i]:=-b[i];) $ i=1..neqs;            // B must be a table 

  (if p[i]=FAIL or degree(p[i])>n+1 then
      error("this system does not seem to be linear")
   end_if:) $ i=1..neqs;

  // convert to floats after conversion to polynomials! Otherwise 
  // float(x+2*y=b) -> x + 2.0*y = b, i.e., the integer coefficient 1 would survive! 
  if method=2 // numeric mode 
  then if map( float(eqsSymbolicParameters),domtype ) 
          minus {DOM_FLOAT,DOM_COMPLEX,DOM_INT} <> {}
       then method:=1;
            warning("symbolic coefficient found, switching to symbolic mode");
       else p:=map(p, mapcoeffs, float); 
            B:=map(B, float);
       end_if;
  end_if;

  if method=1 and eqsSymbolicParameters<>{}
      then use_normal:=TRUE else use_normal:=FALSE
  end_if;

  // count sudiagonal bands and store in lowerbands:  
  //  op(lterm(p[i]),[1,2])     = leading exponent    
  //  n+2-op(lterm(p[i]),[1,2]) = lowest index of Variables in p[i] 
  //  Watch out for poly(0,[x]) resulting in tmp=FAIL 

  lowerbands:=0: 
  for i from 1 to neqs do
      tmp:=op(lterm(p[i]),[1,2]);
      if tmp<>FAIL then lowerbands:=max(lowerbands,i-np2+tmp); end_if;
  end_for;

  (triv[j]:=FALSE;) $ j=1..max(n,neqs);
  constraints:={};

  check:=proc(b) // routine to check right hand side of equations 0=b 
  begin b:=subs(b,float(0)=0);
        if iszero(b) then return(null()) end_if;
        if not hastype(b,{DOM_EXPR,DOM_IDENT,"_index"})
        then userinfo(1,"system has no solution [numeric::linsolve]"): 
             return(FAIL):
        end_if;
        if showassumptions=FALSE
        then userinfo(1,"no solution found [numeric::linsolve]"): 
             return(FAIL):
        end_if;
        return(b);
  end_proc;

  PivotAssumptions:= {};
  for j from 1 to n do
    np2mj:=np2-j;
    jpl:=min(j+lowerbands,neqs):
// ------------------- search for pivot element ------------------ 

    // ------- initialization: collect piv[j],..,piv[jpl] -------- 
    pivot:=0;pivindex:=j; minlength:=RD_INF;
    (piv[i]:=0;) $ i=j..max(neqs,n):
    /* store element j from p[i] = i.th row in piv[i] and 
      truncate p[i] to become the remainder of the row */
    for i from j to jpl do
        piv[i]:= coeff(p[i],np2mj); 
        // Make sure, vanishing piv[i] are identified as 0:
        if use_normal and j=1 and not iszero(piv[i])
           then piv[i]:=normal(piv[i]); 
        end_if;
        /* Must cut piv[i] from p[i], even if iszero(piv[i]). Note that
          piv[i]=float(0)<>0 might happen. */
        if piv[i]<>0 then p[i]:=lmonomial(p[i],Rem)[2]: end_if;
    end_for;

    // ------- search for best pivot element----------------- 
    // piv[j],..,piv[jpl] are candidates for the pivot elements.
     
    for i from j to jpl do

        if method=1 then  /*symbolic mode: take the "simplest" nonsymbolic
                           element<>0 as pivot */
           if (not iszero(piv[i])) then
             // look for non-symbolic Pivot elements only 
             if indets(piv[i],PolyExpr)={} then 
               if iszero(p[i])
               then /*remaining row is trivial. This piv[i] is the ideal
                     candidate for a pivot element: accept it! */
                    pivindex:=i; break;   //stop pivot search 
               else /*remaining row is nontrivial. Look for rows
                     with pivot element of smallest complexity */
                    length(piv[i]):
                    if last(1)<minlength 
                       then minlength:=last(1); pivindex:=i:
                    end_if;
               end_if;
             end_if;
           end_if;
        end_if;

        if method=2 then //numeric mode: do row pivoting 
           if not iszero(piv[i]) then
              rownorm:=max(op(map([coeff(p[i])],specfunc::abs))):
              if iszero(rownorm) //remaining row is trivial 
              then pivindex:=i; break;  //stop pivot search 
              else specfunc::abs(piv[i])/rownorm:
                   if pivot<last(1) then pivot:=last(1);pivindex:=i;end_if;
              end_if;          
           end_if;
        end_if;

    end_for;

    /* so far we tried to avoid a symbolic pivot element. If no such
      pivot element exists, we have to start again. This time we must 
      accept symbolic elements. */
    if method=1 and iszero(piv[pivindex]) then
      pivot:=0; pivindex:=j; minlength:=RD_INF;
      for i from j to jpl do
        if not iszero(piv[i]) then
          if iszero(p[i]) then pivindex:=i; break;  
          else length(piv[i]):
               if last(1)<minlength 
                  then minlength:=last(1); pivindex:=i:
               end_if;
          end_if;
        end_if;
      end_for;
    end_if;

// ------- optimal Pivot element found, now exchange rows --------- 

    if j<>pivindex then
      ([p[j],p[pivindex]]):=[p[pivindex],p[j]]:
      ([B[j],B[pivindex]]):=[B[pivindex],B[j]]:
      ([piv[j],piv[pivindex]]):=[piv[pivindex],piv[j]]:
    end_if;

/* ----- if symbolic pivot must be used, this is assumed to be <>0.
        Store this assumption in PivotAssumptions: */

    if method=1 and indets(piv[j],PolyExpr)<>{} and showassumptions=TRUE then
       PivotAssumptions:= PivotAssumptions union {numer(piv[j])};
    end_if;

// ----- check if non-trivial pivot element was found ------- 
    if iszero(piv[j]) 
    then // Entzerrungsschritt: fuege eine weitere Gleichung ein 
         triv[j]:=TRUE; // mark x[j] as free parameter 
         // append equation as additional row 
         neqs:=neqs+1; piv[neqs]:=0; p[neqs]:=p[j]; B[neqs]:=B[j];
         // the number of lower bands was increased 
         lowerbands:=max(lowerbands,neqs-j-1);
         jpl:=min(j+lowerbands,neqs):
    end_if:

// --- Now do the elimination. Column j is stored in piv[j],..,piv[neqs] -- 
    for i from j+1 to jpl do
        if not iszero(piv[i])
           then tmp:=piv[i]/piv[j]:
                case float(tmp)
                of  1.0 do  p[i]:=p[i]-p[j];B[i]:=B[i]-B[j];break;
                of -1.0 do  p[i]:=p[i]+p[j];B[i]:=B[i]+B[j];break;
                otherwise p[i]:=p[i]-mapcoeffs(p[j],_mult,tmp):
                          B[i]:=B[i]-tmp*B[j]:
                end_case;
                if use_normal then 
                         p[i]:=mapcoeffs(p[i],normal):
                         B[i]:=normal(B[i]):
                end_if;
        end_if;
    end_for:
  end_for;
// -- we now have neqs equations for n unknowns, where the first n  
// equations are  upper triangular. 
// Check that the remaining equations are trivial: -- 
  for j from n+1 to neqs do 
      check(B[j]):
      if last(1)=FAIL
      then if showassumptions=FALSE 
              then delete check; return(FAIL)
              else delete check; return([FAIL, [], []])
           end_if;
      else if last(1)<> null() and showassumptions=TRUE then
           constraints:=constraints union {numer(last(1))};
           end_if;
      end_if;
  end_for;
  delete check;
  constraints:=[op(constraints)];

  // -- Now system is in upper triangular form with nonzero diagonal elements 
  // piv[i]. The equations with zero pivot are marked by triv[i]=TRUE      -- 

  // convert univariate polynomials to expressions for fast sparse            
  // backsubstition: coeff*x^(n+2-i) <-> coeff*X[i]. Use expressions to avoid 
  // costly substitutions in the backsubstitution: the kernel simplfies the   
  // known terms automatically (very fast)  

  if triv[n] then X[n]:=vars[n]
             else X[n]:=subs(B[n]/piv[n],float(0)=0);
                  if use_normal then X[n]:=normal(X[n]): end_if:
  end_if;
  for i from n-1 downto 1 do
     if triv[i] then X[i]:=vars[i]
     else
        nTerms:=nterms(p[i]):  // pick out non-trivial coefficients of the poly 
        Xvars:=[X[np2-op(nthterm(p[i],j),[1,2])]$ j=1..nTerms]:
        koeff:=[coeff(p[i])]:
        p[i]:=_plus( koeff[j]*Xvars[j] $ j=1..nTerms):
        if piv[i]=1 
          then X[i]:=subs( (B[i]-p[i])       ,float(0)=0):
          else X[i]:=subs( (B[i]-p[i])/piv[i],float(0)=0):
        end_if;
        if use_normal then X[i]:=normal(X[i]): end_if:
     end_if;
  end_for;
  tmp:=[ (vars[i]=X[i]) $ i=1..n ];
  tmp:= map(tmp, proc()                         // eliminate trivial 
                 begin                          // equations. This   
                   if op(args(),1)=op(args(),2) // version is faster 
                   then null()                  // than building up  
                   else args(1)                 // the solution set  
                   end_if                       // sequentially using
                 end_proc );                    // for i=1..n do     
                                                //   if triv[i] ..   
  constraints:=map(constraints,
                   proc(x) begin 
                    if iszero(x) then null() else x end_if
                   end_proc);
  if showassumptions=FALSE then return(tmp)
  else return([
       tmp, 
        map([op(constraints)],proc() begin args(1)=0 end_proc ),
        map([op(PivotAssumptions)], proc() begin args(1)<>0 end_proc)
              ]);
  end_if;
end_proc:
