// 
// $Date: 2000/05/21 14:56:32 $  $Author: bij $   $Revision: 1.10 $ 
// 
// W.Oevel, 01/04/98 

//------   file  matlinsolve.mu   1.4.98  ------------------------
//-------------------  W. Oevel  ---------------------------------
/*  Help file:

 matlinsolve - solve a system of linear equations numerically

 Call:

 numeric::matlinsolve( A , B <, Symbolic>, <ShowAssumptions> )
                                                                   
 Parameter:

  A        --  an m x n matrix (DOM_ARRAY or category Cat::Matrix)
  B        --  an m x p matrix (DOM_ARRAY or category Cat::Matrix)
               For p=1 the vector B may also be a 1-dimensional array 
               or a list of length m.
  Symbolic --  option
                                                                   
Synopsis:

  numeric::matlinsolve(A,B,<Symbolic>); returns the list [X,Kernelbasis].
  numeric::matlinsolve(A,B,<Symbolic>, ShowAssumptions);
      returns the list [X,Kernelbasis,Constraints, PivotAssumptions].

  (*) X is a special solution of the linear system A*X=B. It is
      an n x p array.
  (*) Kernelbasis represents the general solution of the homogeneous
      linear system A*X=0. It is an n x d array,
      where d is the dimension of the kernel of A. The columns of 
      Kernelbasis form a basis of the kernel. If the kernel is empty,
      then the returned value of Kernelbasis is the null vector,
      represented by array(1..n,1..1,[[0]$n]).
  (*) Constraints is a list of equations for symbolic parameters in A and B,
      which are necessary and sufficient to grant solvability of 
      the linear system A*X=B. If no such contraints arise, then the
      returned value of Constraints is the empty list.
  (*) PivotAssumptions is a list of inequalities [piv1 <>0, piv2<>0, ...].
      These assumptions were implicitly used by matlinsolve to compute
      the solution and the kernel.

  By default all entries in A and B are converted to floats. Non-numeric
  symbolic entries in A are not admissible, unless the option Symbolic
  is used. In the symbolic mode no conversion to floats will occur. We note 
  that the presence of symbolic parameters in the matrix A will slow down 
  the computation drastically.

  Due to roundoff errors some or all basis vectors in the kernel of A may be
  missed in the numeric mode.

  matlinsolve returns [FAIL,NIL], if no solution of A*X=B exists. If B contains
  symbolic parameters, then Constraints is the set of linear equations
  for the entries of B needed to grant solvability.

  matlinsolve uses Gaussian Elimination with row pivoting. Internally
  a sparse representation of the matrices is used.
  
  Examples: 

>> A:=array(1..2,1..2,[[1,2],[3,4]]): B:=[3,7]:
>> numeric::matlinsolve(A,B);

                       -- +-     -+  +-   -+ --
                       |  |  1.0  |  |  0  |  |
                       |  |       |, |     |  |
                       |  |  1.0  |  |  0  |  |
                       -- +-     -+  +-   -+ --

>> numeric::matlinsolve(A,B,Symbolic);
 
                    -- +-   -+  +-   -+ --
                    |  |  1  |  |  0  |  |
                    |  |     |, |     |  |
                    |  |  1  |  |  0  |  |
                    -- +-   -+  +-   -+ --

>> A:=array(1..3,1..3,[ [1,1,0],[0,1,1],[0,0,1]]):
>> B:=array(1..3,1..3,[ [1,0,0],[0,1,0],[0,0,1]]):
>> InverseOfA:=numeric::matlinsolve(A,B,Symbolic)[1];

                              +-           -+
                              |  1, -1,  1  |
                              |             |
                              |  0,  1, -1  |
                              |             |
                              |  0,  0,  1  |
                              +-           -+

>> A:=Dom::Matrix()(3,3,[[2,2,3],[1,1,2],[3,3,5]]):
>> B:=Dom::Matrix()(3,1,[sin(y)^2,cos(y)^2,x]):
>> ([X,Kernel,Constraints,PivAssupmt]):=
       numeric::matlinsolve(A,B,Symbolic,ShowAssumptions);

-- +-                 -+
|  |          2        |  +-    -+
|  |  5 sin(y)  - 3 x  |  |  -1  |
|  |                   |  |      |         2             2
|  |         0         |, |   1  |, [cos(y)  - x + sin(y)  = 0]
|  |                   |  |      |
|  |                2  |  |   0  |
|  |  2 x - 3 sin(y)   |  +-    -+
-- +-                 -+

        --
         |
         |
         |
   , []  |
         |
         |
         |
        --

>> check_the_result:= A*Dom::Matrix()(X)-B,
                      A*Dom::Matrix()(Kernel); 

                  +-                       -+
                  |           0             |  +-   -+
                  |                         |  |  0  |
                  |            2         2  |  |     |
                  |  x - cos(y)  - sin(y)   |, |  0  |
                  |                         |  |     |
                  |           0             |  |  0  |
                  +-                       -+  +-   -+

 See also:  linalg::matlinsolve, linsolve, numeric::linsolve
                                                                  */
//-----------------------------------------------------------------

numeric::matlinsolve:=proc(A,b)
local neqs,n,method,use_normal,i,j,tmp,B,cb,check,
      showassumptions, PivotAssumptions,
      x,X,Xvars,y,nTerms,koeff,np2,np2mj,p,kdim,Kernel,
      piv,pivot,pivindex,rownorm,constraints,
      lowerbands,jpl,minlength;
begin
  if args(0)<2 then error("expecting at least two arguments"); end_if;
  if args(0)>4 then error("expecting at most four arguments"); end_if;
  if has({args()},Symbolic) then method:=1 else method:=2 end_if;
  if has({args()},ShowAssumptions)
     then showassumptions:=TRUE
     else showassumptions:=FALSE
  end_if;
  // convert to domtype DOM_ARRAY for speed 
  if A::dom::hasProp(Cat::Matrix)=TRUE then A:= expr(A) end_if; 
  if domtype(A) <> DOM_ARRAY then
     error("matrix must be an array or of category 'Cat::Matrix'");
  end_if:
  if op(A,[0,1])<>2 then error("argument is not a matrix");end_if;
  neqs:=op(A,[0,2,2]):  // row dimension = number of equations 
  n:=op(A,[0,3,2]):     // col dimension = number of unknowns  

  // convert to domtype DOM_ARRAY for speed 
  if b::dom::hasProp(Cat::Matrix)=TRUE 
  then b:= expr(b); 
       if not contains({1,2},op(b,[0,1]))
          then error("wrong format of vector");end_if;
  end_if;

  if ( domtype(b)=DOM_ARRAY and op(b,[0,1])=1 )
     or domtype(b)=DOM_LIST
     then if neqs<>nops(b) then
             error("dimensions of matrices/vectors not compatible"); 
          end_if;
          b:=array(1..neqs,1..1,[[b[i]]$ i=1..neqs]);
  end_if;

  if not contains({DOM_ARRAY,DOM_LIST},domtype(b))=TRUE 
     then error("2nd argument must be an array, a list, or of category 'Cat::Matrix'");
  end_if:
  if neqs <> op(b,[0,2,2]) then 
     error("dimensions of matrices/vectors not compatible"); 
  end_if;
  cb:=op(b,[0,3,2]); // col dimension of b 

  // force evaluation of A and b (both are arrays now) 
  ((A[i,j]; A[i,j]:= eval(%);) $ i=1..neqs ) $ j=1..n;
  ((b[i,j]; b[i,j]:= eval(%);) $ i=1..neqs ) $ j=1..cb;

  use_normal:=FALSE;  // default 

  if method=2 then
    tmp:=map(A,float):
    B:=map(b,float):
    if {op(map(tmp,domtype))} // union {op(map(B,domtype))} 
        minus {DOM_FLOAT,DOM_COMPLEX,DOM_INT} <> {} then
        warning("symbolic coefficient found, switching to symbolic mode");
        method:=1;B:=NIL;
    else A:=tmp; b:=B; B:=NIL:
    end_if;  
  end_if;
  if method=1 then
    if (indets(A)<>{}) 
    or has({op(map(A,domtype))},DOM_NIL)
    or has({op(map(A,domtype))},DOM_EXPR)
      then use_normal:=TRUE 
    end_if;
  end_if;

  A:=subs(A,float(0)=0); b:=subs(b,float(0)=0);

  if method=1 then
    userinfo(1,"symbolic mode, no pivoting unless necessary");
  end_if;
  if method=2 then userinfo(1,"using column pivoting"); end_if;

  // -- initialization done --- 

  kdim:=0;  // dimension of kernel 
  np2:=n+2;
  lowerbands:=0:       // count sudiagonal bands and store in lowerbands: 
  x:=genident("x"):    // free identifiers as unknown in polynomials to   
  y:=genident("y"):    // avoid conflicts with global variables           
  for i from 1 to neqs do  // convert equations to univariate polynomials 
      tmp:=0:
      for j from 1 to n do                     // x[j]->x^(n+2-j),j=1.. n 
         if not iszero(A[i,j]) then 
                tmp:=tmp+A[i,j]*x^(np2-j): 
                if i-j>lowerbands then lowerbands:=i-j; end_if;
         end_if;
      end_for;
      p[i]:=poly(tmp,[x]):
      if use_normal then p[i]:=mapcoeffs(p[i],normal);end_if;
      tmp:= b[i,1];
      for j from 2 to cb do           // store rows of b as poly in y 
         if not iszero(b[i,j]) then tmp:=tmp+b[i,j]*y^(j-1); end_if;
      end_for;
      B[i]:=poly(tmp,[y]):
  end_for;

  constraints:= {};

  check:=proc(b) // routine to check right hand side of equations 0=b 
  local koeff, tmp;
  begin b:=subs(b,float(0)=0); 
        if iszero(b) then return(null()) end_if;
        if showassumptions=FALSE then return(FAIL); end_if;
        // Now showassumptions = TRUE and iszero(b) = FALSE.
        // Investigabe, whether b has symbolic parameters.
        // If b is numeric, then return FAIL (the system is not solvable),
        // otherwise insert contraint 0=b into constraints.
        koeff:=[coeff(b)]:     // b is a polynomial in y 
        for tmp in koeff do    
          if not iszero(tmp) then
            if hastype(tmp,{DOM_EXPR,DOM_IDENT,"_index"})
               then constraints:=constraints union {numer(tmp)};
               else return(FAIL):
            end_if;
          end_if;
        end_for;
        return(b);
  end_proc;

 //  Conversion of matrices to sparse univariate poly representation done 
 //  Start with Gauss-Elimination 

  PivotAssumptions:= {};
  for j from 1 to n do
    np2mj:=np2-j;
    jpl:=min(j+lowerbands,neqs):
// ------------------- search for pivot element ------------------ 

   // ------- initialization: collect piv[j],..,piv[jpl] -------- 

    pivot:=0;pivindex:=j; minlength:=RD_INF;
    (piv[i]:=0;) $ i=j..max(neqs,n):
    /* store element j from p[i] = i.th row in piv[i] and
      truncate p[i] to become the remainder of the row */
    for i from j to jpl do
        piv[i]:= coeff(p[i],np2mj);
        // Make sure, vanishing piv[i] are identified as 0:
        if use_normal and j=1 and not iszero(piv[i])
           then piv[i]:=normal(piv[i]);
        end_if;
        /* Must cut piv[i] from p[i], even if iszero(piv[i]). Note that
          piv[i]=float(0)<>0 might happen. */
        if piv[i]<>0 then p[i]:=lmonomial(p[i],Rem)[2]: end_if;
    end_for;

    // ------- search for best pivot element----------------- 
    // piv[j],..,piv[jpl] are candidates for the pivot elements.
    for i from j to jpl do

        if method=1 then  /*symbolic mode: take the "simplest" nonsymbolic
                           element<>0 as pivot */
           if (not iszero(piv[i])) then
             // look for non-symbolic Pivot elements only 
             if indets(piv[i],PolyExpr)={} then
               if iszero(p[i])
               then /*remaining row is trivial. This piv[i] is the ideal
                     candidate for a pivot element: accept it! */
                    pivindex:=i; break;   //stop pivot search 
               else /*remaining row is nontrivial. Look for rows
                     with pivot element of smallest complexity */
                    length(piv[i]):
                    if last(1)<minlength
                       then minlength:=last(1); pivindex:=i:
                    end_if;
               end_if;
             end_if;
           end_if;
        end_if;

        if method=2 then //numeric mode: do row pivoting 
           if not iszero(piv[i]) then
              rownorm:=max(op(map([coeff(p[i])],specfunc::abs))):
              if iszero(rownorm) //remaining row is trivial 
              then pivindex:=i; break;  //stop pivot search 
              else specfunc::abs(piv[i])/rownorm:
                   if pivot<last(1) then pivot:=last(1);pivindex:=i;end_if;
              end_if;
           end_if;
        end_if;

    end_for;

    /* so far we tried to avoid a symbolic pivot element. If no such
      pivot element exists, we have to start again. This time we must
      accept symbolic elements. */
    if method=1 and iszero(piv[pivindex]) then
      pivot:=0; pivindex:=j; minlength:=RD_INF;
      for i from j to jpl do
        if not iszero(piv[i]) then
          if iszero(p[i]) then pivindex:=i; break;
          else length(piv[i]):
               if last(1)<minlength
                  then minlength:=last(1); pivindex:=i:
               end_if;
          end_if;
        end_if;
      end_for;
    end_if;

// ------- optimal Pivot element found, now exchange rows --------- 

    if j<>pivindex then
      ([p[j],p[pivindex]]):=[p[pivindex],p[j]]:
      ([B[j],B[pivindex]]):=[B[pivindex],B[j]]:
      ([piv[j],piv[pivindex]]):=[piv[pivindex],piv[j]]:
    end_if;

/* ----- if symbolic pivot must be used, this is assumed to be <>0.
        Store this assumption in PivotAssumptions: */

    if method=1 and indets(piv[j],PolyExpr)<>{} and showassumptions=TRUE then
       PivotAssumptions:= PivotAssumptions union {numer(piv[j])};
    end_if;

// ----- check if non-trivial pivot element was found ------- 
    if iszero(piv[j]) 
    then // Entzerrungsschritt: fuege eine weitere Gleichung ein 
         neqs:=neqs+1; piv[neqs]:=0; p[neqs]:=p[j]: B[neqs]:=B[j];
         kdim:=kdim+1;
         piv[j]:=1:p[j]:=poly(0,[x]): B[j]:=poly(y^(cb-1+kdim),[y]);
         // the number of lower bands was increased 
         lowerbands:=max(lowerbands,neqs-j-1);
         jpl:=min(j+lowerbands,neqs):
    end_if:

// --- Now do the elimination. Column j is stored in piv[j],..,piv[neqs] -- 
    for i from j+1 to jpl do
        if not iszero(piv[i])
           then tmp:=piv[i]/piv[j]:
                case float(tmp)
                of  1.0 do  p[i]:=p[i]-p[j];B[i]:=B[i]-B[j];break;
                of -1.0 do  p[i]:=p[i]+p[j];B[i]:=B[i]+B[j];break;
                otherwise p[i]:=p[i]-mapcoeffs(p[j],_mult,tmp):
                          B[i]:=B[i]-mapcoeffs(B[j],_mult,tmp):
                end_case;
                if use_normal then 
                         p[i]:=mapcoeffs(p[i],normal):
                         B[i]:=mapcoeffs(B[i],normal):
                end_if;
        end_if;
    end_for:
  end_for;

// -- we now have neqs (>=n) equations for n unknowns, where the first n  
// equations are  upper triangular. 
// Check that the remaining equations are trivial: -- 

  for j from n+1 to neqs do
      if check(B[j])=FAIL
      then if showassumptions=FALSE
              then delete check; return([FAIL, NIL])
              else delete check; return([FAIL, NIL, [], []])
           end_if;
      end_if;
  end_for;
  constraints:=[op(constraints)];
  delete check;

// -- Now system is in upper triangular form with diagonal = piv[1],..,piv[n] -- 

  // convert univariate polynomials to expressions for fast sparse backsubstition:  
  // coeff*x^(n+2-i) <-> coeff*x[i] 
  X[n]:=subs(op(B[n],1)/piv[n],float(0)=0);
  if use_normal then X[n]:=normal(X[n]): end_if:
  for i from n-1 downto 1 do
     nTerms:=nterms(p[i]):  // pick out only non-trivial coefficients of the poly 
     Xvars:=[X[np2-op(nthterm(p[i],j),[1,2])]$ j=1..nTerms]:
     koeff:=[coeff(p[i])]:
     p[i]:=_plus( koeff[j]*Xvars[j] $ j=1..nTerms):
     if piv[i]=1 
       then X[i]:=subs( (op(B[i],1)-p[i])       ,float(0)=0):
       else X[i]:=subs( (op(B[i],1)-p[i])/piv[i],float(0)=0):
     end_if;
  end_for;
  // Now X[i] = polynomial expression in y , need to pick out the various 
  // y powers. Reconvert expressions to polys in y for faster access      
  // via coeff. Do not normalize X here, because poly would undo normal   
  for i from 1 to n do X[i]:=poly(X[i],[y]);end_for;

  if kdim<>0   // kernel elements are stored in the high y powers 
  then Kernel:=array(1..n,1..kdim,
              [[coeff(X[i],cb-1+j)$ j=1..kdim]$ i=1..n]);
  else Kernel:= array(1..n,1..1,[[0]$n]) 
  end_if;

  // special solution is stored in the low y powers 
  X:=array(1..n,1..cb,[[coeff(X[i],j-1)$ j=1..cb]$ i=1..n]);

  // That's it. Final simplifications: 
  if use_normal then X:=map(X,normal); Kernel:=map(Kernel,normal); end_if:

  constraints:=map(constraints,
                   proc(x) begin
                    if iszero(x) then null() else x end_if
                   end_proc);
  if showassumptions=FALSE 
  then return([X,Kernel])
  else constraints:=map([op(constraints)],proc() begin args(1)=0 end_proc);
       PivotAssumptions:=map([op(PivotAssumptions)],proc() begin args(1)<>0 end_proc);
       return([X,Kernel,constraints, PivotAssumptions]);
  end_if;
end_proc:
