// 
// $Date: 2000/05/21 14:56:34 $ $Author: bij $ $Revision: 1.5 $ 
// 
// W. Oevel,  20/10/99 

/*
numeric::polysysroots -- numercal roots of system of polynomial equations


Call: 
       numeric::polysysroots(eqs, <vars>);

Parameters: eqs  -- univariate equation (may have symbolic parameter)
                    or set or list of multivariate equations.
                    Also expressions or DOM_POLY polynomials are
                    accepted and interpreted as homogeneous equations.

Options:    vars -- indeterminate or a set or list of indeterminates

Return value: a set of lists {[x1=..,x2=..,..], 
                              [x1=..,x2=..,..],
                              ..}
              of numerical roots

Details: 
        This routine calls the symbolic solver solve, which computes 
        a Groebner basis in triangular form. 
        The last univariate RootOf object of the Groebner basis
        is solved numerically, the results are substituted into the
        other elements of the Groebner basis.

Examples:

>> numeric::polysysroots(x^2=PI),
   numeric::polysysroots(x^2=PI, x),
   numeric::polysysroots(x^2-PI),
   numeric::polysysroots(x^2-PI, x),
   numeric::polysysroots(poly(x^2-PI, [x])),
   numeric::polysysroots(poly(x^2-PI, [x]), x);

{[x = -1.77245385], [x = 1.77245385]},
   {[x = -1.77245385], [x = 1.77245385]},
   {[x = -1.77245385], [x = 1.77245385]},
   {[x = -1.77245385], [x = 1.77245385]},
   {[x = -1.77245385], [x = 1.77245385]},
   {[x = -1.77245385], [x = 1.77245385]}

>> numeric::polysysroots(x^2=exp(y), x),
   numeric::polysysroots(x^2-exp(y), x),
   numeric::polysysroots(poly(x^2-exp(y), [x]), x),
   numeric::polysysroots(x^2=exp(y), [x]),
   numeric::polysysroots(x^2-exp(y), [x]),
   numeric::polysysroots(poly(x^2-exp(y), [x]), [x]);

            1/2                1/2
{[x = exp(y)   ], [x = - exp(y)   ]},
               1/2                1/2
   {[x = exp(y)   ], [x = - exp(y)   ]},
               1/2                1/2
   {[x = exp(y)   ], [x = - exp(y)   ]},
               1/2                1/2
   {[x = exp(y)   ], [x = - exp(y)   ]},
               1/2                1/2
   {[x = exp(y)   ], [x = - exp(y)   ]},
               1/2                1/2
   {[x = exp(y)   ], [x = - exp(y)   ]}

>> numeric::polysysroots(x^2=exp(c), [x,y]),
   numeric::polysysroots(x^2-exp(c), [x,y]),
   numeric::polysysroots(poly(x^2-exp(c), [x]), [x,y]),
   numeric::polysysroots(poly(x^2-exp(c), [x]), [x,y]);

            1/2                1/2
{[x = exp(c)   ], [x = - exp(c)   ]},
               1/2                1/2
   {[x = exp(c)   ], [x = - exp(c)   ]},
               1/2                1/2
   {[x = exp(c)   ], [x = - exp(c)   ]},
               1/2                1/2
   {[x = exp(c)   ], [x = - exp(c)   ]}


>> DIGITS:=3:
   numeric::polysysroots({x^2+y+1, y^2+x+1=2}),
   numeric::polysysroots({x^2+y+1, y^2+x+1=2}, {x,y}),
   numeric::polysysroots({x^2+y+1, y^2+x+1=2}, [y,x]),
   numeric::polysysroots({x^2+y+1, y^2+x+1=2}, [x,y]),
   numeric::polysysroots({x^2+y+1, poly(y^2+x-1)}, [x,y]);

{[y = -1.2, x = -0.453],
   [y = -1.0, x = 0.0],
   [y = 1.1 - 0.665 I, x = 0.226 + 1.46 I],
   [y = 1.1 + 0.665 I, x = 0.226 - 1.46 I]},
{[y = -1.2, x = -0.453],
   [y = -1.0, x = 0.0], [y = 1.1 - 0.665 I, x = 0.226 + 1.46 I],
   [y = 1.1 + 0.665 I, x = 0.226 - 1.46 I]},
{[y = -1.2, x = -0.453],
   [y = -1.0, x = 0.0], [y = 1.1 - 0.665 I, x = 0.226 + 1.46 I],
   [y = 1.1 + 0.665 I, x = 0.226 - 1.46 I]},
{[x = 0.0, y = -1.0],
   [x = - 0.453 + 0.0 I, y = - 1.2 + 1.16e-10 I],
   [x = 0.226 + 1.46 I, y = 1.1 - 0.665 I],
   [x = 0.226 - 1.46 I, y = 1.1 + 0.665 I]},
{[x = 0.0, y = -1.0],
   [x = - 0.453 + 0.0 I, y = - 1.2 + 1.16e-10 I],
   [x = 0.226 + 1.46 I, y = 1.1 - 0.665 I],
   [x = 0.226 - 1.46 I, y = 1.1 + 0.665 I]}

>> DIGITS:=2:
   numeric::polysysroots({x^2+y+exp(z), y+x+1=exp(z)}, {x,y}),
   numeric::polysysroots({x^2+y+exp(z), y+x+1=exp(z)}, [y,x]),
   numeric::polysysroots({x^2+y+exp(z), y+x+1=exp(z)}, [x,y]),
   numeric::polysysroots({x^2+y+exp(z), poly(y+x+1-exp(z))}, [x,y]);

{[y = exp(z) - 0.5 (5.0 - 8.0 exp(z))    - 1.5,
                             1/2
   x = 0.5 (5.0 - 8.0 exp(z))    + 0.5],
                                      1/2
 [ y = exp(z) + 0.5 (5.0 - 8.0 exp(z))    - 1.5,
                                   1/2
   x = 0.5 - 0.5 (5.0 - 8.0 exp(z))   ]}, ...
}

>> DIGITS:=2: z:=1.0:
   numeric::polysysroots({x^2+y+exp(z), y+x+1=exp(z)}, {x,y}),
   numeric::polysysroots({x^2+y+exp(z), y+x+1=exp(z)}, [y,x]),
   numeric::polysysroots({x^2+y+exp(z), y+x+1=exp(z)}, [x,y]),
   numeric::polysysroots({x^2+y+exp(z), poly(y+x+1-exp(z))}, [x,y]);

  {[y = 1.2 - 2.0 I, x = 0.5 + 2.0 I],
   [y = 1.2 + 2.0 I, x = 0.5 - 2.0 I]},
  {[y = 1.2 - 2.0 I, x = 0.5 + 2.0 I],
   [y = 1.2 + 2.0 I, x = 0.5 - 2.0 I]},
  {[x = 0.5 - 2.0 I, y = 1.2 + 2.0 I],
   [x = 0.5 + 2.0 I, y = 1.2 - 2.0 I]},
  {[x = 0.5 - 2.0 I, y = 1.2 + 2.0 I],
   [x = 0.5 + 2.0 I, y = 1.2 - 2.0 I]}

*/

numeric::polysysroots:= proc(eqs)
   local s, vars, uvars;
 begin
   userinfo(10, "numeric::polysysroots called with args ".expr2text(args()));
   if args(0)<1 then error("expecting at least one argument") end_if;
   if args(0)>2 then error("expecting at most two arguments") end_if;
   s:= eqs; // copy of original form of equations
   // convert equation(s) to list of equation(s)
   if domtype(eqs)=DOM_SET  then eqs:= [op(eqs)]; end_if;
   if domtype(eqs)<>DOM_LIST then eqs:= [eqs]; end_if;
   // convert equations to expressions:
   eqs:= subs(eqs, float(0)=0);
   eqs:= map(eqs, proc(eq) begin if type(eq)= "_equal" then 
                  op(eq,1)-op(eq,2) else eq end_if; end_proc);
   // eliminate trivial equations
   eqs:= select(eqs, _not@iszero);

   // convert unknowns to list of unknowns:
   if args(0)=2
   then vars:= args(2):
        uvars:= vars;
        if domtype(vars)=DOM_SET   then vars:= [op(vars)]; end_if;
        if domtype(vars)<>DOM_LIST then vars:= [vars]; end_if;
   else vars:= [op(numeric::indets(eqs))];
   end_if;
   //handle trivial equations in the same way as solve:
   if nops(eqs)=0 then return(solve(s, uvars)); end_if;
   if vars=[] then error("no indeterminate(s)"); end_if;

   //check that system is polynomial w.r.t. vars
   if has(map(eqs,testtype,Type::PolyExpr(vars)), FALSE) then
     if args(0)=2
       then error("equation(s) are not polynomial in specified indeterminates")
       else error("equation(s) not polynomial")
      end_if;
   end_if;

   // Only need to call solvelib::allvalues:
   s:=op(solvelib::allvalues(solve(eqs, vars, BackSubstitution = FALSE)));
 
   if has([s], FAIL) 
      then return({})
      else // clean up round-off garbage via complexRound
          return(map({s}, map, map, numeric::complexRound));
   end_if;
end_proc:

// end of file 
