/*
  numeric::sort -- sort a list of numerical values

Call:  numeric::sort(list_of_values)

Input is a list of real or complex values,
output is a sorted list of float values.

Sorting criterion:
  1) From small real parts to large real parts
  2) If tie, then from small abs(imaginary part) 
     to large abs(imaginary part)
  3) If tie (i.e., a complex conjugate pair)
     then positive imaginary part first

This routine is used to sort the output of
  numeric::eigenvalues,
  //numeric::eigenvectors,
  numeric::polyroots,
  numeric::polyroots2,
  numeric::singularvalues
  //numeric::singularvectors


Examples:

  >> numeric::sort([PI, 0, PI-I, PI+I, 2, sqrt(2)]);

     [0.0, 1.414213562, 2.0, 3.141592653, 3.141592653 + 1.0 I,
         3.141592653 - 1.0 I]

  >> numeric::sort([-2.0, 2+10*I, 2-12*I, 0.3*I, 0.3]);

            [-2.0, 0.3 I, 0.3, 2.0 + 10.0 I, 2.0 - 12.0 I]
*/

numeric::sort:= proc(d) begin
    if args(0)<>1 then error("expecting one argument") end_if;
    if domtype(d)<> DOM_LIST then error("expecting a list") end_if;
    // convert to float
    d:= map(d, float);
    // sort real floats by kernel sort, watch out for future float(0)=0
    if {op(map(d,domtype))} minus {DOM_FLOAT, DOM_INT} = {}
       then return(sort(d));
    end_if;
    // check for symbolic objects
    if {op(map(d,domtype))} minus 
       {DOM_INT, DOM_RAT, DOM_FLOAT, DOM_COMPLEX} <> {}
       then error("cannot sort symbolic objects") 
    end_if;
    // final sorting
    sort(d, proc(a,b) local absa, absb; begin
            if domtype(a)=DOM_COMPLEX or
               domtype(b)=DOM_COMPLEX 
            then if op(a,1)<op(b,1) then return(TRUE) end_if;
                 if op(a,1)=op(b,1) then
                    absa:= specfunc::abs(a);
                    absb:= specfunc::abs(b);
                    if absa < absb then return(TRUE) end_if:
                    if absa = absb and Im(a)>Im(b) 
                       then return(TRUE)
                    end_if:
                 end_if;
            else if a<b then return(TRUE) end_if;
            end_if;
            return(FALSE)
         end_proc);
end_proc:

