/*===========================================================================

numlib::contfrac -- continued fractions arithmetic

contfrac(e <,p>)	computes a continued fraction for e (with precision p)

contfrac::approx(cf)		gives an interval that contains cf

contfrac::unapprox(a,b)	gives a continued fraction for the interval [a,b]

contfrac::linear(x,a,b,c,d)	computes (a*x+b)/(c*x+d) where x is a continued fraction

a + b			adds two continued fractions

a * b			multiplies two continued fractions

1/a			inverts a continued fraction

a^n			computed the (integer) power of a continued fraction

Convention: "..." represents a real between 1 and infinity, thus
            for example the continued fraction of 0 is 1/..., that
            of an integer n is n+1/...

Examples:

>> export( numlib,contfrac ):
>> a := contfrac(PI,5);

                                  1               
                        ---------------------- + 3
                                1                 
                        ------------------ + 7    
                              1                   
                        ------------- + 15        
                            1                     
                        --------- + 1             
                         1                        
                        --- + 292                 
                        ...                       

>> b := contfrac(sqrt(5),4);

                                   1             
                          ------------------- + 2
                                 1               
                          --------------- + 4    
                               1                 
                          ----------- + 4        
                             1                   
                          ------- + 4            
                           1                     
                          --- + 4                
                          ...                    

>> contfrac::approx(a);

                        103993/33102, 104348/33215


>> contfrac::unapprox(%); 

                                  1            
                           ---------------- + 3
                                1              
                           ------------ + 7    
                              1                
                           ------- + 15        
                            1                  
                           --- + 1             
                           ...                 

(we lost some precision, which is normal, because ... does not
 have a precise value)

Now we get a CF for the gold-ratio (1+sqrt(5))2 from that of sqrt(5),
via the fractional linear form L(x) = (x+1)/2:

>> contfrac::linear(b,1,1,0,2);

                                   1                           
            ----------------------------------------------- + 1
                                 1                             
            ------------------------------------------- + 1    
                               1                               
            --------------------------------------- + 1        
                             1                                 
            ----------------------------------- + 1            
                           1                                   
            ------------------------------- + 1                
                         1                                     
            --------------------------- + 1                    
                       1                                       
            ----------------------- + 1                        
                     1                                         
            ------------------- + 1                            
                   1                                           
            --------------- + 1                                
                 1                                             
            ----------- + 1                                    
               1                                               
            ------- + 1                                        
             1                                                 
            --- + 1                                            
            ...                                                

>> a+b; CF for PI+sqrt(5)

                                   1               
                        ----------------------- + 5
                                 1                 
                        ------------------- + 2    
                               1                   
                        --------------- + 1        
                             1                     
                        ----------- + 1            
                           1                       
                        ------- + 1                
                         1                         
                        --- + 5                    
                        ...                        

>> a*b; CF for PI*sqrt(5)

                                  1          
                             ------------ + 7
                                1            
                             ------- + 40    
                              1              
                             --- + 3         
                             ...             

>> 1/a; CF for 1/PI

                                    1             
                        --------------------------
                                  1               
                        ---------------------- + 3
                                1                 
                        ------------------ + 7    
                              1                   
                        ------------- + 15        
                            1                     
                        --------- + 1             
                         1                        
                        --- + 292                 
                        ...                       

>> a^3; CF for PI^3

                                  1           
                            ------------- + 31
                               1              
                            ------- + 159     
                             1                
                            --- + 3           
                            ...               

==========================================================================*/

alias(contfrac = numlib::contfrac):

contfrac:=newDomain("numlib::contfrac"):
contfrac::info:= "domain 'contfrac': arithmetic with continued fractions":
contfrac::new:=
proc(e)
  local l,n,pn,qn,pnold,qnold,hi,tmp;
  save DIGITS;
begin
  if args(0)>=2 then
    DIGITS:=args(2)
  end_if;
  if not contains({DOM_FLOAT,DOM_RAT,DOM_INT},type(e)) then
    e:=float(e);
    if type(e)<>DOM_FLOAT then
      return(contfrac::function(args()))
    end_if
  end_if;
  pnold:=1; qnold:=0;
  n:=trunc(e);
  pn:=n; qn:=1;
  l:=[n];
  hi := float(10^(DIGITS/2));
  while qn<hi do
    if iszero(e-n) then break end_if; // for rationals 
    e:=1/(e-n);
//  replaced the following to make sure that next n>0
//  (note that exact new_e = 1/(old_e-trunc(old_e)) > 1 !) Walter
//  if e<0 then // wrong sign due to loss of precision, stop here
    if e<1 then // wrong sign due to loss of precision, stop here
      break
    end_if;
    n:=trunc(e);
    l:=append(l,n);
    tmp:=n*pn+pnold; pnold:=pn; pn:=tmp;
    tmp:=n*qn+qnold; qnold:=qn; qn:=tmp;
  end_while;
  new(contfrac,l)
end_proc:

// continued fraction of a function, for example e = sin(z) 
contfrac::function :=
proc(e)
  local z,ord;
begin
  z:=indets(e);
  if nops(z)=1 then
    z:=op(z)
  else
    error("too many variables")
  end_if;
  if args(0)>=2 then
    ord:=args(2)
  else
    ord:=null()
  end_if;
  e:=series(e,z,ord);
  Series::Puiseux::contfrac(e)
end_proc:

contfrac::approx:=
proc(cf, n) // gives the nth rational approximation 
  local l,r,s,i;
begin
  l:=extop(cf,1);
  if args(0)=1 or n>nops(l) then
    n:=nops(l)
  end_if;
  if l=[] then
    0, 0
  else
    r:=op(l,n);
    s:=op(l,n)+1;
    for i from n-1 downto 1 do
      r:=op(l,i)+1/r;
      s:=op(l,i)+1/s;
    end_for;
    min(r,s),max(r,s)
  end_if
end_proc:


contfrac::print:=
proc(cf) 
  local l,res,i;
begin
  l:=extop(cf,1);
   // vordefiniertes Standard-Domain fuer spezielle Ausgaben
  res:=stdlib::Exposed("...");
  for i from nops(l) downto 1 do
    res:=op(l,i)+1/res
  end_for
end_proc:

contfrac::linear:=proc(x,a,b,c,d) // compute (ax+b)/(cx+d) [Zippel, page 34] 
local l,q,t,m,extract;
begin
   extract:=proc() local p,t; begin 
      while c<>0 do
         p:=trunc((a+b)/(c+d));
         if p<>trunc(a/c) then break end_if;
         m:=append(m,p);
         t:=a; a:=c; c:=t-p*c;
         t:=b; b:=d; d:=t-p*d;
      end_while
   end_proc;
   l:=extop(x,1); m:=[]; // for result 
   for q in l do // first try to extract some partial quotients
      extract();
      t:=a; a:=a*q+b; b:=t;
      t:=c; c:=c*q+d; d:=t;
   end_for;
   extract();
   new(contfrac,m)
 end_proc:

// from Zippel, page 38 
// computes the linear form (a*x*y+b*x+c*y+d)/(e*x*y+f*x+g*y+h) 
contfrac::linear2 := proc(x,y,a,b,c,d,e,f,g,h)
local lx,ly,m,extract,tmp,maxi,s,r;
begin
   maxi:=proc(a,b) begin 
      if b=0 then if a=0 then -1 else FAIL end_if else trunc(a/b) end_if
   end_proc;
   extract:=proc() local t,hi; begin
      while TRUE do
         // the minimum is 
         t:=trunc((a+b+c+d)/(e+f+g+h));
         hi:=maxi(a,e);
         if hi=FAIL then break end_if;
         // if a=e=0 then the form is (b*x+c*y+d)/(f*x+g*y+h) 
         // and the maximum is either b/f or c/g 
         hi:=max(maxi(b,f),maxi(c,g));
         if has(hi,FAIL) then break end_if;  
         if t<>hi then break end_if;
         m:=append(m,t);
         tmp:=a; a:=e; e:=tmp-t*e;
         tmp:=b; b:=f; f:=tmp-t*f;
         tmp:=c; c:=g; g:=tmp-t*g;
         tmp:=d; d:=h; h:=tmp-d*h;
      end_while
   end_proc;
   lx:=extop(x,1);
   ly:=extop(y,1);
   m:=[];
   for r in lx do
      // first try to extract: the linear form being monotonic, 
      // the extremes are obtained for x=y=1 and for x=y=infinity 
      // that is (a+b+c+d)/(e+f+g+h) and a/e 
      extract();
      // then absorb a partial quotient from x 
      tmp:=a; a:=r*a+c; c:=tmp;
      tmp:=b; b:=r*b+d; d:=tmp;
      tmp:=e; e:=r*e+g; g:=tmp;
      tmp:=f; f:=r*f+h; h:=tmp;
   end_for;
   for s in ly do
      extract();
      tmp:=a; a:=s*a+b; b:=tmp;
      tmp:=c; c:=s*c+d; d:=tmp;
      tmp:=e; e:=s*e+f; f:=tmp;
      tmp:=g; g:=s*g+h; h:=tmp;
   end_for;
   extract();
   new(contfrac,m)
end_proc:

contfrac::unapprox:=
proc(a,b) // find the contfrac x \in [a,b] (we can have b<a) 
  local m,t;
begin
  m:=[];
  repeat
    if (t:=trunc(a))<>trunc(b) then
      break
    end_if;
    m:=append(m,t);
    if iszero(a-t) or iszero(b-t) then
      break
    end_if;
    a:=1/(a-t); b:=1/(b-t);
  until FALSE end_repeat;
  new(contfrac,m,1)
end_proc:

contfrac::_plus:=proc(x,y) 
local l;
begin 
   // contfrac::linear2(x,y,0,1,1,0,0,0,0,1) 
   l:=zip([contfrac::approx(x)],[contfrac::approx(y)],_plus);
   contfrac::unapprox(op(l))
end_proc:

contfrac::_mult:=proc(x,y) 
local l;
begin 
   l:=zip([contfrac::approx(x)],[contfrac::approx(y)],_mult);
   contfrac::unapprox(op(l))
end_proc:

contfrac::_invert:=proc(x) local l;
begin
   l:=extop(x,1);
   if op(l,1)=0 then l:=subsop(l,1=null()) else l:=[0,op(l)] end_if;
   /* extsubsop(cf,1=l) */ // does not work 
   new(contfrac,l)
end_proc:

contfrac::_power:=proc(x,n) local y;
begin
   if type(n)<>DOM_INT then FAIL
   elif n<0 then contfrac::_invert(x^(-n))
   elif n=0 then contfrac(1)
   else
      y:=x;
      while n>1 do y:=y*x; n:=n-1 end_while;
      y
   end_if
end_proc:

unalias(contfrac):

// end of file 
