/*++ ---------------- repDiff.mu ---------------------

Author: Winfried Fakler,  University of Paderborn
Date Created: 18 July  2000
Last Update : 19 July  2000

Description:
This file contains functions for finding power series solutions of ordinary
differential equations with initial conditions by the method of repeated
differentiating.

Functions:

 - used Parameter:

 - ode::repdiff
 - ode::repeatedDifferentiating

See:

Example:
ode::repdiff(diff(y(x),x)+sin(x)/x,y,x,[0,0],6)
ode::repdiff(diff(y(x),x)^2+sin(x)/x,y,x,[0,0],6)
ode::repdiff(sqrt(diff(y(x),x))+sin(x)/x,y,x,[0,0],6)
ode::repdiff(diff(y(x),x,x)+sin(x)/x,y,x,[0,1,2],6)
ode::repdiff(diff(y(x),x)-y(x)^2-x^2,y,x,[0,1],6)
ode::repdiff(diff(y(x),x)^2-y(x)^2-x^2,y,x,[0,1],6)
ode::repdiff(diff(y(x),x,x)^(1/2)+diff(y(x),x)-sin(x)/x,y,x,[0,1,2],6)
ode::repeatedDifferentiating(ode({diff(y(x),x,x)^(1/2)-sin(x)/x,
      y(0)=1,D(y)(0)=2},y(x)),x=0,7)
ode::repeatedDifferentiating(ode(diff(y(x),x)+sin(x)/x,y(x)),x=a)
++*/

ode::repeatedDifferentiating:=
  proc(aa)  // ode(..),x=a,<order>  oder: eq,y(x),<inits>,x=a,<order>
           // ode(..),x,<order>    oder: eq,y(x),<inits>,x,<order>
    local components,icset,eq,i,Lhs,a,f,icl,dmax,e;
  begin
    components:=ode::getComponents(args());
//print("first components ",components);
    if domtype(aa)=ode then
      if args(0)>1 and domtype(components[5])=DOM_SET then
        error("Illegal arguments")
      elif args(0)=1 then
        error("Expansion point is missing")
      end_if
    elif nops(components)>4 and domtype(components[5])=DOM_SET then
      components[5]:=map(components[5],ode::makeInitialCondition,components[2],
                         components[3]); // minus {FAIL};
      if contains(components[5],FAIL) then
        error("Invalid initial conditions");
      end_if;
      delete components[4];
    end_if;
//print("second components ",components);
    // now: components = sys, depVars, indepVar, inits, <x < =a >> , <order>
    if domtype(components[1])=DOM_SET or nops(components[2])>1 then
      error("Only scalar equations are treated");
    end_if;
    if nops(components)=4 then
      error("Expansion point is missing");
    elif domtype(components[5])=DOM_IDENT and components[5]=components[3] then
      components[5]:=components[5]=0;
    elif testtype(components[5],Type::Equation(Type::Unknown,
                                               Type::Arithmetical))=TRUE 
      and lhs(components[5])=components[3] then
      // its ok
    else
      error("Expansion point is missing")
    end_if;
    if nops(components)>5 and testtype(components[6],Type::PosInt)=FALSE then
      error("Order must be a positive integer");
    else
      components:=components,ORDER;
    end_if;
//print("final components ",components);
    // now: components = eq, y, x, inits, x=a, order
    // it remains to test the initial conditions:
    icset:={};
    a:=rhs(components[5]);
    f:=components[2];
    for eq in components[4] do
      Lhs:=op(eq,1);
      i:=0;
      if a<>op(Lhs) then
        error("Initial conditions and expansion point do not going together")
      end_if;
      Lhs:=op(Lhs,0); // This should be f or D(...(D(f))...)
      case type(Lhs)
        of DOM_IDENT do break;
        of "D" do
          repeat
            Lhs:=op(Lhs);
            i:=i+1;
          until type(Lhs)<>"D"
          end_repeat;
          break;
        otherwise error("Invalid initial conditions")
      end_case;
      if f<>Lhs then
        error("Invalid initial conditions");
      end_if;
      icset:=icset union {[i,op(eq,2)]}; // [i,b] means (D@@i)(f)(a)=b
    end_for;
    // make appropriate initial condition list for ode::repdiff
    if icset<>{} then
      dmax:=max(op(map(icset,e->op(e,1))));
      icl:=[a,(D@@i)(f)(a) $i=0..dmax];
      (icl[op(e,1)+2]:=op(e,2)) $ e in icset;
    else
      icl:=[a];
    end_if;
    ode::repdiff(components[1],f,components[3],icl,components[6]);
  end_proc:

ode::repdiff:=
  proc(eq,y,x,ivs=[],order=ORDER)
    local n,f,s,e,res,r,r0,i,sl0;
  begin
    // determine order of eq
    n:=ode::getOrder(eq,y(x));
    // generate missing initial values
    if ivs=[] then
      ivs:=[0,(D@@i)(y)(0) $i=0..n-1];
    elif nops(ivs)<n+1 then
      ivs:=ivs.[(D@@i)(y)(ivs[1]) $ i=nops(ivs)-1..n-1];
    end_if;
    // handle trivial cases (first stage)
    if n>=order then 
      return({taylor(_plus(ivs[i+2]/fact(i)*(x-ivs[1])^i $i=0..order-1),
                     x=ivs[1],order)});
    elif n=-infinity then
      return({taylor(0,x=ivs[1],order)});
    end_if;
    // resolve equation
    s:=solvelib::discreteSolve(eq,{diff(y(x),x$n)});
    if domtype(s)<>DOM_SET or nops(s)=0 then
      error("Cannot resolve equation for ".expr2text(diff(y(x),x$n)));
    end_if;
    // handle trivial cases (second stage)
    if n=0 then return(map(s,e->taylor(op(op(e),2),x=ivs[1],order))) end_if;
    // step through any resolution
    res:={};
    r0:=_plus(ivs[i+2]/fact(i)*(x-ivs[1])^i $i=0..n-1);
    sl0:=[x=ivs[1],diff(y(x),x$i)=ivs[i+2] $i=0..n-1];
    for e in s do
      f:=op(op(e),2); //  y^(n)=f(x,y,y'',..y^(n-1))
      if traperror((r:=r0+eval(subs(f,sl0))/fact(n)*(x-ivs[1])^n))<>0 then
        // expansion around singular point
        if ode::getOrder(f,y(x))<0 then
          // treat special cases
          case n
            of 1 do return({series(int(f,x),x=ivs[1],order)+ivs[2]}); break;
            otherwise
              for i from 1 to n do
                f:=expr(series(int(f,x),x=ivs[1],order)+ivs[n-i+2]);
              end_for;
              return({taylor(f,x=ivs[1],order)});
          end_case;
        else
          error("Only expansions around regular points are treated.")
        end_if;
      end_if;
      // repeated differentiation
      for i from n+1 to order-1 do
        f:=eval(subs(diff(f,x),e));
        r:=r+eval(subs(f,sl0))/fact(i)*(x-ivs[1])^i;
      end_for;
      res:=res union {taylor(r,x=ivs[1],order)};
    end_for;
    res;
  end_proc:


    