/*++ ---------------- wronsk.mu ---------------------

Author: Winfried Fakler,  University of Karlsruhe
Date Created: 13 September 1996
Last Update : 21 December  1999

Description:
This file contains functions for computing the Wronskian matrix and 
determinant of given functions or of a given linear differential equation.
In particular, this functions can be used to create linear homogeneous
differential equations.

Functions: 

 - used Parameter:
    ++ Ly, y(x)   :DOM_EXPR
    ++ v_i        :DOM_EXPR
    ++ x,y        :DOM_IDENT
    ++ R          :DOM_DOMAIN
    ++ Ly    = ordinary linear differential equation 
    ++ y(x)  = the operator (function) of Ly  
    ++ y     = name of the operator of Ly
    ++ x     = independant variable (of the operator y)
    ++ n     = order of Ly
    ++ v_i   = function in x 
    ++ R     = differential ring

 - ode::wronskianMatrix([v_i], x, < R >)
    ++ returns the Wronskian matrix of the functions v_i  w.r.t. x.
    ++ When the option R is given, then the specified differential ring will
    ++ be choosed for reprensenting the matrix entries, otherwise 
    ++ Dom::ExpressionField(id, iszero@normal) will be used instead. 
 - ode::wronskian([v_i], x, < R >)
   ode::wronskian(Ly, y(x))
    ++ returns the Wronskian (determinant) of the functions v_i  w.r.t. x or
    ++ the Wronskian of the equation Ly.
 - ode::wronskianOf(Ly,y,x,n)
    ++ returns the wronskian of Ly, which is exp(-int(a_(n-1)/a_n, x)) if
    ++ a_n and a_(n-1) are the coefficients of diff(y(x),x$n) and
    ++ diff(y(x),x$(n-1)).
    ++ NOTE: no check of arguments!                                       

>> ode::wronskian([sqrt(x),1/sqrt(x),y(x)],x);      

                   diff(y(x), x, x)   y(x)   diff(y(x), x)
                 - ---------------- + ---- - -------------
                          x              3         2
                                      4 x         x

>> ode::wronskianMatrix([sqrt(x),1/sqrt(x),y(x)],x);                    

                +-                                      -+
                |     1/2       1                        |
                |    x   ,     ----,         y(x)        |
                |               1/2                      |
                |              x                         |
                |                                        |
                |     1          1                       |
                |   ------,  - ------,   diff(y(x), x)   |
                |      1/2        3/2                    |
                |   2 x        2 x                       |
                |                                        |
                |      1        3                        |
                |  - ------,  ------,  diff(y(x), x, x)  |
                |       3/2      5/2                     |
                |    4 x      4 x                        |
                +-                                      -+


++*/

//Dom::ExpressionField: // to avoid warnings about uninitialized values

ode::wronskianMatrix:= proc(vis, x, R=Dom::ExpressionField(id, iszero@normal))
  //option: R 
  local i, mat;
  begin
    if R::hasProp(Dom::BaseDomain) = FAIL 
       or not(R::hasProp(Cat::CommutativeRing) and
              R::hasProp(Cat::DifferentialRing) and
              iszero( R::characteristic() )) 
       then error("illegal coefficient ring") end_if;
    mat := [map(vis, R::new)];
    for i from 1 to nops(vis)-1 do
      mat := append(mat, map(mat[i], diff, x)) 
    end_for;
    (Dom::Matrix(R))::create(mat)
  end_proc:

ode::wronskian:=
  proc()
    local eq;
  begin
    if domtype(args(1))=DOM_LIST then
      expr(linalg::det(ode::wronskianMatrix( args() )));
    elif args(0)=2 then
        eq:=ode::isLODE(args(),Lode);
        if eq<>FALSE then
          ode::wronskianOf(eq);
        else
          error("Not a linear ordinary differential equation");
        end_if;
    else
      error("wrong no of args");
    end_if;
  end_proc:

ode::wronskianOf:=
  proc(eq,y,x,n)
  begin
    normal(expand(exp(-int(coeff(eq,[diff(y(x),x$(n-1))],1)/
                           coeff(eq,[diff(y(x),x$n)],1),x))
                  )
           )
  end_proc:
