/*  ode::autonomousEq(eq,y,x,n) solves the autonomous equation 
   eq=0 of order n in y(x),
   using the algorithm described in Zwillinger, section 36 page 153.

   Output: FAIL or a list of solutions for y(x).

   Example: 
   >> ode::autonomousEq(diff(y(x),x,x)-diff(y(x),x)-2*y(x)*diff(y(x),x),y,x,2);
*/


ode::autonomous:=
  proc()
    local components;
  begin
    components:=ode::getComponents(args());
    if nops(components[2])>1 then
      error("autonomous method works only for scalar equations.")
    end_if;
    components[4]:=ode::order(components[1],{components[2]});
    ode::autonomousEq(components);
  end_proc:
    
ode::autonomousEq :=
 proc(eq,y,x,n) 
   local i,u,solu,soly,s,sol;
 begin
   userinfo(2,"trying to recognize an autonomous equation");
   u:=genident("u");
   eq:=eval(subs(eq,diff(y(x),x$i)=ode::auto_transform(y,u,i) $i=1..n,y(x)=y));
   if has(eq,x) then FAIL 
   else 
      userinfo(1,"autonomous equation");
      sysassign(ode::depth,ode::depth+1);
      solu:=ode::solve_eq(eq,u,y);
      sysassign(ode::depth,ode::depth-1);
      if type(solu)="solve" or has(solu,FAIL) then
         userinfo(1,"autonomous method failed");
         return(FAIL)
      end_if;
      userinfo(2,"solutions for ".expr2text(u)."(y) are",solu);
      // now each solution corresponds to u that is y'(x) 
      s:=eval(subs(eq,u(y)=0));
      /* if s=0, then u(y) is a factor of eq, therefore u=0 (y=C) is 
        already in solu */
      if s=0 then soly:={} 
      else soly:=solvelib::discreteSolve(s,y); //,MaxDegree=1);
        if calledFromWithinDSolve = TRUE and not ode::isResultOfSolveOK(soly)
          then userinfo(4,"autonomous method worked, but only get the ".
                        "implicit solution ".expr2text(soly)." and skip");
        end_if;
      end_if;
      if type(solu)="_union" then 
        solu:=map(solu,()->(if type(args(1))=DOM_SET
	                       then args(1) else {args(1)} end_if)) 
      end_if;
      for sol in solu do
         if sol=0 then s:={genident("C")} // y'=0 
         elif sol=FAIL then s:={}
         else
            userinfo(2,"integrating ".expr2text(1/sol)." in ".expr2text(y));
            // level needed for u=(y^2 + 1)^(3/2) for instance 
            s:=int(1/sol,y); //s:=level(s,2);
            s:=solvelib::discreteSolve(s=x+genident("C"),y);//,PrincipalValue);
                                                      //MaxDegree=1);
            userinfo(2,"solution is",s);
         end_if;
         //if s<>FAIL then soly:=soly union s
         if (calledFromWithinDSolve = TRUE and ode::isResultOfSolveOK(s))
           or (calledFromWithinDSolve <> TRUE and
               ode::isResultOfSolveOK(s,"AllowImplicit"))
           then soly:=soly union s
         else
           userinfo(10,
                    "problems occured in autonomous, solution(s) will be lost")
         end_if;
      end_for;
      if soly<>{} then
        userinfo(1,"autonomous method worked");
        soly;
      else FAIL
      end_if
   end_if
end_proc:

// gives the value of diff(y(x),x$n) in terms of u(y)=diff(y(x),x) 
ode::auto_transform:=proc(y,u,n) local e;
begin
   e:=u(y);
   while n>1 do e:=expand(diff(e,y)*u(y)); n:=n-1 end_while;
   e
end_proc:
