//  $Date: 2000/11/28 18:53:00 $ $Author: fakler $ $Revision: 1.9 $ 

/*++ ---------------- new.mu ---------------------

Authors: Winfried Fakler, University of Paderborn
         Paul Zimmermann, INRIA Lorraine

Date Created: 199?
Last Update : 14 April 2000


Description:
This file contains functions for creating and handling objects of type ode.
A valid differential equation must at least contain one given unknown 
differential function.

Functions:
                               
 - used Parameter:
    ++ a        :this domain 
    ++ eq,unk   :DOM_EXPR
    ++ sys,unks :DOM_SET
    ++ indepVar :DOM_IDENT
    ++ depVars  :DOM_IDENT or DOM_SET
    ++ a        = an element of domain ode
    ++ eq       = ordinary differential equation in unk
    ++ unk      = unknown function
    ++ sys      = set of ordinary differential equations together with initial
    ++            conditions              
    ++ unks     = set of unknown functions in one independent variable
    ++ indepVar = independent variable
    ++ depVars  = dependent variable or set of dependent variables

 - ode::new(eq,unk)
   ode::new(sys,unks)
    ++ creates from an ordinary differential equation eq in unk or a system
    ++ of ordinary differential equations in the unknowns unks an element of
    ++ the domain ode.
    ++ Currently the internal representation is a sequence of
             - set of differential expressions (representing the equations)
               or a differential expression
             - set of dependent variables or dependent variable
             - independent variable
             - set of initial conditions
 - ode::getComponents(a,..)
   ode::getComponents(sys,unks,..)
    ++ returns a sequence of the internal representation together with
    ++  all remaining args
    ++ for the domain element a. It is assumed that a is a valid object of
    ++ type ode. 
 - ode::makeEquation(eq,depVars,indepVar)
    ++ returns FAIL or a valid differential expression representing a 
    ++ differtial equation in at least one depVars[i](indepVar).
 - ode::makeInitialCondition(eq,depVars,indepVar)
    ++ returns FAIL or a (not proper checked!) initial condition.
    ++ Such intial conditions are at least of the form 
    ++  D(..D(depVars[i])..)=e or depVars[i](e1)=e2 
    ++ where the ei's are not allowed to contain depVars.
 - ode::print(ode)
    ++ prints an element of ode out.
 - ode::splitSys(sys,depVars,indepVar)
    ++ returns a sequence of two sets, where the first set contains the 
    ++ differential expressions and the second the initial conditions.

Examples:
>> ode({y'(x)+1,z'(x)+y(x),D(y)(0)=2,y(0)=1},{y(x),z(x)})
>> ode::getComponents(%)

{y(x) + diff(z(x), x), diff(y(x), x) + 1}, {y, z}, x,

   {y(0) = 1, D(y)(0) = 2}
                 
++*/

ode::new:=
  proc(eq, unk=FAIL)
    local indepVar,components;
  begin
    if args(0)=0 then error("Wrong number of arguments") end_if;
    if type(eq)=ode then return(eq) end_if;

    if args(0)<>2 then error("Wrong number of arguments") end_if;
    // argument testing 
    case type(unk)
      of "function" do
        if nops(unk)<>1 then
          error("Function must be univariate")
        else
          indepVar:=op(unk,1);
        end_if;
        break
      of DOM_LIST do unk:={op(unk)};
      of DOM_SET do
        if map(unk, type)<>{"function"} then
          error("Unknowns must be functions")
        end_if;
        map(unk,op);
        if nops(%)<>1 then
          error("Unknowns must be univariate functions ".
                "in the same variable")
        else
          indepVar:=op(%,1)
        end_if;
        break
      of DOM_FAIL do error("Missing unknowns");
      otherwise
        error("Illegal unknowns")
    end_case;
    case domtype(eq)
      of DOM_EXPR do
        break
      of DOM_LIST do eq:={op(eq)};
      of DOM_SET do
        if map(eq, domtype)<>{DOM_EXPR} then
          error("Illegal equations")
        end_if;
        break
      otherwise
        error("First argument must be equation or set of equations")
    end_case;
    // end of argument testing 
    
    components:=ode::makeComponents(eq,unk);
    new(ode, components)
  end_proc:


ode::makeEquation:=
  proc(eq,depVars,indepVar)
  begin
    if has(eq,indepVar) then
      eq:=rewrite(eq,diff);
      if has(eq,D) then return(FAIL) end_if;
      if type(eq)="_equal" then eq:=op(eq,1)-op(eq,2) end_if;
      if has(eq,depVars(indepVar)) then eq
      else FAIL
      end_if
    else
      FAIL
    end_if
  end_proc:


ode::makeInitialCondition:=
  proc(eq,depVars,indepVar)
    local stepThroughD, newvar, res;
  begin
    stepThroughD:=a->case type(a)
                       of "D" do stepThroughD(op(a,1)); break;
                       of "function" do FALSE; break;
                       of DOM_IDENT do
                         if not has(a,depVars) then
                           FALSE
                         else
                           TRUE
                         end_if; break;
                       otherwise
                         FALSE
                     end_case;
                         
    if has(eq,indepVar) then FAIL
    elif type(eq)<>"_equal" then FAIL
    elif type(op(eq,1))="function" and
         has(op(eq,1),depVars) then
      if not has(op(eq,2),depVars) then
        case type(op(op(eq,1),0))
          of DOM_IDENT do
            if not has(op(op(eq,1),0),depVars) then
              error("Invalid initial condition")
            end_if; break;
          of "D" do
            if not stepThroughD(op(op(eq,1),0)) then
              error("Invalid initial condition")
            end_if; break;
        end_case;
        eq
      else // case: e.g. y(1)=z(3) or y'(1)=y(1)
        newvar:=genident("_v");
        res:=ode::makeInitialCondition(lhs(eq)=newvar,depVars,indepVar),
             ode::makeInitialCondition(rhs(eq)=newvar,depVars,indepVar);
        if has({res}, FAIL) then
          FAIL
        else
          res
        end_if;
      end_if;
    else FAIL
    end_if
  end_proc:


ode::splitSys:=
  proc(sys,depVars,indepVar)
    local eqs,inits,eq,h;
  begin
    eqs:=null(); inits:=null();
    if domtype(sys)<>DOM_SET then sys:={sys} end_if;
    for eq in sys do
      if (h:=ode::makeEquation(eq,depVars,indepVar))=FAIL then
        if (h:=ode::makeInitialCondition(eq,depVars,indepVar))=FAIL then
          error("Invalid equation or initial condition contained")
        else
          inits:=inits,h;
        end_if
      else
        eqs:=eqs,h;
      end_if
    end_for;
    {eqs},{inits}
  end_proc:


ode::makeComponents:=
  proc(sys,unks)
    local inits,depVars,indepVar,l,z,f,eqs,components;
  begin
    case type(unks)
        
      of DOM_SET do
        if nops(unks)=1 then unks:=op(unks)
        else // system of differential equations
          l:={}; z:={};
          for f in unks do
            if type(f)="function" then
              if type((depVars:=op(f,0)))=DOM_IDENT and nops(f)=1 then
                l:=l union {depVars};
                z:=z union {op(f)};
                next
              end_if
            end_if;
            error("Invalid unknown function")
          end_for;
          if nops(z)<>1 then
            error("Wrong number of independent variables") end_if;
          z:=op(z);
          
          // split equations and initial conditions
          eqs:=ode::splitSys(sys,l,z);
          sys:=eqs[1]; inits:=eqs[2];
          if nops(sys)<>nops(l) then
            error("Number of equations differs from number of unknowns")
          end_if;
          components:=sys,l,z,inits;
          break;
        end_if;
        
      of "function" do // only one unknown function
        depVars:=op(unks,0);
        if not has(sys,depVars) then
          error("Differential function does not appear") end_if;
        if nops(unks)<>1 then
          l:=ode::diffvars(sys,depVars);
          if nops(l)=1 then  //only one variable is differential 
            z:=op(l);
            l:=[op(unks)];
            sys:=ode::subsvars(sys,depVars,contains(l,z),l);
            unks:=depVars(z);
          else
            error("Resolution of partial differential equations ".
                  "not yet implemented")
          end_if
        end_if;
        indepVar:=op(unks);
        if {type(depVars),type(indepVar)}<>{DOM_IDENT} then
          error("Invalid arguments") end_if;
        if depVars=indepVar then
          error("Dependent and independent variable should differ") end_if;
        eqs:=ode::splitSys(sys,depVars,indepVar);
        inits:=eqs[2];
        eqs:=eqs[1];
        if nops(eqs)=1 then components:=op(eqs),depVars,indepVar,inits;
        else error("Number of equations differs from number of functions")
        end_if; break;
      otherwise
        error("Invalid arguments")
    end_case;
    components
  end_proc:

ode::getComponents:=
  proc(a)
    local sys,unks,components,restargs;
  begin
    restargs:=null();
    if args(0)=0 then error("Wrong number of arguments") end_if;
    if domtype(a)=ode then
      components:=op(a);
      if args(0)>1 then restargs:=args(2..args(0)) end_if;
    elif args(0)<2 then error("Wrong number of arguments")
    else
      sys:=args(1);
      unks:=args(2);
      components:=ode::makeComponents(sys,unks);
      if args(0)>2 then restargs:=args(3..args(0)) end_if;
    end_if;
    components,restargs
  end_proc:

ode::print:=
  proc(a)
    local components;
  begin
    components:=ode::getComponents(a);
    hold(ode)(if domtype(components[1])=DOM_SET then
                components[1] union components[4]
              elif nops(components[4])=0 then
                components[1]
              else
                {components[1]} union components[4]
              end_if,
              map(components[2],e->e(components[3])))
  end_proc:


























































































































































