/* ode::separate(eq,y,z,n) tries to solve the n-th order equation eq=0 for y(z)
  either recognizing a direct n-th order equation (D@@n)(y)(x)=f(x)
  or separate equations

  Examples: 
  >> ode::separate(diff(y(x),x,x)-exp(x),y,x,2);

                           {C4 + C3 x + exp(x)}

  >> ode::separate(diff(y(x),x)+y(x)^2,y,x,1);

                                {   1    }
                                { ------ }
                                { x - C5 }

  >> ode::separate(x+y(x)+1+(2*x+2*y(x)-1)*diff(y(x),x),y,x,1);

                 solve(x + 2 y + C6 + 3 ln(x + y - 2), y)
*/

ode::separate := proc(eq,y,z,n)
local s,i,l,inty,u;
save yp;
begin
   // first try to recognize (D@@n)(y)(x)=f(x) 
   // cf Zwillinger page 147 
   delete yp;
   s:=eval(subs(eq,diff(y(z),z$n)=yp));
    // using option 'PrincipalValue' leads to less efficiency and
    // in many cases to more bad results
   s:=solvelib::discreteSolve(s,yp,MaxDegree=1); //,PrincipalValue);
   if ode::isResultOfSolveOK(s,"NoRootOf") and nops(s)=1 then
      s:=op(s);
      if not has(s,{y,FAIL}) then // try to integrate directly 
	 userinfo(1,"trivial n-th order equation");
	 for i from 1 to n do
	    userinfo(2,"integrand is",s);
	    s:=int(s,z)+genident("C");
	 end_for;
	 return({s})
      elif n=1 then // try to separate 
         if type(s)<>"_mult" then s:=hold(_mult)(s) end_if;
         ([l,s,inty]):=split(s,has,y); 
	 l:=subs(l,y(z)=y);
	 if not has(l,z) then // int(dy/l)=int(s,z) 
	    userinfo(2,"separate equation:",diff(y(z),z)/l=s);
            inty:=int(1/l,y);
            if type(inty)="ln" then // put the integration constant in the log 
               s:=op(inty)=genident("C")*exp(int(s,z))
            else s:=inty=int(s,z)+genident("C")
            end_if;
            s:=solvelib::discreteSolve(s,y); //,PrincipalValue); ,MaxDegree=1);
            if calledFromWithinDSolve <> TRUE or
               //ode::isResultOfSolveOK(s,"AllowImplicit")
               ode::isResultOfSolveOK(s)
              then userinfo(1,"separate method worked");
              return(s);
            else userinfo(4,"separate method worked, but only get the ".
                          "implicit solution ".expr2text(s)." and skip");
            end_if;
         else // try with a change of function 
            l:=Factored::convert_to(factor(l),DOM_EXPR);
            if type(l)="_mult" then u:=op(select(l,has,y),1) else u:=l end_if;
            l:=s*l;
            u:=solvelib::discreteSolve(u,y); //,PrincipalValue); ,MaxDegree=1);
            if type(u)=DOM_SET and nops(u)<>0 then
               u:=op(u,1);
               l:=subs(l,y=y+u);
               // we have now a DE for yy=y-u 
               if not has(l,z) and not has((i:=diff(u,z)),z) then
		  // y' = yy' + i = l 
                  s:=int(1/(l-i),y);
                  s:=subs(s,y=y-u)-z;
                  if type(s)="_plus" then
                     s:=select(s,not testtype,Type::Constant)
                  end_if;
                  s:=solvelib::discreteSolve(s+genident("C"),y,MaxDegree=1);
                                             // ,PrincipalValue);
                  if calledFromWithinDSolve <> TRUE or
                     //ode::isResultOfSolveOK(s,"AllowImplicit")
                     ode::isResultOfSolveOK(s)
                    then userinfo(1,"separate method worked");
                    return(s);
                  else userinfo(4,"separate method worked, but only get the ".
                                "implicit solution ".expr2text(s)." and skip");
                  end_if;
               end_if
            end_if;
	 end_if
      end_if
   end_if;
   FAIL
end_proc:
