// $Date: 2000/11/21 20:41:10 $ $Author: fakler $ $Revision: 1.22 $ 

/* main file of the ODE library solver. This solver works as follows:
  the user calls solve(ode(...)) which calls in turn ode::solve(...)
  by the overloading mechanism of solve, and which calls ode::dsolve(...)
  (see below for details).

  The main reference for the methods used in this library is the book
  of Daniel Zwillinger, Handbook of Differential Equations, Academic Press.
*/


// the following (protected) identifier is used to detect if methods are
// called from within the solver. This becomes necessary due to some lacks
// in the algebraic solver. Thus one can avoid implicit results in many cases.
// E.g. ode::separate will often work and is one of the first tried methods,
// but produce mostly implicit results (due to the algebraic solver) which
// in turn yields horrible results of the ode solver.
// sysassign(calledFromWithinDSolve,FALSE): //yet predefined in LIBFILES/stdlib.mu


/* ode::solve(sys,yofz) solves the system (or equation) sys with
  respect to yofz.
  Input:  sys is either a single equation (or expression) or a set of equations
          yofz is either of the form y(z) or a set of such unknown functions
  Output: a (possibly empty) list of solutions. In the case of one unknown
          function, each solution is an expression. In the case of several 
          unknown functions, each solution is a list of assignments for these 
          functions.

  Examples: ode::solve(diff(y(x),x,x)=0,y(x));
            ode::solve({diff(y(x),x,x)=0,y(0)=a,D(y)(0)=b},y(x));
  ode::solve({diff(y1(t),t)=9*y1(t)+2*y2(t),diff(y2(t),t)=y1(t)+8*y2(t)},
              {y1(t),y2(t)});
*/

ode::solve:=
  proc()
    save z;
    local t,Z,prop,sol,components;
  begin

    components:=ode::getComponents(args());
    if nops(components)>4 then
      warning("Additional argument(s) ".
              expr2text(op(components,5..nops(components))).
              " are ignored")
    end_if;
    
    // handle properties of independent variable
    t:=components[3];
    Z:=genident("__z"); // saved variable z could already have a property
    if property::hasprop(z) then //assume(Z,prop1) end_if;
      //assume(Z,prop1)  // does'nt work due to an error in assume
      zip([Z],[getprop(z)],assume);
    end_if;
    delete z;
    if property::hasprop(t) then
      prop:=getprop(t);
      if prop=Type::Complex then
        //assume(z,prop) // does'nt work due to an error in assume
        zip([z],[Type::Complex],assume);
      elif not(property::Fail((prop:=prop and Type::Real))) then
        if prop=Type::Zero then
          prop:=Type::Complex;
          if ode::printWarningsFlag then
            warning("Cannot solve this ODE when ".expr2text(t).
                    " has property ".expr2text(getprop(t)).".\n".
                    "While solving, we will assume ".expr2text(t).
                    " has property ".expr2text(prop).".");
          end_if;
        elif is(t,Type::Rational)=TRUE then
          prop:=Type::Real;
          if ode::printWarningsFlag then
            warning("Cannot solve this ODE when ".expr2text(t).
                    " has property ".expr2text(getprop(t)).".\n".
                    "While solving, we will assume ".expr2text(t).
                    " has property ".expr2text(prop).".");
          end_if;
        end_if;
        //assume(z,prop)
        zip([z],[prop],assume)
      elif ode::printWarningsFlag then
        warning("Cannot assume ".expr2text(t).
                " is a real variable,\n".
                "since that would be inconsistent to ".
                "user given properties")
      end_if
    else
      assume(z,Type::Real)
    end_if;
    if t<>z then components:=op(subs([components],[z=Z,t=z])) end_if;    
    sol:=subs(ode::solve_intern(components),[z=t,Z=z]);
    unassume(Z);
    if has(sol,FAIL) then hold(solve)(ode(args())) else sol end_if
  end_proc:


ode::solve_intern :=
  proc(sys)
    save calledFromWithinDSolve;
    local depVars,indepVar,inits,components;
  begin
    if domtype(sys)=ode then
      components:=ode::getComponents(args());
      sys:=components[1];
      depVars:=components[2];
      indepVar:=components[3];
      inits:=components[4];
    else
      depVars:=args(2);
      indepVar:=args(3);
      inits:=args(4);
    end_if;
    
    // set flag for calling methods within dsolve
    sysassign(calledFromWithinDSolve,TRUE);
  
    // depth in routines of ODE, to avoid infinite loops 
    sysassign(ode::depth,0);
    
    if nops(depVars)>1 then
      ode::solve_sys(sys,depVars,indepVar,inits);
    else
      ode::solve_eq(sys,depVars,indepVar,inits);
    end_if;
  end_proc:

ode::isResultOfSolveOK:=
  proc(r,o=" ")  // option: "AllowImplicit", "NoRootOf"
  begin
    if type(r)=DOM_SET then
      if o="NoRootOf" and has(r,RootOf) then return(FALSE)
      else return(TRUE) end_if;
    end_if;
    if type(r)="solve" and o="AllowImplicit" then return(TRUE) end_if;
    FALSE
  end_proc:
    

// end of file 
