/* solves the equation
  l[1]*y(z)+l[2]*diff(y(z),z)+...+l[n+1]*diff(y(z),z$n)+l[n+2]=0 */

ode::solve_linear :=
proc(l,z,n)
  local inho,i,k,y,partsol,lc,gensol,eq,dummy,cis;
  save Y, yy;
// global ode::depth; 
begin
  delete Y, yy;
  dummy:=" "; // initial value
  if ode::depth>ode::maxDepth then return(FAIL) end_if;
  userinfo(1,"linear ordinary differential equation of order ".n);
  if n=0 then // l[1]*y(z)+l[2]=0 
    if iszero(l[1]) then return({}) else return({-l[2]/l[1]}) end_if
  end_if;
  userinfo(2,_plus(l[i+1]*diff(Y(z),z$i)$i=0..n)+l[n+2]=0);
  // try first exact_nth_order which deals with the inhomogeneous coeff. 
  y:=ode::exact_nth_order(l,z,n);
  if y<>FAIL then return(y) end_if;
  inho:=l[n+2]; delete l[n+2]; lc:=l[n+1]; // leading coefficient 
  y:=FAIL;
  if not has(l,z) then
    userinfo(1,"with constant coefficients");
    y:=ode::constant(l,z,n)
  elif n=1 then y:=[ode::firstord(l[1]*yy(z)+l[2]*diff(yy(z),z),yy,z)]
  elif map({op(l)},testtype,Type::PolyExpr(z,Type::Rational))={TRUE} then
    userinfo(1,"with polynomial coefficients");
    eq:=_plus(l[i+1]*diff(Y(z),z$i)$i=0..n);
    userinfo(3,"try to solve homogeneous part",eq=0);
    y:=ode::euler(l,z,n);  // Note, euler is also included in the next lines
    if y=FAIL then         // too, but yields easier (and slower!!) results
      if n=2 then
        userinfo(1,"looking for Liouvillian solutions, secondOrderCase");
        y:=ode::secondOrder(eq,Y,z)
      elif n=3 then
        userinfo(1,"looking for Liouvillian solutions, redThirdOrderCase");
        y:=ode::thirdOrder(eq,Y,z)
      else
        // the following lines are currently not general possible, although
        // it would be faster, since the complete solvers is designed to
        // yield generic solutions, while in the case here a list of
        // the solutions is needed for correctly solve the inhomogeneous
        // equation
        if inho=0 then
          userinfo(1,"looking for rational solutions");
          y:=ode::ratsols(eq,Y,z,n);
          if y={} then
            userinfo(1,"looking for exponential solutions");
            y:=ode::expsols(eq,Y,z,n);
          end_if;
          if 0<nops(y) and nops(y)<n then
            userinfo(1,"using reduction method");
            return(ode::reduction(eq,Y(z),op(y,1)))
          end_if;
        else // equation is inhomogeneous
          userinfo(1,"looking for Liouvillian solutions, partRedHigherCase");
          y:=ode::liouvillianSolutions(eq,Y(z));
          if nops(y)<>n and not has(y,RootOf) then
            if ode::printWarningsFlag then
             warning("found not all symbolic solutions of homogenous part. \n".
                     "Found Solutions are ".expr2text(y));
            end_if;
            return(FAIL);
          end_if;
        end_if;
      end_if;
      // doing some simplifications
      if has(y,int) then // in the methods for liouvillian solutions in most
        // cases integrals are remains unevaluated by the call hold(int)(..) 
        userinfo(4,"try to simplify found solutions");
        if traperror((dummy:=ode::simplifySolutions(y)),
                     ode::simplifyTime)=0 then
          y:=dummy;
          if has(y,int) then
            dummy:=hold(Unsimplified);
          else
            dummy:=" ";
          end_if;
        else
          if ode::debugWarningsFlag then
            warning("error occured in ode::simplifySolutions(".expr2text(y).
                    ") whereby ".expr2text(z)." has property ".
                    expr2text(getprop(z)));
          end_if;
          dummy:=hold(Unsimplified);
        end_if;
      end_if;
    end_if;
  else
    userinfo(1,"with non constant coefficients");
  end_if;
  userinfo(2,"solution of homogeneous part is",y);
  if y=FAIL then return(ode::adjoint(l,z,inho)) end_if;
  cis:=[genident("C") $ i=1..n];
  gensol:=_plus(op(zip(cis,[op(y)],_mult)));
  // a solution of the homogeneous part of the linear DE is C1*y[1]+...+Cn*y[n]
  // we now find a particular solution of the full equation 
  // using the method described on page 314 of Zwillinger   
  // (Variation of parameters)                              
  if inho<>0 and has(y,RootOf) then
    if ode::printWarningsFlag then 
    warning("implicit solutions found, cannot solve inhomogenous equation. \n".
            "Solutions of the homgeneous part are ".expr2text(y));
    end_if;
    return(FAIL)
  end_if;
  userinfo(4,"searching for a particular solution");
  partsol:=ode::variationOfConstants(y, z, inho, n, lc, dummy);
  if partsol=FAIL then return(FAIL) end_if;
  if dummy=hold(Unsimplified) then // prevent evaluations
    {collect(gensol+partsol,cis)};
  else  
    partsol:=expand(partsol);
    if has(partsol,{sin,cos}) then
      k:=combine(partsol,sincos);
      if k<>partsol and has(k,hold(int)) then k:=eval(k) end_if;
      partsol:=k
    end_if;
    gensol:=gensol+partsol;
    if has(gensol,exp) and has(gensol,I) then // do some simplifications
      gensol:=
        intlib::complexLnConversion(intlib::applyEulerIdentity(gensol,z),z);
    end_if;
    {collect(gensol,cis)};
  end_if;
end_proc:

/*
Input: sys is a set of solutions for the homogeneous equation
       z is the independent variable
       inho is the inhomogeneous term
       n is the order of the equation
       lc is the leading coefficient of the equation
       notEval (optional) = Unsimplified to let occuring integrals unevaluated 
Output: FAIL or a particular solution of the inhomogeneous equation
*/
ode::variationOfConstants :=
proc(sys, z, inho, n, lc, notEval=" ") 
  local k, y, i;
  save yy, up;
begin
  delete yy, up;
  // method described on page 314 of Zwillinger (Variation of parameters) 
  if inho=0 then 0
  else
    y:=[op(sys)];
    sys:={_plus(diff(y[k],z$i)*up[k] $ k=1..n) $ i=0..n-2,
          _plus(diff(y[k],z$(n-1))*up[k] $ k=1..n)+inho/lc},{up[k] $ k=1..n};
    userinfo(2,"system to solve is",sys);
    sys:=linsolve(sys);
    if sys=FAIL then return(FAIL) end_if;
    userinfo(2,"solution is",sys);
    // to prevent possible errors or to save time
    if notEval=hold(Unsimplified) then
      for k from 1 to n do
        yy[k]:=hold(int)(subs(up[k],sys),z);
      end_for;
    else // try to evaluate the integrals
      for k from 1 to n do
        userinfo(2,"integrating",up[k]);
        yy[k]:=int(eval(subs(up[k],sys)),z);
        userinfo(2,"solution is",yy[k]);
      end_for;
    end_if;
    _plus(yy[k]*y[k] $ k=1..n);
  end_if;
end_proc:

ode::constant :=
proc(l,z,n) local char_eq,i;
  save X;
begin
  delete X;
   char_eq:=_plus(l[i+1]*X^i $ i=0..n);
   userinfo(2,"solving characteristic equation",char_eq);
   l:=solvelib::discreteSolve(char_eq,X,hold(Multiple));
   l:=[expand(l)];
   userinfo(2,"solutions are",l);
   if nops(l)<>n then return(FAIL) end_if;
   ode::vspace(l,z);
end_proc:
