/* Input: a n-th order ODE in y(x).
   Output: a solution or FAIL.
   Uses the dependent variable missing method, cf Zwillinger chap. 43 p. 175
Examples: 
>> ode::ymissing(diff(y(x),x,x)+diff(y(x),x)-x,y,x,2);

                      {            2                }
                      {           x                 }
                      { C11 - x + -- - exp(C10 - x) }
                      {           2                 }

From Murphy, G.M., Ordinary Differential Equations and their solutions. 
Van Nostrand, Princeton, 1960, example 306 (non-linear second order):
>> eq:=diff(y(x),x,x)^3+diff(y(x),x)*(-2*diff(y(x),x)-12*x*diff(y(x),x,x)):
>> ode::ymissing(eq,y,x,2);
{C4, C4 + x^4*19/54, C4 + int(solve(int(x^(-3)*RootOf((x^(-3)*u11)*yp*(-12\
) + (x^(-3)*u11)^2*(-38) + (x^(-3)*u11)^3*27 + yp^3 + (x^(-3)*u11)*yp^2*9 \
+ (x^(-3)*u11)^2*yp*27, yp)^(-1), u11) = C3 + ln(x), u11), x)}
Remark: the above general solution could be simplified to C1*(x-C1)^3+C2
*/
ode::ymissing :=
proc(eq,y,x,n)
  local i,sol,C,lord;
  save yp;
begin
  delete yp;
   eq:=subs(eq,diff(y(x),x$i)=yp[i] $ i=1..n);
   if has(eq,y(x)) then FAIL
   else
      userinfo(1,"using dependent variable missing method");
      lord:=1; while not has(eq,yp[lord]) do lord:=lord+1 end_while;
      // lord is the lowest derivative order in y(x) 
      eq:=subs(eq,yp[i]=diff(y(x),x$(i-lord)) $ i=lord..n);
      userinfo(2,"solving new equation",eq);
      sysassign(ode::depth,ode::depth+1);
      sol:=ode::solve_eq(eq,y,x);
      sysassign(ode::depth,ode::depth-1);
      if has(sol,FAIL) then return(FAIL) end_if;
      for i from 1 to lord do
         C:=genident("C");
         sol:=ode::mapsol(sol,proc(t) begin int(t,x)+C end_proc);
         sol:=eval(sol);
      end_for;
      userinfo(1,"dependent variable missing method worked");
      sol
   end_if
end_proc:

