domain output::Priority
  inherits Dom::BaseDomain;
   
  about := "Domain of Priority elements";

  testtype := FALSE;

/*--------------------------------------------------------------------+
|  Method:                      new               (internal method)  |
|                                                                    |
|    Creates a Priority element.                                     |
|		                                                     |
|  Arguments:                                                        |
|		                                                     |
|    1  Expression		an arbitrary expression              |
|                                                                    |
|  Internal representation:                                          |
|                                                                    |
|    1  the type of the argument with special care on numbers        |
|                                                                    |
|  Return value:                the new Priority element             |
+--------------------------------------------------------------------*/
  new :=   
  proc() begin (
    if domtype(args()) = DOM_INT and args() < 0 then
      return(dom::Plus);
    end_if;
    if not contains({DOM_EXPR, DOM_RAT, DOM_COMPLEX}, domtype(args())) or 
       not contains(dom::value, type(args())) then
      return(dom::Stmt);
    end_if;
		
    new(dom, type(args()));
  ) end_proc;


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
|  Method:                    set                (interface method)  |
|                                                                    |
|    Set the priority of type(arg1) to arg2                          |
|		                                                     |
|  Arguments:                                                        |
|		                                                     |
|    1  Expression	      arbitrary expression                   |
|    2  Type::Rational        the new priority                       |
|                                                                    |
|  Return value:              the old priority of expression         |
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
  set :=
  proc()
    local op0, oldval;
  begin
    if testargs() then
      if args(0) <> 2 then
        error("wrong numbe\tr of args"):
      end_if;
      if domtype(args(1)) <> DOM_EXPR then
	error("expression expected");
      end_if;
      if not testtype(args(2), Type::Rational) then
        error("real number expected");
      end_if;
    end_if;  // testargs  
    
    op0 := type(args(1));
    oldval := dom::value[op0];
    dom::value[op0] := args(2);
    
    if testtype(oldval, Type::Rational)  then
      oldval
    else
      0
    end_if
  end_proc;


  print := proc(elem)
	   begin
	     dom::value[extop(elem,1)]
	   end_proc;

  
/*--------------------------------------------------------------------+
|  Method:                    compare             (internal method)  |
|                                                                    |
|    Compare 2 priorities with a given operator. Does the  work of   |
|    the following relational operators.                             |
|		                                                     |
|  Arguments:                                                        |
|		                                                     |
|    1   Operator              a relational operator		     |
|    2,3 Priority              two priorities                        |
|                                                                    |
|  Return value:               TRUE or FALSE                         |
+--------------------------------------------------------------------*/
  compare :=
  proc()
    local left, right, index1, index2;
  begin
    index1 := op(args(2),1);
    index2 := op(args(3),1);
    if contains(dom::value, index1) then
      left := dom::value[index1]
    else
      left := 0 
    end_if;
    if contains(dom::value, index2) then
      right := dom::value[index2]
    else
      right := 0 
    end_if;
    bool(args(1)(left, right));
  end_proc;


//+++++++++++++++++++  the relational operators:  ++++++++++++++++++

  _less    := proc() begin dom::compare(_less,    args()) end_proc;
  _leequal := proc() begin dom::compare(_leequal, args()) end_proc;
  _unequal := proc() begin dom::compare(_unequal, args()) end_proc;
  _equal   := proc() begin dom::compare(_equal,   args()) end_proc;

    
//----- internal table of the priority values for comparisions. ------

  value := table(
    "Null"     =  0, // dummy for default value 
    "_stmtseq"   =    1,  "_exprseq"     =    2,  "_or"        =  100,
    "_and"       =  200,  "_seqgen"      =  300,  "_less"      =  400,
    "_leequal"   =  400,  "_equal"       =  400,  "_unequal"   =  400,    
    "_range"     =  500,  "_union"       =  600,  "_minus"     =  700,
    "_intersect" =  800,  "_mod"         =  900,  "mods"       =  900,    
    "modp"       =  900,  "_div"         =  900,  "_plus"      = 1000,
    DOM_COMPLEX  = 1000,  "_mult_intern" = 1100,  DOM_RAT      = 1100,
    "_mult"      = 1100,  "_power"       = 1200,  "fact"       = 1300,
    "_not"       = 1400,  "_fconcat"     = 1500,  "_fnest"     = 1600,
    "_concat"    = 1700,  "_index"       = 1800,  "prime"      = 1900,
    "slot"    = 2000,

    "_assign"    = 1000000,    "_break"     = 1000000,    
    "_case"      = 1000000,    "_for"       = 1000000,
    "_for_down"  = 1000000,    "_for_in"    = 1000000,
    "_if"        = 1000000,    "_next"      = 1000000,    
    "_quit"      = 1000000,    "_repeat"    = 1000000,
    "_while"     = 1000000,    "function"   = 1000000 
  );


//----------- predefined Entries; may be used as Constants -----------

  Null       := new(output::Priority, "Null");
  Stmtseq    := new(output::Priority, "_stmtseq");
  Exprseq    := new(output::Priority, "_exprseq");
  Or         := new(output::Priority, "_or");
  And        := new(output::Priority, "_and");
  Seqgen     := new(output::Priority, "_seqgen");
  Relational := new(output::Priority, "_less");
  Range      := new(output::Priority, "_range");
  Union      := new(output::Priority, "_union");
  Minus      := new(output::Priority, "_minus");
  Intersect  := new(output::Priority, "_intersect");
  Mod        := new(output::Priority, "_mod");
  Plus       := new(output::Priority, "_plus");
  Mult       := new(output::Priority, "_mult");
  Power      := new(output::Priority, "_power");
  Not        := new(output::Priority, "_not");
  Fconcat    := new(output::Priority, "_fconcat");
  Concat     := new(output::Priority, "_concat");
  Slot       := new(output::Priority, "slot");
  Index      := new(output::Priority, "_index");
  Stmt       := new(output::Priority, "_if");
end_domain:
