// $Date: 2000/09/04 09:01:44 $ $Author: bij $ $Revision: 1.13 $

/* output::tableForm --- printing a set or list in table form

tableForm(CONT <, seperator> <, OPTIONS >)

CONT      : set or list
seperator : (optional) string
OPTIONS   : (optional) one or a set of options

tableForm(CONT) prints the objects of CONT in table form.
The width of the table depends on the size of TEXTWIDTH.

If seperator is given then it is appended to each object.
Appending spaces to the separator results additionally space
between columns.

OPTIONS is one or a set of

Unquoted            - strings will be printed without quotes
Unique              - all columns are of the same width
Left, Center, Right - the entries will be aligned left, center or right
Width = w           - the table has the max width 'w'
Column = c          - the table has 'c' columns
Sort = procedure    - the entries will be sorted with the given procedure
                      'procedure' is no procedure => no sorting
Output = file       - output into a file
                      if 'file' is a string => file named 'file' will be
                      opened and overwritten and closed after writing
                      if 'file' is a file descriptor the table will be
                      appended to 'file' without closing 'file'
Format = 

Without "Sort" the objects will be converted to strings an then sorted
alphabetically.

Examples:

>> output::tableForm(anames(DOM_PROC))
_check_global _seqin     alias    domattr funcattr loadproc newfuncarg 
newpurefunc   patchlevel pathname protect read     unalias  unprotect 
version       warning                                                  

>> output::tableForm(map(anames(DOM_PROC), expr2text), ", ", {Right})
"_check_global",     "_seqin",       "alias",    "domattr", "funcattr",
     "loadproc", "newfuncarg", "newpurefunc", "patchlevel", "pathname",
      "protect",       "read",     "unalias",  "unprotect",  "version",
      "warning"                                             

>> output::tableForm(map(anames(DOM_PROC), expr2text), "  ", {Center, Unique, Unquoted})
_check_global     _seqin          alias         domattr       funcattr     
  loadproc      newfuncarg     newpurefunc    oldinterval    patchlevel    
  pathname        protect         read          unalias       unprotect    
   version        warning
*/

alias(allOptions = {hold(Unquoted), hold(Unique), hold(Left), hold(Center), hold(Right),
                    hold(Width), hold(Column), hold(Output)}):
alias(allOptionsPar = {hold(Width), hold(Column), hold(Sort), hold(Output)}):

output::tableForm:=
  proc(CONT = [], SEP = " ", OPT = {None})
    local SORT, COLW, COLWM, COLN, k, PE, FILE, FC, SOPT;
  begin
    if testargs() then
      if not (type(CONT) in {DOM_LIST, DOM_TABLE, DOM_DOMAIN, DOM_SET}) then
        error("wrong type of first arg ('DOM_LIST', 'DOM_TABLE', 'DOM_DOMAIN' or 'DOM_SET' expected)")
      elif not (OPT in allOptions
                or (type(OPT) = "_equal" and op(OPT, 1) in allOptionsPar)
                or type(OPT) = DOM_SET) then
        error("wrong option name")
      elif type(SEP) <> DOM_STRING then
        if not (SEP in allOptions
                or (type(SEP) = "_equal" and op(SEP, 1) in allOptionsPar)
                or type(SEP) = DOM_SET) then
          error("wrong option name")
        elif OPT <> {hold(None)} then
          error("wrong type of second arg ('DOM_STRING' expected)")
        end_if
      end_if
    end_if;

    if type(SEP) <> DOM_STRING then // OPTIONS without SEPARATOR
      OPT:= SEP;
      SEP:= " "
    end_if;
    if type(OPT) <> DOM_SET then
      OPT:= {OPT}
    end_if;
    if OPT intersect {hold(Right), hold(Center)} = {} then
      OPT:= OPT union {hold(Left)}
    end_if;
    SOPT:= select(OPT, X -> type(X) = "_equal");
    OPT:= OPT minus SOPT;
    SOPT:= table(op(SOPT));

    // sort method
    SORT:= proc(X)
             local s;
           begin
             s:= select(X, x -> type(x) = "_equal" and op(x, 1) = hold(Sort));
             if s <> {} then
               if type(op(extop(s, 1), 2)) in {DOM_PROC, DOM_EXEC, DOM_FUNC_ENV} then
                 return(op(extop(s, 1), 2))
               else
                 FAIL
               end_if
             else
               return((X, Y) -> expr2text(X) < expr2text(Y))
             end_if
           end_proc(OPT);

    // file
    FC:= FALSE;
    FILE:= proc(X)
             local s;
           begin
             s:= select(X, x -> type(x) = "_equal" and op(x, 1) = hold(Output));
             if s <> {} then
               if domtype(op(extop(s, 1), 2)) = DOM_STRING then
                 if (s:= fopen(Text, op(extop(s, 1), 2), Write)) <> FAIL then
                   FC:= TRUE;
                   return(s)
                 end_if
               elif domtype(op(extop(s, 1), 2)) = DOM_INT
                    and traperror(fprint(Unquoted, op(extop(s, 1), 2), "")) = 0 then
                 return(op(extop(s, 1), 2))
               end_if
             end_if;
             0
           end_proc(OPT);

    if type(CONT) <> DOM_LIST then
      if type(CONT) = DOM_SET then
        CONT:= [op(CONT)];
        if SORT <> FAIL then
          CONT:= sort(CONT, SORT)
        end_if
      elif type(CONT) = DOM_TABLE then
        PE:= [];
        for k in CONT do
          
        end_for;
        return()
      elif type(CONT) = DOM_DOMAIN then
      end_if
    else
      // DOM_LIST
    end_if;

    // container empty
    if nops(CONT) = 0 then
      return()
    end_if;

    CONT:= map(CONT, X -> _if(contains(OPT, hold(Unquoted))
                              and type(X) = DOM_STRING, X, expr2text(X)));

    // col width
    // determine number of columns
    COLWM:= 0;
    for k in CONT do
      if length(k) > COLWM then
        COLWM:= length(k)
      end_if
    end_for;
    COLWM:= COLWM + length(SEP);

    if COLWM < 1 then
      warning("the table is to wide");
      COLWM:= 1
    end_if;
    
    //
    if contains(SOPT, hold(Column)) then
      COLN:= SOPT[hold(Column)]
    elif contains(SOPT, hold(Width)) then
      COLN:= SOPT[hold(Width)] div COLWM
    else
      COLN:= TEXTWIDTH div COLWM
    end_if;

    if COLN < 1 then
      warning("the table is to wide");
      COLN:= 1
    end_if;
      
    COLW:= [COLWM $ COLN];
    if not contains(OPT, hold(Unique)) then
      (COLW[i]:= max(op(map(zip(CONT, [k $k = 1..nops(CONT)],
                                (X, Y) -> if ((Y - 1) mod COLN) = (i - 1) then X else null() end),
                            length)), 1)
       + length(SEP)) $ i=1..COLN
    end_if;
    
    // empty entries
    if (k:= nops(CONT) - (nops(CONT) div COLN)*COLN) > 0 then
      (CONT:= append(CONT, "")) $ i = 1..COLN - k
    end_if;
    
    PE:=
      proc(X, Y, i)
        local sep;
      begin
        if X <> "" then
          if Y = "" or Y = FAIL then
            sep:= (_concat(" " $ length(SEP)))
          else
            sep:= SEP
          end_if;
          if i = COLN and contains(OPT, hold(Left)) then
            X.sep
          else
            if contains(OPT, hold(Right)) then
              (" " $ COLW[i] - length(sep) - length(X)).X.sep
            elif contains(OPT, hold(Center)) then
              Y:= COLW[i] - length(sep) - length(X);
              (" " $ Y div 2).X.sep.(" " $ Y - (Y div 2))
            else
              X.sep.(" " $ COLW[i] - length(sep) - length(X))
            end_if
          end_if
        else
          " " $ COLW[i]
        end_if
      end_proc;

    for k from 1 to nops(CONT) div COLN do
      fprint(Unquoted, FILE,
             PE(op(CONT, (k - 1)*COLN + i), op(CONT, (k - 1)*COLN + i + 1), i) $ i = 1..COLN);
    end_for;

    if FC then fclose(FILE) end_if;
    
    null()
  end_proc:
