/* -------------------------------------------------------
   plot:: HOrbital(n, l, m) erzeugt ein plot-Objekt, das
   das Wasserstoff-Orbital mit den Quantenzahlen n,l,m
   darstellt.

(  HOrbital erzeugt intern ein Plot-Objekt der Form
   plot::Surface3d([x,y,z], theta = [0,PI], pphi = [0,2*PI],
       Color = [Function, mycolor],
       ObjektOptionen)
)
  
Call:     plot::HOrbital(n, l, m, Option1, Option2, ...)

Parameters:
          n   -   Hauptquantenzahl = 1, 2, 3, ...
          l   -   Drehimpulsquantenzahl = 0, 1, .. n-1
          m   -   Magnetquantenzahl = -l, -l+1, .. , 0, .. , l-1, l

Options:
          Option1, Option2, ... 
            - alle Optionen fuer plot::Surface3d sind zulaessig und
              benutzen die entsprechenden Defaults.
              Lediglich der Default-Wert fuer Color ist auf
              Color = [Function, f] gestellt, wobei f intern
              erzeugt wird (siehe Details unten).
              Dieser Default wird bei Angabe von Color = ...
              aber wie ueblich ueberschrieben.

ReturnValue: an object of type plot::Surface3d

Examples:
      n:= 4: l:= 2: m:= 1:
      plot::HOrbital(n,l,m); 
      plot(%, Labeling = FALSE, Ticks = None);
      plot::HOrbital(n,l,m, Color=RGB::Blue);
      plot(%, Labeling = FALSE, Ticks = None);
      plot::HOrbital(n,l,m, Color=[Flat, RGB::Blue]);
      plot(%, Labeling = FALSE, Ticks = None);
      plot::HOrbital(n,l,m, Grid = [20,30]);
      plot(%, Labeling = FALSE, Ticks = None);
      plot::HOrbital(n,l,m, Grid = [10,10], Smoothness = [2,2]);
      plot(%, Labeling = FALSE, Ticks = None);
      plot::HOrbital(n,l,m, Grid = [30,10], Smoothness = [0,1], 
                     Title = "Untertitel am Objekt");
      plot(%, Labeling = FALSE, Ticks = None);
      plot::HOrbital(4, 1, 0, Grid = [30, 30], Smoothness = [1, 1],
                     Title = "ein genaueres Bild");
      plot(%, Labeling = FALSE, Ticks = None);
      plot::HOrbital(6, 4, 1, Grid = [40, 40], Smoothness = [2, 2],
                     Title = "sehr genau und sehr langsam");
      plot(%, Labeling = FALSE, Ticks = None);

Details:

Was wird hier dargestellt? 
Zu gegebenen Quantenzahlen n, l, m ist die quantenmechanische Aufenthalts-
wahrscheinlichkeit eines Elektrons durch das Betragsquadrat der komplexen 
Wellenfunktion Psi(x,y,z) gegeben, die in den blichen Polarkoordinaten

      x(r, phi, theta) = r * cos(phi) * sin(theta),
      y(r, phi, theta) = r * cos(phi) * sin(theta),
      z(r, phi, theta) = r * cos(theta) ,

      r = 0..infinity,  phi = 0 .. 2*PI,  theta = 0 .. PI

von der Form  Psi(x,y,z) = R(r) * Y(phi, theta)  ist. Das angezeigte Bild 
ist eine Visualisierung des Realteils des Winkelanteils Y(phi, theta): die
dargestellte Flche ist durch die Parametrisierung 

      x(phi, theta) = Re(Y(phi, theta))^2 * cos(phi) * sin(theta),

      y(phi, theta) = Re(Y(phi, theta))^2 * cos(phi) * sin(theta),

      z(phi, theta) = Re(Y(phi, theta))^2 * cos(theta)


gegeben. Der Abstand eines Flchenpunktes zum Nullpunkt ist damit  

Re(Y(phi, theta))^2, also: die Aufenthaltswahrscheinlichkeit ist in denjenigen

Richtungen besonders hoch, welche durch die "ausladenden Keulen" der 

Orbitalflche angedeutet werden.


Der Radialanteil R(r) der Wellenfunktion Psi(x,y,z)=R(r)*Y(phi, theta) geht nur

in die Einfrbung ein: hohe Werte von R(r)^2 liefern helle Farben, kleine 

Werte dunkle Farben. Die Farbe Rot deutet an, dass der Realteil Re(Psi) der 

Wellenfunktion hier positiv ist, Blau deutet negative Werte von Re(Psi) an.


------------------------------------------------------- */

plot::HOrbital:= proc(n,l,m)
local r, theta, pphi, R, Y_r, original_m, P, F,
      Rextrema, Rmax, tmpp, ct, Yextrema, Ymax,
      x, y, z, mycolor, colorOption, myPlotOptions, i;
option escape;
begin
  //-------------------------------------------------------
  // Test der Eingabe-Parameter
  //-------------------------------------------------------
  if (not testtype(n, DOM_INT)) or
     (not testtype(l, DOM_INT)) or
     (not testtype(m, DOM_INT)) then 
     error("Quantenzahlen muessen ganzzahlig sein");
  end_if;
  if n<=0 then
     error("die Hauptquantenzahl muss positiv sein");
  end_if;
  if l<0 or l>=n then
     error("0 <= Drehimpulsquantenzahl < Hauptquantenzahl verletzt");
  end_if;
  original_m:= m;
  m:= abs(m); // Realteil der Wellenfunktion fuer -m und m sind gleich
              // Imaginaerteile haben anderes Vorzeichen
  if m>l then
     error("|Magnetquantenzahl| <= Drehimpulsquantenzahl verletzt");
  end_if;

  //-------------------------------------------------------
  // Bereitstellen der adjungierten Kugelflaechen-Funktionen
  // und der konfluenten Reihe:
  //-------------------------------------------------------
  //-------------------------------------------------------
  // adjungierte Kugelfunktionen:
  // P(l,m,theta) = sin(theta)^m* (m.te Ableitung des
  //                               l.ten Legendre-Polynoms
  //                               an der Stelle cos(theta))
  //-------------------------------------------------------
  // Wird nur 3mal ausgewertet, also tut's folgende arg
  // aufwendige Berechnungsweise:

  P:= proc(l,m,theta) local x; 
  begin x:=genident("x"):
        m:= abs(m);
        sin(theta)^m* subs(
                 diff(expr(orthpoly::legendre(l,x)),x$m),
                 x=cos(theta));
  end_proc:

  //-------------------------------------------------------
  // Konfluente Reihe F(a,b,x) = 1 + a/b *x + a*(a+1)/b/b(+1)* x^2/2! + ...
  // Wird fuer den hier interessierenden Fall a = 0, -1, -2, ..
  // zum (Laguerre-) Polynom.
  // Einfache Implementation, liefert fuer a <> 0,-1,-2,.. immer 1.
  //-------------------------------------------------------
  F:= proc(a,b,x) local s,f,i;
  begin s:=1: f:=1;
        if testtype(-a, Type::NonNegInt) then
          for i from 1 to -a do
           f:= f*(a+i-1)/(b+i-1): s:= s + f*x^i/i!
          end_for;
        end_if;
        return(s);
  end_proc:

  //-------------------------------------------------------
  // Radialanteil R der Wellenfunktion mit Quantenindex n,l:
  //-------------------------------------------------------
  r:= genident("r");
  R:= r^l*exp(-r)*F(l+1-n,2*l+2,2*r);

  //-------------------------------------------------------
  // Realteil des Winkelanteils der Wellenfunktion mit Quantenindex l,m:
  //-------------------------------------------------------
  theta:= genident("theta");
  pphi:= genident("phi");
  Y_r:= P(l,m,theta)*cos(m*pphi);

  //-------------------------------------------------------
  // Der Realteil der quantenmechanischen Wellenfunktion Psi
  // in Kugelkoordinaten ist nun:
  //     Psi(r, theta, phi)= R(r)*Y_r(theta, phi)
  //-------------------------------------------------------

  //-------------------------------------------------------
  // Schritt 1:  finde maximalen Betrag des Radialanteils R
  // (dieser Wert wird fuer die Einfaerbung benoetigt):
  //-------------------------------------------------------
  // Suche Nullstellen der Ableitung:
  Rextrema:=numeric::solve(expand(diff(R,r)/exp(-r)),r):
  // ignoriere nicht-reelle Nullstellen, falls vorhanden:
  Rextrema:=select(map(Rextrema,float), testtype, DOM_FLOAT):
  // Werte R an den Nullstellen der Ableitung aus:
  Rmax:= max(op(map(Rextrema union {0.0}, 
             proc(rr) begin abs(float(subs(R,r=rr))) end_proc))):

  //-------------------------------------------------------
  // Schritt 2: finde maximalen Betrag des theta-Anteils P(l,m,theta)
  // der Kugelflaechenfunktionen Y_r (dieser Wert wird fuer die
  // Heuristik der Einfaerbung benoetigt).
  //-------------------------------------------------------
  ct:= genident("ct"):
  // tmpp ist proportional zur theta-Ableitung von P(l,m,theta)
  // fuer m>=0:
  tmpp:= subs(m*ct*P(l,m,theta)-(1-ct^2)*P(l,m+1,theta), 
                             [sin(theta)=1,cos(theta)=ct]):
  // Suche Nullstellen der Ableitung:
  Yextrema:= numeric::solve(tmpp,ct):
  // ignoriere nicht-reelle Nullstellen, falls vorhanden:
  Yextrema:= select(map(Yextrema,float), testtype, DOM_FLOAT):
  // Werte P(l,m,r) an den Nullstellen der Ableitung aus:
  Ymax:= max(op(map(Yextrema union {1.0, -1.0}, 
         proc(ct) begin abs(P(l,m,arccos(ct))) end_proc))):

  //-------------------------------------------------------
  // Schritt 3: Benutze uebliche Polarkoordinaten:
  //   x=r*cos(phi)*sin(theta)   ,   phi = 0 .. 2*PI
  //   y=r*sin(phi)*sin(theta)   ,   theta = 0 .. PI
  //   z=r*cos(theta)
  // Der Betrag r des Vektors [x,y,z] mit Polarwinkeln theta, phi
  // ist das Quadrat der Kugelflaechenfunktion Y_r(l,m,theta,phi).
  // Durch Ymax ist dieser Vektor auf Laenge <= 1 normiert (dies
  // wird nur fuer die Heuristik zur Einfaerbung gemacht).
  // Die Hauptquantenzahl n geht nur in den Radialanteil der
  // Wellenfunktion (d.h., in die Einfaerbung) ein.
  //-------------------------------------------------------
  x:= (Y_r/Ymax)^2*cos(pphi)*sin(theta):
  y:= (Y_r/Ymax)^2*sin(pphi)*sin(theta):
  z:= (Y_r/Ymax)^2*cos(theta):

  //-------------------------------------------------------
  // Schritt 4: eigene Funktion zur Einfaerbung des Orbitalals:
  //   Radialteil R(n,l,r) gross --> hell
  //   Radialteil R(n,l,r) klein --> dunkel
  //   Wellenfunktion positiv    --> rot
  //   Wellenfunktion negativ    --> blau
  //-------------------------------------------------------
  mycolor:= proc(x,y,z,ttt,ppp) 
  local rscale, rr, t, s, red, green, blue;
  begin 
      // Benutzt Quantenzahlen n, l als "globale" Variablen!
      // Rmax = max(abs(R(n,l,r), r = 0..infinity ) wird ebenfalls
      // global berechnet. 
      // Durch Normung von x,y,z ist r^2 = x^2+y^2+z^2 <=1.
      // Heuristik: setze rscale so, dass die wesentlichen
      // Oszillationen von R(n,l,r) sich auf dem Bereiche
      // 0.0 <= r <= rscale abspielen !
      
      // Einfachste Heuristik: (funktioniert, aber nicht optimal)
      // rscale:= n + 2*l; 

      // Etwas bessere Heuristik:
      rscale:= 2.0*n;  // halbwegs ok fuer l<>0.
      // Korrigiere fuer l=0, n<1 und n>6:
      if l=0 and n=0 then rscale:= 2.0 end_if;
      if l=0 and n>6 then rscale:= 2.0*n - 2*float(sqrt(n)) end_if;

      // Betrag des Vektors [x,y,z]
      rr:= (x^2+y^2+z^2)^(1/2);

      // t = Radialanteil(rr):
      t:= float(subs(R, r = rscale*rr)/Rmax); //  abs(t) <= 1.0 !

      // Vorzeichen der Wellenfunktion:
      if float(t*subs(Y_r, [theta=ttt, pphi=ppp]))>0
         then s:=1 else s:=-1
      end_if;

      t:=t^2;  // t = Radialanteil(rr)^2:
      // Faerbung gemaess Intensitaet des Radialanteils:
      // rot bei positivem Vorzeichen der Wellenfunktion,
      // blau bei negativem Vorzeichen:
      if s>=0
        then red:=  0.9*t^0.1 + 0.09; blue:= 0.0; green:= 0.0;
        else blue:= 0.9*t^0.1 + 0.09; red:=  0.0; green:= 0.0;
      end_if;
      [red,green,blue] 
  end_proc:

  //--------------------------------------------------------------
  // Schritt 5: Plot
  //--------------------------------------------------------------

  colorOption:= Color = [hold(Function), mycolor]; // Default
  myPlotOptions:= [args(4..args(0))];
  for i from 1 to args(0)-3 do
      if has(myPlotOptions[i], hold(Color))
         then colorOption:= null();
              break;
      end_if;
  end_for;
  plot::Surface3d([x, y, z],
                  theta = 0..PI,
                  pphi = 0..2*PI,
                  colorOption,
                  op(myPlotOptions)
  );
// plot3d(Axes = None,
//        BackGround = [1.0, 1.0, 1.0], // Hintergrund weiss
//        ForeGround = [0.0, 0.0, 0.0], // Beschriftung etc. schwarz
//        Title = "Quantenzahlen: ".
//                "n = ".expr2text(n).", ".
//                "l = ".expr2text(l).", ".
//                "m = ".expr2text(original_m),
//        [Mode = Surface, [x,y,z], theta = [0,PI], pphi = [0,2*PI], 
//        colorOption,
//        Style = [ColorPatches, AndMesh],
//        op(myPlotOptions)]);
// null():
end_proc:

