// kg, 08/07/94
// Lsys -- the domain of Lindenmeyer systems

/*
Example: (dragon curve)
>> l:= plot::Lsys(90,"L","L"="L+R+","R"="-L-R","L"=Line,"R"=Line):
>> l::generations := 10:
>> plot(l)
*/

domain plot::Lsys

type:= "graphprim";
dimension:= () -> 2;
info:= "plot::Lsys -- graphical primitive for Lindenmeyer systems";

/*
new -- create new L-system

new(deg,start,rule...)

deg   - turtle degree (real number)
start - starting rule (non-empty string)
rule  - rule of the form '<lhs>=<rhs>'

The <lhs> must be a string of length 1. The <rhs> may be a string,
a turtle primitive or a color value. A turtle primitive is one of the
identifiers 'Line', 'Move', 'Left', 'Right', 'Push' or 'Pop'. A color value
is a list with 3 floats in the range 0..1 which define the
red-, green- and blue value of the color.
*/

new:= proc(deg:Type::Real, start:DOM_STRING)
    local gen, i, j, p, startrule, pr, rule, prim, r, Lhs, Rhs, prims, rules, b, g, fdeg;
    option escape;
begin
    if args(0) < 3 then error("wrong no of args") end_if;
    gen:= 5;  // default no of generations
    fdeg:= float(deg);

    // parse start rule
    if length(start) = 0 then
	error("illegal start rule")
    end_if;
    p:= revert([ substring(start,j,1) $ j=0..length(start)-1 ]);
    startrule:= subsop(proc(T) name start; local r, f; begin
	    gen:= gen-1;
	    if gen = 0 then r:= prim else r:= rule end_if;
	    f:= _fconcat(op(map("p", r)));
	    T:= f(T);
	    gen:= gen+1;
	    T
	end_proc, [4,3,2,1,1,1] = p);

    // parse rules
    prim:= proc() begin id end_proc;
    rule:= proc(r) begin prim(r) end_proc;
    pr:= {};
    prims:= {};
    rules:= {};

    for i from 3 to args(0) do
	r:= args(i);
	if type(r) <> "_equal" then
	    error("illegal rule")
	end_if;
	if not testtype(op(r), Type::Product(DOM_STRING,
		    Type::Union(DOM_STRING, DOM_IDENT,
			Type::ListOf(Type::Real,3,3)))) then
	    error("illegal rule")
	end_if;
	Lhs:= op(r,1);
	if length(Lhs) <> 1 then
	    error("illegal left hand side of rule")
	end_if;
	Rhs:= op(r,2);
	case domtype(Rhs)
	of DOM_IDENT do
	    if contains(prims, Lhs) then
		error("multiple turtle Lhs")
	    end_if;
	    prims:= prims union {Lhs};
	    // rule for drawing primitive
	    case (Rhs)
	    of hold(Line) do
		pr:= pr union {"F"};
		prim(Lhs):= proc(T) name Prim; begin T::line(1.0) end_proc;
		break;
	    of hold(Move) do
		pr:= pr union {"f"};
		prim(Lhs):= proc(T) name Prim; begin T::move(1.0) end_proc;
		break;
	    of hold(Push) do
		pr:= pr union {"["};
		prim(Lhs):= proc(T) name Prim; begin T::push() end_proc;
		break;
	    of hold(Pop) do
		pr:= pr union {"]"};
		prim(Lhs):= proc(T) name Prim; begin T::pop() end_proc;
		break;
	    of hold(Left) do
		pr:= pr union {"+"};
		prim(Lhs):= proc(T) name Prim; begin T::left(fdeg) end_proc;
		break;
	    of hold(Right) do
		pr:= pr union {"-"};
		prim(Lhs):= proc(T) name Prim; begin T::right(fdeg) end_proc;
		break;
	    otherwise error("illegal right hand side of rule")
	    end_case;
	    break;

	of DOM_LIST do
	    if contains(prims, Lhs) then
		error("multiple turtle lhs")
	    end_if;
	    prims:= prims union {Lhs};
	    Rhs:= map(Rhs, float);
	    if min(op(Rhs)) < 0.0 or max(op(Rhs)) > 1.0 then
		error("illegal color")
	    end_if;
	    prim(Lhs):= subsop(proc(T) name Color; begin T::color("c") end_proc,
			    [4,1]=Rhs);
	    break;

	otherwise
	    if contains(rules, Lhs) then
		error("multiple rule lhs")
	    end_if;
	    rules:= rules union {Lhs};
	    // grammar rule
	    p:= revert([ substring(Rhs,j,1) $ j=0..length(Rhs)-1 ]);
	    rule(Lhs):= subsop(proc(T) name Rule; local r, f; begin
		    gen:= gen-1;
		    if gen = 0 then r:= prim else r:= rule end_if;
		    f:= _fconcat(op(map("p", r)));
		    T:= f(T);
		    gen:= gen+1;
		    T
		end_proc, [4,3,2,1,1,1] = p);
	end_case;
    end_for;

    // insert default drawing primitive rules
    if not contains(pr, "F") then
	prim("F"):= proc(T) name Prim; begin T::line(1.0) end_proc;
    end_if;
    if not contains(pr, "f") then
	prim("f"):= proc(T) name Prim; begin T::move(1.0) end_proc;
    end_if;
    if not contains(pr, "[") then
	prim("["):= proc(T) name Prim; begin T::push() end_proc;
    end_if;
    if not contains(pr, "]") then
	prim("]"):= proc(T) name Prim; begin T::pop() end_proc;
    end_if;
    if not contains(pr, "+") then
	prim("+"):= proc(T) name Prim; begin T::left(fdeg) end_proc;
    end_if;
    if not contains(pr, "-") then
	prim("-"):= proc(T) name Prim; begin T::right(fdeg) end_proc;
    end_if;
        
    // deg, start, args(i) are used for printing only
    r:= new(dom, startrule, g->(gen:=g), deg, start, args(i) $ i=3..args(0));
    delete deg, start, i, j, p, startrule, pr, Lhs, Rhs, prims, rules, b, g;
    r; r:= FAIL; %2
end_proc;

// print -- return expression to print instead of L-system l
print:= proc(l:dom) local i; begin
    hold(plot::Lsys)(extop(l,i) $ i=4..extnops(l))
end_proc;

// slot -- set or get 'generations' slot of L-system l
slot:= proc(l: dom, s: DOM_STRING, v=null() : Type::PosInt)
begin
    if s <> "generations" then error("invalid slot") end;
    if args(0) = 3 then
	extop(l,2)(v);
	l;
    else
        // relies on the fact that 'gen' is the first local var
	// of the enclosing proc-env of extop(l,1), which is the
	// proc-env of the 'new' method!
        op(op(extop(l,1), 12), 6);
    end_if;
end_proc;

// getPlotdata -- execute L-system l and return turtle path
getPlotdata := proc(l:dom) begin
    plot::Turtle::getPlotdata(extop(l,1)(plot::Turtle()));
end_proc;

end_domain:

// end of file
