// $Date: 2000/11/14 10:05:11 $  $Author: frankp $   $Revision: 1.25 $
// igel
// completely rewritten by frankp (21.09.2000)

/* -------------------------------------------------------------
    Curve3d -- The graphical primitive for three-dimensional curves

    Syntax:

    Curve3d([x_t, y_t, z_t], t = tr, <, op1, op2, ...>)

    x_t, y_t, z_t   : arithmetical expressions
    t               : DOM_IDENT
    tr              : either form 'a..b' or list of two arithmetical expressions.
    op1, op2, ...   : plot3d options, see plot2d and plot3d for details.

    Example:

    >> plot::Curve3d( [sin(t), cos(t), t], t = 0..2*PI )

    >> plot::Curve3d( [sin(t), cos(t), t], t = 0..2*PI, Color = RGB::Blue )

----------------------------------------------------------------*/
alias( DEFAULTOPTIONS = plot::Curve3d::defaultOptions ):
alias( OPTIONNAMES = plot::Curve3d::optionNames ):

plot::Curve3d := newDomain("plot::Curve3d"):
plot::Curve3d::info := "plot::Curve3d -- graphical primitive for three-dimensional curves":
plot::Curve3d::type := "graphprim":
plot::Curve3d::dimension:= () -> 3:

plot::Curve3d::new:= proc(f, r)
    local i, opt, curve, range, options;
begin    
    if args(0) < 2 then 
        error("expecting at least two arguments");
    end_if;

    if not testtype(f, Type::ListOf(Type::Arithmetical, 3, 3)) then
        error("1st argument: expecting a list of three arithmetical expressions");
    elif (range:= plot::conv2range(r)) = FAIL then
        error("2nd argument: expecting form 'ident = a..b'");
    end_if;

    options := DEFAULTOPTIONS;  // default options (a table)
    opt := [args(3..args(0))];  // user options (a list)
    // check user options and override the default options:
    for i from 1 to nops(opt) do
        if (r:= plot::Curve3d::checkOption(opt[i]))[1] = FALSE then
            error(r[2])
        else
            options[r[2]]:= r[3]
        end_if
    end_for;

    curve:= newDomain( expr2text(genident("plot::Curve3d_")) );

    // attributes:
    curve::term           := f;
    curve::range          := range;
    curve::plotdata       := [];
    curve::refreshPlotdata:= TRUE:
    curve::options        := options;

    // return the new curve:
    return( new(plot::Curve3d, curve) )
end_proc:

plot::Curve3d::copy := proc(f)
begin
    return(extsubsop(f, 1=newDomain(genident("plot::Curve3d_"), extop(f, 1))));
end_proc:

// not documented:
plot::Curve3d::diff := proc(c)
    local cc;
begin
    if args(0) = 1 then return(c) end_if;

    cc:= plot::Curve3d::copy(c);
    cc::term := map(cc::term, diff, args(2..args(0)));
    cc::refreshPlotdata:= TRUE;
    return( cc )
end_proc:

plot::Curve3d::modify := proc(object)
    local o, m, options;
begin
    m:= plot::Curve3d::copy(object);
    options:= [args(2..args(0))];    

    ( slot(m, expr2text(op(o, 1)), op(o, 2)) ) $ o in options;

    m::refreshPlotdata := TRUE;
    return(m)
end_proc:

plot::Curve3d::getPlotdata := proc(f)
    local e, x, c;
begin
    e:= extop(f, 1);
    if e::refreshPlotdata = FALSE then return( e::plotdata ) end_if;

    c:= e::term;
    x:= plot::range2list(e::range);

    e::refreshPlotdata:= FALSE;
    // create the plot data:
    e::plotdata:= [
        [hold(Mode) = hold(Curve), c, x] . [op(e::options)]
    ]
end_proc:

plot::Curve3d::print:= proc(c)
begin
    c:= extop(c,1);
    hold(plot::Curve3d)(c::term,c::range)
end_proc:

plot::Curve3d::op := proc(c, n)
begin
    if args(0) < 1 or args(0) > 2 then
        error("invalid number of arguments")
    end_if;

    c:= extop(c,1);
    c:= [c::term, c::range];
    if args(0) = 1 then op(c) else op(c, n) end_if
end_proc:

plot::Curve3d::nops:= c -> 2:

plot::Curve3d::_index:= proc(c:plot::Curve3d,n:Type::PosInt)
begin
    if args(0) <> 2 then
        error("expecting two arguments")
    end_if;

    c:= extop(c,1);
    c:= [c::term, c::range];
    if n > nops(c) then
        error("index out of range")
    else
        return( c[n] )
    end_if
end_proc:

// Note: Do not return an error message in the case of an invalid call,
//       because then the object would be destroyed! Hence a warning 
//       message is issued, and the call of this method has no effect 
//       on the object.
plot::Curve3d::set_index := proc(c:plot::Curve3d, n:Type::PosInt, v)
    local e;
begin
    if args(0) <> 3 then
        warning("wrong number of arguments; assignment ignored");
        return(c)
    end_if; 

    e:= extop(c,1);
    case n
    of 1 do
        if v <> e::term then 
            if not testtype(v,Type::ListOf(Type::Arithmetical, 3, 3)) then
                warning("3rd argument: expecting a list of three arithmetical expressions; assignment ignored");
                return(c)
            end_if;
            e::refreshPlotdata:= TRUE;
            e::term:= v;
        end_if;
        return( extsubsop(c,1=e) )
    of 2 do
        if v <> c::range then 
            if (v:= plot::conv2range(v)) = FAIL then
                warning("3rd argument: expecting form 'ident = a..b'; assignment ignored");
                return(c)
            end_if;
            e::refreshPlotdata:= TRUE;
            e::range:= v;
        end_if;
        return( extsubsop(c,1=e) )
    otherwise
        warning("index out of range; assignment ignored");
        return(c)
    end_case;
end_proc:

// Note: Do not return an error message in the case of an invalid call,
//       because then the object would be destroyed! Hence a warning 
//       message is issued, and the call of this method has no effect 
//       on the object.
plot::Curve3d::slot:= proc(o,s,v=null())
    local e, t, _s;
begin
    if args(0) = 3 then
        // slot assignment:
        e:= extop(o,1);

        // first handle all user-accessable attributes:
        case s
        of "refreshPlotdata" do
            if v <> TRUE and v <> FALSE then
                warning("attribute 'refreshPlotdata': expecting either 'TRUE' or 'FALSE'; assignment ignored");
                return( o )
            else
                e::refreshPlotdata:= v;
                return( extsubsop(o,1=e) )
            end_if
        of "range" do
            if (v:= plot::conv2range(v)) = FAIL then
                error("attribute 'range': expecting form 'x = a..b'; assignment ignored")
            else
                e::range:= v;
                e::refreshPlotdata:= TRUE;
                return( extsubsop(o,1=e) )
            end_if
        of "term" do
            if not testtype(v,Type::ListOf(Type::Arithmetical,3,3)) then
                warning("attribute 'term': expecting a list of three arithmetical expressions; assignment ignored");
                return( o )
            else
                e::term:= v;
                e::refreshPlotdata:= TRUE;
                return( extsubsop(o,1=e) )
            end_if
       of "options" do
            if domtype(v) <> DOM_TABLE then
                warning("attribute 'options': expecting a table of plot options; assignment ignored");
                return( o )
            else
                e::options:= v;
                e::refreshPlotdata:= TRUE;
                return( extsubsop(o,1=e) )
            end_if
        end_case;

        // an option given?
        t:= plot::Curve3d::checkOption(text2expr(s) = v);
        if t[1] = FALSE then
            warning( t[2]."; assignment ignored" );
            return(o)
        else
            e::options[t[2]]:= t[3];
            e::refreshPlotdata:= TRUE;
            return( extsubsop(o,1=e) )
        end_if
    else
        // slot access:
        e:= extop(o,1);
        if (t:= slot(e,s)) <> FAIL then
            return(t)
        else
            _s:= text2expr(s);
            if not contains(OPTIONNAMES,_s) then
                error("unknown attribute or option '".s."'")
            elif type((t:= e::options[_s])) <> "_index" then
                return(t)
            else
                return(FAIL)
            end_if
        end_if
    end_if
end_proc:

// not documented:
plot::Curve3d::move := proc(s, x, y, z)
    local sc, p;
begin
    p:= plot::Point(s::term);
    p:= plot::Point::move(p, x, y, z);

    sc:= plot::Curve3d::copy(s);
    sc::term:= [op(p)];
    return(sc)    
end_proc:

// not documented:
plot::Curve3d::rotate := proc(s, xphi, yphi, zphi)
    local sc, p;
begin    
    p:= plot::Point(s::term);        
    p:= plot::Point::rotate(p, xphi, yphi, zphi);

    sc:= plot::Curve3d::copy(s);
    sc::term:= [op(p)];
    return( sc )
end_proc:

plot::Curve3d::checkOption := proc(opt)
    local value;
begin
    if type(opt) <> "_equal" then
        return( [FALSE,"options must have the form 'OptionName = value'"] )
    end_if;

    value:= op(opt,2); opt:= op(opt,1);
    if not contains( OPTIONNAMES,opt ) then
        return( [FALSE,"unknown option name '".expr2text(opt)."'"] )
    else
        // First, handle special cases for certain options:
        // Color : Beside the standard format 'Color = [{Height,Flat},[r,g,b]]',
        //         allow 'Color = [r,g,b]' which corresponds to '[Flat,[r,g,b]]'.
        if opt = hold(Color) then
            if (value:= plot::getColorFormat(value)) = FAIL then
                return( [FALSE,"option 'Color': invalid color specification"] )
            end_if
        end_if;

        return( [TRUE,opt,value] )
    end_if
end_proc:

OPTIONNAMES:= {
  // options for 2-dimensional objects:
  hold(Color),hold(LineStyle),hold(LineWidth),hold(PointStyle),
  hold(PointWidth),hold(Title),hold(TitlePosition),
  // options for 2d-curves:
  hold(Grid),hold(Smoothness),hold(Style)
}:

DEFAULTOPTIONS:= table():
// Default options for objects of the domain plot::Curve3d:
DEFAULTOPTIONS[hold(Grid)]:= [100]:

