// $Date: 2000/11/14 10:05:12 $  $Author: frankp $   $Revision: 1.24 $
// igel
// completely rewritten by frankp (21.09.2000)

/* -------------------------------------------------------------
    Function3d -- The graphical primitive for three-dimensional 
                  function plots

    Syntax:

    Function3d(f_xy, x = xr, y = yr <, op1, op2, ...>)

    f_xy          : arithmetical expressions (depending on x and y)
    x, y          : DOM_IDENT
    xr, yr        : either form 'a..b' or list of two arithmetical expressions.
    op1, op2, ... : list of plot3d options, see plot2d and plot3d for details.

    Example:

    >> plot::Function3d( sin(t + u), t = 0..2*PI, u = -1..1 )

    >> plot::Function3d( sin(t + u), t = 0..2*PI, u = -1..1, Color = RGB::Blue )

----------------------------------------------------------------*/
alias( DEFAULTOPTIONS = plot::Function3d::defaultOptions ):
alias( OPTIONNAMES = plot::Function3d::optionNames ):

plot::Function3d := newDomain("plot::Function3d"):
plot::Function3d::info := "plot::Function3d -- graphical primitive for three-dimensional function graphs":
plot::Function3d::type := "graphprim":
plot::Function3d::dimension:= () -> 3:

plot::Function3d::new:= proc(f, rx, ry)
    local i, opt, function, xrange, yrange, options;
begin    
    if args(0) < 3 then
        error("expecting at least three arguments");
    elif not testtype(f, Type::Arithmetical) then
        error("1st argument: expecting an arithmetical expression")
    end_if:

    xrange := plot::conv2range(rx);
    yrange := plot::conv2range(ry);
    if xrange = FAIL then
        error("2nd argument: expecting form 'ident = a..b'")
    end_if;
    if yrange = FAIL then
        error("3rd argument: expecting form 'ident = a..b'")
    end_if;

    options := DEFAULTOPTIONS;  // default options (a table)
    opt := [args(4..args(0))];  // user options (a list)
    // check user options and override the default options:
    for i from 1 to nops(opt) do
        if (rx:= plot::Function3d::checkOption(opt[i]))[1] = FALSE then
            error(rx[2])
        else
            options[rx[2]]:= rx[3]
        end_if
    end_for;

    function := newDomain( expr2text(genident("plot::Function3d_")) );
    
    // attributes:
    function::term           := f;
    function::range1         := xrange;
    function::range2         := yrange;
    function::refreshPlotdata:= TRUE;
    function::plotdata       := [];
    function::options        := options;

    return( new(plot::Function3d, function) )
end_proc:

plot::Function3d::copy := proc(f)
begin
    return(extsubsop(f, 1=newDomain(genident("plot::Function3d_"), extop(f, 1))));
end_proc:

plot::Function3d::modify := proc(f)
    local o, m, options;
begin
    m := plot::Function3d::copy(f);
    options := [args(2..args(0))];

    ( slot(m, expr2text(op(o, 1)), op(o, 2)) ) $ o in options;

    m::refreshPlotdata := TRUE;
    return(m)
end_proc:

plot::Function3d::getPlotdata := proc(f)
    local e, c, x, y;
begin
    e:= extop(f, 1);
    if e::refreshPlotdata = FALSE then return( e::plotdata ) end_if;

    c:= e::term;    
    x:= plot::range2list(e::range1);
    y:= plot::range2list(e::range2);

    e::refreshPlotdata:= FALSE;
    // create the plot data:
    e::plotdata:= [
        [hold(Mode) = hold(Surface), [op(x,1),op(y,1),c], x, y] . [op(e::options)]
    ]
end_proc:

plot::Function3d::print:= proc(f)
begin
    f:= extop(f,1);
    hold(plot::Function3d)(f::term,f::range1,f::range2)
end_proc:

plot::Function3d::op := proc(c, n)
begin
    if args(0) < 1 or args(0) > 2 then
        error("invalid number of arguments")
    end_if;

    c:= extop(c,1);
    c:= [c::term,c::range1,c::range2];
    if args(0) = 1 then op(c) else op(c, n) end_if
end_proc:

plot::Function3d::nops := c -> 3:

plot::Function3d::_index := proc(c:plot::Function3d,n:Type::PosInt)
begin
    if args(0) <> 2 then
        error("expecting two arguments")
    end_if;

    c:= extop(c,1);
    c:= [c::term,c::range1,c::range2];
    if n > nops(c) then
        error("index out of range")
    else
        return( c[n] )
    end_if
end_proc:

// Note: Do not return an error message in the case of an invalid call,
//       because then the object would be destroyed! Hence a warning 
//       message is issued, and the call of this method has no effect 
//       on the object.
plot::Function3d::set_index := proc(c:plot::Function3d, n:Type::PosInt, v)
    local e;
begin
    if args(0) <> 3 then
        warning("wrong number of arguments; assignment ignored");
        return(c)
    end_if; 

    e:= extop(c,1);
    case n
    of 1 do
        if v <> e::term then 
            if not testtype(v,Type::Arithmetical) then
                warning("3rd argument: expecting an arithemtical expression; assignment ignored");
                return(c)
            end_if;
            e::refreshPlotdata:= TRUE;
            e::term:= v;
        end_if;
        return( extsubsop(c,1=e) )
    of 2 do
        if v <> e::range1 then 
            if (v:= plot::conv2range(v)) = FAIL then
                warning("3rd argument: expecting form 'ident = a..b'; assignment ignored");
                return(c)
            end_if;
            e::refreshPlotdata:= TRUE;
            e::range1:= v;
        end_if;
        return( extsubsop(c,1=e) )
    of 3 do
        if v <> c::range2 then 
            if (v:= plot::conv2range(v)) = FAIL then
                warning("3rd argument: expecting form 'ident = a..b'; assignment ignored");
                return(c)
            end_if;
            e::refreshPlotdata:= TRUE; 
            e::range2:= v;
        end_if;
        return( extsubsop(c,1=e) )
    otherwise
        warning("index out of range; assignment ignored");
        return(c)
    end_case
end_proc:

// Note: Do not return an error message in the case of an invalid call,
//       because then the object would be destroyed! Hence a warning 
//       message is issued, and the call of this method has no effect 
//       on the object.
plot::Function3d::slot:= proc(o,s,v=null())
    local e, t, _s;
begin
    if args(0) = 3 then
        // slot assignment:
        e:= extop(o,1);

        // first handle all user-accessable attributes:
        case s 
        of "refreshPlotdata" do
            if v <> TRUE and v <> FALSE then
                warning("attribute 'refreshPlotdata': expecting either 'TRUE' or 'FALSE'; assignment ignored");
                return( o )
            else
                e::refreshPlotdata:= v;
                return( extsubsop(o,1=e) )
            end_if
        of "range1" do
        of "range2" do
            if (v:= plot::conv2range(v)) = FAIL then
                warning("attribute '".s."': expecting form 'x = a..b'; assignment ignored");
                return( o )
            else
                e::refreshPlotdata:= TRUE;
                return( extsubsop(o,1=slot(e,s,v)) )
            end_if
        of "term" do
            if not testtype(v,Type::Arithmetical) then
                warning("attribute 'term': expecting an arithmetical expression; assignment ignored");
                return( o )
            else
                e::term:= v;
                e::refreshPlotdata:= TRUE;
                return( extsubsop(o,1=e) )
            end_if
        of "options" do
            if domtype(v) <> DOM_TABLE then
                warning("attribute 'options': expecting a table of plot options; assignment ignored");
                return( o )
            else
                e::options:= v;
                e::refreshPlotdata:= TRUE;
                return( extsubsop(o,1=e) )
            end_if
        end_case;

        // an option given?
        t:= plot::Function3d::checkOption(text2expr(s) = v);
        if t[1] = FALSE then
            warning( t[2]."; assignment ignored" );
            return(o)
        else
            e::options[t[2]]:= t[3];
            e::refreshPlotdata:= TRUE;
            return( extsubsop(o,1=e) )
        end_if
    else
        // slot access:
        e:= extop(o,1);
        if (t:= slot(e,s)) <> FAIL then
            return(t)
        else
            _s:= text2expr(s);
            if not contains(OPTIONNAMES,_s) then
                error("unknown attribute or option '".s."'")
            elif type((t:= e::options[_s])) <> "_index" then
                return(t)
            else
                return(FAIL)
            end_if
        end_if
    end_if
end_proc:

plot::Function3d::checkOption := proc(opt)
    local value;
begin
    if type(opt) <> "_equal" then
        return( [FALSE,"options must have the form 'OptionName = value'"] )
    end_if;

    value:= op(opt,2); opt:= op(opt,1);
    if not contains( OPTIONNAMES,opt ) then
        return( [FALSE,"unknown option name '".expr2text(opt)."'"] )
    else
        // First, handle special cases for certain options:
        // Color : Beside the standard format 'Color = [{Height,Flat},[r,g,b]]',
        //         allow 'Color = [r,g,b]' which corresponds to '[Flat,[r,g,b]]'.
        if opt = hold(Color) then
            if (value:= plot::getColorFormat(value)) = FAIL then
                return( [FALSE,"option 'Color': invalid color specification"] )
            end_if;
        end_if;

        return( [TRUE,opt,value] )
    end_if
end_proc:

OPTIONNAMES:= {
  // options for 3-dimensional objects:
  hold(Color),hold(LineStyle),hold(LineWidth),hold(PointStyle),
  hold(PointWidth),hold(Title),hold(TitlePosition),
  // options for surfaces:
  hold(Grid),hold(Smoothness),hold(Style)
}:

DEFAULTOPTIONS:= table():
// Default options for objects of the domain plot::Function3d:
DEFAULTOPTIONS[hold(Grid)]          := [20,20]:

