// $Date: 2000/11/15 08:41:09 $  $Author: frankp $   $Revision: 1.13 $
// igel
// renamed to plot::Group and completely rewritten by frankp (22.09.2000)

/* -----------------------------------------------------------
    Group -- The graphical primitive for a group of graphicals primitives

    Syntax:

    Group(o1 <, o2, ...> <, op1, op2, ...>)

    o1, o2, ...  : graphical objects
    op1, op2, ...: list of plot2d options, see plot2d and plot3d for details.

    Example:
    >> plot::Group( plot::Ellipse2d( plot::Point(1, 1), i, 2*i ) $ i = 1..10 )

    >> plot::Group( plot::Ellipse2d( plot::Point(1, 1), i, 2*i ) $ i = 1..10, LineWidth = 12 )

    >> myObjects := plot::Ellipse2d( plot::Point(1, 1), i, 2*i , Color = RBG::ColorList[i]) ) $ i = 1..10

    >> plot::Group( myObjects, Color = RGB::Blue )
    >> plot::Group( myObjects, Color = RGB::Green )

----------------------------------------------------------------*/ 
plot::Group:= newDomain("plot::Group"):
plot::Group::info:= "plot::Group -- a group of graphical primitives":
plot::Group::type:= "graphprim":
plot::Group::dimension:= g -> (extop(g,1))::dimension:

plot::Group::new := proc()
    local e, opt, dimen, objects, group;
begin
    if args(0) = 0 then
        error("no arguments given")
    end_if;

    objects:= [args()];
    e:= split(objects, e -> type(e) = "graphprim");
    objects:= e[1];
    opt:= e[2]; // user specified options
    if nops(objects) = 0 then
        error("no graphical primitive given")
    end_if;

    // determine the dimension of the group:
    dimen:= {op(map(objects, e -> e::dom::dimension(e)))};
    if nops(dimen) > 1 then
        error("objects have different dimensions");
    else 
        dimen:= op(dimen,1)
    end_if;

    // transmit the options to each object:
    for e in opt do
        objects:= map( objects, o -> slot( o,expr2text(op(e,1)),op(e,2) ) )
    end_for;

    group:= newDomain( expr2text(genident("plot::Group_")) );

    // attributes:
    group::plotdata       := [];
    group::objects        := objects;
    group::dimension      := dimen;
    
    return(new(plot::Group, group))
end_proc:

// not documented:
plot::Group::move := proc(g, p1, p2, p3)
    local c;
begin
    c:= plot::Group::copy(g);
    c::objects := map(c::objects, o -> o::dom::move(o, p1, p2, p3));
    return(c)
end_proc:

// not documented:
plot::Group::rotate := proc(g, p1, p2, p3)
    local c;
begin
    c:= plot::Group::copy(g);
    c::objects:= map(c::objects,  i -> i::dom::rotate(i, p1, p2, p3));
    return(c)
end_proc:

plot::Group::copy := proc(o)
    local l, g;
begin
    g:= extop(o,1);
    l:= map( g::objects, x -> x::dom::copy(x) );
    g:= newDomain(genident("plot::Group_"), g);
    g::objects:= l;
    return( extsubsop(o, 1=g) )
end_proc:

plot::Group::modify := proc(g)
    local e, o, m, options, objects;
begin
    m:= plot::Group::copy(g);
    e:= extop(m,1);

    options:= [args(2..args(0))];
    objects:= e::objects;
    // transmit the options to each object:
    for o in options do
        objects:= map( objects, e -> slot( e,expr2text(op(o,1)),op(o,2) ) )
    end_for;
    e::objects:= objects;
    return( extsubsop(m,1=e) )
end_proc:

plot::Group::print := () -> hold(plot::Group)():

plot::Group::expose := g -> hold(plot::Group)(op(map((extop(g,1))::objects,expose))):

plot::Group::op := proc(g, n)
begin
    if args(0) < 1 or args(0) > 2 then
        error("invalid number of arguments")
    end_if;

    g:= (extop(g,1))::objects;
    if args(0) = 1 then op(g) else op(g, n) end_if
end_proc:

plot::Group::nops:= g -> nops((extop(g,1))::objects):

plot::Group::_index:= proc(g:plot::Group,n:Type::PosInt)
begin
    if args(0) <> 2 then
        error("expecting two arguments")
    end_if;
    g:= (extop(g,1))::objects;
    if n > nops(g) then
        error("index out of range")
    else
        return( g[n] )
    end_if
end_proc:

// Note: Do not return an error message in the case of an invalid call,
//       because then the object would be destroyed! Hence a warning 
//       message is issued, and the call of this method has no effect 
//       on the object.
plot::Group::set_index := proc(g:plot::Group, n:Type::PosInt, v)
    local e, l;
begin
    if args(0) <> 3 then
        warning("expecting three arguments; assignment ignored");
        return(g)
    elif type(v) <> "graphprim" then
        warning("3rd argument: expecting an object of type \"graphprim\"; assignment ignored");
        return(g)
    end_if; 
 
    e:= extop(g,1);
    l:= e::objects;
    if n > nops(l) then
        warning("index out of range; assignment ignored") ;
        return(g)
    else
        e::objects:= subsop( l,n=v );
        return( extsubsop(g,1=e) )
    end_if
end_proc:

plot::Group::getPlotdata := proc(o)
    local plotdata;
begin
    o:= extop(o,1);
    plotdata:= map( o::objects, oo -> op(oo::dom::getPlotdata(oo)) );
    o::plotdata:= plotdata
end_proc:

// Note: Do not return an error message in the case of an invalid call,
//       because then the object would be destroyed! Hence a warning 
//       message is issued, and the call of this method has no effect 
//       on the object.
plot::Group::slot:= proc(o,s,v=null())
    local e, t;
begin
    if args(0) = 3 then
        // slot assignment:
        e:= extop(o,1);

        // first handle all user-accessable attributes:
        case s 
        of "objects" do
            if domtype(v) <> DOM_LIST and map({op(v)},type) <> {"graphprim"} then
                warning("attribute 'objects': expecting a list of objects of type \"graphprim\"; assignment ignored");
                return( o )
            elif map({op(v)}, x -> x::dom::dimension(x)) <> {e::dimension} then
                warning("attribute 'objects': objects have invalid dimensions; assignment ignored");
                return( o )
            else
                e::objects:= v;
                return( extsubsop(o,1=e) )
            end_if
        otherwise
            // transmit the attribute or option to each object of the group:
            e::objects:= map( e::objects, o -> slot( o,s,v ) );
            return( extsubsop(o,1=e) )
        end_case;
    else
        // slot access:
        e:= extop(o,1);
        if (t:= slot(e,s)) <> FAIL then
            return(t)
        else
            error("unknown attribute '".s."'")
        end_if
    end_if
end_proc:

