// $Date: 2000/11/14 10:05:13 $  $Author: frankp $   $Revision: 1.23 $
// igel
// completely rewritten by frankp (22.09.2000)

/* -----------------------------------------------------------
    Point -- The graphical primitive for two- and three-dimensional points

    Syntax:
    Point(x, y <, z> <, op1, op2, ...>)

    x, y, z      : arithmetical expressions
    op1, op2, ...: plot2d options, see plot2d and plot3d for details.

    Example:
    >> plot::Point(1, 2, 3)

    >> plot::Point(1, 2)

    >> plot::Point(1, 2, PointWidth = 12)

    Note: The plot structure of a point p is [Mode=List,[p],Color=...,...],
          i.e., the color is not stored in p!
----------------------------------------------------------------*/ 
alias( DEFAULTOPTIONS = plot::Point::defaultOptions ):
alias( OPTIONNAMES = plot::Point::optionNames ):

plot::Point:= newDomain("plot::Point"):
plot::Point::info:= "plot::Point -- graphical primitive for two- and three-dimensional points":
plot::Point::type:= "graphprim":
plot::Point::dimension:= p -> nops((extop(p,1))::coords):

plot::Point::new:= proc(x, y, z)
    local c, p, i, opt, options;
begin
    if args(0) = 0 then
        error("no arguments given")
    end_if;

    if testtype(x,Type::Arithmetical) then
        if not testtype(y,Type::Arithmetical) then
            error("2nd argument: expecting an arithmetical expression")
        elif args(0) > 2 and testtype(z,Type::Arithmetical) then
            c:= [x,y,z];
            opt:= [args(4..args(0))];
        else
            c:= [x,y];
            opt:= [args(3..args(0))];
        end_if;

        options:= DEFAULTOPTIONS; // default options (a table)
        // check user options and move override the default options:
        for i from 1 to nops(opt) do
            if (z:= plot::Point::checkOption(opt[i]))[1] = FALSE then
                error(z[2])
            else
                options[z[2]]:= z[3]
            end_if
        end_for;

        p:= newDomain( expr2text(genident("plot::Point_")) );

        // attributes:
        p::plotdata       := [];
        p::refreshPlotdata:= TRUE;
        p::coords         := c;
        p::options        := options;

        return(new(plot::Point, p))  
    elif (p:= plot::Point::convert(x)) = FAIL then
        error("invalid point specification")
    else
        options:= DEFAULTOPTIONS; // default options (a table)
        opt:= [args(2..args(0))]; // user options (a list)
        // check user options and override the default options:
        for i from 1 to nops(opt) do
            if (z:= plot::Point::checkOption(opt[i]))[1] = FALSE then
                error(z[2])
            else
                options[z[2]]:= z[3]
            end_if
        end_for;
        c:= extop(p,1);
        c::options:= options;
        return(extsubsop( p,1=c ))
    end_if
end_proc:

plot::Point::convert := proc(o)
    local p, n;
begin
    case domtype(o)
    of plot::Point do
        return(o)
    of DOM_POINT do
        p:= newDomain( expr2text(genident("plot::Point_")) );
        p::plotdata:= [];
        p::refreshPlotdata:= TRUE;
        p::options:= DEFAULTOPTIONS;
        // keep the points to the attribute 'coords':
        n:= nops(o);
        p::coords:= [op(o,1..n-1)];
        p:= new(plot::Point, p);
        // the last operand of a 'DOM_POINT' is either NIL or
        // a RGB color (force argument checking!)
        if domtype(op(o,n)) = DOM_LIST then
            p::Color:= op(o,n)
        end_if;
        return( p )
    of DOM_LIST do
        if testtype(o, Type::ListOf(Type::Arithmetical, 2, 3)) then
            p:= newDomain( expr2text(genident("plot::Point_")) );
            p::plotdata:= [];
            p::refreshPlotdata:= TRUE;
            // keep the points to the entry 'coords':
            p::coords:= o;
            p::options:= DEFAULTOPTIONS;
            return(new(plot::Point, p))
        end_if
    otherwise
        return(FAIL)
    end_case
end_proc:

plot::Point::convert_to := proc(p, T)
    local e;
begin
    case T
    of plot::Point do
        return(p)
    of DOM_POINT do
        e:= extop(p,1);
        if type((T:= e::options[Color])) <> "_index" then
            return( point(op(float(e::coords)), hold(Color) = T[2]) )
        else
            return( point(op(float(e::coords))) )
        end_if
    of DOM_LIST do
        e:= extop(p,1);        
        return(e::coords)
    // not documented:
    of matrix do
        e:= extop(p,1);
        return(matrix(map(e::coords, i -> [i])))
    otherwise
        return(FAIL)
    end_case
end_proc:

plot::Point::expr := p -> plot::Point::convert_to(p,DOM_POINT):

plot::Point::copy := proc(f)
begin
    return(extsubsop(f, 1=newDomain(genident("plot::Point_"), extop(f, 1))))
end_proc:

// not documented:
plot::Point::rotate := proc(p, xphi, yphi, zphi)
    local A, pc, turnx, turny, turnz;
begin
    turnx:= matrix([[1, 0, 0], [0,cos(xphi), sin(xphi)],[0, -sin(xphi), cos(xphi)]]):
    turny:= matrix([[cos(yphi), 0, sin(yphi)],[0,1,0],[-sin(yphi), 0, cos(yphi)]]):
    turnz:= matrix([[cos(zphi), sin(zphi),0],[-sin(zphi), cos(zphi), 0],[0,0,1]]): 
    A:= turnx * turny * turnz * plot::Point::convert_to(p, matrix);

    pc:= plot::Point::copy(p);
    pc::coords := [op(A)];
    pc::refreshPlotdata:= TRUE;
    return(pc)
end_proc:

// not documented:
plot::Point::move := (p,x,y,z) -> p + plot::Point(x,y,z):

// not documented:
plot::Point::_subtract := proc(p1:plot::Point, p2:plot::Point)
begin
    return( plot::Point::_plus(p1, plot::Point::_negate(p2)) )
end_proc:

// not documented:
plot::Point::_mult := proc(r, p)
    local i;
begin
    case args(0)
    of 0 do
       return( 1 )
    of 1 do
       return(r)
    otherwise
        if domtype(r) = plot::Point and domtype(p) = plot::Point then
            r:= extop(r,1); p:= extop(p,1);
            r:= r::coords; p:= p::coords;
        
            if nops(r) <> nops(p) then
                error("incompatible dimension")
            end_if;

            return( _mult(_plus( (r[i] * p[i]) $ i = 1..nops(p) ), args(3..args(0)) ) );
        elif domtype(p) = plot::Point then
            p:= extop(p,1);
            p:= p::coords;

            return( _mult( plot::Point( r * p[i] $ i = 1..nops(p) ), args(3..args(0)) ) );
        elif domtype(r) = plot::Point then
            r:= extop(r,1);
            r:= r::coords;
    
            return( _mult( plot::Point( p * r[i] $ i = 1..nops(r) ), args(3..args(0)) ) );
        else
            return( _mult(_mult(r,p),args(3..args(0))) )
        end_if
    end_case
end_proc:

// not documented:
plot::Point::abs := proc(p)
    local i;    
begin
    p:= extop(p,1);
    p:= p::coords;
    return( sqrt(_plus( p[i]^2 $ i = 1..nops(p) )) )
end_proc:

// not documented:
plot::Point::_negate := proc(p)
    local q;
begin
    q:= map( (extop(p,1))::coords,_mult,-1 );
    return( plot::Point(op(q)) )
end_proc:

// not documented:
plot::Point::_plus := proc()
    local p, i, n;
begin
    p := [args()];

    if map({op(p)},domtype) <> {plot::Point} then
        error("invalid argument (expecting objects of the domain 'plot::Point')")
    else
        p:= map(p, e -> (extop(e,1))::coords);
        n := { (op@map)(p, nops) };
        if nops(n) > 1 then
            error("incompatible dimension");
        else
            return(plot::Point(
                _plus( (p[j])[i] $ j = 1..nops(p) ) $ i = 1..op(n,1)  
            ))
        end_if
    end_if
end_proc:

plot::Point::modify := proc(p)
    local o, m, options;
begin
    m:= plot::Point::copy(p);
    options := [args(2..args(0))];

    ( slot(m, expr2text(op(o, 1)), op(o, 2)) ) $ o in options;
    
    m::refreshPlotdata := TRUE;
    return(m)    
end_proc:

plot::Point::getPlotdata := proc(p)
    local e;
begin
    e:= extop(p,1);
    if e::refreshPlotdata = FALSE then return( e::plotdata ) end_if;

    p:= point( op(float(e::coords)) );
    e::refreshPlotdata:= FALSE;
    // create the plot data:
    e::plotdata:= [ [hold(Mode) = hold(List), [p]] . [op(e::options)] ]
end_proc:

plot::Point::print := p -> hold(plot::Point)(op((extop(p,1))::coords)):

plot::Point::op := proc(p, n)
begin
    if args(0) < 1 or args(0) > 2 then
        error("invalid number of arguments")
    end_if;

    p:= (extop(p,1))::coords;
    if args(0) = 1 then op(p) else op(p, n) end_if
end_proc:

plot::Point::nops := p -> nops((extop(p,1))::coords):

plot::Point::_index := proc(p,n:Type::PosInt)
begin
    if args(0) <> 2 then
        error("expecting two arguments")
    end_if;

    p:= (extop(p,1))::coords;
    if n > nops(p) then
        error("index out of range")
    else
        return( p[n] )
    end_if
end_proc:

// Note: Do not return an error message in the case of an invalid call,
//       because then the object would be destroyed! Hence a warning 
//       message is issued, and the call of this method has no effect 
//       on the object.
plot::Point::set_index := proc(p, n:Type::PosInt, v)
    local e, l;
begin
    if args(0) <> 3 then
        warning("wrong number of arguments; assignment ignored");
        return(p)
    elif not testtype(v,Type::Arithmetical) then
        warning("3rd argument: expecting an arithmetical expression; assignment ignored");
        return(p)
    end_if; 
 
    e:= extop(p,1);
    l:= e::coords;
    if n > nops(l) then
        warning("index out of range; assignment ignored");
        return(p)
    else
        e::refreshPlotdata:= TRUE;
        e::coords:= subsop( l,n=v );
        return( extsubsop(p,1=e) )
    end_if
end_proc:

// Note: Do not return an error message in the case of an invalid call,
//       because then the object would be destroyed! Hence a warning 
//       message is issued, and the call of this method has no effect 
//       on the object.
plot::Point::slot:= proc(o,s,v=null())
    local e, t, _s;
begin
    if args(0) = 3 then
        // slot assignment:
        e:= extop(o,1);

        // first handle all user-accessable attributes:
        case s 
        of "refreshPlotdata" do
            if v <> TRUE and v <> FALSE then
                warning("attribute 'refreshPlotdata': expecting either 'TRUE' or 'FALSE'; assignment ignored");
                return( o )
            else
                e::refreshPlotdata:= v;
                return( extsubsop(o,1=e) )
            end_if
        of "coords" do
            t:= nops(e::coords);
            if not testtype(v,Type::ListOf(Type::Arithmetical,t,t)) then
                warning("attribute 'coords': expecting a list of ".t." arithmetical expressions; assignment ignored");
                return( o )
            else
                e::coords:= v;
                e::refreshPlotdata:= TRUE;
                return( extsubsop(o,1=e) )
            end_if
        of "options" do
            if domtype(v) <> DOM_TABLE then
                warning("attribute 'options': expecting a table of plot options; assignment ignored");
                return( o )
            else
                e::options:= v;
                e::refreshPlotdata:= TRUE;
                return( extsubsop(o,1=e) )
            end_if
        end_case;

        // an option given?
        t:= plot::Point::checkOption(text2expr(s) = v);
        if t[1] = FALSE then
            warning( t[2]."; assignment ignored" );
            return(o)
        else
            e::options[t[2]]:= t[3];
            e::refreshPlotdata:= TRUE;
            return( extsubsop(o,1=e) )
        end_if
    else
        // slot access:
        e:= extop(o,1);
        if (t:= slot(e,s)) <> FAIL then
            return(t)
        else
            _s:= text2expr(s);
            if not contains(OPTIONNAMES,_s) then
                error("unknown attribute or option '".s."'")
            elif type((t:= e::options[_s])) <> "_index" then
                return(t)
            else
                return(FAIL)
            end_if
        end_if
    end_if
end_proc:

plot::Point::checkOption := proc(opt)
    local value;
begin
    if type(opt) <> "_equal" then
        return( [FALSE,"options must have the form 'OptionName = value'"] )
    end_if;

    value:= op(opt,2); opt:= op(opt,1);
    if not contains( OPTIONNAMES,opt ) then
        return( [FALSE,"unknown option name '".expr2text(opt)."'"] )
    else
        // First, handle special cases for certain options:
        // Color : Beside the standard format 'Color = [{Height,Flat},[r,g,b]]',
        //         allow 'Color = [r,g,b]' which corresponds to '[Flat,[r,g,b]]'.
        case opt
        of hold(Color) do
            if (value:= plot::getColorFormat(value)) = FAIL then
                return( [FALSE,"option 'Color': invalid color specification"] )
            end_if
        end_case;

        return( [TRUE,opt,value] )
    end_if
end_proc:

OPTIONNAMES:= {
  // options for graphical objects:
  hold(Color),hold(LineStyle),hold(LineWidth),hold(PointStyle),
  hold(PointWidth),hold(Title),hold(TitlePosition)
}:

DEFAULTOPTIONS:= table():
// Default options for objects of the domain plot::Point:
// Example: DEFAULTOPTIONS[hold(Color)]:= RGB::Blue:

