// $Date: 2000/11/14 10:05:14 $  $Author: frankp $   $Revision: 1.15 $
// igel
// renamed to plot::Pointlist and completely rewritten by frankp (22.09.2000)

/* -----------------------------------------------------------
    Pointlist -- The graphical primitive for lists of either two- or three-
                 dimensional points

    Syntax:
    Pointlist(p1 <,p2, ...> <, op1, op2>)

    p1, p2, ...     : either lists of two or three arithmetical expressions or 
                      expressions of type plot::Point
    op1, op2, ...   : plot2d or plot3d options, see plot2d and plot3d for details.

    Pointlist offers a special option named 'DrawMode' that can either have one of the
    following values: Connected, Vertical, Horizontal or None or it can be a
    set of them, i.e.: 'DrawMode = {Connected, Vertical}'.

    Example:
    >> plot::Pointlist( plot::Point(i, i^2) $ i = 1..10 )

    >> plot::Pointlist( [i, i^2] $ i = 1..10, DrawMode = Connected)
----------------------------------------------------------------*/ 
alias( DEFAULTOPTIONS = plot::Pointlist::defaultOptions ):
alias( OPTIONNAMES = plot::Pointlist::optionNames ):

plot::Pointlist := newDomain("plot::Pointlist"):
plot::Pointlist::info:= "plot::Pointlist -- graphical primitive for representing a sequence of points":
plot::Pointlist::type:= "graphprim":
plot::Pointlist::dimension:= p -> nops( ((extop(p,1))::objects[1])::coords ):

plot::Pointlist::new := proc()
    local p, i, opt, options, points;
begin
    if args(0) = 0 then
        error("no arguments given");
    end_if;

    points:= [args()];
    for i from 1 to nops(points) do
        if (p:= plot::Point::convert(points[i])) = FAIL then
            break
        else
            points[i]:= p
        end_if
    end_for;

    opt:= [op(points,i..nops(points))]; // user specified options (a list)
    points:= [op(points,1..i-1)];
    if points = [] then
        error("no points given, or illegal specification of points")
    else
        p:= map( {op(points)}, e -> e::dom::dimension(e) );
        if nops(p) <> 1 then
            error("points have different dimension")
        end_if
    end_if;

    options:= DEFAULTOPTIONS; // default options (a table)
    // check user options and override the default options:
    for i from 1 to nops(opt) do
        if (p:= plot::Pointlist::checkOption(opt[i]))[1] = FALSE then
            error(p[2])
        else
            options[p[2]]:= p[3]
        end_if
    end_for;

    p:= newDomain( expr2text(genident("plot::Pointlist_")) );

    // attributes:
    p::plotdata       := [];
    p::refreshPlotdata:= TRUE;
    p::objects        := points;
    p::options        := options;

    return(new(plot::Pointlist, p))  
end_proc:

plot::Pointlist::print := () -> hold(plot::Pointlist)():

plot::Pointlist::expose := p -> hold(plot::Pointlist)(op(map((extop(p,1))::objects,expose))):

plot::Pointlist::copy := proc(o)
    local l, e;
begin
    e:= extop(o,1);
    l:= map( e::objects, x -> x::dom::copy(x) );
    e:= newDomain(genident("plot::Pointlist_"), e);
    e::objects:= l;
    return( extsubsop(o, 1=e) )
end_proc:

plot::Pointlist::op := proc(p, n)
begin
    if args(0) < 1 or args(0) > 2 then
        error("invalid number of arguments")
    end_if;

    p:= (extop(p,1))::objects;
    if args(0) = 1 then op(p) else op(p, n) end_if
end_proc:

plot::Pointlist::nops := p -> nops((extop(p,1))::objects):

plot::Pointlist::_index := proc(p:plot::Pointlist,n:Type::PosInt)
begin
    if args(0) <> 2 then
        error("expecting two arguments")
    end_if;

    p:= (extop(p,1))::objects;
    if n > nops(p) then
        error("index out of range")
    else
        return( p[n] )
    end_if
end_proc:

// Note: Do not return an error message in the case of an invalid call,
//       because then the object would be destroyed! Hence a warning 
//       message is issued, and the call of this method has no effect 
//       on the object.
plot::Pointlist::set_index := proc(s:plot::Pointlist, n:Type::PosInt, v)
    local l, e;
begin
    if args(0) <> 3 then
        warning("wrong number of arguments; assignment ignored");
        return(s)
    elif domtype(v) <> plot::Point then
        warning("3rd argument: expecting an object of the domain 'plot::Point'; assignment ignored");
        return(s)
    end_if;
 
    e:= extop(s,1);
    if nops((e::objects[1])::coords) <> nops(v) then
        warning("3rd argument: point has an invalid dimension; assignment ignored");
        return(s)
    end_if; 

    l:= e::objects;
    if n > nops(l) then
        warning("index out of range; assignment ignored");
        return(s)
    else
        e::objects:= subsop( l,n=v );
        e::refreshPlotdata:= TRUE;
        return( extsubsop(s,1=e) )
    end_if
end_proc:

plot::Pointlist::getPlotdata := proc(c)
    local e, mode, drawMode, points, options, vertexes;
begin
    e:= extop(c,1);
    if e::refreshPlotdata = FALSE then return( e::plotdata ) end_if;

    // subroutine:
    drawMode:= proc(mode)
        name plot::Pointlist::getPlotdata;
        local i;
    begin
        case mode
        of hold(Connected) do
            return( polygon( op(vertexes) ) );
        of hold(Vertical) do
            return( polygon( vertexes[i],subsop(vertexes[i],2=0) ) $ i = 1..nops(vertexes) );
        of hold(Horizontal) do
            return( polygon( vertexes[i],subsop(vertexes[i],1=0) ) $ i = 1..nops(vertexes) );
        otherwise
            error("option 'DrawMode': expecting either 'Connected', 'Vertical', 'Horizontal', or a set of them");
        end_case
    end_proc:


    // create the plot data of the points:
    options:= e::options;
    mode:= options[hold(DrawMode)];
    // remove specific Pointlist options:
    delete options[hold(DrawMode)];
    options:= [op(options)]; // convert table into a list

    points:= map( e::objects, 
        proc(o)
            local p, e, point_options;
        begin
            p:= point( op(float([op(o)])) );
            // build the list of plot options:
            point_options:= [op(o::options)];
            for e in options do
                if not has(point_options,op(e,1)) then
                    point_options:= append(point_options, e)
                end_if
            end_for;
            // create the plot data:
            [hold(Mode) = hold(List), [p]] . point_options
        end
    );

    e::refreshPlotdata:= FALSE;
    // create the plot data:
    if mode = hold(None) then
        e::plotdata:= points
    else

        // To create the list of vertexes from the sequence of points,
        // convert each point to an object of the domain DOM_POINT.
        // Note: Plot options (such as the color information) are taken
        // from the options of the given point-list object 'c'.
        vertexes:= map( e::objects, p -> point(op(float([op(p)]))) );
        if domtype(mode) = DOM_SET then
            e::plotdata:= points . [[hold(Mode) = hold(List), map([op(mode)], drawMode)] . options]
        else
            e::plotdata:= points . [[hold(Mode) = hold(List), [drawMode(mode)]] . options]
        end_if
    end_if
end_proc:

plot::Pointlist::modify := proc(p)
    local o, options;
begin
    p:= plot::Pointlist::copy(p);
    options:= [args(2..args(0))];

    ( slot(p, expr2text(op(o, 1)), op(o, 2)) ) $ o in options;
    
    p::refreshPlotdata := TRUE;
    return(p)
end_proc:

// Note: Do not return an error message in the case of an invalid call,
//       because then the object would be destroyed! Hence a warning 
//       message is issued, and the call of this method has no effect 
//       on the object.
plot::Pointlist::slot:= proc(o,s,v=null())
    local e, t, _s;
begin
    if args(0) = 3 then
        // slot assignment:
        e:= extop(o,1);

        // first handle all user-accessable attributes:
        case s 
        of "refreshPlotdata" do
            if v <> TRUE and v <> FALSE then
                warning("attribute 'refreshPlotdata': expecting either 'TRUE' or 'FALSE'; assignment ignored");
                return( o )
            else
                e::refreshPlotdata:= v;
                return( extsubsop(o,1=e) )
            end_if
        of "objects" do
            if not testtype(v,Type::ListOf(plot::Point)) then
                warning("attribute 'objects': expecting a list of objects of the domain 'plot::Point'; assignment ignored");
                return( o )
            elif map({op(v)}, x -> x::dom::dimension(x)) <> {nops( (e::objects[1])::coords )} then
                warning("attribute 'objects': points have invalid dimensions; assignment ignored");
                return( o )
            else
                e::objects:= v;
                e::refreshPlotdata:= TRUE;
                return( extsubsop(o,1=e) )
            end_if
        of "options" do
            if domtype(v) <> DOM_TABLE then
                warning("attribute 'options': expecting a table of plot options; assignment ignored");
                return( o )
            else
                e::options:= v;
                e::refreshPlotdata:= TRUE;
                return( extsubsop(o,1=e) )
            end_if
        end_case;

        // an option given?
        t:= plot::Pointlist::checkOption(text2expr(s) = v);
        if t[1] = FALSE then
            warning( t[2]."; assignment ignored" );
            return(o)
        else
            e::options[t[2]]:= t[3];
            e::refreshPlotdata:= TRUE;
            return( extsubsop(o,1=e) )
        end_if
    else
        // slot access:
        e:= extop(o,1);
        if (t:= slot(e,s)) <> FAIL then
            return(t)
        else
            _s:= text2expr(s);
            if not contains(OPTIONNAMES,_s) then
                error("unknown attribute or option '".s."'")
            elif type((t:= e::options[_s])) <> "_index" then
                return(t)
            else
                return(FAIL)
            end_if
        end_if
    end_if
end_proc:

plot::Pointlist::checkOption := proc(opt)
    local value, admit;
begin
    if type(opt) <> "_equal" then
        return( [FALSE,"options must have the form 'OptionName = value'"] )
    end_if;

    value:= op(opt,2); opt:= op(opt,1);
    if not contains( OPTIONNAMES,opt ) then
        return( [FALSE,"unknown option name '".expr2text(opt)."'"] )
    else
        // First, handle special cases for certain options:
        // Color : Beside the standard format 'Color = [{Height,Flat},[r,g,b]]',
        //         allow 'Color = [r,g,b]' which corresponds to '[Flat,[r,g,b]]'.
        // DrawMode: Connected, Vertical, Horizontal or None, or a set of them
        case opt
        of hold(Color) do
            if (value:= plot::getColorFormat(value)) = FAIL then
                return( [FALSE,"option 'Color': invalid color specification"] )
            end_if;
            break
        of hold(DrawMode) do
            admit:= {hold(Connected),hold(Vertical),hold(Horizontal),hold(None)};
            if domtype(value) = DOM_SET and value minus admit <> {} 
               or domtype(value) <> DOM_SET and not contains(admit,value)
            then
                return( [FALSE,"option 'DrawMode': expecting either 'Conntected', 'Vertical', 'Horizontal' or 'None', or a set of them"] )
            end_if
        end_case;

        return( [TRUE,opt,value] )
    end_if
end_proc:

OPTIONNAMES:= {
  // options for graphical objects:
  hold(Color),hold(LineStyle),hold(LineWidth),hold(PointStyle),
  hold(PointWidth),hold(Title),hold(TitlePosition),
  // options for plot::Pointlist:
  hold(DrawMode)
}:

DEFAULTOPTIONS:= table():
// Default options for objects of the domain plot::Pointlist:
DEFAULTOPTIONS[hold(DrawMode)]:= hold(None):


