// $Date: 2000/11/14 10:05:14 $  $Author: frankp $   $Revision: 1.22 $
// igel
// completely rewritten by frankp (22.09.2000)

/* -----------------------------------------------------------
    Polygon -- The graphical primitive for two- and three-dimensional polygons

    Syntax:

    Polygon(p1 <, p2, ...> <, op1, op2, ... >)

    p1, p2, ...  : either graphical object of type plot::Point or lists of
                   two or three arithmetical expressions
    op1, op2, ...: plot2d and polygon options, see plot2d, plot3d and polygon 
                   for details.

    Example:
    >> plot::Polygon( plot::Point(1,2), plot::Point(10, 2), plot::Point(10, 10) )

    >> plot::Polygon( plot::Point(1,2), plot::Point(10, 2), plot::Point(10, 10), Closed = TRUE )

    Note: The plot structure of a polygon p is [Mode=List,[p],Color=...,...],
          i.e., the color is not stored in p!

----------------------------------------------------------------*/ 
alias( DEFAULTOPTIONS = plot::Polygon::defaultOptions ):
alias( OPTIONNAMES = plot::Polygon::optionNames ):

plot::Polygon := newDomain("plot::Polygon"):
plot::Polygon::info:= "plot::Polygon -- graphical primitive for two- and three-dimensional polygons":
plot::Polygon::type:= "graphprim":
plot::Polygon::dimension:= p -> nops( ((extop(p,1))::objects[1])::coords ):

plot::Polygon::new:= proc(p)
    local e, i, d, opt, options, points;
begin
    if args(0) = 0 then 
        error("no arguments given")
    end_if;

    case domtype(p)
    of DOM_POLYGON   do
        p:= plot::Polygon::convert(p)
    of plot::Polygon do
        opt:= [args(2..args(0))];
        if opt = [] then 
            return( p )
        else
            options:= DEFAULTOPTIONS;  // default options (a table)
            // check user options and move override the default options:
            for i from 1 to nops(opt) do
                if (d:= plot::Polygon::checkOption(opt[i]))[1] = FALSE then
                    error(d[2])
                else
                    options[d[2]]:= d[3]
                end_if
            end_for;
            e:= extop(p,1);
            e::options:= options;
            return( extsubsop(p,1=e) )
        end_if
    otherwise
        points:= [args()];
        for i from 1 to nops(points) do
            if (p:= plot::Point::convert(points[i])) = FAIL then
                break
            else
                points[i]:= p
            end_if
        end_for;
        opt:= [op(points,i..nops(points))];
        if (points:= [op(points,1..i-1)]) = [] then
            error("no points specified")
        elif nops( {op(map(points, x -> x::dom::dimension(x)))} ) <> 1 then
            error("points have different dimensions")
        end_if;

        options:= DEFAULTOPTIONS;  // default options (a table)
        // check user options and move override the default options:
        for i from 1 to nops(opt) do
            if (e:= plot::Polygon::checkOption(opt[i]))[1] = FALSE then
                error(e[2])
            else
                options[e[2]]:= e[3]
            end_if
        end_for;

        p:= newDomain( expr2text(genident("plot::Polygon_")) );

        // attributes:
        p::plotdata       := [];
        p::refreshPlotdata:= TRUE;
        p::objects        := points;
        p::options        := options;

        return(new(plot::Polygon, p))  
    end_case
end_proc:

plot::Polygon::convert := proc(o)
    local p;
begin
    case domtype(o)
    of plot::Polygon do
        return(o)
    of DOM_POLYGON do
        p:= newDomain( expr2text(genident("plot::Polygon_")) );
        p::plotdata:= [];
        p::refreshPlotdata:= TRUE;
        p::options:= DEFAULTOPTIONS;

        o:= split([op(o)], t -> domtype(t) = DOM_POINT);
        p::objects:= map(o[1], plot::Point::convert);
        p::options[hold(Closed)]:= o[2][2];
        p::options[hold(Filled)]:= o[2][3];
        if o[2][1] <> NIL then p::options[hold(Color)]:= [hold(Flat),o[2][1]] end_if;

        return(new(plot::Polygon, p))  
    otherwise
        return(FAIL)
    end_case
end_proc:

plot::Polygon::convert_to := proc(p, T)
    local e, color, points;
begin
    e:= extop(p,1);
    case T
    of plot::Polygon do
        return(p)
    of DOM_POLYGON do
        if contains( (points:= map(e::objects,plot::Point::convert_to,DOM_POINT)),FAIL ) <> 0 then
            return( FAIL )
        elif type((color:= e::options[hold(Color)])) = "_index" then
            color:= null()
        else
            color:= hold(Color) = color[2]
        end_if;
        return(polygon(op(points), color, 
            hold(Closed) = e::options[hold(Closed)], 
            hold(Filled) = e::options[hold(Filled)]
        ))
    otherwise
        return(FAIL)
    end_case
end_proc:

plot::Polygon::expr := p -> plot::Polygon::convert_to(p,DOM_POLYGON):

plot::Polygon::copy := proc(p)
    local l, e;
begin
    e:= extop(p,1);
    l:= map( e::objects, x -> x::dom::copy(x) );
    e:= newDomain(genident("plot::Polygon_"), e);
    e::objects:= l;
    return( extsubsop(p, 1=e) )
end_proc:

plot::Polygon::modify := proc(p)
    local o, m, options;
begin
    m:= plot::Polygon::copy(p);
    options:= [args(2..args(0))];

    ( slot(m, expr2text(op(o, 1)), op(o, 2)) ) $ o in options;

    m::refreshPlotdata:= TRUE;
    return(m) 
end_proc:

/*-- 
    Notes: In the two-dimensional case, plot2d already calls 
    plotlib::polygonplot to plot closed or filled polygons 
    with more than three points!
--*/
plot::Polygon::getPlotdata := proc(p)
    local e, color, points, options;
begin
    e:= extop(p, 1);
    if e::refreshPlotdata = FALSE then return( e::plotdata ) end_if;

    points:= map(e::objects, plot::Point::convert_to, DOM_POINT);
    if contains( points,FAIL ) <> 0 then
        error("invalid plot data")
    end_if;

    options:= e::options;
    if type(options[hold(Color)]) <> "_index" then 
        color:= hold(Color) = options[hold(Color)][2]
    else
        color:= null()
    end_if;

    p:= [polygon(op(points), 
      hold(Filled) = options[hold(Filled)], hold(Closed) = options[hold(Closed)], color
    )];

    // remove specific Polygon options and options with value NIL:
    delete options[hold(Color)], options[hold(Filled)], options[hold(Closed)];

    e::refreshPlotdata:= FALSE;
    // create the plot data:
    e::plotdata:= [[hold(Mode) = hold(List), p] . [op(options)]]
end_proc:

plot::Polygon::print := () -> hold(plot::Polygon)():

plot::Polygon::expose := p -> hold(plot::Polygon)(op(map((extop(p,1))::objects,expose))):

// undocumented:
plot::Polygon::rotate := proc(p, p1, p2, p3)
    local c, e;
begin
    c:= plot::Polygon::copy(p);
    e:= extop(c,1);
    e::objects:= map(e::objects, plot::Point::rotate, p1, p2, p3);
    e::refreshPlotdata:= TRUE;
    return(extsubsop(c,1=e))
end_proc;

// undocumented:
plot::Polygon::move := proc(p, p1, p2, p3)
    local c, e;
begin
    c:= plot::Polygon::copy(p);
    e:= extop(c,1);
    e::objects := map(e::objects, plot::Point::move, p1, p2, p3);
    e::refreshPlotdata:= TRUE;
    return(extsubsop(c,1=e))
end_proc;

plot::Polygon::op := proc(p, n)
begin
    if args(0) < 1 or args(0) > 2 then
        error("invalid number of arguments")
    end_if;

    p:= (extop(p,1))::objects;
    if args(0) = 1 then op(p) else op(p, n) end_if
end_proc:

plot::Polygon::nops := p -> nops((extop(p,1))::objects):

plot::Polygon::_index := proc(p:plot::Polygon,n:Type::PosInt)
begin
    if args(0) <> 2 then
        error("expecting two arguments")
    end_if;

    p:= (extop(p,1))::objects;
    if n > nops(p) then
        error("index out of range")
    else
        return( p[n] )
    end_if
end_proc:

// Note: Do not return an error message in the case of an invalid call,
//       because then the object would be destroyed! Hence a warning 
//       message is issued, and the call of this method has no effect 
//       on the object.
plot::Polygon::set_index := proc(s:plot::Polygon, n:Type::PosInt, v)
    local l, e;
begin
    if args(0) <> 3 then
        warning("wrong number of arguments; assignment ignored");
        return(s)
    elif domtype(v) <> plot::Point then
        warning("3rd argument: expecting an object of the domain 'plot::Point'; assignment ignored");
        return(s)
    end_if; 
 
    e:= extop(s,1);
    l:= e::objects;
    if n > nops(l) then
        warning("index out of range; assignment ignored");
        return(s)
    else
        e::objects:= subsop( l,n=v );
        e::refreshPlotdata:= TRUE;
        return( extsubsop(s,1=e) )
    end_if
end_proc:

// Note: Do not return an error message in the case of an invalid call,
//       because then the object would be destroyed! Hence a warning 
//       message is issued, and the call of this method has no effect 
//       on the object.
plot::Polygon::slot:= proc(o,s,v=null())
    local e, t, _s;
begin
    if args(0) = 3 then
        // slot assignment:
        e:= extop(o,1);

        // first handle all user-accessable attributes:
        case s 
        of "refreshPlotdata" do
            if v <> TRUE and v <> FALSE then
                warning("attribute 'refreshPlotdata': expecting either 'TRUE' or 'FALSE'; assignment ignored");
                return( o )
            else
                e::refreshPlotdata:= v;
                return( extsubsop(o,1=e) )
            end_if
        of "objects" do
            if not testtype(v,Type::ListOf(plot::Point)) then
                warning("attribute 'objects': expecting a list of objects of the domain 'plot::Point'; assignment ignored");
                return( o )
            elif map({op(v)}, x -> x::dom::dimension(x)) <> {nops((e::objects[1])::coords)} then
                warning("attribute 'objects': points have invalid dimensions; assignment ignored");
                return( o )
            else
                e::objects:= v;
                e::refreshPlotdata:= TRUE;
                return( extsubsop(o,1=e) )
            end_if
        of "options" do
            if domtype(v) <> DOM_TABLE then
                warning("attribute 'options': expecting a table of plot options; assignment ignored");
                return( o )
            else
                e::options:= v;
                e::refreshPlotdata:= TRUE;
                return( extsubsop(o,1=e) )
            end_if
        end_case;

        // an option given?
        t:= plot::Polygon::checkOption(text2expr(s) = v);
        if t[1] = FALSE then
            warning( t[2]."; assignment ignored" );
            return(o)
        else
            e::options[t[2]]:= t[3];
            e::refreshPlotdata:= TRUE;
            return( extsubsop(o,1=e) )
        end_if
    else
        // slot access:
        e:= extop(o,1);
        if (t:= slot(e,s)) <> FAIL then
            return(t)
        else
            _s:= text2expr(s);
            if not contains(OPTIONNAMES,_s) then
                error("unknown attribute or option '".s."'")
            elif type((t:= e::options[_s])) <> "_index" then
                return(t)
            else
                return(FAIL)
            end_if
        end_if
    end_if
end_proc:

plot::Polygon::checkOption := proc(opt)
    local value;
begin
    if type(opt) <> "_equal" then
        return( [FALSE,"options must have the form 'OptionName = value'"] )
    end_if;

    value:= op(opt,2); opt:= op(opt,1);
    if not contains( OPTIONNAMES,opt ) then
        return( [FALSE,"unknown option name '".expr2text(opt)."'"] )
    else
        // First, handle special cases for certain options:
        // Color : Beside the standard format 'Color = [{Height,Flat},[r,g,b]]',
        //         allow 'Color = [r,g,b]' which corresponds to '[Flat,[r,g,b]]'.
        case opt
        of hold(Color) do
            if (value:= plot::getColorFormat(value)) = FAIL then
                return( [FALSE,"option 'Color': invalid color specification"] )
            end_if;
            break
        of hold(Closed) do
        of hold(Filled) do
            if value <> TRUE and value <> FALSE then
                return( [FALSE,"option '".expr2text(opt)."': expecting either 'TRUE' or 'FALSE'"] )
            end_if
        end_case;

        return( [TRUE,opt,value] )
    end_if
end_proc:

OPTIONNAMES:= {
  // options for graphical objects:
  hold(Color),hold(LineStyle),hold(LineWidth),hold(PointStyle),
  hold(PointWidth),hold(Title),hold(TitlePosition),
  // options for polygons:
  hold(Closed),hold(Filled)
}:

DEFAULTOPTIONS:= table():
// Default options for objects of the domain plot::Polygon:
DEFAULTOPTIONS[hold(Closed)]:= FALSE:
DEFAULTOPTIONS[hold(Filled)]:= FALSE:

