// $Date: 2000/11/14 10:05:15 $  $Author: frankp $   $Revision: 1.19 $
// igel
// completely rewritten by frankp (22.09.2000)

/* -----------------------------------------------------------
    Rectangle2d -- The graphical primitive for two-dimensional rectangles

    Syntax:

    Rectangle2d(ll, w, h, <, op1, op2, ...>)

    ll              : either list or point (specifiying the (l)ower (l)eft  
                      corner)
    w, h            : arithmetical expressions (specifiying width and height)
    op1, op2, ...   : plot2d options, see plot2d and plot3d for details.

    Example:
    >> plot::Rectangle2d( [1, 2], 3, 3 )

    >> plot::Rectangle2d( [1, 2], 3, 3, Filled = TRUE )

    >> plot::Rectangle2d( [1, 2], 3, 3, Color = RGB::Green )

----------------------------------------------------------------*/
alias( DEFAULTOPTIONS = plot::Rectangle2d::defaultOptions ):
alias( OPTIONNAMES = plot::Rectangle2d::optionNames ):

plot::Rectangle2d := newDomain("plot::Rectangle2d"):
plot::Rectangle2d::info:= "plot::Rectangle2d -- graphical primitive for rectangles":
plot::Rectangle2d::type:= "graphprim":
plot::Rectangle2d::dimension:= () -> 2:

plot::Rectangle2d::new := proc(m, w, h)
    local i, o, opt, options, rectangle;
begin
    if args(0) < 3 then
        error("expecting at least three arguments");
    end_if;

    case domtype(m)
    of DOM_LIST do
        if nops(m) <> 2 or not testtype(m[1], Type::Arithmetical) or not  testtype(m[2], Type::Arithmetical) then
            error("1st argument: expecting a list of two arithmetical expressions")
        end_if;
        break
    of DOM_POINT do
        if nops(m) <> 3 then
            error("1st argument: expecting a two-dimensional point")
        end_if;
        m:= [op(m,1..2)];
        break
    of plot::Point do
        if nops(m) <> 2 then
            error("1st argument: expecting a two-dimensional point")
        end_if;
        m:= [op(m)];
        break
    otherwise
        error("1st argument: expecting a list of two-arithmetical expressions, or a two-dimensional point")
    end_case;

    if not testtype(w, Type::Arithmetical) then
        error("2nd argument: expecting an arithmetical expression");
    elif not testtype(h, Type::Arithmetical) then
        error("3rd argument: expecting an arithmetical expression");
    end_if;

    options:= DEFAULTOPTIONS;         // default options (a table)
    opt:= [args(4..args(0))];  // user options (a list)
    // check user options and override the default options:
    for i from 1 to nops(opt) do
        if (o:= plot::Rectangle2d::checkOption(opt[i]))[1] = FALSE then
            error(o[2])
        else
            options[o[2]]:= o[3]
        end_if
    end_for;

    rectangle:= newDomain( expr2text(genident("plot::Rectangle2d_")) );

    // attributes:   
    rectangle::corner         := m;
    rectangle::width          := w;
    rectangle::height         := h;
    rectangle::refreshPlotdata:= TRUE;
    rectangle::plotdata       := [];
    rectangle::options        := options;

    return(new(plot::Rectangle2d, rectangle))
end_proc:

plot::Rectangle2d::print := proc(e)
begin
    e:= extop(e,1);
    hold(plot::Rectangle2d)(e::corner,e::width,e::height)
end_proc:

plot::Rectangle2d::getPlotdata := proc(r)
    local e, c, w, h, options, color;
begin
    e:= extop(r,1);
    if e::refreshPlotdata = FALSE then return( e::plotdata ) end_if;

    c:= e::corner;
    w:= e::width;
    h:= e::height;

    options:= e::options;
    if type((color:= options[hold(Color)])) <> "_index" then
        color:= hold(Color) = color[2] // 2nd element of the list
    else
        color:= null()
    end_if;

    // remove specific Rectangle2d options:
    delete options[hold(Filled)], options[hold(Color)];

    r:= point(op(c));
    r:= [
      polygon(r, point(c[1], c[2] + h), point(c[1] + w,c[2] + h), point(c[1] + w, c[2]), r, hold(Filled) = e::options[hold(Filled)], color)
    ];

    e::refreshPlotdata:= FALSE;
    // create the plot data:
    e::plotdata:= [ [hold(Mode = List),r] . [op(options)] ]
end_proc:

plot::Rectangle2d::copy := proc(c)
begin
    return(extsubsop(c, 1=newDomain(genident("plot::Rectangle2d_"), extop(c, 1))))
end_proc:

plot::Rectangle2d::modify := proc(r)
    local o, m, options;
begin
    m := plot::Rectangle2d::copy(r);
    options := [args(2..args(0))];

    ( slot(m, expr2text(op(o, 1)), op(o, 2)) ) $ o in options;
    
    m::refreshPlotdata := TRUE;
    return(m)
end_proc:

plot::Rectangle2d::op := proc(c, n)
begin
    if args(0) < 1 or args(0) > 2 then
        error("invalid number of arguments")
    end_if;

    c:= extop(c,1);
    c:= [c::corner,c::width,c::height];
    if args(0) = 1 then op(c) else op(c, n) end_if
end_proc:

plot::Rectangle2d::nops := () -> 3:

plot::Rectangle2d::_index := proc(c:plot::Rectangle2d,n:Type::PosInt)
begin
    if args(0) <> 2 then
        error("expecting two arguments")
    end_if;

    c:= extop(c,1);
    c:= [c::corner,c::width,c::height];
    if n > nops(c) then
        error("index out of range")
    else
        return( c[n] )
    end_if
end_proc:

// Note: Do not return an error message in the case of an invalid call,
//       because then the object would be destroyed! Hence a warning 
//       message is issued, and the call of this method has no effect 
//       on the object.
plot::Rectangle2d::set_index := proc(c:plot::Rectangle2d, n:Type::PosInt, v)
    local e;
begin
    if args(0) <> 3 then
        warning("wrong number of arguments; assignment ignored");
        return(c)
    end_if; 

    e:= extop(c,1);
    case n
    of 1 do
        if v <> c::Corner then 
            if not testtype(v, Type::ListOf(Type::Arithmetical,2,2)) then
                warning("3rd argument: expecting a list of two arithmetical expressions; assignment ignored");
                return(c)
            end_if;
            e::refreshPlotdata:= TRUE;
            e::corner:= v;
        end_if;
        return( extsubsop(c,1=e) )
    of 2 do
        if v <> c::width then 
            if not testtype(v,Type::Arithmetical) then
                warning("3rd argument: expecting an arithmetical expression; assignment ignored");
                return(c)
            end_if;
            e::refreshPlotdata:= TRUE;
            e::width:= v;
        end_if;
        return( extsubsop(c,1=e) )
    of 3 do
        if v <> e::height then 
            if not testtype(v,Type::Arithmetical) then
                warning("3rd argument: expecting an arithmetical expression; assignment ignored");
                return(c)
            end_if;
            e::refreshPlotdata:= TRUE;
            e::height:= v;
        end_if;
        return( extsubsop(c,1=e) ) 
    otherwise
        warning("index out of range; assignment ignored");
        return(c)
    end_case
end_proc:

// Note: Do not return an error message in the case of an invalid call,
//       because then the object would be destroyed! Hence a warning 
//       message is issued, and the call of this method has no effect 
//       on the object.
plot::Rectangle2d::slot:= proc(o,s,v=null())
    local e, t, _s;
begin
    if args(0) = 3 then
        // slot assignment:
        e:= extop(o,1);

        // first handle all user-accessable attributes:
        case s 
        of "refreshPlotdata" do
            if v <> TRUE and v <> FALSE then
                warning("attribute 'refreshPlotdata': expecting either 'TRUE' or 'FALSE'; assignment ignored");
                return( o )
            else
                e::refreshPlotdata:= v;
                return( extsubsop(o,1=e) )
            end_if
        of "corner"  do
            if not testtype(v, Type::ListOf(Type::Arithmetical,2,2)) then
                warning("3rd argument: expecting a list of two arithmetical expressions; assignment ignored");
                return( o )
            else
                e::corner:= v;
                e::refreshPlotdata:= TRUE;
                return( extsubsop(o,1=e) )
            end_if
        of "width" do
        of "height" do
            if not testtype(v, Type::Arithmetical) then
                warning("attribute '".s."': expecting an arithmetical expression; assignment ignored");
                return( o )  
            else
                e::refreshPlotdata:= TRUE;
                extsubsop(o,1=slot(e,s,v))
            end_if
        of "options" do
            if domtype(v) <> DOM_TABLE then
                warning("attribute 'options': expecting a table of plot options; assignment ignored");
                return( o )
            else
                e::options:= v;
                e::refreshPlotdata:= TRUE;
                return( extsubsop(o,1=e) )
            end_if
        end_case;

        // an option given?
        t:= plot::Rectangle2d::checkOption(text2expr(s) = v);
        if t[1] = FALSE then
            warning( t[2]."; assignment ignored" );
            return(o)
        else
            e::options[t[2]]:= t[3];
            e::refreshPlotdata:= TRUE;
            return( extsubsop(o,1=e) )
        end_if
    else
        // slot access:
        e:= extop(o,1);
        if (t:= slot(e,s)) <> FAIL then
            return(t)
        else
            _s:= text2expr(s);
            if not contains(OPTIONNAMES,_s) then
                error("unknown attribute or option '".s."'")
            elif type((t:= e::options[_s])) <> "_index" then
                return(t)
            else
                return(FAIL)
            end_if
        end_if
    end_if
end_proc:

plot::Rectangle2d::checkOption := proc(opt)
    local value;
begin
    if type(opt) <> "_equal" then
        return( [FALSE,"options must have the form 'OptionName = value'"] )
    end_if;

    value:= op(opt,2); opt:= op(opt,1);
    if not contains( OPTIONNAMES,opt ) then
        return( [FALSE,"unknown option name '".expr2text(opt)."'"] )
    else
        // First, handle special cases for certain options:
        // Color : Beside the standard format 'Color = [{Height,Flat},[r,g,b]]',
        //         allow 'Color = [r,g,b]' which corresponds to '[Flat,[r,g,b]]'.
        if opt = hold(Color) then
            if (value:= plot::getColorFormat(value)) = FAIL then
                return( [FALSE,"option 'Color': invalid color specification"] )
            end_if;
        end_if;

        return( [TRUE,opt,value] )
    end_if
end_proc:

OPTIONNAMES:= {
  // options for graphical objects:
  hold(Color),hold(LineStyle),hold(LineWidth),hold(PointStyle),
  hold(PointWidth),hold(Title),hold(TitlePosition),
  // options for plot::Rectangle2d:
  hold(Filled)
}:

DEFAULTOPTIONS:= table():
// Default options for objects of the domain plot::Rectangle2d:
DEFAULTOPTIONS[hold(Color)] := [hold(Flat),RGB::Red]: // needed if Filled=TRUE
DEFAULTOPTIONS[hold(Filled)]:= FALSE:

