// $Date: 2000/11/14 10:05:15 $  $Author: frankp $   $Revision: 1.30 $
// igel
// completely rewritten by frankp (21.09.2000)

/* -----------------------------------------------------------
    Scene -- the graphical primitive for two- and three-dimensional graphical scenes

    Syntax:

    Scene(o1 <, o2, ...> <, op1, op2, ...>)

    o1, o2, ...  : graphical objects 
    op1, op2, ...: plot2d or plot3d options, see plot2d and plot3d 
                   for details.

    Example:
    >> plot::Scene( plot::Point(1,2), plot::Curve2d( [sin(x), cos(x)], x = 0..2*PI ) )

    >> plot::Scene( 
         plot::Point(1,2), plot::Curve2d( [sin(x), cos(x)], x = 0..2*PI ), 
         Scaling = Constrained 
       )
----------------------------------------------------------------*/ 
alias( DEFAULTOPTIONS2D = plot::Scene::defaultOptions2d ):
alias( OPTIONNAMES2D = plot::Scene::optionNames2d ):
alias( DEFAULTOPTIONS3D = plot::Scene::defaultOptions3d ):
alias( OPTIONNAMES3D = plot::Scene::optionNames3d ):

plot::Scene:= newDomain("plot::Scene"):
plot::Scene::info:= "plot::Scene -- graphical primitive for graphical scenes":
plot::Scene::dimension:= s -> (extop(s,1))::dimension:

plot::Scene::new:= proc()
    local o, i, opt, objects, options, scene, dimen;
begin
    if args(0) = 0 then
        error("no argument given");
    end_if;

    objects:= split( [args()],x -> type(x) = "graphprim" );
    opt:= objects[2];
    if (objects:= objects[1]) = [] then
        error("expecting a graphical primitive, i.e., an object of type \"graphprim\"")
    end_if;

    dimen:= {op(map(objects, e -> e::dom::dimension(e)))};
    if nops(dimen) > 1 then
        error("objects have different dimensions");
    else 
        dimen:= op(dimen,1)
    end_if;
    
    // default options (a table)
    if dimen = 2 then options := DEFAULTOPTIONS2D else options := DEFAULTOPTIONS3D end_if;
    // check user specified options and override the default options:
    for i from 1 to nops(opt) do
        if dimen = 2 then
            if (o:= plot::Scene::checkOption2d(opt[i]))[1] = FALSE then
                error(o[2])
            else
                options[o[2]]:= o[3]
            end_if
        elif (o:= plot::Scene::checkOption3d(opt[i]))[1] = FALSE then
            error(o[2])
        else
            options[o[2]]:= o[3]
        end_if
    end_for;

    scene:= newDomain( expr2text(genident("plot::Scene_")) );

    // attributes:
    scene::plotdata       := [];
    scene::options        := options;
    scene::objects        := objects;
    scene::dimension      := dimen;

    return( new(plot::Scene, scene) )
end_proc:

plot::Scene::copy := proc(s)
    local l, o;
begin
    o:= extop(s,1);
    l:= map( o::objects, x -> x::dom::copy(x) );
    o:= newDomain(genident("plot::Scene_"), o);
    o::objects:= l;
    return( extsubsop(s, 1=o) )
end_proc:

plot::Scene::modify := proc(s)
    local o, m, options;
begin
    m:= plot::Scene::copy(s);
    options:= [args(2..args(0))];

    ( slot(m, expr2text(op(o, 1)), op(o, 2)) ) $ o in options;
    
    return(m)
end_proc:

/*--
    Note: getPlotdata of graphical primitives must
    return the list [[Mode=...,...]<, [Mode=...]]
--*/
plot::Scene::getPlotdata := proc(s)
    local e, plotdata;
begin
    e:= extop(s, 1);
    plotdata:= map( e::objects, o -> op(o::dom::getPlotdata(o)) );
    if plotdata = [] then 
        // the empty plot
        e::plotdata:= []
    else
        e::plotdata:= [ op(e::options),op(plotdata) ]
    end_if
end_proc:

plot::Scene::print := () -> hold(plot::Scene)():

plot::Scene::expose := s -> hold(plot::Scene)(op(map((extop(s,1))::objects,expose))):
        
plot::Scene::op := proc(s, n)
begin
    if args(0) < 1 or args(0) > 2 then
        error("invalid number of arguments")
    end_if;

    s:= (extop(s,1))::objects;
    if args(0) = 1 then op(s) else op(s, n) end_if
end_proc:

plot::Scene::nops := s -> nops(s::objects):

plot::Scene::_index := proc(s:plot::Scene,n:Type::PosInt)
begin
    if args(0) <> 2 then
        error("expecting two arguments")
    end_if;
    s:= (extop(s,1))::objects;
    if n > nops(s) then
        error("index out of range")
    else
        return( s[n] )
    end_if
end_proc:

// Note: Do not return an error message in the case of an invalid call,
//       because then the object would be destroyed! Hence a warning 
//       message is issued, and the call of this method has no effect 
//       on the object.
plot::Scene::set_index := proc(s:plot::Scene, n:Type::PosInt, v)
    local l;
begin
    if args(0) <> 3 then
        warning("wrong number of arguments; assignment ignored");
        return(s)
    elif type(v) <> "graphprim" then
        warning("3rd argument: expecting an object of type \"graphprim\"; assignment ignored");
        return(s)
    end_if; 
 
    l:= (extop(s,1))::objects;
    if n > nops(l) then
        warning("index out of range; assignment ignored");
        return(s)
    else
        s::objects:= subsop( l,n=v );
        return( s )
    end_if
end_proc:

// Note: Do not return an error message in the case of an invalid call,
//       because then the object would be destroyed! Hence a warning 
//       message is issued, and the call of this method has no effect 
//       on the object.
plot::Scene::slot:= proc(o,s,v=null())
    local e, t, _s;
begin
    if args(0) = 3 then
        // slot assignment:
        e:= extop(o,1);

        // first handle all user-accessable attributes:
        case s 
        of "objects" do
            if domtype(v) <> DOM_LIST and map({op(v)},type) <> {"graphprim"} then
                warning("attribute 'objects': expecting a list of objects of type \"graphprim\"; assignment ignored");
                return( o )
            elif map({op(v)}, x -> x::dom::dimension(x)) <> {e::dimension} then
                warning("attribute 'objects': objects have invalid dimensions; assignment ignored");
                return( o )
            else
                e::objects:= v;
                return( extsubsop(o,1=e) )
            end_if
        of "options" do
            if domtype(v) <> DOM_TABLE then
                warning("attribute 'options': expecting a table of plot options; assignment ignored");
                return( o )
            else
                e::options:= v;
                return( extsubsop(o,1=e) )
            end_if
        end_case;

        // an option given?
        if e::dimension = 2 then 
            t:= plot::Scene::checkOption2d(text2expr(s) = v);
        else
            t:= plot::Scene::checkOption3d(text2expr(s) = v);
        end_if;
        if t[1] = FALSE then
            warning( t[2]."; assignment ignored" );
            return(o)
        else
            e::options[t[2]]:= t[3];
            return( extsubsop(o,1=e) )
        end_if
    else
        // slot access:
        e:= extop(o,1);
        if (t:= slot(e,s)) <> FAIL then
            return(t)
        else
            _s:= text2expr(s);
            if e::dimension = 2 then
                if not contains(OPTIONNAMES2D,_s) then
                    error("unknown attribute or option '".s."'")
                end_if
            elif not contains(OPTIONNAMES3D,_s) then
                error("unknown attribute or option '".s."'")
            end_if;

            if type((t:= e::options[_s])) <> "_index" then
                return(t)
            else
                return(FAIL)
            end_if
        end_if
    end_if
end_proc:

plot::Scene::checkOption2d := proc(opt)
    local value;
begin
    if type(opt) <> "_equal" then
        return( [FALSE,"options must have the form 'OptionName = value'"] )
    end_if;

    value:= op(opt,2); opt:= op(opt,1);
    if not contains( OPTIONNAMES2D,opt ) then
        return( [FALSE,"unknown option name '".opt."'"] )
    else
        // First, handle special cases for certain options:
        // Color : Beside the standard format 'Color = [{Height,Flat},[r,g,b]]',
        //         allow 'Color = [r,g,b]' which corresponds to '[Flat,[r,g,b]]'.
        if opt = hold(Color) then
            if (value:= plot::getColorFormat(value)) = FAIL then
                return( [FALSE,"option 'Color': invalid color specification"] )
            end_if;
        end_if;

        return( [TRUE,opt,value] )
    end_if
end_proc:

plot::Scene::checkOption3d := proc(opt)
    local value;
begin
    if type(opt) <> "_equal" then
        return( [FALSE,"options must have the form 'option = value'"] )
    end_if;

    value:= op(opt,2); opt:= op(opt,1);
    if not contains( OPTIONNAMES3D,opt ) then
        return( [FALSE,"unknown option name '".expr2text(opt)."'"] )
    else
        // First, handle special cases for certain options:
        // Color : Beside the standard format 'Color = [{Height,Flat},[r,g,b]]',
        //         allow 'Color = [r,g,b]' which corresponds to '[Flat,[r,g,b]]'.
        if opt = hold(Color) then
            if (value:= plot::getColorFormat(value)) = FAIL then
                return( [FALSE,"option 'Color': invalid color specification"] )
            end_if;
        end_if;

        return( [TRUE,opt,value] )
    end_if
end_proc:

OPTIONNAMES2D:= {
   hold(Arrows), hold(Axes), hold(AxesOrigin), hold(AxesScaling),
   hold(BackGround), hold(Discont), hold(Font), hold(FontFamily),
   hold(FontSize), hold(FontStyle), hold(ForeGround), hold(GridLines),
   hold(GridLinesWidth), hold(GridLinesColor), hold(GridLinesStyle),
   hold(Labeling), hold(Labels), hold(LineStyle), hold(LineWidth),
   hold(RealValuesOnly), hold(PointStyle), hold(PointWidth), hold(PlotDevice),
   hold(Scaling), hold(Ticks), hold(Title), hold(TitlePosition),
   hold(ViewingBox)
}:

DEFAULTOPTIONS2D:= table():
// Default options for objects of the domain plot::Surface3d:
DEFAULTOPTIONS2D[hold(Arrows)]         := FALSE:
DEFAULTOPTIONS2D[hold(Axes)]           := hold(Origin):
DEFAULTOPTIONS2D[hold(AxesOrigin)]     := hold(Automatic):
DEFAULTOPTIONS2D[hold(AxesScaling)]    := [hold(Lin),hold(Lin)]:
DEFAULTOPTIONS2D[hold(BackGround)]     := RGB::White:
DEFAULTOPTIONS2D[hold(Discont)]        := FALSE:
DEFAULTOPTIONS2D[hold(ForeGround)]     := RGB::Black:
DEFAULTOPTIONS2D[hold(GridLines)]      := hold(None):
DEFAULTOPTIONS2D[hold(GridLinesWidth)] := 1:
DEFAULTOPTIONS2D[hold(GridLinesColor)] := RGB::Gray:
DEFAULTOPTIONS2D[hold(GridLinesStyle)] := hold(DashedLines):
DEFAULTOPTIONS2D[hold(Labeling)]       := TRUE:
DEFAULTOPTIONS2D[hold(LineStyle)]      := hold(SolidLines):
DEFAULTOPTIONS2D[hold(LineWidth)]      := 1:
DEFAULTOPTIONS2D[hold(RealValuesOnly)] := FALSE:
DEFAULTOPTIONS2D[hold(PointStyle)]     := hold(FilledSquares):
DEFAULTOPTIONS2D[hold(PointWidth)]     := 30:
DEFAULTOPTIONS2D[hold(Scaling)]        := hold(UnConstrained):
DEFAULTOPTIONS2D[hold(Ticks)]          := hold(Automatic):
DEFAULTOPTIONS2D[hold(TitlePosition)]  := hold(Above):
DEFAULTOPTIONS2D[hold(ViewingBox)]     := hold(Automatic):

OPTIONNAMES3D:= {
   hold(Arrows), hold(Axes), hold(AxesOrigin), hold(AxesScaling),
   hold(BackGround), hold(CameraPoint), hold(FocalPoint),
   hold(Font), hold(FontFamily),
   hold(FontSize), hold(FontStyle), hold(ForeGround), 
   hold(Labeling), hold(Labels), hold(LineStyle), hold(LineWidth),
   hold(PointStyle), hold(PointWidth), hold(PlotDevice),
   hold(Scaling), hold(Ticks), hold(Title), hold(TitlePosition),
   hold(ViewingBox)
}:
DEFAULTOPTIONS3D:= table():
// All options for 3-dimensional graphical scenes (a value NIL means that 
// the corresponding option will not be taken into account when building 
// the plotdata):
DEFAULTOPTIONS3D[hold(Arrows)]         := FALSE:
DEFAULTOPTIONS3D[hold(Axes)]           := hold(Box):
DEFAULTOPTIONS3D[hold(AxesOrigin)]     := hold(Automatic):
DEFAULTOPTIONS3D[hold(AxesScaling)]    := [hold(Lin),hold(Lin),hold(Lin)]:
DEFAULTOPTIONS3D[hold(BackGround)]     := RGB::White:
DEFAULTOPTIONS3D[hold(CameraPoint)]    := hold(Automatic):
DEFAULTOPTIONS3D[hold(FocalPoint)]     := hold(Automatic):
DEFAULTOPTIONS3D[hold(ForeGround)]     := RGB::Black:
DEFAULTOPTIONS3D[hold(Labeling)]       := TRUE:
DEFAULTOPTIONS3D[hold(LineStyle)]      := hold(SolidLines):
DEFAULTOPTIONS3D[hold(LineWidth)]      := 1:
DEFAULTOPTIONS3D[hold(PointStyle)]     := hold(FilledSquares):
DEFAULTOPTIONS3D[hold(PointWidth)]     := 30:
DEFAULTOPTIONS3D[hold(Scaling)]        := hold(UnConstrained):
DEFAULTOPTIONS3D[hold(Ticks)]          := hold(Automatic):
DEFAULTOPTIONS3D[hold(TitlePosition)]  := hold(Above):
DEFAULTOPTIONS3D[hold(ViewingBox)]     := hold(Automatic):

