// $Date: 2000/11/14 10:05:15 $  $Author: frankp $   $Revision: 1.26 $
// igel
// completely rewritten by frankp (21.09.2000)

/* -------------------------------------------------------------
    Surface3d -- The graphical primitive for three-dimensional surfaces

    Syntax:

    Surface3d([x_tu, y_tu, z_tu], t = tr, u = ur <, op1, op2, ...>)

    x_tu, y_tu, z_tu : arithmetical expressions in t an u.
    t, u             : DOM_IDENT
    tr, ur           : either form 'a..b' or list of two arithmetical expressions.
    op1, op2, ...    : plot3d options, see plot2d and plot3d for details.
 
    Example:
    >> plot::Surface3d( [u, sin(t), cos(t)], t = 0..2*PI, u = -1..1 )

    >> plot::Surface3d( [u, sin(t), cos(t)], t = 0..2*PI, u = -1..1, Color = RGB::Blue )

----------------------------------------------------------------*/
alias( DEFAULTOPTIONS = plot::Surface3d::defaultOptions ):
alias( OPTIONNAMES = plot::Surface3d::optionNames ):

plot::Surface3d := newDomain("plot::Surface3d"):
plot::Surface3d::info := "plot::Surface3d -- graphical primitive for surfaces":
plot::Surface3d::type := "graphprim":
plot::Surface3d::dimension := () -> 3:

plot::Surface3d::new:= proc(f, rx, ry)
    local i, opt, surface, xrange, yrange, options;
begin    
    if args(0) < 3 then 
        error("expecting at least three arguments")
    end_if;

    if not testtype(f, Type::ListOf(Type::Arithmetical, 3, 3)) then
        error("1st argument: expecting a list of three arithmetical expressions")
    end_if;

    xrange := plot::conv2range(rx);
    yrange := plot::conv2range(ry);

    if xrange = FAIL then
        error("2nd argument: expecting form 'ident = a..b'");
    end_if;
    if yrange = FAIL then
        error("3rd argument: expecting form 'ident = a..b'");
    end_if:

    options := DEFAULTOPTIONS;  // default options (a table)
    opt := [args(4..args(0))];  // user options (a list)
    // check user options and override the default options:
    for i from 1 to nops(opt) do
        if (rx:= plot::Surface3d::checkOption(opt[i]))[1] = FALSE then
            error(rx[2])
        else
            options[rx[2]]:= rx[3]
        end_if
    end_for;
    
    surface := newDomain( expr2text(genident("plot::Surface3d_")) );
   
    // attributes:
    surface::term           := f;
    surface::range1         := xrange;
    surface::range2         := yrange;
    surface::refreshPlotdata:= TRUE;
    surface::plotdata       := [];
    surface::options        := options;

    return( new(plot::Surface3d, surface) )
end_proc:

plot::Surface3d::copy := proc(f)
begin
    return(extsubsop(f, 1=newDomain(genident("plot::Surface3d_"), extop(f, 1))));
end_proc:

plot::Surface3d::modify := proc(s)
    local o, m, options;
begin
    m := plot::Surface3d::copy(s);
    options := [args(2..args(0))];

    ( slot(m, expr2text(op(o, 1)), op(o, 2)) ) $ o in options;

    m::refreshPlotdata := TRUE;
    return(m)
end_proc:

plot::Surface3d::getPlotdata := proc(f)
    local x, y, e, c;
begin
    e := extop(f, 1);
    if e::refreshPlotdata = FALSE then return( e::plotdata ) end_if;

    c:= e::term;
    x:= plot::range2list(e::range1);
    y:= plot::range2list(e::range2);

    e::refreshPlotdata:= FALSE;
    // create the plot data:
    e::plotdata:= [
        [hold(Mode) = hold(Surface), c, x, y] . [op(e::options)]
    ]
end_proc:

plot::Surface3d::print:= proc(c)
begin
    c:= extop(c,1);
    hold(plot::Surface3d)(c::term,c::range1,c::range2)
end_proc:

// not documented:
plot::Surface3d::move := proc(s, x, y, z)
    local sc, p;
begin
    p:= plot::Point((extop(s,1))::term);
    p:= plot::Point::move(p, x, y, z);

    sc:= plot::Surface3d::copy(s);
    sc::term:= [op(p)];
    sc::refreshPlotdata:= TRUE;
    return(sc)
end_proc:

// not documented:
plot::Surface3d::rotate := proc(s, xphi, yphi, zphi)
    local sc, p;
begin    
    p:= plot::Point((extop(s,1))::term);        
    p:= plot::Point::rotate(p, xphi, yphi, zphi);

    sc:= plot::Surface3d::copy(s);
    sc::term := [op(p)];
    sc::refreshPlotdata:= TRUE;
    return(sc)
end_proc:

plot::Surface3d::op := proc(c, n)
begin
    if args(0) < 1 or args(0) > 2 then
        error("invalid number of arguments")
    end_if;

    c:= [c::term,c::range1,c::range2];
    if args(0) = 1 then op(c) else op(c, n) end_if
end_proc:

plot::Surface3d::nops := c -> 3:

plot::Surface3d::_index := proc(c:plot::Surface3d,n:Type::PosInt)
begin
    if args(0) <> 2 then
        error("expecting two arguments")
    end_if;

    c:= extop(c,1);
    c:= [c::term,c::range1,c::range2];
    if n > nops(c) then
        error("index out of range")
    else
        return( c[n] )
    end_if
end_proc:

// Note: Do not return an error message in the case of an invalid call,
//       because then the object would be destroyed! Hence a warning 
//       message is issued, and the call of this method has no effect 
//       on the object.
plot::Surface3d::set_index := proc(c:plot::Surface3d, n:Type::PosInt, v)
    local e;
begin
    if args(0) <> 3 then
        warning("wrong number of arguments; assignment ignored");
        return(c)
    end_if; 

    e:= extop(c,1);
    case n
    of 1 do
        if v <> e::term then 
            if not testtype(v, Type::ListOf(Type::Arithmetical, 3, 3)) then
                warning("3rd argument: expecting a list of three arithmetical expressions; assignment ignored");
                return(c)
            end_if;
            e::refreshPlotdata:= TRUE;
            e::term:= v
        end_if;
        return( extsubsop(c,1=e) )
    of 2 do
        if v <> e::range1 then 
            if (v:= plot::conv2range(v)) = FAIL then
                warning("3rd argument: expecting form 'ident = a..b'; assignment ignored");
                return(c)
            end_if;
            e::refreshPlotdata:= TRUE;
            e::range1:= v
        end_if;
        return( extsubsop(c,1=e) )
    of 3 do
        if v <> e::range2 then 
            if (v:= plot::conv2range(v)) = FAIL then
                warning("3rd argument: expecting form 'ident = a..b'; assignment ignored");
                return(c)
            end_if;
            e::refreshPlotdata:= TRUE;
            e::range2:= v
        end_if;
        return( extsubsop(c,1=e) )
    otherwise
        warning("index out of range; assignment ignored");
        return(c)
    end_case
end_proc:

// Note: Do not return an error message in the case of an invalid call,
//       because then the object would be destroyed! Hence a warning 
//       message is issued, and the call of this method has no effect 
//       on the object.
plot::Surface3d::slot:= proc(o,s,v=null())
    local e, t, _s;
begin
    if args(0) = 3 then
        // slot assignment:
        e:= extop(o,1);

        // first handle all user-accessable attributes:
        case s 
        of "refreshPlotdata" do
            if v <> TRUE and v <> FALSE then
                warning("attribute 'refreshPlotdata': expecting either 'TRUE' or 'FALSE'; assignment ignored");
                return( o )
            else
                e::refreshPlotdata:= v;
                return( extsubsop(o,1=e) )
            end_if
        of "range1" do
        of "range2" do
            if (v:= plot::conv2range(v)) = FAIL then
                warning("attribute '".s."': expecting form 'ident = a..b'; assignment ignored");
                return( o )
            else
                e::refreshPlotdata:= TRUE;
                return( extsubsop(o,1=slot(e,s,v)) )
            end_if
        of "term" do
            if not testtype(v,Type::ListOf(Type::Arithmetical,3,3)) then
                warning("attribute 'term': expecting a list of three arithmetical expressions; assignment ignored");
                return( o )
            else
                e::term:= v;
                e::refreshPlotdata:= TRUE;
                return( extsubsop(o,1=e) )
            end_if
        of "options" do
            if domtype(v) <> DOM_TABLE then
                warning("attribute 'options': expecting a table of plot options; assignment ignored");
                return( o )
            else
                e::options:= v;
                e::refreshPlotdata:= TRUE;
                return( extsubsop(o,1=e) )
            end_if
        end_case;

        // an option given?
        t:= plot::Surface3d::checkOption(text2expr(s) = v);
        if t[1] = FALSE then
            warning( t[2]."; assignment ignored" );
            return(o)
        else
            e::options[t[2]]:= t[3];
            e::refreshPlotdata:= TRUE;
            return( extsubsop(o,1=e) )
        end_if
    else
        // slot access:
        e:= extop(o,1);
        if (t:= slot(e,s)) <> FAIL then
            return(t)
        else
            _s:= text2expr(s);
            if not contains(OPTIONNAMES,_s) then
                error("unknown attribute or option '".s."'")
            elif type((t:= e::options[_s])) <> "_index" then
                return(t)
            else
                return(FAIL)
            end_if
        end_if
    end_if
end_proc:

plot::Surface3d::checkOption := proc(opt)
    local value;
begin
    if type(opt) <> "_equal" then
        return( [FALSE,"options must have the form 'OptionName = value'"] )
    end_if;

    value:= op(opt,2); opt:= op(opt,1);
    if not contains( OPTIONNAMES,opt ) then
        return( [FALSE,"unknown option name '".expr2text(opt)."'"] )
    else
        // First, handle special cases for certain options:
        // Color : Beside the standard format 'Color = [{Height,Flat},[r,g,b]]',
        //         allow 'Color = [r,g,b]' which corresponds to '[Flat,[r,g,b]]'.
        if opt = hold(Color) then
            if (value:= plot::getColorFormat(value)) = FAIL then
                return( [FALSE,"option 'Color': invalid color specification"] )
            end_if;
        end_if;

        return( [TRUE,opt,value] )
    end_if
end_proc:

OPTIONNAMES:= {
  // options for 3-dimensional objects:
  hold(Color),hold(LineStyle),hold(LineWidth),hold(PointStyle),
  hold(PointWidth),hold(Title),hold(TitlePosition),
  // options for surfaces:
  hold(Grid),hold(Smoothness),hold(Style)
}:

DEFAULTOPTIONS:= table():
// Default options for objects of the domain plot::Surface3d:
DEFAULTOPTIONS[hold(Grid)]          := [20,20]:


