// $Date: 2000/11/16 08:17:47 $  $Author: frankp $   $Revision: 1.3 $

/*++
    vector -- graphical object for vectors

    Syntax:

    vector(p1, p2, r, l<, options>)

    p1, p2: either graphical objects of domain type plot::Point or
            lists of two or three arithmetical expressions 
    r, l  : arithmetical expressions
    
    Return Value:
    A graphical object of the domain type plot::Polygon

    Details:
    * plot::vector returns a polygon consisting of three lines.
      Theses lines are connecting the points p1 with p2 and
      the endpoints of the arrow with the point p2.
    * r specifies the angle between the line and the arrow.
    * l specifies the length of the arrow.

    Examples:
    >> 
++*/
plot::vector := proc(p1, p2, r, l)
    local length_x, length_y, length_l, alpha, x, y, end1, end2;
begin
    r:= float(r):
    if args(0) < 4 then
        error("expecting at least four arguments")
    elif (p1:= plot::Point::convert(p1)) = FAIL then
        error("1st argument: two- or three-dimensional point expected")
    elif (p2:= plot::Point::convert(p2)) = FAIL then
        error("2nd argument: two- or three-dimensional point expected")
    elif testargs() then
        if p1::dom::dimension(p1) <> p2::dom::dimension(p2) then
            error("dimensions of the points mismatch")
        elif not testtype(r,Type::Real) then
            error("3rd argument: real value expected")
        elif not testtype(l,Type::Real) then
            error("4th argument: real value expected")
        end_if;

        if not testtype([op(p1)],Type::ListOf(Type::Real)) or
           not testtype([op(p2)],Type::ListOf(Type::Real))
        then
            error("coordinates of the points must be real values")
        end_if
    end_if;

    // Calculate the length of the line connecting p1 and p2:
    length_x := op(p2, 1) - op(p1, 1):
    length_y := op(p2, 2) - op(p1, 2):
    length_l := sqrt(float(length_x^2 + length_y^2)):

    // Calculate the angle between the line (point1, point2) and the
    // horizontal axes.
    if abs(length_l) > 10^(-DIGITS) then
        if abs(length_y) > 10^(-DIGITS) then
            alpha := sign(length_y)*arccos(length_x/length_l):
        else
            alpha := arccos(length_x/length_l):
        end_if:
    else
        alpha := 0:
    end_if;

    // Calculate the endpoints of the arrow:
    length_x := l*length_l;
    length_y := l*length_l;
    x    := op(p2, 1) - length_x*cos(alpha+r):
    y    := op(p2, 2) - length_y*sin(alpha+r):
    end1 := plot::Point(x, y):
    x    := op(p2, 1) - length_x*cos(alpha-r):
    y    := op(p2, 2) - length_y*sin(alpha-r):
    end2 := plot::Point(x, y):

    plot::Polygon( p1, p2, end1, p2, end2, args(5..args(0)) )
end_proc:

