// kg, 08/07/94
// Turtle -- 2d turtle graphic primitives

/*
Operands of a turtle:
1 - a table of procedures which refer to a procedure environment
    holding the turtles state.
    These procedures may be used to change the turtles state.

Example: (a triangle)
>> t:= plot::Turtle():
>> t::line(1):
>> t::right(120):
>> t::line(1):
>> t::right(120):
>> t::line(1):
>> plot(t)
*/

domain plot::Turtle

type:= "graphprim";
dimension:= () -> 2;
info:= "plot::Turtle -- primitive for two-dimensional turtle graphics";

// The state of a turtle is stored in a procedure environment.
// The turtle object holds a table with procedures which may be
// used to change the state of the turtle.
// A new turtle starts at (0,0) heading north, green color, 
// empty stack.
new:= proc() local t; begin
    if args(0) <> 0 then error("no args allowed") end_if;
    t:= new(dom,
        // this procedure creates the environment holding the
        // tutles state
        proc() local x, y, dir, path, color, stateStack, this; 
        option escape;
        begin
            x:= 0.0; y:= 0.0; dir:= 0.5*float(PI);
            path:= []; color:= [0.0,1.0,0.0]; stateStack:=[];

            // The following procedures are used to change the
            // turtles state. They mey be regarded as methods
            // of the turtle object (in OOP speak).
            table(
            // __init__ -- set 'this' var
            "__init__" = proc(t:dom) begin
                this:= t;
            end_proc,

            // print -- print turtle
            "print" = proc() begin
                "(".expr2text(x).",".expr2text(y).") -> ".
                    expr2text(dir*180.0/float(PI))
            end_proc,

            // color -- set actual turtle color
            "color" = proc(c: Type::ListOf(Type::Real, 3, 3)) begin
		c:= map(c, float);
                if testargs() then
        	    if max(op(c)) > 1.0 then
        	        error("wrong color value")
        	    end_if;
        	    if min(op(c)) < 0.0 then
        	        error("wrong color value")
        	    end_if;
                end_if;
                color:= c;
                this;
            end_proc,

            // left -- turn left about d without changing position
            "left" = proc(d:Type::Real) begin
                dir:= dir + float(d*PI/180.0);
                this;
            end_proc,
        
            // right -- turn right about d without changing position
            "right" = proc(d:Type::Real) begin
                dir:= dir - float(d*PI/180.0);
                this;
            end_proc,
        
            // move -- move about length l without drawing
            "move" = proc(l:Type::Real) begin
                l:= float(l);
                x:= x + l*cos(dir);
        	y:= y + l*sin(dir);
                this;
            end_proc,

            // line -- draw line with length l and actual color #
            "line" = proc(l:Type::Real) local p; begin
                l:= float(l);
                p:= point(x,y);
                x:= x + l*cos(dir);
                y:= y + l*sin(dir);
                path:= append(path, polygon(p, point(x,y), Color=color));
                this;
            end_proc,

            // push -- push actual state onto stack
            "push" = proc() begin
                stateStack:= [[x,y,dir,color], stateStack];
                this;
            end_proc,
        
            // pop -- pop new turtle state from stack
            "pop" = proc() local state; begin
                if nops(stateStack) > 0 then
                    state:= stateStack[1];
                    x:= state[1]; y:= state[2]; dir:= state[3]; color:= state[4];
                    stateStack:= stateStack[2];
                end_if;
                this;
            end_proc,

            // getPlotdata -- return turtle path as polyline
            "getPlotdata" = proc() begin
                [[ Mode=List, path ]]
            end_proc
            )
        end_proc());
    extop(t,1)["__init__"](t);
    t;
end_proc;

// testtype -- test if expression is turtle
testtype:= proc(e,T) begin
    if T = dom then
	if domtype(e) = dom then
	    return(TRUE)
	end_if
    end_if;
    FAIL
end_proc;

// print -- print turtle
print:= proc(t:dom) begin
    extop(t,1)["print"]()
end_proc;

// getPlotdata -- return turtle path as polyline
getPlotdata:= proc(t:dom) begin
    extop(t,1)["getPlotdata"]()
end_proc;

// slot -- access turtle methods
slot:= proc(t:dom, n:DOM_STRING) begin
    if args(0) = 2 then
        if contains(extop(t,1), n) then
            extop(t,1)[n]
        else error("unknown slot")
        end_if
    elif args(0) = 3 then error("can not set turtle slots")
    else error("illegal no of args")
    end_if
end_proc;

end_domain:

// end of file
