// $Date: 2000/10/30 18:01:44 $  $Author: frankp $   $Revision: 1.5 $

/* Comments:

    1) Neue Diagrammtypen:
       Points      : Nur Datenpunkte werden gezeichnet 
                     Point ist fuer 2D und 3D implementiert.
       LinesPoints : Zeichnen der Verbindungslinien und der 
                     Datenpunkte
       CurvesPoints: Zeichnen der Verbindungskurven mittels
                     kubischer Splines und der Datenpunkte

    2) Die Achsenbeschriftung wurde vollstaendig ueberarbeitet. 

    3) Bzgl. Points, Lines, Curves, LinesPoints und CurvesPoints
       koennen nun nicht-equidistante Datenpunkte gezeichnet
       werden. Ausserdem kann die Anzahl der Datenpunkte in
       den Datenlisten unterschiedlich sein.
       Auch kann anstelle einer Liste ein array angegeben werden.
       Wird ein Objekt vom Typ Dom::Matrix angegeben, so wird
       dieses immer als 2D-Array interpretiert, unabh. davon,
       ob es sich u.U. nur um einen Spalten- oder Zeilen-Vektor
       handelt, d.h. nur der Diagrammtyp Points steht derzeit fuer
       Dom::Matrix zur Verfuegung.

    ToDo:
    * Es koennen mittels SceneOptions = [ ... ] beliebige
      Szene-Optionen an plot2d, plot3d weitergereicht werden.
      Diese koennen sich intern mit den fest eingestellten
      Szene-Optionen wie z.B. Axes = None etc. beissen und
      zum Fehler in plot2d, plot3d fuehren.
      Die fest eingestellten Szene-Optionen muessen als
      Defaults realisiert werden, die dann gegebenenfalls
      von user-Szene-Optionen ueberschrieben werden.

      Done. (frankp)

    * Implementierungsbedarf fuer "Surface": wir brauchen
      noch eine 2d-Spline-Approximation von Daten, um das
      unselige (und langsame) numeric::lagrange ersetzen
      zu koennen. In "Curves" ist dies bereits geschehen:
      altes numeric::lagrange --> numeric::cubicSpline.
      Zudem stimmt die Farbverwaltung nicht!
*/

/*++
    plot::data  --  Plots data in different formats

    Call:
    data(Mode, list<, options>)

    Parameters:
    mode   : either 'Pichart2d', 'Pichart3d', 'Lines', 'LinesPoints', 'Curves', 
             'CurvesPoints', 'Columns', 'Beams', 'Surface' or 'Plain'
    list   : list of real values
    options: scene options, including the special options 
             'Colors = ...' and 'Titles = ...'

    Return Value:
    An object of domain type plot::Scene

    Examples:
    >> export(plot): export(RGB):

    >> plot(data( Piechart2d, [5,12,38,14,25] ))

    >> plot(data( Piechart3d, [5,12,38], Colors=[RoyalBlue,VioletRed,GreenPale] ))

    >> plot(data(Lines, 
        [[5,10,24,-3],[6,5,2,18],[19,45,12,-10]], Colors=[Red,Green,Blue]
       ))

    >> plot(data(LinesPoints, 
        [[[1,5],[2,10],[4,24],[7,-3]], [[0,6],[1,5],[2,2],[3,18]],
         [[-2,19],[1,45],[2,12],[7,-10]]], 
        Colors=[Red,Green,Blue]
       ))

    >> plot(data(Curves, 
        [[5,10,24,-3],[6,5,2,18],[19,45,12,-10]], Colors=[Red,Green,Blue]
       ))

    >> plot(data(CurvesPoints, 
        [[5,10,24,-3],[6,5,2,18],[19,45,12,-10]], Colors=[Red,Green,Blue]
       ))

    >> plot(data(Columns, 
        [[5,10,24,-3],[6,5,2,18],[19,45,12,-10]], Colors=[Red,Green,Blue]
       ))

    >> plot(data(Columns, [[5,10,24,-3,6,5,2,18]]))

    >> plot(data(Beams, 
        [[5,10,24,-3],[6,5,2,18],[19,45,12,-10]], Colors=[Red,Green,Blue]
       ))

    >> plot(data(Plain, [[5,10,24,-3],[6,5,2,18],[19,45,12,-10]], 
        Colors=[Red,Red,Green])
       )

    >> plot(data(Surface, [[5,10,24,-3],[6,5,2,18]]))
++*/
plot::data:= proc(dtyp, l)
    local s, c, i, t, p, tspec, cspec, options;
begin
    if args(0) < 1 then
        error("expecting diagram type and a list of data")
    else
        l:= plot::data_checkformat(l, dtyp);

        // get options:
        c:= RGB::ColorList; // default color list
        t:= []; // Titles
        options:= [args(3..args(0))];
        s:= nops(options);
        tspec:= cspec:= FALSE;
        for i from 1 to nops(options) do
            if type( (p:= options[i]) ) <> "_equal" then
                error("specification of the form 'option = value' expected")
            end_if;
            case op(p,1)
            of hold(Colors) do 
                c:= op(p,2);
                if domtype(c) <> DOM_LIST or nops(c) < nops(l) then
                    error("option 'Colors': list with ".nops(l)." RGB color values expected")
                end_if;
                options[i]:= FAIL;
                cspec:= TRUE;
                break
            of hold(Titles) do 
                t:= op(p,2);
                if not testtype(t,Type::ListOf(DOM_STRING,nops(l) $ 2)) then
                    error("option 'Titles': list with ".nops(l)." strings")
                end_if;
                options[i]:= FAIL;
                tspec:= TRUE;
            end_case;

            if cspec = TRUE and tspec = TRUE then break end_if;
        end_for;
        options:= subs( options,FAIL=null() )
    end_if;

    if testargs() then
        case dtyp 
        of hold(Piechart2d) do
        of hold(Piechart3d) do
            if not testtype(l, Type::ListOf(Type::Real,1)) or contains(map(l,specfunc::sign),-1) <> 0 then
                error("1st argument: list of positive real numbers expected")
            end_if;
            break
        of hold(Lines)        do
        of hold(Points)       do
        of hold(LinesPoints)  do
        of hold(Curves)       do
        of hold(CurvesPoints) do
            if testtype(l, Type::ListOf(Type::ListOf(Type::Real,1),1)) then
               if _or(nops(l[i])<>nops(l[i+1])$i=1..nops(l)-1) then
                   error("lists must have the same number of elements")
               end_if
            elif testtype(l, Type::ListOf(Type::ListOf(DOM_LIST),1)) then
                if _or((nops(l[i][j])<>nops(l[i][j+1])$j=1..nops(l[i])-1)$i=1..nops(l)) then
                    error("lists must have the same number of elements")
                end_if
            else 
                error("2nd argument: a list of lists expected")
            end_if;
            break
        of hold(Columns) do
        of hold(Beams)   do
        of hold(Plain)   do
        of hold(Surface) do
            if not testtype(l, Type::ListOf(Type::ListOf(Type::Real,1),1)) then
                error("2nd argument: list of lists expected")
            elif _or(nops(l[i])<>nops(l[i+1])$i=1..nops(l)-1) then
                error("lists must have the same number of elements")
            end_if
        end_case
    end_if;

    case dtyp
    of hold(Piechart2d) do
        return( plot::data_Piechart2d(l, c, t, options) )
    of hold(Piechart3d) do
        return( plot::data_Piechart3d(l, c, t, options) )
    of hold(Points) do
        return( plot::data_Lines(l, c, t, 1, options) )
    of hold(Lines) do
        return( plot::data_Lines(l, c, t, 2, options) )
    of hold(LinesPoints) do
        return( plot::data_Lines(l, c, t, 3, options) )
    of hold(Columns) do
        return( plot::data_Columns(l, c, t, options) )
    of hold(Beams) do
        return( plot::data_Beams(l, c, t, options) )
    of hold(Plain) do
        return( plot::data_Plain(l, c, t, options) )
    of hold(Surface) do
        return( plot::data_Surface(l, c, t, options) )
    of hold(Curves) do
        return( plot::data_Curves(l, c, t, 1, options) )
    of hold(CurvesPoints) do
        return( plot::data_Curves(l, c, t, 2, options) )
    otherwise
        error("unknown diagram type")
    end_case
end_proc: 

/*--
    Piechart2d(list, lrgb, lstring, options)

    list   : list of positive real values
    lrgb   : list of colors (RGB values)
    lstring: list of strings (the legende)
    options: 3d-plot (!) options for a graphical scene

    Return Value:
    An object of the domain type plot::Scene.
--*/
plot::data_Piechart2d:= proc(l,c,t,options=[])
    name plot::data::Piechart2d;
    local a, b, db, fsin, fcos, grid, i, obj, p, rgb1,
          s, u, v, segment_angle, scene_default;
begin
    rgb1:= c -> [max(0,c[1]-0.2), max(0,c[2]-0.2), max(0,c[3]-0.2)];
    [u,v] := [genident("u"),genident("v")];
     
    fsin := slot(sin, "float"): 
    fcos := slot(cos, "float"): 
 
    s := _plus(op(l)); obj := null(); 
    p := float(PI); b := 0; 
    t := plot::data_set_titles(t, l, nops(l));
    // divide each slice into segements of 6 degrees each.
    // This angle is small enough to generate a smooth slice
    segment_angle:= 0.1047197551; // = float(6/180*PI) = 6 degrees

    for i from 1 to nops(l) do
        db:= l[i]/s*2*p;
        // use at least 2 segments per slice, i.e., grid >=3
        grid:= max(3, round(db/segment_angle)); 
        a := b; b := b+db;
        obj := obj, plot::Surface3d(
            [v*fcos(u),v*fsin(u), 0],u=[a,b],v=[0,1],
            hold(Style) = [hold(ColorPatches),hold(Only)], hold(Color) = [hold(Flat), c[i]],
            hold(Grid) = [grid,2], hold(Title) = t[i]
        )
    end_for;

    // set options:
    scene_default:= 
        [hold(Axes) = hold(None), hold(Scaling) = hold(Constrained),
         hold(CameraPoint) = [0,-0.01,1], hold(FocalPoint) = [0,0,0],
         hold(Title) = ""];
    if options <> [] then
        if has(options,hold(Axes)) then scene_default[1]:= FAIL end_if;
        if has(options,hold(Scaling)) then scene_default[2]:= FAIL end_if;
        if has(options,hold(CameraPoint)) then scene_default[3]:= FAIL end_if;
        if has(options,hold(FocalPoint)) then scene_default[4]:= FAIL end_if;
        if has(options,hold(Title)) then scene_default[5]:= FAIL end_if;
        options:= subs( scene_default, FAIL=null()) . options;
    else
        options:= scene_default
    end_if;

    return( plot::Scene(obj, op(options)) )
end_proc:

/*--
    Piechart3d(list, lrgb, lstring, options)

    list   : list of positive real values
    lrgb   : list of colors (RGB values)
    lstring: list of strings (the legende)
    options: 3d-plot options for a graphical scene

    Return Value:
    An object of the domain type plot::Scene.
--*/
plot::data_Piechart3d:= proc(l, c, t, options = [])
    name plot::data::Piechart3d;
    local a, b, db, fsin, fcos, grid, i, obj, p, rgb1,
          s, u, v, segment_angle, scene_default;
begin
    rgb1 := proc(c)
    begin
        c[1] := max(0,c[1]-0.2);
        c[2] := max(0,c[2]-0.2);
        c[3] := max(0,c[3]-0.2);
        c
    end_proc:

    [u,v] := [genident(),genident()];
     
    fsin:= slot(sin, "float"): 
    fcos:= slot(cos, "float"): 
 
    s:= _plus(op(l)); obj := null(); 
    p:= float(PI); b := 0; 
    t:= plot::data_set_titles(t, l, nops(l)*2);
    // divide each slice into segements of 6 degrees each.
    // This angle is small enough to generate a smooth slice
    segment_angle:= 0.1047197551; // = float(6/180*PI) = 6 degrees

    for i from 1 to nops(l) do
        db:= l[i]/s*2*p;
        // use at least 2 segments per slice, i.e., grid >=3
        grid:= max(3, round(db/segment_angle)); 
        a:= b; b:= b + db;
        obj:= obj, plot::Surface3d(
            [v*fcos(u),v*fsin(u), 0],u=[a,b],v=[0,1],
            hold(Style) = [hold(ColorPatches),hold(Only)], hold(Color) = [hold(Flat), c[i]],
            hold(Grid) = [grid,2], hold(Title)=t[2*(i-1)+1]
        );
        obj:= obj, plot::Surface3d([fcos(u),fsin(u), v], u=[a,b], v=[-0.2,0],
            hold(Style) = [hold(ColorPatches),hold(Only)], hold(Color) = [hold(Flat), rgb1(c[i])],
            hold(Grid)=[grid,2], hold(Title)=t[2*i]
        )
    end_for;

    // set options:
    scene_default:= [hold(Axes) = hold(None), hold(Title) = ""];
    if options <> [] then
        if has(options,hold(Axes)) then scene_default[1]:= FAIL end_if;
        if has(options,hold(Title)) then scene_default[2]:= FAIL end_if;
        options:= subs( scene_default, FAIL=null()) . options;
    else
        options:= scene_default
    end_if;

    return( plot::Scene(obj, op(options)) )
end_proc:

/*--
    dataplot_Lines
--*/
plot::data_Lines:= proc(l, c, t, typ, options=[])
    name plot::data::Lines;
    local i, j, m, objs, scene_default;
begin
    // typ: 1 -> Points, 2 -> Lines, 3 -> LinesPoints 
    m:= nops(l); 
    if domtype(l[1][1]) = DOM_LIST then
        case nops(l[1][1])
        of 2 do
            break;
        of 3 do
            if typ <> 1 then
                error("only diagram type 'Points' available for three-dimensional lists")
            else
                // set options:
                scene_default:= 
                    [hold(Axes) = hold(Box), hold(Scaling) = hold(UnConstrained),
                     hold(Labeling) = TRUE, hold(Ticks) = plot::data_ticks(l),
                     hold(Title) = ""];
                if options <> [] then
                    if has(options,hold(Axes)) then scene_default[1]:= FAIL end_if;
                    if has(options,hold(Scaling)) then scene_default[2]:= FAIL end_if;
                    if has(options,hold(Labeling)) then scene_default[3]:= FAIL end_if;
                    if has(options,hold(Ticks)) then scene_default[4]:= FAIL end_if;
                    if has(options,hold(Title)) then scene_default[5]:= FAIL end_if;
                    options:= subs( scene_default, FAIL=null()) . options;
                else
                    options:= scene_default
                end_if;

                t:= plot::data_set_titles(t, l, 2*m);
                return(plot::Scene(
                    plot::data_point3d_list(l, t, c), op(options)
                ));
            end_if;
        otherwise
            error("lists of two or three numbers expected")
        end_case
    end_if;

    objs:= null(); 
    t:= plot::data_set_titles(t, l, 2*m);

    // typ: 1 -> Points, 2 -> Lines, 3 -> LinesPoints 

    if type(l[1][1]) <> DOM_LIST then
        // [y1,..., yN] -> [[0,y1],...[N-1, yN]]
        l := [ [ [i, l[j][i+1]] $ i=0..nops(l[j])-1 ] $ j=1..m ]
    end_if;

    // l = [ [ [x_1_1, y_1_1], ..., [x_1_m1, y_1_n1] ], ... , ]

    if typ = 1 or typ = 3 then
        objs := plot::data_point2d_list(l, t, c) 
    end_if;

    if typ = 2 or typ = 3 then
        objs := objs,
          plot::Polygon(
            (plot::Point(l[j][i][1],l[j][i][2]),
             plot::Point(l[j][i+1][1],l[j][i+1][2])
            ) $ i=1..nops(l[j])-1, hold(Color) = c[j],
            hold(Title) = t[m+j], hold(TitlePosition) = [0,0]) $ j=1..m
    end_if;

    // set options:
    scene_default:= 
        // Ticks=min(max(2,n-2),10)
        [hold(Axes) = hold(Corner), hold(Scaling) = hold(UnConstrained),
         hold(Labeling) = TRUE, hold(Ticks) = plot::data_ticks(l),
         hold(Labels) = ["",""], hold(Title) = ""];
    if options <> [] then
        if has(options,hold(Axes)) then scene_default[1]:= FAIL end_if;
        if has(options,hold(Scaling)) then scene_default[2]:= FAIL end_if;
        if has(options,hold(Labeling)) then scene_default[3]:= FAIL end_if;
        if has(options,hold(Ticks)) then scene_default[4]:= FAIL end_if;
        if has(options,hold(Labels)) then scene_default[5]:= FAIL end_if;
        if has(options,hold(Title)) then scene_default[6]:= FAIL end_if;
        options:= subs(scene_default, FAIL=null()) . options;
    else
        options:= scene_default
    end_if;

    return( plot::Scene(objs, op(options)) )
end_proc:

/*--
    dataplot_Curves

    to do: we presently assume equidistant data points, all
           sublists presently must have the same number of points

    Should be generalized to:
    *) non-equidistant data points
    *) each sub-list of data may have different nodes,
       the number of nodes may differ between sub-lists

    Works now if lists are specified as follows:
    [ [[x_1_1, y_1_1],...,[x_1_N1, y_1_N1]], 
      ... , 
    [[x_M_1, y_M_1],...,[x_M_NM, y_M_NM]] ]   
--*/
plot::data_Curves:= proc(l, c, t, typ, options=[])
    name plot::data::Curves;
    local i, j, m, n, objs, x, xmax, xmin, ymax, ymin, SplineFunc, 
          scene_default;
begin
    m:= nops(l); x:= genident();

    if domtype(l[1][1]) <> DOM_LIST then
        // [y1,..., yN] -> [[0,y1],...[N-1, yN]]
        l := [ [ [i, l[j][i+1]] $ i=0..nops(l[j])-1 ] $ j=1..m ]
    end_if;

    t := plot::data_set_titles(t, l, 2*m);
    (SplineFunc[j]:= numeric::cubicSpline(op(l[j])))      $ j=1..m;
    (xmin[j] := float(min(l[j][i][1] $ i=1..nops(l[j])))) $ j=1..m;
    (xmax[j] := float(max(l[j][i][1] $ i=1..nops(l[j])))) $ j=1..m;

     
    objs := plot::Curve2d([x, hold(fp::apply)(SplineFunc[j],x)], 
        x=[xmin[j],xmax[j]], hold(Color)=[hold(Flat),c[j]],
        hold(Title)=t[m+j]
    ) $ j=1..m;
/*     
    objs := plot::Curve2d([x, hold(fp::apply)(SplineFunc[j],x)], 
        x=[xmin[j],xmax[j]], hold(Color)=[hold(Flat),c[j]],
        hold(Title)=t[m+j]
    ) $ j=1..m;
*/
    if typ = 2 then
        // the data points
        objs := objs , plot::data_point2d_list(l, t, c)       
    end_if;

    // Folgende 2 Zeilen nur interessant, falls kleinster
    // und groesster y-Wert ermittelt werden soll.
    n := plot::data_curve_min_max(SplineFunc, xmin, xmax, 100);
    ymin := n[1]; ymax := n[2];

    // set options:
    scene_default:= 
        [hold(Axes) = hold(Corner), hold(Scaling) = hold(UnConstrained), hold(Labeling) = TRUE,
         hold(Labels) = ["",""], hold(Ticks) = plot::data_ticks(l,ymin, ymax),
         hold(Title) = ""];
    if options <> [] then
        if has(options,hold(Axes)) then scene_default[1]:= FAIL end_if;
        if has(options,hold(Scaling)) then scene_default[2]:= FAIL end_if;
        if has(options,hold(Labeling)) then scene_default[3]:= FAIL end_if;
        if has(options,hold(Labels)) then scene_default[4]:= FAIL end_if;
        if has(options,hold(Ticks)) then scene_default[5]:= FAIL end_if;
        if has(options,hold(Title)) then scene_default[6]:= FAIL end_if;
        options:= subs(scene_default, FAIL=null()) . options;
    else
        options:= scene_default
    end_if;

    return( plot::Scene(objs, op(options)) )
end_proc:

/* --
    dataplot_Surface

    Dies ist noch zu ueberarbeiten: die Poly-Interpolation per 
    numeric::lagrange sollte durch eine Spline-Interpolation 
    ersetzt werden, eine entsprechende Lib-Funktion existiert
    aber noch nicht. W. Oevel
    
    to do: we presently assume equidistant data points, all
           sublists presently must have the same number of points

    Should be generalized to:
    *) non-equidistant data points
    *) each sub-list of data may have different nodes,
       the number of nodes may differ between sub-lists
--*/
plot::data_Surface:= proc(l, c, t, options=[])
    name plot::data::Surface;
    local j, m, n, x, y, interpoly, scene_default;
begin    
    m := nops(l); n := nops(l[1]);
    t := plot::data_set_titles(t, l, 1);
    [x, y] := [genident(), genident()];
    // the symbolic interpolation polynomial
    interpoly:= numeric::lagrange( [[j$j=0..m-1],[j$j=0..n-1]], array(1..m,1..n,l),[x,y]):
    interpoly:= expr(interpoly);

    // set options:
    scene_default:= 
        [hold(Axes) = hold(Box), hold(Scaling) = hold(UnConstrained), 
         hold(Labeling) = FALSE, hold(Title) = ""];
    if options <> [] then
        if has(options,hold(Axes)) then scene_default[1]:= FAIL end_if;
        if has(options,hold(Scaling)) then scene_default[2]:= FAIL end_if;
        if has(options,hold(Labeling)) then scene_default[3]:= FAIL end_if;
        if has(options,hold(Title)) then scene_default[4]:= FAIL end_if;
        options:= subs( scene_default, FAIL=null()) . options;
    else
        options:= scene_default
    end_if;

    return(plot::Scene(
        // data points
        (plot::Point(i, j, l[i+1][j+1], 
            hold(PointWidth) = 40, hold(PointStyle) = hold(FilledCircles)
            //, hold(Title) = t[i*m+j]
        ) $ i=0..m-1) $ j=0..n-1,
        // interpolating surface  
        plot::Surface3d([x, y, interpoly], x=[0,m-1], y=[0,n-1], 
            hold(Grid) = [m, n], hold(Style) = [hold(HiddenLine),hold(Mesh)], hold(Title) = "",
            hold(Color) = c[1]
        ),
        // scene options
        op(options)
    ))
end_proc:

/*--
    dataplot_Columns

    Rechtecke muessen durch zwei Polygone, die jeweils ein Dreieck bilden,
    dargestellt werden, da derzeit nur Dreiecke mit einer Farbe gefuellt
    werden koennen. Ein weiteres Polygon bildet den Rahmen des Rechtecks.
    Ein Rechteck besteht daher aus drei Polygonen.

    Das gesamte Saeulen- oder Balkendiagramm koennte durch ein Objekt dar-
    gestellt werden. Man koennte dann aber nur zwei Beschriftungen eingeben,
    eine fuer die Szene und eine fuer das Objekt. Aus diesem Grund wird
    jedes Polygon als ein eigenes Objekt dargestellt. Demnach koennen
    3*m*n+1 Beschriftungen eingefuegt werden.
--*/
plot::data_Columns:= proc(l, c, t, options=[])
    name plot::data::Columns;
    local i, j, m, n, nc, offset, d, scene_default;
begin
    m := nops(l); n := nops(l[1]); nc := nops(c);
    offset := (if m = 1 then 0.2 else 1 end_if);
    t := plot::data_set_titles(t, l, 3*m*n);
    if m = 1 then
        d := (max(op(l[1]))-min(op(l[1])))/50.0
    end_if;

    // set options:
    scene_default:= 
        [hold(Axes) = hold(Origin), hold(Scaling) = hold(UnConstrained), hold(Labeling) = TRUE,
         hold(Labels) = ["",""], hold(Ticks) = plot::data_ticks(l), hold(Title) = ""];
    if options <> [] then
        if has(options,hold(Axes)) then scene_default[1]:= FAIL end_if;
        if has(options,hold(Scaling)) then scene_default[2]:= FAIL end_if;
        if has(options,hold(Labeling)) then scene_default[3]:= FAIL end_if;
        if has(options,hold(Labels)) then scene_default[4]:= FAIL end_if;
        if has(options,hold(Ticks)) then scene_default[5]:= FAIL end_if;
        if has(options,hold(Title)) then scene_default[6]:= FAIL end_if;
        options:= subs( scene_default, FAIL=null()) . options;
    else
        options:= scene_default
    end_if;

    return(plot::Scene(
        (plot::data_rectangle(
            j+i*(m+offset),0,j+1+i*(m+offset),l[j+1][i+1],
            (if m=1 and nc=n then c[i+1] else c[j+1] end_if),t[i*m+j+1],
            (if m=1 then /*0.1,0.3,TRUE*/ 0.1,d,TRUE else null() end_if)
        ) $ j=0..m-1) $ i=0..n-1,
        op(options)
    ))
end_proc:

/*--
    dataplot_Beams
--*/
plot::data_Beams:= proc(l, c, t, options=[])
    name plot::data::Beams;
    local i, j, m, n, nc, offset, scene_default;
begin
    m := nops(l); n := nops(l[1]); nc := nops(c);
    offset := (if m = 1 then 0.2 else 1 end_if);
    t := plot::data_set_titles(t, l, 3*m*n);

    // set options:
    scene_default:= 
        [hold(Axes) = hold(None), hold(Scaling) = hold(UnConstrained), hold(Title) = ""];
    if options <> [] then
        if has(options,hold(Axes)) then scene_default[1]:= FAIL end_if;
        if has(options,hold(Scaling)) then scene_default[2]:= FAIL end_if;
        if has(options,hold(Title)) then scene_default[3]:= FAIL end_if;
        options:= subs( scene_default, FAIL=null()) . options;
    else
        options:= scene_default
    end_if;

    return(plot::Scene((plot::data_rectangle(
               0,j+i*(m+offset),l[j+1][i+1],j+1+i*(m+offset),
               (if m=1 and nc=n then c[i+1] else c[j+1] end_if),t[i*m+j+1],
               (if m=1 then 0.3,0.1,FALSE else null() end_if)
       ) $ j=0..m-1) $ i=0..n-1,
       plot::Polygon(plot::Point(0,0), plot::Point(0,(m+offset)*n-offset), hold(Color)=RGB::Black),
       op(options)
    ))
end_proc:

/*--
    dataplot_Plain
--*/
plot::data_Plain:= proc(l, c, t, options=[])
    name plot::data::Plain;
    local i, j, m, n, scene_default;
begin
    // set options:
    scene_default:= 
        [hold(Axes) = hold(Box), hold(Scaling) = hold(UnConstrained), hold(Ticks) = 0,
         hold(Title) = ""];
    if options <> [] then
        if has(options,hold(Axes)) then scene_default[1]:= FAIL end_if;
        if has(options,hold(Scaling)) then scene_default[2]:= FAIL end_if;
        if has(options,hold(Ticks)) then scene_default[3]:= FAIL end_if;
        if has(options,hold(Title)) then scene_default[4]:= FAIL end_if;
        options:= subs( scene_default, FAIL=null()) . options;
    else
        options:= scene_default
    end_if;

    m:= nops(l); n:= nops(l[1]);
    return(plot::Scene(    
        ((plot::Polygon(plot::Point(i,j,l[j+1][i+1]), plot::Point(i,j+1,l[j+2][i+1]),
            plot::Point(i+1,j+1,l[j+2][i+2]), hold(Color) = c[j+1], hold(Filled) = TRUE),
        
        plot::Polygon(plot::Point(i,j,l[j+1][i+1]),plot::Point(i+1,j,l[j+1][i+2]),
            plot::Point(i+1,j+1,l[j+2][i+2]), hold(Color) = c[j+1], hold(Filled) = TRUE),
        
        plot::Polygon(plot::Point(i,j,l[j+1][i+1]), plot::Point(i,j+1,l[j+2][i+1]),
            plot::Point(i+1,j+1,l[j+2][i+2]), plot::Point(i+1,j,l[j+1][i+2]),
            plot::Point(i,j,l[j+1][i+1]), hold(Color) = c[m]))
        $ i = 0..n-2) $ j = 0..m-2,
        op(options)
    ))
end_proc:

/*-----------------------------------------------------------*/
/*-------------------- Utilities ----------------------------*/
/*-----------------------------------------------------------*/
plot::data_set_titles:= proc(t, l, n)
    name plot::data;
    local i, j, k, m;
begin
    if (k:=contains(t, hold(Values))) <> 0 then
        delete t[k];
        if testtype(l[1], DOM_LIST) then
            t := t.[(expr2text(l[i][j])$i=1..nops(l))$j=1..nops(l[1])]
        else
            t := t.[expr2text(l[i])$i=1..nops(l)]
        end_if
    end_if;
    m:= nops(t);
    if m < n then t := t.[ "" $ i = 1..n-m] end_if;
    return(t)
end_proc:

/*-----------------------------------------------------------*/

plot::data_rectangle := proc(a,b,c,d,co,ti,sx,sy,vertical)
    name plot::data;
begin
    plot::Polygon(plot::Point(a,b), plot::Point(a,d), plot::Point(c,d), 
        hold(Filled) = TRUE, hold(Color) = co, hold(Title) = ti, hold(TitlePosition) =[0,0]
    ),
    plot::Polygon(plot::Point(a,b), plot::Point(c,b), plot::Point(c,d), 
        hold(Filled) = TRUE, hold(Color) = co
    ),
    (if args(0) = 6 then
        // Draw frame 
        plot::Polygon(plot::Point(a,b), plot::Point(a,d), plot::Point(c,d), plot::Point(c,b), plot::Point(a,b), hold(Color) = RGB::Black)
    else
        // Draw shadow frame 
        if vertical then
            if d < 0 then sy := -sy; end_if;
            plot::Polygon(plot::Point(a+sx,b),plot::Point(a+sx,d+sy),plot::Point(c+sx,d+sy), hold(Filled) = TRUE, hold(Color) = RGB::Gray),
            plot::Polygon(plot::Point(a+sx,b),plot::Point(c+sx,b),plot::Point(c+sx,d+sy), hold(Filled) = TRUE, hold(Color) = RGB::Gray)
        else
            if c < 0 then sx := -sx; end_if;
            plot::Polygon(plot::Point(a,b-sy),plot::Point(a,d-sy),plot::Point(c+sx,d-sy), hold(Filled) = TRUE, hold(Color) = RGB::Gray), 
            plot::Polygon(plot::Point(a,b-sy),plot::Point(c+sx,b-sy),plot::Point(c+sx,d-sy), hold(Filled) = TRUE,hold(Color) = RGB::Gray)
        end_if
    end_if)
end_proc:

/*-----------------------------------------------------------*/

plot::data_ticks:= proc(l, ymin=FAIL, ymax=FAIL)
    name plot::data;
    local i, tx, ty, tz;
begin
    if type(l[1][1]) = DOM_LIST then
        tx := [ l[i][j][1] $ j = 1..nops(l[i]) $ i = 1..nops(l) ];
        tx := plot::data_xyzticks(tx);
        ty := [ l[i][j][2] $ j = 1..nops(l[i]) $ i = 1..nops(l) ];
        ty := plot::data_xyzticks(ty, ymin, ymax);
        if nops(l[1][1]) = 2 then
            return([ tx, ty ])
        else
            tz := [ l[i][j][3] $ j = 1..nops(l[i]) $ i = 1..nops(l) ];
            tz := plot::data_xyzticks(tz);
            return([ tx, ty, tz ])
        end_if
    else 
        ty := [ op(l[i]) $ i = 1..nops(l) ];
        return([ 0, plot::data_xyzticks( ty ) ])
    end_if
end_proc:

/*-----------------------------------------------------------*/

plot::data_xyzticks:= proc(l, ymin=FAIL, ymax=FAIL)
    name plot::data;
    local d, i, j, n;
begin
    // Convert ty to a set to remove multiple entries
    // and re-convert into a list
    l:= sort( [op({op(l)})] );  
    n:= nops(l);
    // Maximal tick marks: 40 (Internal upper bound!)
    if n > 40 then
       // 10 (arbitrary) tick marks 
       return( 10 )
    end_if;
    if type(ymin) = DOM_FLOAT then
        d:= (ymax-ymin)/20.0
    else
        d:= (l[n]-l[1])/20.0
    end_if;
    j:= l[1];
    for i from 2 to n do
        if l[i] - j < d then
            l[i]:= l[i] = ""
        else
            j:= l[i]
        end_if
    end_for;

    return(l)
end_proc:

/*-----------------------------------------------------------*/

plot::data_curve_min_max := proc(p, a, b, g)
    name plot::data;
    local j, ma, mi, v, x;
begin
    mi:= infinity; ma:= -infinity;
    for j from 1 to nops(p) do
        for x from a[j] to b[j] step (b[j]-a[j])/g do
            v := p[j](float(x));
            if v < mi then 
                mi := v
            elif v > ma then
                ma := v
            end_if
        end_for
    end_for;
    return( [mi, ma] )
end_proc:

/*-----------------------------------------------------------*/

plot::data_point2d_list := proc(l, t, c)
    name plot::data;
    local i, j;
begin
    plot::Pointlist(                     
        plot::Point(l[j][i][1],l[j][i][2],hold(Color)=c[j]) $ i = 1..nops(l[j]), 
        hold(Title) = t[j], hold(TitlePosition) = [0,0],
        hold(PointWidth) = 40, hold(PointStyle) = hold(FilledCircles)
    ) $ j=1..nops(l)
end_proc:

/*-----------------------------------------------------------*/

plot::data_point3d_list := proc(l, t, c)
    name plot::data;
    local i, j;
begin
    return(plot::Pointlist(
        plot::Point(l[j][i][1],l[j][i][2],l[j][i][3],hold(Color) = c[j])$ i = 1..nops(l[j]),
        hold(Title) = t[j], hold(TitlePosition) = [0,0]
    ) $ j = 1..nops(l))
end_proc:

/*-----------------------------------------------------------*/

plot::data_checkformat := proc(l, dtyp)
    name plot::data;
    local j, check;
begin
    check:= proc(l, dtyp)
        local a, b, dim, i, j, nl, t;
    begin
        if domtype(l) = DOM_ARRAY or 
            l::dom::hasProp( Cat::Matrix ) = TRUE 
        then
            if (dim := op(l,[0,1])) > 2 then 
                error("one- or two-dimensional array or matrix expected")
            end_if
        else
            return( l )
        end_if;

        (a[i] := op(l, [0, i+1, 1])) $ i=1..dim;
        (b[i] := op(l, [0, i+1, 2])) $ i=1..dim;

        nl := [];
        if dim = 1 then
            t := {hold(Points), hold(Lines), hold(Curves), hold(LinesPoints), hold(CurvesPoints)};
            for i from a[1] to b[1] do
                if testtype(l[i], Type::Real) then
                    if contains(t, dtyp) then
                        nl := append(nl, [i, l[i]])
                    else
                        nl := append(nl, l[i])
                    end_if
                end_if
            end_for
        else // dim = 2
            for i from a[1] to b[1] do
                for j from a[2] to b[2] do
                    if testtype(l[i,j], Type::Real) then
                        nl := append(nl, [i, j, l[i,j]])
                    end_if
                end_for
            end_for
        end_if;

        return(nl)
    end_proc:

    if type(l) = DOM_LIST then
        [ check(l[j], dtyp) $ j = 1..nops(l) ]
    else
        [ check(l, dtyp) ]
    end_if
end_proc:

