// $Date: 2000/09/22 16:31:24 $  $Author: frankp $   $Revision: 1.3 $

/*++
    inequality - plot inequalities

    Call:

    plot::inequality(ineq, left..right, bottom..top 
             <, n>
             <, Colors = [col1, col2, col3]>,
             <, opt1 ...> 
    );

    Parameters

    ineq             -- a list [f1,f2,..] of real functions of two arguments
    left, right, bottom, top 
                     -- numerical real values
    n                -- optional nonnegative integer
    col1, col2, col3 -- color names (RGB values)
    opt1             -- plot option(s)

    Synopsis

    inequality serves for displaying points (x,y) in the rectangle 
    Q = [left,right] x [bottom,top] satisfying the inequalities

            f1(x,y)>=0 and  f2(x,y)>=0 and  ...           (1)

    specified by ineq = [f1,f2,..].

    The rectangle Q is divided into 2^n times 2^n subrectangles.
    The default value for n is 6.

    A subrectangle is displayed with the color col1, if all its points
    (x,y) satisfy f(x,y)>0 for all functions f in ineq. Consequently, 
    it is guaranteed that points of this color satisfy (1).

    A subrectangles is displayed with the color col3, if there is a function 
    f in ineq such that all points in the subrectangle satisfy f(x,y)<0.
    Consequently, it is guaranteed that points of this color do not satisfy (1).

    The remaining subrectangles are displayed with the color col2. They cover 
    the boundary of the region defined by (1).

    The default values are Colors = [RGB::Green, RGB::Yellow, RGB::Red].

    The plot is constructed via

    plot2d( option_sequence, [Mode = List , [ ... ] ])

    where [...] is a sequence of polygons. option_sequence may consist of 
    all valid scene options in plot2d. The default values are

    SceneOptions= [Scaling=UnConstrained, Labeling=TRUE, Labels=["",""]]

    Interval arithmetic is used to check the inequalities. The input functions 
    must be suitable for this kind of arithmetic, i.e., the calls 
         f1(Dom::Interval(left..right),Dom::Interval(bottome..top)) 
    etc. must produce valid intervals.

    Examples:

    >> f1:= proc(x,y) begin x^2+y^2-1 end_proc:
    >> plot::inequality([f1], -1..1, -1..1);

    >> f2:=proc(x,y) begin (cos(x)-y) end_proc: 
    >> f3:=proc(x,y) begin (cos(x)+y) end_proc:
    >> plot::inequality([f2, f3], -PI..PI, -2..2, 5);

    >> scene_options:= [Scaling = Constrained, Labeling = TRUE, Axes = Box,
             BackGround = RGB::White, ForeGround=RGB:: Black]:
    >> colors:=[RGB::Red, RGB::Black, RGB::White]:
    >> plot::inequality([f1, f2, f3], -2..2, -1..1, 5, Colors = colors,
             SceneOptions = scene_options ):

    Note: (W.Oevel)
    The following implementation works for 1.4 as well as for 1.5.
    For efficiency it avoids Dom::Interval as far as possible.
    Dom::Interval is only used through (Dom::Interval)::sign
    Still: it's terribly slow.
    There's no guarantee that for n=0 and
    (Dom::Interval)::sign(g(IX, IY)))=0
    there really are at least one point in the subrectangle!
++*/
plot::inequality:= proc(f, rx, ry)
  local x, y, i, left, right, bottom, top, g, n, 
        PColor, ZColor, NColor, options;
begin
    if args(0) < 3 then error("expecting at least 3 arguments") end_if;
  
    if testtype( f,Type::ListOf(Type::Arithmetical,1) ) then
        x:= indets(f);
        case x
        of 0 do
            x:= genident("x"); y:= genident("y");
            break
        of 1 do
            x:= op(x,1); y:= genident("y");
            break
        of 2 do
            y:= op(y,2); x:= op(x,1);
            break
        otherwise
            error("too many unknowns")
        end_case;
        f:= map(f, proc(g) begin (X,Y) -> eval(subs(g,x=X,y=Y)) end)
    elif not testtype(f, Type::ListOf(Type::Function, 1)) then
        error("expecting list of functional expressions")
    end_if;

    if type(rx) <> "_range" then 
        error("2nd argument: range of the form 'x0..x1' expected")
    elif type(ry) <> "_range" then 
        error("3rd argument: range of the form 'y0..y1' expected") 
    end_if;

    left:= float(op(rx,1)): right:= float(op(rx,2)):
    bottom:= float(op(ry,1)): top:= float(op(ry,2)):
    if map({left,right,bottom,top},domtype) <> {DOM_FLOAT}
        then error("plot range must be specified by numerical values"); 
    end_if;
    if left>=right or bottom>=top then error("invalid range") end_if;

    // default values for optional arguments: 
    n:=6; PColor:=RGB::Green: ZColor:=RGB::Yellow: NColor:=RGB::Red:

    // override default values by optional arguments: 
    options:= [args(4..args(0))];
    if options <> [] and domtype(options[1]) = DOM_INT then
        if n < 1 then
            error("3rd argument: nonnegative integer expected")
        else
            n:= options[1]; delete options[1];
        end_if
    end_if;

    for i from 1 to nops(options) do
    g:= options[i];
        if type(g)="_equal" and op(g,1)=hold(Colors) then
            if not testtype(op(g,2),Type::ListOf(Type::ListOf(Type::Real,2,3),3,3)) then
                error("option 'Colors': list of three RGB color values expected")
            else
                [PColor, ZColor, NColor]:= op(g,2);
                delete options[i];
                break
            end_if
        end_if
    end_for;

    return( plot::Group(plot::inequality_main(
        f,left,right,bottom,top, n, PColor, ZColor,NColor
    ),op(options)) )
end_proc:

// returns plotdata for plot::inequality
plot::inequality_main:= proc(f, left, right, bottom, top, n, PColor, ZColor, NColor)
  local g, v, vv, IX, IY, mx, my;
begin
  IX:= Dom::Interval(left..right): 
  IY:= Dom::Interval(bottom..top):
  vv:=1;
  for g in f do
    //!! For expression g replace g(IX,IY) by subs(g, {x=IX, y=IY}, Unsimplified) 
    if (v:=(Dom::Interval)::sign(g(IX,IY)))={-1} then
        return(plot::Polygon(
                plot::Point(left,bottom), plot::Point(right,bottom), plot::Point(right,top), 
                hold(Filled)=TRUE, hold(Color)=NColor
            ),
            plot::Polygon(
                plot::Point(left,bottom), plot::Point(left,top), plot::Point(right,top), 
                hold(Filled)=TRUE, hold(Color)=NColor
            )
        ) 
    end_if;
    vv:= min(vv, op(v));
  end_for;
  if vv = 1 then // inequalities are definitely satisfied for all points 
    return(plot::Polygon(
            plot::Point(left,bottom), plot::Point(right,bottom), plot::Point(right,top), 
            hold(Filled)=TRUE, hold(Color)=PColor
        ),
        plot::Polygon(
            plot::Point(left,bottom), plot::Point(left,top), plot::Point(right,top), 
            hold(Filled)=TRUE, hold(Color)=PColor
        )
    );
  else /*vv=0 */ // inequalities may be satisfied for some points 
     if n=0 then
        return(plot::Polygon(
                plot::Point(left,bottom), plot::Point(right,bottom), plot::Point(right,top),
                hold(Filled)=TRUE, hold(Color)=ZColor
            ),
            plot::Polygon(
                plot::Point(left,bottom), plot::Point(left,top), plot::Point(right,top), 
                hold(Filled)=TRUE, hold(Color)=ZColor
            )
        );
     else // have a closer look, i.e., investigate subrectangeles 
          mx:=(left+right)/2; my:=(bottom+top)/2;
          plot::inequality_main(f, left, mx, bottom,my, n - 1,PColor, ZColor, NColor), 
          plot::inequality_main(f, left, mx, my,   top, n - 1,PColor, ZColor, NColor), 
          plot::inequality_main(f, mx,right, bottom,my, n - 1,PColor, ZColor, NColor), 
          plot::inequality_main(f, mx,right, my,   top, n - 1,PColor, ZColor, NColor);
     end_if
  end_if
end_proc:

