/*++
Author: W. Oevel 
1st version: 3.9. 2000


plot::ode - plot the numerical solution of an ODE

Intro:  the ODE  dY/dt = f(t,Y), Y(t0)=Y0 is solved
        by numeric::odesolve. 
        A mesh of sample points Y0=Y(t0), Y1=Y(t1), .. is
        generated. 

        These sample points (or components of these points)
        can be plotted as curves or points

Calls:
  plot::ode([t0,t1,..], f, Y0, 
            <method,> <Stepsize = h,> <RelativeError = tol,> 
            [F1 <, Style = style1> <, Color = c1>],
            [F2 <, Style = style2> <, Color = c2>], ...)

Parameters:
    t0, t1, .. - sample points for the "time" parameter: real numerical values
    f          - a procedure (t, Y) -> f(t,Y) representing
                 the rhs of the ODE dY/dt = f(t, Y)
    Y0         - the initial condition: a list of numerical values
    method            - see numeric::odesolve
    Stepsize = tol    - see numeric::odesolve
    RelativeError = h - see numeric::odesolve

    F1, F2, .. - the 'generators of plot data':
                 procedures (t, Y) -> F(t, Y) providing
                 the numerical data that are to be plotted
                 (see the details below).
                 These functions must accept a float t
                 and a list Y of floats. The list Y
                 represents the solutions vector Y(t)
                 of the ODE.
                 The functions F1, F2, .. must produce lists
                 with numerical values. All lists must be
                 either of length 2 (this produces a 2D plot)
                 or of length 3 (this produces a 3D plot).
    style1, style2, ..
               - the style in which the solution curve is plotted:
                 either 'Points' (points only),
                 or 'Lines' (curve with straight line segments)
                 or 'Splines' (curve via cubic spline interpolation)
    c1, c2, .. - RGB colors

Return Value: an object of type plot::Group

Details:
   - A mesh of sample points Y0=Y(t0), Y1=Y(t1), .. is
     generated. 
     The functions F1, F2 etc. are called at these points,
     each generating a data sequence 
     (F1(t0, Y0), F1(t1, Y1), ...),
     (F2(t0, Y0), F2(t1, Y1), ...).
     Each of these sequences is regarded as a sequence
     of mesh points in R^2 (if F1, F2, .. produce lists with 2 elements)
     or in R^3 (if F1, F2, .. produce lists with 3 elements)
     Each sequence is regarded a curve, which is plotted in the
     specified style and the specified color.
     E.g.:
        F1 := (t, Y) -> [t, Y[1]]    --> x-axis == t
                                         y-axis == Y[1]
        F2 := (t, Y) -> [Y[1], Y[2]] --> x-axis == Y[1]
                                         y-axis == Y[2]
        F2 := (t, Y) -> [t, h(Y)]    --> x-axis == t
                                         y-axis == some function h(Y)

   - A combination of lines and sample points can be generated
     by using the same 'generator' twice with different styles,
     e.g.,
     plot::ode([t0,..], f, Y0, [F1, Style = Points], [F1, Style = Lines])
   - For a spline interpolation, the mesh points t0, t1, ... must
     be monotonically increasing.
     For Style = Points or Style = Lines the mesh points can be
     arbitrary.

Example1:
  We consider Y = [y1, y2, y3], dY/dt = [y2*y3-y2^2*y1, ..], Y(0)=[1,1,1]:
  We plot the components Y[2] and Y[3] against t as a 3D-curve:

>> obj:= plot::ode(
         [i $ i=0..4],
         (t, Y) -> [Y[2]*Y[3] - Y[2]^2*Y[1],
	            Y[2]*Y[1]^2 + Y[3]^2, 
		    -Y[1]*Y[2] - Y[2]*Y[3]],
	 [1, 1, 1], 
         [(t, Y) -> [t,Y[2], Y[3]], Style = Points, Color = RGB::Red],
	 [(t, Y) -> [t,Y[2], Y[3]], Style = Lines, Color = RGB::Green],
         [(t, Y) -> [t,Y[2], Y[3]], Style = Splines, Color = RGB::Blue]
        ):
>> obj::PointWidth:= 50:
>> plot(obj, Scaling = UnConstrained):

Example2: We plot the solution of dy/dt = f(t,y) together with
          a vector field plot of the right hand side of the ODE.
>> f:= (t, y) -> t*sin(t + y^2) - y:
>> p1 := plot::vectorfield([1, f(t, y)], t = 1..3, y = 0..1,
                           Grid = [10, 5], Color = RGB::Black):
>> p2 := plot::ode([i/3 $ i=0..12], (t,Y) -> [f(t, Y[1])], [0],
                [(t, Y) -> [t, Y[1]], Style = Points, Color = RGB::Red],
                [(t, Y) -> [t, Y[1]], Style = Splines, Color = RGB::Blue]
	       ):
>> plot(p1, p2, Scaling = Constrained, 
              GridLines = [Steps = 0.25, Steps = 0.5],
              Ticks = [Steps = 1.0, Steps = 0.5]):
++*/
plot::ode:= proc(t, f, Y0)
  local odesolveoptions, FF, dummy, N,
        F, Styles, Colors, AllStyles, 
        i, j, k, n, nn, Y, primitive,
        dimension, XX, YY, ZZ, tt;
begin
  if args(0) < 4 then error("expecting at least 4 arguments") end_if:

  // ----  checking input parameters "odesolve input"  ------------
  if type(t) <> DOM_LIST then
     error("expecting a list of sample points as first argument") 
  end_if;
  if {op(map(t, domtype@float))} <> {DOM_FLOAT} then
     error("cannot convert the sample points to real floats")
  end_if:

  [FF, odesolveoptions, dummy]:= 
      split([args(4..args(0))], x -> bool(domtype(x) = DOM_LIST));
  odesolveoptions:= op(odesolveoptions);


  // ----  checking input parameters "plot data generators" ---
  N:= nops(FF): // the number of curves to be plotted
  if N = 0 then
     error("need a procedure to generate plot data")
  end_if:

  // collect data generators, styles, and colors:
  F:= [0$N]:
  Styles := [Lines $ N]:
  Colors := [RGB::Black $ N]:
  AllStyles := [Points, Lines, Splines]:
  for j from 1 to N do
     F[j]:= FF[j][1];
     if domtype(F[j]) <> DOM_PROC then
        error("expecting a procedure to generate plot data"):
     end_if;
     for k from 2 to nops(FF[j]) do
         if type(FF[j][k]) <> "_equal" then
            error("expecting an equation 'Style = style' or 'Color = color',". 
                  " got '".expr2text(FF[j][k])."'");
         end_if;
         case op(FF[j][k], 1) 
         of Style do
            Styles[j]:= op(FF[j][k], 2);
            if {Styles[j]} minus {op(AllStyles)} <> {} then
               nn:= nops(AllStyles):
               error("unknown plot style '".expr2text(Styles[j])."'; ".
                     "available styles are ".
                     _concat(("'". expr2text(op(AllStyles,j)) . "', ")
                                  $ j = 1..nn-1,
                              "and '". expr2text(op(AllStyles,nn)) . "'"
                            )
                     );
            end_if:
            break;
         of Color do Colors[j]:= op(FF[j][k], 2);
            if domtype(Colors[j]) <> DOM_LIST then
               error("illegal color specification");
            end_if:
            break;
         otherwise
            error("expecting arguments of the form " .
                  "'[procedure, Style = style, Color = RGB color]'"):
         end_case;
     end_for:
  end_for:

  // --- the work starts ------------

  // --- solve the ode: ---
  n:= nops(t);  // the number of sample points
  if n < 2 then 
     error("the first argument must be a list with at least 2 sample points")
  end_if;

  Y:= [[t[1], Y0], 0 $ n-1]:  // initialize container for sample points
  userinfo(2, "starting to solve the ode"):
  for i from 2 to n do
     userinfo(3, "computing ".expr2text(i)."-th sample point:");
     Y[i]:= [t[i], numeric::odesolve(t[i-1]..t[i], f, Y[i-1][2], odesolveoptions)];
     userinfo(3, "--- time: ".expr2text(t[i]) . ", solution: ". expr2text(Y[i][2]));
  end_for:
  // --- the ode is solved. ----

  // --- generate plot data ----
  primitive:= [0 $ N];  // initialize container for primitives
  for j from 1 to N do  // the j-th curve generated
                        // by F[j] with Styles[j], Color[j]
     case Styles[j]
     of Points do
        primitive[j]:= plot::Pointlist(
                 plot::Point(F[j](op(Y[i]))) $ i=1..n,
                 Color = Colors[j]);
        break;
     of Lines do
        primitive[j]:= plot::Pointlist(
                 plot::Point(F[j](op(Y[i]))) $ i=1..n,
                 Color = Colors[j],
                 DrawMode = Connected);
        break;
     of Splines do
        dimension:= nops(F[j](op(Y[1]))):
        case dimension
        of 2 do
           XX:= numeric::cubicSpline([t[i], F[j](op(Y[i]))[1]] $ i = 1..n):
           YY:= numeric::cubicSpline([t[i], F[j](op(Y[i]))[2]] $ i = 1..n):
           tt:= genident("t"):
           primitive[j]:= plot::Curve2d(
	         [XX(tt), YY(tt)],
                 tt = t[1]..t[n],
		 Color = Colors[j],
		 Title = "",
		 Grid = [5*n]);
           break;
        of 3 do
           XX:= numeric::cubicSpline([t[i], F[j](op(Y[i]))[1]] $ i = 1..n):
           YY:= numeric::cubicSpline([t[i], F[j](op(Y[i]))[2]] $ i = 1..n):
           ZZ:= numeric::cubicSpline([t[i], F[j](op(Y[i]))[3]] $ i = 1..n):
           tt:= genident("t"):
           primitive[j]:= plot::Curve3d(
	         [XX(tt), YY(tt), ZZ(tt)],
                 tt = t[1]..t[n], 
		 Color = Colors[j], 
		 Title = "", 
		 Grid = [5*n]);
           break;
        otherwise error("generators of plot data must produce lists with " .
                        "2 or 3 elements"):
        end_case;
        break;
     otherwise
        error("unknown style"):
     end_case;
  end_for;

  plot::Group(primitive[j] $ j = 1..N);

end_proc:
