// $Date: 2000/11/01 13:49:22 $  $Author: frankp $   $Revision: 1.7 $

/*--
    getColorFormat(c)
    c: either 'Color = ...' or a list.

    Details:
    getColorFormat(c) converts c as follows:
                   c                    |          Result
      ----------------------------------|---------------------------
      [r,g,b]                           |  [Flat, [r,g,b]]
      [{Flat,Height}, [r,g,b]]          |  [{Flat,Height}, [r,g,b]]
      [Function, f]                     |  [Function, f]
      Color = [r,g,b]                   |  Color = [Flat, [r,g,b]]
      Color = [{Flat,Height}, [r,g,b]]  |  Color = [{Flat,Height}, [r,g,b]]
      Color = [Function,f]              |  Color = [Function, f]

    If this is not possible, then FAIL is returned.
--*/
plot::getColorFormat := proc(c)
    local f;
begin
    if type(c) <> "_equal" then
        f:= id
    elif op(c,1) <> hold(Color) then
        return( FAIL )
    else
        f:= x -> hold(Color) = x;
        c:= op(c,2)
    end_if;
    
    if testtype(c, Type::ListOf(Type::Arithmetical, 3, 3)) then
        return( f([hold(Flat),c]) )
    elif domtype(c) <> DOM_LIST or c = [] then
        return( FAIL )
    else
        case c[1]
        of hold(Flat) do
            case nops(c)
            of 2 do
                if testtype(c[2],Type::ListOf(Type::Arithmetical, 3, 3)) then
                    return( f(c) )
                else
                    return( FAIL )
                end_if
            of 1 do
                return( f(c) )
            otherwise
                return( FAIL )
            end_case
        of hold(Height) do
            case nops(c)
            of 1 do
                return( f(c) )
            of 3 do
                if testtype(c[2],Type::ListOf(Type::Arithmetical, 3, 3)) and
                   testtype(c[3],Type::ListOf(Type::Arithmetical, 3, 3))
                then
                    return( f(c) )
                end_if
            otherwise
                return( FAIL )
            end_case
        of hold(Function) do
            if nops(c) <> 2 then
                return( FAIL )
            else
                return( f(c) )
            end_if
        otherwise
            return( FAIL )
        end_case
    end_if
end_proc:

/*--
    conv2range(r)
    r : either form 'x = a..b' or 'x = [a,b]'

    Checks whether r has the form 'x = a..b' or 'x = [a,b]' with
    x an identifier and a and b arithemtical expressions. FAIL
    is returned if this is not the case.

    If r has the form 'x = [a,b]', then it is converted into
    the form 'x = a..b'. 
--*/
plot::conv2range := proc(e)
    local r, a, b;
begin
    if type(e) <> "_equal" or domtype(op(e,1)) <> DOM_IDENT then
    	return( FAIL )
    end_if;
    r:= op(e,2);
    case type(r)
    of DOM_LIST do
    	if nops(r) <> 2 then return( FAIL ) end_if;
	break
    of "_range" do
        break
    otherwise
    	return( FAIL )
    end_case;
    a:= op(r,1); b:= op(r,2);
    if not testtype(a,Type::Arithmetical) or has(a,op(e,1)) or
       not testtype(b,Type::Arithmetical) or has(b,op(e,1)) 
    then
    	return( FAIL )
    end_if;
    return( op(e,1) = a..b )
end_proc:

/*--
    Takes a range of the form 'x=a..b' and converts it into the form 'x=[a,b]'.

    For internal use only.
--*/
plot::range2list := r -> subsop(r,2=[op(r,[2,1]),op(r,[2,2])]):

/*--
    Takes a range of the form 'x=[a,b]' and converts into the form 'x=a..b'.

    For internal use only.
--*/
plot::list2range:= l -> subsop(l,2=op(l,[2,1])..op(l,[2,2])):


