// $Date: 2000/10/30 18:01:48 $  $Author: frankp $   $Revision: 1.6 $

// W. Oevel
/*++
    plot::vectorfield - generates a plot of two-dimensional vector field

    Call:
    vectorfield([v1,v2], x = a..b, y = c..d<, options>)

    v1, v2    : x- and y-component of the vectorfield (functional expressions
                in x and y
    x,y       : identifiers
    a, b, c, d: arithmetical expressions
    options   : plot options for two-dimensional graphical objects

    Return Value:
    A graphical object of the domain type plot::Group.

    Details:
    * The form of the arrows used for displaying the vectors
      is built for the scene option Scaling = Constrained.
      For Scaling = UnConstrained you can resize the plot window
      and change the scaling. 
      ** The directions of the vectors change in the correct way. **
      However, the tips of the arrows will be distorted and may 
      look strange (this is unavoidable)
               ^                               __^_
              /|\                                | \
               |       may be distorted to       |
               |                                 |
       So the user is advised to use Scaling = Constrained for
       a nice graphical appearance of the arrows, whenever this
       is appropriate.
       However, if the scale ratio  (ymax-ymin)/(xmax-xmin) is 
       not close to 1, then Scaling = Constrained is usually not 
       appropriate. If you do not like the arrows, then you must
       reparametrize the vector field such that the scale ratio
       in the new x-y parameters is close to 1.

    Examples:  
    We demonstrate a plot of the vector field v(x,y) = [1, sin(x)+cos(y)].

    >> DIGITS:=5:
    >> field:= plot::vectorfield(
        [1,sin(x)+cos(y)], x=0..3, y=1..2,
        Grid=[30,20], Color = [Flat, RGB::Red]
       ):

    It is the directional field associated with the ode  y'(x) = sin(x)+cos(y).  
    We insert a curve representing the numerical solution of this ode into this  
    plot. We compute the numerical solution of y'(x)=sin(x)+cos(y), y(1)=1.2 
    via numeric::odesolve. DIGITS is increased to get rich sample of points:

    >> DIGITS:=20: 
    >> f:= proc(x,y) begin [sin(x)+cos(y[1])] end_proc:
    >> data:= numeric::odesolve(1..2, f, [1.2], Alldata):
    >> curve:= plot::Polygon(
        ([data[i][1], data[i][2][1]], [data[i+1][1], data[i+1][2][1]]
           ) $ i=1..nops(data)-1, 
        Color = [Flat, RGB::Blue]
       ):

    Define a circle with center (1.5 ,1.5) and radius 1/2: 
    >> circle:= plot::Ellipse2d( [1.5,1.5],1/2,1/2 ):

    Plot the three objects in a single plot:  
    >> plot(
         field, circle, curve, Axes=Box, Labeling = TRUE, Scaling = Constrained
       )

    Do it again, this time with Scaling = UnConstrained:
    >> plot(
         field, circle, curve, Axes=Box, Labeling = TRUE, Scaling = UnConstrained
       )
++*/
plot::vectorfield:= proc(f,xr,yr)
    local i, t, opt, grid, calc_pointlist, convf, grid_pos;
begin
    calc_pointlist := proc(list,opt)
       local f_x, f_y, points, grad_x, grad_y,
             act_x, xmin, xmax, xgrid, delta_x,
             act_y, ymin, ymax, ygrid, delta_y,
             delta_xy, ratio, rat_x, rat_y;
    begin
        // Check the arguments of list, left out:
        xmin    := float(op(list, [2,2,1])):
        xmax    := float(op(list, [2,2,2])):
        ymin    := float(op(list, [3,2,1])):
        ymax    := float(op(list, [3,2,2])):
        xgrid   := op(list, [4,2,1]):
        ygrid   := op(list, [4,2,2]):
        delta_x := (xmax - xmin)/(xgrid-1):
        delta_y := (ymax - ymin)/(ygrid-1):
        delta_xy:= min(delta_x,delta_y); // Walter, 18.5.99

        points  := []:
    
        f_x := convf(op(list, [1,1]), op(list, [2,1]), op(list, [3,1]));
        f_y := convf(op(list, [1,2]), op(list, [2,1]), op(list, [3,1]));

        for act_x from xmin to xmax step delta_x do
            for act_y from ymin to ymax step delta_y do
              // Evaluate the expression expr_xy at the point
              // act_x, act_y).
             grad_x := float(f_x(act_x, act_y));
             grad_y := float(f_y(act_x, act_y));

             ratio  := (2.5*(grad_x^2+grad_y^2)^(1/2)):
             if not iszero(ratio) then ratio := 1.0/ratio; end;
             rat_x  := delta_xy*ratio*grad_x:
             rat_y  := delta_xy*ratio*grad_y: // Walter, 18.5.99
                                                      
             points := append(points, 
                plot::vector(
                    plot::Point(act_x-rat_x, act_y-rat_y),
                    plot::Point(act_x+rat_x, act_y+rat_y),
                    PI/7, 0.5
                )
             );
            end_for:
        end_for:
        return( plot::Group(op(points), op(opt)) )
    end_proc:

    convf := proc(f, x, y)
        option escape;
    begin
        if not contains({DOM_PROC, DOM_EXEC, DOM_FUNC_ENV}, domtype(f)) then 
           if contains({DOM_EXPR,DOM_IDENT,DOM_INT,DOM_RAT,DOM_FLOAT}, domtype(f)) then
              return( proc(dx,dy) name plot::vectorfield::convf;
                  begin subs(f,x=dx,y=dy) end 
              )
           else
              error("illegal argument: ".expr2text(f))
           end_if 
        end_if
    end_proc:

    if testargs() then
        if args(0) < 3 then
            error("at least three arguments expected")
        elif not testtype(f,Type::ListOf(Type::Arithmetical,2,2)) then
            error("1st argument: list of two arithmetical expressions expected")
        elif type(xr) <> "_equal" or domtype(op(xr,1)) <> DOM_IDENT or not testtype(op(xr,2),Type::Real..Type::Real) then
            error("2nd argument: form 'x = a..b' with real values for 'a' and 'b' expected")
        elif type(yr) <> "_equal" or domtype(op(yr,1)) <> DOM_IDENT or not testtype(op(yr,2),Type::Real..Type::Real) then
            error("3rd argument: form 'y = c..d' with real values for 'c' and 'd' expected")
        elif op(xr,1) = op(yr,1) then
            error("range for ".expr2text(op(xr,1))." specified twice")
        end_if
    end_if;

    // extract option 'Grid', or set its default value if not given
    opt:= [args(4..args(0))];
    grid:= hold(Grid) = [20,20];
    if has(opt,hold(Grid)) then
        grid_pos:= 0;
        for i from 1 to nops(opt) do
            if type(opt[i]) = "_equal" and op(opt[i],1) = hold(Grid) then
                t:= op(opt[i],2);
                if testargs() and (not testtype(t,Type::ListOf(DOM_INT,2,2)) or t[1] < 2 or t[2] < 2) then
                    error("option 'Grid': list of two integers greater than one expected")
                end_if;
                grid:= opt[i];
                grid_pos:= i;
                break
            end_if
        end_for;
        if grid_pos <> 0 then
            delete opt[grid_pos];
            if testargs() and has(opt,hold(Grid)) then
                error("option 'Grid' specified twice")
            end_if
        end_if
    end_if;
    return( calc_pointlist([args(1..3),grid],opt) )
end_proc:

