// kg, 08/07/94
// Lsys -- the domain of Lindenmeyer systems

domain plotlib::Lsys

/*
new -- create new L-system

new(deg,start,rule...)

deg   - turtle degree (real number)
start - starting rule (non-empty string)
rule  - rule of the form '<lhs>=<rhs>'

The <lhs> must be a string of length 1. The <rhs> may be a string,
a turtle primitive or a color value. A turtle primitive is one of the
identifiers 'Line', 'Move', 'Left', 'Right', 'Push' or 'Pop'. A color value
is a list with 3 floats in the range 0..1 which define the
red-, green- and blue value of the color.
*/

new:= proc(deg:Type::Real, start:DOM_STRING)
    local i, j, p, startrule, pr, rule, prim, r, Lhs, Rhs, prims, rules, b, g, fdeg, gen;
    option escape;
begin
    if args(0) < 3 then error("wrong no of args") end_if;
    fdeg:= float(deg);

    // parse start rule
    if length(start) = 0 then
	error("illegal start rule")
    end_if;
    p:= revert([ substring(start,j,1) $ j=0..length(start)-1 ]);
    startrule:= subsop(proc(T) name start; local r, f; begin
	    gen:= gen-1;
	    if gen = 0 then r:= prim else r:= rule end_if;
	    f:= _fconcat(op(map("p", r)));
	    T:= f(T);
	    gen:= gen+1;
	    T
	end_proc, [4,3,2,1,1,1] = p);

    // parse rules
    prim:= proc() begin id end_proc;
    rule:= proc(r) begin prim(r) end_proc;
    pr:= {};
    prims:= {};
    rules:= {};

    for i from 3 to args(0) do
	r:= args(i);
	if type(r) <> "_equal" then
	    error("illegal rule")
	end_if;
	if not testtype(op(r), Type::Product(DOM_STRING,
		    Type::Union(DOM_STRING, DOM_IDENT,
			Type::ListOf(Type::Real,3,3)))) then
	    error("illegal rule")
	end_if;
	Lhs:= op(r,1);
	if length(Lhs) <> 1 then
	    error("illegal left hand side of rule")
	end_if;
	Rhs:= op(r,2);
	case domtype(Rhs)
	of DOM_IDENT do
	    if contains(prims, Lhs) then
		error("multiple turtle Lhs")
	    end_if;
	    prims:= prims union {Lhs};
	    // rule for drawing primitive
	    case (Rhs)
	    of hold(Line) do
		pr:= pr union {"F"};
		prim(Lhs):= proc(T) name Prim; begin plotlib::Turtle::line(T, 1.0) end_proc;
		break;
	    of hold(Move) do
		pr:= pr union {"f"};
		prim(Lhs):= proc(T) name Prim; begin plotlib::Turtle::move(T, 1.0) end_proc;
		break;
	    of hold(Push) do
		pr:= pr union {"["};
		prim(Lhs):= proc(T) name Prim; begin plotlib::Turtle::push(T) end_proc;
		break;
	    of hold(Pop) do
		pr:= pr union {"]"};
		prim(Lhs):= proc(T) name Prim; begin plotlib::Turtle::pop(T) end_proc;
		break;
	    of hold(Left) do
		pr:= pr union {"+"};
		prim(Lhs):= proc(T) name Prim; begin plotlib::Turtle::left(T, fdeg) end_proc;
		break;
	    of hold(Right) do
		pr:= pr union {"-"};
		prim(Lhs):= proc(T) name Prim; begin plotlib::Turtle::right(T, fdeg) end_proc;
		break;
	    otherwise error("illegal right hand side of rule")
	    end_case;
	    break;

	of DOM_LIST do
	    if contains(prims, Lhs) then
		error("multiple turtle lhs")
	    end_if;
	    prims:= prims union {Lhs};
	    Rhs:= map(Rhs, float);
	    if min(op(Rhs)) < 0.0 or max(op(Rhs)) > 1.0 then
		error("illegal color")
	    end_if;
	    prim(Lhs):= subsop(proc(T) name Color; begin plotlib::Turtle::color(T,"r","g","b") end_proc,
			    [4,2]=Rhs[1], [4,3]=Rhs[2], [4,4]=Rhs[3]);
	    break;

	otherwise
	    if contains(rules, Lhs) then
		error("multiple rule lhs")
	    end_if;
	    rules:= rules union {Lhs};
	    // grammar rule
	    p:= revert([ substring(Rhs,j,1) $ j=0..length(Rhs)-1 ]);
	    rule(Lhs):= subsop(proc(T) name Rule; local r, f; begin
		    gen:= gen-1;
		    if gen = 0 then r:= prim else r:= rule end_if;
		    f:= _fconcat(op(map("p", r)));
		    T:= f(T);
		    gen:= gen+1;
		    T
		end_proc, [4,3,2,1,1,1] = p);
	end_case;
    end_for;

    // insert default drawing primitive rules
    if not contains(pr, "F") then
	prim("F"):= proc(T) name Prim; begin plotlib::Turtle::line(T, 1.0) end_proc;
    end_if;
    if not contains(pr, "f") then
	prim("f"):= proc(T) name Prim; begin plotlib::Turtle::move(T, 1.0) end_proc;
    end_if;
    if not contains(pr, "[") then
	prim("["):= proc(T) name Prim; begin plotlib::Turtle::push(T) end_proc;
    end_if;
    if not contains(pr, "]") then
	prim("]"):= proc(T) name Prim; begin plotlib::Turtle::pop(T) end_proc;
    end_if;
    if not contains(pr, "+") then
	prim("+"):= proc(T) name Prim; begin plotlib::Turtle::left(T, fdeg) end_proc;
    end_if;
    if not contains(pr, "-") then
	prim("-"):= proc(T) name Prim; begin plotlib::Turtle::right(T, fdeg) end_proc;
    end_if;
        
    // deg, start, args(i) are used for printing only
    r:= new(dom, startrule, g->(gen:=g), deg, start, args(i) $ i=3..args(0));
    delete deg, start, i, j, p, startrule, pr, Lhs, Rhs, prims, rules, b, g;
    r; r:= FAIL; %2
end_proc;

/*
print -- return expression to print instead of L-system l
*/

print:= proc(l:dom) local i; begin
    hold(plotlib::Lsys)(extop(l,i) $ i=4..extnops(l))
end_proc;

/*
plot -- execute L-system l for gen generations and plot turtle path
*/

plot:= proc(l:dom, gen:Type::PosInt)
begin
    if args(0) <> 2 then error("wrong no of args") end_if;
    extop(l,2)(gen);
    plotlib::Turtle::plot(extop(l,1)(plotlib::Turtle()));
end_proc;

/*
path -- execute L-system l for gen generations and return turtle path
*/

path:= proc(l:dom, gen:Type::PosInt)
begin
    if args(0) <> 2 then error("wrong no of args") end_if;
    extop(l,2)(gen);
    plotlib::Turtle::path(extop(l,1)(plotlib::Turtle()));
end_proc;

end_domain:

// end of file
