/*--

plotlib::animate - create an animation of 2D/3D plots of functions
   
Calling sequence:

   animate2d(t=[c,d], Frames=frames, Format=format, plot2d_options)
   animate3d(t=[c,d], Frames=frames, Format=format, plot3d_options)
   animate(plot_func, t=[c,d], Frames=frames, Format=format, plot_func_options)
   
Parameter:

   t      - an identifier
   c, d   - real numbers
   frames - an integer > 0
   format - Gif 
   
Summary:   

- animate2d(t=[c,d], [Mode=Curve, [x, F(x,t)], x=[a,b]]) is a typical
  call to the animate2d function where F is a real function in x and t,
  and [a,b] specifies the horizontal real range on which F is plotted
  while [c,d] specifies how the frame coordinate varies from one frame
  to the next. 
   
- The animate function provides support for two/three-dimensional plots
  of one or more functions specified as expressions, procedures, parametric
  functions, or lists of points/polygons.

- The frame parameter range must be specified and must be the first argument
  of animate2d/animate3d. The plotting function must be the first argument
  and the frame parameter range must be the second argument of animate
  respectively. The next two arguments are optional. The default value of
  Frames is 10, the default value of Format is Gif.

- Note that the animate functions generates only the corresponding frames.
  Use show_animation to animate the frames. To delete the generated frames
  use delete_frames.

- To animate a sequence of GIF pictures MuPAD uses the external program
  xanim by Mark Podlipec (podlipec@wellfleet.com). 
  You can get the program via anonymous ftp, for example:
  ftp ftp.uni-paderborn.de: /unix/X11/ftp.x.org/R5contrib/xanim264.tar.gz

Examples:

>> export(plotlib):

>> animate3d(t=[1,2], Axes=None, [Mode=Surface,
          [x, y, cos(t*x)*sin(t*y)], x=[-PI,PI],y=[-PI,PI],
          Style=[HiddenLine, Mesh]]):
>> show_animation(%): 
>> plotlib::delete_frames(%2):
 
>> animate2d(t=[1,2], [Mode=Curve,[x, sin(t*x)], x=[-10,10],Grid=[50]],
                   [Mode=Curve,[x, cos(t*x)], x=[-10,10],Grid=[50]]):
 
>> animate(polarplot, t=[1,2], Frames=20, Axes=None, 
           [[sin(x*t),x], x=[-4,4], Grid=[100]]):

See also:

plot2d, plot3d
--*/

plotlib::AnimationNr   := 1:

plotlib::animate := proc()
  local a, arguments, c, d, filename, format, frames,
        frame_nr, i, plotFunc, s, start, t;
begin
  if sysname() <> "UNIX" then
	error("this function only runs on a unix platform")
  end_if;

  frames := plotlib::FramesDefault; format := plotlib::FormatDefault;
  plotFunc := args(1); start := 3;

  i := args(2);
  if not testtype(i, "_equal") then
     error("First argument must be the frame parameter range.")
  end_if;
  s := op(i, 1);
  if not (testtype(s, DOM_IDENT) and 
     testtype(op(i, 2), DOM_LIST) and nops(op(i, 2)) = 2) then
     error("frame parameter range must be of type t = [a,b]")
  end_if;
  c := op(i,[2,1]); d := op(i,[2,2]);

  i := args(3);
  if testtype(i, "_equal") then
     if op(i, 1) = Frames then
        frames := op(i, 2); start := start + 1;
        if not (testtype(frames, DOM_INT) and frames > 0) then
           error("Frames must be an integer > 0.")
        end_if
      end_if
  end_if;

  i := args(start);  
  if testtype(i, "_equal") then
      if op(i, 1) = Format then
        format := op(i, 2); start := start + 1;
        if not (format = Gif) then
           error("Format must be Gif.")
        end_if
      end_if 
  end_if;

  
  arguments:= args(start..args(0));
  filename := plotlib::tmpDir."/".format.getpid();
  filename := filename."_".plotlib::AnimationNr."." ;
  frame_nr := 1;
  sysassign( plotlib::AnimationNr, plotlib::AnimationNr + 1 );
  
  print(Unquoted, "".frames." Frames to generate ...");
  for t from c to d step (d-c)/(frames-1) do
      a := subs(arguments, s=t); 
      plotFunc(PlotDevice=[filename.(frame_nr+100), format], a);
      print(Unquoted, "Frame ".frame_nr." generated.");
      frame_nr := frame_nr+1
  end_for;

  [filename, format]

end_proc:

plotlib::animate2d := proc()
begin
   plotlib::animate(plot2d, args())
end_proc:

plotlib::animate3d := proc()
begin
   plotlib::animate(plot3d, args())
end_proc:

plotlib::show_animation := proc(list)
    local tool;
begin
  if sysname() <> "UNIX" then
        error("this function only runs on a unix platform")
  end_if;

  case list[2]
    of Gif    do tool := plotlib::AnimateGif;    break;
    otherwise    error("Can't animate ".expr2text(list[2])." pictures.")
  end_case;
  system(tool." ".list[1]."*");
  list
end_proc:

plotlib::delete_frames := proc(list)
begin 
  system("rm -f ".list[1]."*");
  list
end_proc:
