/* --------------------------------------------------------------------------
  plotlib::fieldplot - displays a 2 dimensional vector field
 
  call:   plotlib::fieldplot(<scene_options ,> object_1 <, object_2, ...>);
 
         fieldplot(arguments) is essentially the same call
         as plot2d(arguments), i.e., scene_options and
         object_1, object_2 etc. can be built in the same
         way as described for plot2d (see help(plot2d)).
         In addition, a further type of object is allowed,
         which represents a vectorfield. It is a list of
         the following form:
 
         [[v1(x,y),v2(x,y)], x=[xmin,xmax], y=[ymin,ymax], <options>]
 
       v1(x,y) - x-component of the vectorfield (expression of x,y)
       v2(x,y) - y-component of the vectorfield (expression of x,y)
       x       - the first independent variable
       xmin    - minimum value for the range in which x is evaluated
       xmax    - maximum value for the range in which x is evaluated
       y       - the second independent variable
       ymin    - minimum value for the range in which y is evaluated
       ymax    - maximum value for the range in which y is evaluated
       options - used to influence the graphical represen-
                 tation of the vectorfield object. All options listed
                 for plot2d() in the reference manual can
                 be used here.

    Note: the form of the arrows used for displaying the vectors
          is built for the scene option Scaling = Constrained.
          For Scaling = UnConstrained you can resize the plot window
          and change the scaling. 
          ** The directions of the vectors change in the correct way. **
          However, the tips of the arrows will be distorted and may 
          look strange (this is unavoidable)
               ^                               __^_
              /|\                                | \
               |       may be distorted to       |
               |                                 |
          So the user is advised to use Scaling = Constrained for
          a nice graphical appearance of the arrows, whenever this
          is appropriate.
          However, if the scale ratio  (ymax-ymin)/(xmax-xmin) is 
          not close to 1, then Scaling = Constrained is usually not 
          appropriate. If you do not like the arrows, then you must
          reparametrize the vector field such that the scale ratio
          in the new x-y parameters is close to 1.

Example:  
    We demonstrate a plot of the vector field v(x,y) = [1, sin(x)+cos(y)].       

    >> vectorfield_object:= [[1,sin(x)+cos(y)], x=[0,3], y=[1,2], 
                             Grid=[30,20], Color = [Flat, RGB::Red]]:

    It is the directional field associated with the ode  y'(x) = sin(x)+cos(y).  
    We insert a curve representing the numerical solution of this ode into this  
    plot. We compute the numerical solution of y'(x)=sin(x)+cos(y), y(1)=1.2 
    via numeric::odesolve. DIGITS is increased to get rich sample of points:

    >> DIGITS:=20: 
    >> f:= proc(x,y) begin [sin(x)+cos(y[1])] end_proc:
    >> data:= numeric::odesolve(1..2, f, [1.2], Alldata):
    >> curve_object:= [Mode = List,
    >>                 [polygon( point( data[i][1], data[i][2][1]),
    >>                          point( data[i+1][1], data[i+1][2][1])
    >>                         ) $ i=1..nops(data)-1
    >>                 ], Color = [Flat, RGB::Blue]
    >>                ]:

    Define a circle with center (1.5 ,1.5) and radius 1/2: 
    >> circle_object:= [Mode = Curve, [1.5+0.5*cos(s),1.5+0.5*sin(s)],
    >>                  s=[0,2*PI], Grid = [100], 
    >>                  Color = [Flat, RGB::Green] ]:

    Plot the 3 objects in a single plot:  
    >> DIGITS:=5:
    >> plotlib::fieldplot(Axes=Box, Labeling = TRUE, Scaling = Constrained,
                          vectorfield_object, curve_object, circle_object);

    Do it again, this time with Scaling = UnConstrained:

    >> plotlib::fieldplot(Axes=Box, Labeling = TRUE, Scaling = UnConstrained,
                          vectorfield_object, curve_object, circle_object);


----------------------------------------------------------------------*/

plotlib::fieldplot := proc()
    local i, j, list, calc_pointlist, convf;
begin

    calc_pointlist := proc(list)
       local f_x, f_y, points, i, grad_x, grad_y,
             act_x, xmin, xmax, xgrid, delta_x,
             act_y, ymin, ymax, ygrid, delta_y,
    	        delta_xy, ratio, rat_x, rat_y, result;
    begin
        /*
        | Check the arguments of list, left out
        */
        
        xmin    := float(op(list, [2,2,1])):
        xmax    := float(op(list, [2,2,2])):
        ymin    := float(op(list, [3,2,1])):
        ymax    := float(op(list, [3,2,2])):
        xgrid   := op(list, [4,2,1]):
        ygrid   := op(list, [4,2,2]):
        delta_x := (xmax - xmin)/(xgrid-1):
        delta_y := (ymax - ymin)/(ygrid-1):
        delta_xy:= min(delta_x,delta_y); // Walter, 18.5.99

        points  := []:
    
        f_x := convf(op(list, [1,1]), op(list, [2,1]), op(list, [3,1]));
        f_y := convf(op(list, [1,2]), op(list, [2,1]), op(list, [3,1]));

        for act_x from xmin to xmax step delta_x do
            for act_y from ymin to ymax step delta_y do
	        /*
	        | Evaluate the expression expr_xy at the point
	        | (act_x, act_y).
	        */
             grad_x := float(f_x(act_x, act_y));
             grad_y := float(f_y(act_x, act_y));

             ratio  := 1.0/(2.5*(grad_x^2+grad_y^2)^(1/2)):
             rat_x  := delta_xy*ratio*grad_x:
             rat_y  := delta_xy*ratio*grad_y: // Walter, 18.5.99
                                                      
             points := append(points, plotlib::vector(
                                        point(act_x-rat_x, act_y-rat_y),
                                        point(act_x+rat_x, act_y+rat_y),
                                        PI/7, 0.5
                                      )
                              );
	    end_for:
        end_for:
        result := [Mode = List, points, (list[i] $ i = 5..nops(list))]:
        return(result);
    end_proc:

    convf := proc(f, x, y)
        option escape;
    begin
        if not contains({DOM_PROC, DOM_EXEC, DOM_FUNC_ENV}, domtype(f)) then 
           if contains({DOM_EXPR,DOM_IDENT,DOM_INT,DOM_RAT,DOM_FLOAT}, domtype(f)) then
              return( proc(dx,dy) name plotlib::fieldplot::convf;
                  begin subs(f,x=dx,y=dy) end 
              )
           else
              error("Illegal argument: ".expr2text(f))
           end_if 
        end_if
    end_proc:
    
    if args(0) = 0 then 
        plot::_plot(Dimension = 3)
    else 
        for i from 1 to args(0) do 
            list[i]:=args(i); 
            /*
            | Check whether args(i) is an object or a scene-option.
            */
            if testtype(list[i], DOM_LIST) 
               and op(list[i][1],1)<>Mode  // allow other objects such as
               then                        // [Mode = Curve, ...] to be processed
                                           // as usual by plot2d
	        list[i] := calc_pointlist(list[i]):
            end_if; 
        end_for; 
	plot2d(list[j] $ j = 1..nops(list))
    end_if
end_proc:


// ----- end of plotlib::fieldplot -------- 
