/* original: B. Fuchssteiner 
  modified: W. Oevel 12/10/98 --------------------------------------------------
  Help: file:    plotlib::inequalityplot
 
  inequality - plot inequalities

  Call:

  plotlib::inequalityplot(ineq, left..right, bottom..top 
                          <, n>
                          <, Colors = [col1, col2, col3]>,
                          <, SceneOptions = [option_sequence]> );

  Parameters

  ineq             -- a list [f1,f2,..] of real functions of two arguments
  left, right, bottom, top 
                   -- numerical real values
  n                -- optional non-negative integer
  col1, col2, col3 -- color names (RGB values)
  option_sequence  -- a sequence of equations representing the
                      scene options of a plot2d command

  Synopsis

  inequalityplot serves for displaying points (x,y) in the rectangle 
  Q = [left,right] x [bottom,top] satisfying the inequalities

            f1(x,y)>=0 and  f2(x,y)>=0 and  ...           (1)

  specified by ineq = [f1,f2,..].

  The rectangle Q is divided into 2^n times 2^n subrectangles.
  The default value for n is 6.

  A subrectangle is displayed with the color col1, if all its points
  (x,y) satisfy f(x,y)>0 for all functions f in ineq. Consequently, 
  it is guaranteed that points of this color satisfy (1).

  A subrectangles is displayed with the color col3, if there is a function 
  f in ineq such that all points in the subrectangle satisfy f(x,y)<0.
  Consequently, it is guaranteed that points of this color do not satisfy (1).

  The remaining subrectangles are displayed with the color col2. They cover 
  the boundary of the region defined by (1).

  The default values are Colors = [RGB::Green, RGB::Yellow, RGB::Red].

  The plot is constructed via

   plot2d( option_sequence, [Mode = List , [ ... ] ])

  where [...] is a sequence of polygons. option_sequence may consist of 
  all valid scene options in plot2d. The default values are

  SceneOptions= [Scaling=UnConstrained, Labeling=TRUE, Labels=["",""]]

  Interval arithmetic is used to check the inequalities. The input functions 
  must be suitable for this kind of arithmetic, i.e., the calls 
      f1(Dom::Interval(left..right),Dom::Interval(bottome..top)) 
  etc. must produce valid intervals.

  Examples:

  >> f1:= proc(x,y) begin x^2+y^2-1 end_proc:
  >> plotlib::inequalityplot([f1], -1..1, -1..1);

  >> f2:=proc(x,y) begin (cos(x)-y) end_proc: 
  >> f3:=proc(x,y) begin (cos(x)+y) end_proc:
  >> plotlib::inequalityplot([f2, f3], -PI..PI, -2..2, 5);

  >> scene_options:= [Scaling = Constrained, Labeling = TRUE, Axes = Box,
                      BackGround = RGB::White, ForeGround=RGB:: Black]:
  >> colors:=[RGB::Red, RGB::Black, RGB::White]:
  >> plotlib::inequalityplot([f1, f2, f3], -2..2, -1..1, 5,
                             Colors = colors,
                             SceneOptions = scene_options ):

----------------------------------------------------------------------*/
/*
  the following implementation works for 1.4 as well as for 1.5.
  For efficiency it avoids Dom::Interval as far as possible.
  Dom::Interval is only used through (Dom::Interval)::sign
  Still: it's terribly slow.
  There's no guarantee that for n=0 and
  (Dom::Interval)::sign(g(IX, IY)))=0
  there really are at least one point in the subrectangle !
  W. Oevel
*/

plotlib::_inequalityplotdata:=        // returns plotdata for inequalityplot 
proc(f, left, right, bottom, top, n, PColor, ZColor, NColor)
  local g, v, vv, IX, IY, mx, my;
begin
  IX:= Dom::Interval(left..right): 
  IY:= Dom::Interval(bottom..top):
  vv:=1;
  for g in f do
    //!! For expression g replace g(IX,IY) by subs(g, {x=IX, y=IY}, Unsimplified) 
    if (v:=(Dom::Interval)::sign(g(IX,IY)))={-1} then
        return(polygon(point(left,bottom), point(right,bottom), 
                       point(right,top), Filled=TRUE, Color=NColor),
               polygon(point(left,bottom), point(left,top),
                       point(right,top), Filled=TRUE, Color=NColor)) 
    end_if;
    vv:= min(vv, op(v));
  end_for;
  if vv = 1 then // inequalities are definitely satisfied for all points 
    polygon(point(left,bottom), point(right,bottom), 
            point(right,top), Filled=TRUE, Color=PColor),
    polygon(point(left,bottom), point(left,top),
            point(right,top), Filled=TRUE, Color=PColor);
  else /*vv=0 */ // inequalities may be satisfied for some points 
     if n=0 then
          polygon(point(left,bottom), point(right,bottom),
                  point(right,top),Filled=TRUE, Color=ZColor),
          polygon(point(left,bottom), point(left,top), 
                  point(right,top), Filled=TRUE, Color=ZColor);
     else // have a closer look, i.e., investigate subrectangeles 
          mx:=(left+right)/2; my:=(bottom+top)/2;
          plotlib::_inequalityplotdata(f, left, mx, bottom,my, n - 1,PColor, ZColor, NColor), 
          plotlib::_inequalityplotdata(f, left, mx, my,   top, n - 1,PColor, ZColor, NColor), 
          plotlib::_inequalityplotdata(f, mx,right, bottom,my, n - 1,PColor, ZColor, NColor), 
          plotlib::_inequalityplotdata(f, mx,right, my,   top, n - 1,PColor, ZColor, NColor);
     end_if;
  end_if;
end_proc:

plotlib::inequalityplot:=
proc(f, rx, ry)
  local left, right, bottom, top, g, Arg, n, PColor, ZColor, NColor,
  Plotoptions_;
begin
  if args(0)<3 then error("expecting at least 3 arguments") end_if;
  if args(0)>6 then error("expecting at most 6 arguments") end_if;
  if domtype(f)<>DOM_LIST then error("first argument should be a list"); end_if;
  if type(rx)<>"_range" then error("range should be specified as x0..x1") end_if;
  if type(ry)<>"_range" then error("range should be specified as y0..y1") end_if;
  for g in f do
    if domtype(g)<>DOM_PROC then 
       error("inequalty f(x,y)<=0 to be specified by a function f"); end_if;
  end_for;
  left:= float(op(rx,1)): right:= float(op(rx,2)):
  bottom:= float(op(ry,1)): top:= float(op(ry,2)):
  if map({left,right,bottom,top},domtype) <> {DOM_FLOAT}
     then error("plot range must be specified by numerical values"); end_if;
  if left>=right or bottom>=top then error("invalid range") end_if;

  // default values for optional arguments: 
  n:=6; PColor:=RGB::Green: ZColor:=RGB::Yellow: NColor:=RGB::Red:
  Plotoptions_:= Scaling=UnConstrained, Labeling=TRUE, Labels=["",""];

  // override default values by optional arguments: 
  if args(0)>3 then
     for Arg in [args(4..args(0))] do
         if domtype(Arg)=DOM_INT 
            then n:=Arg;
            elif type(Arg)="_equal" and op(Arg,1)=Colors 
                      and type(op(Arg,2)) = DOM_LIST
                 then PColor:=op(Arg,[2,1]): 
                      ZColor:=op(Arg,[2,2]):
                      NColor:=op(Arg,[2,3]):
                 elif type(Arg)="_equal" and op(Arg,1)=hold(SceneOptions)
                      and type(op(Arg,2)) = DOM_LIST
                      then Plotoptions_:=op(op(Arg,2));
                      else error("invalid option");
         end_if;
     end_for;
  end_if;
  plot2d(Plotoptions_ ,
         [Mode=List, [plotlib::_inequalityplotdata(
               f,left,right,bottom,top, n, PColor, ZColor,NColor)] ]
        )
end_proc:
