// $Date: 2000/05/02 17:02:03 $ $Author: frankp $ $Revision: 1.9 $


/*++
faclib::factor_expr -- factor expression with given factorizer

faclib::factor_expr(p, fac)

p   - polynomial expression over the rationals
fac - factorizer

Returns a list of the form [c, f1, e1,...,fn, en] such that p=c*f1^e1*...
*fn^en and c is the content of p.

fac must be a procedure which takes a rational polynomial q and returns
a list of the form [c, f1, e1,...,fn, en] such that the fi are polynomials
and q = c*f1^e1*...*fn^en.
++*/

faclib::factor_expr:=
proc(p, fac)
  local l, i, j, f, e, c, newp;
begin
  // special cases which may save some time...
  case type(p)
    of DOM_INT do
    of DOM_RAT do
      return([p]);

    of "_power" do
      j:= op(p,2);
      if domtype(j) <> DOM_INT then
        return([1, p, 1])
      end_if;
      l:= faclib::factor_expr(op(p,1), fac);
      return([ l[1]^j, (l[2*i], l[2*i+1]*j) $ i=1..(nops(l) div 2) ]);

    of "_mult" do
      c:= FAIL;
      f:= table();
      // find common extension containing all coefficients
      // of all factors
      newp:= polylib::makerat([op(p)]);
      if type(op(newp,[1,1]))<>DOM_POLY then
        // all indets are algebraic over Q, each el. of op(newp,1) is a unit
        return([p])
      end_if;

    // factors of the numerator
      
      for e in op(newp,1) do
        e:= fac(e);
        c:= if c=FAIL then e[1] else c*e[1] end_if;
        if iszero(c) then return( [c] ) end_if;

        // collect factors in f:
        for j from 2 to nops(e) step 2 do
          if contains(f, e[j]) then
            f[e[j]]:= f[e[j]] + e[j+1]
          else
            f[e[j]]:= e[j+1]
          end_if
        end_for;
      end_for;

      // factors of the denominator
      // denoted by negative multiplicities
      
      for e in op(newp,2) do
        e:= fac(e);
        c:= c/ e[1];
        if iszero(c) then error("division by zero") end_if;

         for j from 2 to nops(e) step 2 do
          if contains(f, e[j]) then
            // this may well happen:
            // the product had not been normalized
            f[e[j]]:= f[e[j]] - e[j+1]
          else
            f[e[j]]:= -e[j+1]
          end_if
        end_for;
      end_for;
        
      // deleting zero entries
      f:=select(f, x-> not iszero(op(x,2)));
      return(subs(map([ c, op(op(f,i)) $ i=1..nops(f) ], expr), op(newp,3)));
  end_case;

  newp:=normal(p);
  if contains({DOM_INT, DOM_RAT, "_power", "_mult"}, type(newp)) then
    // simple case after normalizing, start from top
    return(faclib::factor_expr(newp, fac))
  end_if;

  newp:=polylib::makerat(newp);

  if type(op(newp,1))<>DOM_POLY then
    // all indets are algebraic over Q, p is a unit
    return([p])
  end_if;
    
  // eventually factor using fac
  f:= fac(op(newp,1));
  e:= fac(op(newp,2));
  l:= f . [(op(e,2*i), -op(e,2*i+1)) $ i=1..nops(e) div 2];
  // adjust unit
  l[1]:=l[1]/op(e,1);
  
  // back substitution:
  return( subs(map(l, expr), op(newp,3)) )
end_proc:

