// $Date: 2000/05/21 14:57:57 $ $Author: bij $ $Revision: 1.7 $ 


// polylib::makerat


/* transforms an expression or a list or set of expressions
   into a quotient of polynomials (or simply field elements)
   over a suitable extension of the rationals,


   returns:
   - numerator polynomial or field element, or list/set of these
   - denominator polynomial or field element, or list set of these
   - list of substitutions

*/


polylib::makerat:=
proc(l)
  local
  F:DOM_DOMAIN,
  newl:DOM_LIST,
  subst:DOM_LIST,
  revsubst:DOM_LIST,
  i:DOM_INT,
  d:DOM_INT,
  eq:"_equal",
  RHS, base, expo,
  X: DOM_IDENT,
  inds:DOM_SET,
  f,
  alpha;
  
begin
  if type(l)=DOM_POLY then
    if op(l,3)<> hold(Expr) then
      error("Polynomial must have coefficient ring Expr")
    end_if;
    newl:=polylib::makerat(expr(l));
    // reorder indets such that the original indets are in front
    inds:=op(l,2);
    if type(op(newl,1))=DOM_POLY then
      inds:= inds.[op({op(op(op(newl,1),2))} minus {op(inds)})];
      F:=op(newl, [1,3]);
    else // l is a unit
      F:=domtype(op(newl,1))
    end_if;
    return(poly(op(newl,1), inds, F), poly(op(newl, 2), inds, F), op(newl, 3))
  end_if;
  F:=Dom::Rational;
  rationalize(l, FAIL, FAIL, {}, Recursive);
  [newl, subst]:=[op(%)];
  inds:=indets(newl, RatExpr);
   // the indets of the poly to be generated 
   /* initially, these are all identifiers (both
     the previously existing ones and those generated by rationalize) */

  revsubst:=map(subst, s-> op(s,2)=op(s,1));

  
  for i from nops(subst) downto 1 do
    // for every substitution, try to regard the right hand side
    // as algebraic over F
    // if this succeeds, extend F; the left hand side of the
    // substitution then represents an algebraic number and
    // can thus be removed from the set of indeterminates
    
    // start from behind, because there are the simplest right hand sides
    // e.g. for nested roots, the inner root comes after the outer
    eq:= op(subst, i);
    X:= op(eq,1); // the newly created identifier
    case type((RHS:=op(eq,2)))
      of DOM_COMPLEX do
        // RHS must be I
        F:=Dom::AlgebraicExtension(F, poly(X^2+1, [X], F));
        inds:=inds minus {X};
        break
      of "_power" do
        /* The radicand on the right hand side must be an expression that can
           be turned into a rational expression by applying reverse
           substitutions op(revsubst, i+1..nops(subst)) */
        base:= op(RHS,1);
        expo:=op(RHS, 2);
        if type(expo)<>DOM_RAT then
          // only for rational exponents, the extension is algebraic
          break
        end_if;
        base:=subs(base, op(revsubst, i+1..nops(revsubst)));
        base:=F::convert(base);
        if base=FAIL then
          // do not handle function field case, i.e. roots
          // involving indeterminates
          break
        else
          d:=denom(expo);
          // to adjoin: a d-th root of base
          factor(poly([[F::one, d], [-base,0]],[X], F));
          // sort the factors
          f:=(sort([op(%, 2*i) $i=1..nops(%) div 2],
                (u,v)-> degree(u)<=degree(v)));
          // test whether the smallest factor is linear
          select(f, u-> degree(u)=1);
          if %<>[] then
            // some d-th root is already contained in F
            f:=%;
            if nops(f)=1 then
              alpha:=-coeff(op(f,1), 0)
            else
              // there are several d-th roots
              map(f, coeff, 0);
              map(%, _negate@expr);
              // apply substitutions to get expressions
              // find out which choice of alpha agrees
              // with the choices for roots of previous minimal
              // polynomials
              // since denesting radicals is difficult, just use float
              alpha:=sort(%, (u,v)-> abs(float(subs(u, subst)^expo -RHS)) <=
                                     abs(float(subs(v, subst)^expo -RHS)))[1]
            end_if;
           // replace current substitution
          // alpha satisfies alpha^d = base
          // i.e. (with n=numer(expo))
          // alpha^n = base^(n/d) = the original right hand side  
            subst[i]:=(X=expr(alpha)^numer(expo));
              // substitute also in newl
            newl:=subs(newl, X=expr(alpha)^numer(expo))
          else
            // to be improved: no selection of factor
            f:=op(f,1);
            // field has to be extended
            F:=Dom::AlgebraicExtension(F, f);
            // no new substitution, but the expression
            // newl must be adjusted
            newl:=subs(newl, X=X^numer(expo))
          end_if; // % = [] ?
        end_if; // base = FAIL ?
        
          // replace current substitution
          // alpha satisfies alpha^d = base
          // i.e. (with n=numer(expo))
          // alpha^n = base^(n/d) = the original right hand side  
          inds:= inds minus {X}; // X is algebraic
          
      // otherwise, the right hand side is considered
      // transcendental: F need not be extended, and
      // the left hand side remains among the indeterminates  
    end_case
  end_for;

    // if no alg. element was found, the computation can go on
    // over Expr instead of the rationals
  if F=Dom::Rational then
    hold(Expr)
  else
    F
  end_if;
  if nops(inds)>0 then
    if type(l)=DOM_LIST or type(l)=DOM_SET then
      map(map(newl,numer), poly, [op(inds)], %),
      map(map(newl,denom), poly, [op(inds)], %),
      subst
    else
      poly(numer(newl), [op(inds)], %),
      poly(denom(newl), [op(inds)], %),
      subst
    end_if
  else
    if type(l)=DOM_LIST or type(l)=DOM_SET then
      map(newl, F::convert), map(newl, F::one), // not really a function call
      subst
    else
      F(newl), F::one, subst
    end_if
  end_if
end_proc:


