// $Date: 2000/10/30 13:52:51 $ $Author: bij $ $Revision: 1.48 $

/*++
after prog::trace(f) where f is a procedure or a function environment, 
all the calls of f will print their input and output.

Examples:

>> prog::trace(has): has(x,y);
enter 'has'                            with args   : x, y
leave 'has'                            with result : FALSE

                                   FALSE

>> prog::untrace(has): has(x,y);

                                   FALSE

>> fib := proc(n) begin if n<2 then n else fib(n-1)+fib(n-2) end_if end_proc:
>> prog::trace(fib): fib(2);
enter 'fib'                            with args   : 2
enter 'fib'                            with args   : 1
leave 'fib'                            with result : 1
enter 'fib'                            with args   : 0
leave 'fib'                            with result : 0
leave 'fib'                            with result : 1

                                     1

>> g:= x->x + 1: prog::trace(g): g(3);
enter 'g'                              with args   : 3
leave 'g'                              with result : 4

                                     4

>> prog::trace(domtype(infinity)):
>> infinity+3;
enter 'stdlib::Infinity::_plus'        with args   : infinity, 3
enter 'stdlib::Infinity::thisDomain'   with args   : infinity
leave 'stdlib::Infinity::thisDomain'   with result : TRUE
enter 'stdlib::Infinity::thisDomain'   with args   : 3
leave 'stdlib::Infinity::thisDomain'   with result : FALSE
enter 'stdlib::Infinity::thisDomain'   with args   : infinity
leave 'stdlib::Infinity::thisDomain'   with result : TRUE
enter 'stdlib::Infinity::thisDomain'   with args   : 3
leave 'stdlib::Infinity::thisDomain'   with result : FALSE
enter 'stdlib::Infinity::removeRe'     with args   : 3
leave 'stdlib::Infinity::removeRe'     with result : 0
leave 'stdlib::Infinity::_plus'        with result : infinity

                                 infinity
++*/

/* This table will contain the original code of all traced functions and is
   used for prog::untrace */
prog::traceTab := table():

// Option Tree
prog::trace_depth:= -1:
// Option Mem
prog::BYTES:= table():

alias(AllOptions = {hold(Backup), hold(Force), hold(None), hold(Mem),
                    hold(Plain), hold(NoArgs), hold(Depth)}):
// param options    hold(Depth) = d

prog::trace:=
proc(_f_ = (()->()))
  option hold, escape;
  local ff, p, doPrint, trace_internal, trace_proc, funcTab,
        bytes_used_in, bytes_used_out, OX, OPT, SOPT, SFORMAT;
begin
  if args(0) = 0 then
    warning("no object to trace");
    return()
  end_if;

  /* check if function is already traced */
  if contains(prog::traceTab, _f_) then
    if args(args(0)) = hold(Force)
       or (domtype(args(args(0))) = DOM_SET and contains(args(args(0)), hold(Force))) then
      warning("backup of object '".expr2text(_f_)."' will be replaced");
      //
    elif args(args(0)) = hold(Backup)
       or (domtype(args(args(0))) = DOM_SET and contains(args(args(0)), hold(Backup))) then
      warning("backup of object '".expr2text(_f_)."' will be traced");
      sysevalassign(_f_, prog::traceTab[_f_], 2)
    else
      warning("object '".expr2text(_f_)."' is already traced");
      return()
    end_if
  end_if;

  SFORMAT:=
    proc(STR)
    begin
      substring(STR."                               ", 0..31)
    end_proc;

  // ---- internal routines ---- //
  doPrint := 
  /* print all procedures in the argument list in short Form */
  proc(aaa)
    name doPrint;
    option noDebug;
  begin
    map(aaa, x->(if domtype(x) = DOM_PROC then DOM_PROC::print(x) else expr2text(x) end_if));
    zip(%, [", "$nops(%)-1], _exprseq, "");
    _concat(op(%))
  end_proc;
  
  trace_internal :=    /* trace for kernel functions */
  proc(p,n)
    option escape, noDebug;
  begin
    subsop(proc()
             local  lastResult, _x_1_, _arguments_, _name_;
             option noDebug;
           begin
             _x_1_ := p;
             _name_ := SFORMAT("'".n."'");
             _arguments_ := args();
             fprint(Unquoted, 0, "enter "._name_." with args   : ".doPrint([_arguments_]));
             hold(_x_1_)(_arguments_);
             lastResult := eval(%);
             fprint(Unquoted, 0, "leave "._name_." with result : ".doPrint([lastResult]) );
             lastResult;
           end_proc, 6 = _f_)
  end_proc;
	       
  bytes_used_in:=
    proc(n)
      name bytes_used;
    begin
      if not(contains(prog::BYTES, n)) then sysassign(prog::BYTES[n], []) end_if;
      sysassign(prog::BYTES[n], append(prog::BYTES[n], op(bytes(), 1)));
      null()
    end_proc;

  bytes_used_out:=
    proc(n)
      local used;
    begin
      used:= op(prog::BYTES[n], nops(prog::BYTES[n]));
      sysassign(prog::BYTES[n], subsop(prog::BYTES[n], (nops(prog::BYTES[n])) = null()));
      if prog::BYTES[n] = [] then sysdelete(prog::BYTES[n]) end_if;
      used:= trunc((op(bytes(), 1) - used)/1000.0);
      expr2text(used)
    end_proc;

  trace_proc :=      /* trace for procedures */
  proc(p, n)
    option escape;
    name trace_proc;
    local trace_exit;
  begin
//     /* check if function is already traced */
//     if contains(prog::traceTab, p) then
//       return(p)
//     end_if;
    
    if contains(SOPT, hold(Depth)) then
      trace_exit :=    /* print result on exit ; also used for return */
        subsop(proc(n)
                 option escape;
                 name trace_exit;
               begin
                 proc()
                 begin
                   if prog::trace_depth < NIL then
                     fprint(Unquoted, 0, NIL, "leave ".SFORMAT("'".n."'"),
                            " with result ", "[".bytes_used_out(n)." kB]", ": ".doPrint([args()]))
                   end_if;
                   NIL;
                   args():
                 end_proc:
               end_proc,
               [4,4,1,2,3] = _if(not contains(OPT, hold(Plain)),
                               hold(_concat("  ") $ prog::trace_depth), ""),
               [4,4,1,1,2] = SOPT[hold(Depth)],
               [4,4,2] = _if(not contains(OPT, hold(Plain)) or contains(SOPT, hold(Depth)),
                             hold(sysassign(prog::trace_depth, prog::trace_depth - 1)), null()),
               _if(contains(OPT, hold(Mem)), null(), [4,4,1,2,6] = ""),
               _if(contains(OPT, hold(NoArgs)), ([4,4,1,2,5] = "", [4,4,1,2,7] = ""), null())
               )
    else
      trace_exit :=    /* print result on exit ; also used for return */
        subsop(proc(n)
                 option escape;
                 name trace_exit;
               begin
                 proc()
                 begin
                   fprint(Unquoted, 0, NIL, "leave ".SFORMAT("'".n."'"),
                          " with result ", "[".bytes_used_out(n)." kB]", ": ".doPrint([args()]));
                   NIL;
                   args():
                 end_proc:
               end_proc,
               [4,4,1,3] = _if(not contains(OPT, hold(Plain)),
                               hold(_concat("  ") $ prog::trace_depth), ""),
               [4,4,2] = _if(not contains(OPT, hold(Plain)),
                             hold(sysassign(prog::trace_depth, prog::trace_depth - 1)), null()),
               _if(contains(OPT, hold(Mem)), null(), [4,4,1,6] = ""),
               _if(contains(OPT, hold(NoArgs)), ([4,4,1,5] = "", [4,4,1,7] = ""), null())
               )
    end_if;

    /* create new Proc with trace-token, enter-info, old code, exit-info */
    if contains(SOPT, hold(Depth)) then
      subsop(p, 4 =
	     hold(_stmtseq)(
                            hold(id)(n),
                            NIL,
                            hold(_if)(TRUE,
                                      hold(fprint)(Unquoted, 0,
                                                   hold(_concat)
                                                   (NIL,
                                                    "enter ",
                                                    SFORMAT("'".n."'"),
                                                    " with args   : ",
                                                    hold(NIL([args()])))),
                                      NIL),
                            NIL,
                            trace_exit(n),
                            hold(%(%2))),
             _if(contains(OPT, hold(Mem)), [4,1,0] = bytes_used_in, null()),
             _if(not contains(OPT, hold(Plain)) or contains(SOPT, hold(Depth)),
                 [4, 2] = hold(sysassign(prog::trace_depth, prog::trace_depth + 1)), null()),
             [4,3,2,3,1] = _if(not contains(OPT, hold(Plain)),
                             hold(_concat)(hold(_concat("  " $ prog::trace_depth))), ""),
             [4,3,2,3,5,0] = doPrint,
             [4,3,1] = hold(_less)(hold(slot)(prog, "trace_depth"), SOPT[hold(Depth)]),
	     [4,4] = hold(_exprseq)(subs(op(p,4),hold(return)=return@trace_exit(n), Unsimplified)),
             _if(contains(OPT, hold(NoArgs)), ([4,3,2,3,4] = "", [4,3,2,3,5] = ""), null()),
             //[4,3,1,2] = SOPT[hold(Depth)],
             //6 = _f_,
	     Unsimplified)
    else
      subsop(p, 4 =
	     hold(_stmtseq)(
                            hold(id)(n),
                            NIL,
                            hold(fprint)(Unquoted, 0,
                                         hold(_concat)
                                         (NIL,
                                          "enter ",
                                          SFORMAT("'".n."'"),
                                          " with args   : ",
                                          hold(NIL([args()])))),
                            NIL,
                            trace_exit(n),
                            hold(%(%2))),
             _if(contains(OPT, hold(Mem)), [4,1,0] = bytes_used_in, null()),
             _if(not contains(OPT, hold(Plain)),
                 [4, 2] = hold(sysassign(prog::trace_depth, prog::trace_depth + 1)), null()),
             [4,3,3,1] = _if(not contains(OPT, hold(Plain)),
                             hold(_concat)(hold(_concat("  " $ prog::trace_depth))), ""),
             [4,3,3,5,0] = doPrint,
	     [4,4] = hold(_exprseq)(subs(op(p,4),hold(return)=return@trace_exit(n), Unsimplified)),
             _if(contains(OPT, hold(NoArgs)), ([4,3,3,4] = "", [4,3,3,5] = ""), null()),
             //6 = _f_,
	     Unsimplified)
    end_if
  end_proc;
// --------------------------- 

  // functions that cannot be traced
  if contains({hold(fprint), hold(hold)}, _f_) then
    error("this function cannot be traced")
  elif contains({hold(_concat), hold(length), hold(Pref), hold(strmatch),
                 hold(anames), hold(type), hold(substring), hold(domtype),
                 hold(contains), hold(sysassign)}, _f_) then
    warning("this function should not be traced, unexpected sideeffects breaks out")
  elif Pref::typeCheck() <> hold(None) and contains({hold(testtype)}, _f_) then
    error("this function cannot be traced, when 'Pref::typeCheck' is not 'None'")
  end_if;

  ff:=context(_f_);
  prog::init(ff);

  OPT:= {}; OX:= 0;
  if domtype(args(args(0))) <> DOM_SET then
    if contains(AllOptions, args(args(0))) then
      OPT:= {args(args(0))};
      OX:= 1
    end_if
  elif map(args(args(0)), X -> _if(type(X) = "_equal", op(X, 1), X)) minus AllOptions = {} then
    OPT:= args(args(0));
    OX:= 1
  else
    OPT:= map(args(args(0)), X -> _if(type(X) = "_equal", op(X, 1), X)) minus AllOptions;
    error("unknown option"._if(nops(OPT) > 1, "s ", " ").expr2text(OPT))
  end_if;
  SOPT:= select(OPT, X -> type(X) = "_equal");
  OPT:= OPT minus SOPT;
  SOPT:= table(op(SOPT));
  
  case domtype(ff)
    /*  ff is a internal function or procedure with slot's */
    of DOM_FUNC_ENV do
      sysassign(prog::traceTab[_f_], ff):
      case domtype(op(ff,1))
	of DOM_PROC do
	  stdlib::syseval(evalassign(_f_,
				     subsop(ff, 1=trace_proc(op(ff, 1),
							     doPrint([_f_]))), 1));
	  break;
	  
	of DOM_EXEC do
	  stdlib::syseval(evalassign(_f_,
				     subsop(ff, 1=trace_internal(op(ff, 1),
								 doPrint([_f_]))), 1));
	  break;
	   
      end_case;

      ff:=context(_f_);  /* get changes from last step in ff */
      if domtype(op(ff,3)) = DOM_TABLE then
	 /* trace all slot's of the DOM_FUNC_ENV */
	funcTab := op(ff,3);
	map(funcTab, context);
	ff:=context(_f_);  /* get changes from last step in ff */
	funcTab := op(ff,3);
	map([op(select(funcTab, x->bool(domtype(op(x,2))=DOM_PROC)))],
            proc(x)
              local Pr;
            begin
              Pr:= op(x,2);
              funcTab[op(x,1)] := trace_proc(Pr, expr2text(op(Pr,6)))
            end_proc);
	stdlib::syseval(evalassign(_f_,subsop(ff,3=funcTab), 1));
      end_if;
      break;
       
    of DOM_PROC do
      if type(_f_) = "slot"
         and domtype(context(op(_f_, 1))) = DOM_DOMAIN then
        sysassign(prog::traceTab[_f_], ff);
        stdlib::syseval((evalassign(_f_,trace_proc(ff, expr2text(_f_)), 1)))
      else
        sysassign(prog::traceTab[_f_], ff):
        stdlib::syseval((evalassign(_f_,trace_proc(ff, expr2text(_f_)), 1)))
      end_if;
      break;

    of DOM_DOMAIN do
      if args(0) >= 2 + OX then
        if domtype(args(2)) = DOM_SET or domtype(args(2)) = DOM_LIST then
          for p in args(2) do
            if domtype(p) <> DOM_STRING then
              p:= expr2text(p)
            end_if;
            if p <> "print"
               and domtype(slot(ff,p)) = DOM_PROC then
              if contains(prog::traceTab, hold(slot)(_f_, p))
                 and not contains(OPT, hold(Force)) then
                warning("method '".expr2text(hold(slot)(ff, p))."' is already traced")
              else
                sysassign(prog::traceTab[hold(slot)(_f_,p)], slot(ff, p));
                sysassign(slot(ff,p),
                          trace_proc(slot(ff,p), expr2text(hold(slot)(ff,p))))
              end_if
            else
              warning("no object to trace [".expr2text(ff)."::".p."]")
            end_if
          end_for
        else
          error("wrong type of second arg ('DOM_SET' or 'DOM_LIST' expected)")
        end_if
      else // trace all Methods
        for p in ff do
          sysassign(prog::traceTab[_f_], "traced domain");
          // not traced methods
          if contains(ff, "noTrace") then
            OX:= map(slot(ff, "noTrace"), expr2text)
          else
            OX:= {}
          end_if;
	  // do not trace the method "print" !
	  if op(p,1) <> "print" and
             not contains(OX, op(p,1)) and
	     domtype(op(p,2)) = DOM_PROC then
            // traced?
            if contains(prog::traceTab, hold(slot)(_f_,op(p,1)))
               and not contains(OPT, hold(Force)) then
              warning("method '".expr2text(hold(slot)(ff,op(p,1)))."' is already traced")
            else
	      sysassign(prog::traceTab[hold(slot)(_f_,op(p,1))], op(p,2));
	      sysassign(slot(ff,op(p,1)),
		        trace_proc(op(p,2), expr2text(hold(slot)(ff,op(p,1)))))
            end_if
	  end_if
        end_for
      end_if;
      break
    of DOM_SET do
      map(_f_, prog::trace, OPT); break
    otherwise
      warning("no object to trace (".expr2text(_f_).")")
  end_case;

  null()
end_proc:

prog::traceall:=
  proc()
    local GF, DF;
  begin
    // DOM_PROC /////////////////////////////////////////////////////////////
    userinfo(Text, 1, "pre-init all DOM_PROC");
    map(sort([op(select(stdlib::anames(3), (X)->(type(context(X)) = DOM_PROC)))],
	     (X, Y)->(expr2text(X)<expr2text(Y))),
	prog::init);
    // DOM_FUNC_ENV /////////////////////////////////////////////////////////
    userinfo(Text, 1, "pre-init all DOM_FUNC_ENV");
    map(sort([op(select(stdlib::anames(3), (X)->(type(context(X)) = DOM_FUNC_ENV)))],
	     (X, Y)->(expr2text(X)<expr2text(Y))),
	prog::init);
    // DOM_DOMAIN ///////////////////////////////////////////////////////////
    userinfo(Text, 1, "pre-init most DOM_DOMAIN");
    GF:= sort([op(select(stdlib::anames(3), (X)->(type(context(X)) = DOM_DOMAIN)))],
	      (X, Y)->(expr2text(X)<expr2text(Y)));
    for DF in GF do
      DF:= context(DF);
      if domtype(DF) = DOM_DOMAIN then
	//if contains(context(DF), "constructor") then
	  prog::init(DF);
	  sysdelete(prog::init(DF))
	//end_if
      end_if
    end_for;
    null()
  end_proc:

// 
prog::untrace:=
  proc(_f_) local p, tabIndex;
    option hold;
  begin
    if args(0) = 0 then
      //* untrace all traced functions */
      for p in prog::traceTab do
        if prog::traceTab[op(p, 1)] = "traced domain" then
          // 
        else
          sysevalassign(op(p, 1), op(p, 2), 2);
        end_if
      end_for;
      sysassign(prog::traceTab, table());
      return();
    end_if;
  
    if type(context(_f_)) = DOM_DOMAIN then
      //* untrace all procedures in this domain */
      if contains(prog::traceTab, _f_) then
        sysdelete(prog::traceTab[_f_])
      end_if;
      for p in prog::traceTab do
        if type(op(p,1)) = "slot" and op(p,[1,1]) = _f_ then
	  sysevalassign(op(p,1),op(p,2), 2);
	  sysdelete(prog::traceTab[op(p,1)]);
        end_if;
      end_for;
    elif type(_f_) = "slot" then
      if type(_f_) = "slot"
         and domtype(context(op(_f_, 1))) = DOM_DOMAIN then
        sysdelete(prog::traceTab[op(_f_, 1)])
      end_if;
      tabIndex := hold(slot)(op(_f_,1),op(_f_,2));
      if not contains(prog::traceTab, tabIndex) then
        error("function was not traced")
      end_if;
   
      stdlib::syseval(evalassign(tabIndex, prog::traceTab[tabIndex], 2));
      sysdelete(prog::traceTab[tabIndex]);
    else
      if not contains(prog::traceTab, _f_) then
        error("function was not traced")
      end_if;
      stdlib::syseval(evalassign(_f_, prog::traceTab[_f_], 2));
      sysdelete(prog::traceTab[_f_]);
    end_if;
		   
    null()
  end_proc:

prog::traced:=
  proc(X)
    option hold;
    local k, _X;
  begin
    if args(0) = 0 then
      if nops(prog::traceTab) > 0 then
        output::tableForm(sort(map([op(prog::traceTab)], op, 1)), ", ")
      else
        fprint(Unquoted, 0, "No objects traced.")
      end_if
    else
      if contains(prog::traceTab, X) then
        fprint(Unquoted, 0, "A backup of object '", expr2text(X), "' exists.")
      end_if;
      _X:= context(X);
      case domtype(_X)
        of DOM_FUNC_ENV do
          if type(op(_X, [1, 4])) = "_stmtseq"
             and op(_X, [1, 4, 1]) = hold(id)(expr2text(X)) then
            fprint(Unquoted, 0, "Object '", expr2text(X), "' seems to be traced.")
          end_if;
          break
        of DOM_PROC do
          if type(op(_X, 4)) = "_stmtseq"
             and op(_X, [4, 1]) = hold(id)(expr2text(X)) then
            fprint(Unquoted, 0, "Object '", expr2text(X), "' seems to be traced.")
          end_if;
          break
        of DOM_DOMAIN do
          if contains(prog::traceTab, X) or contains(prog::traceTab, _X) then
            fprint(Unquoted, 0, "Domain '", expr2text(X), "' seems to be traced.")
          else
            for k in map(op(_X), op, 1) do
              if contains(prog::traceTab, hold(slot)(X, k)) then
                fprint(Unquoted, 0, "A backup of object '", expr2text(hold(slot)(X, k)), "' exists.")
              end_if;
              if domtype(slot(_X, k)) = DOM_PROC then
                if type(op(slot(_X, k), 4)) = "_stmtseq"
                   and op(slot(_X, k), [4, 1]) = hold(id)(expr2text(hold(slot)(X, k))) then
                  fprint(Unquoted, 0, "Object '", expr2text(hold(slot)(X, k)), "' seems to be traced.")
                end_if
              end_if
            end_for
          end_if;
          break
        otherwise
          fprint(Unquoted, 0, "Nothing known about object '", expr2text(X), "'.")
      end_case
    end_if;
    null()
  end_proc:
