/*
 * series slot of abs - computes series(abs(f), x, n, dir)
 *
 * If dir <> Left, Right, Real ==> simply returns sign(1/f)*series(f).
 * Otherwise expands f, writes it as f0 * x^d * (1 + f1), simplifies
 * sign(1/f0) * sign(x)^(-d) by taking into account the direction,
 * and returns sign(1/f0) * sign(x)^(-d) * f0 * x^d * series(abs(1 + f1)).
 *
 * The code is completely analogous to sign::series.
 *
 * Author: Juergen Gerhard, 18.5.00
 */

abs := slot(abs, "series", proc(f, x, n, dir)
  local t, d, f0, f1, X, s, tr, ti;
begin 

  // handle special case abs(x) separately for efficiency reasons
  if f = x then
    case dir
    of Right do
      return(Series::Puiseux::new(x, x, n));
    of Left do
      return(Series::Puiseux::new(-x, x, n));
    otherwise
      return(Series::Puiseux::scalmult(Series::Puiseux::new(x, x, n),
                                       sign(1/x), 0));
    end_case
  end_if;

  if dir <> FAIL then // directional expansion

    // replace x by a new variable X and assign the properties implied
    // by the direction to it
    X := genident();
    case dir
    of Left do
      assume(X < 0);
      break
    of Right do
      assume(X > 0);
      break
    of Real do
      assume(X, Type::Real);
      break
    end_case;

    // recursively expand the argument
    t := Series::series(eval(subs(f, x = X)), X, n, dir);

    if domtype(t) = Series::Puiseux and (d := ldegree(t)) <> FAIL then

      // write t as f0 * x^d * (1 + f1), then abs(f) =
      // sign(1/f0) * sign(x)^(-d) * f0 * x^d * abs(1 + f1)
      f0 := lcoeff(t);
      f1 := Series::Puiseux::scalmult(lmonomial(t, Rem)[2], 1/f0, -d);

      s := f0 * sign(1/f0) * sign(X)^(-d);
      s := eval(subs(s, X = x));

      // return s * x^d * sqrt((1 + Re(f1)^2) + Im(f1)^2)

      // some shortcuts
      if normal(eval(subs(f, x = X)) - f0*X^d) = 0 then // f = f0*x^d
        unassume(X);
        return(Series::Puiseux::scalmult(Series::Puiseux::one(n, x), s, d));
      elif ldegree(f1) = FAIL then // f = f0 * (1 + O(..))
        unassume(X);
        return(Series::Puiseux::scalmult(1 + subs(f1, X = x), s, d));
      end_if;
    
      // split f1 into real and imaginary part tr + ti
      tr := subs(map(f1, Re), X = x);
      ti := subs(f1 - tr, X = x);
      f1 := subs(f1, X = x);
      unassume(X);

      if ldegree(ti) = FAIL then // ti = O(..)
        return(Series::Puiseux::scalmult(1 + f1 + ti, s, d))
      else
        return(Series::Puiseux::scalmult(((1 + tr)^2 - ti^2)^(1/2), s, d))
      end_if

    end_if;

    unassume(X);

  end_if;

  // default: move sign(1/f) into the 0th coefficient
  return( Series::Puiseux::scalmult(Series::series(f, x, n, dir),
                                    sign(1/f), 0) );

end_proc):
