// $Date: 2000/11/23 09:31:11 $ $Author: frankp $ $Revision: 1.42 $ 
// frankp, 1997

/*++
        gseries  --  a domain for generalized series expansions 

        Call:
        gseries(ex, x [=a] [,ord] [,direction])

        Parameters:
        ex       : expression in x
        x        : identifier
        a        : (optional) real arithmetical expression (default: 0)
        ord      : (optional) positive integer
        direction: (optional) option 'Left', or 'Right' (default: Right)

        Synopsis:
        Compute the asymptotic series of ex of order ord with respect to x 
        at the point a. If a is missing then x is assumed to goes to 0+.  
        If ord is missing then the value of ORDER is used (6 by default).
        If an expansion could not be computed, then FAIL is returned.

        A generalized series expansion is of the domain 'Series::gseries', 
        and has 3 operands:

        1. A list containing lists of the factors for each product in ex;
        2. the bigO term;
        3. an expression e depending of the series variable x
           such that ex tends to infinity for x = a+.

        Example: 3*exp(1/x) + y*ln(x) + O(x^4) in x = infinity
                    -> [[3,exp(1/x)],[y,ln(x)]], x^4, x


        The algorithm based on the dissertation of Dominik Gruntz:
        "On Computing Limits in a Symbolic Manipulation System", 
        submitted to the SWISS FEDERAL INSTITUTE OF TECHNOLOGY ZUERICH
        in 1995.

        This domain is used by 'series' and 'asympt'.
++*/

Series::gseries:= newDomain("Series::gseries"):
Series::gseries::info:="domain for generalized series expansions":
Series::gseries::interface:= {}:

alias( gseries=Series::gseries ):

/*--
    new  --  the main procedure for computing the expansion
--*/
gseries::new := proc(ex, x, ord=ORDER)
    local a, tmp, asy, XX, X, LIMITx, direction, siszero;
begin
    LIMITx:= genident("LIMITx");
    assume( LIMITx > 0 );

    if args(0) < 2 or args(0) > 4 then 
        error("wrong no of args")
    elif type(x) = "_equal" then
        a:= op(x,2);
        x := op(x, 1); 
        if domtype(x) <> DOM_IDENT then
            error("invalid series variable")
        elif not testtype(a,Type::Constant) and not testtype(a,Type::Arithmetical) then
            error("expansion point: not an arithmetical expression")
        elif not testtype(a,Type::Numeric) and a <> infinity and a <> -infinity then
            case is(a,Type::Real) 
            of TRUE    do break
            of FALSE   do error("expansion point must be a real arithmetical expression")
            of UNKNOWN do error("expansion point cannot be checked on being real")
            end_case
        end_if
    elif domtype(x) <> DOM_IDENT then 
        error("invalid series variable") 
    else 
        a:= 0 // x -> +0
    end_if;

    case args(0)
    of 2 do
        direction:= hold(Right);
        break
    of 3 do
        if not testtype(ord,Type::PosInt) then
            direction:= ord;
            ord:= ORDER;
            if not contains({hold(Right),hold(Left)},direction) then
                error("unknown or invalid direction")
            elif a = infinity and direction = hold(Right) or
                 a = -infinity and direction = hold(Left)
            then
                 error("inconsistent direction")
            end_if
        else
            direction:= hold(Right);
        end_if;
        break
    of 4 do
        if not testtype(ord,Type::PosInt) then
            error("3rd argument, the order of the series expansion, must be a positive integer")
        else
            direction:= args(4);
            if not contains({hold(Right),hold(Left)},direction) then
                error("unknown or invalid direction")
            elif a = infinity and direction = hold(Right) or
                 a = -infinity and direction = hold(Left)
            then
                error("inconsistent direction")
            end_if
        end_if
    end_case;

    // variable transformation: Find X(x) such that X(x) -> infinity for x -> a+
    if a = infinity then 
        XX:= LIMITx; X:= x
    elif a = -infinity then 
        XX:= -LIMITx; X:= -x
    elif direction = hold(Left) then
        XX:= a-1/LIMITx; X:= -1/(x-a)
    else
        XX:= a+1/LIMITx; X:= 1/(x-a)
    end_if;

    if not has(ex,x) then
        // return the exact expansion ex:
        return(new(gseries,[[1,ex]],0,1/X))
    end_if;

    // LIMITx -> infinity:
    asy:= eval(subs(ex,x=XX));

    // keep the current 'iszero' function of Series::Puiseux: 
    siszero:= Series::Puiseux::iszero;
    // suspend warning messages printed by the MRV algorithm:
    tmp:= stdlib::limit::printWarnings(FALSE);

    if traperror( (asy:= gseries::asympt( (stdlib::limit)::simp2explog(asy,LIMITx), LIMITx , ord )) ) <> 0 or asy = FAIL then
        sysassign( Series::Puiseux::iszero, siszero ):
        stdlib::limit::printWarnings(tmp);
        return( FAIL )
    else
        sysassign( Series::Puiseux::iszero, siszero ):
        stdlib::limit::printWarnings(tmp);

        asy := map(asy,subs,LIMITx=X);
        return( new(gseries,op(asy,1),op(asy,2),X) )
    end_if
end_proc:

/*-----------------------------------------------------------------*/
/*----------------------- INTERNAL METHODS ------------------------*/
/*-----------------------------------------------------------------*/

/*--
    gseries::asympt(ex,x,n)

    ex : expression in x
    x  : identifier
    n  : order of series computation

    Compute the asymptotic series of ex, assuming that x goes to
    infinity. This function returns FAIL if an expansion can not
    be computed, and the list [aex,ord] otherwise where aex is
    an expansion of ex in x and ord the corresponding big-O term.
--*/
gseries::asympt := proc(ex,x,ord)
    option escape;
    local S, s, i, t, lc, Ex, omega;
begin
    Ex := ex;  // save expression to avoid infinite loops

    userinfo(1,"(1)  Find mrv set of ",ex);

    if args(0) = 3 then
        S := stdlib::limit::getMRV( ex,x );
        if contains( {UNKNOWN,FAIL,"tryIV"},S ) then
            userinfo(1,"** getMRV fails");
            return( FAIL )
        elif domtype(S) <> DOM_SET then
            return( [ [[S,1]],0] )
        elif nops(S) = 0 then
            return( [gseries::mkgseries(ex,x),x^ord] )
        end_if
    else
        S:= args(4);
        t:= select( S, (x,y) -> has(y,x), ex );
        if t = {} then
            S:= stdlib::limit::getMRV( ex,x );
            if contains( {UNKNOWN,FAIL,"tryIV"},S ) then
                userinfo(3,"MRV: ** determining MRV set fails");
                return( FAIL )
            elif domtype(S) <> DOM_SET then
                return( [ [[S,1]],0] )     
            end_if
        end_if
    end_if;


    if S intersect {x} = {x} then // Upward Movement
        t:= gseries::asympt(
          stdlib::limit::upmove(ex,x), x, ord, stdlib::limit::upmove(S,x)
        );
        if t = FAIL then return( FAIL ) else return( stdlib::limit::downmove(t,x) ) end_if
    end_if;

    userinfo(1,"(2)  Rewrite it");

    s := stdlib::limit::rewrite( S,x );
    if s = FAIL then
        return( FAIL )
    end_if;
    omega:= op(s,1);
    S := op(s,2);
    ex := eval(subs(ex,op(S,i)=op(s,[3,i]) $ i=1..nops(S),omega=stdlib::limit::OMEGA)); 


    userinfo(1,"(3)  Compute the Puiseux series in OMEGA=0+ of ",ex);
    // Series computation:
    sysassign( Series::Puiseux::iszero, proc(c) begin
        c:= subs( normal(c), stdlib::limit::OMEGA=omega);
        if iszero(simplify(c,exp)) or iszero(simplify(c,ln)) then
            return( TRUE )
        else
            return( FALSE )
        end_if
    end ):
    s := Series::series(ex,stdlib::limit::OMEGA,ord,hold(Right));

    if s = FAIL then
        userinfo(1,"** series computation fails");
        return( FAIL )
    elif domtype(s) = gseries then
        userinfo(3,"     type is gseries...");
        t := gseries::lterm(s);
        if type(t) = "_power" and op(t,1) = stdlib::limit::OMEGA then
            userinfo(3,"     series has no rational exponents");
            //  can't be handled by now
            return( FAIL )
        else
            userinfo(2,"** that should not happen");
            return( FAIL )
        end_if
    end_if;

    userinfo(2,"Result is ",expr(s));

    s:= extsubsop( s,4=subs(extop(s,4),stdlib::limit::OMEGA=omega) );
    s:= extsubsop( s,5=subs(extop(s,5),stdlib::limit::OMEGA=omega) );
    // old: 
    //    s := eval(subs(s,stdlib::limit::OMEGA=omega)); 
    // Does not work anymore, as subs does not allow to substitute arbitrary
    // expressions (anyway, this should have not work even in 1.4?!)
    if (t:= Series::Puiseux::lmonomial(s)) = Ex then
        // series approximation is an exact one:
        /*
        t:= gseries::convert(s);
        return( [extop(t,1),extop(t,2)] )
        */
        return( [gseries::mkgseries(t,x),0] )
    end_if;

    i:= extop(s,2)/extop(s,1) - 1; // ldegree(s) - 1
    lc:= [];
    repeat
        i:= i + 1;
        t:= Series::Puiseux::coeff(s,i);
        S:= (if t = FAIL or has(t,x) then TRUE else FALSE end );
        if not S then 
            if not Series::Puiseux::iszero(t) then lc:= lc.[[t,omega^i]] end_if
        end_if
    until S end_repeat;

    if t = FAIL then
        userinfo(2,"all coefficients are constant");
        return( [lc,omega^Series::Puiseux::order(s)] )
    end_if;

    Ex:= FALSE;
    repeat
        if not has(t,x) then
            if not Series::Puiseux::iszero(t) then lc:= lc.[[t,omega^i]]; end_if;
            i := i + 1;
            t := Series::Puiseux::coeff(s,i)
        else
            S := gseries::asympt(t,x,ord);
            if S = FAIL then 
                return( FAIL ) 
            // elif nops(S[1]) = 1 and op(S,[1,1,1])*op(S,[1,1,2]) = t then
            elif extop(S,2) = 0 then
                // constant expansion
                lc := lc.gseries::insFactor(S[1],omega^i);
                i := i + 1;
                t := Series::Puiseux::coeff(s,i)
                // continue
            else
                // break
                Ex:= TRUE
            end_if
        end_if
    until t = FAIL or Ex = TRUE end_repeat;

    if t = FAIL then
        return( [lc,omega^i] )
    else
        return( [lc.gseries::insFactor(S[1],omega^i), S[2]*omega^i] )
    end_if
end_proc:

/*--
    gseries::insFactor(L,f)

    L: DOM_LIST
    f: any
--*/
gseries::insFactor:= proc(L,f) 
    local j;
begin 
    return( subsop(L,j=[L[j][1],L[j][2]*f] $ j=1..nops(L)) )
end_proc:

/*--
    gseries::mkgseries(ex,x) 

    Returns a list containing respectively that coefficients of an
    addend of ex that are constant with respect to x and the product
    of all subterms of ex dependend on x.
    If ex is a constant (in respect to x), an error message occurs.
--*/
gseries::mkgseries := proc(ex,x)
    local lc, f, g, xx, cc;
begin
    ex := expand(ex);
    lc := [];
    if type(ex) = "_plus" then
        for f in ex do
            if type(f) = "_mult" then
                cc := 1; xx := 1;
                for g in f do
                    if has(g,x) then xx := xx*g 
                    else cc := cc*g 
                    end_if
                end_for;
                // return( lc.[[cc,xx]] )
                lc:= lc . [[cc,xx]]
            elif has(f,x) then 
                // return( lc.[[1,f]] )
                lc:= lc . [[1,f]]
            else 
                // return( lc.[[f,1]] )
                lc:= lc . [[f,1]] 
            end_if
        end_for;
        return(lc)
    elif has(ex,x) then
        return( [[1,ex]] )
    else 
        error("illegal argument")
    end_if
end_proc:

/*--
    scalmult(a,c,f) -- multiply coefficients by c and terms by f 
--*/
gseries::scalmult := proc(a,c,f) 
    local l,i;
begin
    if args(0)<3 then f:=1 end_if;
    l:=extop(a,1);
    return( extsubsop(a,1=[[l[i][1]*c,l[i][2]*f]$i=1..nops(l)],2=extop(a,2)*f) )
end_proc:

/*--
    compare( a,b,x )  --  get the dominan expression

    a, b : arithmetical expressions
    x    : DOM_IDENT

    This method returns 1, if a >> b, 0 if a = b, and -1 if a << b
    for x -> 0.

    a >> b means that for x -> 0+, b tends to zero faster than a.
    Example: x >> x^2, exp(x) >> x, a >> 0 for any a <> 0.
--*/
gseries::compare := proc(a,b,x)
    local c;
begin
    if a = b then 
        return( 0 )
    elif iszero(b) = TRUE then 
        return( 1 )
    elif iszero(a) = TRUE then 
        return( -1 )
    else
        c:= normal(a/b);
        if c = 1 then
            return( 0 )
        elif c = x then 
            return( -1 )
        else
            c:= limit( c,x=0,hold(Right) );
            if c = FAIL or type(c) = "limit" or domtype(c) = Dom::Interval then
                error("cannot get dominant function")
            elif iszero(c) then 
                return( -1 )
            elif domtype(c) = stdlib::Infinity then
                return( 1 )
            else
                return( 0 )
            end_if
        end_if
    end_if
end_proc:

/*-----------------------------------------------------------------*/
/*------------------------ PUBLIC METHODS -------------------------*/
/*-----------------------------------------------------------------*/

/*--
    gseries::create(l,bigO,x=a)

    l    - list containing lists of the factors for each product in ex.
    bigO - the bigO term, possibly zero.
    x    - identifier
    a    - arithemtical expression, not depending on x.

    Returns an object of the type 'Series::gseries'. This slot
    should be used with caution, because no argument checking is
    performed. It should only be used to create a gseries without
    calling the algorithm implemented by the domain 'gseries'.
--*/
gseries::create:= proc(l,bigO,dirx)
    local x, a;
begin
    if domtype(dirx) <> DOM_IDENT then
        if type(dirx) <> "_equal" then
            return( FAIL )
        end_if;

        x:= op(dirx, 1); 
        a:= op(dirx, 2);
        case a 
        of -infinity do
            x:= -x;
            break
        of infinity do
            break
        otherwise
            x:= 1/(x-a)
        end_case;

        return( new(gseries,l,bigO,x) )
    else
        return( new(gseries,l,bigO,dirx) )
    end_if
end_proc:

/*--
    map  --  maps onto the coefficients
--*/
gseries::map := proc(ex)
    local l, j, param;
begin
    param:= args(2..args(0));

    l := extop(ex,1);
    ( l[j][1] := map(l[j][1],param) ) $ j=1..nops(l);
    return( extsubsop(ex,1 = l) )
end_proc: 

/*--
    print  --  print an asymptotic series
--*/
gseries::print := proc(s) 
    local i, e;
begin 
    e:= extop(s,1);
    if extop(s,2) = 0 then
        if nops(e) = 1 then
            return( _mult(op(e[i])) $ i=1..nops(e) )
        else
            return( hold(_plus)(_mult(op(e[i])) $ i=1..nops(e)) )
        end_if
    else
        return( hold(_plus)(_mult(op(e[i])) $ i=1..nops(e),hold(O)(extop(s,2))) )
    end_if
end_proc:

/*--
    convert  --  convert a Puiseux series to an asymptotic series
--*/
gseries::convert := proc(a)
    local bo,k,l,i,x,ll,p,sgn;
begin
    case domtype(a)
    of gseries do
        return( a )
    of Series::Puiseux do
        bo:= op(a,1);
        k:= op(a,2);
        l:= op(a,4);
        x:= Series::Puiseux::indet(a);
        p:= Series::Puiseux::point(a);
        if domtype(p) = stdlib::Infinity then
            sgn := -1
        else
            sgn := 1
        end_if;
        ll:=[];
        for i from 1 to nops(l) do
            if iszero(l[i]) = FALSE then 
                ll:= append(ll,[l[i],x^(sgn*(k+i-1)/bo)]) 
            end_if
        end_for;
        return(
            new(gseries,ll,x^(sgn*extop(a,3)/bo),
                if domtype(p)=stdlib::Infinity then sign(p)*x else 1/(x-p) end
            )
        )
    otherwise
        return( FAIL )
    end_case
end_proc:

/*--
    convert_to  --  convert a generalized series to another domain
--*/
gseries::convert_to := proc(s,T)
    local e;
begin
    if domtype(T) <> DOM_DOMAIN then T:= domtype(T) end_if;

    case T
    of gseries do
        return( s )
    of DOM_POLY do
        e:= gseries::expr( s );
        return( poly(e,[op(indets(extop(s,3)))]) )
    of DOM_EXPR do
        return( gseries::expr(s) )
    otherwise
        return( FAIL )
    end_case
end_proc:

gseries::lmonomial := proc(a)
    local l;
begin
    if args(0) <> 1 then
        error("expecting one argument")
    end_if;

    l:=extop(a,1);
    if nops(l)=0 then 
        return( FAIL )
    else 
        return( l[1][1]*l[1][2] )
    end_if
end_proc:

/*--
    lterm(s)  --  leading term of s 

    Example: lterm(3*exp(1/x)+y*ln(x)+O(x^4)) -> exp(1/x) 
--*/
gseries::lterm := proc(a)
    local l;
begin
    if args(0) <> 1 then
        error("expecting one argument")
    end_if;

    l:=extop(a,1);
    if nops(l) = 0 then 
        return( FAIL )
    else 
        return( l[1][2] )
    end_if
end_proc:

gseries::ldegree := proc(a)
    local l,e,x;
begin
    if args(0) <> 1 then
        error("expecting one argument")
    end_if;

    x:= genident("x"): // create a local series variable

    l:= extop(a,1);
    while l <> [] do
        e:= l[1];
        if Series::Puiseux::iszero(e[1]) then
            delete l[1];
        elif e[2] = 1/extop(a,3) then
            return( 1 )
        elif type(e[2]) = "_power" and op(e[2],1) = 1/extop(a,3) 
             and not has( op(e[2],2), extop(a,3) )
        then
            return( op(e[2],2) )
        else
            e:= series( eval(subs(e[1]*e[2],extop(a,3)=x)),x=infinity ); 
            case domtype(e)
            of gseries do 
                return( 0 )
            of Series::Puiseux do
                return( Series::Puiseux::ldegree( e ) )
            otherwise
                return( FAIL )
            end_case
        end_if
    end_while;
    return( FAIL )
end_proc:

/*--
    lcoeff(s)  --  leading coefficient of s

    Example: lcoeff(3*exp(1/x)+y*ln(x)+O(x^4)) -> 3 
--*/
gseries::lcoeff := proc(a)
    local l;
begin
    if args(0) <> 1 then
        error("expecting one argument")
    end_if;

    l:=extop(a,1);
    if nops(l)=0 then 
        return( FAIL )
    else 
        return( l[1][1] )
    end_if
end_proc:

gseries::coeff := proc(a,n)
    local l;
begin
    case args(0)
    of 1 do
        return( op(extop(a,1)) )
    of 2 do
        if not testtype(n,Type::PosInt) then 
           error("2nd argument must ba a positive integer") 
        else
            l:=extop(a,1);
            if n > nops(l) then 
                return( FAIL )
            else 
                return( l[n][1] )
            end_if
        end_if
    otherwise
        error("expecting one or two argument(s)")
    end_case
end_proc:

gseries::nthterm := proc(a,n)
    local l;
begin
    if args(0) <> 2 then
        error("expecting two arguments")
    elif not testtype(n,Type::PosInt) then 
        error("2nd argument must be a positive integer") 
    else
        l:= extop(a,1);
        if n > nops(l) then 
            return( FAIL )
        else 
            return( l[n][2] )
        end_if
    end_if
end_proc:

gseries::nthmonomial := proc(a,n)
    local l;
begin
    if args(0) <> 2 then
        error("expecting two arguments")
    elif not testtype(n,Type::PosInt) then 
        error("2nd argument must be a positive integer") 
    else
        l:= extop(a,1);
        if n > nops(l) then 
            return( FAIL )
        else 
            return( l[n][1]*l[n][2] )
        end_if
    end_if
end_proc:

gseries::expr := proc(a)
    local l,i;
begin
    l:= extop(a,1);
    return( _plus(_mult(op(l[i])) $ i=1..nops(l)) )
end_proc:

gseries::_plus := proc(a,b)
    local f, c;
begin
    case args(0)
    of 0 do
        return( 0 )
    of 1 do
        return( a )
    of 2 do
        if domtype(a) <> gseries then a:= gseries::convert(a) end_if;
        if domtype(b) <> gseries then b:= gseries::convert(b) end_if;

        if not has(extop(a,1),{extop(a,3),extop(b,3)}) then 
            if gseries::ldegree(b) = 0 then
                a:= op(extop(a,1),1); // term of degree 0
                f:= op(extop(b,1),1);
                f:= subsop( f,2=a[1]*a[2]+f[1]*f[2],1=1 );
                return( extsubsop( b,1= subsop(extop(b,1),1=f) ) )
            else
                return( extsubsop(b,1=extop(a,1).extop(b,1)) )
            end_if
        elif not has(extop(b,1),{extop(b,3),extop(a,3)}) then 
            if gseries::ldegree(a) = 0 then
                b:= op(extop(b,1),1); // term of degree 0
                f:= op(extop(a,1),1);
                f:= subsop( f,2=b[1]*b[2]+f[1]*f[2],1=1 );
                return( extsubsop( a,1= subsop(extop(a,1),1=f) ) )
            else
                return( extsubsop(a,1=extop(b,1).extop(a,1)) )
            end_if
        elif extop(a,3) = extop(b,3) then
            f:= op(indets(extop(a,3)),1);
            case extop(a,3)
            of  f do
                return( gseries::new( expr(a)+expr(b),f=infinity ) )
            of -f do
                return( gseries::new( expr(a)+expr(b),f=-infinity ) )
            otherwise
                // extop(a,3) = 1/(x-c), x->c
                c:= expand(1/extop(a,3)-f);
                return( gseries::new( expr(a)+expr(b),f=-c ) )
            end_case
        else
            return( FAIL )
        end_if
    otherwise
        a:= _plus( a,b );
        if f = FAIL then
            return( FAIL )
        else
            return(_plus( a,args(3..args(0)) ))
        end_if
    end_case
end_proc:

gseries::_subtract:= (x,y) -> gseries::_plus( x,gseries::scalmult(y,-1,1) ):

gseries::_mult := proc(a,b)
    local err1,err2,la,lb,ea,eb;
begin
    case args(0)
    of 0 do
        return( 1 )
    of 1 do
        return( a )
    of 2 do
        if domtype(a) <> gseries then 
            a:= gseries::convert(a);
            if domtype(a) <> gseries then return( FAIL ) end_if
        end_if;
        if domtype(b) <> gseries then
            b:=gseries::convert(b);
            if domtype(b) <> gseries then return( FAIL ) end_if
        end_if;

        la:= extop(a,1); lb:= extop(b,1);
        ea:= extop(a,2); eb:= extop(b,2);
        if la = [] then err1:= ea else err1:= la[1][2] end_if;
        if lb = [] then err2:= eb else err2:= lb[1][2] end_if;
        err1:= err1*eb; err2:= err2*ea;
        if gseries::compare(err1,err2,extop(a,3)) = -1 then err1:=err2 end_if;

        // now err1 is the dominant function:
        if nops(la) = 1 and nops(lb)=1 then 
            return( extsubsop(a,1=[zip(la[1],lb[1],_mult)],2=err1) )
        elif nops(la) = 1 and ea = 0 then
            return( extsubsop(a,1=map(lb,zip,la[1],_mult), 2=err1) )
        elif nops(lb) = 1 and eb = 0 then
            return( extsubsop(b,1=map(la,zip,lb[1],_mult), 2=err1) )
        else 
            return( FAIL )
        end_if
    otherwise
        a:= _mult( a,b );
        if a = FAIL then
            return( FAIL )
        else
            return(_mult( a,args(3..args(0)) ))
        end_if
    end_case
end_proc:

/*--
    subs( s,x=... )  --  substitute on the terms of s

    This method applies the function 'subs' to the
    first operand of s, i.e., the terms of s.
--*/
gseries::subs := proc(s) 
begin 
    new(gseries,map( extop(s,1),subs,args(2..args(0)) ),extop(s,2),extop(s,3) )
end_proc:

gseries::subsex := proc(s) 
begin 
    new(gseries,map( extop(s,1),subsex,args(2..args(0)) ),extop(s,2),extop(s,3) )
end_proc:

/*--
    set_var( s,x,y,a )  --  substitute the variable x by y, where y -> a.
--*/
gseries::set_var := proc(s,x,y,a) 
begin 
    new(gseries, op(subs( [extop(s)], x =
            if domtype(a)=stdlib::Infinity then sign(a)/y else y-a end
        ))
    )
end_proc:

gseries::diff:= () -> error("sorry, not implemented yet"):

gseries::iszero:= proc(s)
begin
    if extop(s,2) <> 0 then return( FALSE ) end_if;
    return(iszero( expand(_plus( op(map( extop(s,1),x -> _mult(x[1],x[2]) )) )) ))
end_proc:

gseries::TeX := s -> generate::TeX(gseries::print(s)):

gseries::rectform:= FAIL:

gseries::_power:= proc(s,n)
begin
    if domtype(n) <> DOM_INT or n < 0 then
        // _invert not implemented yet
        return( FAIL )
    elif n = 0 then 
        return(extsubsop(s,1 = [[1,1]]))
    elif n = 1 then
        return(s)
    else
        return(gseries::_mult(s $ n))
    end_if
end_proc:

unalias( gseries ):

