/*
 * series slot of sign - computes series(sign(f), x, n, dir)
 *
 * If dir <> Left, Right, Real ==> simply returns sign(f) + O(x^n).
 * Otherwise expands f, writes it as f0 * x^d * (1 + f1), simplifies
 * sign(f0) * sign(x)^d by taking into account the direction,
 * and returns sign(f0) * sign(x)^d * series((1 + f1) / abs(1 + f1)).
 *
 * Author: Juergen Gerhard, 18.5.00
 */

sign := slot(sign, "series", proc(f, x, n, dir)
  local t, d, f0, f1, X, s, tr, ti;
begin 

  // handle special case sign(x) separately for efficiency reasons
  if f = x then
    case dir
    of Right do
      return(Series::Puiseux::one(n, x));
    of Left do
      return(-Series::Puiseux::one(n, x));
    otherwise
      return(Series::Puiseux::scalmult(Series::Puiseux::one(n, x),
                                       sign(x), 0));
    end_case
  end_if;

  if dir <> FAIL then // directional expansion

    // replace x by a new variable X and assign the properties implied
    // by the direction to it
    X := genident();
    case dir
    of Left do
      assume(X < 0);
      break
    of Right do
      assume(X > 0);
      break
    of Real do
      assume(X, Type::Real);
      break
    end_case;

    // recursively expand the argument
    t := Series::series(eval(subs(f, x = X)), X, n, dir);

    if domtype(t) = Series::Puiseux and (d := ldegree(t)) <> FAIL then
      
      // write t as f0 * x^d * (1 + f1), then sign(f) =
      // sign(f0) * sign(x)^d * (1 + f1) / abs(1 + f1)
      f0 := lcoeff(t);
      f1 := Series::Puiseux::scalmult(lmonomial(t, Rem)[2], 1/f0, -d);

      s := sign(f0) * sign(X)^d;
      s := eval(subs(s, X = x));

      // return s * (1 + f1) / sqrt((1 + Re(f1)^2) + Im(f1)^2)

      // some shortcuts
      if normal(eval(subs(f, x = X)) - f0*X^d) = 0 then // f = f0*x^d
        unassume(X);
        return(Series::Puiseux::scalmult(Series::Puiseux::one(n, x), s, 0));
      elif ldegree(f1) = FAIL then // f = f0 * x^d * (1 + O(..))
        unassume(X);
        return(Series::Puiseux::scalmult(1 + subs(f1, X = x), s, 0));
      end_if;
    
      // split f1 into real and imaginary part tr + ti
      tr := subs(map(f1, Re), X = x);
      ti := subs(f1 - tr, X = x);
      f1 := subs(f1, X = x);
      unassume(X);

      if ldegree(ti) = FAIL then // ti = O(..)
        return(Series::Puiseux::scalmult(1 + ti, s, 0))
      else
        return(Series::Puiseux::scalmult(
                 (1 + f1) / ((1 + tr)^2 - ti^2)^(1/2), s, 0))
      end_if

    end_if;

    unassume(X);

  end_if;

  // default: move sign(f) into the 0th coefficient
  return( Series::Puiseux::scalmult(Series::Puiseux::one(n, x),
                                    hold(sign)(f), 0) );

end_proc):
