/* */
/* $Date: 2000/11/30 17:43:58 $ $Author: stefanw $ $Revision: 1.49 $ */
/* */

/* solvelib::ineqs(inequation, va)

   solve inequation with respect to variable va

   va should have been assumed real before, to allow better
   handling of inequation

*/



/* rewritten. stefanw 18/08/1998 */



solvelib::ineqs :=
proc(inequation,var)
  local crit_pts, ineq, eq, sls, sol, realvals;

begin
  userinfo(3, "Inequality solver called for ".expr2text(inequation));

  if inequation::dom::ineqs <> FAIL then
    return(inequation::dom::ineqs (args()))
  end_if;

  if args(0)<2 then
    error("Wrong number of arguments")
  end_if;
    
  case type(inequation)
    of DOM_SET do
    of DOM_LIST do
      error("Call ineqsys instead!")
    of "_equal" do
      error("Inequality solver cannot solve equations!")
    of "_unequal" do
      error("For inequations of type _unequal, call solve instead")
  end_case;


  ineq:=subsop(inequation,
               1=rewrite(op(inequation,1)-op(inequation,2), piecewise), 2=0);

  userinfo(8, "Rewriting gives ".expr2text(ineq));
  
  /* now ineq has the form expression <= 0 or expression < 0 */

  // handle constant inequalities first
  
  if indets(inequation) minus Type::ConstantIdents={} then
    case is(inequation)
      of TRUE do
        return(Dom::Interval(-infinity, infinity))
      of FALSE do
        return({})
      of UNKNOWN do
        return(hold(solve)(inequation))
      otherwise
        error("Unexpected result from is")
    end_case
  end_if;
  
  if contains( Type::ConstantIdents,var ) then
    error("Invalid variable (mathematical constant was given)")
  end_if; 

  userinfo(5, "Solving for variable ".expr2text(var));

  eq := normal(op(ineq,1));

  if type(eq)=piecewise then
    userinfo(20, "solving piecewise defined inequation by case analysis");
    return(_union(map(op(eq),
                      /* for every branch [cond, expression],
                         intersect the set of all numbers that satisfy cond
                         and the solution of expression < (<=) 0 */
                      proc(branch)
                      begin
                        userinfo(20, "Handling case ".expr2text(op(branch,1)));
                        solve(op(branch,1), var) intersect
                        solve(subsop(ineq, 1=op(branch,2)), var)
                      end_proc
                        )))
  end_if;
    
  // if eq takes on complex values, only those values of var for
  // which eq is real can be solutions

  realvals:= solvelib::preImage(eq, var, solvelib::BasicSet(Dom::Real));

  userinfo(10, "Function takes on real values for ".expr2text(realvals));

  // special case: few real values 

  if type(realvals)=DOM_SET then
    // subdivide the points for which ineq takes on real values into
    // solutions, non-solutions and others
    
    sls:=split(realvals, x-> is(subs(ineq, var=x)) );

    if sls[3]={} then
      return(sls[1])
    else
      return(sls[1] union hold(_intersect)(hold(solve)(ineq, var), sls[3]))
    end_if;
    
  end_if;

  
  crit_pts:= discont(eq, var, Dom::Real);

  userinfo(5, "Discontinuities are ".expr2text(crit_pts));

  sol:=solve(eq, var, hold(Domain)=Dom::Real);

  // all reals are solutions: strict inequality cannot be solvable

  if sol=solvelib::BasicSet(Dom::Real) then
    if type(ineq)="_less" then
      return({})
    else
      return(realvals)
    end_if
  end_if;
  
  if type(sol)=hold(_intersect) then
    // to be careful, assume all solutions to be real
    sol:=select(sol, _unequal, solvelib::BasicSet(Dom::Real))
  end_if;
  
  userinfo(5, "Solving equation gives ".expr2text(sol));
  
  if type(sol)="solve" then
    userinfo(5, "Solve not powerful enough");
    return(hold(solve)(ineq,var))
  end_if;


  crit_pts intersect solvelib::BasicSet(Dom::Real);

  if domtype(%)<>DOM_EXPR then
    // no _union or _intersect, operation was successful
    crit_pts:= %
  end_if;

  userinfo(10, "Critical points are ".expr2text(crit_pts));
  
  sls:= sol union crit_pts;

  sls:= solvelib::ineqWithEqSolved(sls, ineq, var, sol);

  if type(sls)="solve" then
    return(sls)
  else  
    userinfo(10, "Intersecting with set of real values ".expr2text(realvals));
    sls intersect realvals
  end_if

end_proc:



/*
solvelib::ineqWithEqSolved

solves an inequation of type "_less" for which the corresponding 
equation has been investigated

parameters:

sls  -  set of all points for which a change of sign could occur
ineq -  inequation
var  - variable to solve for
sol  -  set of solutions of the equation

*/


solvelib::ineqWithEqSolved:=
proc(sls, ineq, var, sol)
local newineq;

begin

case type(sls)
  of DOM_SET do
    break // handle below
  of solvelib::BasicSet do
    if sls=solvelib::BasicSet(Dom::Complex)
    or sls=solvelib::BasicSet(Dom::Real) then
      return({})
    end_if;
    // Q : impossible by continuity
    // Z : not yet implemented
  of piecewise do
    return(map(sls, solvelib::ineqWithEqSolved, ineq, var, sol))
  of "_intersect" do
      return(hold(solve)(ineq, var))
  otherwise
    userinfo(5,"Cannot handle sets of type ".expr2text(type(sls)));
    return(hold(solve)(ineq, var))
end_case;
   
  userinfo(10, "Solution must be some union of intervals");
  userinfo(10, "Possible borders are: ".expr2text(sls));
  
  if map(sls, x-> indets(x) minus Type::ConstantIdents)={{}} then
    sls:=sort([op(sls)],
              proc()
              begin
                float(args(1))<float(args(2))
              end_proc)
  else
    /* parametric case */
    sls:= solvelib::conditionalSort([op(sls)])
  end_if;

  userinfo(10, "Sorting points gives: ".expr2text(sls));

  
  if sls=[] then
    userinfo(20, "Inequation is satisfied iff it is satisfied for ".
                 expr2text(var). " = 0 ");
    newineq:= subs(ineq, var=0);
    return(piecewise([newineq, solvelib::BasicSet(Dom::Real)],
                      [not ((op(newineq,1)-op(newineq,2)) in
                            solvelib::BasicSet(Dom::Real)),
                       hold(solve)(ineq, var)],
                       [not newineq, {}]))
  end_if;

   if type(sls)=piecewise then
    sls:=map([op(sls)],
             proc(x)
               name solveDependingOnSorting;
                   begin
                     solvelib::intermediatevaluetheorem(op(x,2),
                                       ineq, var, sol, op(x,1));
                     if type(%)=piecewise then
                       op(piecewise::restrict(%,op(x,1)))
                     else
                       [op(x,1), %]
                     end_if;
                     userinfo(10, "Branch ".expr2text(x)." gives ".
                              expr2text(%));
                     eval(%2);
             end_proc);
    sls:=piecewise(op(sls))
  else
    sls:=solvelib::intermediatevaluetheorem(sls, ineq, var, sol, NIL)
  end_if;

  userinfo(10, "Intermediate value theorem solver returns ".expr2text(sls));
  
  return(sls)

end_proc:

 
/*
solvelib::intermediatevaluetheorem

applies the intermediate value theorem to the inequality ineq

parameters:

pointlist  - a list of all points where the inequality could be discontinous,
             or begin to take on non-real values, or where the difference
             left hand side minus right hand side could be zero.
             The points must be sorted in ascending order.
ineq       - inequality (of type "_less" or "_leequal")
var        - identifier (variable to solve for)
sol        - set of zeroes of the equation left hand side minus right and side

output: set of solutions of the inequation
                          
*/


solvelib::intermediatevaluetheorem:=
proc(pointlist, ineq, var, sol /* , condition */)
local interval, sls;

begin
  // if condition<>NIL then assume(condition) end_if;
  // risky because of side effects
  sls:=[-infinity].pointlist.[infinity];

  /* make borders closed if border point is a solution and <=0 was asked */
  // works only for finite sets

  if type(ineq)="_leequal" and type(sol)=DOM_SET then
    sls:=map(sls, proc(u)
                  begin
                    if contains(sol, u) then [u] else u end_if
                  end_proc)
  end_if;

  userinfo(10, "Interval borders are: ".expr2text(sls));
  
  interval:= solvelib::intervals(sls);

  userinfo(10, "Intervals are: ".expr2text(interval));

  solvelib::selectintervals(interval, ineq, var)
 
end_proc:


/*
solvelib::intervals

input: a sorted list of points
returns a list of intervals separating the points

*/                          

solvelib::intervals:=
proc(sls)

begin
  zip([op(sls,1..nops(sls)-1)],[op(sls,2..nops(sls))],
            Dom::Interval);
end_proc:


/*

solvelib::selectintervals(intervallist, ineq, var)

input: a list of intervals, an inequation, and a variable such that for
each interval, either none or all of its points satisfy the
inequation when substituted for var

output: union of all intervals whose elements satisfy the inequation

*/                          


solvelib::selectintervals:=
proc(intervallist, ineq, var)
local sol;

begin
  userinfo(10, expr2text(nops(intervallist))." intervals to check");
  userinfo(10, "Intervals are: ".expr2text(intervallist));

  if nops(intervallist)=0 then
    return({})
  end_if;
  
  sol:=map(intervallist,
           proc(x)
             name testinterval;
             local pt,newineq,newvar,solution;
           begin
             if type(x)<>Dom::Interval then
               // x is a finite set: there are no interior points to check
               return({})
             end_if;
             
           // select a point pt in the given interval: evaluating
           // the inequality at that point gives TRUE iff the inequality
           // is satisfied in the entire interval
             if Dom::Interval::left(x)=-infinity then
               pt:=Dom::Interval::right(x)-1
             elif Dom::Interval::right(x)=infinity then
               pt:=Dom::Interval::left(x)+1
             else
               pt:=(Dom::Interval::right(x)+
                    Dom::Interval::left(x))/2
             end_if;
           userinfo(10, "Checking ".expr2text(x).
                    " at ".expr2text(pt));
           newineq:=eval(subs(ineq, var=pt));
           userinfo(10, "Substituting gives ".expr2text(newineq));
           if indets(newineq) minus Type::ConstantIdents={} then
             userinfo(15, "Inequation has no variables");
             map(newineq,float);
             if not has(%, I) and bool(%) then  /* and must be lazy here! */
               userinfo(15, expr2text(%)." is TRUE");
               x
             else
               userinfo(15, expr2text(%)." is FALSE");
               {}
             end_if
           else
             newvar:=op(indets(newineq),1);
             solution:=solve(newineq, newvar);
             userinfo(6, "Solving ". expr2text(newineq)." for ".
                      expr2text(newvar)." gives ". expr2text(solution));
             if solution={} then
               {}
             elif solution=Dom::Interval(-infinity, infinity) then
               x
             elif type(solution)="solve" then
               piecewise([newineq, x],[not newineq,{}])
             else
               _in(newvar, solution);
               piecewise([%, x], [not %, {}])
             end_if;
           end_if;
           end_proc
);

  userinfo(10, "Intervals: ".expr2text(intervallist));
  userinfo(10, "Solutions for each interval: ".expr2text(sol));
 
  
  
 if nops(sol)=0 then
   {}
 elif nops(sol)=1 then
   sol[1]
 else
   _union(op(sol))
 end_if
end_proc:
  


solvelib::ineqsys:=
proc(ineqs, var)
local sls;
  
begin
  sls:=map(ineqs, solve,var);
  if has(map(sls, type), "solve") then
    return(hold(solve)(ineqs,var))
  end_if;
  sls := _intersect(op(sls));
  return(sls);
end_proc:
                      

/* end of file */






























