
// solvelib::makeSystem(l, r, x)

// converts l=r into a system of equations

/*

   l - expression
   r - expression
   x - identifier

   r must be independent of x


   returns:

   - a system of algebraic equations eqsystem
   - a set of remaining (non-algebraic) equations
   - a list of variables vars
   - a list of conditions

  such that the coordinate labelled by x=...
  of every solution of the eqsystem and the nonalg. eq.
  for vars whose i-th
  entry satisfies the i-th condition is also a solution to l=r 


*/

solvelib::makeSystem:=
proc(l, r, x:DOM_IDENT)
  local equation, subslist, i, j, s, rsid, nthroot, n,
  eqsystem: DOM_SET,
  nonalgebraic: DOM_SET,
  vars: DOM_LIST,
  conditions: DOM_LIST;
  
begin
  // turn l into a rational equation, with many variables
  rationalize(l, FAIL, subexpr-> not has(subexpr, x) or subexpr=x, {},
              Recursive);
  [equation, subslist]:=[op(%,1), op(%,2)];
  // subslist is a list of equations new_variable = irrational expression

  eqsystem:= { equation = r};
  nonalgebraic:={};
  vars:=[x];
  conditions:=[TRUE];
  
  for i from nops(subslist) downto 1 do
    s:= subslist[i];  // i-th equation
    // substitute in former operands wherever possible
    for j from 1 to i-1 do
      subslist[j]:=subs(subslist[j], rhs(s)=lhs(s))
    end_for;
    rsid:=rhs(s);
    /* if indets(rsid) intersect {op(vars)}={} then
      next
    end_if; */ 
    case type(rsid)
      of "_plus" do
      of "_mult" do
        // polynomial subexpressions are ok
        break;
      of DOM_COMPLEX do
      //  if rsid<>I then
          // should not happen (rationalize should not return this)
          error("Unexpected result from rationalize")
      //  end_if;
      //  eqsystem:= eqsystem union {lhs(s)^2=-1};
      //  vars:= append(vars,lhs(s));
      //  conditions:= append(conditions, Im(lhs(s))>0);
      //  break;
      of "_power" do
        if type(op(rsid,2))=DOM_INT then
          // ok
          break
        end_if;
        if not has(op(rsid, 2), x) then
          // exponent is independent of x
          
        // if var = a^(p/q), write this as var^q=a^p
        // this in not an equivalence; hence it must be
        // checked whether the resulting var can be a q-th root at all
          eqsystem:= eqsystem union
                     { lhs(s)^denom(op(rsid, 2)) = op(rsid,1)^numer(op(rsid,2))};
          vars:= append(vars, lhs(s));
          nthroot:=lhs(s);
          n:=denom(op(rsid, 2));
          if n=2 then
            conditions:= append(conditions,
                                Re(nthroot)>=0 and Im(nthroot)>=0
                                or Im(nthroot)>0)
          else
            conditions:= append(conditions,
                                arg(Re(nthroot), Im(nthroot)) >= 0 and
                                arg(Re(nthroot), Im(nthroot)) <
                                2*PI/denom(op(rsid,2)))
          end_if;
          break
        else
          userinfo(10, "exponent depends on ".expr2text(x));
          nonalgebraic:=nonalgebraic union {s};
          vars:=append(vars, lhs(s));
        end_if
      otherwise
        userinfo(10, "Cannot handle right hand side ".
                 expr2text(rhs(s)));
        nonalgebraic:= nonalgebraic union {s};
        vars:=append(vars, lhs(s));
    end_case;
    
  end_for;

  eqsystem, nonalgebraic, vars, conditions 
end_proc:
