// $Date: 2000/10/31 14:28:16 $ $Author: stefanw $ $Revision: 1.22 $ 

// kg, 13/01/94 

/*++
arccos -- the inverse cosine

arccos(x)

x - expression

NB:  The branch cuts are the real intervals (-infinity, -1) and
     (1, infinity). 
     Note: for x >=1 :  arccos(x) =  limit(arccos(x+eps*I),eps=0, eps>0) 
                                  = -limit(arccos(x-eps*I),eps=0, eps>0) 
           for x <=-1:  arccos(x) =  limit(arccos(x-eps*I),eps=0, eps>0) 

     I.e., on the left real line the values coincide with the limits from
     below, on the right real line with the limit from above. OK.

     Consistent with arccos(x)+arccos(-x) = PI. OK.
++*/

arccos:=
proc(x)
  local f;
  name arccos;
  option noDebug;
begin
  if args(0) = 0 then error("no arguments given")
  elif x::dom::arccos <> FAIL then return(x::dom::arccos(args()))
  elif args(0) <> 1 then error("wrong no of args") 
  end_if;

  if domtype(x) = DOM_SET then
      return(map(x,arccos))
  elif not testtype(x,Type::Arithmetical) then
      error("argument must be of 'Type::Arithmetical'")
  end_if;

  case type(x)
    of DOM_INT do
    of DOM_RAT do
	if x < 0 then return(PI - arccos(-x)) end_if;
	break;
    
    of DOM_FLOAT do
	return(slot(arccos,"float")(x));

    of DOM_COMPLEX do
	if domtype(op(x,1)) = DOM_FLOAT or
	   domtype(op(x,2)) = DOM_FLOAT then
	    return(slot(arccos,"float")(x))
	end_if;
	if op(x,1) = 0 then return(PI/2 - arcsin(x)) end_if;
	break;

    of "_mult" do
	f:= op(x, nops(x));
	if testtype(f, Type::Real) then
	    if f < 0 then return(PI - arccos(-x)) end_if;
	end_if;
	break;

    of "cos" do // arccos : [-1..1] --> [PI..0] 
       if (is(op(x)>=0) and is(op(x)<=PI))=TRUE then return(op(x))
       elif type((f:=floor(Re(op(x))/PI)))=DOM_INT then
          case f mod 2
          of 0 do return(op(x)-f*PI)
          of 1 do return((f+1)*PI-op(x))
          end_case;
       end_if; 
    end_case;

    procname(x)
end_proc:

arccos( 0):=  PI/2:  
arccos( 1):= 0:
arccos(-1):= PI:
arccos( 1/2):=   PI/3:
arccos(-1/2):= 2*PI/3:
arccos( (3^(1/2))/2):=   PI/6:
arccos(-(3^(1/2))/2):= 5*PI/6:
arccos( (2^(1/2))/2):=   PI/4:
arccos(-(2^(1/2))/2):= 3*PI/4:
arccos( 1/4*2^(1/2)*(5+5^(1/2))^(1/2)):=   PI/10:
arccos(-1/4*2^(1/2)*(5+5^(1/2))^(1/2)):= 9*PI/10:
arccos( 1/4*2^(1/2)*(5-5^(1/2))^(1/2)):= 3*PI/10:
arccos(-1/4*2^(1/2)*(5-5^(1/2))^(1/2)):= 7*PI/10:
arccos( ((2-2^(1/2))^(1/2))/2):= 3*PI/8:
arccos(-((2-2^(1/2))^(1/2))/2):= 5*PI/8:
arccos( ((2+2^(1/2))^(1/2))/2):=   PI/8:
arccos(-((2+2^(1/2))^(1/2))/2):= 7*PI/8:
arccos( 1/4*(5^(1/2)+1)):=   PI/5:
arccos(-1/4*(5^(1/2)+1)):= 4*PI/5:
arccos( 1/4*(5^(1/2)-1)):= 2*PI/5:
arccos(-1/4*(5^(1/2)-1)):= 3*PI/5:
arccos( (6^(1/2)-2^(1/2))/4):= 5*PI/12:
arccos(-(6^(1/2)-2^(1/2))/4):= 7*PI/12:
arccos( (6^(1/2)+2^(1/2))/4):=   PI/12:
arccos(-(6^(1/2)+2^(1/2))/4):=11*PI/12:

arccos( infinity):=   -I*infinity:
arccos(-infinity):= PI+I*infinity:
arccos( I*infinity):= PI/2 - I*infinity:
arccos(-I*infinity):= PI/2 + I*infinity:

arccos:= funcenv(arccos):
arccos:= subsop(arccos, 2=op(specfunc::arccos, 2)):
arccos:= slot(arccos,"info","arccos -- the inverse cosine"):
arccos:= slot(arccos,"print","arccos"):
arccos:= slot(arccos,"type","arccos"):
arccos:= slot(arccos, "float", specfunc::arccos):

arccos:= slot(arccos,"inverse", "cos"):

arccos:= slot(arccos,"diff",proc() begin 
    op(args(1),1);
    -diff(%, args(2..args(0))) / sqrt(1-%^2)
 end_proc ):

arccos:= slot(arccos, "complexDiscont",
              Dom::Interval(-infinity, -1) union Dom::Interval(1, infinity)):
arccos:= slot(arccos, "realDiscont", {}):
arccos:= slot(arccos, "undefined", {}):

arccos:= slot(arccos, "series",
    loadproc(slot(arccos,"series"),pathname("SERIES"),"arccos")
):

arccos:= slot(arccos,"rectform",
    loadproc(slot(arccos,"rectform"),pathname("STDLIB","RECTFORM"),"acos")
):


// end of file 
