/*++
arg -- the polar angle ("the argument") of a complex number

arg(x, y)

x, y - real quantifiers, representing the complex number x+I*y

arg(x,y) computes the principal value of the argument of the complex
number x+I*y with -PI < arg(x,y) <= PI.

Mathematical identities:
  1) arg(x,y) = ln(sign(x+I*y))/I
  2) arg(x,y) = arctan(y/x) + PI/2*sign(y)*(1-sign(x))
  
1) holds for any finite real x, y
2) holds for any x <> 0, y<>0  (unfortunately, sign(0) = 0 making
                                this formula wrong for y=0, x<0)

The value for x=y=0 is undefined. 

!!!!!!!!  To do: the discont attributes are incorrect. !!!!!!!!!!
!!!!!!!!  To do: the sign attribute is missing. !!!!!!!!!!
++*/

arg:= proc(x, y)
  option noDebug;
  local floatconvert, sx, sxknown, sy, syknown, xx, yy, dummy;
  name arg;
begin
  // ---------- argument check -----------------------
  if args(0) <> 2 then
       error("wrong no of args") 
  elif x::dom::arg <> FAIL then
       return(x::dom::arg(args()))
  elif y::dom::arg <> FAIL then
       return(y::dom::arg(args()))
  end_if;

  // ----------  map to sets -------------------------
  // This is quite complicated, because both arguments have
  // to be handled
  case [type(x),type(y)]
    of [DOM_SET, DOM_SET] do
        return({arg(sx, sy) $ sx in x $ sy in y});
    of [DOM_SET, "_union"] do
    of ["_union", DOM_SET] do
        return(_union((arg(sx, sy) $ sx in x) $ sy in y));
    of ["_union", "_union"] do

        // arg({a} union A, C union D)
        //  -> arg({a},C) union arg({a},D) union arg(A,C) union arg(A,D)
        // Make sure that A, C, D are interpreted as sets, i.e.,
        // arg({a},C) --> arg(a, C), not {arg(a,C)} !

        [x, xx, dummy]:= split(x, testtype, DOM_SET):
        if type(xx) <> "_union" then xx:= [xx] end_if:
        [y, yy, dummy]:= split(y, testtype, DOM_SET):
        if type(yy) <> "_union" then yy:= [yy] end_if:
 
        return(_union({(arg(sx, sy) $ sx in x ) $ sy in y},
                       (arg(sx, sy) $ sx in xx) $ sy in y,
                       (arg(sx, sy) $ sx in x ) $ sy in yy,
                       (arg(sx, sy) $ sx in xx) $ sy in yy ));
  end_case;

  case type(x)
    of DOM_SET do
    of "_union" do
       // y cannot be a set, if x is a set
       return(map(x, arg, y));
  end_case;

  case type(y)
    of DOM_SET do
    of "_union" do
       // x cannot be a set, if y is a set
       return(map(y, (y,x) -> arg(x,y), x))
  end_case;
  // --------------------------------------------------

  if not testtype(x,Type::Arithmetical) then
     error("first argument must be of 'Type::Arithmetical'")
  end_if;
  if not testtype(y,Type::Arithmetical) then
     error("second argument must be of 'Type::Arithmetical'")
  end_if;

  if domtype(float(x)) = DOM_COMPLEX then
    error("first argument must be real")
  end_if;
  if domtype(float(y)) = DOM_COMPLEX then
    error("second argument must be real")
  end_if;
  // --------------------------------------------------

  if domtype(x) = DOM_FLOAT or domtype(y) = DOM_FLOAT 
     then floatconvert:= float;
     else floatconvert:= id;
  end_if:

  sx:= sign(x); sxknown:= bool(domtype(sx)=DOM_INT);
  sy:= sign(y); syknown:= bool(domtype(sy)=DOM_INT);

  if iszero(sx) and iszero(sy) then
     // each value between -PI < phi <= +PI would make sense here:
     return(floatconvert(0));
  end_if:
  if iszero(sx) /*and syknown*/ then // sy <> 0 here !!
        return(floatconvert(sy*PI/2));
  end_if:
  if iszero(sy) /*and sxknown*/ then // sx <> 0 here !!
        return(floatconvert((1-sx)*PI/2));
  end_if:

  // now sx<>0 and sy<>0 in the following !!

  // ----- handle infinity ----------
  if has(x,infinity) then
     if has(y,infinity)
        then return(procname(x,y))
        else if sxknown and syknown then
                return(floatconvert(sy*(1-sx)*PI/2));
             end_if;
     end_if;
  end_if;
  if has(y,infinity) then
     if syknown then
        return(floatconvert(sy*PI/2));
     end_if;
  end_if;

  // ------------------------------------------
  // explicit result in terms of arctan, if
  // sign(x) and sign(y) are known. Note that
  // sign has already taken care of properties
  // of x, y set via assume.

  //Do not use the following piece of code to avoid
  //return values involving a symbolic sign(x). This
  //is not differentiable at x = 0 (and arctan(y/x)
  //is not smooth at x=0), so differentiation and 
  // series expansion would give odd results.
  //if syknown and not iszero(sy) then 
  //   // the following formula holds for all (x,y) with y<>0
  //   return(floatconvert(arctan(y/x) + PI/2*sy*(1-sx)));
  //end_if:

  //Instead, use this simplication only if sign(x) is known,
  //or is x is known to be non-zero via assume(x<>0)

  if (sxknown or is(x, Type::NonZero)=TRUE) and syknown then
     // note that sx<>0 and sy<>0 here
     return(floatconvert(arctan(y/x) + PI/2*sy*(1-sx)));
  end_if:
  
  // In the right half plane we do not need to know the
  // sign of y, because arg(x,y) = arctan(y/x) for any y.

  if sx = 1 then
     return(floatconvert(arctan(y/x)))
  end_if:

  // Last step: sign(x) or sign(y) are not known, so return
  // an unevaluated call arg(x,y). However, the following
  // "projective" simplification such as arg(2*x,2*y) = arg(x,y)
  // can still be done for any symbolic x, y.
  // The idea is: normalize the first argument

  if type(x) = "_mult" then
     sx:= op(x, nops(x));
     if is(sx>0) = TRUE then 
        x:= x/sx; y:= y/sx;
     end_if:
     if is(sx<0) = TRUE then 
        x:= -x/sx; y:= -y/sx;
     end_if:
  end_if;
  return(procname(x,y));
end_proc:

//---------------------------------------------------------------------
arg:= funcenv(arg):
arg:= slot(arg,"print","arg"):
arg:= slot(arg,"info",
           "arg -- the polar angle of a complex number [try ?arg for help]"):
arg:= slot(arg,"type","arg"):
//---------------------------------------------------------------------

// arg(0, 0):= 0:

//---------------------------------------------------------------------
// for version arg(z)=arg(x+I*y) the following would be correct:
//arg:= slot(arg, "realDiscont", {0}):  
//arg:= slot(arg, "complexDiscont", Dom::Interval(-infinity, [0])):

//---------------------------------------------------------------------
arg:=slot(arg,"float",proc(x,y)
begin
   if args(0)=2 
      then arg(float(x),float(y))
      else error("wrong no of args") 
   end_if
end_proc):

//---------------------------------------------------------------------
/* Obsolete: this is now a part of arg itself:
arg:= slot(arg,"simplify",proc(e)
    local x, y;
begin
    x:= simplify(op(e,1));
    y:= simplify(op(e,2));
    if is(x>0) = TRUE then return(arctan(y/x)) end_if:
    if is(x<0) = TRUE then
       if iszero(y) then return(PI) end_if;
       if is(y<0)=TRUE or is(y>0)=TRUE then
          return(arctan(y/x)+PI*sign(y) ) 
       end_if:
    end_if;
    return(arg(x,y) )
end_proc):
*/

//---------------------------------------------------------------------
arg:= slot(arg,"diff", proc() local x, y;
           begin 
              x:= op(args(1),1);
              y:= op(args(1),2);
              diff(y/x, args(2..args(0))) *x^2/(x^2+y^2)
           end_proc ):
//---------------------------------------------------------------------

arg:= slot(arg, "series",
    loadproc(slot(arg,"series"),pathname("SERIES"),"arg")
):

//---------------------------------------------------------------------

arg:= slot(arg, "rectform",
 proc(xx, yy)
   local x, y;
 begin
   x:= rectform::new( xx );
   y:= rectform::new( yy );
   if extop(x,3) <> 0 or extop(y,3) <> 0 then
     new(rectform,0,0,hold(arg)(rectform::expr(x),rectform::expr(y)))
   else // assume that arg has been correctly called with two real arguments 
       new( rectform,hold(arg)(xx,yy),0,0 )
   end_if
 end_proc):
//---------------------------------------------------------------------
arg:= slot(arg,"Re", proc(x,y) begin hold(arg)(x,y) end_proc):
//---------------------------------------------------------------------
arg:= slot(arg,"Im", proc(x,y) begin 
         if domtype(x)=DOM_FLOAT or
            domtype(y)=DOM_FLOAT 
           then return(float(0))
           else return(0)
         end_if:
         end_proc):
//---------------------------------------------------------------------
// arg:= slot(arg,"expand", proc(x,y) begin ... end_proc):
// arg:= slot(arg,"sign", proc(x, y) begin ... end_proc):

// end of file 
