// Author:  T.C. Scott Feb 2000
/*++
        besselJ(v,z) is the bessel function of the first kind:

        besselJ(v,z) = 1/gamma(v+1)*(z/2)^v*0F1(v+1; -z^2/4)

            = (z/2)^v * sum( (-z^2/4)^k/k!/gamma(v+1+k), k=0..infinity)
++*/

besselJ:=
proc(vv,z)
  local cv,sv,fv,fz,v,r,r2,t1,t2,z2,sz,Vv,zz,dummy;
  option noDebug;
begin
  v:=vv;
  if args(0) = 0 then
    error("no arguments given")
  elif z::dom::besselJ <> FAIL then
    return( z::dom::besselJ(args()) )
  elif contains( {DOM_BOOL,DOM_FAIL,DOM_NULL,DOM_NIL},domtype(v) ) then
    error("invalid operand")
  end_if;

 
  // ----------  map to sets -------------------------
  // This is quite complicated, because both arguments have
  // to be handled
  case [type(v),type(z)]
    of [DOM_SET, DOM_SET] do
        return({besselJ(sv, sz) $ sv in v $ sz in z});
    of [DOM_SET, "_union"] do
    of ["_union", DOM_SET] do
        return(_union((besselJ(sv, sz) $ sv in v) $ sz in z));
    of ["_union", "_union"] do

        // besselJ({a} union A, C union D)
        //  -> besselJ({a},C) union besselJ({a},D) union besselJ(A,C) union besselJ(A,D)
        // Make sure that A, C, D are interpreted as sets, i.e.,
        // besselJ({a},C) --> besselJ(a, C), not {besselJ(a,C)} !

        [v, Vv, dummy]:= split(v, testtype, DOM_SET):
        if type(Vv) <> "_union" then Vv:= [Vv] end_if:
        [z, zz, dummy]:= split(z, testtype, DOM_SET):
        if type(zz) <> "_union" then zz:= [zz] end_if:

        return(_union({(besselJ(sv, sz) $ sv in v ) $ sz in z},
                       (besselJ(sv, sz) $ sv in Vv) $ sz in z,
                       (besselJ(sv, sz) $ sv in v ) $ sz in zz,
                       (besselJ(sv, sz) $ sv in Vv) $ sz in zz ));
  end_case;

  case type(v)
    of DOM_SET do
    of "_union" do
       // z cannot be a set, if v is a set
       return(map(v, besselJ, z));
  end_case;

  case type(z)
    of DOM_SET do
    of "_union" do
       // v cannot be a set, if z is a set
       return(map(z, (z,v) -> besselJ(v,z), v))
  end_case;
  // --------------------------------------------------


  if not testtype(z,Type::Arithmetical) or
    not testtype(v,Type::Arithmetical)
    then
    error("arguments must be of type 'Type::Arithmetical'")
  end_if: 

  fv:= domtype(float(v)):
  fz:= domtype(float(z)):
  if (fv=DOM_FLOAT or fv=DOM_COMPLEX) and (fz=DOM_FLOAT or fz=DOM_COMPLEX)
   and has([map((op(v),op(z)),type)],DOM_FLOAT) then
         return(slot(besselJ,"float")(v,z))
  end_if:

  if iszero(z) then
     if iszero(v) then
            return(1)
      elif (fv=DOM_FLOAT or fv=DOM_COMPLEX) and (Re(v) > 0 or domtype(v)=DOM_INT) then
            return(0):
      elif (fv=DOM_FLOAT or fv=DOM_COMPLEX) and not(Re(v) > 0 or domtype(v)=DOM_INT) then
            error("singularity"):
      else
            procname(v,z);
      end_if:
  end_if:

  if v=1/2 then
      return(2^(1/2)/PI^(1/2)/z^(1/2)*sin(z));
  elif v=-1/2 then
      return(2^(1/2)/PI^(1/2)/z^(1/2)*cos(z));
  elif type(v) <> DOM_INT and type(2*v)=DOM_INT then
      // might want to limit this expansion to say abs(v) <= 10
      // and let "expand" take care of cases > 10.
      //cv:= cos((v+1/2)*PI/2);
      //sv:= sin((v+1/2)*PI/2);
      case (v+1/2) mod 4
        of 0 do cv:= 1: sv:= 0: break;
        of 1 do cv:= 0: sv:= 1: break;
        of 2 do cv:=-1: sv:= 0: break;
        of 3 do cv:= 0: sv:=-1: break;
      end_case;

      if v<1/2
          then v:= -v-1/2:
          else v := v-1/2;
      end_if:
      z2:= z^2:
      t1:= 1:
      t2:= 1:
      return(
         2^(1/2)/PI^(1/2)/sqrt(z)*((cos(z)*cv+sin(z)*sv)*
      _plus(1,(r2:=2*r;t1:=-t1/8*(v+r2-1)*(v+r2)*(-v+r2-1)*(-v+r2-2)/z2/r/(r2-1))
                    $ r=1..trunc(v/2))
        +(cos(z)*sv-sin(z)*cv)* v*(v+1)/2/z*
      _plus(1,(r2:=2*r;t2:=-t2/8*(v+r2)*(v+r2+1)*(-v+r2)*(-v+r2-1)/z2/r/(r2+1))
                    $ r=1..trunc((v-1)/2))));
  else 
	procname(v,z) 
  end_if
end_proc:

besselJ := funcenv(besselJ):
besselJ:= slot(besselJ,"type","besselJ"):
besselJ:= slot(besselJ,"print","besselJ"):
besselJ:= slot(besselJ,"info",
  "besselJ(v,z) -- bessel function of first kind, order v and argument z"
):
besselJ::Content      := (Out, data) -> Out::Capply(Out::CbesselJ, map(op(data), Out)):

besselJ := slot(besselJ, "float",
    loadproc(slot(besselJ, "float"),  pathname("STDLIB","FLOAT"),"besselJ")
):

besselJ := slot(besselJ,"diff",
    loadproc(slot(besselJ,"diff"),pathname("STDLIB","DIFF"),"besselJ")
):

besselJ := slot(besselJ,"series",
    loadproc(slot(besselJ,"series"),pathname("SERIES"),"besselJ")
):

besselJ:= slot(besselJ,"expand",
    loadproc(slot(besselJ,"expand"),pathname("STDLIB","EXPAND"),
        "besselJ"
    )
):


// end of file 
