// T.C. Scott, Feb. 2000
/*++
	besselK(v,z) is the modified bessel function of the second kind.
 	It satisfes the modified bessel equation: 
	       2                2    2
	      x  y'' + x y' - (x  + v ) y = 0
++*/

besselK:=proc(v,z)
local fv,fz,i,n,res,t,twoz,sv,sz,vv,zz,dummy;

begin
  if args(0) = 0 then
    error("no arguments given")
  elif z::dom::besselK <> FAIL then
    return( z::dom::besselK(args()) )
  elif args(0) <> 2 then
    error("expecting two arguments")
  end_if;

  // ----------  map to sets -------------------------
  // This is quite complicated, because both arguments have
  // to be handled
  case [type(v),type(z)]
    of [DOM_SET, DOM_SET] do
        return({besselK(sv, sz) $ sv in v $ sz in z});
    of [DOM_SET, "_union"] do
    of ["_union", DOM_SET] do
        return(_union((besselK(sv, sz) $ sv in v) $ sz in z));
    of ["_union", "_union"] do

        // besselK({a} union A, C union D)
        //  -> besselK({a},C) union besselK({a},D) union besselK(A,C) union besselK(A,D)
        // Make sure that A, C, D are interpreted as sets, i.e.,
        // besselK({a},C) --> besselK(a, C), not {besselK(a,C)} !

        [v, vv, dummy]:= split(v, testtype, DOM_SET):
        if type(vv) <> "_union" then vv:= [vv] end_if:
        [z, zz, dummy]:= split(z, testtype, DOM_SET):
        if type(zz) <> "_union" then zz:= [zz] end_if:

        return(_union({(besselK(sv, sz) $ sv in v ) $ sz in z},
                       (besselK(sv, sz) $ sv in vv) $ sz in z,
                       (besselK(sv, sz) $ sv in v ) $ sz in zz,
                       (besselK(sv, sz) $ sv in vv) $ sz in zz ));
  end_case;

  case type(v)
    of DOM_SET do
    of "_union" do
       // z cannot be a set, if v is a set
       return(map(v, besselK, z));
  end_case;

  case type(z)
    of DOM_SET do
    of "_union" do
       // v cannot be a set, if z is a set
       return(map(z, (z,v) -> besselK(v,z), v))
  end_case;

  // --------------------------------------------------
  if not testtype(z,Type::Arithmetical) or
    not testtype(v,Type::Arithmetical)
    then
    error("arguments must be of type 'Type::Arithmetical'")
  end_if;

  fv:= domtype(float(v)):
  fz:= domtype(float(z)):
  if (fv=DOM_FLOAT or fv=DOM_COMPLEX) and (fz=DOM_FLOAT or fz=DOM_COMPLEX)
   and has([map((op(v),op(z)),type)],DOM_FLOAT) then
         return(slot(besselK,"float")(v,z))
  end_if:

  if iszero(z) then
        error("singularity"):
  end_if:
  if type(v) <> DOM_INT and type(2*v)=DOM_INT then
    // might want to limit this expansion to say abs(v) <= 10
    // and let "expand" take care of cases > 10.
    // Worked out from 10.2.15 and 10.2.16 of A and S
    n := trunc(abs(v));
    twoz:= 2*z:
    t:= 1:     // i=0
    res:= sqrt(PI/2)/sqrt(z)*exp(-z)*
             _plus(1, (t:= t*(n+i)*(n+1-i)/i/twoz) $ i=1..n);
    return(factor(res))
  end_if:
  procname(v,z)
end_proc:

besselK := funcenv(besselK):
besselK:= slot(besselK,"type","besselK"):
besselK:= slot(besselK,"print","besselK"):
besselK:= slot(besselK,"info",
  "besselK(v,z) -- modified bessel function of second kind (order v, argument z)"
):
besselK::Content      := (Out, data) -> Out::Capply(Out::CbesselK, map(op(data), Out)):

besselK := slot(besselK, "float",
    loadproc(besselK::float,pathname("STDLIB","FLOAT"),"besselK")
):

besselK := slot(besselK,"diff",
    loadproc(slot(besselK,"diff"),pathname("STDLIB","DIFF"),"besselK")
):

besselK := slot(besselK,"series",
    loadproc(slot(besselK,"series"),pathname("SERIES"),"besselK")
):

// end of file 
