/*++
        beta -- the Beta function

        beta( x, y ) := gamma( x ) * gamma( y ) / gamma( x + y )

++*/

beta :=
proc(x,y) 
  option noDebug;
  local isInteger;
begin
  if args(0) <> 2 then
    error("expecting two arguments")
  elif x::dom::beta <> FAIL then
    return(x::dom::beta(args()))
  end_if;

  case type(x)
    of DOM_SET do
    of "_union" do
      return(map(x, beta, y))
  end_case;
  case type(y)
    of DOM_SET do
    of "_union" do
      return(map(y, beta, x))
  end_case;
  
  if not testtype(x,Type::Arithmetical) or
     not testtype(y,Type::Arithmetical) then
    error("arguments must be of type 'Type::Arithmetical'")
  end_if;

  // If one argument is a float, then try to convert the
  // other one to a float as well:
  case domtype(x)
  of DOM_FLOAT do 
  of DOM_COMPLEX do
     if domtype(op(x,1))=DOM_FLOAT or
        domtype(op(x,2))=DOM_FLOAT
     then y:= float(y);
     end_if:
  end_case;
  case domtype(y)
  of DOM_FLOAT do 
  of DOM_COMPLEX do
     if domtype(op(y,1))=DOM_FLOAT or
        domtype(op(y,2))=DOM_FLOAT
     then x:= float(x);
     end_if:
  end_case;

  // return symbolic calls of beta with a sorted sequence
  // of arguments to guarantee beta(x,y)=beta(y,x):
  [x,y]:= sort([x,y]);

  isInteger:= proc(x) begin // x=integer or x=float(integer)?
    bool(domtype(x)=DOM_INT)    
    or (testtype(x, Type::Real) and iszero(frac(x)))
  end_proc:

  if iszero(x) or iszero(y) then error("singularity")
  elif iszero(x-1) then return(1/y)
  elif iszero(y-1) then return(1/x)
  elif isInteger(x) and   // x = integer or x = float(integer)
       isInteger(y) then  // y = integer or y = float(integer)
    if x + y <= 0
    then // can reflect negative argument to positive argument
       if x<0 and 0<y then return((-1)^y*beta(1-x-y,y)) end_if;
       if y<0 and 0<x then return((-1)^x*beta(x,1-x-y)) end_if;
       error("singularity");
    else 
       if (x<0 and 0<y) or (y<0 and 0<x) then 
          error("singularity")
       end_if:
    end_if:
  elif isInteger(x) and x <= 0 and testtype(y, Type::Numeric) then
       error("singularity")
  elif isInteger(y) and y <= 0 and testtype(x, Type::Numeric) then
       error("singularity")
  elif isInteger(x+y) and x+y <= 0 then
       return(0)
  end_if;

  // neither x nor y nor x+y is a negative integer:
  if testtype(x, Type::Numeric) and 
     testtype(y, Type::Numeric)
  then return(gamma(x)*gamma(y)/gamma(x+y)):
  else return(procname(x, y))
  end_if:
end_proc:

beta := funcenv(beta):
beta:= slot(beta, "print", "beta"):
beta:= slot(beta, "info",  "beta -- the Beta function beta(x,y)=gamma(x)*gamma(y)/gamma(x+y)"):
beta:= slot(beta, "type",  "beta"):

beta := slot(beta, "diff", proc() begin 
  args(2..args(0)); op(args(1), 2); op(args(1), 1);
  (diff(%1, %3)*psi(%1)+diff(%2, %3)*psi(%2)-diff(%1+%2, %3)*psi(%1+%2))*
  beta(%1, %2)
 end_proc ):

beta := slot(beta, "expand", 
	proc() begin
          expand(gamma(args(1))*gamma(args(2))/gamma(args(1)+args(2))) end_proc ):

beta := slot(beta, "float", proc(a, b) begin beta(float(a), float(b)) end_proc):
