/* The dilogarithm function: dilog(x) = int(ln(t)/(1-t),t=1..x)

  Reference: Abramowitz & Stegun, Handbook of Mathematical Functions,
	section 27.7

related to to Li2(x) polylog(2, 1-x) = int(-log(1-t)/t,t=0..x)
Ref: L. Lewin, Dilogarithms and associated functions, Macdonald, London, 1958
     L. Lewin, Polylogarithms and associated functions, 1981
From Harald Anlauf <anlauf@crunch.ikp.physik.th-darmstadt.de>.
     Other notation ? polylog(n,x) == Li_n(x)
                      polylog(1,x) = -ln(1-x)
                      int( polylog(n,t)/t, t=0..x) = polylog(n+1,x)
                      int(ln(1-x)/x) = -polylog(2,x)
*/

dilog :=
proc(x)
  option noDebug;
  local fx;
begin
  if args(0) = 0 then
    error("no arguments given")
  elif x::dom::dilog<>FAIL then
    return(x::dom::dilog(args()))
  elif args(0)<>1 then
    error("wrong number of arguments")
  end_if;

  case type(x)
    of DOM_SET do
    of "_union" do
      return(map(x, dilog))
  end_case;
  
  if not testtype(x,Type::Arithmetical) then
    error("argument must be of 'Type::Arithmetical'")
  end_if;

  case domtype(x)
    of DOM_FLOAT do
      return(slot(dilog,"float")(x)):
    of DOM_COMPLEX do
      if type(op(x,1)) = DOM_FLOAT or type(op(x,2)) = DOM_FLOAT then
        return(slot(dilog,"float")(x)  ):
      end_if:
  end_case;

  if testtype(x, Type::Numeric) then
  // Simplifications for exact numeric input: only return symbolic
  // calls with arguments in the right semi-circle Re(x)>=0, abs(x)<1
    fx:= float(x):
    if op(fx,1)<0  // i.e., Re(x) < 0
    then // reflect from Re(x)<0 to Re(x)>1
         return(PI^2/6 - dilog(1-x) - ln(x)*ln(1-x))
    end_if:
    if specfunc::abs(fx) > 1 
    then // use Landen-identity to reduce abs(x)>1 to abs(x)< 1
         // Note that this identity is valid for all arguments,
         // but has an exceptional form on the left real semi-axis:
         if testtype(x, Type::Real) and x < 0 
         then // this case should not arise, because Re(x)<0)
              // is reflected to Re(x)>1 above
              return(-dilog(1/x) - ln(x)^2/2 - 2*PI*I*ln(1-1/x) )
         else return(-dilog(1/x) - ln(x)^2/2 )
         end_if;
    end_if:
  end_if:

  procname(x)
end_proc:

dilog(0):=PI^2/6:
dilog(1/2):=PI^2/12-ln(2)^2/2:
dilog(1):=0:
dilog(2):=-PI^2/12:
dilog(infinity):=-infinity:
dilog(I):= PI^2/16 - I*CATALAN - I*PI/4*ln(2):
dilog(-I):= PI^2/16 + I*CATALAN + I*PI/4*ln(2):
dilog(1-I):= -PI^2/48 + I*CATALAN:
dilog(1+I):= -PI^2/48 - I*CATALAN:

dilog:=funcenv(dilog):
dilog:=slot(dilog,"print","dilog"):
dilog:=slot(dilog,"info","dilog -- the dilogarithm function dilog(x) = int(ln(t)/(1-t),t=1..x)"):
dilog:=slot(dilog,"type","dilog"):

dilog:=slot(dilog,"diff",proc() begin
   op(args(1)); ln(%)*diff(%,args(2))/(1-%)
 end_proc ):

dilog:=slot(dilog, "complexDiscont", Dom::Interval(-infinity, 0)):
dilog:=slot(dilog, "realDiscont", {}):
dilog:=slot(dilog, "undefined", {}):

/* We only have a library routine for float(polylog). So
   use dilog(x) = polylog(2, 1-x), i.e. call
   polylog::float() provided by D. Duparc, 1998.05.31 */

dilog:= slot(dilog,"float",
proc(z)
begin
  z:= float(z);
  if not has({DOM_FLOAT,DOM_COMPLEX}, domtype(z))
     then return(hold(dilog)(z))
  end_if;
  if specfunc::abs(1-I-z)<1/2 or specfunc::abs(1+I-z)<1/2
  then return( float(PI^2/6-ln::float(z)*ln::float(1-z)) 
              -slot(polylog,"float")(2,z));
  else return(slot(polylog,"float")(2,1-z));
  end_if;
end_proc):

dilog:= slot(dilog,"series",
    loadproc(slot(dilog,"series"),pathname("SERIES"),"dilog")
):
