/* ------------------------------------------
Author: W. Oevel
1.st version: 16.10.2000

specfunc::inverf:  The inverse of the erf function on the real line:
y:= specfunc::inverf(x) is the solution of erf(y) = x.

Call(s):    specfunc::inverf(x)

Parameters: x - a numerical real value from the interval [-1, 1]

Return Value: for -1 < x < 1: a real floating point value
              for x =-1 or -1.0:  -infinity
              for x = 1 or  1.0:   infinity

Details: inverf is used to compute alpha-quantiles of the 
         statistical normal distribution

Warning: this is a utility function without interface and
         help page! It does not check its argument! So
         make sure that the calling function provides a
         *numerical* real x with 0 <= x <= 1.

Remark: works also for x > 1 and x < -1, if x is
        not too far away from 1 and -1, respectively.
        But there is no guarantee that the result makes sense!

Examples:
>>  specfunc::inverf(0)

                              0.0

>>  specfunc::inverf(PI/30 + sqrt(2)/3)

                         0.5654841632

>>  specfunc::inverf(1)

                           infinity

------------------------------------------ */

//--------------------------------------------------------------------
// Solve erf(y) = x.
// Some estimates for the solution close to x = 1:
//  For x = 1 - eps,  eps small:
//  x = erf(y) = 1 - 1/sqrt(PI) *exp(-y^2) * ( 1/y -1/2/y^3+3/4/y^5 + .. )
//  --> exp(y^2) = 1/sqrt(PI)/(1-x)* 1/y* (1 -1/2/y^2 + 3/4/y^4)
//  First approximations are given by:
//  y^2 = -ln(sqrt(PI)) -ln(1-x) - ln(y) + ln(1-1/2/y^2 +3/4/y^4 + .. )  
//--------------------------------------------------------------------

specfunc::inverf:= proc(x)
local fx, macheps, absx, sqrtPI, i, maxit, y, y0, dy;
save DIGITS;
begin
   if iszero(x-1) or iszero(x-float(1))
        then return(infinity);
   end_if;
   if iszero(x) then return(float(0)) end_if;
   if float(x)<0 then return(-specfunc::inverf(-x)) end_if;
   macheps:= float(10^(-DIGITS));
   DIGITS:= DIGITS+2;
   sqrtPI:= float(PI)^(1/2):
   fx:= float(x):
   absx:= specfunc::abs(fx):

   //---------------------------------------------------------------------
   if absx<0.7
   then // Find good starting point for Newton-Iteration, when |x| is small: 
        sqrtPI:= sqrtPI/2:
        x:= fx:
        y:=x*sqrtPI;
        y:=y*(1+y^2/3);
        if absx < 10*macheps then return(y) end_if;
        dy:= y;
        // do Newton search: 
        while specfunc::abs(dy) > macheps*specfunc::abs(y) do
          dy:= sqrtPI*(erf(y)-x)*exp(-y^2);
          y:= y-dy;
        end_while;
        return(y);
   end_if;

   //---------------------------------------------------------------------
   // Now x >= 0.7.
   //---------------------------------------------------------------------

   //---------------------------------------------------------------------
   // Find good starting point for Newton-Iteration, when |x-1| is small: 
   // Solve erfc(y) = 1 - x rather than erf(y) = x, because erfc(y)
   // is better conditioned than erf(y) for large y:
   //---------------------------------------------------------------------
   x:= float(1-x):  // i.e., 0 <= x <= 0.3
   if iszero(x) then return(infinity); end_if;
   absx:= specfunc::abs(x):
   if absx>0.2 then // for 0.2 < x <= 0.3
        y0:= (-ln::float(sqrtPI) - ln::float(x))^(1/2); 
   elif absx>0.01 then // for 0.01 < x <= 0.2
        y0:= -ln::float(sqrtPI) - ln::float(x); 
        y0:= (y0 - ln::float(y0)/2)^(1/2);
   else // for 0.0 < x < 0.01
        y0:= -ln::float(sqrtPI) - ln::float(x); 
        y0:= (y0 - ln::float(y0)/2 +ln::float(1-1/2/y0))^(1/2);
   end_if;
   
   //---------------------------------------------------------------------
   // try direct Newton search first to avoid the overhead
   // of calling numeric::realroot:
   y:= y0:
   dy:= y;
   sqrtPI:= sqrtPI/2:
   maxit:= 10;
   for i from 1 to maxit do
       dy:= sqrtPI*(x-erfc::float(y))*exp::float(y^2);
        y:= y-dy;
       if specfunc::abs(dy) <= macheps*specfunc::abs(y) then
          return(y);
       end_if;
   end_for:
   //---------------------------------------------------------------------
   // The Newton iteration did not converge after maxit steps.
   // Finally, try sure-fire bisectioning/quadratic interpolation
   // implemented by numeric::realroot.
   // For 0.7 <= x < 1 the solution y of erf(y) = x satisfies
   //                   0.85*y0 < y < 1.1*y0,
   // where y0 is the previous starting point for the Newton iteration.
   // With this bounding interval, numeric::realroot *must* succeed:
   y:= genident():
   numeric::realroot(erfc(y) - x, y = 0.85*y0 .. 1.1*y0); 
end_proc:
