// $Date: 2000/05/30 10:07:51 $  $Author: walter $   $Revision: 1.15 $ 

/*--
usage:

reg(data, model, v, p, <Symbolic>)
reg(data, model, v, p, StartingValues = p_init, <Symbolic>)

// The following syntax is obsolete, but still accepted,
// because TCI uses it:
reg("advanced", data, model, v, p, p_init, <Symbolic>)

nonlinear least-squares-fit of an arbitrary function
(nonlinear concerning independent variables and nonlinear concerning 
 the fit-parameters)
--*/
/*--
parameters:
data  data - this may be a sample with specified columns 
                      or a sequence of columns
                      or a list of rows.
             the number of columns depends on the number of 
             independent variables.
             the order of columns is:
             independent variables (x_i), dependent variable (y),
                      errors/standard deviations of the dependent 
                      variable (sigma_y)
             the sigma_y are optional
model        the function to fit: an expression
v            list of independent variables
p            list of fit-parameters
p_init       list of starting values for fit-parameters
Symbolic     this option prevents conversion of input data
             to floats. Only has an effect, if the regression
             is linear, i.e., if the model depends linearly
             on the fit-parameters
--*/

stats::reg := proc()
  local llll, x, y, fitf, fitderiv, xvar_, xvar, param, 
        linearRegression,
        N, n, i, j, k, m, p, a, deltaa, al, alpha, alphal,
        lambda, bet, betav, chis, chisqu, chisqu_old, w,
        minv, det, counter_out, counter_in, macheps,
        iter, conv, max_out, sigmay, nvars,
        dummy, dummy2; //, errout;
save DIGITS;

  /*
  x[i]          x-data
  y[i]          y-data
  fitf          fitting function
  fitderiv[i]  fitting function's derivatives
  xvar_         user's fitting variable
  xvar          symbolic fitting variable in fitting function
  param[i]      user's fitting parameters
  N             number of (x,y)-pairs
  n             number of fitting parameters
  i, j, k, m    indices
  p[i]          symbolic fitting parameters in fitting function
  a[i]          current values of the fitting parameters
  aout[i]       list of converged parameters
  errout[i]     list of errors of those parameters (sigma_aout)
  deltaa[i]     vector in parameter space (step to next parameter)
  al[i,j]       symbolic matrix alpha
  alpha[i,j]    matrix alpha
  alphal[i,j]   matrix alpha(lambda)
  lambda        marquard-factor
  bet[i]        symbolic gradient vector
  betav[i]      gradient vector
  chis          symbolic chisquare
  chisqu        chisquare
  w             weighting factor
  minv()        matrix inversion of a symmetric square matrix
  det           determinant of alphal
  counter_out   counter for outer loop
  counter_in    counter for inner loop
  macheps       10^(-DIGITS) = relative machine precision
  iter          FALSE if converged
  conv          number of converged parameters in current loop
  max_out       maximum number of outer loops
  sigmay        TRUE if x-y-data includes sigma_y
  */

begin

  if args(0) < 4 then
     //need:  data, model, variables, parameters
     error("expecting at least four arguments");
  end_if:

  DIGITS:= DIGITS + 2;


  // --------------------------------------------------------
  // utility function to invert matrix and calculate det 

  minv:=proc()
    local i, j, ol, ii, ji, i_ol, j_ol, h, maxentry, store;
    // alphal[i,j], n, det are treated global 
  begin
    // outer loop 
    det:=1;
    for ol from 1 to n do
      i_ol:=-1;
      j_ol:=-1;

      // order matrix (get largest elements in diagonal) 

      while j_ol < 0 do
        while i_ol < 0 do

          // find largest entry in remaining matrix 
          maxentry:=0;
          for i from ol to n do
            for j from ol to n do
              if abs(alphal[i,j]) >= abs(maxentry) then  
                maxentry:=alphal[i,j];
                ii[ol]:=i;
                ji[ol]:=j;
              end_if;
            end_for
          end_for;

          if iszero(maxentry) then
            det:=0; 
            alphal:=FAIL; 
            return([]); 
          end_if;

          i:=ii[ol];
          i_ol:=i - ol;
        end_while;

        // move largest entry to alphal[ol, ol] 
        if i_ol > 0 then
          for j from 1 to n do
            store:=alphal[ol, j];
            alphal[ol, j]:=alphal[i, j];
            alphal[i, j]:=-store;
          end_for;
        end_if;

        j:=ji[ol];
        j_ol:=j - ol;
      end_while;

      if j_ol > 0 then
        for i from 1 to n do
          store:=alphal[i, ol];
          alphal[i, ol]:=alphal[i, j];
          alphal[i, j]:=-store;
        end_for;
      end_if;

      // inversion of ordered matrix 

      for i from 1 to n do
        if _not(i = ol) then
          alphal[i, ol]:=-alphal[i, ol]/maxentry
        end_if
      end_for;

      for i from 1 to n do
        for j from 1 to n do
          if _not(i = ol) then
            if _not(j = ol) then
              alphal[i, j]:=alphal[i, j] + alphal[i, ol]*alphal[ol, j]
            end_if
          end_if
        end_for
      end_for;

      for j from 1 to n do
        if _not(j = ol) then
          alphal[ol, j]:=alphal[ol, j]/maxentry
        end_if
      end_for;

      det:=det*maxentry;
      alphal[ol, ol]:=1/maxentry;
    end_for;
    // end of outer loop 

    // restore original arrangement of entries 
    for h from 1 to n do
      ol:= n - h + 1;

      j:=ii[ol];
      if ol < j then
        for i from 1 to n do
          store:=alphal[i, ol];
          alphal[i, ol]:=-alphal[i, j];
          alphal[i, j]:=store
        end_for
      end_if;

      i:=ji[ol];
      if ol < i then
        for j from 1 to n do
          store:=alphal[ol, j];
          alphal[ol, j]:=-alphal[i, j];
          alphal[i, j]:=store
        end_for
      end_if;

    end_for
  end_proc;
  // ------------- end of utility minv ---------------------------

  // -------------------------------------------------------------
  // Starting values provided by the user ?
  // If not, set default StartingValues = [1.0, 1.0, ..]
  // Note that StartingValues = [0.0, 0.0, ..] would be
  // a rather unfortunate choice, because for model functions
  // with polynomial dependence on the fit parameters the
  // curvature matrix will be singular, if all parameters 
  // vanish!

  //-----------------------------------
  // obsolete input syntax, used by TCI.
  if args(1) = "advanced" then
    return(stats::reg(args(2..args(0)-1), StartingValues = args(args(0))));
  end_if;
  //-----------------------------------

  if not has([args(args(0))], StartingValues) then
    return(stats::reg(args(), StartingValues = [float(1) $ nops(args(args(0)))]))
  end_if;

  // -------------------------------------------------------------
  // now we can assume that starting values
  // for the fit parameters are provided.

  //----------------------------------------------------------------------
  // check the arguments

  if testargs() then
    if domtype(args(args(0)-3)) <> DOM_EXPR then 
      error("expecting a model function of type DOM_EXPR")
    elif domtype(args(args(0)-2)) <> DOM_LIST then
      error("expecting independent variables provided as a list")
    elif domtype(args(args(0)-1)) <> DOM_LIST then
      error("expecting fit-parameters provided as a list")
    elif type(args(args(0))) <> "_equal" or
         op(args(args(0)), 1)<> StartingValues or
         domtype(op(args(args(0)), 2))<> DOM_LIST then
      error("expecting starting values for the fit-parameters provided as a list")
    end_if
  end_if;

  //----------------------------------------------------------------------
  // get the arguments

  fitf :=args(args(0)-3);// the model function
  xvar_:=args(args(0)-2);// the independent variables
  param:=args(args(0)-1);// the fit parameters
  a:=op(args(args(0)),2);// starting values for the fit parameters 
  a:= map(a, float);
  nvars:=nops(xvar_);    // number of variables
  n:=nops(param);        // number of fit parameters

  //----------------------------------------------------------------------
  // further check of the arguments

  if testargs() then
     map(xvar_, proc(x)
            name stats::reg::checkVariables;
            begin 
              if domtype(x) <> DOM_IDENT and type(x) <> "_index" then
                 error("Expecting symbolic variables"):
              end_if;
            end_proc);
     map(param, proc(p)
            name stats::reg::checkParameters;
            begin 
              if domtype(p) <> DOM_IDENT and type(p) <> "_index" then
                 error("Expecting symbolic parameters"):
              end_if;
            end_proc);
     map(a, proc(a)
            name stats::reg::checkStartingValues;
            begin 
              if domtype(a) <> DOM_FLOAT and domtype(a) <> DOM_COMPLEX then
                 error("Starting values are not numerical"):
              end_if;
            end_proc);
   end_if;

  //----------------------------------------------------------------------
  // First try to get the data: use FALSE as first argument: Note that
  // an intentional error occurs, if weights are provided.
  // This intentional error should not lead to abortion.

  llll:=stats::getdata(FALSE, "numeric_only", nvars+1, args(1..args(0)-4));
  if domtype(llll) = DOM_STRING then  // an error occurred, because weights are provided
       llll:=stats::getdata(FALSE, "numeric_only", nvars+2, args(1..args(0)-4));
       sigmay:=TRUE;
  else sigmay:=FALSE;
  end_if;
  if domtype(llll) = DOM_STRING then error(llll) end_if;
  
  N:=nops(llll[1]);      // the number of sample points

  //----------------------------------------------------------------------
  // First check, whether the equations for the parameters are degenerate.
  // This is definitely the case, if N < n or if the model function does
  // not contain one of the parameters:

  if N < n then 
     error("need as least as many sample points as parameters") 
  end_if;
  for i from 1 to n do
    if iszero((diff(fitf, op(param, i))))
      then error("model function does not depend on parameter ".expr2text(op(param,i)))
    end_if
  end_for;

  //----------------------------------------------------------------------
  // get the data and convert to floats

  for i from 1 to nvars do
    x[i]:=map(llll[i], float); // x[i] = list of sample points for the i-th variable
  end_for;
  y:=map(llll[nvars+1],float); // y = list of sample points for the dependent variable
  if sigmay then 
       w:= map(llll[nvars+2], float); // w = list of weights
  else w:= [float(1) $ N];
  end_if;

  //----------------------------------------------------------------------
  // Substitute the symbols xvar_ and param provided by the user
  // by our own names xvar and p, respectively:

  xvar:=genident("X");
  p:=genident("p");
  fitf:=subs(fitf, [xvar_[i]=xvar[i] $i=1..nvars, op(param, i)=p[i] $i=1..n]);

  //---------------------------------------------------------
  // calculate all partial derivatives fitderiv[i] of function to fit.

  fitderiv:= [diff(fitf, p[i]) $ i=1..n];

  //---------------------------------------------------------
  // Check, whether the regression is linear, 
  // i.e., whether fitf is linear in the parameters,
  // i.e., whether fitderiv does not contain p

  if has(fitderiv, p) then
       linearRegression:= FALSE;
  else linearRegression:= TRUE;
  end_if;

  //---------------------------------------------------------
  // special implementation for linear regression. Need to
  // solve a linear system of equations only *once*. Avoid the
  // iterative process for non-linear systems below:

  if linearRegression then

     userinfo(1, "using special routine for linear regression");
     bet:= [0 $ n];
     for i from 1 to N do
         dummy:= eval(subs(fitderiv, [xvar[m]=x[m][i] $m=1..nvars]));
         dummy2:=eval(subs(fitf,     [xvar[m]=x[m][i] $m=1..nvars]));
         for k from 1 to n do
           bet[k]:=bet[k] + w[i]* ((y[i] - dummy2)*dummy[k]);
         end_for;
     end_for;
     a:=  numeric::linsolve(bet, [p[i] $ i=1..n]);
     if a = FAIL or  // matrix not invertible
        nops(a) <> n // rank of matrix was not maximal
        then
        return(FAIL)
     end_if;

     // compute quadratic error
     // Substitute optimized parameters into model function
     fitf:= eval(subs(fitf, a));
     chis:=float(0);
     for i from 1 to N do
       // substitute sample data into model function
       dummy2:=eval(subs(fitf, [xvar[m]=x[m][i] $m=1..nvars]));
       chis:=chis + float(w[i]*abs(y[i] - dummy2)^2);
     end_for;

     // convert [p[1]= value1, p[2]= value2, ..] to [value1, value2, ..]:
     a:= map(a, op, 2);

     return([a, chis]);

  end_if: // end of linear regression

  //---------------------------------------------------------
  // Now do the non-linear case:
  //---------------------------------------------------------
  // calculate symbolic matrix al[i,j] 
  // and symbolic gradient bet[i,j] 

  for k from 1 to n do
    bet[k]:=0;
    for j from 1 to n do
        al[j, k]:=0
    end_for;
  end_for;
  for i from 1 to N do
    dummy2:=eval(subs(fitf, [xvar[m]=x[m][i] $m=1..nvars]));
    dummy:= eval(subs(fitderiv, [xvar[m]=x[m][i] $m=1..nvars]));
    
    for k from 1 to n do
      bet[k]:=bet[k] + w[i]* ( (y[i] - dummy2)*dummy[k] );
      for j from 1 to n do
        al[j, k]:=al[j, k] + w[i]*eval(dummy[j]*dummy[k]);
      end_for
    end_for
  end_for;
  //---------------------------------------------------------
  // calculate symbolic chisquare chis 

  chis:=0;
  for i from 1 to N do
    // Substitute sample data into model function.
    // (the parameters remain symbolically)
    dummy2:=eval(subs(fitf, [xvar[m]=x[m][i] $m=1..nvars]));
    chis:=chis + w[i]*abs(y[i] - dummy2)^2;
  end_for;

  //---------------------------------------------------------
  // calculate initial chisquare(a) 

  chisqu_old:=eval(subs( chis, [p[i]=a[i] $i=1..n]));
  //chisqu_old:=Re(float(chisqu_old)); // this is vital for dealing with trigonometric functions 
  chisqu_old:=float(chisqu_old);
  if chisqu_old > float(0) then 
    chisqu:=2*chisqu_old
  elif iszero(chisqu_old) then 
    chisqu:=float(10);
  else
    return(FAIL) // illegal starting parameters 
  end_if;

  //---------------------------------------------------------
  // Do the numerical search:
  macheps:= 10.0^(2-DIGITS):
  max_out:=500;
  iter:=TRUE;
  counter_out:=0;
  lambda:=0.001;

  /** begin outer loop (searching parameter space) **/

  while (iter and counter_out < max_out) do

    counter_out:=counter_out+1;

    // calculate alpha and betav 

    alpha:=eval(subs(al,  [p[i]=a[i] $i=1..n]));
    betav:=eval(subs(bet, [p[i]=a[i] $i=1..n]));
    // alpha:=map(map(alpha, float), Re);
    alpha:=map(alpha, float);
    //betav:=map(map(betav, float), Re);
    betav:=map(betav, float);

    counter_in:=0;

    /** begin inner loop (computing step in parameter space (deltaa)) **/

    while chisqu > chisqu_old do

      counter_in:=counter_in+1;
      if counter_in > 500 then 
        userinfo(1, "no convergence in inner loop!");
        return(FAIL)
      end_if;

      // alphal(lambda) 

      alphal:=alpha;
      for i from 1 to n do
        alphal[i,i]:=(1 + lambda)*alphal[i,i]
      end_for;

      // invert alphal(lambda) 

      minv();
      if iszero(det) then 
        userinfo(1, "matrix of curvature is singular!");
        return(FAIL)
      end_if;

      // calculate deltaa 

      for j from 1 to n do
        deltaa[j]:=0;
        for k from 1 to n do
          deltaa[j]:=deltaa[j] + betav[k]*alphal[j, k];
        end_for
      end_for;

      // calculate chisquare(a+deltaa) 

      chisqu:=eval(subs( chis, [p[i]=a[i]+deltaa[i] $i=1..n]));
      // chisqu:=Re(float(chisqu));
      chisqu:=float(chisqu);

      if chisqu > chisqu_old then 
        lambda:=lambda*10
      end_if;

    end_while;

    /** end inner loop **/

    // check convergence 

    conv:=0;
    for i from 1 to n do
      if specfunc::abs(deltaa[i]) <= macheps*(1+specfunc::abs(a[i])) then 
        conv:=conv+1
      end_if
    end_for;
    if conv = n then
      iter:=FALSE
    end_if;

    lambda:=lambda/10;
    chisqu_old:=chisqu;
    chisqu:=chisqu+10;
    for i from 1 to n do
      a[i]:=a[i]+deltaa[i]
    end_for;
  end_while;

  /** ---------------- end outer loop ------------------ **/

  if counter_out = max_out then 
    userinfo(1, "poor convergence!");
    return(FAIL);
  end_if;
  if chisqu_old < 0 then return(FAIL) end_if;

  /* ------ totgeschaltet ------------------------------------
  // // calculate alphal(lambda=0) and invert ( => error matrix )
  //
  // alphal:=eval(subs( al, [p[i]=a[i] $i=1..n]));
  // alphal:=map(map(alphal, float), Re);
  // 
  //  minv();
  //  if iszero(det) then 
  //     errout:=FAIL;
  //  else
  //    errout:=[0$n];
  //    for i from 1 to n do
  //      errout[i]:=sqrt(alphal[i, i]);
  //    end_for;
  //  end_if;
  // 
  //  if errout <> FAIL then
  //    if not iszero(_plus(op(map(map(errout, float), Im)))) then
  //      return(FAIL);
  //    end_if
  //  end_if;
  //
  // // convert table a to list aout:
  //
  // aout:= [0 $ n];
  // for i from 1 to n do
  //   aout[i]:=a[i];
  // end_for;

  // // errout is the list of uncertainties for the fit parameters 
  // return([aout, errout, chisqu_old]);
  -------------------------------------------------------- */

  return([a, chisqu_old]);

end_proc:

// end of file 
