// $Date: 2000/09/16 19:35:43 $ $Author: jngerhar $ $Revision: 1.15 $ 

/*--
	ln/simplify -- the function attribut "simplify" for ln

	Rules (take properties into account):

	1) ln( posint1/posint2 * r )  -->  ln(posint1) - ln(posint2) + ln(r)
	2) ln( exp(real) * r )        -->  real + ln(r)
        3) ln( positive^real * r )    -->  real*ln(positive) + ln(r)
           ln( negative^even * r )    -->  even*ln(-negative) + ln(r)
           ln( negative^odd  * r )    -->   odd*ln(-negative) + ln(-r)
        4) ln( positive * r )         -->  ln(positive) + ln(r)
           ln( negative * r )         -->  ln(-negative) + ln(-r)

        5) ln( y^n )                  -->  n*ln(y) if
                                           -1 < n <= 1 or
                                           n = -1 and is(y <= 0) = FALSE or
                                           -2 <= n <= 2 and Re(y) > 0
        6) ln( exp(z) )               -->  z if -PI < Im(z) <= PI

        The ideology is to have only positive arguments in ln's where
        possible.
--*/

ln := slot(ln, "simplify", proc(x: "ln")
  name ln::simplify;
  local l, t, r, k, z, y, n;
begin

  // recursively simplify the argument of ln
  l := simplify(op(x));

  // convert the result into a list of factors
  if type(l) = "_mult" then
    l := [op(l)];
  else
    l := [l];
  end_if;

  // determine all factors that are recognized as positive or negative by "is"
  l := split(l, proc(a) begin
      is(a > 0) or is(a < 0)
  end_proc);

  r := op(l[2])*op(l[3]);
  l := l[1];
  // now x = ln(_mult(op(l)) * r), and every element of l "is" either >0 or <0

  // handle positive or negative factors
  k := 0;
  for t in l do
    case type(t)

    of DOM_RAT do // rule 1)
      // ln(numer/denom) = ln(numer) - ln(denom)
      if t < 0 then
        t := -t;
        r := -r;
      end_if;
      k := k + ln(op(t, 1)) - ln(op(t, 2));
      break;

    of "exp" do // rule 2)
      if is(op(t), Type::Real) = TRUE then
	// t "is" positive and ln(t) = op(t)
        k := k + op(t);
      else
        r := r * t;
      end_if;
      break;

    of "_power" do // rule 3)
      [y, n] := [op(t)];
      if is(y > 0) = TRUE then
        // ln(y^n * r) = n*ln(y) + ln(r) if y > 0 and n is real
        if is(n, Type::Real) = TRUE then
          k := k + n*ln(y);
        else
          r := r * t;
        end_if;
      elif is(y < 0) = TRUE then
        // ln(y^n * r) = n*ln(-y) + ln(r) if y < 0 and n is even
        // ln(y^n * r) = n*ln(-y) + ln(-r) if y < 0 and n is odd
	if is(n, Type::Even) = TRUE then
	  k := k + n*ln(-y);
        elif is(n, Type::Odd) = TRUE then
	  k := k + n*ln(-y);
          r := -r;
        else
          r := r * t;
        end_if;
      else
        r := r * t;
      end_if;
      break;
      
    otherwise // rule 4)
      if is(t < 0) = TRUE then
        t := -t;
        r := -r;
      end_if;
      k := k + ln(t);

    end_case;
  end_for;
  // now x = k + ln(r)

  // handle the remainder ln(r)
  case type(r)

  of "_power" do // rule 5
    [y, n] := [op(r)];
    /* ln(y^n)=n*ln(y) when -1< n <=1
       and for n = -1 if y is not real nonpositive
       and for -2 <= n <= 2 if Re(y) > 0
    */
    if (is(-1 < n) = TRUE and is(n <= 1) = TRUE) or
       (is(n = -1) = TRUE and is(y <= 0) = FALSE) or
       (is(-2 <= n) = TRUE and is(n <= 2) = TRUE and is(Re(y) > 0) = TRUE) then
       return(k + n*ln(y));
    end_if;
    break;

  of "exp" do // rule 6
    // ln(exp(z)) = z if -PI < Im z <= PI
    z := op(r);
    if (is(Im(z) > -PI) = TRUE and is(Im(z) <= PI) = TRUE) or
       contains({-1, 0, 1}, sign(z)) then
      return(k + z);
    end_if;

  end_case;

  return(k + ln(r));

end_proc):

// old code, now obsolete, to be removed

//ln := slot(ln, "simplify",
( proc(x: "ln")
//    name ln::simplify;
    local t, k, _y, _x, expo;
 begin
   _x:= op(x);
   case type(_x)
     of DOM_RAT do
       return(ln(op(_x,1))-ln(op(_x,2)))
     of "_power" do
	/* ln(x^n)=n*ln(x) when x is real nonnegative
          or if n=1/2, because if x = r*exp(I*t) with r>=0 
	  and -PI<t<=PI, then x^(1/2)=sqrt(r)*exp(I*t/2)
          and ln(x^(1/2))=1/2*ln(r)+I*t/2 = ln(x)/2.

          This is true in general for -1<n<1
          and for n=-1 if x is not real nonnegative 
	*/
       if (is(op(_x,1)>=0)=TRUE and is(op(_x,2), Type::Real)=TRUE) or
         (is(-1<op(_x,2))=TRUE and is(op(_x,2)<1)=TRUE) or
         (bool(op(_x,2)=-1) and is(op(_x,1)<0)=FALSE)
         then
         return(
                simplify(op(_x,2)) *
                simplify(ln(simplify(op(_x,1))))
                )
       end_if;
      break
   of "_mult"  do
      t:=simplify(_x);
      if type(t)="_mult" then
	    /* CAUTION: ln(a*b) is not always ln(a)+ln(b),
              for example if a=b=-1, ln(a*b)=0 but ln(a)+ln(b)=2*I*PI */
            // split ln(t) = ln( positive * unknown), where positive is real positive
            // Return ln(positive*unknown) = ln(positive) + ln(unknown);
            k:= 0;  // k = ln(positive)
            for _y in t do
               if is(_y>0)=TRUE then
                     k:= k + simplify(ln(_y));
                     t:=t/_y
               elif is(_y<0)=TRUE then 
                     // do not always use simplify(ln(-y)) here, because
                     // this can lead to an infinite recursion, e.g.
                     // assume(x<0): simplify(ln(-x));
                     if type(-_y) <> "_mult" then
                          k:= k + simplify(ln(-_y))
                     elif type(_y) = "_power" then
                          // simplify ln(-1/y^p) with assume(y < 0):
                          expo:= op(_y, 2);
                          if is(expo, Type::Odd) = TRUE and
                             is(expo < 0)=TRUE then
                             // ln( 1/y^(-expo) * rest)
                             //=ln( (-y)^expo * (-rest))
                             //= expo*ln(-y) + ln(-rest);
                             k:= k + expo*ln(-op(_y,1))
                          else
                             k:= k + ln(-_y);
                          end_if:
                     else k:= k + ln(-_y);
                     end_if:
                     t:=t/(-_y)
               end_if
            end_for;
            if k <> 0 then
                 return(k + ln(t));
            else return(ln(t))
            end_if;
       elif t <> _x then
            return( simplify(ln(t)) )
       else
            return( x )
       end_if
    of "exp" do 
	/* ln(exp(x))=x when x is real 
	  and more generally when -PI < Im x <= PI
	*/
       t:= op(_x,1);
	if (is( t,Type::Real ) = TRUE) 
            or (is(Im(t) > -PI) = TRUE and is(Im(t) <= PI) = TRUE) then
	    return( t ) 
	else
	    k:= round( -Im(t)/2/PI );
       if type(k) = DOM_INT then 
		    return( t+2*k*PI*I )
       elif contains({-1,0,1},sign(t)) then
		    return( t )
	    end_if
	end_if;
 end_case;

 t:= simplify(_x);
 if is(t<0) = TRUE then
    return(I*PI + simplify(ln(-t))) 
 end_if;
 if t <> _x then
   t:= ln(t);
   if type(t) = "ln" then
     slot(ln,"simplify")(t)
   else
     t
   end_if
 else
   x
 end_if
end_proc):
