// $Date: 2000/11/30 11:48:50 $ $Author: stefanw $ $Revision: 1.35 $ 

/* zimmerma, 01/08/95 */

/*

	discont  --  determines the discontinuities of a function


	discont(f,x, F) returns a set of (possible) discontinuities 
	of f, where f is regarded as a function over F.

	discont(f,x=a..b, F) returns the discontinuities in a..b, where a<=b.

   F may be either Dom::Real or Dom::Complex.

   If F is not given, Dom::Complex is used unless is(Global,
   Type::Real) is true. 


	discont looks for an attribute "realDiscont" (or "complexDiscont",
   respectively) of the operator of f in order
   to determine the discontinuities of f viewed as a real
(resp. complex) function. 
    

	For example to say that the only 
	discontinuity of dirac is in 0, you write 
	dirac:=slot(dirac,"realDiscont",{0})


   discont(f, x, undefined), or discont(f, x=a..b, undefined)
   returns the set of all complex numbers (or all numbers
   in the interval [a,b], respectively) for which f is undefined.

   In this case, discont looks for an attribute "undefined" of the operator.
   


*/

alias(Z=solvelib::BasicSet(Dom::Integer)):

discont :=
proc(f,x, F)
  local i, s, r;
begin

  //  overloading for domains 
  if f::dom::discont <> FAIL then
    return(f::dom::discont(args()))
  end_if;

  /* main */

  if testargs() then
    if args(0)<2 then
      error("Two or three arguments expected")
    end_if;
    case type(x)
      of DOM_IDENT do break; // ok 
      of "_equal" do
        if map(x,type)<>(DOM_IDENT="_range") then
          error("x=a..b expected as 2nd argument")
        end_if;
        break
      otherwise
        error("invalid second argument")
    end_case;
  end_if;

  if args(0)=3 then
    case F
      of undefined do  
      of Dom::Real do
      of Dom::Complex do
        break;
      of R_ do
        F:= Dom::Real; break
      of C_ do
        F:= Dom::Complex; break
      otherwise  
        error("Third argument must be Dom::Real or Dom::Complex")
    end_case
  elif args(0)<>2 then
    error("Two or three arguments expected")
  else
    // third argument not specified
    if is(hold(Global), Type::Real)=TRUE then
      F:=Dom::Real
    else
      F:=Dom::Complex
    end_if
  end_if;

  // handle case that a certain range is given
  
  if type(x)="_equal" then
    r:=op(x,2);
    case type((s:=discont(f,op(x,1), F)))
      of piecewise do
        userinfo(10, "Discontinuities depend on parameters: ".
                 expr2text(s));
        return(map(s,_intersect, Dom::Interval([op(r,1), op(r,2)]) ))
      of "solve" do
        return(procname(args()))
      of "_union" do
        return(map(s, _intersect, Dom::Interval([op(r,1), op(r,2)])))
      otherwise
        userinfo(10, "The set of discontinuities is ".expr2text(s));
        return(s intersect Dom::Interval([op(r,1), op(r,2)]))
    end_case 
  end_if;
  
 // overloading for function environments 
  if type((s:=eval(op(f,0))))=DOM_FUNC_ENV then
    case F
      of Dom::Real do
        s::realDiscont; break
      of Dom::Complex do
        s::complexDiscont; break
      of undefined do
        s::undefined
    end_case;
    
    if %<>FAIL then
        return(solvelib::preImage(op(f), x, %)
               union discont(op(f),x, F))
    end_if;
  end_if;


  // do not handle RootOf's
  // (would cause problems in property mechanism)
  
  if hastype(f, "RootOf") then
    return(procname(args()))
  end_if;
  
  
  case type(f)
    of DOM_INT do
    of DOM_RAT do
    of DOM_IDENT do
    of DOM_COMPLEX do
    of DOM_FLOAT do
      return({})
    of "_index" do
      return((if not has(f,x) then return({}) else FAIL end_if))
    of "_plus" do
//  of "igamma" do   // igamma has singular points and branch cuts
    of "_mult" do
      return(_union(discont(op(f,i),x, F)$i=1..nops(f)))
    of "_power" do

       // special case F=undefined
       // f(x)^g(x) is undefined iff f(x) or g(x) is undefined
       // or f(x) is zero and g(x) is negative
       // for efficiency reasons, we handle the case g(x)=const separately
      if F=undefined then
        if not has(op(f,2), x) then
          return(piecewise([op(f,2)>=0, discont(op(f,1), x, F)],
                           [op(f,2)<0, discont(op(f,1), x, F) union
                            solve(op(f,1), x)]))
        else
          // x occurs in the exponent
          // we have to call solve in oirder to decide
          // exponent >=0 / exponent <0
          return(
                 solve(op(f, 2)<0, x) intersect
                            solve(op(f,1), x)
                 union discont(op(f,2), x, F)
                 union discont(op(f,1), x, F)
                 )
        end_if
      end_if;
      
       // f=a^b
       // special cases for b integer and independent of x
      if not has(op(f,2), x) and is(op(f,2), Type::Integer)=TRUE
        then
         // x -> x^b is everywhere continous for b>0, and has
         // its only discontinuity at zero for b<=0
        return(piecewise([op(f,2)>0, discont(op(f,1), x, F)],
                         [op(f,2)<=0, discont(op(f,1), x, F) union
                          solve(op(f,1), x)]))
      end_if;
      
      // special case: real base, b independent of x
      // then, as a function of the reals, x -> x^b
      // is everywhere continous for b>0, and discontinous only
      // at zero for b<=0
      if F=Dom::Real and not has(op(f,2), x) then
        // test whether base is real
        i:= genident();
        eval(hold(assume)(i, Type::Real));
        if is(subs(op(f, 1), x=i), Type::Real)=TRUE then
          return(piecewise([op(f,2)>0, discont(op(f,1), x, F)],
                         [op(f,2)<=0, discont(op(f,1), x, F) union
                          solve(op(f,1), x)]))
        end_if;
      end_if;
        
       // = exp(b*ln(a)) if a is not zero
       // discontinuities are those of b and those of ln(a)
      return(discont(op(f,2), x, F) union discont(op(f,1), x, F)
             union solvelib::preImage(op(f,1), x, ln::complexDiscont))
       // functions without singularities 
     of "besselJ" do
       f:=[op(f,2)];
     of "sin" do
     of "cos" do
     of "abs" do
     of "exp" do
     of "sinh" do
     of "cosh" do
     of "erf" do
    of "erfc" do
    of "Re" do
    of "Im" do  
       return(discont(op(f), x, F))
   // functions with a discontinuity in 0 
     of "besselY" do
       f:=[op(f,2)];
       // fall through
     of "dirac" do
     of "heaviside" do
    of "sign" do
      if F=undefined then
        return(discont(op(f), x, F))
      else  
        return(discont(op(f),x, F) union solve(op(f),x))
      end_if;
   // functions with a discontinuity in 1 
     of "zeta" do
       return(discont(op(f),x, F) union solve(op(f)=1,x))
       // functions with discontinuities in all integers
       // to be corrected for F=Dom::Complex !!!
     of "floor" do
     of "frac" do
     of "ceil" do
    of "trunc" do
      if F=undefined then
        return(discont(op(f), x, F))
      else
        return(discont(op(f), x, F) union
               solvelib::preImage(op(f), x, Z))
      end_if
    of "round" do
      if F=undefined then
        return(discont(op(f), x, F))
      else         
        i:=genident();
        return(discont(op(f), x, F) union
              solvelib::preImage(op(f), x, Dom::ImageSet(i+1/2, i, Z)))
      end_if
   // functions with a discontinuity in -n 
     of "psi" do
      return(discont(op(f),x, F) union
             solvelib::preImage(op(f), x,
                                Z intersect Dom::Interval(-infinity..[0])))
     of "fact" do
       return(discont(rewrite(f,gamma),x, F))
   // functions with a discontinuity in k*PI 
     of "cot" do
       i:= genident("i");
       return(discont(op(f),x, F) union
             solvelib::preImage(op(f),x ,
             Dom::ImageSet(i*PI,i, Z)))
   // functions with a discontinuity in PI/2+k*PI 
     of "tan" do
       i:= genident("i");
       return(discont(op(f),x, F) union
              solvelib::preImage(op(f), x,
              Dom::ImageSet(PI/2+i*PI,i, Z)))
   // logarithm: rewrite by ln
   // assume the base to be positive and not equal 1
   // either the base is constant and positive anyway, or it is an identifier
   // in the latter case, replace with a positive identifier   
    of "log" do
      if type(op(f,1))=DOM_IDENT then
        i:=genident("b");
        eval(hold(assume)(i>0));
      // einkommentieren, wenn property-Problem sich erledigt hat:  
      //  eval(hold(assume)(i<>1, hold(_and)));
        return(subs
               (
                discont(rewrite(subs(f, op(f,1)=i),ln), subs(x, op(f,1)=i),F),
                i=op(f,1)))
      else  
        return(discont(rewrite(f, ln), x, F))
      end_if;
   // integrals    
     of "int" do
       return(discont(op(f,1),x, F) union
              (if type(op(f,2))="_equal" then
                 discont(op(f,[2,2,1]),x, F)
                 union discont(op(f,[2,2,2]),x, F)
               else
                 {}
               end_if))
   end_case;

  procname(args())
end_proc:

// end of file 
