// $Date: 2000/05/27 16:34:27 $  $Author: frankp $  $Revision: 1.28 $

/*++
    factor -- return the factorization of multivariate polynomial in Z or
              factorization of univariate polynomial in Zp, output is given in form
              [u, f1, e1, ..., fn, en], (p=u*f1^e1*...*fn^en, u is the constant
              content), fi has the same form of input p. 

    factor(p)

    p - an expression or a polynomial
++*/

factor:= proc(p)
  local _p_, f, Factors, ff, i, pp, q, t, xx, Float, factored;
begin
    if args(0) <> 1 then
      error("one argument expected");
    end_if;
 
    // test the type of the input expression, and
    // change the suitable one to polynomial form  
    case domtype(p)
    of DOM_EXPR do
      if hastype(p, DOM_FLOAT, {DOM_COMPLEX}) then
        _p_:= numeric::rationalize( p, Restore );
        f:= faclib::factor_expr( _p_, x -> Factored::convert_to(factor(x),DOM_LIST) );
        (f[2*i]:= float( f[2*i])) $ i=1..nops(f) div 2
      else
        f:=faclib::factor_expr( p, x -> Factored::convert_to(factor(x),DOM_LIST) )
      end_if;
      return(Factored::create( f,"irreducible" ))
    of DOM_POLY do
      // the result type of the factorization:
      factored:= l -> Factored::create( l,"irreducible" );

      if degree(p) = 0 then
        return(factored( [coeff(p)] ))
      end_if;

      // remove indets which have 0 degree
      xx:= op(p,2);
      f:= select(xx, x -> degree(p,x)>0 );
      if f <> xx then
        p:= poly( p,f );
        _p_:= Factored::convert_to( factor(p),DOM_LIST );
        return(factored( 
            [_p_[1], (poly(_p_[2*i],xx), _p_[2*i+1]) $ i=1..(nops(_p_) div 2) ]
        ))
      end_if;
      // replace indets that are not in DOM_IDENT
      f:= select(xx,v -> domtype(v) <> DOM_IDENT );
      if f <> [] then
        f:=[(f[i]=genident())$i=1..nops(f)];
        p:=poly(subs(op(p,1..3),f));
        _p_:=Factored::convert_to( factor(p),DOM_LIST );
        f:=[(op(f[i],2)=op(f[i],1))$i=1..nops(f)];
        return(factored( subs(_p_,f) ))
      end_if;
      if nterms(p)=1 then
        return(factored( [lcoeff(p),(poly(xx[i],op(p,2..3)),degree(p,xx[i]))$ i=1..nops(xx)] ))
      end_if;
      break;
    of DOM_INT do
      return(Factored::create( stdlib::ifactor(p),"irreducible",Dom::Integer ))
    of DOM_RAT do
      f:= stdlib::ifactor(denom(p));
      return(Factored::create(
        stdlib::ifactor(numer(p)).[(op(f,2*i), -op(f,2*i+1)) $ i=1..nops(f) div 2],
        "irreducible", Dom::Integer
      ))
    of DOM_COMPLEX do
      // the result type of the factorization:
      factored:= l -> Factored::create( l,"irreducible" );

      if type(op(p,1))=DOM_FLOAT or type(op(p,2))=DOM_FLOAT then
        return(factored( [1,p,1] ))
      else
        return(factored( [p] ))
      end_if
    of DOM_IDENT do
    of DOM_FLOAT do  
      return(Factored::create( [1,p,1],"irreducible" ))
    otherwise
      // overloaded?
      if (f:= p::dom::factor) <> FAIL then
        // Note: Only for test purposes, because the method should already
        //       an object of type Factored!
        f:= f(p);
        if domtype(f) = DOM_LIST then
          return(Factored::create( f,"irreducible" ))
        else
          return(f)
        end_if
      else
        error("Illegal argument: ".expr2text(p))
      end_if
    end_case;

    // only case DOM_POLY was not left using return, so
    // p must be of type DOM_POLY now!
  
    if (q:=op(p,3)) <> Expr then
      if op(q,0) <> IntMod then
        f:= faclib::domfact(p);
        if domtype(f) = DOM_LIST then
          return(Factored::create( f,"irreducible" ))
        else
          return( f )
        end_if
      end_if;

      q:= op(p,[3,1]);
      // factorization in Zq for ...
      if isprime(q) then
        // .. multivariate polynomials over the field Zq:
        if nops((xx:=op(p,2)))=1 then
          // univariate polynomial given:
          ff:=[lcoeff(p)];
          p:=multcoeffs(p,1/op(ff));
          if degree((t:=nthterm(p,nterms(p))))<>0 then
            ff:=append(ff,poly(op(xx),xx,IntMod(q)),degree(t));
            p:=divide(p,t,Exact);
          end_if;

          if degree(p)=1 then
            return(factored( append(ff,p,1) ))
          else
            return(factored( ff.faclib::factor_mod(p,op(xx),q) ))
          end_if
        else
          // multivariate polynomial given:
          p:=faclib::multifactor(subsop(p,3=Dom::IntegerMod(q)));
          p:=subsop(p,1=expr(p[1]));
          return(factored( subs(p,Dom::IntegerMod(q)=IntMod(q)) ))
        end_if;
      elif nops(op(p,2)) > 1 then
        error("Multivariate factorization modulo composite numbers not yet implemented")
      else
        // ... univariate polynomials over the ring Zq:
        return(factored( [1].faclib::factor_mod_composite(p,op(p,[2,1]),q) ))
      end_if
    else
      // the result type of the factorization:
      factored:= l -> Factored::create( l,"irreducible" );

      // case op(p,3) = Expr
      if not testtype(p, Type::PolyOf(Type::Rational)) then
        if hastype(p, DOM_FLOAT, {DOM_COMPLEX, DOM_POLY}) then
          _p_:= mapcoeffs( p, numeric::rationalize,  Restore );
          Float:= l -> map(l,
               proc(x)
                 name factor_poly2float;
               begin
                 if domtype(x)=DOM_POLY then
                   mapcoeffs(x,float)
                 else
                   x
                 end_if
               end_proc)
        else
          _p_:=p;
          Float:=id
        end_if;

        f:= faclib::factor_expr( expr(_p_), x -> Factored::convert_to(factor(x),DOM_LIST) );
        Factors:=[f[1], (poly(f[2*i], op(p,2)), f[2*i+1]) $ i=1..nops(f) div 2];
        // remove constant factors
        i:=2;
        while i<= nops(Factors) do
          if degree(Factors[i])=0 then
            Factors[1]:=Factors[1]*expr(Factors[i])^Factors[i+1];
            delete Factors[i+1];
            delete Factors[i];
          else
            i:=i+2
          end_if
        end_while;
        return(factored( Float(Factors)) )
      end_if;
      // factorization in Z
      Factors:=[icontent(p)*sign(lcoeff(p))];
      p:= multcoeffs(p,1/op(Factors));
      if op((t:=faclib::monomial(p)),1)<>[] then
        Factors:=Factors.op(t,1);
        p:= divide(p, poly(op(t,2), op(p,2..3)), Exact);
        // PZ : added the following line because divide does not
        //      reduce the number of indets, which causes a loop
        //      for example with factor(a^2*k-2*a*k+k)
        p:=poly(expr(p));
      end_if;
      pp:=faclib::pfactor(p,0);
      // check sign
      if expr(poly(_mult(_power(expr(op(pp,2*i-1)),op(pp,2*i)) $ i=1..nops(pp)/2))) <> expr(p) 
      then
        Factors[1]:= -Factors[1];
      end_if;
    end_if;

    Factors:=Factors.pp;
    return(factored(
        subsop( Factors, (((2*i)=subsop(op(Factors,2*i),2=xx)) $ i=1..(nops(Factors)-1)/2) )
    ))
end_proc:

