// $Date: 2000/12/08 12:36:27 $ $Author: frankp $ $Revision: 1.80 $
// frankp, 11.01.1995

/*
        limit  -- computing the limit of expressions

        limit(expr,x[=a] [,dir])

        expr -- expression in x
        x    -- limit variable x (an identifier)
        a    -- (optional) limit point (possible infinity or -infinity)
                default is 0
        dir  -- (optional) direction: one of 'Left' or 'Right'

        The limit function attempts to compute the limiting value of expr
        as x goes to a.
   
        If dir is not specified, the limit is the real bidirectional limit,
        except in the case where the limit point is infinity or -infinity.
        In this case the limit is from the left to infinity and from the 
        right to -infinity, respectivly.
   
        limit only consider real valued functions, and any assigned 
        variables in the expression are assumed to be real.

        limit returns 'undefined' if the limit is not defined, or 
        FAIL if ORDER is not big enough for series.  

        If the limit could not be computed then limit returns
        unevaluated, i.e. the result is of type "limit".
*/

limit:= proc(f,pointt,dir=hold(Bidirect))
    local t, s, x, lp;
begin
    // allow overloading of the first argument of limit:
    if f::dom::limit <> FAIL then return(f::dom::limit( args() )) end_if;

    if testargs() then
        if args(0) < 2 or args(0) > 3 then
            error("expecting two or three arguments")
        end_if;
        if type(pointt) = "_equal" then
            if domtype(op(pointt,1)) <> DOM_IDENT then
                error("invalid limit variable (expecting an identifier)")
            end_if;
            if has( op(pointt,2), op(pointt,1) ) then
                error("expansion variable appears in limit point")
            elif not testtype( op(pointt,2),Type::Arithmetical ) then
                error("limit point must be of 'Type::Arithmetical'")
            end_if
        elif domtype(pointt) <> DOM_IDENT then
            error("invalid limit variable (expecting an identifier)")
        end_if;
        if not contains({hold(Right),hold(Left),hold(Bidirect)},dir) then
            error("unknown limit direction")
        elif op(pointt,2) = infinity and dir = hold(Right) or
             op(pointt,2) = -infinity and dir = hold(Left)
        then
             error("inconsistent limit direction")
        elif op(pointt,2) <> infinity and op(pointt,2) <> -infinity and has( op(pointt,2),infinity ) then
             error("invalid limit point")
        end_if
    end_if;

    case type( f ) 
    of "limit" do
        x:= op(f,1);
        t:= genident();
        if type(pointt) = "_equal" then
            lp:= op(pointt,2);
            if lp = infinity then
                assume( t > 0 );
            elif lp = -infinity then
                assume( t < 0 );
            elif testtype(lp,Type::Real) then
                case sign(lp)
                of 1 do
                    assume( t>0 );
                    if lp-1 > 0 then 
                        assume( t>1 ) 
                    elif lp-1 < 0 then 
                        assume( t<1, _and ) 
                    end_if;
                    break
                of -1 do
                    assume( t<0 ); 
                    break
                of 0 do
                    case dir 
                    of hold(Left) do 
                        assume( t<0 );
                        break
                    of hold(Right) do
                        assume( t>0 );
                        break
                    of hold(Bidirect) do
                        assume( t,Type::Real );
                    end_case;
                    break
                otherwise
                    assume( t,Type::Real );
                end_case
            else
                return( procname(args()) )
            end_if;

            x:= subs( x,op(pointt,1)=t );
            s:= limit( x,op(f,2..nops(f)) );
        else
            case dir
            of hold(Left) do 
                assume( t<0 );
                break
            of hold(Right) do
                assume( t>0 );
                break
            of hold(Bidirect) do
                assume( t,Type::Real );
            end_case;

            x:= subs( x,op(pointt,1)=t );
            s:= limit( x,op(f,2..nops(f)) );
        end_if;

        if s = f then
            return( procname(args()) )
        elif type(s) = "limit" then
            return( procname(subs(s,t=op(pointt,1)),pointt) )
        else
            s:= limit( subs(s,t=op(pointt,1)),pointt );
            return( s )
        end_if
    of DOM_PROC do
    of DOM_EXEC do
    of Series::Puiseux do
    of Series::gseries do
    of "series" do
    of "int" do
    of DOM_SET do
    of DOM_LIST do
    of DOM_DOMAIN do
    of DOM_POINT do
    of DOM_POLYGON do
    of DOM_ARRAY do
    of "diff" do
        return( procname(args()) )
    end_case;

    if type(pointt) = "_equal" then
        x:= op(pointt,1);
        lp:= op(pointt,2)
    else
        x:= pointt;
        lp:= 0
    end_if;

    if lp <> infinity and lp <> -infinity and dir = hold(Bidirect) then
        if traperror((s:= stdlib::limit::limit_main( f,x,lp,hold(Right) ))) <> 0 then
            stdlib::limit::reset();
            return( procname(args()) )
        end_if;
        // s:= stdlib::limit::limit_main( f,x,lp,hold(Right) );

        case s
        of UNKNOWN do
            stdlib::limit::reset();
            return( FAIL )
        of FAIL do
            stdlib::limit::reset();
            return( procname(args()) )  
        otherwise
            if x = infinity then 
                stdlib::limit::reset();
                return( s ) 
            end_if;
            if traperror((t:= stdlib::limit::limit_main( f,x,lp,hold(Left) ))) <> 0 then
                stdlib::limit::reset();
                return( procname(args()) )
            end_if;

            stdlib::limit::reset();
            case t
            of UNKNOWN do
                return( FAIL )
            of FAIL do
                return( procname(args()) )  
            otherwise
                if t = s then return( s ) else return( undefined ) end_if
            end_case
        end_case
    else
        // Note: the limit points infinity and -infinity will be handled in 
        // the subroutine 'limit_main':
        if traperror((s:= stdlib::limit::limit_main( f,x,lp,dir ))) <> 0 then
            stdlib::limit::reset();
            return( procname(args()) )
        end_if;
        // s:= stdlib::limit::limit_main( f,x,lp,dir );

        stdlib::limit::reset();
        case s
        of UNKNOWN do
            return( FAIL )
        of FAIL do
            return( procname(args()) )  
        otherwise
            return( s )
        end_case
    end_if
end_proc:

limit:= funcenv(limit,NIL,table(
    "type" = "limit", "info" = "limit -- calculate limits [try ?limit for options]",
    "print" = "limit"
)):
limit::Content := proc(Out, data)
    local direction, condition, tendsto, bvar;
begin
    if type(op(data, 2)) = "_equal" then
        tendsto := Out(op(data, [2,2])); 
        bvar := Out(op(data, [2,1]));
    else
        tendsto := Out(0); 
        bvar := Out(op(data, 2));
    end_if;
      
    if nops(data) = 3 then
        if op(data, 3) = hold(Right) then
            direction := Out::Ctendsto("type"="above");
        elif op(data, 3) = hold(Left) then
            direction := Out::Ctendsto("type"="below");
        else
            direction := Out::Ctendsto("type"="two-sided");
        end_if;
        condition := Out::Ccondition(
            Out::Capply(direction, bvar, tendsto)
        );
    else
        condition := Out::Clowlimit(tendsto);
    end_if;

    Out::Capply(Out::Climit, Out::Cbvar(bvar), condition, Out(op(data,1)));
end_proc:

/*********************************************************************************/
/*******  The domain stdlib::limit used for a namespace for the           ********/
/*******  limit routines.                                                 ********/
/*********************************************************************************/
alias( LIMIT = stdlib::limit ):

LIMIT:= newDomain("stdlib::limit"):

LIMIT::printWarningsFlag:= TRUE:
LIMIT::printWarnings:= proc(flag)
    local tmp;
begin
    if args(0) = 0 then 
        return( LIMIT::printWarningsFlag )
    else
        case flag
        of TRUE do
        of FALSE do
            tmp:= LIMIT::printWarningsFlag;
            sysassign( LIMIT::printWarningsFlag, flag );
            return( tmp )
        otherwise
            error("invalid argument")
        end_case
    end_if
end_proc:

LIMIT::setuserinfo:= proc() begin 
    // set userinfo
    setuserinfo(slot(LIMIT,"limit_main"), 10 );
    setuserinfo(slot(LIMIT,"limitMRV"), 10 );
    setuserinfo(slot(LIMIT,"lterm"), 10 );
    setuserinfo(slot(LIMIT,"getMRV"), 10 );
    setuserinfo(slot(LIMIT,"rewrite"), 10 );
    setuserinfo(slot(LIMIT,"maxMRV"), 10 );
    TRUE
end_proc:

LIMIT::reset := proc() begin 
    // clear remember tables
    sysassign(slot(LIMIT,"limitMRV"), subsop( slot(LIMIT,"limitMRV"),5=NIL ));
    sysassign(slot(LIMIT,"lterm"), subsop( slot(LIMIT,"lterm"),5=NIL ));
    sysassign(slot(LIMIT,"getMRV"), subsop( slot(LIMIT,"getMRV"),5=NIL ));
    sysassign(slot(LIMIT,"rewrite"), subsop( slot(LIMIT,"rewrite"),5=NIL ));
    sysassign(slot(LIMIT,"maxMRV"), subsop( slot(LIMIT,"maxMRV"),5=NIL ));
    sysassign(slot(LIMIT,"sign"), subsop( slot(LIMIT,"sign"),5=NIL ));
    sysassign(slot(LIMIT,"simp2explog"), subsop( slot(LIMIT,"simp2explog"),5=NIL ));
    sysassign(slot(LIMIT,"limitIV_main"), subsop( slot(LIMIT,"limitIV_main"),5=NIL ));
    sysassign(slot(LIMIT,"limitIV_cmp"), subsop( slot(LIMIT,"limitIV_cmp"),5=NIL ));
    sysassign(slot(LIMIT,"limitIV_join"), subsop( slot(LIMIT,"limitIV_join"),5=NIL ));
end_proc:

LIMIT::limit_main:= proc( ex,x,lp,dir )
    local __LIMITx, nex, s, t, S, ld, order;
begin
    if not has(ex,x) then return( ex ) end_if;

    __LIMITx:= genident("__LIMITx");
    assume(__LIMITx,Type::Real);

    // ********************************
    // Try series computation directly:
    // ********************************
    userinfo(3,"First, try series computation directly ...");

    // we transform the problem to the case __LIMITx -> +0:
    if dir = hold(Left) then
        if lp = infinity then
            t:= 1/__LIMITx
        else
            t:= -__LIMITx+lp 
        end_if
    else 
        if lp = infinity then
            t:= 1/__LIMITx
        elif lp = -infinity then
            t:= -1/__LIMITx
        else
            t:= __LIMITx+lp 
        end_if
    end_if;
    if traperror( (nex:= eval(subs(ex,x=t))) ) <> 0 then
        return( FAIL )
    end_if;

    s:= FAIL;
    for order from 2 to ORDER do
        // an error occurs if the requested order is too small!
        if traperror((s:= series(nex, __LIMITx, order, hold(Right), hold(NoWarning)))) = 0 then 
            if domtype(s) <> Series::Puiseux then 
                // try the MRV algorithm ...
                break
            elif (S:= ldegree(s)) <> FAIL then
                break
            end_if
        end_if
    end_for;
    if domtype(s) = Series::Puiseux and S <> FAIL then
        case sign(S)
        of  0 do
            ld:= lmonomial(s);
            if has(ld,__LIMITx) then
                if ld = expr(s) then
                    // avoid infinite loop, continue with the MRV algorithm ...
                    break
                else
                    return( LIMIT::limit_main(ld,__LIMITx,0,hold(Right)) )
                end_if
            else
                return( ld )
            end_if
        of  1 do
            return( 0 )
        of -1 do
            // NOTE: the leasing coefficient could contain
            // __LIMITx! Example: limit(ln(x)/x,x=0,Left).
            // However, it is guaranteed that the lcoeff does not grow 
            // stronger than the rest of the leading monomial. So the
            // limit is infinity*sign(lcoeff).
            s:= lcoeff(s);
            if s = FAIL then
                // try the MRV algorithm!
                // Example: series((sqrt(2*x)-4)/(sqrt(x+1)-3),x=8,2,Right)
                break
            elif not has(s,__LIMITx) then
                return( sign(s)*infinity )
            else
                s:= sign(LIMIT::limit_main(s,__LIMITx,0,hold(Right)));
                if s = 1 or s = -1 then 
                    return( s*infinity ) 
                end_if;
                // continue with the MRV algorithm
                break
            end_if
        otherwise
            if LIMIT::printWarningsFlag = TRUE then
                warning("cannot determine sign of ".expr2text(S) );
            end_if;
            return( FAIL )
        end_case;
        // continue with the MRV algorithm ...
        userinfo(3,"Direct series computation fails. Enter MRV algorithm ...")
    end_if;

    // ********************
    // Enter MRV algorithm:
    // ********************
    // we transform the problem to the case __LIMITx -> infinity:
    assume( __LIMITx > 1 ); // in order to have ln(__LIMITx) positive ...

    // trap possible error due to evaluation with variable (see above)
    if lp = -infinity then
        if traperror((ex:= eval(subs(ex,x=-__LIMITx)) )) <> 0 then
            return( FAIL )
        end_if
    elif lp = infinity then
        if traperror((ex:= eval(subs(ex,x=__LIMITx)) )) <> 0 then
            return( FAIL )
        end_if
    elif dir = hold(Right) then	
        if traperror((ex:= eval(subs(ex,x=lp+1/__LIMITx)) )) <> 0 then
            return( FAIL )
        end_if
    elif traperror((ex:= eval(subs(ex,x=lp-1/__LIMITx)) )) <> 0 then
        return( FAIL )
    end_if;

    nex:= simplify( ex,exp ); 
    nex:= simplify( nex,ln );
    nex:= LIMIT::simp2explog( nex,__LIMITx ); 

    if has(ex,LIMIT::setPreProcessFunc) then
        nex:= LIMIT::preProcess( nex,__LIMITx );
    end_if;

    s:= LIMIT::limitMRV(nex,__LIMITx);
    case s
    of FAIL    do
    of UNKNOWN do
        return( s )
    of "tryIV" do
        // **********************
        // Try interval calculus:
        // **********************
        s:= LIMIT::limitIV_main( ex,__LIMITx );
        if s = FAIL or s = UNKNOWN then
            return( s )
        elif domtype(s) = DOM_LIST and nops(s) = 1 then
            s:= s[1];
            t:= normal( s[1] - s[2] );
            if iszero(t) = TRUE or is(t = 0) = TRUE then
                return( s[1] )
            elif has(s,infinity) then
                return( FAIL )
            else
                return(Dom::Interval( s ))
            end_if
        else
            return( FAIL )
        end_if
    otherwise
        return( s )
    end_case
end_proc:

/*********************************************************************************/
/******** The following routines implement limit computation by interval  ********/
/******** arithmetic, based on the following paper:                       ********/
/******** John Shackell: "Asymptotic estimation of oscillation functions  ********/
/******** using an interval calculus".                                    ********/
/*********************************************************************************/

/*
        limitIV_main(f,x)

        f : expression in x
        x : limit variable

        Tries to determine the limit of f(x) for x->infinity using
        an interval calculus. Return values are either
        a) a single value, the limit of f(x) for x->infinity.
        b) an interval [a,b] such that f(x) is in [a,b] for large x. 
        c) FAIL, if the limit could not be determined.

        The interval calculus implemented in Dom::Interval is used. 
        It handles parameters and takes care of possibly given 
        properties on identifiers (see assume/is).
*/
LIMIT::limitIV_main:= proc(f,x)
    local c, s, t, e;
    option remember;
begin
    if f = x then
        return( [[infinity,infinity]] )
    else
        if not has(f,x) then 
            return( [[f,f]] )
        elif hastype(f,DOM_COMPLEX) then
            return( FAIL )
        end_if
    end_if;

    c:= LIMIT::limitMRV( f,x );
    case c 
    of FAIL do
    of "tryIV" do
        t:= type(f);
        case t
        of "_plus"  do
        of "_mult"  do
        of "_power" do
            t:= LIMIT::limitIV[t];

            s:= NIL;
            for e in f do
                c:= LIMIT::limitIV_main( e,x );
                if c = FAIL or c = UNKNOWN then return( c )
                elif s = NIL then s:= c
                else s:= t( s,c )
                end_if
            end_for;
            return( s )
        otherwise
            if type((s:= LIMIT::limitIV[t])) <> "_index" then
                c:= LIMIT::limitIV_main( op(f,1),x );
                if c = FAIL or c = UNKNOWN then
                    return( c )
                else
                    return( s(c) )
                end_if
            else
                return( FAIL )
            end_if
        end_case 
    of UNKNOWN do
        // ORDER of series computation seems to be not big enough
        return( UNKNOWN )
    otherwise
        return( [[c,c]] )
    end_case
end_proc:

LIMIT::limitIV_cmp:= proc(a, b)
    option remember;
begin 
    if a[1] = b[1] then
        if a[2] = b[2] then
            return(TRUE)
        else
            return(bool(sign(a[2]-b[2])=-1))
        end_if
    else
        return(bool(sign(a[1]-b[1])=-1))
    end_if
end_proc:

/*--
        Liste von Intervallen wird sortiert, zusammengefasst und die 
        Reihenfolge der Intervallgrenzen werden gerichtet.
--*/
LIMIT::limitIV_join:= proc(liste)
    local i, a, b, c, s;
    option remember;
begin
    for i from 1 to nops(liste) do
        // Reihenfolge innerhalb eines Intervalls richten:
        a:= liste[i];
        if sign(a[1]-a[2]) = 1 then
            liste[i]:= [a[2],a[1]]
        end_if;
    end_for;

    liste:= sort(liste,LIMIT::limitIV_cmp);

    // ueberlappende Intervalle vereinigen:
    i:= 1;
    a:= op(liste,i);
    b:= op(liste,i+1);
    while a <> FAIL and b <> FAIL do
        if ((a[1] = -infinity) and (a[2] = -infinity)) or
           ((a[1] =  infinity) and (a[2] =  infinity)) 
        then
            delete liste[i]
        elif ((b[1] = -infinity) and (b[2] = -infinity)) or
             ((b[1] =  infinity) and (b[2] =  infinity)) 
        then
            delete liste[i+1];
        elif ((a[2] = -infinity) and (b[1] = -infinity)) or
             ((a[2] =  infinity) and (b[1] =  infinity)) 
        then
            c:= [a[1],b[2]];
            liste[i]:= c;
            delete liste[i+1]; 
        elif (s:= sign(a[2]-b[1])) = 0 or s = 1 then 
            // Intervalle ueberlappen sich:
            c:= [a[1],0]; 
            if (a[2] = infinity) and (b[2] = infinity) then
                // rechte Intervallgrenze festlegen:
                c[2]:= a[2];
            elif (s:= sign(a[2]-b[2])) = 0 or s = 1 then
                c[2]:= a[2];
            else
                c[2]:= b[2];
            end_if:

            liste[i]:= c;
            delete liste[i+1];   
        else
            i:= i + 1;
        end_if;

        a:= op(liste,i);
        b:= op(liste,i+1); 
    end_while;

    return(liste)
end_proc:

LIMIT::limitIV:= table():
/*--
        Invertierung von Intervallisten
--*/
LIMIT::limitIV["_invert"]:= proc(liste)
    local a, l, sa1, sa2;
begin
    l:= [];
    for a in liste do
        if sign(a[1]-a[2]) = 1 then
            // Reihenfolge richten:
            a:= [a[2],a[1]];
        end_if;

        sa1:= sign(a[1]); sa2:= sign(a[2]);
        if sa1 = 0 and sa2 = 0 then       
            // Sonderfall: Division durch Null:
            l:= l . [[-infinity,-infinity],[infinity,infinity]]
        elif sa1 = -1 and sa2 = 1 then            
            // Sonderfall: Null liegt im Intervall ==> zwei Intervalle
            l:= l . [[-infinity,1/a[1]],[1/a[2],infinity]]
        elif sa1 = 0 then                        
            l:= l . [[1/a[2],infinity]]
        elif sa2 = 0 then                        
            l:= l . [[-infinity,1/a[1]]]
        else
           l:= l . [[1/a[2],1/a[1]]]
        end_if
    end_for;
    
    return( LIMIT::limitIV_join(l) )
end_proc:

/*--
        Addition zweier Intervallisten:
--*/
LIMIT::limitIV["_plus"]:= proc(l1,l2)
    local a, b, s;
begin
    s:= [];
    for a in l1 do
        for b in l2 do
            s:= s . [[a[1] + b[1],a[2] + b[2]]]
        end_for
    end_for;
 
    return( LIMIT::limitIV_join(s) )
end_proc:

/*--
        Multiplikation zweier Intervallisten:
--*/
LIMIT::limitIV_multIV:= proc(a,b)
    local sa1,sa2,sb1,sb2,c,mi,ma;
begin
     // Reihenfolge richten:
     if sign( a[1]-a[2] ) = 1 then a:= [a[2],a[1]] end_if;
     if sign( b[1]-b[2] ) = 1 then b:= [b[2],b[1]] end_if;

     sa1:= sign(a[1]); sa2:= sign(a[2]);
     sb1:= sign(b[1]); sb2:= sign(b[2]);
     if sa1 = 0 and sa2 = 0 and b[1] = -infinity and b[2] = infinity then
         return( [-infinity,infinity] )
     elif sb1 = 0 and sb2 = 0 and a[1] = -infinity and a[2] = infinity then
         return( [-infinity,infinity] )
     elif sa1 = 0 and sa2 = 0 and b[2] = infinity then
         return( [0,infinity] )
     elif sb1 = 0 and sb2 = 0 and a[2] = infinity then
         return( [0,infinity] )
     elif sa1 = 0 and sa2 = 0 and b[1] = -infinity then
         return( [-infinity,0] )
     elif sb1 = 0 and sb2 = 0 and a[1] = -infinity then
         return( [-infinity,0] )
     elif sa2 = 0 and b[2] = infinity then
         if sa1 * sa2 = -1 then 
             return( [-infinity,0] )
         else 
             return( [-infinity,a[1]*b[1]] )
         end_if
     elif sb2 = 0 and a[2] = infinity then
         if sa1*sb1 = -1 then 
             return( [-infinity,0] )
         else 
             return( [-infinity,a[1]*b[1]] )
         end_if
     elif sb1 = 0 and a[1] = - infinity then
         if sa2*sb2 = -1 then 
             return( [-infinity,0] )
         else
             return( [-infinity,a[2]*b[2]] )
         end_if
     elif sa1 = 0 and b[1] = -infinity then
         if sa2*sb2 = -1 then
             return( [-infinity,0] )
         else 
             return( [-infinity,a[2]*b[2]] )
         end_if
     elif sa1 = 0 and b[2] = infinity then
         if sa2*sb1 = 1 then 
             return( [0,infinity] )
         else 
             return( [a[2]*b[1],infinity] )
         end_if
     elif sb1 = 0 and a[2] = infinity then
         if sa1*sb2 = 1 then 
             return( [0,infinity] )
         else 
             return( [a[1]*b[2],infinity] )
         end_if
     elif sb2 = 0 and a[1] = -infinity then
         if sa2*sb1 = 1 then 
             return( [0,infinity] )
         else
             return( [a[2]*b[1],infinity] )
         end_if
     elif sa2 = 0 and b[1] = -infinity then
         if sa1*sb2 = 1 then 
             return( [0,infinity] )
         else 
             return( [a[1]*b[2],infinity] )
         end_if
     else
         mi:= (ma:= a[1]*b[1]);
         c:= a[1]*b[2];
         if sign(c-mi) = -1 then mi := c end_if;
         if sign(c-ma) = 1  then ma := c end_if;
         c:= a[2]*b[1];
         if sign(c-mi) = -1 then mi := c end_if;
         if sign(c-ma) = 1  then ma := c end_if;
         c:= a[2]*b[2];
         if sign(c-mi) = -1 then mi := c end_if;
         if sign(c-ma) = 1  then ma := c end_if;
         return( [mi,ma] )
     end_if; 
end_proc:
LIMIT::limitIV["_mult"]:= proc(l1,l2)
    local a, b, m;
begin
    m:= [];
    for a in l1 do
        for b in l2 do
            m:= m . [LIMIT::limitIV_multIV(a,b)];
        end_for;
    end_for;
  
    return( LIMIT::limitIV_join(m) )
end_proc:

/*--
        Potenzierung zweier Intervallisten:
--*/
LIMIT::limitIV_powerIV:= proc(a,b)
begin
    if b[1] <> b[2] then 
        // Exponent darf kein Intervall sein!
        return( FAIL )
    end_if;

    // Reihenfolge richten:
    if sign(a[1]-a[2]) = 1 then a:= [a[2],a[1]] end_if;
  
    case sign(b[1])
    of 0 do
        return( [[1,1]] )
    of -1 do
        return(LIMIT::limitIV_invert( LIMIT::limitIV_powerIV( a,[-b[1],-b[1]] ) ))
    otherwise
        if sign(a[1]*a[2]) <> 0 then
            if sign(a[1]*a[2]) = -1 and sign(a[1]^b[1]) = 1 then
                return( [[0,a[2]^b[1]]] )
            else
                return( [[a[1]^b[1],a[2]^b[1]]] )
            end_if
        else
            return( [[a[1]^b[1],a[2]^b[1]]] )
        end_if
    end_case
end_proc:
LIMIT::limitIV["_power"]:= proc(l1,l2)
    local a, b, p, tmp;
begin
    p:= [];
    for a in l1 do
        for b in l2 do
            tmp:= LIMIT::limitIV_powerIV(a,b);
            if tmp = FAIL then 
                return( FAIL )
            else 
                p:= p . tmp
            end_if
        end_for
    end_for;
  
    return( LIMIT::limitIV_join(p) )
end_proc:

/*--
        Exponentialfunktion auf Intervalliste:
--*/
LIMIT::limitIV["exp"]:= proc(liste)
    local i, a, expIV;
begin
    // Erweiterte Exponentialfunktion:
    expIV:= proc(a)
    begin
        case a
        of infinity  do return(infinity)
        of -infinity do return(0)
        otherwise
            return( exp(a) )
        end_case
    end;
    // ---------------------------------
    for i from 1 to nops(liste) do
        a:= liste[i];
        liste[i]:= [expIV(a[1]),expIV(a[2])]
    end_for;
  
    return( LIMIT::limitIV_join(liste) )
end_proc:

/*--
        Logarithmus auf Intervalliste:
--*/
LIMIT::limitIV["ln"]:= proc(liste)
    local i, a, t1, t2, lnIV;
begin
    // Erweiterte Logarithmusfunktion:
    lnIV:= proc(a)
    begin
        if a <> -infinity and sign(a) = -1 then
            return(FAIL)
        end_if;

        if (a = infinity) or (a = -infinity) then
            return(infinity)
        elif iszero(a) = TRUE then
            return(-infinity)
        else
            return(ln(a))
        end_if;
    end:
    // ---------------------------------
    for i from 1 to nops(liste) do
        a := liste[i];

        t1:= lnIV(a[1]); t2:= lnIV(a[2]);
        if t1 = FAIL or t2 = FAIL then
            return( FAIL )
        elif sign(a[1]*a[2]) <> 0 then
            if sign(a[1]*a[2]) = -1 then
                if sign(t1-t2) = 1 then
                    liste[i]:= [-infinity,t1]
                else
                    liste[i]:= [-infinity,t2]
                end_if;
            else
                liste[i]:= [t1,t2]
            end_if;
        else 
            liste[i]:= [t1,t2]
        end_if;
    end_for;
  
    return( LIMIT::limitIV_join(liste) )
end_proc:

/*--
        Absolutfunktion auf Intervalliste:
--*/
LIMIT::limitIV["abs"]:= proc(liste)
    local i, a;
begin
    for i from 1 to nops(liste) do
        a:= liste[i];

        if sign(a[1]*a[2]) <> 0 then
            if sign(a[1]*a[2]) = -1 then
                if sign(abs(a[1])-abs(a[2])) = 1 then
                    a:= [0,abs(a[1])]
                else
                    a:= [0,abs(a[2])]
                end_if
             end_if
         else
             a:= [abs(a[1]),abs(a[2])]
         end_if;
         liste[i]:= a
    end_for;
  
    return( LIMIT::limitIV_join(liste) )
end_proc:

/*--
        Sinus auf Intervalliste:
--*/
LIMIT::limitIV["sin"]:= proc(liste)
    local i, a, sinIV;
begin
    // Sinus eines Intervalls:
    sinIV:= proc(a)
        local a1, a2, npi;
    begin
        a1 := float(a[1]); a2 := float(a[2]);
        if domtype(sign(a1)) <> DOM_INT or domtype(sign(a2)) <> DOM_INT then
            // sign can't be computed
            return( FAIL )
        elif sign(a[1]-a[2]) = 1 then 
            // Reihenfolge richten:
            a:= [a[2],a[1]];
            ([a1,a2]):= [a2,a1]
        end_if;
  
        npi:= float(PI);
        if a1 = infinity or a1 = -infinity  or a2 = infinity or a2 = -infinity 
           // or (a2-a1 >= float(2*PI)) or (float(a1/PI+0.5) <= floor(float(a2/PI-0.5)))
           or contains( {0,1},sign(a2-a1-2*npi) ) or contains( {-1,0},sign((a1/npi+0.5) - floor(a2/npi-0.5)) )
        then
            return([-1,1])
        // elif (floor(float(a2/(2*PI)-0.25)) > float(a1/(2*PI)-0.25)) then
        elif sign(floor(a2/(2*npi)-0.25) - (a1/(2*npi)-0.25)) = 1 then
            if sign(sin(a[1])-sin(a[2])) = -1 then
                return([sin(a[1]),1]) 
            else 
                return([sin(a[2]),1])
            end_if
        // elif floor(a2/(2*PI)+0.25) > a1/(2*PI)+0.25 then
        elif sign(floor(a2/(2*npi)+0.25) - (a1/(2*npi)+0.25)) = 1 then
            if sign(sin(a[1])-sin(a[2])) = 1 then
                return([-1,sin(a[1])])
            else 
                return([-1,sin(a[2])])
            end_if
        else
            if sign(sin(a[1])-sin(a[2])) = -1 then
                return([sin(a[1]),sin(a[2])])
            else 
                return([sin(a[2]),sin(a[1])] )
            end_if
        end_if
    end:
    // ---------------------------------
    for i from 1 to nops(liste) do  
        a := liste[i];
        if (liste[i] := sinIV(a)) = FAIL then
            return( FAIL )
        end_if
    end_for;
  
    return( LIMIT::limitIV_join(liste) )
end_proc:

/*--
        Cosinus auf Intervalliste:
--*/
LIMIT::limitIV["cos"]:= proc(liste)
begin
    LIMIT::limitIV["sin"]( LIMIT::limitIV["_plus"]( 
        [[PI/2,PI/2]], LIMIT::limitIV["_mult"]( [[-1,-1]],liste )
    ) )
end_proc:

/*--
        Tangens auf Intervalliste:
--*/
LIMIT::limitIV["tan"]:= proc(liste)
    local f1, f2;
begin
  /*
     LIMIT::limitIV["_mult"](
        LIMIT::limitIV["sin"](liste),
        LIMIT::limitIV["_invert"](LIMIT::limitIV["cos"](liste))
    )
  */
    // take care of the result FAIL!
    if (f2:= LIMIT::limitIV["cos"](liste)) = FAIL then
        return( FAIL )
    elif (f2:= LIMIT::limitIV["_invert"](f2)) = FAIL then
        return( FAIL )
    elif (f1:= LIMIT::limitIV["sin"](liste)) = FAIL then
        return( FAIL )
    else
        return( LIMIT::limitIV["_mult"](f1,f2) )
    end_if
end_proc:

/*--
        Sinushyperbolikus auf Intervalliste:
--*/
LIMIT::limitIV["sinh"]:= proc(liste)
    local i, a;
begin
    for i from 1 to nops(liste) do  
        a := liste[i];
        liste[i] := [sinh(a[1]),sinh(a[2])]
    end_for;
  
    return( LIMIT::limitIV_join(liste) )
end_proc:

/*--
        Cosinushyperbolikus auf Intervallliste:
--*/
LIMIT::limitIV["cosh"]:= proc(liste)
    local i, a, coshIV;
begin
    // Cosinushyperbolikus auf Intervall:
    coshIV:= proc(a)
    begin
        if sign(a[1]*a[2]) = -1 then
            if sign(cosh(a[1])-cosh(a[2])) = 1 then
                a:= [1,cosh(a[1])]
            else
                a := [1,cosh(a[2])]
            end_if
        else
            a:= [cosh(a[1]),cosh(a[2])]
        end_if;

        // Reihenfolge richten:
        if sign(a[1]-a[2]) = 1 then a:= [a[2],a[1]] end_if
    end:
    // ---------------------------------
    for i from 1 to nops(liste) do  
        a:= liste[i];
        liste[i]:= [coshIV(a[1]),coshIV(a[2])]
    end_for;
  
    return( LIMIT::limitIV_join(liste) )
end_proc:

/*--
        Tangeshyperbolikus auf Intervallliste:
--*/
LIMIT::limitIV["tanh"]:= proc(liste)
    local i, a;
begin
    for i from 1 to nops(liste) do  
        a:= liste[i];
        liste[i]:= [sinh(a[1]),sinh(a[2])]
    end_for;
  
    return( LIMIT::limitIV_join(liste) )
end_proc:

/*--
        Arcussinus auf Intervallliste:
--*/
LIMIT::limitIV["arcsin"]:= proc(liste)
    local i, a;
begin
    for i from 1 to nops(liste) do  
        a:= liste[i];
        if sign(a[1]+1) = -1 or sign(a[2]-1) = 1 then
            return(FAIL)
        end_if;

        liste[i]:= [arcsin(a[1]),arcsin(a[2])];
    end_for;
  
    return( LIMIT::limitIV_join(liste) )
end_proc:

/*--
        Arcuscosinus auf Intervallliste:
--*/
LIMIT::limitIV["arccos"]:= proc(liste)
    local i, a;
begin
    for i from 1 to nops(liste) do  
        a:= liste[i];
        if sign(a[1]+1) = -1 or sign(a[2]-1) = 1 then
            return(FAIL)
        end_if;

        liste[i] := [arccos(a[2]),arccos(a[1])]
    end_for;
  
    return( LIMIT::limitIV_join(liste) )
end_proc:

/*--
        Arcustangens auf Intervallliste:
--*/
LIMIT::limitIV["arctan"]:= proc(liste)
    local i, a;
begin
    for i from 1 to nops(liste) do  
        a:= liste[i];
        liste[i]:= [arctan(a[1]),arctan(a[2])]
    end_for;
  
    return( LIMIT::limitIV_join(liste) )
end_proc:

/*--
        Arcussinushyperbolikus auf Intervallliste:
--*/
LIMIT::limitIV["arcsinh"]:= proc(liste)
    local i, a, arcsinhIV;
begin
    // Arcussinushyperbolikus auf Intervall:
    arcsinhIV:= proc(a)
    begin
        case a
        of infinity  do return(infinity)
        of -infinity do return(-infinity)
        otherwise
            return(arcsinh(a))
        end_case
    end:
    // ---------------------------------
    for i from 1 to nops(liste) do  
        a:= liste[i];
        liste[i]:= [arcsinhIV(a[1]),arcsinhIV(a[2])]
    end_for;

    return( LIMIT::limitIV_join(liste) )
end_proc:

/*--
        Arcuscosinushyperbolikus auf Intervallliste:
--*/
LIMIT::limitIV["acosh"]:= proc(liste)
    local i, a, arccoshIV;
begin
    // Arcuscosinushyperbolikus auf Intervall:
    arccoshIV:= proc(a)
    begin
        if a = infinity then return(infinity) else return(arccosh(a)) end_if
    end:
    // ---------------------------------
    for i from 1 to nops(liste) do  
        a:= liste[i];
        if sign(a[1]-1) = -1 then
            return(FAIL)
        end_if;

        liste[i]:= [arccoshIV(a[1]),arccoshIV(a[2])];
    end_for;
  
    return( LIMIT::limitIV_join(liste) )
end_proc:

/*--
        Arcustangenshyperbolikus auf Intervallliste:
--*/
LIMIT::limitIV["arctanh"]:= proc(liste)
    local i, a, arctanhIV;
begin
    // Arcustangenshyperbolikus auf Intervall:
    arctanhIV := proc(a)
    begin
        if sign(a+1) = 0 then
            return(-infinity)
        elif sign(a-1) = 0 then
            return(infinity)
        else
            return(arctanh(a))
        end_if
    end:
    // ---------------------------------
    for i from 1 to nops(liste) do  
        a:= liste[i];
        if sign(a[1]+1) = -1 or sign(a[2]-1) = 1 then
            return(FAIL)
        end_if;

        liste[i]:= [arctanhIV(a[1]),arctanhIV(a[2])];
    end_for;
 
    return( LIMIT::limitIV_join(liste) )
end_proc:

/*--
        Signumfunktion auf Intervallliste:
--*/
LIMIT::limitIV["sign"]:= proc(liste)
    local i, a, l, sa1, sa2;
begin
    l:= [];
    for i from 1 to nops(liste) do  
        a:= liste[i];

        sa1:= sign(a[1]); sa2:= sign(a[2]);
        if sa1*sa2 = -1 then
            l := l . [[-1,-1],[0,0],[1,1]]
        elif sa1 = 0 and sa2 = 0 then
            l := l . [[0,0]]
        elif sa1*sa2 = 0 then
            l := l . [[sa1,sa1],[sa2,sa2]]
        else
            l := l . [[sa1,sa1]]
        end_if
    end_for;
  
    return( LIMIT::limitIV_join(l) )
end_proc:

/*--
        Fehlerfunktion auf Intervallliste:
--*/
LIMIT::limitIV["erf"]:= proc(liste)
    local i, a;
begin
    for i from 1 to nops(liste) do  
        a:= liste[i];
        liste[i]:= [erf(a[1]),erf(a[2])];
    end_for;
  
    return( LIMIT::limitIV_join(liste) )
end_proc:

/*--
        komplementaere Fehlerfunktion auf Intervallliste:
--*/
LIMIT::limitIV["erfc"]:= proc(liste)
    local i, a;
begin
    for i from 1 to nops(liste) do  
        a:= liste[i];
        liste[i]:= [erfc(a[2]),erfc(a[1])];
    end_for;
  
    return( LIMIT::limitIV_join(liste) )
end_proc:


/******************************************************************************/
/*********  The followig routines implement the MRV algorithm based ***********/
/*********  on the paper of Gaston H. Gonnet and Dominik Gruntz:    ***********/
/*********  "Limit Computation in Computer Algebra", Nov. 1992.     ***********/
/******************************************************************************/
// Global Variables:
LIMIT::OMEGA:= genident("stdlib::limit::OMEGA"): // global identifier without a value

// Rewrites powers to exp-log representation:
LIMIT::simp2explog := proc(ex,__LIMITx)
    name (stdlib::limit)::simp2explog;
    option remember;
    local t;
begin
    case domtype(ex)
    of DOM_EXPR do
        case type(ex)
        of "_index" do 
            return( ex )
        of "_power" do
            if testtype(op(ex,2),Type::Numeric) then
                return( LIMIT::simp2explog(op(ex,1),__LIMITx)^op(ex,2) )
            else
                t:= LIMIT::simp2explog( op(ex,1),__LIMITx );
                case sign(t)
                of  0 do
                    return( 0 )
                of -1 do
                    // avoid rewriting to exponential form, it would cause
                    // complex values ...
                    return( t^op(ex,2) )
                otherwise
                    return( exp(LIMIT::simp2explog(op(ex,2),__LIMITx)*ln(t)) )
                end_case
            end_if;
        of "_mult"   do
        of "_plus"   do
            return( eval(map(ex,LIMIT::simp2explog,__LIMITx)) )
        end_case;
        t:= eval(op(ex,0));
        if domtype(t) = DOM_FUNC_ENV then
            if nops(ex) = 1 then
                return(t( LIMIT::simp2explog(op(ex,1),__LIMITx) ))
            else
                return(t( op(map([op(ex)],LIMIT::simp2explog,__LIMITx)) ))
            end_if
        elif is(ex = 0) = TRUE then 
            return( 0 )
        else
            return( ex )
        end_if
    of DOM_IDENT do
        if is(ex = 0) = TRUE then return( 0 ) end_if
    otherwise
        return( ex )
    end_case
end_proc:

/*
        limitMRV(ex,x)

        ex - expression
        x  - limit variable

        Calls lterm and analyzes the leading
        coefficient of the asymptotic expansion (in w), i.e.
        returns the limit of ex for x -> infinity, if possible.

        Return Values:
        l       - the limit (arithmetical expression).
        FAIL    - limit cannot be computed.
        UNKNOWN - ORDER for series seems to be not big enough.
        "tryIV" - MRV algorithm fails, but interval calculus may succeed.
*/
LIMIT::limitMRV:= proc(ex,__LIMITx)
    name stdlib::limit::limitMRV;
    option remember;
    local lc, s; 
begin
    userinfo(2,"MRV: compute limit of ",ex);
    lc:= LIMIT::lterm( ex,__LIMITx );
    if contains( {UNKNOWN,FAIL,"tryIV"},lc ) then
        return( lc )
    else
        // (d) Analyze the leading degree
        s:= sign(lc[3]);
        if s = 0 then 
            return( lc[1] )
        elif s = 1 then 
            return( 0 )
        elif s = -1 then
            s:= LIMIT::sign(lc[1],__LIMITx);
            if s = FAIL then
                if indets(lc[1]) minus Type::ConstantIdents minus {__LIMITx} <> {} then
                    if LIMIT::printWarningsFlag = TRUE then
                        warning("cannot determine sign of ".expr2text(lc[1]));
                    end_if;
                    return( FAIL )
                else
                    // use interval calculus, as for the example (-1/2)^n,n->oo!
                    return( "tryIV" )
                end_if
            elif iszero(s) then
                // Happens if series returns zero as the leading coefficient
                // of the series expansion!
                return( FAIL )
            else
                return( s*infinity )
            end_if
        else
            if LIMIT::printWarningsFlag = TRUE then
                warning("cannot determine sign of ".expr2text(lc[3]));
            end_if;
            return( FAIL )
        end_if
    end_if
end_proc:


/*
        upmove(ex,x)
        downmove(ex,x)

        Subroutines for LIMIT::lterm.
*/
LIMIT::upmove:= (f,x) -> eval(subs(f,[ln(x)=x,x=exp(x)])):
LIMIT::downmove:= (f,x) -> eval(subs(f,[exp(x)=x,x=ln(x)])):

/*
        lterm(ex,x)

        ex -- expression
        x  -- limit variable
        S  -- (optional) mrv set

        Compute the list [c0,w,e0], where c0*w^e0 is the leading
        term of an asymptotic expansion of ex for x -> infinity.

        Return Values:
        [c0,w,e0] - c0*w^e0 as the leading term
        FAIL      - limit cannot be computed.
        UNKNOWN   - ORDER for series seems to be not big enough.
        "tryIV"   - MRV algorithm fails, but interval calculus may succeed.
*/
LIMIT::lterm:= proc(ex,__LIMITx)
    name stdlib::limit::lterm;
    local i, S, s, lc, le, t, ex0, siszero, omega;
    option remember;
begin
    ex0:=ex; // used to avoid infinite loops

    if not has(ex,__LIMITx) then return( [ex,1,0] ) end_if;

    // (a) Find the most rapidly varying subexpressions

    if args(0) = 2 then
        userinfo(3,"MRV: (1) Find MRV set");
        userinfo(4,"         of ",ex);
        S:= LIMIT::getMRV(ex,__LIMITx);
        if contains( {UNKNOWN,FAIL,"tryIV"},S ) then
            userinfo(3,"MRV: ** determining MRV set fails");
            return( S )
        elif domtype(S) <> DOM_SET then
            // the limit could be determined directly!
            return( [S,__LIMITx,0] )
        elif nops(S) = 0 then
            return( [ex,1,0] )
        end_if
    else
        S:= args(3);
        t:= select( S, (x,y) -> has(y,x), ex );
        if t = {} then 
            S:= LIMIT::getMRV(ex,__LIMITx);
            if contains( {UNKNOWN,FAIL,"tryIV"},S ) then
                userinfo(3,"MRV: ** determining MRV set fails");
                return( S )
            elif domtype(S) <> DOM_SET then
                // the limit could be determined directly!
                return( [S,__LIMITx,0] )
            end_if
        end_if
    end_if;

    if S intersect {__LIMITx} = {__LIMITx} then // Upward Movement
        omega:= LIMIT::lterm(
            LIMIT::upmove(ex,__LIMITx), __LIMITx, LIMIT::upmove(S,__LIMITx)
        );
        if omega = FAIL then return( FAIL ) else return( LIMIT::downmove(omega,__LIMITx) ) end_if
    end_if;


   // (b) Choose one expression in this set and call it w. Rewrite the    
   //     other expressions in that set as g(x)=w^p h(x) and consider all 
   //     expressions  independent of w as constants.                     

    userinfo(3,"MRV: (2) Rewrite MRV set ");
    userinfo(4,"         i.e. rewrite ", S);

    s:= LIMIT::rewrite( S,__LIMITx );
    if contains( {UNKNOWN,FAIL,"tryIV"},s ) then
        return( s )
    end_if;
    
    userinfo(4,"MRV:      result is ",s);
    omega:= op(s,1);
    S:= op(s,2);
    ex:= combine(eval(subs(ex,op(S,i)=op(s,[3,i]) $ i=1..nops(S),omega=LIMIT::OMEGA)));


    // (c) Compute the leading term of the Puiseux series in w around w = 0+ 

    userinfo(3,"MRV: (3) Compute series");
    userinfo(4,"         in OMEGA=0+ of ",ex);

    if not has(ex,LIMIT::OMEGA) then 
        lc:= ex; le:= 0;
    else
        // Series computation:
        siszero:= Series::Puiseux::iszero;
        sysassign(Series::Puiseux::iszero, 
            proc(c) begin
                // apply 'expr' because iszero can be applied to the whole
                // domain object of domain type Series::Puiseux!
                c:= normal(expr(c));
                c:= eval(subs(c, [ln(stdlib::limit::OMEGA)=ln(omega),stdlib::limit::OMEGA=omega]));
                if iszero(c) or iszero(simplify(c,exp)) or iszero(simplify(c,ln)) then
                    return( TRUE )
                else
                    return( FALSE )
                end_if
            end 
        ):

        s:= series(ex,LIMIT::OMEGA,hold(Right));
        if domtype(s) = Series::Puiseux then
            // The sum of two Series::Puiseux is normalized, but
            // without calling Puiseux::iszero!
            s:= Series::Puiseux::normal(s);
        end_if;
        sysassign( Series::Puiseux::iszero, siszero ):

        case domtype(s)
        of Series::Puiseux do
            le:= Series::Puiseux::ldegree(s);
            lc:= Series::Puiseux::lcoeff(s);

            if lc = FAIL then
                if Series::Puiseux::order(s) > 0 then
                    // the limit is zero
                    return( [0,omega,1] )
                else
                    if LIMIT::printWarningsFlag = TRUE then
                        warning("ORDER seems to be not big enough for series computation");
                    end_if;
                    return( UNKNOWN )
                end_if
            end_if;
            break
        of Series::gseries do
            userinfo(4,"MRV: series is of type 'Series::gseries': ",s);
            t:= Series::gseries::lterm(s);
            if not has(t,LIMIT::OMEGA) then
                le:= 0;
                lc:= Series::gseries::lmonomial(s)
            else
                // handle special gseries expansions, such as x^n:
                case type(t)
                of "ln" do
                    userinfo(4,"MRV: expression has a log-singularity");
                    le:= 0;
                    lc:= Series::gseries::lmonomial(s);
                    break
                of "_power" do 
                    if op(t,1) = LIMIT::OMEGA or op(t,1) = -LIMIT::OMEGA then
                        userinfo(4,"MRV: series has real exponents");
                        le:= sign(Re(op(t,2)));
                        if domtype(le) <> DOM_INT then
                            // can't determine leading degree
                            return( FAIL )
                        end_if;
                        lc:= Series::gseries::lcoeff(s);
                        if op(t,1) = -LIMIT::OMEGA then
                            lc:= lc * (-1)^le
                        end_if
                    else
                        // can't handle given case
                        return( FAIL )
                    end_if;
                    break
                of "exp" do 
                    // case exp(b*ln(+/-LIMIT::OMEGA)) ?
                    t:= op(t,1);
                    if type(t) = "_mult" then
                        t:= [op(t)];
                        i:= 0;
                        while (i:= contains( map(t,type),"ln",i+1 )) <> 0 and op(t[i],1) <> LIMIT::OMEGA and op(t[i],1) <> -LIMIT::OMEGA do end_while;
                        if i = 0 then
                            // can't handle given case
                            return( FAIL )
                        else
                            le:= sign( Re(_mult(op(t,1..i-1),op(t,i+1..nops(t)))) );
                            if domtype(le) = DOM_INT then
                                lc:= Series::gseries::lcoeff(s);
                                if op(t[i],1) = -LIMIT::OMEGA then
                                    // exp(a*ln(-w)) = exp(I*PI)^a * exp(a*ln(w)) = (-1)^a * exp(a*ln(w))
                                    lc:= lc * (-1)^le
                                end_if
                            else
                                // can't determine leading degree
                                return( FAIL )
                            end_if
                        end_if
                    else
                        userinfo(3,"MRV: can't handle given case");
                        return( FAIL )
                    end_if;
                    break
                otherwise
                    lc:= Series::gseries::lmonomial(s);
                    le:= 0
                end_case
            end_if;
            break
        otherwise
            userinfo(3,"MRV: ** series computation fails");
            userinfo(4,"MRV: ** on ",ex );
            return( FAIL )
        end_case
    end_if;

    if has( lc,LIMIT::OMEGA ) then
        lc:= eval(subs(lc, [ln(LIMIT::OMEGA)=ln(omega),LIMIT::OMEGA=omega] ));
        lc:= simplify(lc,ln);
        // the simplification routine could have rewrite exp-log terms exp(b*ln(a)) 
        // to the form a^b, so make sure that lc has the correct form:
        lc:= LIMIT::simp2explog( lc,__LIMITx );
    elif not testtype(lc,Type::Numeric) then
        lc:= simplify(lc);
        // the simplification routine could have rewrite exp-log terms exp(b*ln(a)) 
        // to the form a^b, so make sure that lc has the correct form:
        lc:= LIMIT::simp2explog( lc,__LIMITx );
    end_if;

    userinfo(4,"MRV: lcoeff is ",lc, "; ldegree is ",le);
    if le = 0 then
        if lc = ex0 then
            userinfo(3,"entered in infinite loop!");
            return( FAIL ) // to avoid infinite loops
        else
            return( LIMIT::lterm(lc,__LIMITx) )
        end_if
    else
        return( [lc,omega,le] )
    end_if
end_proc:

/*
        getMRV(f,x)

        f -- expression
        x -- limit variable

        Find the most rapidly varying subexpressions at x=infinity
        (see section 3.1).

        Return Values:
        S       - set of MRV subexpressions (possibly empty).
        FAIL    - limit cannot be computed.
        UNKNOWN - ORDER for series seems to be not big enough.
        "tryIV" - MRV algorithm fails, but interval calculus may succeed.
*/
LIMIT::getMRV:= proc(f,__LIMITx)
    name stdlib::limit::getMRV;
    option remember;
    local s, t, g;
begin
    if not has(f,__LIMITx) then
        return( {} )
    elif f = __LIMITx then
        return( {__LIMITx} )
    else
        case type(f)
        of "_mult" do
        of "_plus" do
            t:= {};
            for g in select( {op(f)},has,__LIMITx ) do
                s:= LIMIT::getMRV(g,__LIMITx);
                if contains( {UNKNOWN,FAIL,"tryIV"},s ) then
                    return( s )
                elif domtype(s) <> DOM_SET then
                    // the limit could be determined directly!
                    t:= t union { {s} }
                else 
                    t:= t union {s}
                end_if
            end_for;
            return(LIMIT::maxMRV( t,__LIMITx ))
    
        of "_power" do
            if not has(op(f,2),__LIMITx) then
                return( LIMIT::getMRV(op(f,1),__LIMITx) )
            /* frankp: gibt Probleme, wenn {0} nach maxMRV geschickt wird! 
            elif iszero(op(f,1)) = TRUE then
                return( {0} )
            */
            elif iszero(op(f,1)) = TRUE then
                // frankp: Diese Fall sollte nicht auftreten!
                return( FAIL )
            else
                return( LIMIT::getMRV(exp(op(f,2)*ln(op(f,1))),__LIMITx) )
            end_if
    
        of "ln" do
            if f = ln(__LIMITx) then return( {__LIMITx} ) 
            else 
                return( LIMIT::getMRV(op(f,1),__LIMITx) )
            end_if
    
        of "exp" do
            if f = exp(__LIMITx) then return( {f} ) end_if;

            t:= op(f,1);
            s:= LIMIT::limitMRV(t,__LIMITx);
            if contains( {UNKNOWN,FAIL,"tryIV"},s ) then
                return( s )
            elif s = infinity or s = -infinity then
                // if the limit of t is real, then exp(t)>=0
                // The new assume/is implementation should handle this case.
                // if is(s,Type::Real)=TRUE then assume(f>0,_and) end_if;
                s:= LIMIT::getMRV(t,__LIMITx);
                if contains( {UNKNOWN,FAIL,"tryIV"},s ) then
                    return( s )
                end_if;
                return(LIMIT::maxMRV({{exp(t)},s},__LIMITx))
            elif not has(s,infinity) then
                // The new assume/is implementation should handle this case.
                // if is(s,Type::Real)=TRUE then assume(f>0,_and) end_if;
                return( LIMIT::getMRV(t,__LIMITx) )
            else
                return( FAIL )
            end_if
    
        of "sin"  do
        of "cos"  do
            t:= LIMIT::limitMRV(op(f,1),__LIMITx);
            if contains( {UNKNOWN,FAIL,"tryIV"},t ) then
                return( t )
            elif domtype(t) = stdlib::Infinity then
                // t = infinity or t = -infinity: try interval calculus!
                return( "tryIV" )
            else
                return( LIMIT::getMRV(op(f,1),__LIMITx) )
            end_if
    
        of "tan"    do
        of "arcsin" do
        of "arccos" do
        of "cot"    do
            t:= LIMIT::limitMRV(op(f,1),__LIMITx);
            if contains( {UNKNOWN,FAIL,"tryIV"},t ) then
                return( t )
            elif t <> infinity and t <> -infinity then
                return( LIMIT::getMRV(op(f,1),__LIMITx) )
            else
                // essential singularity
                return( FAIL )
            end_if
    
        of "arctan" do
            if has( {op(f)},__LIMITx ) then
                return( FAIL )
            else
                return( {} )
            end_if
    
        of "arcsinh" do
        of "arccosh" do
        of "psi"     do
            t:= LIMIT::limitMRV(op(f,1),__LIMITx);
            if contains( {UNKNOWN,FAIL,"tryIV"},t ) then
                return( t )
            else
                s:= sign(t);
                if s = 1 or s = 0 then
                    return( LIMIT::getMRV(op(f,1),__LIMITx) )
                else
                    userinfo( 1,"** ".type(f).": can't handle point ",t);
                    return( FAIL )
                end_if
            end_if
    
        of "arctanh" do
            t:= LIMIT::limitMRV(op(f,1),__LIMITx);
            if contains( {UNKNOWN,FAIL,"tryIV"},t ) then
                return( t )
            elif contains( {1,-1},t ) then
                // singularity
                return( FAIL )
            else
                return( LIMIT::getMRV(op(f,1),__LIMITx) )
            end_if

        of "Ei" do
        of "Si" do
            return( LIMIT::getMRV(op(f,1),__LIMITx) )

        of "besselJ" do // besselJ(n,x)
            t:= LIMIT::limitMRV(op(f,2),__LIMITx);
            if contains( {UNKNOWN,FAIL,"tryIV"},t ) then
                return( t )
            elif t = -infinity then
                // essential singularity 
                return( FAIL )
            elif t = infinity or not has(t,infinity) then
                return( LIMIT::getMRV(op(f,2),__LIMITx) )
            else 
                // complex*infinity or sign(..)*infinity
                return( FAIL )
            end_if
        of "dilog" do
            // series is powerfull enough to handle this case:
            return( LIMIT::getMRV(op(f,1),__LIMITx) )
        // -------------------------------------------------------
        // functions which were treated by the pre-processing step
        of "erf"           do
        of "limit::erfs"   do
        of "gamma"         do
        of "limit::gammas" do
            return( LIMIT::getMRV(op(f,1),__LIMITx) )
        end_case
    end_if;

    return( FAIL )
end_proc:

/*
        rewrite(S,x)

        S -- mrv set of f in respect to x
        x -- limit variable

        Returns [w,S,l] where w represents OMEGA and l is 
        the list of elements of S rewritten in OMEGA.
        The i.th element of l is the i.th element of S rewritten
        in OMEGA (i=1..|S|).

        Return Values:
        [w,S,l] - see above.
        FAIL    - limit cannot be computed.
        UNKNOWN - ORDER for series seems to be not big enough.
        "tryIV" - MRV algorithm fails, but interval calculus may succeed.
*/
LIMIT::rewrite:= proc(S,__LIMITx)
    name stdlib::limit::rewrite;
    option remember;
    local _S_, t, s, i, w, INVw, f, c, l;
begin
    if not has( S,__LIMITx ) then return( [__LIMITx,S,S] ) end_if;
    
    
    // extract the set _S_ of expressions which contains subexpressions
    // of the same comparability class:
    _S_:= {};
    for i from 1 to nops(S) do
        s := op(S,i);
        w := select( S minus {s}, has, s );
        if w <> {} then
            _S_:= _S_ union w
        end_if
    end_for;
    // the complement of _S_:
    S:= S minus _S_;

    w:= op(sort( [op(S)],proc() begin 
        if length(args(1)) < length(args(2)) then TRUE else FALSE end_if
    end_proc ),1);
    if type(w) <> "exp" then
        error("illegal element in mrv set: ".expr2text(w))
    end_if;

    S:= (sort([op(_S_)],
      (x,y) -> (if has(x,y) then TRUE else FALSE end_if)
    )) . [op(S minus {w})];

    s:= LIMIT::sign(op(w),__LIMITx);
    if s = 1 then 
        t:= -op(w); w := 1/w
    elif s = -1 then 
        t:= op(w);
    else
        error("w = ".expr2text(w)." has illegal sign: ".expr2text(s))
    end_if;
    INVw := exp(-t);

    userinfo(4,"MRV: selected OMEGA of ",_S_, "is ",w);

    l := [];
    for f in S do
        // The assume/is facility needs some help ...
        if type(f) = "exp" then 
            f:= op(f,1) 
        else 
            f:= simplify(ln(f),ln)
        end_if;

        c:= LIMIT::lterm( f/t,__LIMITx );

        if contains( {UNKNOWN,FAIL,"tryIV"},c ) then
            userinfo(4,"MRV: ** unable to compute the leading term");	
            return( c )
        elif c[3] <> 0 then
            userinfo(4,"MRV: ** elements are not in the same comparability class!");
            return( FAIL )
        else
            l:= append( l,exp(f-c[1]*t)*LIMIT::OMEGA^c[1] );
        end_if
    end_for;

    return( [w,S.[INVw],l.[1/LIMIT::OMEGA]] ) // append 1/w to S
end_proc:

/*
        maxMRV(s,x)

        s -- set containing sets of rapidly varying terms in x
        x -- limit variable

        Find the more rapidly varying set (see section 3.2).

        Return Values:
        S       - the more rapidly varying set.
        FAIL    - limit cannot be computed.
        UNKNOWN - ORDER for series seems to be not big enough.
        "tryIV" - MRV algorithm fails, but interval calculus may succeed.
*/
LIMIT::maxMRV:= proc(l,__LIMITx)
    name stdlib::limit::maxMRV;
    option remember;
    local f, g, s;
begin
    l:= l minus {{}};
    case nops(l)
    of 0 do
        return( {} )
    of 1 do
        return( op(l) )
    of 2 do
        // choose representatives:
        f:= op(op(l,1),1); g:= op(op(l,2),1);

        // The new assume/is implementation may handle these cases:
        // if type(f)="exp" then assume(op(f),Type::Real,_and) end_if;
        // if type(g)="exp" then assume(op(g),Type::Real,_and) end_if;

        // The assume/is facility needs some help ...
        if type(f) = "exp" then 
            f:= op(f,1) 
        else 
            f:= simplify( ln(f),ln );
        end_if;
        if type(g) = "exp" then 
            g:= op(g,1) 
        else 
            g:= simplify( ln(g),ln ); 
        end_if;

        s:= LIMIT::limitMRV( normal(f/g),__LIMITx );
        if contains( {UNKNOWN,FAIL,"tryIV"},s ) then
            return( s )
        elif domtype(s) = stdlib::Infinity then 
            // s = +/- infinity:
            return( op(l,1) )
        elif iszero(s) = TRUE then 
            return( op(l,2) )
        else 
            return( op(l,1) union op(l,2) )
        end_if
    otherwise
        f:= op(l,1);
        g:= LIMIT::maxMRV( l minus {f},__LIMITx );
        if contains( {UNKNOWN,FAIL,"tryIV"},g ) then
            return( g )
        else
            return(LIMIT::maxMRV( {f,g},__LIMITx))
        end_if
    end_case
end_proc:

/*
        sign(f,x)

        f -- expression in x
        x -- ident

        Compute the sign of ex, assuming that x goes to infinity.
        Returns FAIL if this cannot be determined.
*/
LIMIT::sign:= proc(ex,__LIMITx)
    name stdlib::limit::sign;
    option remember;
    local s, t, S;
begin
    userinfo(4,"MRV: compute the sign of ",ex);
    if ex = __LIMITx then return( 1 ) end_if;

    if not has(ex,__LIMITx) then
        s:= sign(ex);
        if not has(s,sign) then
            return( s )
        else
            return( FAIL )
        end_if
    end_if;

    case type(ex)
    of "_mult"  do
        S:= 1;
        for t in ex do
            s:= LIMIT::sign(t,__LIMITx);
            if s = FAIL then 
                return( FAIL )
            else
                S:= S * s
            end_if
        end_for;
        return( S )
    of "_power" do
        if domtype(op(ex,2)) = DOM_INT and modp(op(ex,2),2) = 0 then 
            return( 1 )
        elif op(ex,1) = __LIMITx then
            return( 1 )
        elif LIMIT::sign(op(ex,1),__LIMITx) = 1 then
            return( 1 )
        else
            t:= LIMIT::lterm( ex, __LIMITx );
            if t = FAIL then
                return( FAIL )
            else
                return( LIMIT::sign( t[1], __LIMITx ) )
            end_if
        end_if
    of "_plus"do
        t:= LIMIT::lterm(ex,__LIMITx);
        if t = FAIL then
            return( FAIL )
        else
            return( LIMIT::sign(t[1],__LIMITx) )
        end_if
    of "exp" do 
        return( 1 )
    otherwise
        t:= LIMIT::lterm( ex, __LIMITx );
        if t = FAIL then
            return( FAIL )
        else
            return( LIMIT::sign( t[1],__LIMITx ) )
        end_if
    end_case
end_proc:

/*
        subsFUNC(ex,f1,f2)

        ex -- expression
        f1 -- string (type of function)
        f2 -- function

        Substitutes all function calls f(x) of type f1 to 
        function calls f2(x).
*/
LIMIT::subsFUNC:= proc(ex,f1,f2) 
    option remember;
    name stdlib::limit::subsFUNC;
begin
    case domtype(ex) 
    of Series::Puiseux do
        return( eval(map(ex,LIMIT::subsFUNC,f1,f2)) )
    of DOM_EXPR do
        if type(ex) = f1 then ex := f2(op(ex)) end_if;
        return( eval(map(ex,LIMIT::subsFUNC,f1,f2)) )
    otherwise
        return( ex )
    end_case
end_proc:

/********************************************************************
        preProcess(ex,x)

        ex -- expression
        x  -- limit variable

        Define the pre-processing algorithms for some
        smooth functions. 

        These rules are defined as methods of the domain LIMIT::preProcess.
        The methods are called by the type of the regarded function. Two 
        pre-pocessing procedures must be defined, one named by the type of 
        the original function, the second by the type of the corresponding 
        smooth function (for example "erf" and "erfs").
*******************************************************************/
LIMIT::preProcess:= proc(ex,__LIMITx)
    name stdlib::limit::preProcess;
    local p;
begin
    if domtype(ex) = DOM_EXPR then
        if (p:= contains(LIMIT::setPreProcessFunc,op(ex,0))) > 0 and has(ex,__LIMITx) then
            return( slot(LIMIT::DpreProcess,expr2text( LIMIT::setPreProcessFunc[p] ))(
                op(map([op(ex)],LIMIT::preProcess,__LIMITx )), __LIMITx
            ) )
        else
            return( map(ex,LIMIT::preProcess,__LIMITx ) )
        end_if
    else
        return( ex )
    end_if
end_proc:

/*--
        Note: 'hold' is requiered here; otherwise the pre-processing step
        only works for the *object* and not for its name. Thus, if
        for instance, the remember table of the object was changed
        afterwards, the process never works again!
--*/
LIMIT::setPreProcessFunc:= [
    hold(abs),hold(sign),hold(erf),hold(gamma),hold(sinh),hold(cosh),hold(tanh),hold(coth),
    hold(arctan),hold(fact),hold(log),hold(igamma),hold(erfc),hold(binomial)
]:

LIMIT::DpreProcess:= newDomain("stdlib::limit::DpreProcess"):

/*--
        erf(x) = 1-exp(-x^2)*erfs(x)  for x->infty
        erf(x) = -1+exp(-x^2)*erfs(-x)  for x->-infty
--*/
LIMIT::DpreProcess::erf:= proc(ex,x)
    local s;
begin
    s:= limit(ex,x=infinity);
    case s 
    of FAIL do
        return( FAIL )
    of infinity do
        return( 1-exp(-ex^2)*LIMIT::erfs(ex) )
    of -infinity do
        return( -1+exp(-ex^2)*LIMIT::erfs(-ex) )
    otherwise
        return( hold(erf)(ex) )
    end_case
end_proc:
slot(LIMIT::DpreProcess,"limit::erfs"):= proc(ex,x)
    local s;
begin
    s:= limit(ex,x=infinity);
    if s = FAIL then
        return( FAIL )
    elif domtype(s) <> stdlib::Infinity then
        // s <> infinity and s <> -infinity:
        s:= sign(s);
        if s = 1 then 
            return( (1-erf(ex))*exp(ex^2) )
        elif s = -1 then 
            return( (1+erf(-ex))*exp(ex^2) )
        else 
            return( FAIL )
        end_if
    else
        return( subsop(hold(X)(ex),0=LIMIT::erfs) )
    end_if
end_proc:

/*--
        erfc(x) = 1-erf(x)
--*/
LIMIT::DpreProcess::erfc:= proc(ex,x)
begin
    1-LIMIT::DpreProcess::erf(ex,x)
end_proc:

LIMIT::DpreProcess::igamma:= proc(a,ex,x)
    local s;
begin
    s:= limit(ex,x=infinity);
    case s 
    of FAIL do
        return( FAIL )        
    of infinity do
        // igamma(a,infinity) -> 0
        return( 1/x )
    otherwise
        return( hold(igamma)(a,ex) )
    end_case
end_proc:

/*--
        gamma(x) = exp( gammas(x) )  for x->infty
--*/
LIMIT::DpreProcess::gamma:= proc(ex,x)
    local s;
begin
    s:= limit(ex,x=infinity);
    if s = FAIL then 
        return( FAIL )
    elif s = infinity then
        return( exp(LIMIT::gammas(ex)) )
    else
        return( hold(gamma)(ex) )
    end_if
end_proc:
slot(LIMIT::DpreProcess,"limit::gammas"):= proc(ex,x)
    local s;
begin
    s:= limit(ex,x=infinity);
    if s = FAIL then 
        return( FAIL )
    elif domtype(s) <> stdlib::Infinity then
        // s <> infinity and s <> -infinity:
        if sign(s) = 1 then 
            return( ln(gamma(ex)) )
        else 
            return( FAIL )
        end_if
    else
        return( subsop(hold(X)(ex),0=LIMIT::gammas) )
    end_if
end_proc:

/*--
                   /  1  if sign(x) -> 1
        sign(x) = <
                   \  -1 if sign(x) -> -1
--*/
LIMIT::DpreProcess::sign:= proc(ex,x)
    local t;
begin
    t:= LIMIT::sign(ex,x);
    if t = 1 then 1 
    elif t = -1 then -1
    else hold(sign)(ex)
    end_if
end_proc:

/*--
        abs(x)  = x/sign(x)
--*/
LIMIT::DpreProcess::abs:= proc(ex,x) begin  
    ex/LIMIT::DpreProcess::sign(ex,x)
end_proc:

/*--
        fact(x) = gamma(x+1)
--*/
LIMIT::DpreProcess::fact:= proc(ex,x) 
begin
    LIMIT::DpreProcess::gamma(ex+1,x)
end_proc:

/*--
        binomial(n,k) = n*gamma(n)/(k*gamma(k)*(n-k)*gamma(n-k)
--*/
LIMIT::DpreProcess::binomial:= proc(n,k,x) 
begin
    n*LIMIT::DpreProcess::gamma(n, x) / k / LIMIT::DpreProcess::gamma(k, x) / (n-k) / LIMIT::DpreProcess::gamma(n-k, x)
end_proc:

LIMIT::DpreProcess::arctan:= proc(ex,x)
    local t;
begin
    if args(0) = 3 then
         // two-argument arcus tangens
         ex:= x/ex; x:= args(3)
    end_if;
    case (t:= limit(ex,x=infinity))
    of -infinity do
        return( -PI/2 )
    of infinity do
        return( PI/2 )
    of FAIL do
        return( FAIL )
    otherwise
        return( hold(arctan)(t) )
    end_case
end_proc:

LIMIT::DpreProcess::sinh:= x -> (exp(x)-exp(-x))/2:
LIMIT::DpreProcess::cosh:= x -> (exp(x)+exp(-x))/2:
LIMIT::DpreProcess::tanh:= x -> (exp(x)-exp(-x))/(exp(x)+exp(-x)):
LIMIT::DpreProcess::coth:= x -> (exp(x)+exp(-x))/(exp(x)-exp(-x)):
LIMIT::DpreProcess::log:= (b,x) -> ln(x)/ln(b):

/*******************************************************************************/

//------------------------------------------
//        The function limit::erfs(x)
//------------------------------------------
LIMIT::erfs:= funcenv( proc() begin procname(args()) end ):

LIMIT::erfs:= slot( LIMIT::erfs, "print", "limit::erfs" ):

LIMIT::erfs:= slot( LIMIT::erfs, "series", 
    proc(f,x,n)
        local t, s, k;
    begin
        t:= Series::series(f,x,n);
        if not testtype(t,Type::Series(Puiseux)) then   
            error("unable to compute series expansion")
        end_if;
        k:= ldegree(t);
        if k = 0 then
            t := expr(t);
            s := Series::series((1-erf(t))*exp(t^2),x,n);
            return( LIMIT::subsFUNC(s,"erf",e -> 1-exp(-e^2)*LIMIT::erfs(e)) )
        elif k < 0 then 
            // f goes to +/- infinity
            if sign(lcoeff(t)) = 1 then 
                // f goes to +infinity
                s:= Series::Puiseux::create(
                    1,1,n,[(1/sqrt(PI)*(-1/4)^s*fact(2*s)/fact(s),0) $ s=0..iquo(n-1,2)],x
                );
                return( Series::Puiseux::_fconcat(Series::series(1/f,x,n),s) )
            end_if
        else
            return( Series::unknown(LIMIT::erfs(f),x,n) )
       end_if
    end_proc
):

LIMIT::erfs:= slot( LIMIT::erfs, "type", "limit::erfs" ):

LIMIT::erfs:= slot( LIMIT::erfs, "diff", 
    proc(f,x)
    begin
        case args(0)
        of 1 do 
            return( f )
        of 2 do
            f:= op(f);
            if has(f,x) then
                return( (-2/sqrt(PI)+2*LIMIT::erfs(f)*f)*diff(f,x) )
            else 
                return( 0 )
            end_if
        otherwise
            return( diff( diff(f,x),args(3..args(0)) ) )
        end_case
    end_proc
):

LIMIT::erfs:= slot( LIMIT::erfs, "stdlib::limit::sign", 
    proc(f,x)
        local s;
    begin
        s:= LIMIT::limitMRV(f,x);
        if contains( {FAIL,UNKNOWN,"tryIV"},s ) then
            return( FAIL )
        else
            s:= sign(s);
            if s = 0 or s = 1 then return( 1 ) else return( FAIL ) end_if
        end_if
    end_proc
):

//------------------------------------------
//        The function limit::gammas(x)
//------------------------------------------
LIMIT::gammas:= funcenv( proc() begin procname(args()) end ):

LIMIT::gammas:= slot( LIMIT::gammas, "print", "limit::gammas" ):

LIMIT::gammas:= slot( LIMIT::gammas, "series",
    proc(f,x,n)
        local t, s, k;
    begin
        t:= Series::series(f,x,n);
        if not testtype(t,Type::Series(Puiseux)) then
            error("unable to compute series expansion")
        end_if;
        k:= ldegree(t);
        if k = 0 then
            t:= expr(t);
            s:= Series::series(ln(gamma(t)),x,n);
            return( LIMIT::subsFUNC(s,"gamma",e -> exp(LIMIT::gammas(e))) )
        elif k < 0 then 
            // f goes to +/- infinity
            s:= Series::Puiseux::create(
                1,-1,13,[(-ln(x)-1),1/2*ln(x)+1/2*ln(2*PI),1/12,0,-1/360,0,1/1260,0,-1/1680, 
                0,1/1188,0,-691/360360],x
            );
            return( Series::series( subs(expr(s),x=1/expr(t)),x,n ) )
        else
            return( Series::unknown(LIMIT::gammas(f),x,n) )
       end_if
    end_proc
):

LIMIT::gammas:= slot(LIMIT::gammas, "type", "limit::gammas" ):

LIMIT::gammas:= slot(LIMIT::gammas, "diff",
    proc(f,x)
    begin
        case args(0)
        of 1 do 
            return( f )
        of 2 do
            f:= op(f);
            if has(f,x) then
                return( psi(f)*diff(f,x) )
            else
                return( 0 )
            end_if
        otherwise
            return( diff( diff(f,x),args(3..args(0)) ) )
        end_case
    end_proc
):

LIMIT::gammas:= slot( LIMIT::gammas, "stdlib::limit::sign",
    proc(f,x)
        local s;
    begin
        s:= LIMIT::limitMRV(f,x);
        if contains( {FAIL,UNKNOWN,"tryIV"},s ) or sign(s) <> 1 then
            return( FAIL )
        else
            return( 1 )
        end_if
    end_proc):

/*******************************************************************************/

unalias( LIMIT ):

