// $Date: 2000/11/30 19:21:27 $ $Author: stefanw $ $Revision: 1.85 $ 

// piecewise - the domain of piecewise defined functions 


// an object of type piecewise has one operand:

// 1 - a list of branches


// the value of the piecewise defined function is the first expression
// the condition of which is satisfied
// the expression may be any object; however, the user is responsible
// not to use any operations that do not make sense







// branch - the domain of branches
// each branch has two operands
// 1. a condition
// 2. an expression





alias(BRANCH=stdlib::branch):

BRANCH:=newDomain("stdlib::branch"):
BRANCH::info:="branch - the domain of branches":

BRANCH::testtype:= (x, T) -> bool(type(x)=T):


BRANCH::new:=
proc(cond, expression)
begin
  if args(0)=2 then
    piecewise::simplifyCondition(cond);
    case %
      of TRUE do
        return(new(BRANCH, TRUE, expression))
      of FALSE do
        return(new(BRANCH, FALSE, undefined))
      otherwise
        return(new(BRANCH, %, expression))
    end_case;
  end_if;  

  if args(0)=0 then
    return(new(BRANCH, FALSE, undefined))
  end_if;
  
  if args(0)<>1 then
    error("Wrong number of arguments")
  end_if;
  
  
  BRANCH::convert(cond);
  if %=FAIL then
    error("Cannot convert to type branch")
  else
    %
  end_if
end_proc:

BRANCH::convert:=
proc(x)
begin
  if args(0)=0 then
    return(new(BRANCH, FALSE, undefined))
  end_if;
  
  case type(x)
    of BRANCH do
      return(x)                       
    of piecewise do
      if nops(x)<>1 then
        return(FAIL)
      else  
        return(op(x,1))
      end_if
    of DOM_LIST do
      if nops(x)<>2 then
        return(FAIL)
      else  
        return(BRANCH::new(op(x,1), op(x,2)))
      end_if;
    otherwise
      FAIL
  end_case
end_proc:


BRANCH::condition:= br -> extop(br,1);

BRANCH::expression:= br -> extop(br,2);

BRANCH::apply:= 
proc(br)
  local errlevel, result;
begin
  errlevel:=traperror(
                      (result:=new(BRANCH, extop(br,1),
                                   args(2)(extop(br,2), args(3..args(0)))))
                      );
  if errlevel<>0 then
  /*  warning("Leaving out branch ".expr2text(extop(br,1)).
            " which produced an error"); */
    userinfo(1, "Leaving out branch ".expr2text(br).
             " which produced an error");
    return(new(BRANCH, FALSE, undefined))
  end_if;
  result
end_proc:

BRANCH::applyToCondition:=
proc(br)
begin
  new(BRANCH, piecewise::simplifyCondition
      (args(2)(extop(br,1),args(3..args(0)))),
      op(br,2))
end_proc:


BRANCH::has:= (br, obj) -> has([extop(br)], obj):

BRANCH::print:=
proc(br)
  local L1, L2, len;
begin
  L1 := expr2text(BRANCH::expression(br));
  L2 := expr2text(BRANCH::condition(br));
  len := length(L1);
  stringlib::format(L1, len)."  if ".
           L2."\n";
end_proc:

piecewise:=newDomain("piecewise"):

piecewise::info:="Piecewise -- the domain of piecewise defined functions":

piecewise::new:=
proc()
  option hold;
  local i, result, newbranch, argument, errlev;
begin
  if args(0)=0 then
    return(undefined)
  end_if;

  if args(0)<>1 then
    // convert each argument to branch
    result:=null();
    for i from 1 to args(0) do
      errlev:=traperror((argument:= context(args(i))));
      if errlev<>0 then
        userinfo(2, output::ordinal(i)." branch skipped because of an error");
        next
      end_if;
      if type(argument)=DOM_LIST then
        newbranch:=BRANCH(argument[1], argument[2])
      else              
        newbranch:=BRANCH(argument)
      end_if;
        case op(newbranch, 1)
          of TRUE do
            return(op(newbranch,2))
          of FALSE do
            break
          otherwise
            result:=result, newbranch
        end_case
    end_for;
    if nops(result)=0 then
      return(undefined)
    else  
      return(piecewise::flatten(new(piecewise, result)))
    end_if;
  end_if;

  errlev:=traperror((argument:=context(args(1))));
  if errlev<>0 then
    return(undefined)
  end_if;

  if type(argument)="_exprseq" then
    // force flattening
    return(eval(hold(piecewise::new)(argument)))
  end_if;
  
  piecewise::convert(argument);
  if %=FAIL then
    error("Cannot convert to type piecewise")
  else
    piecewise::flatten(%)
  end_if
end_proc:

piecewise::convert:=
proc(x)
  local br:BRANCH;
begin
  if args(0)=0 then
    error("Wrong number of arguments")
  end_if;
  
  case type(x)
    of piecewise do
      return(x)
    of BRANCH do
      br:=x;
      break
    otherwise
      br:=BRANCH(x);
  end_case;
  // here br is of type branch
  case BRANCH::condition(br)
    of TRUE do
      return(BRANCH::expression(br))
    of FALSE do
      return(undefined)
    otherwise
      return(new(piecewise, br))
  end_case;
end_proc:

piecewise::convert_to:=
proc(x:piecewise, T:DOM_DOMAIN)

begin
  case T
    of piecewise do
      return(x)
    of DOM_LIST do
      return([extop(x)])
    otherwise
      FAIL
  end_case
end_proc:

      
piecewise::testtype:= (x, T) -> bool(type(x)=T):

piecewise::condition:=
proc(pw:piecewise,n:DOM_INT)
// returns the n-th condition 
begin
  BRANCH::condition(op(pw,n))
end_proc:

piecewise::expression:=
proc(pw:piecewise,n:DOM_INT)
// returns the expression that is valid if the n-th condition holds 
begin
  BRANCH::expression(op(pw,n))
end_proc:


// piecewise::simplifyCondition
// a replacement for property::simpex that also uses the solver


piecewise::simplifyCondition:=
proc(cond)
  local newcond, newcond2, S, inds;

begin
  if contains({"_and", "_or", "_not"}, type(cond)) then
    return(property::simpex(map(cond, piecewise::simplifyCondition)))
  end_if;

  // emergency exit for very long conditions
  if length(cond)>40 and nops(indets(cond) minus Type::ConstantIdents)>1 then
    return(cond)
  end_if;
  
  newcond:=property::simpex(cond);
  
  inds:= indets(newcond) minus Type::ConstantIdents;
  if nops(inds)<>1 or type(newcond)="_in" then
    return(newcond)
  end_if;
  if contains({"_equal", "_less", "_leequal", "_unequal"}, type(newcond)) then
    S:=solve(newcond, op(inds,1));
    if S=C_ then
      return(TRUE)
    end_if;
    if type(S)="_minus" and op(S,1)=C_ then
      return(not property::simpex(op(inds,1) in op(S,2)))
    end_if;
    newcond2:=op(inds,1) in S;
    property::simpex(newcond2)
  else
    newcond
  end_if;
end_proc:
  

piecewise::nops:= extnops;

piecewise::op:= extop;

/*++

	map - application of a function to all expressions

++*/


piecewise::map:=
proc(pw)
begin
  piecewise::flatten(piecewise::new(
     BRANCH::apply(op(pw,i), args(2..args(0))) $i=1..nops(pw)))
end_proc:

/*

   mapMap - map a "map" command to all branches
            but simply map to objects that are not piecewise

*/


piecewise::mapMap:=
proc(pw)
begin
  if type(pw)<>piecewise then
    map(pw, args(2..args(0)))
  else
    map(pw, map, args(2..args(0)))
  end_if
end_proc:



/*

 mapOp - map an "op" command to all branches
         simply extract an op of objects that are not piecewise

*/

piecewise::mapOp:=
proc(pw)
  begin
  if type(pw)<>piecewise then
    op(pw, args(2..args(0)))
  else
    map(pw, op, args(2..args(0)))
  end_if
end_proc:


/*++

	mapConditions - application of a function to all
			conditions

++*/

piecewise::mapConditions:=
proc(pw)
  local i, result;
begin
  result:=null();
  for i from 1 to nops(pw) do
    BRANCH::applyToCondition(op(pw,i), args(2..args(0)));
    case BRANCH::condition(%)
      of TRUE do
        return(BRANCH::expression(%))
      of FALSE do
        break;
      otherwise
        result:=result, %
    end_case;
  end_for;
  if nops(result)=0 then
    return(undefined)
  else
    return(piecewise::flatten(new(piecewise, result)))
  end_if;
end_proc:


/* piecewise::selectConditions(pw, f)

keep only those conditions for which f(condition)=TRUE

*/

piecewise::selectConditions:=
proc(pw,f)
local argv;
begin
  argv:=args();
  select([extop(pw)], x-> f(op(x,1), op(argv, 3..nops(argv))));
  if nops(%)=0 then
    undefined
  else
    piecewise::new(op(%))
  end_if;
end_proc:

/* piecewise::splitConditions(pw, f)

split pw into a list of three piecewise objects depending on whether
f(condition) gives TRUE, FALSE, or UNKNOWN

*/

piecewise::splitConditions:=
proc(pw, f)
local argv;
begin
  argv:=args();
  split([op(pw)], x-> f(op(x,1), op(argv, 3..nops(argv))));
  map(%, u-> if nops(u)>0 then
               piecewise::new(op(u))
             else
               undefined
             end_if)
end_proc:
  
/*++ 
	zip - argumentwise application of a law of composition 
	for two or more piecewise defined functions
++*/

piecewise::zip:=
proc(pwf, pwg, f)
  local i,j,result,cond,g,conds_so_far;
begin
  if type(pwf)<>piecewise then
    g:= x -> f(pwf, x);
    return(piecewise::map(pwg,g))
  elif type(pwg)<>piecewise then
    return(piecewise::map(pwf, f, pwg))
  end_if;
  result:=[];
  conds_so_far:=[];
  for i from 1 to nops(pwf) do
    for j from 1 to nops(pwg) do
      cond:=piecewise::simplifyCondition(piecewise::condition(pwf,i) and
      piecewise::condition(pwg,j));
      // possibility check 
      case cond
        of TRUE do
          return(f(piecewise::expression(pwf,i),
                          piecewise::expression(pwg,j)))
        of FALSE do
          break
        otherwise
          if contains(conds_so_far, cond)=0 then
            result:=append(result,
                           BRANCH(cond,
                            f(piecewise::expression(pwf,i),
                              piecewise::expression(pwg,j))));
            conds_so_far:=append(conds_so_far, cond)
          end_if
      end_case;
	end_for
  end_for;
  if nops(result)= 0 then
    undefined
  else
    piecewise::new(op(result))
  end_if
end_proc:


piecewise::evaluate:=
proc(pw)
  local i, result;
begin
  result:=null();
  pw:=piecewise::flatten(pw);
  for i from 1 to nops(pw) do
    piecewise::simplifyCondition(piecewise::condition(pw,i));
    case %
      of TRUE do
        return(eval(piecewise::expression(pw, i)))
      of FALSE do
        break
      otherwise
        result:=result, BRANCH(%, eval(piecewise::expression(pw, i)))
    end_case;
  end_for;
  if nops(result)=0 then
    return(undefined)
  else  
    new(piecewise, result)
  end_if
end_proc:


piecewise::subs:=
proc(pw:piecewise)
begin  
  piecewise::new(subs(extop(pw), args(2..args(0))))
end_proc:
                   
piecewise::insert:=
proc(pwf, l, no: Type::PosInt)

begin
  if no>nops(pwf)+1 then
    error("Illegal position")
  end_if;
  BRANCH::new(l);
  new(piecewise,
      op(pwf, 1..no-1), l, op(pwf, no..nops(pwf)))
end_proc:

piecewise::remove:=
proc(pwf, no: Type::PosInt)

begin
if nops(pwf)<no then
	error("Illegal position")
end_if;
new(piecewise, op(pwf,1..no-1), op(pwf, no+1..nops(pwf)))
end_proc:


/*

disregardPoints: assume every condition of type x=C to be FALSE
and every condition of type x<>C to be TRUE

*/

piecewise::disregardPoints:=
proc(pw)
  local subsCond, result;
begin
  /* subsCond - replace conditions of type _unequal by TRUE
     and conditions of type _equal by FALSE */
  
  subsCond:=
  proc(cond)
    name piecewise::subsCond;
  begin
    case type(cond)
      of "_equal" do
        return(FALSE)
      of "_unequal" do
        return(TRUE)
      of "_or" do 
      of "_not" do
      of "_and" do
        return(map(cond, subsCond))
      otherwise
        cond
    end_case
  end_proc;

  result:=piecewise::mapConditions(pw, subsCond);
  delete subsCond;
  result
end_proc:


// assumeIdentsTRUE(pw, identset) - assume that all conditions in which
// one of the identifiers in identset appears is TRUE


piecewise::assumeIdentsTRUE:=
proc(pw, identset:DOM_SET)
  local subsCond, result;
begin
  subsCond:=
  proc(cond)

  begin
     case type(cond)
      of "_or" do 
      of "_not" do
      of "_and" do
        return(map(cond, subsCond))
      otherwise
        if indets(cond) intersect identset <> {} then
          return(TRUE)
        else
          return(cond)
        end_if
    end_case
  end_proc;

  result:=piecewise::mapConditions(pw, subsCond);  
  delete subsCond;
  result
end_proc:


//  operations on conditions //


// innerOfCondition: restrict condition to interior points 

piecewise::innerOfCondition:=
proc(c)

begin
  case type(c)
    of "_equal" do
      return(FALSE)
    of "_unequal" do
    of "_less" do
      return(c)
    of "_leequal" do
      return(_less(op(c)))
    of "_and" do
    of "_or" do
    of "_not" do
      return(map(c, piecewise::innerOfCondition))
    otherwise
      error("Cannot handle condition of type ".expr2text(type(c)))
  end_case
end_proc:
   
piecewise::borderOfCondition:=
proc(c)
begin
  case type(c)
    of "_equal" do
      return(c)
    of "_unequal" do
    of "_leequal" do
    of "_less" do
      return(_equal(op(c)))
    of "_or" do
      return(map(c, piecewise::borderOfCondition))
    of "_and" do
      return(piecewise::borderOfCondition(op(c,1)) or
             piecewise::borderOfCondition(op(c,2))
             )
    of "_not" do
      return(piecewise::borderOfCondition(op(c,1)))
    otherwise
      error("Cannot handle condition of type ".expr2text(type(c)))
  end_case
end_proc:

piecewise::outerOfCondition:=
proc(c)
begin
  piecewise::innerOfCondition(not c)
end_proc:

piecewise::satisfiesCondition:=
proc(x, value, c)
begin
  is(subs(c, x=value))
end_proc:



//  arithmetical/ set theoretic operations (argumentwise) 

piecewise::_plus:=
proc() 
begin
  if args(0)=1 then
    args(1)
  elif args(0)>2 then
    piecewise::_plus(args(1), _plus(args(2..args(0))))
  else // two arguments 
    piecewise::zip(args(), _plus)
  end_if
end_proc :

piecewise::_mult:=
proc() 
begin
  if args(0)=1 then
    args(1)
  elif args(0)>2 then
    piecewise::_mult(args(1), _mult(args(2..args(0))))
  else // two arguments 
    piecewise::zip(args(),_mult)
  end_if
end_proc :

piecewise::_union:=
proc() 
begin
  if args(0)=1 then
    args(1)
  elif args(0)>2 then
    piecewise::_union(args(1), _union(args(2..args(0))))
  else // two arguments 
    piecewise::zip(args(),_union)
  end_if
end_proc :

piecewise::_intersect:=
proc() 
begin
  if args(0)=1 then
    args(1)
  elif args(0)>2 then
    piecewise::_intersect(args(1), _intersect(args(2..args(0))))
  else // two arguments 
    piecewise::zip(args(),_intersect)
  end_if
end_proc :

piecewise::_negate:=proc() begin piecewise::map(args(1), _negate) end_proc :

piecewise::_invert:=proc() begin piecewise::map(args(1), _invert) end_proc :

piecewise::_subtract:=proc() begin piecewise::zip(args(),_subtract) end_proc :

piecewise::_divide:=proc() begin piecewise::zip(args(),_divide) end_proc :

piecewise::_power:=proc() begin piecewise::zip(args(),_power) end_proc :

piecewise::_div:=proc() begin piecewise::zip(args(),_div) end_proc :

piecewise::_mod:=proc() begin piecewise::zip(args(),_mod) end_proc :

piecewise::_minus:=proc() begin piecewise::zip(args(),_minus) end_proc:

piecewise::_less:=(x,y) -> piecewise::zip(x,y, _less):

piecewise::_leequal:=(x,y) -> piecewise::zip(x,y,_leequal):

// overload solvelib::Union
/* piecewise::Union(pw, x, xset)
   
   returns the union of all pw(x), x varying over xset

   There are two possibilities:
   a) the conditions in pw do not depend on x, i.e. the applicable
      branch is the same for all x \in xset. Then the Union may be
      computed separately for every branch

   b) at least one condition depends on x. Then the result consists of
      the Union over the elements of xset that satisfy the condition
      (only the respective branch must be evaluated) plus the Union
      over all the other elements

*/

piecewise::Union:=
proc(pw:piecewise, x:DOM_IDENT, xset)
  local dep, indep, dummy, dU, iU, newcond;
begin
  if xset={} then
    return({})
  end_if;
  [dep, indep, dummy] := piecewise::splitConditions(pw, has, x);
  if dep<>undefined then
    userinfo(20, expr2text(nops(dep))."conditions depend on variable");
    dU:=piecewise::depUnion(dep, x, xset)
  else
    dU:=undefined
  end_if;
  if dU<>undefined and type(dU)<>piecewise then
   // branch has been recognized to be the only possible
    return(dU)
  end_if;
  iU:=piecewise::indepUnion(indep, x, xset);
// operation cannot give undefined or unconditionally def. object,
// since conditions are not checked in indepUnion 
  if dU=undefined then
    return(iU)
  end_if;

// restrict the dependend conditions by demanding that the indep.
// conditions do not apply

  newcond:= if iU=undefined then
              TRUE
            else
              not _and(piecewise::condition(iU, i) $i=1..nops(iU))
            end_if;
  piecewise::_concat(piecewise::restrict(dU, newcond),iU)
end_proc:

piecewise::indepUnion:= (pw, x, xset) -> map(pw, solvelib::Union, x, xset):

piecewise::depUnion:=
proc(pw:piecewise, x:DOM_IDENT, xset)
begin
_union(solvelib::Union(piecewise::expression(pw, i), x, xset intersect 
      solve(piecewise::condition(pw, i), x)) $i=1..nops(pw))
end_proc:


// special functions

piecewise::sin:= x -> piecewise::map(x, sin):
piecewise::cos:= x -> piecewise::map(x, cos):
piecewise::tan:= x -> piecewise::map(x, tan):
piecewise::exp:= x -> piecewise::map(x, exp):
piecewise::ln := x -> piecewise::map(x, ln):
piecewise::arcsin:= x-> piecewise::map(x, arcsin):
piecewise::arccos:= x -> piecewise::map(x, arccos):
piecewise::arctan:= x -> piecewise::map(x, arctan):
piecewise::sinh := x -> piecewise::map(x, sinh):
piecewise::cosh := x-> piecewise::map(x, cosh):
piecewise::tanh := x -> piecewise::map(x, tanh):
piecewise::coth := x-> piecewise::map(x, coth):
piecewise::arcsinh:= x -> piecewise::map(x, arcsinh):
piecewise::arccosh:= x -> piecewise::map(x, arccosh):
piecewise::arctanh:= x -> piecewise::map(x, arctanh):
piecewise::arccoth:= x -> piecewise::map(x, arccoth):
piecewise::sech:= x-> piecewise::map(x, sech):
piecewise::csch:= x-> piecewise::map(x, csch):
piecewise::sign := x -> piecewise::map(x, sign):
piecewise::abs  := x-> piecewise::map(x, abs):
piecewise::heaviside:= x -> piecewise::map(x, heaviside):

piecewise::Re := x -> piecewise::map(x, Re):
piecewise::Im := x -> piecewise::map(x, Im):

/*++    
	restrict - impose an additional restriction
		   on the piecewise function
++*/

piecewise::restrict:=
proc(pw, cond)

begin
  if type(pw) <> piecewise then
    piecewise(BRANCH(cond, pw))
  else
    piecewise::mapConditions(pw, _and, cond)
  end_if
end_proc:

piecewise::_concat:=
proc()
  local argv;
begin
  argv:=select([args()], _unequal, undefined);
  // the everywhere undefined function is neutral
  // element w.r.t. concatenation
  if map({op(argv)}, type)<>{piecewise} then
    error("All arguments must be of type piecewise")
  end_if;
  new(piecewise,op(map(argv, extop))) // flatten !
end_proc:

/*++
	simplify - try to decide the truth of the condition and
	remove unnecessary conditions
++*/ 

piecewise::simplify:=
proc(pww:piecewise)
  local i, j, result, cond, xpr,pw;
  
begin
  result:=[];
  pw:=piecewise::flatten(pww);
  for i from 1 to nops(pw) do
    // cond:=is(piecewise::condition(pw, i));
    // if cond=UNKNOWN then
      cond:=piecewise::simplifyCondition(piecewise::condition(pw,i));
    // end_if;
    case cond
      of FALSE do
        break; // do not enter this branch into result 
      of TRUE do
        piecewise::expression(pw,i);
        return(eval(%))
      otherwise
        // test whether the same object already exists 
        xpr:=simplify(piecewise::expression(pw, i));
        xpr:=piecewise::subsEqualities(xpr, cond);
        j:=contains(map(result, op, 2), xpr);
        if j>0 then
          // combine old and new condition
          cond:=piecewise::simplifyCondition(result[j][1] or piecewise::condition(pw, i));
          if cond=TRUE then
            return(xpr)
          elif cond=FALSE then
            // should not happen, since result[j][1] was not FALSE
            delete result[j]
          else  
            result[j][1]:=cond;
            result[j][2]:=piecewise::subsEqualities(result[j][2], cond)
          end_if
        else  // cond has still its original value! 
          result:=append(result, [cond, xpr])
        end_if;
    end_case;
  end_for;
  if result=[] then
    undefined
  else
    new(piecewise, op(map(result, BRANCH::new)))
  end_if
end_proc:


// piecewise::flatten - split up conditions and expressions that
// are of type piecewise themselves


piecewise::flatten:=
proc(pw)
begin
  if type(pw)<>piecewise then
    pw
  else  
     new(piecewise, (piecewise::flattenCondition(op(pw,i));
            op(map(%,piecewise::flattenExpression)))
            $i=1..nops(pw))
  end_if
end_proc:

// flattenCondition - takes a branch [cond, expression] of a piecewise
// flattens cond if cond is piecewise itself, does nothing otherwise
// returns a list of branches

piecewise::flattenCondition:=
proc(branch:stdlib::branch):DOM_LIST
  local cond;
begin
  if type(BRANCH::condition(branch))<>piecewise then
    return([branch])
  end_if;
  cond:=piecewise::flatten(BRANCH::condition(branch));
  // the operands of cond are branches [condition.i1, condition.i2]
  // i.e. conditions valid under other conditions
  // turn them into
  // [condition11 and condition12, ...,condition.n1 and condition.n2]
  map([op(cond)], br -> BRANCH::condition(br) and BRANCH::expression(br))
  // for each i, condition.i1 and condition.i2 ->
  //             [condition.i1 and condition.i2, expression]
  [BRANCH::new(op(%,i), branch[2]) $i=1..nops(%)]
end_proc:


piecewise::flattenExpression:=
proc(branch:stdlib::branch)
  local xpr;
begin
  if type(BRANCH::expression(branch))<> piecewise then
    return(branch)
  end_if;
  xpr:= piecewise::flatten(BRANCH::expression(branch));
  // restrict each condition by the outer condition
  op(piecewise::restrict(xpr, BRANCH::condition(branch)))
end_proc:

  
// piecewise::subsEqualities - simplifies a given expression
// taking into account the given condition

piecewise::subsEqualities:=
proc(xpr, cond)
local i, j, result, Lhs, Rhs, newexpr, newcond;
  
begin
  case type(cond)
    of "_equal" do
      subs(xpr, cond);
      break
    // case _less etc.: use assume 
    of "_less" do
    of "_leequal" do
      Lhs:= op(cond, 1);
      Rhs:= op(cond, 2);
      if not has(xpr,Lhs) and not has(xpr, Rhs) then
        return(xpr)
      end_if;
      i:=genident(); j:=genident();
      newexpr:=xpr;
      if testtype(Lhs, Type::Numeric)<>TRUE then
        newexpr:=subs(newexpr, Lhs=i);
        newcond:=subsop(cond, 1=i)
      else
        newcond:=cond
      end_if;
      if testtype(Rhs, Type::Numeric)<>TRUE then
        newexpr:=subs(newexpr, Rhs=j);
        newcond:=subsop(newcond, 2=j)
      end_if;
        // substitute a new variable for the left/right hand side
        // in both the expression and the condition
        // then assume the new condition generated in this way
        // to be true, and simplify the newly generated expression  
      
      /* since assume does not evaluate its arguments */
      eval(subs(hold(assume(dummy)), hold(dummy)=newcond));
      subs(simplify(newexpr), i=Lhs, j=Rhs);
      break
    of "_and" do
      result:=xpr;
      for i in cond do
        result:= piecewise::subsEqualities(result, i)
      end_for;
      result;
      break
    otherwise
      xpr
  end_case
end_proc:



// some methods for expression manipulation, realized by mapping

piecewise::expand:= pw -> map(pw, expand):

piecewise::normal:= pw -> map(pw, normal):

piecewise::combine:= pw -> map(pw, combine, args(2..args(0))):

piecewise::rewrite:= pw -> map(pw, rewrite, args(2..args(0))):


/* solve(pw, x, options) - solves the equation pw=0 for x */

piecewise::solve:=
proc(pw:piecewise, x)
  local otherargs;
begin
  otherargs:=args(3..args(0));
  _union(map(op(pw),
             /* for every branch, fin the set of x that both satisfy the
                given condition and solve the given expression */
             proc(branch)
               name solveBranch;
             begin
               userinfo(30, "Solving case ".
                        expr2text(BRANCH::condition(branch)));
               solve(BRANCH::condition(branch), x) intersect
               solve(BRANCH::expression(branch), x, otherargs)
             end_proc
             ))
end_proc:


/* piecewise::solveConditions

   solve all conditions for the given variable x and reconvert
   the sets of solutions into expressions
*/

piecewise::solveConditions:=
proc(pw:piecewise, x:DOM_IDENT)
begin
  piecewise::mapConditions(pw, cond -> (x in solve(cond,x)))
end_proc:

/*
	discont - determine the discontinuities (the points for which
	          the function is defined on both sides of the point,
		  but the limits from the left/right are not the same
		  or do not agree with the function value at the point
*/

piecewise::discont:=
proc(pw:piecewise, x:DOM_IDENT, F=null())
  local discontlist:DOM_SET,
  i:DOM_INT,
  j:DOM_INT,
  cond, disc;

begin
  if args(0)<2 or args(0)>3 then
    error("Two arguments expected")
  end_if;
    
  discontlist:={};
  for i from 1 to nops(pw) do
    cond:=piecewise::condition(pw, i);
    disc:= solve(piecewise::borderOfCondition(cond, x),x);
    userinfo(10, "Found border points ".expr2text(disc));
    discontlist:= discontlist union
                  select(discont(piecewise::expression(pw, i), x, F),
                         u -> piecewise::satisfiesCondition( x, u, cond)=TRUE );
    for j from 1 to nops(pw) do
      if j<>i then
        userinfo(10, "Checking whether border points are inner points of ".
                 output::ordinal(j)." condition");
        disc:=select(disc, u -> not piecewise::satisfiesCondition
                     (x, u, piecewise::innerOfCondition
                      (piecewise::condition(pw,j))) )
      end_if;
    end_for;
    discontlist:= discontlist union disc;
  end_for;
  userinfo(5, "Discontinuities are ".expr2text(discontlist));
  discontlist
end_proc:

piecewise::indefint :=
proc(pw:piecewise, x:DOM_IDENT)

begin
map(pw, int, x)
end_proc:

piecewise::defint :=
proc(pw:piecewise, eq: "_equal")
  local x, a,b, i:DOM_INT, S, J, cond, res,
  newpw:piecewise, // the rest which cannot be integrated
  covered,         // the subset of [a,b] already covered;
  integral;        // int_{ x \in covered} pw(x) dx
  
begin
  x:=op(eq,1);
  if type(x)<>DOM_IDENT then
    error("First argument must be identifier")
  end_if;
  if type(op(eq,2)) <> "_range" then
    error("Right hand side of second argument must be range")
  end_if;
  a:=op(eq,[2,1]);
  b:=op(eq,[2,2]);

  if is(b<a) then // special case
    return(- piecewise::defint(pw, x=b..a))
  elif b=a then
    return(0)
  end_if;

  // assume that a<=b even if this could not be decided
  
  J:= Dom::Interval([a,b]);
  
  // and the subset already integrated is empty
  // the integral is zero 

  covered:={};
  integral:=0;
  newpw:=[];
  
  for i from 1 to nops(pw) do
    S:=solve((cond:=piecewise::condition(pw,i)), x) intersect
        (J minus covered);
    userinfo(20, "Condition ".expr2text(cond)." satisfied over ".
             expr2text(S));


    res:=intlib::intOverSet(piecewise::expression(pw, i), x, S);

    if type(res)<> "intOverSet" then
      userinfo(20, "Integral over ".expr2text(S)." is ".
               expr2text(res));
      covered:=covered union S;
      integral:=integral+ res;
    else  
      newpw:=append(newpw, op(pw, i))
    end_if;
  
    if type(J minus covered)=DOM_SET then
      return(integral)
    end_if;
  end_for;
    if nops(newpw)>0 then
      newpw:=new(piecewise, op(newpw));
      integral + hold(intlib::intOverSet)(newpw, x, J minus covered)
    else
      // parts of [a,b] were not covered because func is not defined there
      integral
    end_if
end_proc:

piecewise::int :=
proc(pw:piecewise, x)
begin
  if type(x)="_equal" then
    piecewise::defint(pw, x)
  elif type(x)=DOM_IDENT then
    piecewise::indefint(pw, x)
  end_if
end_proc:

piecewise::intOverSet :=
proc(xpr, x, pw:piecewise)

begin
  // the conditions of pw must *not* depend on x
  map(pw, S -> intlib::intOverSet(xpr, x, S))
end_proc:

piecewise::diff :=
proc(pw:piecewise, x)
begin
case args(0)
  of 1 do return(pw)
  of 2 do 
	return(piecewise::map(piecewise::mapConditions
                         (pw, piecewise::innerOfCondition), diff, x))
  otherwise
	piecewise::diff(piecewise::diff(pw, x), args(3..args(0)) )
end_case
end_proc:

piecewise::contains:=
proc(pw:piecewise, x)
begin
  if args(0)<>2 then
    error("Wrong number of arguments")
  end_if;
  piecewise::map(pw, contains, x)
end_proc:

piecewise::has:=
proc(pw:piecewise, x)
begin
has([extop(pw)], x)
end_proc:

// overload _in

piecewise::_in:=
proc(pw:piecewise, S)
begin
  if args(0)<>2 then
    error("Wrong number of arguments")
  end_if;
  _or((piecewise::condition(pw,i) and piecewise::expression(pw,i) in S )
      $i=1..nops(pw))
end_proc:
  
piecewise::set2expr:=
proc(pw:piecewise, x)
  begin
  if args(0)<>2 then
    error("Wrong number of arguments")
  end_if;
  _or((piecewise::condition(pw,i) and x in piecewise::expression(pw,i))
      $i=1..nops(pw))
end_proc:

// since the result of set2prop should be a valid input
// for assume, mapping does not suffice

piecewise::set2prop:=
proc(pw:piecewise, opt)
begin
  [solvelib::set2prop(piecewise::expression(pw, i), opt) $i=1..nops(pw)];
  if has(%, FAIL) then
    FAIL
  else
    _or(op(%))
  end_if
end_proc:

// overload solvelib::getElement
// return FAIL if some conditions cannot be removed

piecewise::getElement:=
proc(pw:piecewise)
begin
  solvelib::getElement(_intersect(piecewise::expression(pw, i) $i=1..nops(pw)))
end_proc:

piecewise::isFinite:=
proc(pw:piecewise):DOM_BOOL
begin
  [solvelib::isFinite(piecewise::expression(pw, i)) $i=1..nops(pw)];
  // is the result independent of the conditions ?
  if nops(%)=1 then
    op(%,1)
  else
    UNKNOWN
  end_if
end_proc:


// overload fp::unapply

piecewise::unapply:=
proc(pw /* ,var1, ...,vark */)
  local listform, Args;
  
begin
  listform:=map(op(pw), u->[op(u,1), op(u,2)]);
  if args(0) = 1 then
    Args:= [ op(indets(pw)) ]
  else
    Args:= [ args(2..args(0)) ]
  end_if;

  if Args = [] then
    subsop(proc() option arrow; begin end, 4=hold(piecewise)(listform),
           Unsimplified)
  else
    subsop(proc() option arrow; begin end, 1=op(Args),
           4=hold(piecewise)
           (subs(listform, [ Args[i]=DOM_VAR(0,1+i) $ i=1..nops(Args) ],
                  Unsimplified)))
  end_if
end_proc:

piecewise::PWif := funcenv(hold(slot)(piecewise, "PWif"),
                           builtin(1100, 3, " if ", "piecewise::PWif")):
piecewise::print := 
proc(pw)
  local i, nargs;
begin
  if PRETTYPRINT=FALSE then
    hold(piecewise)([piecewise::condition(pw,i),
                     piecewise::expression(pw, i)] $i=1..nops(pw))
  else
    nargs := nops(pw);
    hold(piecewise)(piecewise::PWif(piecewise::expression(pw, i), piecewise::condition(pw, i)) $
               i = 1..nargs)
  end_if
end_proc:


piecewise::TeX :=
proc(pw)
begin
  "\\begin{array}{ll}\n".
  _concat((generate::TeX(piecewise::expression(pw,i))." & \\mbox{if} ".
          generate::TeX(piecewise::condition(pw, i))."\\\\\n") $i=1..nops(pw)).
  "\\end{array}"
end_proc:

// Typesetting output of piecewise
piecewise::Content := 
proc(Out, pw)
  local i;
begin
    Out::Cpiecewise(
		 (Out(piecewise::expression(pw, i)),
                 Out::Cpwcondition(Out(piecewise::condition(pw, i))))
                         $ i = 1..nops(pw)
                 )
end_proc:
