/*

Replacement of plotfunc for 2-dimensional function plots

Remarks:
You can specify plot2d objects like 
plotfunc2d([Mode=List,[point(2,2)]],sin(x), x=-PI..PI): 

       
Examples:

plotfunc2d(x^3, x=-3..3):
plotfunc2d(Axes=Corner, Ticks=[Steps=0.5, Steps=10], x^3, x=-3..3):
plotfunc2d(x^3, x=-2..2, y=-4..4):
plotfunc2d(GridLines=Automatic, x^3, x=-2..2, y=-5..5):
plotfunc2d(Ticks=[Steps=[1,2],Steps=2],GridLines=Automatic, 
           x^3, x=-2..2, y=-10..10):
plotfunc2d(ViewingBox=[-2..2,-5..5], x^3, x=-4..4, y=-10..10):
plotfunc2d(ViewingBox=[-3..3,-6..6], x^3, x=-2..2, y=-5..5):

Defined default values for plotting:

plotlib::Options[Scene2d] := { Axes=Origin, Scaling=UnConstrained, 
                               Labels=["x","y"], Labeling=TRUE, 
			       RealValuesOnly=TRUE, Discont=TRUE}:

plotfunc2d(x^3, x=-3..3):

*/

plotfunc2d := proc()
   local arg, axes, check_object, createrange, 
         createtitle, DefaultX, DefaultY,
         discont_mode, gen_x, getrange, grid, gridlines,
	 ind, new_args, nr, plot2d_objects, 
	 real_mode, s, scene_options, t, ti, ticks,
	 x_max, x_min, x_var, xr, 
	 y_max, y_min, y_var, yr, vbox;
begin

   // Definition of a local procedure

   getrange:= proc(rr)
      name plotfunc2d::getrange;
      local r;
    begin
      r:=rr; // to avoid warnings 
      if domtype(op(r,1)) <> DOM_IDENT then error("no variable") end_if;
      if type(op(r,2)) = "_range" then
        r:= (op(r,1) = [op(op(r,2))])
      end_if;
      if domtype(float(op(r,[2,1]))) <> DOM_FLOAT then
        error("illegal plot range")
      end_if;
      if domtype(float(op(r,[2,2]))) <> DOM_FLOAT then
        error("illegal plot range")
      end_if;
      op(r,1) = float(op(r,2))
    end_proc:

    createtitle:= proc()
        name plotfunc2d::createtitle; 
        local s,i;
    begin
        s:= expr2text( args(1) );
        i:= length(s)-1;
        while i > 0 and s[i] = "0" and s[i-1] <> "." do i:= i - 1 end_while;
        substring( s,0..i )
    end_proc:

    createrange := proc(ind, nr, xr)
       name plotfunc2d::createrange;
       local ni;
    begin
       if (ni := nops(ind)) > 1 then 
             error("Functions contain more than one variable.")
       end_if;
       if nr = 0 then
          if ni = 0 then
	     gen_x := TRUE;
	     return( (genident("x") = DefaultX) )
	  else
	     return( (op(ind, 1) = DefaultX) )
	  end_if;
       elif nr = 1 then
          if ni = 0 then
	     return( xr )
	  else
	     if op(ind, 1) = op(xr, 1) then
	        return( xr )
	     else
	        error("More than one variable defined.")
	     end_if
	  end_if
       else
          return( xr )
       end_if;
    end_proc:

    check_object := proc(arg:DOM_LIST)
       name plotfunc2d::check_object;
       local a, b, k;
    begin
       if args(0) = 2 then
          xr := args(2)
       else
          xr := null()
       end_if;
       k := arg[1];
       if nops(arg) < 3 or type(k) <> "_equal" or 
          op(k, 1) <> Mode or op(k, 2) <> Curve then
          return(xr);
       end_if;
       // [ Mode = Curve, [x(t), y(t)], t=[a,b],...]
       k := arg[3];
       if not (testtype(k, "_equal") and
          testtype(float(op(k,2)), Type::ListOf(Type::Real,2,2))) then
          return(xr)
       end_if;
       if domtype(xr) = DOM_NULL then
          return( op(k,1) = float(op(k,2)) );
       end_if;
       a := min(float(op(k, [2,1])), float(op(xr, [2,1]))); 
       b := max(float(op(k, [2,2])), float(op(xr, [2,2])));
       return( op(k, 1)=[a,b] );
   end_proc:

   // Initializiation: Default values (change it if you want)
   axes           := Origin; // Axes=Origin (Posible values: None, Box, Corner)            
   real_mode      := TRUE;   // Set option RealsOnly = real_mode (TRUE or FALSE)
   discont_mode   := FALSE;  // Set option Discont = discont_mode (TRUE or FALSE)
   grid           := 100;    // Grid=n, n integer > 0

   // Do not change the following initialization   
   xr             := null();
   yr             := null();
   x_var          := null();
   y_var          := null();
   y_min          := null();
   y_max          := null();
   vbox           := null();
   gridlines      := null();
   ticks          := null();
   new_args       := [];
   plot2d_objects := null();
   scene_options  := {};
   ind            := {};
   nr             := 0;
   ti             := "";
   gen_x          := FALSE;
   DefaultX       := plotlib::Options[DefaultXRange]; 
   DefaultX       := [op(DefaultX, 1), op(DefaultX, 2)];
   DefaultY       := plotlib::Options[DefaultYRange]; 
   DefaultY       := [op(DefaultY, 1), op(DefaultY, 2)];
   

   for arg in [args()] do

       if type(arg) = "_equal" then

          case type(op(arg, 2)) 
           of "_range" do
	      if nr = 0 then
	         xr := getrange(arg);  
	         x_var := op(xr, 1);
                 x_min := op(xr,[2,1]);
                 x_max := op(xr,[2,2]);
              elif nr = 1 then
	         yr := getrange(arg);
		 y_var := op(yr, 1);
                 y_min := op(yr,[2,1]);
                 y_max := op(yr,[2,2]);
	      else
	         error("vertical and horizontal plotting range already specified.")
	      end_if;
	      nr := nr + 1;
	      break;
           otherwise
              scene_options := scene_options union {arg}
        end_case;
     elif domtype(arg) = DOM_LIST then
        plot2d_objects := plot2d_objects, arg;
	t := check_object(arg, xr);
	if t <> xr then
	   xr := t;
	   if nr = 0 then nr := nr + 1 end_if;
	end_if
     elif domtype(arg) = DOM_SET then
        error("Illegal argument. Sets are not allowed.")
     else
        new_args := append(new_args, arg);
	ind := ind union (indets(arg) minus Type::ConstantIdents);
     end_if;

   end_for;

   xr    := createrange(ind, nr, xr);
   x_var := op(xr, 1);
   x_min := op(xr,[2,1]);
   x_max := op(xr,[2,2]);

   case type(plotlib::Options[Scene2d]) 
     of DOM_SET   do 
        s := plotlib::Options[Scene2d]; break;
     of DOM_LIST  do 
        s := {op(plotlib::Options[Scene2d])}; break;
     otherwise 
        error("plotlib::Options[Scene2d]: set or list expected."); 
   end_case;

   t := select(s, has, {RealsOnly, RealValuesOnly});
   if nops(t) <> 1 then
      error("plotlib::Options[Scene2d]: Only RealsOnly or RealValuesOnly allowed."); 
   end_if;
   s := s minus t;
   for arg in s do
       if type(arg) <> "_equal" then
          error("Illegal option in plotlib::Options[Scene2d].");
       end_if;
       if not has(scene_options, op(arg, 1)) then
          scene_options := scene_options union {arg}
       end_if
   end_for;
   if not has(scene_options, {RealsOnly, RealValuesOnly}) then
      scene_options := scene_options union t
   end_if;

   s := {};
   for arg in scene_options do
       case op(arg, 1)     
         of Grid do
	    grid := op(arg, 2);
	    if domtype(grid) <> DOM_INT or grid < 2 then
	       error("Grid=n, n integer > 1 expected")
	    end_if;
	    break; 
         of ViewingBox do
	    vbox := arg;
            break;
	 of GridLines do
	    gridlines := arg;
	    break;
	 of Ticks do
	    ticks := arg;
	    break;
         otherwise
            s := s union {arg};
       end_case;
   end_for;
   scene_options := s;

   if not has(scene_options, Axes) then
      scene_options := scene_options union {Axes=axes};
   end_if;
   if not has(scene_options, Labels) then
      if domtype(x_var) <> DOM_NULL and not gen_x then 
         s := expr2text(x_var) 
      else
         s := "x"
      end_if;
      if domtype(y_var) <> DOM_NULL then 
         t := expr2text(y_var) 
      else
         t := "y"
      end_if;
      scene_options := scene_options union {Labels=[s, t]};
   end_if;
   if not has(scene_options, {RealsOnly, RealValuesOnly}) then
      scene_options := scene_options union {RealsOnly=real_mode};
   end_if;
   if not has(scene_options, Discont) then
      scene_options := scene_options union {Discont=discont_mode};
   end_if;

   if type(y_min) <> DOM_NULL and type(y_max) <> DOM_NULL then
      scene_options := scene_options union 
                       {ViewingBox=[x_min..x_max,y_min..y_max]}
   end_if;

   s := null(); nr := nops(new_args);
   for arg in new_args do
        s := s, [Mode=Curve, [ x_var, arg ], x_var=[x_min, x_max], Grid=[grid]];
        ti := ti . createtitle(arg); 
        if (nr := nr - 1) > 0 then ti := ti . ", " end_if;
   end_for;
   if not has(scene_options, Title) then
      scene_options := scene_options union {Title=ti};
   end_if;

   s := op(scene_options), s, plot2d_objects;

   if type(vbox) <> DOM_NULL then
      if type(y_var) = DOM_NULL then
         t := plotlib::check_vb(op(vbox, 2));
         s := ViewingBox=[x_min..x_max, t[3]..t[4]], s
      end_if;
      s := plotlib::check_args(s);
      if (t := select([s], has, Ticks)) <> [] then
          s := subs(s, op(t) = null())
      end_if;
      s := vbox , s
   end_if;

   s := gridlines, ticks, s;

   plot2d(s)
   
end_proc:
