/*
series -- series expansions

series(f,x)		f is an expression
series(f,x,n)		x is a variable
series(f,x=v,n)         n is an integer and v an expression independent of x
series(f,x=v,Right)
series(f,x=v,Left)
series(f,x=v,Real)
series(f,x=v,n,Right)
series(f,x=v,NoWarning) - supresses warning "Could not find ..."
series(f,x=v,n,NoWarning)
series(f,x=v,Real,NoWarning)
series(f,x=v,n,Real,NoWarning)

Note: option NoWarning is only relevant within the interface functions
series and Series::main; it is NOT passed to the series attributes.
Thus, if a series attribute itself calls series, then NoWarning has
no effect for this call.
*/

series:= proc(f, x, ordr = FAIL, direc = null(), warn = null())
  local applyFloat, s, ord, dir, war;
begin
  if args(0) = 0 then error("missing arguments") end_if;
  if f::dom::series <> FAIL then return(f::dom::series(args())) end_if;

  // Copy args before modifying them in order to make the procname(args())
  // at the end make work correctly
  ord := ordr;
  dir := direc;
  war := warn;

  // set ord to the current value of ORDER, if no order was given:
  if ord = FAIL then ord := ORDER end_if;

  // handle the case where no order but at least three args are given
  if args(0) >= 3 and domtype(ord) <> DOM_INT and ord <> infinity then
    war := dir;
    dir := ord;
    ord := ORDER;
  end_if;

  case type(f)
  of "int" do
    return(int(series(op(f,1),x,ord,dir,warn),op(f,2..nops(f))))
  of DOM_POLY do
    if args(0) > 5 then 
        error("wrong number of arguments")
    else
        return(Series::Puiseux::frompoly(args()))
    end_if
  end_case;

  case args(0)
    of 1 do
      error("missing series variable")
    of 2 do
    of 3 do
    of 4 do
    of 5 do
      break
    otherwise
      error("wrong no of arguments");
  end_case;

  if domtype(dir) = DOM_NULL then dir := FAIL end_if;

  if ord <> infinity and (domtype(ord) <> DOM_INT or ord < 0) then
    error("expecting nonnegative integer or infinity as order")
  end_if;
  
  if not contains({Left,Right,Real,NoWarning,FAIL},dir) then
    error("invalid direction (expecting Left, Right, or Real)")
  end_if;

  // Handle the case where option NoWarning is given with less than 5
  // arguments
  if dir = NoWarning then
    dir := FAIL;
    war := NoWarning;
  end_if;

  s:=misc::maprec(f,{DOM_FLOAT}=proc()
                                begin
                                  numeric::rationalize(args(1), Minimize)
                                end_proc );
  applyFloat := id;
  // the eval is necessary since for some weird reason f may be
  // not completely evaluated sometimes!
  if eval(f) <> eval(s) then
    applyFloat := float;
  end_if;
  s:= Series::main(s,x,ord,dir,war);

  if s = FAIL then procname(args()) else applyFloat(s) end_if
end_proc:

series:= slot(funcenv(series), "type", "series"):

Series::main := proc(f, p, ord, dir = FAIL, war = FAIL)
   local x,X,x0,xofX,n,s,tmp;
begin
   X:= genident("X"): // series variable (X->0)

   if domtype(p)=DOM_IDENT then 
       x0:=0; x:=p
   elif type(p)="_equal" and domtype(op(p,1))=DOM_IDENT then 
      x:=op(p,1); x0:=op(p,2);
      if not testtype( x0,Type::Arithmetical ) then
          error("invalid series point")
      elif has(x0,x) then error("series point depends on series variable")
      elif x0=infinity then // special case 
         if contains({Right,Real}, dir) then
           error("inconsistent direction")
         else 
           xofX:=1/X;
           dir:=Right
         end_if
      elif x0=-infinity then
         if contains({Left,Real}, dir) then
           error("inconsistent direction")
         else
           xofX:=-1/X;
           dir:=Right
         end_if
      elif has(x0, infinity) then
         error("invalid expansion point")
      else 
         xofX:=X+x0;
// This does destroys any properties x might have before !!!
// It is irrelevant anyway, since s is overwritten below immediately!
	 if dir=Right then assume(x > x0); s:=simplify(f);unassume( x );
	 elif dir=Left then assume(x < x0);s:=simplify(f);unassume( x );
	 elif dir=Real then assume(x, Type::Real);s:=simplify(f);unassume( x );
	 end_if;
      end_if;
   else 
      error("invalid series variable")
   end_if;

   // keep properties of the series variable:
   s:= f;
   if x0=0 then X:=x else s:=subs(s,x=xofX) end_if;
   if ord=infinity then
      if testtype(s, Type::PolyExpr(X)) then
        return(f)
      end_if;
      n := genident("n");
      s:=Series::expr2coeff(s,X,n);
      if s<>FAIL then s:=hold(sum)(s*x^n,n=0..infinity) end_if;
      return(s)
   end_if;

   tmp:= getprop(x);
   if tmp <> x then assume( X,tmp ) end_if;

   s:= Series::series(s,X,ord,dir);
   if s = FAIL then
      if dir = FAIL then
         // undirected expansion
         if war <> NoWarning then
           warning("Could not find undirected series expansion; try option `Left', `Right', or `Real'");
         end_if;
         return(FAIL);
      end_if;
      if dir = Real then
         // twosided expansion
         if war <> NoWarning then
           warning("Could not find real series expansion; try option `Left' or `Right'");
           return(FAIL);
         end_if;
      end_if;
      if traperror((s:=Series::gseries::new(f,p,ord,dir)))=0 then s else FAIL end_if
   elif x0=0 then s 
   else s::dom::set_var(s,X,x,x0)
   end_if
end_proc:

