/*===========================================================================

sum -- (in)definite summation of rational functions

Calling sequence:
sum(f, x)

Parameters:
f -- an expression (rational function)
x -- an

Summary:
sum(f,x) computes the indefinite sum of f(x) with respect to x.

Copyright:
Paul Zimmermann and Inria, November 1994

--------------------------------------------------------------------------

Examples of the implementation of the "sum" function:

1) Polynomial case:

>> sum(1,k);

                                     k

>> sum(k^2,k);

                                     2    3
                                k   k    k
                                - - -- + --
                                6   2    3
 
>> sum(a*x^2+b*x+c,x);
 
                    3
                 a x      / a   b     \    2 /   a   b  \
                 ---- + x | - - - + c | + x  | - - + -  |
                  3       \ 6   2     /      \   2   2  /
 
 
2) Fractional case: 

a) Decomposition into the polynomial part and the rational part:
 
>> sum(k/(k+1),k);
 
                              k - psi(k + 1)
 
>> sum((x^2+1)/(x-1),x);
 
                              2
                         x   x     
                         - + -- + 2 psi(x - 1) 
                         2   2 
 
 
b) Reduced rational expressions:
   (degree of numerator is less than degree of denominator)

>> sum(1/k/(k-2),k);

                                1           1    
                          - --------- - ---------
                            2 (k - 1)   2 (k - 2)


>> sum(1/k/(2*k-1),k);

                          - psi(k) + psi(k - 1/2)


==========================================================================*/
sum:= proc(s, x)
begin
  if args(0) <> 2
     then error("two arguments expected")
     else return( Sum::sum(s,x) )
  end_if
end_proc:

sum:= funcenv(sum):
sum:= slot(sum, "type", "sum"):
sum:= slot(sum, "print", "sum"):
sum:= slot(sum,"int", proc(s,x) begin sum(int(op(s,1),x),op(s,2)) end_proc):
sum::Content      := proc(Out, data)
                     begin
                       if type(op(data, 2)) = "_equal" then
                         if type(op(data, [2,2])) = "_range" then
                           Out::Capply(Out::Csum,
                                     Out::Cbvar(Out(op(data, [2,1]))),
                                     Out::Clowlimit(Out(op(data, [2,2,1]))),
                                     Out::Cuplimit( Out(op(data, [2,2,2]))),
                                     Out(op(data,1)));
                         else
                           Out::Capply(Out::Csum,
                                     Out::Cbvar(Out(op(data, [2,1]))),
                                     Out::Clowlimit(Out(op(data, [2,2]))),
                                     Out(op(data,1)));
                         end_if:
                       else
                         Out::Capply(Out::Csum,
                                     Out::Cbvar(Out(op(data, 2))),
                                     Out(op(data,1)));
                       end_if;
                     end_proc:

//sum:= slot(sum,"float",
// loadproc(slot(sum,"float"),pathname("NUMERIC"),"floatsum")):
sum:= slot(sum,"float",loadproc(numeric::sum,pathname("NUMERIC"),"floatsum")):

sum:= slot(sum,"diff", proc(e) // e = sum(f,x) 
local i,l,_x;
begin
  l:=[args(i)$i=2..args(0)]; // variables to differentiate 
  if map({op(l)},type)={DOM_IDENT} and not has((_x:=op(e,2)),l) then 
	// summation independent from differentiation 
     sum(diff(op(e,1),op(l)),_x)
  else
     hold(diff)(e,op(l))
  end_if
end_proc):

sum:= slot (sum,"expand", proc()
 local s;
 begin
   s:=expand(args(1));
   if type(s)="_plus" then
     _plus(map(s, expand@sum,args(2)))
   elif type(s)="_mult" then
     split(s,has,(if type(args(2))=DOM_IDENT 
                   then args(2)
                   else op(args(2),1)
                  end_if));
     %[2]*hold(sum)(%[1],args(2))
   else
     hold(sum)(args())
   end_if
 end_proc):

//--------------------------------------------------------------------------
Sum:= newDomain("Sum"):
Sum::info:= "Sum: Utilities for definite and indefinite summation":
Sum::interface:= {}:

Sum::sum:= proc(s,x) 
  local f,g,gg,_k_,Rec,RRec,u,_d_,l,_p,n,q,z,toE,_E, clean;
begin
  if has(s,{hold(cos),hold(sin),hold(cosh),hold(sinh)}) then
    f:= rewrite(s,exp)
  else
    f:= s
  end_if;

  if (toE:=has(f,exp)) then
    _E:= genident("_E");
    f:=eval(subs(f,hold(exp) = (x->_E^x)))
  end_if;

  // include, because in some branches _E survived. Walter, 24.10.99
  //  sum( exp(n*x)/n, n=0..infinity);
  clean:= proc(result) begin 
        if toE and has(result, _E) 
        then subs(misc::maprec(combine(result),
                  {"_power"}= proc(x) begin
                    if op(x,1)=_E then exp(op(x,2)) else x end_if
                              end_proc
                               ), _E=exp(1));
        else result;
        end_if;
  end_proc;

  if domtype(x)=DOM_IDENT then
    f:= Sum::indefinite(f,x)
  elif type(x)="_equal" and domtype((_k_:=op(x,1)))=DOM_IDENT then
    if type(op(x,2))="_range" then
      // definite summation 
      if domtype((g:=op(x,[2,2])-op(x,[2,1])))=DOM_INT then // finite range 
        f:= eval(subs(args(1),_k_=_k_+op(x,[2,1])));
        f:= (if g>=0 then
               _plus(eval(subs(f,_k_=u)) $ u=0..g)
             else
               -_plus(eval(subs(f,_k_=u)) $ u=g+1..-1)
             end_if);
        return(eval(%))
      elif has(op(x,2),_k_) then
        error("summation variable appears in bounds")
      elif hastype(op(x,2),DOM_FLOAT) then
        error("floating-point numbers not allowed in bounds")
      end_if;
      g:=Sum::indefinite(f,_k_);
      if type(g)="sum" and type(f)="_plus" then // try to map 
        l:=[];
        for q in f do
          u:=Sum::indefinite(q,_k_);
          if type(u)="sum" then break else l:=append(l,u) end_if
        end_for;
        if nops(l)=nops(f) then g:=_plus(op(l)) end_if;
      end_if;
      if type(g)="sum" then
        userinfo(1,"indefinite summation failed");
        if Sum::ishypergeom(f,_k_) then
          userinfo(1,"summand is hypergeometric");
          // try Zeilberger's algorithm 
          // choose in priority indets that are in the bounds 
          l:=indets(op(x,2));
          l:=[op(l),op(indets(f) minus (l union {_k_}))];
          for n in l do
            if Sum::ishypergeom(f,n) then
              userinfo(1,"trying Zeilberger's algorithm with respect to",n);
              u:=genident("u");
              Rec:=Sum::zeilberger(f,_k_,n,u,op(op(x,2)));
              if Rec=FAIL then
                userinfo(1,"Zeilberger's algorithm fails"); next
              else
                userinfo(1,"Zeilberger's algorithm succeeds")
              end_if;
              gg:=Rec[2]; RRec:=Rec[1];
              _d_:=op(RRec,[1,1])-n;
              _p:= proc(z)
                     name Sum::sum;
                   begin
                     eval(subs( hold(sum)(f,_k_=op(x,2)), n=z ))
                   end_proc;
              if traperror((g:={(u(z)=_p(z))$ z=gg-_d_..gg}))<>0 then
                g:={(u(z)=_p(z))$ z=gg-_d_+1..gg+1}
              end_if;
              Rec:=solve(rec(RRec,u(n),g));
              if type(Rec) = "solve" then
                // We specified one more initial condition than necessary in
                // the previous call.
                // ( This is necessary since otherwise, due to the changed
                //   output format of rec::solve, the
                //   solution returned by rec::solve may still contain free
                //   parameters ... CRUDE HACK!
                //   Example: sum(k*binomial(n, k), k = 0..n) )
                // Since it was not successful, it
                // sometimes helps to specify one less initial condition.
                // Example: sum(binomial(d,j)*binomial(d-r-j-1,r-j),j=0..r)
                // ( Reason: rec::solve uses linsolve which does not recognize
                //   the functional identity of the gamma function)
                if traperror((g:={(u(z)=_p(z))$ z=gg-_d_..gg-1}))<>0 then
                  g:={(u(z)=_p(z))$ z=gg-_d_+1..gg}
                end_if;
                Rec:=solve(rec(RRec,u(n),g));
              end_if:
              userinfo(5, "Solving recurrence equation gives ".
                       expr2text(Rec));
              if not has(Rec,undefined) then
                        if type(Rec)=DOM_SET and nops(Rec)>=1 then
			                  Rec:=op(Rec,1);
                        end_if; 
                delete _p;
                return(clean(Rec))
              end_if;
              delete _p;
            end_if
          end_for;
          // f is hypergeometric but Zeilberger's algorithm fails 
          userinfo(1,"trying definite summation of hypergeometrics");
          g:=Sum::hypergeomtype(f,_k_);
          userinfo(2,"hypergeometric type is",g);
          if g<>FAIL then
            g:=Sum::hypergeomeval(g,op(op(x,2)),f,_k_);
            if g<>FAIL then return(clean(g)) end_if
          end_if;
          if g=FAIL then g:=Sum::recognize_gf(f,_k_,op(op(x,2)));
            if g<>FAIL then return(clean(g)) end_if
          end_if;
          userinfo(1,"definite summation of hypergeometrics failed");
        end_if;

            // not hypergeometric in k or in another variable 
        s:= f; // return changed argument
            f:= FAIL // hold(sum)(f,x)
      else // indefinite summation succeeded 
        userinfo(2,"indefinite summation gave",g);
        // eval is needed to evaluate psi(1,1) into PI^2/6 for example 
        if traperror(
                     (g:=Sum::eval(g,_k_,op(x,[2,2])+1)-Sum::eval(g,_k_,op(x,[2,1])))
                     )=0 then
	        /* level has no effect on DOM_VAR
	        normal(level(g,2)) */
          g;
          f:= normal(eval(%))
        else
          s:= f; // return changed argument
          f:= FAIL // hold(sum)(f,x)
        end_if
      end_if
    elif type((_p:=op(x,2)))="RootOf" then // sum(..., k=RootOf(q,z)) 
      userinfo(1,"sum over RootOf");
      z:=op(_p,2);
      q:=eval(subs(op(_p,1),z=_k_));
      // sum(..., x=RootOf(1, y)) simplifies to zero 
      if contains({DOM_INT,DOM_RAT},type(q)) then
        return(0)
      end_if;
      if testtype(f,Type::RatExpr(_k_)) then // first simplify to a sum over a polynomial 
        userinfo(2,"summand is rational");
        f:= stdlib::RootOf_evala(f,q,_k_)
      end_if;
      if testtype(f,Type::PolyExpr(_k_)) then // evaluate 
        f:= stdlib::RootOf_polysum(f,q,_k_)
      else
        f:= FAIL // hold(sum)(args())
      end_if
    else
      error("invalid second argument")
    end_if
  else
    error("invalid arguments")
  end_if;

  if f = FAIL
     then return( hold(sum)(eval(clean(s)),x) )
     else return(clean(f))
  end_if;
end_proc:

/* the following extension was suggested by Daniel Duparc
  <dduparc@club-internet.fr> */
Sum::recognize_gf:=proc(f,n,a,b) local s,P,x,Q,lc,d;
begin
   P:=genident("P"); x:=genident("x");
   // recognizes sum(P(n)*x^n/n!, n=0..infinity) 
   if a=0 and b=infinity then
      s:=intlib::match(n!*f, _pattern(P,Type::PolyExpr(n))*_pattern(x)^n);
      if s<>FAIL then
         P:=eval(subs(P,s)); x:=eval(subs(x,s)); Q:=0;
         // x^i*exp(x) -> n*(n-1)*...*(n-i+1) = n!/(n-i)! 
         while P<>0 do
            lc:=lcoeff(P,[n]); d:=degree(P,[n]);
            Q:=Q+lc*x^d; P:=expand(P-lc*n!/(n-d)!);
         end_while;
         return(Q*exp(x));
      end_if
   end_if;
   FAIL
end_proc:

Sum::eval := proc(f,x,v) // evaluate value of f at x=v 
begin
   if v=infinity then
      limit(f,x=infinity)
   else /* do no eval here ==> problems if f contains the name "x",
          for example sum(x^k*y^(n-k),k=0..n) or sum(x^k,k=1..2) */
      eval(subs(f,x=v))
   end_if
end_proc:

// returns h such that h(x+1)-h(x)=f(x) 
Sum::indefinite:=proc(f,x)
begin
   userinfo(1,"entering indefinite summation");
   if testtype(f,Type::PolyExpr(x,Type::AnyType)) then Sum::poly(f,x)
   elif testtype(f,Type::RatExpr(x,Type::AnyType)) then Sum::rat(f,x)
   else Sum::gosper(f,x)
   end_if
end_proc:

// p is a polynom in x, returns h such that p = eval(subs(h,x=x+1))-h 
Sum::poly := proc(p,x)
local d,i,b,a;
begin
  if not has(p,x) then p*x
  else // method of indeterminate coefficients 
   p:=poly(p,[x]); d:=degree(p);
   // first tabulate a[i]=(x+1)^i 
   a[1]:=poly(x+1,[x]); for i from 2 to d+1 do a[i]:=a[1]*a[i-1] end_for;
   for i from d+1 downto 2 do
      b[i]:=coeff(p,i-1)/i;
      p:=p+multcoeffs(lmonomial(a[i],Rem)[2],-b[i]);
   end_for;
   b[1]:=coeff(p,0);
   _plus(b[i]*x^i $ i=1..d+1)
  end_if
end_proc:

Sum::rat := proc(f,x)
local num,den,d,dnum,dden;
begin
   num:=numer(f); den:=denom(f);
   dnum:=degree(num,x); dden:=degree(den,x);
   if dnum>=dden then
      d:=divide(poly(num,[x]),poly(den,[x]));
      Sum::poly(expr(d[1]),x)+Sum::rat(expr(d[2])/den,x);
   else // now deg(num)<deg(den) 
      d:=Sum::rat_abramov(f,x);
      if d=FAIL then hold(sum)(f,x) else d end_if
   end_if
end_proc:

/* f is minimal : sum(f,x) can not be written p+sum(r,x)
  where p is a polynomial and denom(r) has lower degree than denom(f)
*/
Sum::to_psi := proc(f,x)
local ff,t,s,a,d,lin;
begin
   ff:=partfrac(f,x,hold(full));
   userinfo(3,"full partial fraction decomposition=",ff);
   if type(ff)="_plus" then ff:=[op(ff)] else ff:=[ff] end_if;
   s:=0;
   for t in ff do
      d:=stdlib::denom(t); // because denom expands the denominator 
      d:=d/maprat(d,proc() begin content(args(1),[x]) end_proc ); // primpart(d,[x]) expands too !!! 
      if type(d)="_power" and testtype(op(d,2),Type::NonNegInt) then lin:=op(d,1)
      else lin:=d
      end_if;
      if testtype(lin,Type::PolyExpr(x,Type::AnyType)) then
         case degree(lin,[x])
         of 1 do a:=coeff(lin,[x],0)/coeff(lin,[x],1);
                 s:=s+subs(Sum::to_psi2(normal(eval(subs(t,x=x-a))),x),x=x+a);
                 break;
         otherwise return(FAIL)
         end_case
      else // non polynomial 
         return(FAIL)
      end_if
   end_for;
   s
end_proc:

Sum::is_linear:=proc(p,x) // checks for a*x+b 
begin
   if testtype(p,Type::PolyExpr(x,Type::AnyType)) then
      if degree(p,x)=1 then return(TRUE) end_if
   end_if;
   FALSE
end_proc:

Sum::to_psi2 := proc(f,x) // the denominator looks like a*x^j 
// f must be in normal form 
local n,j,res,d,c,dd,t;
begin
   res:=0;
   d:=denom(f);
   c:=maprat(d,proc() begin content(args(1),[x]) end_proc );
   d:=normal(d/c); // should be given by primpart 
   if d=-x then d:=x; c:=-c end_if;
   if not ((d=x) or (type(d)="_power" and op(d,1)=x and type(op(d,2))=DOM_INT))
   then return(FAIL) // should not happen 
   end_if;
   dd:=degree(d,x)-1;
   n:=numer(f);
   if type(n)="_plus" then n:=[op(n)] else n:=[n] end_if;
   for t in n do
      j:=dd-degree(t,x);
      res:=res+maprat(t,proc() begin lcoeff(args(1),[x]) end_proc )/c*(-1)^j/fact(j)*psi(x,j);
   end_for;
   res
end_proc:

Sum::rat_abramov := proc(f,x)
begin
   userinfo(1,"entering Abramov's algorithm");
   Sum::abr(f,Sum::dispersion(denom(normal(f)),x),x)
end_proc:

Sum::abr := proc(f,dis,x)
local p,q,cp,vwp,vwm,ba,u,newf;
begin
   userinfo(2,"dispersion=",dis);
   if f=0 then return(0) end_if;
   if dis=0 then return(Sum::to_psi(f,x)) end_if;
   p:=numer(f); q:=denom(f);
   cp:=gcd(q,subs(q,x=x+dis));
   if cp=1 then return(Sum::abr(f,dis-1,x)) end_if;
   vwp:=Sum::abr_part(q,cp);
   vwm:=Sum::abr_part(q,subs(cp,x=x-dis));
   if degree(vwm[1],x)>degree(vwp[1],x) then
      ba:=solvelib::pdioe(vwp,p,x);
      u:=subs(ba[2]/vwp[1],x=x-1);
      newf:=normal(ba[1]/vwp[2]+u);
   else
      ba:=solvelib::pdioe(vwm,p,x);
      u:=-ba[2]/vwm[1];
      newf:=normal(ba[1]/vwm[2]+subs(ba[2]/vwm[1],x=x+1))
   end_if;
   u+Sum::abr(newf,dis-1,x)
end_proc:

Sum::abr_part := proc(p,h)
local g,v,w;
begin
   g:=gcd(p,h); v:=1; w:=p;
   while g<>1 do w:=divide(w,g,Quo); v:=v*g; g:=gcd(w,v) end_while;
   v,w
end_proc:

Sum::dispersion := proc(q,x) // q is a polynomial 
local X,r,i,s;
begin
   X:= genident("X"):
   q:=normal(q/content(q,[x]));
   r:=polylib::resultant(q,subs(q,x=x+X),x);
   r:=factor(r);
   s:={};
   for i from 2 to nops(r) step 2 do
      // this is because iroots does not like other variables 
      if indets(op(r,i))={X} then s:=s union solvelib::iroots(op(r,i)) end_if;
   end_for;
   // no multiplicity for iroots 
   max(op(s))
end_proc:

Sum::ishypergeom := proc(f,n)
begin
   testtype(Sum::ratio(f,n),Type::RatExpr(n))
end_proc:

/* f being hypergeometric in k, puts f(k+1)/f(k) in the form
  (k+a1)...(k+ap)/(k+b1)/.../(k+bq) x/(k+1)
  and returns [a1,...,ap], [b1,...,bq], x,
  cf Wilf page 5 */
Sum::hypergeomtype := proc(f,k)
local g,P,Q,a,b,x,t,d;
begin
   g:=normal(subs(Sum::ratio(f,k),k=k+1));
   P:=Factored::convert_to(factor(numer(g)),DOM_EXPR);
   if type(P)<>"_mult" then P:=[P] end_if;
   a:=[]; b:=[]; x:=1;
   for t in P do
      if not has(t,k) then x:=x*t
      else
         if type(t)="_power" then d:=op(t,2); t:=op(t,1) else d:=1 end_if;
         if degree(t,k)<>1 then return(FAIL) end_if;
         x:=x*coeff(t,k,1);
         a:=append(a,coeff(t,k,0)/coeff(t,k,1) $ d);
      end_if
   end_for;
   Q:=Factored::convert_to(factor(denom(g)),DOM_EXPR);
   if type(Q)<>"_mult" then Q:=[Q] end_if;
   for t in Q do
      if not has(t,k) then x:=x/t
      else
         if type(t)="_power" then d:=op(t,2); t:=op(t,1) else d:=1 end_if;
         if degree(t,k)<>1 then return(FAIL) end_if;
         x:=x/coeff(t,k,1);
         b:=append(b,coeff(t,k,0)/coeff(t,k,1) $ d);
      end_if
   end_for;
   t:=contains(b,1);
   if t=0 then // no factor k+1 in the denominator, add one 
      a:=append(a,1)
   else // delete it 
      delete b[t]
   end_if;
   a,b,x
end_proc:

// sum(f,k=a..b) when sum(f,k={0,1}..infinity) = pFq(P,Q,z) * f(0)
Sum::hypergeomeval := proc(P,Q,z,a,b,f,k)
  local init,kk;
begin
   userinfo(1,"hypergeometric of type ".nops(P)."F".nops(Q));
   if nops(P)=1 and nops(Q)=0 then // 1F0 
      if testtype(a,Type::NonNegInt) and b=infinity then
         userinfo(2,"1F0 with right bound at infinity");
         if P=[0] then
            if a = 0 then return(FAIL) end_if;
            init:=_plus(eval(subs(f,k=kk)) $ kk=1..a-1);
            return(-ln(1-z)*eval(subs(f,k=1))/z-init)
         else
           init:=_plus(eval(subs(f,k=kk)) $ kk=0..a-1);
           return(((1-z)^(-P[1])-init)*eval(subs(f,k=0)))
         end_if
      end_if
   elif nops(P)=0 and nops(Q)=0 then // 0F0 
      if testtype(a,Type::NonNegInt) and b=infinity then
           init:=eval(_plus(eval(subs(f,k=kk)) $ kk=0..a-1));
           return(exp(z)*eval(subs(f,k=0))-init)
      end_if
   elif nops(P)=2 and nops(Q)=1 then // 2F1 
      return((if P=[1/2,1] and Q=[3/2] and a=0 and b=infinity then
         arctan(I*sqrt(z))/I/sqrt(z) * eval(subs(f, k=0))
      elif P=[0,0] and Q=[1] and testtype(a,Type::PosInt) and b=infinity then
	 init:=_plus(eval(subs(f,k=kk)) $ kk=1..a-1);
         polylog(2,z)*eval(subs(f/z,k=1))-init
      else FAIL
      end_if))
   elif nops(P)=3 and nops(Q)=2 then // 3F2 
      if P=[1,1,1] and Q=[2,3/2] and z=1/4 and b=infinity and 
	testtype(a,Type::NonNegInt) then // sum(n!^2/(2*n+2)!,n=a..infinity) 
           /* level has not effect on DOM_VAR
           init:=level(_plus(f $ k=0..a-1));
           return((PI^2/9-2*init)*eval(subs(f,k=0))) 
           */
           _plus(f $ k=0..a-1);
           return((PI^2/9-2*eval(%))*eval(subs(f,k=0)))
      end_if
   end_if;
   FAIL
end_proc:

// computes f/subs(f,n=n-1) and returns a normalized result 
Sum::ratio := proc(f,n) 
local p,a,b,i;
option remember;
begin
   case type(f)
   of "_mult" do return(normal(map(f,Sum::ratio,n)))
   of "_power" do
      if not has(op(f,2),n) then return(Sum::ratio(op(f,1),n)^op(f,2)) 
      elif not has(op(f,1),n) then
	return(op(f,1)^(op(f,2)-subs(op(f,2),n=n-1)))
      else break end_if
   of "fact" do
      if not has((p:=Sum::ratio(rewrite(f,gamma),n)),gamma) then p end_if;
      break
   of "gamma" do
      if not has(f,n) then return(1)
      elif Sum::is_linear((p:=op(f)),n) then
         if type((a:=coeff(p,n,1)))=DOM_INT then
            b:=coeff(p,n,0);
            if a>0 then return(_mult(a*n+b-i$i=1..a))
            elif a<0 then return(1/eval(_mult(a*n+b+i$i=0..-a-1))) // PR-6 
            // a cannot be 0 otherwise p would be independent of n 
            end_if
         end_if
      end_if;
   end_case;
   // other cases 
   if type(f)="function" then
      if op(f,0)=hold(rfact) then
         if not has(op(f,1),n) and Sum::is_linear(op(f,2),n) and
	    coeff(op(f,2),n,1)=1 then return(op(f,1)+op(f,2)-1)
         end_if
      end_if
   elif type(f)="binomial" then return(Sum::ratio(expand(f),n))
   end_if;
   normal(f/expand(subs(f,n=n-1)))
end_proc:

// from GKP pages 224-226 
Sum::gosper := proc(f,_n_) // returns g such that f = subs(g,n=n+1)-g 
local rat,qn,rn,res;
begin
   userinfo(1,"enter Gosper");
   if not has(f,_n_) then f*_n_
   else
      /* write f(k+1)/f(k) = p(k+1)/p(k) * q(k)/r(k+1) (5.117)
        where roots of q and r do not differ by an integer */
      rat:=Sum::ratio(f,_n_);
      // check for geometric summations 
      if not has(rat,_n_) then return(eval(subs(f,_n_=0))*rat^_n_/(rat-1)) end_if;
      if has(rat,{binomial,fact}) then rat:=normal(expand(rat)) end_if;
      userinfo(2,"f(".expr2text(_n_)." / f( ".expr2text(_n_-1)." = ".
expr2text(rat));
      qn:=numer(rat);
      rn:=denom(rat);
      if not (testtype(qn,Type::PolyExpr(_n_,Type::AnyType)) 
         and testtype(rn,Type::PolyExpr(_n_,Type::AnyType))) then
            return(hold(sum)(f,_n_))
      end_if;
      userinfo(1,"input is hypergeometric");
      res:=Sum::gosper2(1,qn,rn,_n_,{},f);
      if res=FAIL then 
         userinfo(1,"Gosper's algorithm fails");
         hold(sum)(f,_n_)
      else normal(res*f)
      end_if
   end_if
end_proc:

/* returns R=g/f such that f = subs(g,n=n+1)-g for f(n)/f(n-1)=p(n)/p(n-1)*qn/rn
  with pn,qn,rn polynomials,
  inds is a set of auxiliary unknowns (used only in Zeilberger).
  As f[n] = g[n+1]-g[n]. we must have f[n]=R[n+1]*f[n+1]-R[n]*f[n],
  i.e. 1 = R[n]*p(n)/p(n-1)*qn/rn-R[n-1]
*/
Sum::gosper2 := proc(pn,qn,rn,n,inds,f)
local X,rnj,res,i,p_zeros,z,j,gn,gnj,dp,R,dR,Q,dQ,d,k0,fn,eq,F,ii,m,unk,ff;
begin
      userinfo(3,"pn=",pn,"qn=",qn,"rn=",rn);

      X:= genident("X");
      rnj:=subs(rn/content(rn,[n]),n=n+X); // do not expand ! 
      res:=polylib::resultant(qn,rnj,n);
      res:=factor(res);
      p_zeros:=[];
      for i from 1 to nops(res) div 2 do
         ff:=op(res,2*i);
         if degree(ff,X)=1 then
            z:=-coeff(ff,X,0)/coeff(ff,X,1);
            if testtype(z,Type::PosInt) then p_zeros:=append(p_zeros,z) end_if
         end_if
      end_for;
      for j from nops(p_zeros) downto 1 do
         gn:=gcd(qn,expand(eval(subs(rn,n=n+p_zeros[j]))));
         qn:=divide(qn,gn,Exact);
         gnj:=gn;
         for i from 1 to p_zeros[j] do
            pn:=expand(pn*gnj);
            gnj:=expand(subs(gn,n=n-i));
         end_for;
         rn:=normal(rn/gnj); // divide(rn,gnj,Exact) ? 
      end_for;
      // make the polynoms monic (cf Prop. 3.2 of [LiPaSt93]) 
      qn:=qn/lcoeff(rn,[n]);
      rn:=rn/lcoeff(rn,[n]);
      qn:=expand(subs(qn,n=n+1));
      dp:=degree(pn,[n]);
      R:=expand(qn+rn); dR:=degree(R,[n]);
      Q:=expand(qn-rn); dQ:=degree(Q,[n]);
      if dR<=dQ then // case 1 
         userinfo(2,"case 1 of Gosper's algorithm");
         if dQ=0 then // polynomial case 
            d:=dp+1
         else
            d:=dp-dQ
         end_if
      else // case 2 : deg(qn)=deg(rn) and lcoeff(qn)=lcoeff(rn) 
         m:=dR; // to have the same notation as [LiPaSt93] 
         k0:=coeff(rn-qn,n,m-1); // (5.2) of [LiPaSt93] 
         if type(k0)<>DOM_INT or m=1 // cf [LiPaSt93] page 255 
         or testtype(f,Type::RatExpr(n,Type::AnyType)) then // case 2a 
            userinfo(2,"case 2a of Gosper's algorithm");
            d:=dp-m+1
         else // case 2b 
            userinfo(2,"case 2b of Gosper's algorithm");
            d:=max(k0,dp-m+1)
         end_if
      end_if;
      userinfo(2,"d=",d);
      if d>=0 then
	 F:= genident("F");
         fn:=_plus(F[ii]*n^ii$ii=0..d);
         // eq. (GE) of [LiPaSt93] p. 248 
         // where qn is q(n+1) 
         // put minus sign before fn because of PR302-303-310 
         eq:=expand(pn+qn*subs(-fn,n=n+1)+rn*fn);
         userinfo(2,"number of equations is",degree(eq,[n])+1);
         eq:=poly(eq, [n]);
         // necessary to avoid problems with e.g. eq=param^(-1)*n 
         eq:={coeff(eq,i)$ i=0..degree(eq)};
         unk:=inds union {F[ii]$ii=0..d};
         if nops(eq)<nops(unk) then eq:=eq union {F[0]=0} end_if;
         userinfo(3,"system to solve is",eq);
         eq:=linsolve(eq,unk);
         userinfo(1,"solution is",eq);
         if eq = FAIL then 
             // system has no solution, or solution is only valid if the parameters satisfy certain constraints 
             return( FAIL )
         else
             /* do the substitution into pn too because it can
               contain some indeterminates C[l] in Zeilberger's alg. */
             fn:=Sum::normal(eval(subs(fn/pn,eq))*rn);
             if has(fn,F) then return( FAIL )
             elif inds={} then return( fn )   // for Sum::gosper 
             else return( fn,eq )             // for Sum::zeilberger 
             end_if
         end_if
      else 
         return( FAIL )
      end_if
end_proc:

// required because normal expands 
Sum::normal := proc(f)
begin
   Sum::factor(numer(f))/Sum::factor(denom(f))
end_proc:

Sum::factor := proc(f)
local i,l;
begin
   l:=factor(f);
   _mult(l[1],l[2*i]^l[2*i+1]$i=1..nops(l) div 2)
end_proc:

// returns a..b such that f=0 for j<a or j>b 
Sum::support := proc(f,j) local a,b,x,k,c,d;
begin
   a:=-infinity; b:=infinity;
   for x in misc::subExpressions(f,"binomial") do
      for k in [op(x,2),op(x,1)-op(x,2)] do
         // k must be >= 0 
	 if testtype(k,Type::PolyExpr(j)) then
	    if degree(k,[j])=1 then
	       d:=coeff(k,[j],1); c:=-coeff(k,[j],0)/d;
	       if is(coeff(k,[j],1)>0) then
		  if a=-infinity then a:=c else a:=max(a,c) end_if
	       elif is(coeff(k,[j],1)<0) then
		  if b=infinity then b:=c else b:=min(b,c) end_if
	       end_if
	    end_if
	 end_if
      end_for
   end_for;
   a..b
end_proc:

// CHANGED (jngerhar, 28.9.00)
// Try to compute limit(f, x = a) without using limit if possible, since
// limit is often much much too slow when f contains parameters.
// Used only in Sum::zeilberger below
Sum::myeval := proc(f, x, a)
  local ff, fa, ord;
begin
  ff := eval(expr(factor(f)));
  // first try subs
  if traperror((fa := eval(subs(ff, x = a)))) = 0 then
    return(eval(expr(factor(fa))));
  else
    // next try series with orders 1,2,3,4
    for ord from 1 to 4 do
      if traperror((fa := series(ff, x = a, ord, Real, NoWarning))) = 0 then
        if domtype(fa) = Series::Puiseux then
          // expand is necessary to simplify products of gammas. This
          // was formerly done by series.
          return(eval(expr(factor(expand(coeff(fa, x, 0))))));
        end_if;
      end_if;
    end_for;
  end_if;
  // Finally, use limit as a last resort
  return(hold(limit)(ff, x = a))
end_proc:

// u is the name of the output recurrence 
Sum::zeilberger:=proc(f,_k,n,u,a,b) // sum(f,k=a..b) 
local ratn,d,l,m,C,p,ratk,res,rnk,cert,va,vb,A,/*fcert,*/corr,i,sup,aa,bb;
begin
   if has(a,n) then return(FAIL) end_if; // not yet treated 

   ratn:=Sum::ratio(f,n); // f(n)/f(n-1) 
   if not has(ratn,{n,_k}) then 
	return(FAIL) 
   elif not testtype(ratn,Type::RatExpr(n)) then 
	return(FAIL)
   elif not testtype(ratn,Type::RatExpr(_k)) then 
	return(FAIL)
   end_if;

   d:=1;
   C:= genident("C");
   repeat
      rnk:=1; m:=1;
      for l from 1 to d do
         m:=m*subs(ratn,n=n+l); // now m=f(n+l)/f(n) 
         rnk:=rnk+C[l]*m;
      end_for;
      // now apply Gosper to rnk*f with respect to k 
      /* i.e. search if F=f(n)+C[1]*f(n+1)+...+C[d]*f(n+d)
        satisfies F(k)=G(k+1)-G(k) */
      userinfo(3,"rnk=",rnk);
      l:=C[l]$ l=1..d;
      p:=numer(rnk);
      ratk:=Sum::ratio(f/denom(rnk),_k);
      res:=Sum::gosper2(p,numer(ratk),denom(ratk),_k,{l},f);
      if res<>FAIL then 
         cert:=eval(subs(res[1]*rnk,res[2])); // g = cert*f 
         userinfo(1,"certificate=",cert);
         /* one should have subs(G,k=k+1)-G-F=0 where
           F=f(n)+C[1]*f(n+1)+...=rnk*f and G=(cert/rnk)*F=cert*f,
           i.e. subs(cert*f,k=k+1)/f-cert-rnk=0 */
         /* thus sum(f,k=a..b)+...+sum(subs(f,n=n+d),k=a..b) 
		= subs(f*cert,k=b+1) - subs(f*cert,k=a) */
         /* only deal with the cases where a is independent from n
           and b is independent from n or b=n */
//         fcert:=Sum::normal(f*cert);
//         /* The following is a VERY CRUDE hack to make the limit
//            computations below work correctly. Remove this as
//            soon as the corresponding limit bug has been fixed!
//         */
//         fcert := numer(fcert)/expand(denom(fcert));
// CHANGED (jngerhar, 28.9.00): the above lines are no longer necessary,
// see the changes below. fcert is no longer used anywhere.
         sup:=Sum::support(f,_k);
         // if a<=op(sup,1) we can extend the lower bound to -infinity 
         aa:=simplify(min(a,op(sup,1)));
//         if traperror((va:=eval(subs(expand(subs(f*cert,_k=a+_k)),_k=0))))<>0 
//         then va:=hold(limit)(fcert,_k=a)
//         end_if;
// CHANGED (jngerhar, 28.9.00): replaced the above three lines by Sum::myeval
         va := Sum::myeval(f*cert, _k, a);
         // if op(sup,2)<=b we can extend the upper bound to infinity 
         bb:=simplify(max(b,subs(op(sup,2),n=n+d)));
//         if traperror((vb:=eval(subs(expand(subs(f*cert,_k=b+1+_k)),_k=0))))<>0
//         then vb:=hold(limit)(fcert,_k=b+1)
//         end_if;
// CHANGED (jngerhar, 28.9.00): replaced the above three lines by Sum::myeval
         vb := Sum::myeval(f*cert, _k, b + 1);
         A:=_plus(u(n),C[l]*u(n+l)$ l=1..d);
         A:=A+eval(va-vb);
         if A=undefined then return(FAIL)
         // correcting terms if b depends on n 
         elif bb<>b and not has(b-n,n) then
	       corr:=_plus((C[l]*eval(subs(f,n=n+l,_k=b+i))$ i=1..l) $ l=1..d);
	       A:=A-corr
         else return(FAIL)
         end_if;
         A:=expand(eval(subs(A,res[2])));
         /* to check: subs(rnk,res[2])-subs(cert*Sum::ratio(f,k),k=k+1)+cert
	   should be zero */
         userinfo(1,"recurrence=",A);
         res:=numer(A);
         res:=-coeff(res,u(n+d),0)/coeff(res,u(n+d),1);
         res:=u(n+d)=collect(res,[u(n+l)$ l=0..d-1],x->Factored::convert_to(factor(x),DOM_EXPR));
         if iszero(eval(subs(denom(cert),n=0))) then d:=d+1 end_if;
         return([res,d])
      else
         userinfo(1,"Gosper's algorithm fails for order ",d);
      end_if;
      d:=d+1; // try a higher order 
   until d>Sum::maxorder end_repeat;
   FAIL
end_proc:

Sum::maxorder:=2:

