// $Date: 2000/07/25 19:15:16 $ $Author: walter $ $Revision: 1.17 $ 

// Author: paul 

/*++
	fourier(f(x),x,y) returns the Exponential Fourier transform of f(x) 
	with respect to y,

	g_e(y) = int(f(x)*exp(I*y*x),x=-infinity..infinity).

	There are also the Fourier Cosine Transform

	g_c(y) = int(f(x)*cos(y*x),x=0..infinity)

	and the Fourier Sine Transform

	g_s(y) = int(f(x)*sin(y*x),x=0..infinity)

	If f is even, then g_e(y)=2*g_c(y); if f is odd, then g_e(y)=2*I*g_s(y),
	thus g_e(f(x)) = g_c(f(x)+f(-x))/2+I*g_s(f(x)-f(-x))/2

	f is any expression, x must be an identifier, y is any expression

	Reference: F. Oberhettinger, Tables of Fourier Transforms and Fourier 
	           Transforms of Distributions, Springer-Verlag, 1990
++*/

alias(Match=intlib::match1):
alias(Match2=intlib::matchassign):
alias(indep(a,x)=_pattern(a,Type::IndepOf(x))):
alias(_p=_pattern):

transform::fourier:= proc(f,x,y) 
   local v, yy, i, tmp;
   save _a,_b_,c,n,_k_,g,z,h,m;
begin
   delete c,m,z,n;

   if args(0) <> 3 then error("expecting three arguments") end_if;

   if f::dom::fourier<>FAIL then 
       // allow overloading
       return(f::dom::fourier(args())) 
    end_if;

   if testargs() then
      if type(f) = "_equal" then
          if not testtype(f,Type::Arithmetical=Type::Arithmetical) then
              error("invalid 1st argument")
          end_if
      elif not testtype(f,Type::Arithmetical) then
          error("1st argument must be of 'Type::Arithmetical'")
      end_if
   end_if;
 
   if domtype(x)=DOM_IDENT then
     if not has(f,x) then 
         return(f*2*PI*dirac(y)) 
     end_if
   elif domtype(x/2/PI)=DOM_IDENT then 
     // frequency form:
     return(2*PI*transform::fourier(f,x/2/PI,2*PI*y))
   else
     error("invalid arguments")
   end_if;

   case type(f)
   of "_equal" do
   of "_plus" do return(map(f,transform::fourier,x,y))
   of DOM_IDENT do return(2*I*PI*dirac(y,1)) // can be only x 
   of "heaviside" do // formula 3.1 p. 199 of Oberhettinger90 
      if (_k_:=Type::Linear(op(f),[x]))<>FALSE then
         return( 
           (PI*dirac(y)+I*sign(_k_[1])/y ) *exp(-I*_k_[2]/_k_[1]*y)
                );
      end_if; 
      break
   of "dirac" do
      if nops([op(f)]) = 1 
        then n:= 0; // f = dirac(..)
        else n:= op(f, 2); // f = dirac(.., n)
      end_if:
      if op(f,1)=x then return( (-I*y)^n) end_if;
      if (_k_:=Type::Linear(op(f,1),[x]))<>FALSE then
         return( 
           (-I*y)^n /abs(_k_[1]) * exp(-I*_k_[2]/_k_[1]*y)
               );
      end_if:
      break;
   of "exp" do
      if (_k_:=Type::Linear(op(f),[x]))<>FALSE then
         // f = exp(a*x + b)
         _a := _k_[1]:
         _b_:= _k_[2]:
	return(2*PI*exp(_b_)*dirac(y - _a*I));
      end_if; 
      if not has((_a:=-op(f)/x^2),x)
	then // formula 3.17 p. 11 (even function) 
         return(sqrt(PI)/sqrt(_a)*exp(-y^2/4/_a))
      end_if; 
      break
   of "sin" do 
      if not has((_a:=op(f)/x),x)
	then return(I*PI*(dirac(y-_a)-dirac(y+_a))) end_if; 
      break
   of "cos" do
      if not has((_a:=op(f)/x),x)
	then return(PI*(dirac(y+_a)+dirac(y-_a))) end_if; 
      break
   of "_power" do // expr^k 
      _k_:=op(f,2);
      if op(f,1)=x and testtype(_k_,Type::PosInt) then
         return(2*PI*(-I)^_k_*dirac(y,_k_))
      elif _k_=-1 then
         delete _a, _b_, c;
         if Match2(op(f,1),indep(_a,x)*x+indep(_b_,x)) then
            return(I*PI*exp(-I*y*_b_/_a)/_a*(2*heaviside(y)-1))
         elif Match2(op(f,1),indep(_a,x)*x^2+indep(_b_,x)*x+indep(c,x)) then
            _b_:=_b_/2/_a; c:=c/_a;
            if _b_=0 then
               if sign(c)=-1 then
                    c:=sqrt(-c);
                    return(-PI/c/_a*sin(c*y)*sign(c*y)) // formula 1.29 p. 5 
               else c:=sqrt(c); 
               //   return(PI/c/_a*exp(-c*y)) # formula 1.25 p. 5 #
                    return(
                           PI/c/_a*exp(-c*y)*heaviside( y)
                          +PI/c/_a*exp( c*y)*heaviside(-y)
                          )
               end_if
            end_if; 
            // 1/(a*x^2+b*x+c) = 1/a*1/((x+b/2/a)^2+c/a-b^2/4/a^2) 
            c:=sqrt(_b_^2-c);
            // now 1/a*1/((x+b)^2-c^2) = 1/a/2/c*(1/(x+b-c)-1/(x+b+c)) 
            if not iszero(c) then
                 return(1/(2*_a*c)*(transform::fourier(1/(x+_b_-c),x,y)
                                  -transform::fourier(1/(x+_b_+c),x,y)))
            else return(procname(args())); // !! TODO, insert formula here
            end_if:
         elif Match2(op(f,1),x^4+indep(_a,x)) then
               _a:=sqrt(sqrt(_a)); // formula 1.51 p. 7 
               return(PI/_a^3*exp(-_a*y/2^(1/2))*sin(PI/4+_a*y/2^(1/2)))
         end_if
      elif not has((_a:=op(f,1)-x^2),x) then
         _a:=sqrt(_a); // (x^2+a^2)^_k_ 
         if testtype((m:=-_k_-1),Type::PosInt) then // formula 1.57 p. 7 
            return(2*PI*(2*_a)^(-m-1)*exp(-_a*y)*y^m/fact(m)*
		   _plus(fact(m+_k_)/fact(m-_k_)/(2*_a*y)^_k_ $ _k_=0..m))
	 elif is(Re(-_k_)>0)=TRUE
	   then // formula 2.7 p. 8, includes formula 1.32 p. 5 
            v:=-_k_-1/2;
            return(2*PI^(1/2)*(y/2/_a)^v/gamma(v+1/2)*besselK(v,_a*y))
         end_if
      end_if;
      break
   of "diff" do
    //if op(op(f,1))=x
    //  then return((-I*y)^(nops(f)-1)*procname(op(f,1),x,y))
    //end_if;
      if op(f,2) = x
	then return(-I*y*transform::fourier(diff(op(f,1),op(f,3..nops(f))),x,y))
      end_if;
      break
   of "_mult" do
      if (_k_:=select(f,has,x))<>f then
         return(f/_k_*transform::fourier(_k_,x,y))
      end_if:
      // now all factors contain x 
      delete n, _k_;
      if Match2(f,x^_p(n,Type::PosInt)*_p(_k_)) then
         // fourier(x^n*f(x),x,y)=(-I)^n*diff(fourier(f(x),x,y),y$n) 
	 yy:= genident("yy");
         // return((-I)^n*eval(subs(diff(transform::fourier(_k_,x,yy),yy$n),yy=y)))
         tmp:= transform::fourier(_k_,x,yy):
         if type(tmp) <> "transform::fourier" then
            tmp:= diff(tmp, yy $ n);
            if type(tmp) <> "diff" or      // beware: if y is not an unknown,
               domtype(y) = DOM_IDENT or   // then do not substitute this into
               type(y) = "_index" then     // a symbolic diff expression!
               return( (-I)^n*eval(subs(tmp, yy = y)) );
            end_if;
         end_if;
      end_if;

      //---------------------------------------------------------
      // fourier(exp(a*x)*f(x), x, y) = fourier(f(x), x, y - I*a)
      //---------------------------------------------------------
      if has(f, exp) then
        for i from 1 to nops(f) do
         if type(op(f,i)) = "exp" and (_k_:=Type::Linear(op(op(f,i)),[x]))<>FALSE
         then // f = ff * stuff, ff = exp(_a*x + _b_)
              _a := _k_[1]:
              _b_:= _k_[2]:
              f:= subsop(f, i = 1):
	      return(exp(_b_)*transform::fourier(f, x, y - I*_a));
          end_if:
        end_for:
      end_if;


      if contains([op(f)],heaviside(x))<>0 then
         return(transform::laplace(f/heaviside(x),x,-I*y))
      end_if:
      delete _a, _k_;
      if (g:=Match(f,dirac(x-indep(_a,x))*_p(_k_)))<>FAIL then
         return(eval(subs(_k_*exp(I*y*x),x=_a,op(g))))
      end_if:
      delete _a, _b_;
      if (g:=Match(f,1/(x^2+indep(_a,x)^2)*1/(x^2+indep(_b_,x)^2)))<>FAIL
	then
         _a:=subs(_a,op(g)); _b_:=subs(_b_,op(g));
         // formula 1.34 p. 5 
	 return(PI/(_a^2-_b_^2)*(exp(-_b_*y)/_b_-exp(-_a*y)/_a)) 
      end_if:
      if (g:=expand(f))<>f then
         return(transform::fourier(g,x,y))
      end_if; 
      break
   of "int" do
      if (_a:=Match(f,hold(int)(_p(g)(_p(z))*_p(h)(x-_p(z)),
				_p(z)=-infinity..infinity)))<>FAIL
	then
         g:=subs(g,op(_a)); h:=subs(h,op(_a));
         return(transform::fourier(g(x),x,y)*transform::fourier(h(x),x,y))
      end_if; 
      break
   of "transform::fourier" do
      // special case:
      if op(f,3)=x then 
         return(2*PI*eval(subs(op(f,1),op(f,2)=-y)))
      end_if; 
  
      // more general case with linear shift of Fourier variable:
      if (_k_:=Type::Linear(op(f,3),[x]))<>FALSE then
         // f:= transform::fourier(g(z),z,_a*x + _b_);
         _a := _k_[1]:
         _b_:= _k_[2]:
         if not iszero(_a) then
            return(2*PI/abs(_a)*exp(-I*_b_*y/_a)*eval(subs(op(f,1), op(f,2)=-y/_a)))
         end_if;
      end_if;
      break
   end_case;
   if (g:=Match(f,x^(2*_p(m,Type::PosInt))/(x^2+indep(z,x))^(_p(n)+1)))<>FAIL
     then
      m:=subs(m,op(g)); n:=subs(n,op(g));
      if testtype(n+1-m,Type::PosInt) then
         // formula 1.55 p. 7 
         return((-1)^(m+n)*PI/fact(n)*eval(subs(diff(z^(m-1/2)*
						exp(-y*z^(1/2)),z$n),op(g))))
      end_if
   elif (g:=Match(f,x^(2*_p(m,Type::PosInt))/
		  (indep(_a,x)+x^(2*_p(n,Type::PosInt)))))<>FAIL
     then
      m:=subs(m,op(g)); n:=subs(n,op(g)); _a:=subs(_a,op(g))^(1/2/n);
      // formula 1.56 p. 7 
      return(PI/n*_a^(2*m-2*n+1)*_plus(sin((2*m+1)*(_k_-1/2)*PI/n+_a*y
	*cos((_k_-1/2)*PI/n))*exp(-_a*y*sin((_k_-1/2)*PI/n))$ _k_=1..n))
   end_if;
   procname(args())
end_proc:

transform::fourier:=funcenv(transform::fourier):
transform::fourier:=slot(transform::fourier,"type","transform::fourier"):
transform::fourier:=slot(transform::fourier,"diff",
                         proc(F, z)
                         local f, x, y;
                         begin
                           // F = transform::fourier(f, x, y)
                           f:= op(F, 1):
                           x:= op(F, 2):
                           y:= op(F, 3):
                           // Mathematically, F does not depend on x:
                           if z = x then return(0) end_if; 
                           return(transform::fourier(diff(f, z), x, y) 
                                 +transform::fourier(f*I*x*diff(y,z), x, y)
                                 );
                         end_proc):
