#ifndef __MCL_client_h__
#define __MCL_client_h__

#include "MKT_mssg.h"
#include "MCO_types.h"

class MCLT_client;

//---------------------------------------------------------------------------
// Formatted MuPAD data

/** Format of MuPAD data.
 */
class MCLT_format : public MTformat {
public:
    /// Create format object. The arguments are copied.
    MCLT_format(const char* name=0, Encoding enc=Ascii);

    /// Create format by copying.
    MCLT_format(const MTformat* format);

    /// Destroy format.
    virtual ~MCLT_format();
};


/** ASCII text string data.
 *
 * The string data is copied.
 */
class MCLT_string_data : public MTmemoryData {
public:
    /// Create text data object. The argument is copied.
    MCLT_string_data(const char* format_name, const char* str);

    /// Destroy data.
    virtual ~MCLT_string_data();

    /// Return data string.
    virtual const char* c_str() const;

protected:
    char* string_;
    MTbufferStream buffer_;
};

/** MuPAD language ASCII string data.
 *
 * The data is copied.
 */
class MCLT_mupad_lang_data : public MCLT_string_data {
public:
    /// Create MuPAD language data object. The argument is copied.
    MCLT_mupad_lang_data(const char* cmd);

    /// Destroy data.
    virtual ~MCLT_mupad_lang_data();
};

/** MCode data stored in a stream.
 */
class MCLT_mcode_data : public MTmemoryData {
public:
    /// Create MCode data object given a stream.
    /// The streams content is copied.
    MCLT_mcode_data(const char* format_name=0, MTmemoryStream* str=0);

    /// Destroy data.
    virtual ~MCLT_mcode_data();

    /// Return S-Pointer parsed from data.
    /// Returns MMMNULL in case of an parsing error.
    virtual S_Pointer s_pointer();

    /// Returns 0 in any case.
    virtual const char* c_str() const;

protected:
    MTheapStream data_;
};


//---------------------------------------------------------------------------
/** Message sent by client.
 *
 * Messages are created by client switches and inserted into a global
 * message queue.
 * Several situations exist where messages must be handled depending on
 * their kind and the kernels state:
 *
 * Init:
 * When the kernel starts up a modal state is entered where only the
 * first init message is handled. It is used to initialize the evaluator.
 *
 * Events:
 * Events are handled "quasi asynchronously" during MCL_handle_event.
 * This function is called often during evaluation and inside of PARI,
 * so one may handle messages which do not need to be synchronized
 * with the evaluator.
 *
 * Commands:
 * Commands are handled in the kernels main command loop. The commands
 * are parsed and evaluated. The client of the command actually executed
 * is called the "active client". Any output is sent to this client.
 *
 * Input:
 * If the MuPAD function input() is called a modal state is entered
 * where only events and input messages from the active client are
 * handled.
 *
 * Textinput:
 * If the MuPAD function textinput() is called a modal state is entered
 * where only events and textinput messages from the active client are
 * handled.
 *
 * Callback:
 * If the MuPAD function client_call() is called a modal state is entered
 * where only events and callback result messages from the active client
 * are handled.
 *
 * Debug:
 * If the MuPAD function debug() is called a modal state is entered
 * where only events and debug commands from the active client are
 * handled.
 *
 * Base class, subclassed for each client message type.
 */
class MCLT_mssg {
public:
    /** Remove first init message from queue and handle it.
     * The message is destroyed afterwards.
     * Returns false if no init message was in the queue. */
    static bool handle_init();

    /** Remove first event message from queue and handle it.
     * The message is destroyed afterwards.
     * Returns false if no event message was in the queue. */
    static bool handle_event();

    /** Remove first command message from queue and handle it.
     * The command message is destroyed afterwards.
     * Returns false if no command message was in the queue. */
    static bool handle_command();

    /** Remove first input message of client from queue and handle it.
     * The input message is destroyed afterwards.
     * Returns false if no matching input message was in the queue. */
    static bool handle_input(MCLT_client*, S_Pointer* result);

    /** Remove first textinput message of client from queue and handle it.
     * The input message is destroyed afterwards.
     * Returns false if no matching input message was in the queue. */
    static bool handle_textinput(MCLT_client*, S_Pointer* result);

    /** Remove first callback result message of client from queue and handle it.
     * The callback result message is destroyed afterwards.
     * Returns false if no matching callback result message was in the queue. */
    static bool handle_call_result(MCLT_client*, bool& error, S_Pointer* result);

    /** Remove first debug message of client from queue and handle it.
     * The debug message is destroyed afterwards.
     * Returns false if no matching debug message was in the queue. */
    static bool handle_debug(MCLT_client*, bool& quit);

    /// Append message at back of queue.
    static void push_back(MCLT_mssg*);

    /// Kill all messages for dead clients.
    static void kill_dead();

    /// Kill all messages from a certain client.
    static void kill_client(MCLT_client*);

    /// Kill all command messages from a certain client.
    static void kill_commands(MCLT_client*);

    /// Does the message queue contain any events?
    static bool has_message();

    /// Create message from client.
    MCLT_mssg(MCLT_client* c=0);

    /// Destroy message.
    virtual ~MCLT_mssg();

protected:
    /// Handle message by calling #handle#, then delete this.
    void do_handle();
    void do_handle(S_Pointer* result);
    void do_handle(bool&error, S_Pointer* result);
    void do_handle(bool &res);

    /** Handle message.
     * The default implementation ignores the message. */
    virtual void handle();

    /** Handle message, returning a S_Pointer as result.
     * The default implementation ignores the message. */
    virtual void handle(S_Pointer* result);

    /** Handle message, returning an error status and a S_Pointer as result.
     * The default implementation ignores the message. */
    virtual void handle(bool&error, S_Pointer* result);

    /** Handle message, setting a boolean value as side-effect.
     * The default implementation ignores the message. */
    virtual void handle(bool &res);

    /** Is this a valid message or a dead one which just has not been 
     * deleted properly? */
    bool is_dead() const;

    /// Is this a init message? Default returns false.
    virtual bool is_init();

    /// Is this a event message? Default returns false.
    virtual bool is_event();

    /// Is this a command message? Default returns false.
    virtual bool is_command();

    /// Is this a input message from a certain client? Default returns false.
    virtual bool is_input(MCLT_client*);

    /// Is this a textinput message from a certain client? Default returns false.
    virtual bool is_textinput(MCLT_client*);

    /// Is this a callback result message from a certain client? Default returns false.
    virtual bool is_call_result(MCLT_client*);

    /// Is this a debug message from a certain client? Default returns false.
    virtual bool is_debug(MCLT_client*);

    /// Remove this message from queue and delete it. Returns next entry in queue.
    MCLT_mssg* unlink();

    static MCLT_mssg* g_first_mssg_;
    static MCLT_mssg* g_last_mssg_;

    MCLT_client* client_;
    bool is_dead_;
    MCLT_mssg* next_mssg_;
    MCLT_mssg* prev_mssg_;
};

// input message base class
class MCLT_input_mssg : public MCLT_mssg {
public:
    MCLT_input_mssg(MCLT_client*, bool canceled);
    virtual ~MCLT_input_mssg();
protected:
    virtual bool is_input(MCLT_client*);
    bool canceled_;
};

// input message in MuPAD language format
class MCLT_mupad_lang_input_mssg : public MCLT_input_mssg {
public:
    MCLT_mupad_lang_input_mssg(MCLT_client*, bool canceled, const char* input);
    virtual ~MCLT_mupad_lang_input_mssg();
protected:
    virtual void handle(S_Pointer* result);
    MCLT_mupad_lang_data input_;
};

// input message in MCode tree format
class MCLT_mcode_input_mssg : public MCLT_input_mssg {
public:
    MCLT_mcode_input_mssg(MCLT_client*, bool canceled, MTmemoryStream* input);
    virtual ~MCLT_mcode_input_mssg();
protected:
    virtual void handle(S_Pointer* result);
    MCLT_mcode_data input_;
};

class MCLT_textinput_mssg : public MCLT_mssg {
public:
    MCLT_textinput_mssg(MCLT_client*, bool canceled, const char* input);
    virtual ~MCLT_textinput_mssg();
protected:
    virtual void handle(S_Pointer* result);
    virtual bool is_textinput(MCLT_client*);
    bool canceled_;
    char* input_;
};

class MCLT_call_result_mssg : public MCLT_mssg {
public:
    MCLT_call_result_mssg(MCLT_client*, bool error, MTmemoryStream* result);
    virtual ~MCLT_call_result_mssg();
protected:
    virtual void handle(bool& error, S_Pointer* result);
    virtual bool is_call_result(MCLT_client*);
    bool error_;
    MTheapStream result_;
};

class MCLT_stop_mssg : public MCLT_mssg {
public:
    MCLT_stop_mssg(MCLT_client*);
    virtual ~MCLT_stop_mssg();
protected:
    virtual void handle();
    virtual bool is_event();
};

class MCLT_init_mssg : public MCLT_mssg {
public:
    MCLT_init_mssg(MCLT_client*, size_t argc, char** argv);
    virtual ~MCLT_init_mssg();
protected:
    virtual void handle();
    virtual bool is_init();
    size_t argc_;
    char** argv_;
};

class MCLT_quit_mssg : public MCLT_mssg {
public:
    MCLT_quit_mssg(MCLT_client*);
    virtual ~MCLT_quit_mssg();
protected:
    virtual void handle();
    virtual bool is_event();
};

class MCLT_textwidth_mssg : public MCLT_mssg {
public:
    MCLT_textwidth_mssg(MCLT_client*, long width);
    virtual ~MCLT_textwidth_mssg();
protected:
    virtual void handle();
    virtual bool is_command();
    long width_;
};

class MCLT_pretty_print_mssg : public MCLT_mssg {
public:
    MCLT_pretty_print_mssg(MCLT_client*, bool pretty);
    virtual ~MCLT_pretty_print_mssg();
protected:
    virtual void handle();
    virtual bool is_command();
    bool pretty_;
};

class MCLT_format_mssg : public MCLT_mssg {
public:
    MCLT_format_mssg(MCLT_client*, const char* output_type, const char* format);
    virtual ~MCLT_format_mssg();
protected:
    virtual void handle();
    virtual bool is_command();
    char* output_type_;
    char* format_;
};

class MCLT_status_wanted_mssg : public MCLT_mssg {
public:
    MCLT_status_wanted_mssg(MCLT_client*);
    virtual ~MCLT_status_wanted_mssg();
protected:
    virtual void handle();
    virtual bool is_event();
};

class MCLT_completion_wanted_mssg : public MCLT_mssg {
public:
    MCLT_completion_wanted_mssg(MCLT_client*, const char* input);
    virtual ~MCLT_completion_wanted_mssg();
protected:
    virtual void handle();
    virtual bool is_command();
    char* input_;
};

class MCLT_dbg_cmd_mssg : public MCLT_mssg {
public:
    MCLT_dbg_cmd_mssg(MCLT_client*, unsigned char cmd, const char* fname=0,
                      unsigned long fpos=0, const char* expr=0);
    virtual ~MCLT_dbg_cmd_mssg();
protected:
    virtual void handle(bool& quit);
    virtual bool is_debug(MCLT_client*);
    unsigned char cmd_;
    char* fname_;
    unsigned long fpos_;
    char* expr_;
};

/* MuPAD command message sent by client.
 *
 * Base class, subclassed for each MuPAD command message type.
 */
class MCLT_command_mssg : public MCLT_mssg {
public:
    MCLT_command_mssg(MCLT_client*, void* cookie,
                      MTexecutionFlag flags=0, const char* format=0);
    virtual ~MCLT_command_mssg();
protected:
    virtual bool is_command();
    virtual MTexecutionFlag flags() const;
    virtual const char* format() const;
    void* cookie_;
    MTexecutionFlag flags_;
    char* format_;
};

class MCLT_mupad_lang_command : public MCLT_command_mssg {
public:
    MCLT_mupad_lang_command(MCLT_client*, void* cookie, const char* cmd,
                            MTexecutionFlag flags=0, const char* format=0);
    virtual ~MCLT_mupad_lang_command();
protected:
    MCLT_mupad_lang_data cmd_;
};

class MCLT_mupad_lang_cmd_mssg : public MCLT_mupad_lang_command {
public:
    MCLT_mupad_lang_cmd_mssg(MCLT_client*, void* cookie, const char* cmd,
                             MTexecutionFlag flags=0, const char* format=0);
    virtual ~MCLT_mupad_lang_cmd_mssg();
protected:
    virtual void handle();
};

class MCLT_mupad_lang_plot_cmd_mssg : public MCLT_mupad_lang_command {
public:
    MCLT_mupad_lang_plot_cmd_mssg(MCLT_client*, void* cookie, const char* cmd);
    virtual ~MCLT_mupad_lang_plot_cmd_mssg();
protected:
    virtual void handle();
};

class MCLT_mcode_cmd_mssg : public MCLT_command_mssg {
public:
    MCLT_mcode_cmd_mssg(MCLT_client*, void* cookie, MTmemoryStream* cmd,
                        MTexecutionFlag flags=0, const char* format=0);
    virtual ~MCLT_mcode_cmd_mssg();
protected:
    virtual void handle();
    MCLT_mcode_data cmd_;
};

//---------------------------------------------------------------------------

/** Client switch.
 *
 * A switch implements the communication of a client with the kernel. The clients
 * are represented as proxies in the kernel (class MCLT_client). The switch recieves
 * incoming client messages, sends outgoing kernel messages and manages the livetime
 * of the client proxies. Incoming messages are translated into internal kernel
 * messages (class MCLT_mssg).
 *
 * Concrete subclasses exist for the different platforms, transport media and protocols.
 */
class MCLT_switch {
public:
    /// Returns true if all switches want to quit if no more clients exist.
    static bool all_switches_quit_without_clients();

    /// Create new client switch.
    MCLT_switch();

    /// Destroy client.
    virtual ~MCLT_switch();

    /// Test if client has a certain feature.
    virtual bool has_feature(const MCLT_client* client, MTclientFeature feature) const;

    /// Initialize kernel.
    virtual void do_init(MCLT_client* client, size_t argc, char* argv[]);

    /// Evaluate command.
    virtual void do_cmd(MCLT_client* client, void* cookie, MTmemoryData* cmd, 
                        MTexecutionFlag flags, const char* format) = 0;

    /// Evaluate plot command.
    virtual void do_plot_cmd(MCLT_client* client, void* cookie, MTmemoryData* cmd) = 0;

    /// Parse input given as mupad language string.
    virtual S_Pointer do_input(MCLT_client* client, bool canceled, MTmemoryData* input) = 0;

    /// Called if debugging mode is stopped by client.
    virtual void on_stop_debug(MCLT_client* client);

    /// Called after a command has been executed.
    virtual void cmd_done(MCLT_client* client) = 0;

    /// Called after a plot command has been executed.
    virtual void plot_cmd_done(MCLT_client* client) = 0;

    /// Called after a debug command has been executed.
    virtual void dbg_cmd_done(MCLT_client* client) = 0;

    /// Send plotting data stored in S-Pointer to UI
    virtual bool put_plot_data(MCLT_client* client, S_Pointer data) = 0;

    /** Refresh debugger display variable list.
     * The default implementation does nothing. */
    virtual void refresh_display_vars(MCLT_client* client);

    /** Refresh debugger breakpoint list.
     * The default implementation does nothing. */
    virtual void refresh_breakpoints(MCLT_client* client);

    /** Refresh debugger call stack list.
     * The default implementation does nothing. */
    virtual void refresh_call_stack(MCLT_client* client);

    /** Query if client really wants to stop command execution.
     * This implementation always returns true. */
    virtual bool ack_stop(MCLT_client* client);

    //@name Sending Messages to the Real Clients
    /** These methods return true if the message was sent successfully.
     */
    //@{
    /// Send client id message.
    virtual bool send_client_id(MCLT_client* client) = 0;

    /// Send kernel ready message.
    virtual bool send_kernel_ready(MCLT_client* client) = 0;

    /// Send output message.
    virtual bool send_output(MCLT_client* client, void* cmd_cookie, const char* output_type,
                             MTmemoryData* output) = 0;

    /// Send finished message.
    virtual bool send_finished(MCLT_client* client, void* cmd_cookie) = 0;

    /// Send complete command message.
    virtual bool send_complete_cmd(MCLT_client* client, void* cmd_cookie);

    /// Sent input wanted message.
    virtual bool send_input_wanted(MCLT_client* client, void* cmd_cookie, bool text_input,
                                   MTmemoryData* prompt) = 0;

    /// Sent callback message.
    virtual bool send_callback(MCLT_client* client, void* cmd_cookie,
                               MTmemoryStream* call) = 0;

    /// Send help message.
    virtual bool send_help(MCLT_client* client, const char* topic) = 0;

    /// Send system message.
    virtual bool send_system(MCLT_client* client, const char* cmd) = 0;

    /// Send textwidth changed message.
    virtual bool send_textwidth_changed(MCLT_client* client, long width) = 0;

    /// Send pretty-print changed message.
    virtual bool send_pretty_print_changed(MCLT_client* client, bool pretty) = 0;

    /// Send format changed message.
    virtual bool send_format_changed(MCLT_client* client, const char* output_type,
                                     const MTformat* format) = 0;

    /// Send select format message.
    virtual bool send_select_format(MCLT_client* client, const char* output_type,
                                    size_t num_formats, const MTformat* formats) = 0;

    /// Send status message.
    virtual bool send_status(MCLT_client* client) = 0;

    /// Send input completion message
    virtual bool send_input_completion(MCLT_client* client, const char* completionList, const char* completion) = 0;

    /// Return value of last completion / only needed for TTY
    virtual char *completion();

    /// Return value of last completion list / only needed for TTY
    virtual char *all_possible_completions();

    /// Send plot result file message
    virtual bool send_plot_result_file(MCLT_client* client, void* cmd_cookie, const char* fname) = 0;

    /// Send debug command wanted message.
    virtual bool send_dbg_cmd_wanted(MCLT_client* client, void* cmd_cookie) = 0;

    /// Send debug position message.
    virtual bool send_dbg_info(MCLT_client* client, unsigned char cmd, const char* fname,
                               long pos, const char* procname, long lev, long depth) = 0;

    /// Send debug display list message.
    virtual bool send_display_list(MCLT_client* client, size_t num_vars, const char* names[],
                                   const MTmemoryData* values[]) = 0;

    /// Send debug breakpoint list message.
    virtual bool send_breakpoint_list(MCLT_client* client, size_t num_bpts, const char* fnames[],
                                      long lines[], const char* conditions[]) = 0;

    /// Send debug call stack message.
    virtual bool send_call_stack(MCLT_client* client, size_t level, size_t num_calls,
                                 const char* fnames[], long lines[], const char* calls[]) = 0;

    /// Send init wanted message.
    virtual bool send_init_wanted(MCLT_client* client) = 0;

    /// Send reset message.
    virtual bool send_reset(MCLT_client* client) = 0;

    /// Send quit message.
    virtual bool send_quit() = 0;
    //@}

protected:
    /// May the kernel quit if no more clients exist?
    virtual bool quit_without_clients();

private:
    static MCLT_switch* g_first_switch_;
    MCLT_switch* next_switch_;
};

//---------------------------------------------------------------------------

/** Kernel client.
 *
 * A client object is a proxy which communicates with the "real client" via
 * a MCLT_switch.
 */
class MCLT_client {
public:
    /// Get client object given a client id.
    static MCLT_client* get_client(long client_id);

    /// Get client object given its name.
    static MCLT_client* get_client(const char* name);

    /// Get first client. Returns 0 if no clients exist.
    static MCLT_client* first_client();

    /// Remove all dead clients.
    static void remove_dead();

    /// Send quit message to all living clients.
    static void quit_all();

    /// Mark all clients of given switch as dead.
    static void remove_switch(MCLT_switch* sw);

    /// Create new client object.
    MCLT_client(MCLT_switch* sw, const char* name=0,
                const char* namesp=0,
		MTlongFormat long_fmt=MCstringLong,
		MTfloatFormat float_fmt=MCieeeFloat,
		MTcoding coding=MCflatCoding|MCidentCoding,
		MTclientFeature* features=0);

    /// Destroy client.
    ~MCLT_client();

    /// Mark client as dead.
    void mark_dead();

    /// Is client alive?
    bool alive() const;

    /// Get next client. Returns 0 if no further client exists.
    MCLT_client* next_client() const;

    /// Get id of client.
    long id() const;

    /// Is this the first client created?
    bool is_first_client() const;

    /// Test if client has a certain feature.
    bool has_feature(MTclientFeature feature) const;

    /// Returns the format of PARI numbers in MCode understood by the client.
    MTlongFormat mcode_long_format() const;

    /// Returns the format of floats in MCode understood by the client.
    MTfloatFormat mcode_float_format() const;

    /// Returns the coding format for MCode as wanted by the client.
    MTcoding mcode_coding() const;

    /// Make this client the actual command handler.
    void activate();

    /// Initialize kernel.
    void do_init(size_t argc, char* argv[]);

    /// Start handling command identified by cookie.
    void on_cmd(void* cookie);

    /// Evaluate command.
    void do_cmd(MTmemoryData* cmd, MTexecutionFlag flags, 
                const char* format);

    /// Evaluate plot command.
    void do_plot_cmd(MTmemoryData* cmd);

    /// Parse input.
    S_Pointer do_input(bool canceled, MTmemoryData* input);

    /// Called if debugging mode is stopped by client.
    void on_stop_debug();

    /// Called after a command has been executed.
    void cmd_done();

    /// Called after a plot command has been executed.
    void plot_cmd_done();

    /// Called after a debug command has been executed.
    void dbg_cmd_done();

    /// Send plotting data stored in S-Pointer to UI.
    bool put_plot_data(S_Pointer data);

    /// Refresh debugger display variable list.
    void refresh_display_vars();

    /// Refresh debugger breakpoint list.
    void refresh_breakpoints();

    /// Refresh debugger call stack list.
    void refresh_call_stack();

    /// Query if client really wants to stop command execution.
    bool ack_stop();

    //@name Sending Messages to the Client
    /** The messages are sent via the clients switch.
     * These methods return true if the message was sent successfully.
     */
    //@{
    /// Send client id message.
    bool send_client_id();

    /// Send kernel ready message.
    bool send_kernel_ready();

    /// Send output message.
    bool send_output(const char* output_type, MTmemoryData* output);

    /// Send finished message.
    bool send_finished();

    /// Send complete command message.
    bool send_complete_cmd();

    /// Sent input wanted message.
    bool send_input_wanted(bool text_input, MTmemoryData* prompt);

    /// Sent callback message.
    bool send_callback(MTmemoryStream* call);

    /// Send help message.
    bool send_help(const char* topic);

    /// Send system message.
    bool send_system(const char* cmd);

    /// Send textwidth changed message.
    bool send_textwidth_changed(long width);

    /// Send pretty-print changed message.
    bool send_pretty_print_changed(bool pretty);

    /// Send format changed message.
    bool send_format_changed(const char* output_type, const MTformat* format);

    /// Send select format message.
    bool send_select_format(const char* output_type, size_t num_formats,
                            const MTformat* formats);

    /// Send status message.
    bool send_status();

    /// Send input completion message
    bool send_input_completion(const char* completionList, const char* completion);

    /** Return value of last completion / only needed for TTY.
     * String returned must be freed by caller. */
    char *completion();

    /** Return value of last completion list / only needed for TTY.
     * String returned must be freed by caller. */
    char *all_possible_completions();

    /// Send plot result file message
    bool send_plot_result_file(const char* fname);

    /// Send init wanted message.
    bool send_init_wanted();

    /// Send debug command wanted message.
    bool send_dbg_cmd_wanted();

    /// Send debug position message.
    bool send_dbg_info(unsigned char cmd, const char* fname, long pos,
                       const char* procname, long lev, long depth);

    /// Send debug display list message.
    bool send_display_list(size_t num_vars, const char* names[],
                           const MTmemoryData* values[]);

    /// Send debug breakpoint list message.
    bool send_breakpoint_list(size_t num_bpts, const char* fnames[],
                              long lines[], const char* conditions[]);

    /// Send debug call stack message.
    bool send_call_stack(size_t level, size_t num_calls,
			 const char* fnames[], long lines[],
			 const char* calls[]);

    /// Send reset message.
    bool send_reset();

    /// Send quit message.
    bool send_quit();
    //@}

protected:
    static MCLT_client* g_first_client_;
    static long g_next_client_id_;

    MCLT_switch* switch_;
    MCLT_client* next_client_;
    MCLT_client* prev_client_;
    long client_id_;
    long context_;
    void* cmd_cookie_;
    char* namespace_;
    char* name_;
    MTlongFormat long_format_;
    MTfloatFormat float_format_;
    MTcoding coding_;
    MTclientFeature* features_;
};

//---------------------------------------------------------------------------

/** Handle operating system and MuPAD client event messages.
 * Yields control to other apps if no events exist.
 * Is called in the diverse message loops of the kernel. */
void MCL_handle_events(bool main_loop);

/** Handle operating system and MuPAD client event messages,
 * but does not yield control to other apps.
 * Is called during longer computations of the kernel. */
void MCL_peek_events();

/** Handle operating system events.
 * @param main_loop called in main loop
 * @param wait yield control to other apps if no events exist */
void MCL_handle_os_events(bool main_loop, bool wait);

/** Returns the active client handling the actual command.
 * May return 0 if there is no such client. */
MCLT_client* MCL_act_client();

/** Returns the active client if it has the given feature.
 * Returns 0 if there is no active client or if is does not have this feature. */
inline MCLT_client* MCL_act_client_if(MTclientFeature feature)
{
    MCLT_client* cl = MCL_act_client();
    if (cl != 0 && cl->has_feature(feature))
        return cl;
    return 0;
}

#endif
