// MCL_lib_switch.h -- client switch for link-library kernel

#ifndef __MCL_lib_switch_h__
#define __MCL_lib_switch_h__

#include "MCL_client.h"
#include "MKT_mssg.h"

class MTlibKernel;

/** Link-library kernel client switch.
 *
 * Client switch for kernel implemented as link-library.
 * A singleton class.
 */

class MCLT_lib_switch : public MCLT_switch {
public:
    /// Get singleton switch object.
    static MCLT_lib_switch& object();

    /// Calls event handler during longer calculations.
    static void handle_events();

    /// Calls fatal exit handler.
    static void handle_fatal_exit();

    // Destroy switch.
    virtual ~MCLT_lib_switch();

    /** Set fatal exit handler.
     * This installs a callback which is executed when the kernel
     * causes a fatal exit. The previous exit handler is returned. */
    void (*fatal_exit_handler(void (*handler)()))();

    /** Set event handler.
     * This installs a callback which is called quite often by the kernel
     * during longer calculations. It may be used to keep a user interface
     * reactive for example. The previous event handler is returned. The
     * user_data are handled back to the handler. */
    void (*event_handler(void (*handler)(void*), void* user_data))(void*);

    /** Init kernel.
     * Must be called before kernel may be used. May be called more than
     * once. Do not use the kernel if false is returned. */
    bool init(const char* progname);

    /** Exit kernel.
     * Must be called when done with the kernel. */
    void exit();

    /// Create new client object for given kernel object, return its ID.
    bool new_client(MTlibKernel* kernel, const char* name, const char* namesp,
		    MTlongFormat long_fmt, MTfloatFormat float_fmt, MTcoding coding,
                    size_t num_features, MTclientFeature features[],
                    bool* init_wanted, long* client_id);

    /// Remove the kernel client from the switches list of known external clients.
    void remove_client(MTlibKernel* kernel);

    /// Evaluate command.
    virtual void do_cmd(MCLT_client* client, void* cookie, MTmemoryData* cmd,
                        MTexecutionFlag flags, const char* format);

    /// Evaluate plot command.
    virtual void do_plot_cmd(MCLT_client* client, void* cookie, MTmemoryData* cmd);

    /// Parse input.
    virtual S_Pointer do_input(MCLT_client* client, bool canceled, MTmemoryData* input);

    /// Called after a command has been executed.
    virtual void cmd_done(MCLT_client* client);

    /// Called after a plot command has been executed.
    virtual void plot_cmd_done(MCLT_client* client);

    /// Called after a debug command has been executed.
    virtual void dbg_cmd_done(MCLT_client* client);

    /// Send plotting data stored in S-Pointer to UI
    virtual bool put_plot_data(MCLT_client* client, S_Pointer data);

    /// Refresh debugger display variable list.
    virtual void refresh_display_vars(MCLT_client* client);

    /// Refresh debugger breakpoint list.
    virtual void refresh_breakpoints(MCLT_client* client);

    /// Refresh debugger call stack list.
    virtual void refresh_call_stack(MCLT_client* client);

    /// Query if client really wants to stop command execution.
    virtual bool ack_stop(MCLT_client* client);

    //@name Sending Messages to the Real Clients
    /** These methods return true if the message was sent successfully.
     */
    //@{
    /// Send client id message.
    virtual bool send_client_id(MCLT_client* client);

    /// Send kernel ready message.
    virtual bool send_kernel_ready(MCLT_client* client);

    /// Send output message.
    virtual bool send_output(MCLT_client* client, void* cmd_cookie, const char* output_type,
                             MTmemoryData* output);

    /// Send finished message.
    virtual bool send_finished(MCLT_client* client, void* cmd_cookie);

    /// Sent input wanted message.
    virtual bool send_input_wanted(MCLT_client* client, void* cmd_cookie, bool text_input, 
                                   MTmemoryData* prompt);
    
    /// Sent callback message.
    virtual bool send_callback(MCLT_client* client, void* cmd_cookie, MTmemoryStream* call);

    /// Send help message.
    virtual bool send_help(MCLT_client* client, const char* topic);

    /// Send system message.
    virtual bool send_system(MCLT_client* client, const char* cmd);

    /// Send textwidth changed message.
    virtual bool send_textwidth_changed(MCLT_client* client, long width);

    /// Send pretty-print changed message.
    virtual bool send_pretty_print_changed(MCLT_client* client, bool pretty);

    /// Send format changed message.
    virtual bool send_format_changed(MCLT_client* client, const char* output_type, 
                                     const MTformat* format);

    /// Send select format message.
    virtual bool send_select_format(MCLT_client* client, const char* output_type, 
                                    size_t num_formats, const MTformat* formats);

    /// Send status message.
    virtual bool send_status(MCLT_client* client);

    /// Send input completion message
    virtual bool send_input_completion(MCLT_client* client, const char* list, 
                                       const char* completion);

    /// Send plot result file message
    virtual bool send_plot_result_file(MCLT_client* client, void* cmd_cookie, const char* fname);

    /// Send debug command wanted message.
    virtual bool send_dbg_cmd_wanted(MCLT_client* client, void* cmd_cookie);

    /// Send debug position message.
    virtual bool send_dbg_info(MCLT_client* client, unsigned char cmd, const char* fname, 
                               long pos, const char* procname, long lev, long depth);

    /// Send debug display list message.
    virtual bool send_display_list(MCLT_client* client, size_t num_vars, const char* names[],
                                   const MTmemoryData* values[]);

    /// Send debug breakpoint list message.
    virtual bool send_breakpoint_list(MCLT_client* client, size_t num_bpts, const char* fnames[],
                                      long lines[], const char* conditions[]);

    /// Send debug call stack message.
    virtual bool send_call_stack(MCLT_client* client, size_t level, size_t num_calls, 
                                 const char* fnames[], long lines[], const char* calls[]);

    /// Send init wanted message.
    virtual bool send_init_wanted(MCLT_client* client);

    /// Send reset message.
    virtual bool send_reset(MCLT_client* client);

    /// Send quit message.
    virtual bool send_quit();
    //@}

protected:
    static MCLT_lib_switch switch_;

    // Create new switch.
    MCLT_lib_switch();

    // Get message handler for external client.
    MTclientMessageHandler* mssg_handler(MCLT_client* client);

    void (*fatal_exit_handler_)();
    void (*event_handler_)(void*);
    void* event_user_data_;
    MTlibKernel** lib_kernel_;
    size_t num_kernels_;
    bool init_;
};

inline MCLT_lib_switch& MCLT_lib_switch::object() {
    return switch_;
}

#endif // ! __MCLT_lib_switch_h__
