#ifndef __MEN_string_h__
#define __MEN_string_h__

/* MEN_string.h -- Support for UTF8 and multi-byte strings.
 *
 * Currently multi-byte strings are supported for WIN32 only.
 * 
 * IMPORTANT: The following assumption is made for multi-byte string
 * encodings: ASCII characters in the range 0..128 are represented by
 * 'themselve', ie. their encoding is simply a single byte ASCII char.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

/** Test if c is a lead byte of a multi-byte character.
 * Returns 0 if it is not a lead byte. */
int MEN_isleadbyte(int c);

/** Increment char pointer to the next logical character in a multi-byte
 * string according to the current locale. */
char* MEN_mbsinc(char* s);
const char* MEN_mbsinc(const char* s);

/** Returns the number of logical characters in a multi-byte
 * string according to the current locale. */
size_t MEN_mbsclen(const char *s);

/** Find first occurence of ASCII character c in in a multi-byte
 * string according to the current locale. */
char* MEN_mbschr(const char* s, int c);

/** Find last occurence of ASCII character c in in a multi-byte
 * string according to the current locale. */
char* MEN_mbsrchr(const char* s, int c);

/** Find first occurence of substring in given string. */
char *MEN_mbsstr(const char *s, const char *substr);

/** Find first occurence of some character from set in given string. */
size_t MEN_mbsspn(const char *s, const char *charset);

/** Find first occurence of some character not from set in given string. */
size_t MEN_mbscspn(const char *s, const char *charset);

/** Find first occurence of some character from set in given string. */
char *MEN_mbspbrk(const char *s, const char *charset);

/** Return length of logical character at given address in a multi-byte
 * string according to the current locale (in bytes). */
size_t MEN_mbscharlen(const char* s);

/** Perform a lowercase comparison of multi-byte strings according to 
 * the current locale. */
int MEN_mbsicmp(const char*, const char*);

/** Convert strings to a long integer value. */
long MEN_mbstol(const char *nptr, char **endptr, int base);

// Implementation -----------------------------------------------------------

#ifdef WIN32

#include <ctype.h>
#include <tchar.h>

inline int MEN_isleadbyte(int c) { return _istleadbyte(c); }

inline char* MEN_mbsinc(char* s) { return _tcsinc(s); }
inline const char* MEN_mbsinc(const char* s) { return _tcsinc(s); }
inline char* MEN_mbschr(const char* s, int c) { return _tcschr(s, c); }
inline char* MEN_mbsrchr(const char* s, int c) { return _tcsrchr(s, c); }
inline char *MEN_mbsstr(const char *s, const char *substr) { return _tcsstr(s, substr); }
inline size_t MEN_mbsspn(const char *s, const char *charset) { return _tcsspn(s, charset); }
inline size_t MEN_mbscspn(const char *s, const char *charset) { return _tcscspn(s, charset); }
inline char *MEN_mbspbrk(const char *s, const char *charset) { return _tcspbrk(s, charset); }
inline int MEN_mbsicmp(const char* a, const char* b) { return _tcsicmp(a, b); }
inline long MEN_mbstol(const char *n, char **end, int b) { return _tcstol(n, end, b); }
inline size_t MEN_mbsclen(const char *s) { return _tcsclen(s); }

#else // ! WIN32

// MBCS are not supported yet!

#ifdef MACINTOSH
extern int strcasecmp(const char *s1, const char *s2);
#endif

inline int MEN_isleadbyte(int c) { return 0; }

inline char* MEN_mbsinc(char* s) { ++s; return s; }
inline const char* MEN_mbsinc(const char* s) { ++s; return s; }
inline char* MEN_mbschr(const char* s, int c) { return strchr(s, c); }
inline char* MEN_mbsrchr(const char* s, int c) { return strrchr(s, c); }
inline char *MEN_mbsstr(const char *s, const char *substr) { return strstr(s, substr); }
inline size_t MEN_mbsspn(const char *s, const char *charset) { return strspn(s, charset); }
inline size_t MEN_mbscspn(const char *s, const char *charset) { return strcspn(s, charset); }
inline char *MEN_mbspbrk(const char *s, const char *charset) { return strpbrk(s, charset); }
inline int MEN_mbsicmp(const char* a, const char* b) { return strcasecmp(a, b); }
inline long MEN_mbstol(const char *n, char **end, int b) { return strtol(n, end, b); }
inline size_t MEN_mbsclen(const char *s) { return strlen(s); }

#endif // ! WIN32

inline size_t MEN_mbscharlen(const char* s)
{
    return MEN_mbsinc(s) - s;
}

#endif // !__MEN_string_h__
