/* $Date: 2000/11/24 10:13:30 $  $Author: tonner $  $Revision: 1.67 $ */

/**************************************************************************/
/* FILE   : MEV_evaluate.h                                                */
/* AUTHOR : Andreas Kemper, Andreas Sorgatz                               */
/* CREATED: 25.04.1991                                                    */
/* CHANGED: 10.09.1997                                                    */
/**************************************************************************/

#ifndef __MEV_evaluate___
#define __MEV_evaluate___


#ifdef DOS_FILENAMES
#  include "MTR_tree.h"
#  include "MEV_evah.h"
#  include "MID_iden.h"
#else
#  include "MTR_tree.h"
#  include "MEV_evaluate_help.h"
#  include "MID_ident.h"
#endif

/**@name Evaluator Error Handling

If an error occures during evaluation the function #MEV_set_error# is
called which sends the error message to the client and makes a longjmp
to the label #MEVV_error_label#. An error code is stored in #MEVV_ERROR#.

One may try an operation and catch errors as follows:
\begin{verbatim}
MEVC_TRY
{
    ... // code which may throw an evaluator error
}
MEVC_CATCH_ERR
{
    ... // cleanup code if an evaluator error was thrown
}
MEVC_END_CATCH
\end{verbatim}
The #MEVC_...# macros above a quite fragile, please be shure to place the
braces correctly.
*/
//@{

/** Error code of last error thrown by #MEV_set_error#.
 * Is set to 0 when the error is handled finally. */
extern long MEVV_ERROR;

/// Label for long jumps after errors during evaluation.
extern jmp_buf MEVV_error_label;

/// Throws an error, i.e. makes a longjmp to #MEVV_error_label#.
#define MEVC_THROW_ERR \
    longjmp(MEVV_error_label, 1)

/** Trys an operation.
 * If an error occures it may be catched via #MEVC_CATCH_ERR#. */
#define MEVC_TRY \
{ \
    jmp_buf MEVV_old_error_label; \
    memcpy(MEVV_old_error_label, MEVV_error_label, sizeof(jmp_buf)); \
    if (setjmp(MEVV_error_label) == 0)

/** Catches an evaluation error.
 * Catches an error after a #MEVC_TRY# block. */
#define MEVC_CATCH_ERR \
    else

/// Marks the end of a #MEVC_CATCH_ERR# block.
#define MEVC_END_CATCH \
    memcpy(MEVV_error_label, MEVV_old_error_label, sizeof(jmp_buf)); \
}

/** Prints an error message and throws an error.

 Prints the error text corresponding to the error number #err_nr#. This
 text may contain formatting information like #printf#. Additional arguments
 after #err_str# are used to print the text.

 Eventually a longjmp to the global label #MEVV_error_label# is performed.

 @param err_nr error code (must be positive)
 @param err_str error text, usually the name of the MuPAD function
   throwing the error
 @param ... additional parameters according to the error text corresponding
   to the error code #err_nr#
 @return does not return
*/
void MEV_set_error( long err_nr, const char *err_str Mva_arglist) MUPC_NORETURN;

/** Prints a warning.

 Prints the warning text corresponding to the warning number #err_nr#. This
 text may contain formatting information like #printf#. Additional arguments
 after #err_str# are used to print the text.

 @param err_nr warning code (must be negative)
 @param err_str error text, usually the name of the MuPAD function
   throwing the error
 @param ... additional parameters according to the text corresponding
   to the warning code #err_nr#
*/
void MEV_warning( long err_nr, const char *err_str Mva_arglist);

/** Saving the global state of the evaluator.

 Is used to restore the evaluators state after an error.
*/
class MEVT_save_state {
public:
    /// Create state saving object.
    MEVT_save_state();

    /// Destroy state saving object.
    ~MEVT_save_state();

    /// Reset evaluators state after an error.
    void error_reset();

    /// Discard saved state and refresh according to actual state.
    void refresh();

private:
    long level;
    long prev_level;
    long digits;
    long order;
    long rounding;
    long history;
    long maxlevel;
    long maxdepth;
    long textwidth;
    short typecheck;
    bool pretty_print;
    S_Pointer read_path;
    S_Pointer write_path;
    S_Pointer lib_path;
};

//@}

/**************************************************************************/
/* MEVTrefreshoption                                                      */
/**************************************************************************/
typedef long MEVTrefreshoption ;

#define MEVC_refreshoption_assign       0
#define MEVC_refreshoption_push         1
#define MEVC_refreshoption_pop          1

/**************************************************************************/
/* Optionen f"ur die domain-Funktionen                                    */
/**************************************************************************/
#define MEEC_domoption_ass_part            1   /*Interne Konstante*/
#define MEEC_domoption_ass                 1
#define MEEC_domoption_exec                0
#define MEEC_domoption_quick               2
#define MEEC_domoption_non_EXT_only        4
#define MEEC_domoption_EXT_only            8
#define MEEC_domoption_standard            0

#define MEEC_answer_ext                    1  /*Es wurde ein EXT-Element */
                                              /*gefunden*/
#define MEEC_answer_non_ext                2  /*Es wurde ein nicht EXT-  */
                                              /*Element gefunden.        */

/**************************************************************************/
/* Optionen f"ur die func_env-Funktionen                                  */
/**************************************************************************/
#define MEEC_func_envoption_standard       0


/**************************************************************************/
/* Deklaration globaler Variablen des Evaluierers.                        */
/**************************************************************************/

extern S_Pointer      MEVV_DOMAIN_ASS_MEM;
extern S_Pointer      MEVV_EVAL_LAST;
extern volatile short MEVV_STATUS;
extern long           MEVV_ASSIGNED;

#ifdef MMMD_LIT_SP
extern S_Pointer MEVV_LIT_DOMAIN_ASS_MEM;
#endif

extern short MEVV_LegalOpTab[MEVC_MAX_OPERATORS][MEVC_MAX_OPERANDS] ;

MTR_DECL_LITERAL(MEVV_INTERPRETER)
MTR_DECL_LITERAL(MEVV_EVAL_EXPR_FUNCTION_NAME)
MTR_DECL_LITERAL(MEVV_ASS_FLOAT)

MTR_DECL_LITERAL( MEVV_DOMAIN_ASS_MINUS )
#define MEVC_DOMAIN_NUM_ASS_MINUS                        1
MTR_DECL_LITERAL( MEVV_DOMAIN_ASS_DIV )
#define MEVC_DOMAIN_NUM_ASS_DIV                          2
MTR_DECL_LITERAL( MEVV_DOMAIN_ASS_NOT )
#define MEVC_DOMAIN_NUM_ASS_NOT                          3
MTR_DECL_LITERAL( MEVV_DOMAIN_ASS_EVALUATE )
#define MEVC_DOMAIN_NUM_ASS_EVALUATE                     4
MTR_DECL_LITERAL( MEVV_DOMAIN_ASS_OP )
#define MEVC_DOMAIN_NUM_ASS_OP                           5
MTR_DECL_LITERAL( MEVV_DOMAIN_ASS_NOPS )
#define MEVC_DOMAIN_NUM_ASS_NOPS                         6
MTR_DECL_LITERAL( MEVV_DOMAIN_ASS_SUBSOP )
#define MEVC_DOMAIN_NUM_ASS_SUBSOP                       7
MTR_DECL_LITERAL( MEVV_DOMAIN_ASS_SUBS )
#define MEVC_DOMAIN_NUM_ASS_SUBS                         8
MTR_DECL_LITERAL( MEVV_DOMAIN_ASS_TYPE )
#define MEVC_DOMAIN_NUM_ASS_TYPE                         9
MTR_DECL_LITERAL( MEVV_DOMAIN_ASS_DOMTYPE )
#define MEVC_DOMAIN_NUM_ASS_DOMTYPE                     10
MTR_DECL_LITERAL( MEVV_DOMAIN_ASS_NAME )
#define MEVC_DOMAIN_NUM_ASS_NAME                        11
MTR_DECL_LITERAL( MEVV_DOMAIN_ASS_POSTEVAL )
#define MEVC_DOMAIN_NUM_ASS_POSTEVAL                    12
MTR_DECL_LITERAL( MEVV_DOMAIN_ASS_KEY )
#define MEVC_DOMAIN_NUM_ASS_KEY                         13
MTR_DECL_LITERAL( MEVV_DOMAIN_ASS_SLOT )
#define MEVC_DOMAIN_NUM_ASS_SLOT                        14
MTR_DECL_LITERAL( MEVV_DOMAIN_ASS_MAKE_SLOT )
#define MEVC_DOMAIN_NUM_ASS_MAKE_SLOT                   15

#define MEVC_NUMBER_OF_QUICK_VALUES                     15

MTR_DECL_LITERAL( MEVV_DOMAIN_ASS_EVALUATEFUNCTION )
MTR_DECL_LITERAL( MEVV_DOMAIN_ASS_INDEX )
MTR_DECL_LITERAL( MEVV_DOMAIN_ASS_SETINDEX )
MTR_DECL_LITERAL( MEVV_DOMAIN_ASS_SETFUNCCALL )
MTR_DECL_LITERAL( MEVV_DOMAIN_ASS_NEW_EXTELEMENT )
MTR_DECL_LITERAL( MEVV_DOMAIN_ASS_SUBSEX )
MTR_DECL_LITERAL( MEVV_DOMAIN_ASS_NEW )
MTR_DECL_LITERAL( MEVV_DOMAIN_ASS_INDETS )
MTR_DECL_LITERAL( MEVV_DOMAIN_ASS_DOM )
MTR_DECL_LITERAL( MEVV_DOMAIN_ASS_LESS )
MTR_DECL_LITERAL( MEVV_DOMAIN_ASS_LEEQUAL )

MTR_DECL_LITERAL( MSTR_args )
MTR_DECL_LITERAL( MSTR_remember )
MTR_DECL_LITERAL( MSTR_noDebug )
MTR_DECL_LITERAL( MSTR_arrow )
MTR_DECL_LITERAL( MSTR_escape )
MTR_DECL_LITERAL( MSTR_hold )
MTR_DECL_LITERAL( MSTR_cyclic )
MTR_DECL_LITERAL( MSTR_Error )
MTR_DECL_LITERAL( MSTR_Warning )
MTR_DECL_LITERAL( MSTR_User )
MTR_DECL_LITERAL( MSTR_Non )
MTR_DECL_LITERAL( MSTR_protecterror )

MTR_DECL_LITERAL( MEEV_base_domains )

extern S_Pointer MEVV_DOMAIN_INDEXARRAY[MEVC_NUMBER_OF_QUICK_VALUES+1] ;
extern S_Pointer* MEEV_domain_array;

/* Steuert die Erzeugung einer unique data representation in der
 * main command loop nach Aufruf von 'share()'. */
extern bool MEEV_execute_share_all;

#ifdef MMMD_GEN_LIT_SP
/* Steuert die Erzeugung von Code fuer literalen S-Pointern in der
 * main command loop nach Aufruf von '_gen_lit()'. */
extern bool MEEV_execute_gen_lit;
#endif // MMMD_GEN_LIT_SP

/**************************************************************************/
/* MuPAD-Preferences (Steuerung von Kern-Einstellungen)                   */
/*                                                                        */
/* Initialisierung der durch Kommandozeilenparameter veraenderbaren Werte */
/* geschieht in MUT_first_init.                                           */
/* S_Pointer werden erst in MUT_init initialisiert (nach der              */
/* Speicheverwaltung !!!)                                                 */
/**************************************************************************/
typedef struct MEVS_pref {
	bool   secure;          /* ATTENTION: no system calls,...      */
	bool   prompt;          /* Ausgabe des Kernel-Prompt: 0/1      */
	bool   echo;            /* Echo der Kernel-Eingabe  : 0/1      */
	char  *promptstrg;      /* Prompt-Zeichenkette als C-String    */
        short  keeporder;       /* interne Ordnung bei Ausgabe bei-    */
                                /* behalten                            */
	short  resprint;        /* Periodische Ausgabe der benutzten   */
	                        /* Resourcen: 0=aus,1=oft..9=selten .  */
	long   maxmem;          /* Abbruch bei berschreiben von Re.   */
	                        /* von xxMB Speicher: 0=kein Test.     */
	long   maxtime;         /* Abbruch bei berschreiben der Be-   */
	                        /* von xxSec Rechenz.: 0=kein Test.    */
	char   floatformat[2];  /* Ausgabe von Floatzahlen: 'g','f'    */
	                        /* oder 'e'.                           */
	S_Pointer callback;     /* is called periodically              */
	S_Pointer evalonexit;   /* is called when MuPAD quits          */
	S_Pointer output;       /* this function will be used for      */
	                        /* output.                             */
	S_Pointer alias;        /* this function will be used for      */
	                        /* substituting output aliasses        */
	S_Pointer postoutput;   /* Nach-Ausgabe z.B. fuer Type-Info    */
	short  typecheck;       /* Typechecking in procedures          */
	S_Pointer postinput;    /* Nach-Eingabe z.B. fuer Type-Info    */
	S_Pointer noprocremtab; /* Einfuegen in procedure remember     */
                                /* table global abschalten             */
        short  moduletrace;     /* Ein Tracemodus fuer Module          */
	S_Pointer useroptions;  /* UserOptionen an den MuPAD Kern.     */
        char   matrixSeparator[5];/* Trennzeichen in Matrizen.         */
	short  trailingZeroes;  /* Nullen in Floats entfernen ?        */
	short  ignoreNoDebug;   /* noDebug-Option ignorieren ?         */
	char   timesDot[10];    /* Multiplikationszeichen \cdot ?      */
	bool   warn15;		/* Parser-Warnungen bzgl. Aenderungen  */
				/* fuer MuPAD 1.5: 0/1		       */
	short  warnLexPE;       /* Parser-Warnung  bzgl. Verwendung    */
				/* lex. Umgebungen: 0/1                */
	short  warnDeadPE;      /* Warnung der Garbage Collection bei  */
				/* toten lex. Umgebungen: 0/1          */
	S_Pointer complete;     /* this function will be used for      */
	                        /* command line completion.            */
        short  interruptLocks;  /* wenn der Wert = 0 ist, dann darf in */
                                /* MEV_interrupt unterbrochen werden.  */
        short  ansi;            /* can we use ANSI in ascii help page? */

} MEVT_pref;

GLOBAL_DECL( MEVT_pref, MEVV_pref) ;

/**************************************************************************/
/* MuPAD-Interrupts (synchron, angestossen durch MEV_eval)                */
/**************************************************************************/

/* Schwellwert fuer Interrupt;  sollte so eingestellt sein, dass 
   MEV_interrupt ca. alle 0.5 Sec aufgerufen wird.  Der aktuelle
   Wert passt in etwa fuer einen 400 MHz Celeron.  Gemessen mit:
   Pref::report(9):  int(sin(x)*exp(x),x) $ i=1..1000: 
                                                                          */
#define MEVC_INTERRUPT_THRESHOLD  200000      

/* Verwaltung synchroner MuPAD-Interrupts: MEVV_interrupt ****/
inline void MEV_disable_interrupt() {MEVV_pref.interruptLocks++;}

inline void MEV_enable_interrupt() {MEVV_pref.interruptLocks--;}

inline bool MEV_interrupt_allowed() {return MEVV_pref.interruptLocks==0;}


GLOBAL_DECL( unsigned long, MEVV_interrupt);

/* Globale user-definierte callback routine: MEVV_callback */

typedef  void (*MEVT_cbfunc)();     /* Typ der CB-Funktion */

GLOBAL_DECL( MEVT_cbfunc, MEVV_callback );


/**************************************************************************/
/* Zeitmessung fuer die vom Parser angestossenen Evaluierungen            */
/**************************************************************************/

GLOBAL_DECL( long, MEVV_calctime );


/**************************************************************************/
/* Weitere Deklarationen                                                  */
/**************************************************************************/

GLOBAL_DECL( S_Pointer, MEVV_ACT_ENV );

/**************************************************************************/
/* Definition von Makros zum Evaluierer.                                  */
/**************************************************************************/

#define MEV_header(s, null_mode, prev_func, eval_type, exec, flat_args, flat_type, args_out, typecheck, dom_mode, dom_start, dom_end, dom_num) \
    if (MEV_header_func(s, null_mode, prev_func, eval_type, exec, flat_args, flat_type, args_out, typecheck, dom_mode, dom_start, dom_end, dom_num) == 1) return(*args_out);


#define MEV_header_1(s, null_mode, prev_func, eval_type, exec, flat_args, flat_type, args_out, typecheck, dom_mode, dom_start, dom_end, dom_num) \
    if (MEV_header_func_1(s, null_mode, prev_func, eval_type, exec, flat_args, flat_type, args_out, typecheck, dom_mode, dom_start, dom_end, dom_num) == 1) return(*args_out);


#define MEV_header_2(s, null_mode, prev_func, eval_type, exec, flat_args, flat_type, typecheck) \
    { S_Pointer args_out; \
      if (MEV_header_func_2(s, null_mode, prev_func, eval_type, flat_args, flat_type, &args_out, typecheck) == 1) return(args_out); }


inline short
MEV_remember_func(S_Pointer exec, S_Pointer args_in, S_Pointer * args_out)
/**************************************************************************/
/* Headerfunktion zur Evaluierung der Argumente mit 'MEV_eval_args' und   */
/* gleichzeitigem Nachschauen in der Remembertafel.                       */
/**************************************************************************/
{
    S_Pointer result;

    /* Lesen aus der Remembertafel */
    if (! MMMequal(*MMMP(&exec, 3), VALUE_READ(MUPV_CAT_NIL)))
    {
        if (MTA_read_table(*MMMP(&exec, 3), args_in, MTAC_FUNC, &result))
        {
            *args_out = result;
            return (1);
        }
    }
    return (0);
}

#define MEV_remember(s,prev_func,eval_type,exec,args_in,args_out) \
    if (MEV_remember_func(exec,args_in, args_out) == 1) \
    { MMMfree(args_in); return(*args_out); }

#define MEV_domain(args_in,prev_func,eval_type,exec,first_ext,dom_mode,dom_start,dom_end,dom_num,args_out)                          \
        if (MEV_domain_func(args_in,prev_func,eval_type,exec,first_ext,dom_mode,dom_start,dom_end,dom_num,args_out) == 1)           \
	   { MMMfree(args_in);                                                                                                      \
	     return(*args_out); }

/*************************************************************************/
/* BOOLEAN MEV_is_CAT_EXT(S_Pointer *s)                                  */
/* Makro, um zu testen, ob es sich um ein CAT_EXT-Element handelt. Das   */
/* Makro liefert TRUE, wenn es sich bei dem Parameter um ein CAT_EXT-    */
/* Element handelt, sonst FALSE.                                         */
/*************************************************************************/
#define MEV_is_CAT_EXT(s) \
                (MMMZ(s, CATEGORY)->cat == CAT_EXT)

/**************************************************************************/
/* Deklaration der Prozeduren des Evaluierers.                            */
/**************************************************************************/

/* Funktionen im Modul MEV_evaluate.c */

S_Pointer	MEV_main_eval		( S_Pointer	s,
					  bool	interactive,
					  bool	print_result,
					  bool	insert_history);

S_Pointer	MEV_execute		( S_Pointer s );

S_Pointer       MEV_eval_any            ( S_Pointer      s,
                                          long           prev_func,
                                          long           eval_type);

MSYT_eval_func*	MEV_get_eval_func	( S_Pointer fe, S_Pointer *ex );
void		MEV_callback            (  );
void		MEV_interrupt	        (  );
S_Pointer	MEV_eval		( S_Pointer	 s,
				          long		 prev_func,
				          long		 eval_type);
S_Pointer	MEV_eval_expr		( S_Pointer	 s,
					  long		 prev_func,
					  long		 eval_type);
S_Pointer       MEV_eval_unknown_function ( S_Pointer s,
                                            long      prev_func,
                                            long      eval_type,
			                    S_Pointer func);
S_Pointer	MEV_eval_args		( S_Pointer	 s,
				          long           null_mode,
					  long		 prev_func,
					  long		 flat_arg,
					  long		 eval_type,
					  long           *first_ext);
S_Pointer	MEV_eval_args_index	( S_Pointer	 s,
					  long		 prev_func,
					  long		 flat_arg,
					  long		 eval_type);
S_Pointer	MEV_eval_index		( S_Pointer	 s,
					  long		 prev_func,
					  long		 eval_type,
					  S_Pointer	 exec );
S_Pointer	MEV_eval_index_val      ( S_Pointer	 s,
					  long		 prev_func,
					  long		 eval_type,
					  S_Pointer	 exec );
S_Pointer 	MEV_index              ( S_Pointer	 s,
					  long		 prev_func,
					  long		 eval_type,
					  S_Pointer	 exec,
					  bool        eval );
S_Pointer	MEV_eval_not		( S_Pointer	 s,
					  long		 prev_func,
					  long		 eval_type,
					  S_Pointer	 exec );
S_Pointer	MEV_eval_ident		( S_Pointer	 s,
					  long		 prev_func,
					  long		 eval_type);
S_Pointer	MEV_eval_var		( S_Pointer	 s,
					  long		 prev_func,
					  long		 eval_type);
S_Pointer	MEV_eval_set		( S_Pointer	 s,
					  long		 prev_func,
					  long		 eval_type);
S_Pointer	MEV_eval_stat_list	( S_Pointer	 s,
					  long		 prev_func,
					  long		 eval_type);
S_Pointer	MEV_eval_dyn_list	( S_Pointer	 s,
					  long		 prev_func,
					  long		 eval_type);
S_Pointer	MEV_eval_proc 		( S_Pointer	 s,
					  long		 prev_func,
					  long		 eval_type);
S_Pointer	MEV_eval_debug		( S_Pointer	 s,
					  long		 prev_func,
					  long		 eval_type);
S_Pointer	MEV_eval_nil		( S_Pointer	 s,
					  long		 prev_func,
					  long		 eval_type);
S_Pointer	MEV_eval_null		( S_Pointer	 s,
					  long		 prev_func,
					  long		 eval_type);
S_Pointer	MEV_eval_failed		( S_Pointer	 s,
					  long		 prev_func,
					  long		 eval_type);
S_Pointer	MEV_eval_ext		( S_Pointer	 s,
					  long		 prev_func,
					  long		 eval_type);
void		MEV_init		(  );
void		MEV_reset		(  );
void		MEV_exit		(  );
S_Pointer	MEV_eval_interpreter	( S_Pointer	 s,
					  long		 prev_func,
					  long		 eval_type,
					  S_Pointer	 proc );
S_Pointer       MEV_options2expr        ( short spec );
S_Pointer	MEV_eval_procdef	( S_Pointer	 s,
					  long		 prev_func,
					  long		 eval_type,
					  S_Pointer      exec);
S_Pointer	MEV_eval_sys_func_args	( S_Pointer	 s,
					  long		 prev_func,
					  long		 eval_type,
					  S_Pointer	 exec );
void     	MEV_check_local_formals ( S_Pointer	 s);
S_Pointer	MEV_eval_act_proc_env	( S_Pointer	 s,
					  long		 prev_func,
					  long		 eval_type,
					  S_Pointer	 exec );
S_Pointer	MEV_eval_eval_entry	( S_Pointer	 s,
					  long		 prev_func,
					  long		 eval_type,
					  S_Pointer	 exec );
long		MEV_eval_bool		( S_Pointer s );
long            MEV_eval_bool1          (S_Pointer s, S_Pointer *result) ;
void		MEV_init_LegalOpTab	(  );

S_Pointer	MEV_eval_gprof		( S_Pointer	 root,
					  long		 prev_func,
					  long		 eval_type,
					  S_Pointer	 exec );
void		MEV_gen_prof_data	( S_Pointer name,
					  S_Pointer args,
					  long depth,
					  long total_time,
					  long args_time );
S_Pointer	MEV_func_env_in		( S_Pointer func, S_Pointer ass );
void		MEV_func_env_insert	( S_Pointer *func,
					  S_Pointer ass,
					  S_Pointer erg );

short MEV_header_func(   S_Pointer         s,
	       		 long              null_mode,
	      		 long              prev_func,
	      		 long              eval_type,
               		 S_Pointer 	   exec,
	       		 long              flat_arg,
			 long              flat_type,
	      		 S_Pointer         *args,
	      		 long              type_check,
			 long              dom_mode,
			 long              dom_start,
			 long		   dom_end,
			 long              dom_num);

short MEV_header_func_1(S_Pointer       s,
	 	        long            null_mode,
	 	        long            prev_func,
	   	        long            eval_type,
          	        S_Pointer 	exec,
	      	     	long            flat_arg,
			long            flat_type,
	           	S_Pointer       *args,
	           	long            type_check,
                  	long            dom_mode,
		   	long            dom_start,
		   	long            dom_end,
                   	long            dom_num);


short MEV_header_func_2(S_Pointer       s,
			long            null_mode,
	 	        long            prev_func,
	   	        long            eval_type,
	      	     	long            flat_arg,
			long            flat_type,
	           	S_Pointer       *args,
	           	long            type_check);


short
MEV_domain_func (  S_Pointer        s,
                   long             prev_func,
                   long             eval_type,
                   S_Pointer        exec,
                   long             first_ext,
                   long             dom_mode,
		   long             dom_start,
		   long             dom_end,
                   long             dom_num,
		   S_Pointer        *args);


long MEV_first_ext(S_Pointer s, long mode);

S_Pointer MEV_float(S_Pointer s);
S_Pointer MEV_bool(S_Pointer s);

/* Funktionen im Modul MEV_operator.c */

S_Pointer MEV_eval_op_plusmult(S_Pointer s, long prev_func, long eval_type, S_Pointer exec) ;
S_Pointer MEV_eval_op_power(S_Pointer s, long prev_func, long eval_type, S_Pointer exec) ;
S_Pointer MEV_eval_op_fconcat(S_Pointer s, long prev_func, long eval_type, S_Pointer exec) ;
S_Pointer MEV_eval_op_fconcatpower(S_Pointer s, long prev_func, long eval_type, S_Pointer exec) ;
S_Pointer
MEV_eval_domainfunccall( S_Pointer s, long prev_func, long eval_type, S_Pointer exec ) ;


/* Funktionen im Modul MEE_utils.c */

S_Pointer MEE_eval_func_env(S_Pointer s, long prev_func, long eval_type, S_Pointer exec) ;
S_Pointer MEE_eval_built_in(S_Pointer s, long prev_func, long eval_type, S_Pointer exec) ;
S_Pointer MEE_copy(S_Pointer s, S_Pointer op, S_Pointer arg1, S_Pointer arg2,
                   S_Pointer eval_arg1, S_Pointer eval_arg2,
                   bool changes, bool two_para) ;
S_Pointer MEE_func_env_slot(S_Pointer env, S_Pointer key, S_Pointer val, long prev_func, long eval_type);
S_Pointer MEE_eval_slot(S_Pointer s, long prev_func, long eval_type, S_Pointer exec) ;
S_Pointer MEE_func_env_ass_callm(S_Pointer ass, S_Pointer fe, S_Pointer seq, long prev_func, long eval_type, long start, long end) ;
S_Pointer MEE_eval_context(S_Pointer s, long prev_func, long eval_type, S_Pointer exec) ;




/* Funktionen im Modul MEE_eval_expr.c */

S_Pointer MEE_eval_exprseq(S_Pointer s, long prev_func, long eval_type, S_Pointer exec);
S_Pointer MEE_eval_or(S_Pointer s, long prev_func, long eval_type, S_Pointer exec);
S_Pointer MEE_eval_and(S_Pointer s, long prev_func, long eval_type, S_Pointer exec);
S_Pointer MEE_eval_seqgen(S_Pointer s, long prev_func, long eval_type, S_Pointer exec);
S_Pointer MEE_eval_less(S_Pointer s, long prev_func, long eval_type, S_Pointer exec);
S_Pointer MEE_eval_leequal(S_Pointer s, long prev_func, long eval_type, S_Pointer exec);
S_Pointer MEE_eval_equal(S_Pointer s, long prev_func, long eval_type, S_Pointer exec);
S_Pointer MEE_eval_unequal(S_Pointer s, long prev_func, long eval_type, S_Pointer exec);
S_Pointer MEE_eval_range(S_Pointer s, long prev_func, long eval_type, S_Pointer exec);
S_Pointer MEE_eval_mod(S_Pointer s, long prev_func, long eval_type, S_Pointer exec);
S_Pointer MEE_eval_modp(S_Pointer s, long prev_func, long eval_type, S_Pointer exec);
S_Pointer MEE_eval_mods(S_Pointer s, long prev_func, long eval_type, S_Pointer exec);
S_Pointer MEE_eval_div(S_Pointer s, long prev_func, long eval_type, S_Pointer exec);
S_Pointer MEE_eval_union(S_Pointer s, long prev_func, long eval_type, S_Pointer exec);
S_Pointer MEE_eval_minus(S_Pointer s, long prev_func, long eval_type, S_Pointer exec);
S_Pointer MEE_eval_intersect(S_Pointer s, long prev_func, long eval_type, S_Pointer exec);
S_Pointer MEE_eval_plus(S_Pointer s, long prev_func, long eval_type, S_Pointer exec);
S_Pointer MEE_eval_mult(S_Pointer s, long prev_func, long eval_type, S_Pointer exec);
S_Pointer MEE_eval_power(S_Pointer s, long prev_func, long eval_type, S_Pointer exec);
S_Pointer MEE_eval_concat(S_Pointer s, long prev_func, long eval_type, S_Pointer exec);
S_Pointer MEE_eval_fconcat(S_Pointer s, long prev_func, long eval_type, S_Pointer exec);

S_Pointer MEE_eval_negate(S_Pointer s, long prev_func, long eval_type, S_Pointer exec);
S_Pointer MEE_eval_invert(S_Pointer s, long prev_func, long eval_type, S_Pointer exec);
S_Pointer MEE_eval_divex(S_Pointer s, long prev_func, long eval_type, S_Pointer exec);
S_Pointer MEE_eval_subtract(S_Pointer s, long prev_func, long eval_type, S_Pointer exec);

S_Pointer MEE_eval_float_mult(S_Pointer s, long prev_func, long eval_type, S_Pointer exec);
S_Pointer MEE_eval_float_plus(S_Pointer s, long prev_func, long eval_type, S_Pointer exec);
S_Pointer MEE_eval_float_power(S_Pointer s, long prev_func, long eval_type, S_Pointer exec);

S_Pointer MEE_eval_lazy_and(S_Pointer s, long prev_func, long eval_type, S_Pointer exec);
S_Pointer MEE_eval_lazy_or(S_Pointer s, long prev_func, long eval_type, S_Pointer exec);
bool      MEE_equal(S_Pointer arg1, S_Pointer arg2);



/* Funktionen im Modul MEE_domain.c */

S_Pointer MEE_get_domain(S_Pointer key, bool *is_new) ;
S_Pointer MEE_read_domain(short cat) ;
void MEE_write_domain(short cat, S_Pointer domain) ;
S_Pointer MEE_eval_new_extelement(S_Pointer s, long prev_func, long eval_type, S_Pointer exec) ;
S_Pointer MEE_eval_new_domain(S_Pointer s, long prev_func, long eval_type, S_Pointer exec) ;
S_Pointer MEE_domain_slot(S_Pointer dom, S_Pointer key, S_Pointer val, long prev_func, long eval_type);
S_Pointer MEE_domain_inquick(S_Pointer dom, long count) ;
S_Pointer MEE_domain_in(S_Pointer dom, S_Pointer ass) ;
void      MEE_domain_initialize_quickreference(S_Pointer result, bool delet) ;
void      MEE_domain_actualize_quickreference(S_Pointer domain, S_Pointer key, 
					      S_Pointer value ) ;
bool   MEE_domain_insert(S_Pointer *dom, S_Pointer ass, S_Pointer value, short force = MIDC_option_standard|MIDC_option_delete) ;
S_Pointer MEE_domain_ass_call1(long option, S_Pointer ass, S_Pointer domelem, long prev_func, long eval_type) ;
S_Pointer MEE_domain_ass_call1quick(long option, long ass, S_Pointer domelem, long prev_func, long eval_type) ;
S_Pointer MEE_domain_ass_callm(long option, S_Pointer ass, S_Pointer seq, long prev_func, long eval_type, long start, long end, long domstart, long domend, long *answer) ;
S_Pointer MEE_domain_ass_callmquick(long option, long ass, S_Pointer seq, long prev_func, long eval_type, long start, long end, long domstart, long domend, long *answer) ;

void MEE_free_unused();


/* Funktionen im Modul MEE_eval_arith.c ***********************/

S_Pointer MEE_mult_const(S_Pointer s, long prev_func);



/* Konstanten und Funktionen im Modul MES_eval_stmt.c ***********************/

#define MESC_VALUE  1
#define MESC_IDENT  2
#define MESC_INDEX  3
#define MESC_FUNC   4

void	  MES_eval		( S_Pointer s );
S_Pointer MES_eval_assign 	( S_Pointer	 s,
				  long		 prev_func,
				  long		 eval_type,
				  S_Pointer	 exec );
S_Pointer MES_eval_delete 	( S_Pointer	 s,
				  long		 prev_func,
				  long		 eval_type,
				  S_Pointer	 exec );
S_Pointer MES_eval_sysdelete 	( S_Pointer	 s,
				  long		 prev_func,
				  long		 eval_type,
				  S_Pointer	 exec );
S_Pointer MES_eval_sysassign    ( S_Pointer      s,
                                  long           prev_func,
                                  long           eval_type,
                                  S_Pointer      exec );
S_Pointer MES_assign_intern     ( S_Pointer      s,
                                  long           prev_func,
                                  long           eval_type,
                                  S_Pointer      exec,
                                  short          force );
S_Pointer MES_delete_intern     ( S_Pointer      s,
                                  long           prev_func,
                                  long           eval_type,
                                  S_Pointer      exec,
                                  short          force );
S_Pointer MES_eval_if 		( S_Pointer	 s,
				  long		 prev_func,
				  long		 eval_type,
				  S_Pointer	 exec );
S_Pointer MES_eval_case 	( S_Pointer	 s,
				  long		 prev_func,
				  long		 eval_type,
				  S_Pointer	 exec );
S_Pointer MES_eval_for 		( S_Pointer	 s,
				  long		 prev_func,
				  long		 eval_type,
				  S_Pointer	 exec );
S_Pointer MES_eval_for_down	( S_Pointer	 s,
				  long		 prev_func,
				  long		 eval_type,
				  S_Pointer	 exec );
S_Pointer MES_eval_for_in	( S_Pointer	 s,
				  long		 prev_func,
				  long		 eval_type,
				  S_Pointer	 exec );
S_Pointer MES_eval_while	( S_Pointer	 s,
				  long		 prev_func,
				  long		 eval_type,
				  S_Pointer	 exec );
S_Pointer MES_eval_repeat	( S_Pointer	 s,
				  long		 prev_func,
				  long		 eval_type,
				  S_Pointer	 exec );
S_Pointer MES_eval_next		( S_Pointer	 s,
				  long		 prev_func,
				  long		 eval_type,
				  S_Pointer	 exec );
S_Pointer MES_eval_break	( S_Pointer	 s,
				  long		 prev_func,
				  long		 eval_type,
				  S_Pointer	 exec );
S_Pointer MES_eval_quit		( S_Pointer	 s,
				  long		 prev_func,
				  long		 eval_type,
				  S_Pointer	 exec );
S_Pointer MES_eval_stmtseq	( S_Pointer	 s,
				  long		 prev_func,
				  long		 eval_type,
				  S_Pointer	 exec );
S_Pointer MES_eval_return	( S_Pointer	 s,
				  long		 prev_func,
				  long		 eval_type,
				  S_Pointer	 exec );
S_Pointer MES_eval_share	( S_Pointer	 s,
				  long		 prev_func,
				  long		 eval_type,
				  S_Pointer	 exec );
S_Pointer MES_eval_gen_lit	( S_Pointer	 s,
				  long		 prev_func,
				  long		 eval_type,
				  S_Pointer	 exec );

S_Pointer MES_new_formal_rem_proc( );
S_Pointer MES_build_name         (S_Pointer	 list );

/* Hilfsfunktionen *******************************************************/

inline S_Pointer
MEV_domtype(S_Pointer s)
{
    S_Pointer dom;
    short cat = MMMZ(&s,CATEGORY)->cat;
    if (cat == CAT_EXT) {
	MMMset(&dom, *MMMP(&s, 0));
    } else {
	dom = MEE_read_domain(cat);
    }
    return dom;
}

inline bool
MEV_level_reached(long depth)
{
    if (depth > MEVC_LEVEL_INF)
	return true;
    if (MEVV_MAXLEVEL != MEVC_LEVEL_INF && depth >= MEVV_MAXLEVEL)
    {
	MEV_set_error(MEVC_RECURSIV_DEFINITION, "See ?MAXLEVEL");
    }
    return (depth >= MEVV_LEVEL);
}

#endif /* __MEV_evaluate___ */
