#ifndef __MKT_lib_kernel_h__
#define __MKT_lib_kernel_h__

#include "MKT_mssg.h"

class MCLT_lib_switch;

/** Clients MuPAD kernel object for link-library kernel.
 *
 * An object of this class represents a client connection to a MuPAD 
 * link-library kernel.
 *
 * For each client process there may exist several objects of this class,
 * connected to exactly one common kernel object.
 *
 * This class is not thread safe.
 */

class MTlibKernel : public MTkernel {
public:
    /** Get a new client connection to the kernel.
     *
     * The kernel is initialized when the first client connection is created.
     *
     * @param mssg_handler messsage handler for the kernel object returned
     * @param kernel returns pointer to kernel object
     * @return true in case of success, false otherwise
     */
    static bool get_kernel(MTclientMessageHandler* mssg_handler,
                           MTlibKernel** kernel);

    /** Destructor, releases connection to kernel.
     * If this is the last client of the kernel the kernel will be quit. */
    virtual ~MTlibKernel();

    /** Get the ID of this client.
     * The ID returned is the ID this client has on the kernels side.
     * @return the client ID, a non-negative number, or -1 if this
     *   object is not connected to the kernel */
    long get_client_id() const;

    // Send a new client message.
    bool send_new_client(const char* name=0,
			 const char* namesp=0,
			 MTlongFormat long_fmt=MCstringLong,
			 MTfloatFormat float_fmt=MCieeeFloat,
			 MTcoding coding=MCflatCoding|MCidentCoding,
			 size_t num_features=0,
			 MTclientFeature features[]=0);

    /// Get message handler of this client.
    MTclientMessageHandler* mssg_handler();

    // Send initialization data.
    bool send_init(size_t argc=0, char* argv[]=0);

    // Send a MuPAD command.
    bool send_cmd(void* cmd_cookie, MTmemoryData* cmd, 
                  MTexecutionFlag flags=0, const char* format=0);

    // Send a completion wanted message.
    bool send_completion_wanted(const char* prefix);

    // Send a plotting command.
    bool send_plot_cmd(void* cmd_cookie, MTmemoryData* cmd);

    // Send debug command.
    bool send_dbg_cmd(unsigned char cmd, const char* fname=0, 
                      long fpos=0, const char* expr=0);

    // Send a response for a MuPAD input() function call.
    bool send_input(bool canceled, MTmemoryData* input=0);

    // Send a response for a MuPAD textinput() function call.
    bool send_textinput(bool canceled, const char* input=0);

    // Send a response for a MuPAD client_call() function call.
    bool send_call_result(bool error, MTmemoryStream* result=0);

    // Send a new textwidth.
    bool send_textwidth(long width);

    // Send a new pretty-print value.
    bool send_pretty_print(bool pretty);

    // Send a format wanted for a output type.
    bool send_format(const char* output_type, const char* format);

    // Send a status request.
    bool send_status_wanted();

    // Send a stop signal.
    bool send_stop();

    // Send a quit message.
    bool send_quit();

private:
    // create kernel
    MTlibKernel(MTclientMessageHandler* mssg_handler);

    long client_id_;           // the id of this client
};

inline long MTlibKernel::get_client_id() const {
    return client_id_; 
}
inline MTclientMessageHandler* MTlibKernel::mssg_handler() { 
    return mssg_handler_; 
}

#endif // !__MKT_lib_kernel_h__
