#ifndef __MPA_parser_h__
#define __MPA_parser_h__
/*
 * MPA_parser.h -- MuPAD Parser and Scanner
 */

#include "llparser.h"

/*
 * Global vars and functions
 */

/* Indizes der Callback-Routinen */
#define MPAC_CB_INPUT_FILTER	0
#define MPAC_CB_SET		1
#define MPAC_CB_LIST		2

#define MPAC_LAST_CB		2  /* Index der letzten Callback-Routine */


#define MSCC_INTERRUPTED 1108
extern short MSCV_ERROR;
extern bool MSCV_cont_mode;

/// name of file actually parsed by 'read'
extern char* MPAV_FILE_NAME;

extern S_Pointer MPAV_CALLBACKS;
extern S_Pointer MPAV_INPUT_STRING;
extern S_Pointer MPAV_parser_config;

MTR_DECL_LITERAL(MPAV_id_arrow)
MTR_DECL_LITERAL(MPAV_id_escape)
MTR_DECL_LITERAL(MPAV_id_seqin)
MTR_DECL_LITERAL(MPAV_id_fnest)
MTR_DECL_LITERAL(MPAV_id_in)
MTR_DECL_LITERAL(MPAV_id_D)
MTR_DECL_LITERAL(MPAV_id_constructor)
MTR_DECL_LITERAL(MPAV_id_procname)
MTR_DECL_LITERAL(MPAV_id_dom)
MTR_DECL_LITERAL(MPAV_id_check_global)

void MPA_init();
void MPA_reset();
void MPA_exit();
void MPA_init_main_parser(bool tty);
void MPA_delete_main_parser();

bool MPA_parse(const char* input, MTexecutionFlag flags, const char* format);
long MPA_completed_chars();
S_Pointer MPA_exec_read(MCLT_iostream &file, char *filename, bool plain, bool do_close);
S_Pointer MPA_parse_input(const char* input);
S_Pointer MPA_parse_finput(struct _MFOT_finfo* finfo);
S_Pointer MPA_parse_ftextinput(MCLT_iostream &file);
S_Pointer MPA_parse_string(S_Pointer string, long show_error_mess);
S_Pointer MPA_parse_in_context(S_Pointer string);

S_Pointer MPA_new_client_context();
void MPA_delete_client_context(S_Pointer ctx);
void MPA_activate_client_context(S_Pointer ctx, bool activate);


/*
 * Parser-Interna
 */

#define MPAC_CHAR_QUEUE_LEN 128

/* Scanner tokens */

typedef int MPAT_token_no;

class MPAT_token {
public:
    MPAT_token();
    MPAT_token(MPAT_token&);
    ~MPAT_token();
    MPAT_token& operator=(MPAT_token&);
    bool is_valid() const { return (no != 0); }
    void clear();

    MPAT_token_no no;       // token number
    int prio;               // priority (for operator tokens)
    S_Pointer val;          // value as read by scanner
    int src_line;           // act line of source file (for debugger)
    int src_col;            // act column of source file
};

/* Abstract parser base class.
 * A parser object holds the actual parser and scanner state. This allows for
 * a recursive call of the parser (via read).
 * The method Getc must be overwritten in a concrete subclass. This method
 * must return returns characters form the input source like the C-stdio
 * function getc.
 * The parser has two modes: interactive and non-interactive. In interactive
 * mode help and system commands may be enterd via ? and ! at the begin of the
 * input.
 */
class MPAT_parser : public LLParser {
public:
    /* Constructor
     * If eval is true statements are evaluated on interactive level.
     * If interactive is true an input may be a help or system command
     *      started with ? or !.
     * src_name is used for syntax error messages, may be nil. The
     *      string is NOT copied!
     * max_errs are the maximal number of errors reported, may be 0.
     */
    MPAT_parser(bool eval=true, bool interactive=false,
                const char* src_name=0, int max_errs=1);
    virtual ~MPAT_parser();

    // Parse input (read via Getc).
    virtual void parse();

    // Get copy of result of parse (may be evaluated). Returns nil in
    // case of an error.
    S_Pointer result();

    // Has an syntax error occured during parsing?
    bool syntax_err();

    // Has an error occured during evaluation?
    bool eval_err();

    // Set name of actual input source (for error messages), the string
    // is NOT copied!
    void set_src_name(const char*name);

    // Get position of actual token
    int src_line() const;
    int src_col() const;

// ------------- Internal methods, called by parser only. ------------------

    // returns true if statements should be evaluated
    virtual bool evaluate();

    // returns true if help queries entered via ? should be evaluated
    virtual bool eval_help();

    // returns true if system commands entered via ! should be evaluated
    virtual bool eval_system();

    // evaluate statement, returns nil in case of an error
    // If colon is false a postinput procedure may be executed.
    // Evaluation takes place only if no syntax error or evaluation error
    // occured before during parse.
    virtual S_Pointer eval_stmt(S_Pointer stmt, bool colon);

    // returns true if parser is in single-step mode
    bool single_step();

    // is called after each interactive stmt, does nothing
    virtual void next_stmt() { }

    // print result, colon is true if the statement was followed by a ':'
    // If an error occured before printing does not take place.
    virtual void print_result(S_Pointer res, bool colon);

    // keep names as idents, do not create variables from them
    bool keep_idents(bool keep);

    // lexical context
    void push_defs(S_Pointer name, S_Pointer args, S_Pointer locals);
    void pop_defs();
    bool make_ident_seq(S_Pointer* s);

    // are debug infos to be created?
    bool mk_dbg(bool);
    bool mk_dbg() const;

    // look ahead 1 token
    int look_ahead();

    // ignore %if statements?
    void pre_if_ignore(bool);
    bool pre_if_ignore() const;

    virtual int lexer();
    virtual void message(int);
    virtual void set_err(int err);
    virtual void err_mssg(int err, ...);
    void warn_env_var(S_Pointer var);

    // create a literal from actual token
    S_Pointer make_var(long& lex_dist);
    S_Pointer make_ident();
    S_Pointer make_string();
    S_Pointer make_key();
    S_Pointer make_number(bool after_dot=false);
    S_Pointer make_op();

#ifdef MDX
    // adding debbugging information
    virtual bool debug();
    unsigned long make_dbx_mark();
    unsigned long make_dbx_info();
#endif

    int prio();
    void check_option(bool& no_debug);
    void set_result(S_Pointer r);

protected:
    // overwrite to read chars from input (like C-stdio function getc)
    virtual int Getc() = 0;

    // is called from parse(), allows to set up initial variable definitions
    // before parsing begins
    virtual void init_definitions();

    // is the result of the evaluation to be inserted into the history?
    virtual bool insert_history() const;

    // get the result output format (0 if actual format should be used)
    virtual const char* result_format() const;

    bool do_eval_;              // evaluate statements on interactive level?
    bool interactive_;          // parsing in interactive mode?
    bool single_step_;          // stop parsing after each stmt
				// (single-step mode)?
    const char* src_name_;      // name of input source for messages
    int max_errs_;              // max. number of errors shown to user
    S_Pointer result_;          // result of call to parse() or nil

    MPAT_token token_;          // actual token read
    MPAT_token saved_token_;    // saved token for lookahead
    bool input_start_;          // true at start of input until first
				// no-white char
    bool saw_slot_;             // scanned a :: operator just before?
    bool saw_dot_;              // scanned a '.[0-9]' just before?
    int src_line_;              // actual line of input
    int src_col_;               // actual column of input
    int err_cnt_;               // number of syntax errors encountered
    bool eval_err_;             // an error occured during evaluation
    bool unexpected_end_;	// did input end unexpectedly?
    bool pretty_print_;         // state of PRETTY_PRINT
    unsigned long dbx_mark_;    // last mdx mark created
    bool keep_idents_;		// parse each name as ident, do not create
				// variables from them

    S_Pointer var_defs_;	// stack of local vars defined
    S_Pointer arg_defs_;	// stack of formal args defined
    S_Pointer name_defs_;	// stack of procedure names parsed
    long impl_var_dist_;        // lex. distance of implicitly defined vars
    bool mk_dbg_;		// true if parser infos are to be created
    int pre_if_ignore_;         // level of %if-parts to ignore

    // character FIFO for scanning
    void push_back(int c);
    int char_queue_[MPAC_CHAR_QUEUE_LEN];
    int char_front_;
    int char_rear_;

    int getc_src();
    int get_char(bool ignore_backsl=true);
    int eatwhite();
    void getline(MPAT_token_no no);
    void getsymbol(int c);
    void getnumber(int c);
    void getstring(int c);
    bool getuserop(int c);
    void gettoken(bool handle_pre = true);
    void make_token(MPAT_token_no n, int p=0, S_Pointer v=MMMNULL);
    void make_keyword();
    S_Pointer var2ident(S_Pointer);
    void set_unexpected_end();
};

inline void MPAT_parser::init_definitions()
{
    // nix
}

inline bool MPAT_parser::keep_idents(bool keep)
{
    bool res = keep_idents_;
    keep_idents_ = keep;
    return res;
}

inline int MPAT_parser::prio()
{
    return token_.prio;
}

inline void MPAT_parser::push_back(int c)
{
    char_queue_[char_rear_] = c;
    char_rear_ = (char_rear_+1) % MPAC_CHAR_QUEUE_LEN;
}

#ifdef MDX
inline bool MPAT_parser::debug()
{
    return MUPV_env.isDebugOn() && ! interactive_ && src_name_ != NULL;
}
#endif

inline bool MPAT_parser::single_step()
{
    return single_step_;
}

inline void MPAT_parser::set_unexpected_end()
{
    if (err_cnt_ == 0 && ! eval_err_)
	unexpected_end_ = true;
}

inline bool MPAT_parser::insert_history() const
{
    return false;
}

inline const char* MPAT_parser::result_format() const
{
    return 0;
}

inline bool MPAT_parser::mk_dbg(bool mdb)
{
    bool res = mk_dbg_;
    mk_dbg_ = mdb;
    return res;
}

inline bool MPAT_parser::mk_dbg() const
{
    return mk_dbg_;
}

inline void MPAT_parser::pre_if_ignore(bool ign)
{
    if (ign) {
	pre_if_ignore_++;
    } else {
	MMMASSERT(pre_if_ignore_ > 0);
	pre_if_ignore_--;
    }
}

inline bool MPAT_parser::pre_if_ignore() const
{
    return pre_if_ignore_ != 0;
}

/* A concrete parser class for streams.
 */
class MPAT_stream_parser : public MPAT_parser {
public:
    MPAT_stream_parser(MCLT_iostream& str, const char* sname, bool plain,
		       bool eval=true);
    void parse_stmt();
protected:
    virtual bool insert_history() const;
    virtual int Getc();
    MCLT_iostream* strp_;
    bool plain_;
};

/* Parser-Konfiguration */

/* Bounds of priorities for user-defined operators */
#define MPAC_MIN_PRIO 1
#define MPAC_MAX_PRIO 2000

/* max. Laenge von Benutzer-Operatoren */
#define MPAC_MAX_OP_LEN 32

/* Operator-Typen */
#define MPAC_PREFIX_OP 1
#define MPAC_POSTFIX_OP 2
#define MPAC_BINARY_OP 4
#define MPAC_NARY_OP 12

MTR_DECL_LITERAL(MPAV_id_Postfix)
MTR_DECL_LITERAL(MPAV_id_Prefix)
MTR_DECL_LITERAL(MPAV_id_Binary)
MTR_DECL_LITERAL(MPAV_id_Nary)

S_Pointer MPA_ident_alias(S_Pointer id);
S_Pointer MPA_call_alias(S_Pointer id, S_Pointer args);
bool MPA_is_call_alias(S_Pointer id);
S_Pointer MPA_vector2list(S_Pointer vec);
const char* MPA_operator_name_at(int i);
void MPA_operator_at(int i, long* typ, long* prio, S_Pointer* val);

#endif // !__MPA_parser_h__
