/* ---------------------------------------------------------------------- */
/* FILE   : MUP_clib.h                                                    */
/* AUTHOR : SciFace Software, Copyright (c) 1999-2000                     */
/* DESCRIP: Interface to access the MuPAD kernel from C/C++ programs      */
/* ---------------------------------------------------------------------- */

#ifndef __MUP_clib__
#define __MUP_clib__

#ifdef __cplusplus
#  define EXTERNC  extern "C"
#else
#  define EXTERNC  
#endif 

/* ---------------------------------------------------------------------- */

EXTERNC int    mupadInit        ( char *libpath ) ;
EXTERNC int    mupadInitArgv    ( int argc, char *argv[] ) ;
 
EXTERNC int    mupadReset       ( ) ;
EXTERNC void   mupadExit        ( ) ;

EXTERNC char*  mupadEvalString  ( char *string, long *errnum ) ;
EXTERNC void   mupadFreeString  ( char *string ) ;

EXTERNC void   mupadEvalFile    ( char *string, long *errnum, FILE *file, 
                                 int   pretty ) ;

EXTERNC void   mupadGarbageCollection ( );

EXTERNC void   mupadSetCallback ( void (*func)() ) ;

/* ---------------------------------------------------------------------- */

#ifdef MALL                                /* use the full link interface */
#  ifndef MAPI 
#    define MAPI                           /* for this MAPI is needed too */
#  endif
#  define MDMD_USE_DYNLINK                 /* use dynamic link context    */
#endif

/* ---------------------------------------------------------------------- */

#ifdef MRUN
#  ifndef MAPI 
#    define MAPI                           /* for this MAPI is needed too */
#  endif
#  define MDM_K(o) mupadKernelEntry[o]     /* use this kernel entry point */

#  define ISMODULE                         /* modules context is needed   */
#  include "MDM_base.h"

extern void       *mupadKernelHandle;      /* defined in the kernel stub  */
extern MDM_jump_t *mupadKernelEntry ;      /* defined in the kernel stub  */

EXTERNC int        mupadLink        ( const char *kpath  = NULL ) ;
EXTERNC int        mupadUnlink      ( const int exitHandler = 0 ) ;
EXTERNC void       mupadExitHandler ( ) ;

#define MUPAD_SAVE_EXIT  1                 /* to be used with mupadUnlink */

#else
#  define MUPD_USE_STALINK                 /* static link library kernel  */
#endif

/* ---------------------------------------------------------------------- */

#ifdef MAPI                                /* use modules MAPI interface  */
#  ifndef MAPI 
#    define ISMODULE                       /* modules context is needed   */
#  endif
#  define SUPPORT_TRYCATCH                 /* use save evaluation context */

#  include "MKT_mapi.h"                    /* include MAPI definitions    */
#  include "MTBfile.h"                     /* Toolbox routines: file      */
#  include "MTBproc.h"                     /* Toolbox routines: process   */
#endif /* MAPI */

/* ---------------------------------------------------------------------- */

#ifdef SUPPORT_TRYCATCH

// ************************************************************************
/** This statement throws a MuPAD error, meaning it jumps ({\tt longjmp})
    to the next upper #MUPAD_TRY# respectively #MUPAD_CATCH_ERR# context.
    The global MuPAD error status variable #MUPAD_ERROR# and the global 
    evaluation status variable #MUPAD_STATUS# may be assigned a value 
    before the error is thrown. 

    {\bf Example:}

    \begin{verbatim}
MUPAD_ERROR = MUPAD_ERROR_USER;
MUPAD_THROW_ERR;
    \end{verbatim}
   
    @memo  Throws a MuPAD error 
    @args  
*/ //**********************************************************************
 
#define MUPAD_THROW_ERR \
   longjmp(MEVV_error_label, 1)


// ************************************************************************
/** This statement marks the begin of a save evaluation context. It must
    directly be followed by a #MUPAD_CATCH_ERR# statement which is 
    executed in case of an error. 

    {\bf Example:}

    \begin{verbatim}
MUPAD_TRY
{
    MF( "/this_leads_to_an_error" ); 
}
MUPAD_CATCH_ERR
{
    fprintf( stderr, "Evaluation failed [%ld].\n", MUPAD_ERROR );
}
MUPAD_END_CATCH
    \end{verbatim}

    @memo  Defines a save MuPAD evaluation context
    @args  { some code using MAPI }
*/ //**********************************************************************

#define MUPAD_TRY \
   {                                                                 \
   jmp_buf MEVV_old_error_label;                                     \
   memcpy (MEVV_old_error_label, MEVV_error_label, sizeof(jmp_buf)); \
   if (setjmp(MEVV_error_label) == 0)


// ************************************************************************
/** This statement marks the begin of a block that is executed if the
    previous #MUPAD_TRY# context has thrown an error. This statement
    must be located directly after a #MUPAD_TRY# block and must directly
    be followed by a #MUPAD_END_CATCH# statement. 

    {\bf Example:}

    \begin{verbatim}
MUPAD_TRY
{
    MF( "/this_leads_to_an_error" );   
} 
MUPAD_CATCH_ERR
{
    fprintf( stderr, "Evaluation failed [%ld].\n", MUPAD_ERROR );
}
MUPAD_END_CATCH
    \end{verbatim}

    @memo  Catches a MuPAD evaluation error
    @args  { some error handling code }
*/ //**********************************************************************

#define MUPAD_CATCH_ERR \
   else


// ************************************************************************
/** This statement marks the end of a #MUPAD_CATCH_ERR# block and is
    absolutely required after each occurancy of a #MUPAD_CATCH_ERR#
    statement. 

    {\bf Example:}

    \begin{verbatim}
MUPAD_TRY
{
    MF( "/this_leads_to_an_error" );   
} 
MUPAD_CATCH_ERR
{
    fprintf( stderr, "Evaluation failed [%ld].\n", MUPAD_ERROR );
}
MUPAD_END_CATCH
    \end{verbatim}

    @memo  Marks the end of a #MUPAD_CATCH_ERR# block
    @args  
*/ //**********************************************************************

#define MUPAD_END_CATCH \
   memcpy(MEVV_error_label, MEVV_old_error_label, sizeof(jmp_buf));  \
   }


// ************************************************************************
/** This variable can be used to read out (and to define) the global
    error status of the MuPAD kernel. It can either be used in a 
    #MUPAD_CATCH_ERR# block to identify the last error that was thrown
    or directly before a #MUPAD_THROW_ERR# statement is executed to
    set a current error code. 

    The default value of the variable #MUPAD_ERROR# is #MUPAD_ERROR_NO# 
    (zero).  The user may set it to the value #MUPAD_ERROR_USER#. 

    {\bf Example:}

    \begin{verbatim}
MUPAD_TRY
{
    MUPAD_ERROR = MUPAD_ERROR_USER;
    MUPAD_THROW_ERR;
} 
MUPAD_CATCH_ERR
{
    fprintf( stderr, "Error status: %ld.\n", MUPAD_ERROR  );
}
MUPAD_END_CATCH
    \end{verbatim}

    @memo  Global MuPAD error status variable
    @args 
*/ //**********************************************************************

#define MUPAD_ERROR MEVV_ERROR

#define MUPAD_ERROR_NO    0
#define MUPAD_ERROR_USER  MEVC_USER_DEFINED_ERROR


// ************************************************************************
/** This variable can be used to read out (and to define) the global 
    evaluation status of the MuPAD kernel. It can either be used in
    a #MUPAD_CATCH_ERR# block to identify a quit signal set during 
    evaluation or to set this signal for example in an application
    callback routine directly before a #MUPAD_THROW_ERR# statement
    is executed. 

    The default value of #MUPAD_STATUS# is #MUPAD_STATUS_CONTINUE#.
    The user may set it to #MUPAD_STATUS_QUIT# before using the
    statement #MUPAD_THROW_ERR# to indicate that the evaluation 
    is to be terminated. 

    {\bf Example:}

    \begin{verbatim} 
MUPAD_TRY    
{    
    MUPAD_ERROR  = MUPAD_ERROR_USER;
    MUPAD_STATUS = MUPAD_STATUS_QUIT;
    MUPAD_THROW_ERR;
}     
MUPAD_CATCH_ERR    
{    
    fprintf( stderr, "Error      status: %ld.\n", MUPAD_ERROR  );
    fprintf( stderr, "Evaluation status: %ld.\n", MUPAD_STATUS );
}
MUPAD_END_CATCH
    \end{verbatim}

    @memo  Global MuPAD evaluation status variable
    @args
*/ //**********************************************************************

#define MUPAD_STATUS MEVV_STATUS

#define MUPAD_STATUS_CONTINUE MEVC_CONTINUED
#define MUPAD_STATUS_QUIT     MEVC_QUIT

#endif /* SUPPORT_TRYCATCH */

/* ---------------------------------------------------------------------- */

/* ---------------------------------------------------------------------- */
/* ---------------------------------------------------------------------- */
#if (defined WIN32) && (defined MULIB)
// The following ifdef block is the standard way of creating macros which make exporting 
// from a DLL simpler. All files within this DLL are compiled with the MUPADD_EXPORTS
// symbol defined on the command line. this symbol should not be defined on any project
// that uses this DLL. This way any other project whose source files include this file see 
// MUPADD_API functions as being imported from a DLL, wheras this DLL sees symbols
// defined with this macro as being exported.
#ifdef MUPADD_EXPORTS
#define MUPADD_API __declspec(dllexport)
#else
#define MUPADD_API __declspec(dllimport)
#endif

// This class is exported from the mupadd.dll
class MUPADD_API CMupadd {
public:
	CMupadd(void);
	// TODO: add your methods here.
};

extern MUPADD_API int nMupadd;

MUPADD_API int fnMupadd(void);
#endif /* WIN32 */
/* ---------------------------------------------------------------------- */
/* ---------------------------------------------------------------------- */


#endif /* __MUP_clib__ */

