/* $Date: 2001/01/29 13:47:45 $  $Author: tonner $  $Revision: 1.12 $ */

/**************************************************************************/
/* FILE   : MUP_env.h                                                     */
/* AUTHOR : Andreas Sorgatz                                               */
/* CREATED: 31.08.1999                                                    */
/* CHANGED: 13.10.2000                                                    */
/* DESCRIP: This file declares the MuPAD Kernel Environment class.        */
/**************************************************************************/

#ifndef __MUP_env__
#define __MUP_env__


#include "MUP_constants.h"


#if (!defined MSDOS)
class MCLT_classic_switch;
#endif


extern int  MUT_set_alarm(long s,void (*h)());  // needed for the watchdog
extern void MUT_reset_alarm();                  // needed for the watchdog
extern void MUP_catch_system_error();           // needed for global use
void        MUP_env_catchWatchdog();            // Global watchdog handler


/**@name MuPAD Kernel Environment
    The MuPAD kernel environment specifies and controlles the behaviour of 
    the kernel.  This includes features like suppressing a banner message,
    specifying the library path, etc. The environment is first set to some
    default values.  It can be changed via command line options as well as
    by directly accessing method and variables of the MUPT\_env class.
*/
//@{

class MCLT_bytestream;                          // external declaration

/// MuPAD kernel environment 
class MUPT_env {  

public:
  /// Sets default values for all options of the MuPAD kernel environment
  void setDefaults ( int argc, char* argv[] );

  /// Checks is MuPAD is registered correctly
  bool isRegistered ( );

  /// Parses command line options to initialise the kernel environment
  int  setCommandLineOptions ( int argc, char* argv[] );

  /// Displays information about the usage of command line options
  void printUsage ();

  /// Prints out the current values of the MuPAD kernel environment.
#if (defined MUP_ENV_DEBUG) || (defined DEBUG_ON)
  void debugDisplay ( );
#endif

  /// Sets/Gets the type of communication switch that is used
  inline void setTermSwitch()  {        terminalMode =  'T' ; };
  inline int   isTermSwitch()  { return(terminalMode == 'T'); };
  inline void setMotifSwitch() {        terminalMode =  'M' ; };
  inline int   isMotifSwitch() { return(terminalMode == 'M'); };
  inline void setXviewSwitch() {        terminalMode =  'X' ; };
  inline int   isXviewSwitch() { return(terminalMode == 'X'); };

#if (!defined MSDOS)
  /// This variables stores the handle of the communication switch.
  /// The classic (old) communication is only used on Macintosh and
  /// UNIX systems.
  MCLT_classic_switch *theSwitch;
#endif

  /// Sets/Gets the debug modes. Note that on Macintosh and Windows
  /// systems a graphical debugger front end is used by default and
  /// thus debugging takes place always in connect mode.
#if (defined MACINTOSH) || (defined MSDOS)
  inline void setDebugOn ()            { setDebugConnect();       };
  inline void setDebugMode (long dm)   { if(dm) debugMode = dm|1; 
                                         else   debugMode = dm  ; };
#else
  inline void setDebugOn()             {        debugMode |=  1 ; };
  inline void setDebugMode (long dm)   {        debugMode  = dm ; };
#endif
  inline void setDebugOff()            {        debugMode  =  0 ; };
  inline void setDebugVerbose()        {        debugMode |=  4 ; };
  inline void setDebugConnect()        {        debugMode |=  9 ; };
  inline void setDebugInteractive()    {        debugMode |= 17 ; };
  inline void setDebugNotInteractive() {        debugMode &= ~16; };
  inline void setDebugQuitRequest()    { 
    dbgQuitRequested_ = true;
    setUserInterrupt();
  };
  inline void resetDebugQuitRequest()    { 
    dbgQuitRequested_ = false;
  };
  inline bool getDebugQuitRequest()    { return(dbgQuitRequested_);};
  inline int  getDebugMode()           { return(debugMode      ); };
  inline int   isDebugOn()             { return(debugMode &   1); };
  inline int   isDebugVerbose()        { return(debugMode &   4); };
  inline int   isDebugConnect()        { return(debugMode &   8); };
  inline int   isDebugInteractive()    { return(debugMode &  16); };

  /// PARI Stacksize in KB
  long pariStackSize;

  /// PARI prime number limit
  long primeNumberLimit;

  /// These streams define the default output and error devices
  MCLT_bytestream *stdOutput;
  MCLT_bytestream *errOutput;

  /// These time stamps are set when the kernel is started up.
  /// Varable sessionBeginSec is used for real time meassuring.
  long sessionBegin;
  long sessionBeginSec;
  
  /// This flag specifies if a banner is printed or not.
  bool printBanner;

  /// Sets/Gets the watchdog modes and installs a new watchdog
  inline void setWatchdogDoTerminate()  { watchdogAction  = 'T' ; };
  inline void setWatchdogThrowError()   { watchdogAction  = 'E' ; };
  inline void setWatchdogMode(char wdm) { watchdogAction  = wdm ; };
  inline char getWatchdogMode()         { return(watchdogAction); };
  void        resetWatchdog()           { MUT_reset_alarm()     ; };
  inline int  installWatchdog(int sec)  
          { return( MUT_set_alarm(sec, &MUP_env_catchWatchdog) ); };

  /// Controls user interrupt in critical kernel sections
  bool setUserInterrupt       ( bool interrupt=true );
  bool setBlockUserInterrupts ( bool block=true );
  bool isUserInterrupt        ( );

  /// Tests if the kernel is in a critical section of initialization
  bool isInitializing( );

  /// Variables to control critical sections of kernel initialization.
  bool inInit;
  bool inPreInit;

  /// Is set to 'true' when the memory management has been set up
  bool isMemUp;

  /// Default new Handler of the kernel
  new_handler origNewHandler;
  
  /// This flag specifies whether the userinit file is read or not. It
  /// is needed when the kernel is launched or reset.  The default value
  /// should be true.
  bool readUserInit;

  /// Process-id of the MuPAD kernel
  int parentPid;

  /// Pathname of the MuPAD kernel binary file
  char *progName;

  /// Pathname of the MuPAD installation directory
  char rootPath [MUPC_MAX_PATH];

  /// Pathname of the MuPAD binary directory
  char binPath [MUPC_MAX_PATH];

  /// Pathname of the MuPAD dynamic modules directory
  char modPath [MUPC_MAX_PATH];

  /// Pathname of the MuPAD library and system init file directory
  char libPath [MUPC_MAX_PATH];

  /// Pathname of the MuPAD user init file and preferences/resources
  /// On Macintosh systems it is not used.  On Windows systems it is
  /// used only if it is defined via a command line option.
  char userPath [MUPC_MAX_PATH];

  /// Pathname of the MuPAD help pages and help index file
  char helpPath      [MUPC_MAX_PATH];
  char helpIndexPath [MUPC_MAX_PATH];

  /// Pathname which can be used to temporary files
  char tempPath      [MUPC_MAX_PATH];
  
  /// Output file when batch mode is active. This variable either 
  /// contains a file name,  the value - for the default output
  /// device or NULL  when the output file could not be opened or
  /// batch mode is inactive.
  char *batchOutputFile;

  /// Stores a user option which may be specified via the command line
  char *userOption;

  /// Methods to access Pref options stored in MUPT_env.
  inline void setSecureKernel(bool value) { SecureKernel = value; };
  inline bool getSecureKernel()           { return(SecureKernel); };
  inline bool getPrintEcho()              { return(PrintEcho   ); };
  inline bool getPrintPrompt()            { return(PrintPrompt ); };
  inline bool getPrintWarn15()            { return(PrintWarn15 ); };

private:
  /// This flag specifies the terminal mode that is used to launch the
  /// kernel. Currently supported modes are (I)nteractive, (M)otif and
  /// (X)view.
  char terminalMode;

  /// This flag specifies the debug modes that are used by the kernel.
  /// Currently the following bits are supported: 1=active, 4=verbose, 
  /// 8=connected, 16=interactive. Bit 4,8,16 requires also to set bit
  /// 1. If no bit is set, debugging in inactive.
  long debugMode;


  /// This flag controls the action of isUserInterrupt.  If true 
  /// and the debugger is active then we want to quit the debugger.
  /// If the debugger is active and this flag is set to false
  /// we switch into step mode.
  bool dbgQuitRequested_;


  /// Stores the action mode of the watchdog timer. Supported modes
  /// are: (E)xit program and (C)atch error code.
  volatile char watchdogAction;

  /// Variables to control user interrupt in critical kernel sections.
  bool UserInt;
  bool BlockUserInt;

  /// Variables that store command line flags which influence options
  /// that can be set with the Pref command. This is needed for the
  /// MuPAD reset command.
  bool SecureKernel;
  bool PrintEcho;
  bool PrintPrompt;
  bool PrintWarn15;

};
//@}

//@Include: MUP_env.cpp

#endif /* __MUP_env__ */

