/*****************************************************************************/
/* pari-arithmetic.h                                                         */
/* PARI arithmetic with dynamic memory allocation and directed roundings     */
/* including 32 bit and 64 bit word support HEADER version 1.2 BETA          */
/* Copyright (C) 1997 by Manfred Radimersky, Bad Pyrmont, Germany            */
/* Email: maradim@uni-paderborn.de                                           */
/*****************************************************************************/
/* NOTE: the integers used here are PARI integers with one exception:        */
/* sign field is 0x80 for 32 bit or 0x8000 for 64 bit: NotANumber            */
/*****************************************************************************/
/* NOTE: the reals used here are PARI reals with the following exceptions:   */
/* sign field is 0x80 for 32 bit or 0x8000 for 64 bit: NotANumber            */
/* sign field is 0xFE for 32 bit or 0xFFFE for 64 bit: NegativeInfinity      */
/* sign field is 0x02 for 32 bit or 0x0002 for 64 bit: PositiveInfinity      */
/*****************************************************************************/
/* IMPORTANT: The original PARI library from the University of Bordeaux in   */
/* France will not work with infinity or NAN values defined here. If you     */
/* want to use this library and the original PARI library simultaneously,    */
/* you will have to make sure yourself that the original PARI routines are   */
/* not called with arguments containing an infinity or NAN value.            */
/*****************************************************************************/
/* MEMORY ORGANIZATION: For efficiency reasons, the following routines make  */
/* use of the fact that the given system is little endian or big endian.     */
/* Surely all modern RISC processor architectures are big endian, but the    */
/* Intel 80386 compatible CISC processors, for instance, are little endian.  */
/* For little endian architectures, you have to define PARI_LITTLE_ENDIAN,   */
/* for big endian architectures, you have to define PARI_BIG_ENDIAN.         */
/*****************************************************************************/
/* MEMORY ALLOCATION: The memory allocation concept used here is entirely    */
/* different from the original PARI library. This concept tries to handle    */
/* dynamic variables very much like static variables and minimizes the count */
/* of heap operations this way. Read the user's manual for more information. */
/*****************************************************************************/

#ifndef PARI_ARITHMETIC_H
#define PARI_ARITHMETIC_H

#if defined(__cplusplus) && !defined(PARI_CPLUSPLUS_COMPILED)
extern "C" {
#endif

/* make sure the target system is marked either little endian or big endian */
#if (!defined(PARI_LITTLE_ENDIAN) && !defined(PARI_BIG_ENDIAN)) \
  || (defined(PARI_LITTLE_ENDIAN) && defined(PARI_BIG_ENDIAN))
  #error THE TARGET SYSTEM MUST BE MARKED EITHER LITTLE ENDIAN OR BIG ENDIAN!
#endif

/* include pari object memory allocation header */
#include "pari-malloc.h"

/* include standard headers */
#include <limits.h>

/* choose word types (#ifdef) */
#ifdef PARI_WORD_LONG_LONG

/* define logical types for pari words */
typedef unsigned long long uword;
typedef signed long long sword;
typedef unsigned long uhalf;
typedef signed long shalf;
typedef unsigned short usign;
typedef signed short ssign;

/* bits in word */
#define WORD_BITS 64
/* binary logarithm of bits in word */
#define LB_WORD_BITS 6
/* bytes in word */
#define WORD_BYTES 8
/* binary logarithm of bytes in word */
#define LB_WORD_BYTES 3

/* signed word minimum */
#define SWORD_MIN -0x8000000000000000ull
/* signed word maximum */
#define SWORD_MAX  0x7FFFFFFFFFFFFFFFull
/* unsigned word most significant bit */
#define UWORD_MSB  0x8000000000000000ull
/* unsigned word maximum */
#define UWORD_MAX  0xFFFFFFFFFFFFFFFFull

/* pari type code none */
#define PARI_TYPE_NONE 0x00000001ull
/* pari type code integer */
#define PARI_TYPE_INT  0x00010001ull
/* pari type code integer */
#define PARI_TYPE_REAL 0x00020001ull

/* maximum pari object length */
#define PARI_LEN_MAX 0x00000000FFFFFFFFll

/* pari object sign shift */
#define PARI_SIGN_SHIFT 48
/* pari sign code negative infinity */
#define PARI_SIGN_NINF 0xFFFEull
/* pari sign code negative */
#define PARI_SIGN_NEG  0xFFFFull
/* pari sign code zero */
#define PARI_SIGN_ZERO 0x0000ull
/* pari sign code positive */
#define PARI_SIGN_POS  0x0001ull
/* pari sign code positive infinity */
#define PARI_SIGN_PINF 0x0002ull
/* pari sign code not a number */
#define PARI_SIGN_NAN  0x8000ull

/* pari exponent bias */
#define PARI_EXP_BIAS 0x0000800000000000ll
/* pari biased exponent minimum */
#define PARI_EXP_MIN  0x0000000000000000ll
/* pari biased exponent maximum */
#define PARI_EXP_MAX  0x0000FFFFFFFFFFFFll

/* decimal logarithm of maximum sword value rounded down */
#define PARI_LOG10_SWORD_MAX 18
/* decimal digits needed for one data word */
#define PARI_UWORD_DEC_DIG 20
/* smallest decimal exponent for 1 <= mantissa < 10 */
#define PARI_DEC_EXP_MIN -42366205509364ll
/* greatest decimal exponent for 1 <= mantissa < 10 */
#define PARI_DEC_EXP_MAX  42366205509363ll
/* greatest decimal exponent string length */
#define PARI_DEC_EXP_LEN 15

/* log10(2) rounded down encoded in two uwords */
#define LOG10_2_MUL  3010299956639811952ull
#define LOG10_2_DIV 10000000000000000000ull

#ifdef PARI_LITTLE_ENDIAN

/* pari object length */
#define parilen(arg)    ((unsigned long*)      arg)[0]
/* pari object type and reference count */
#define parityp(arg)    ((unsigned long*)      arg)[1]
/* pari object sign */
#define parisgn(arg)    ((unsigned short*)     arg)[7]
/* pari integer effective length */
#define parieff(arg)    ((unsigned long*)      arg)[2]
/* pari real sign and biased exponent */
#define parisgnexp(arg) ((unsigned long long*) arg)[1]
/* pari object mantissa most significant word */
#define parimanmsw(arg) ((unsigned long long*) arg)[2]
/* pari real get biased exponent */
#define getpariexp(arg) (((long long*)         arg)[1] & 0x0000FFFFFFFFFFFFll)

#else  /* #ifdef PARI_LITTLE_ENDIAN */

/* pari object length */
#define parilen(arg)    ((unsigned long*)      arg)[1]
/* pari object type and reference count */
#define parityp(arg)    ((unsigned long*)      arg)[0]
/* pari object sign */
#define parisgn(arg)    ((unsigned short*)     arg)[4]
/* pari integer effective length */
#define parieff(arg)    ((unsigned long*)      arg)[3]
/* pari real sign and biased exponent */
#define parisgnexp(arg) ((unsigned long long*) arg)[1]
/* pari object mantissa most significant word */
#define parimanmsw(arg) ((unsigned long long*) arg)[2]
/* pari real get biased exponent */
#define getpariexp(arg) (((long long*)         arg)[1] & 0x0000FFFFFFFFFFFFll)

#endif /* #ifdef PARI_LITTLE_ENDIAN */

/* choose word types (#else) */
#else  /* #ifdef PARI_WORD_LONG_LONG */

/* long has 32 bits */
#if ULONG_MAX == 0xFFFFFFFFul

/* define logical types for pari words */
typedef unsigned long uword;
typedef signed long sword;
typedef unsigned short uhalf;
typedef signed short shalf;
typedef unsigned char usign;
typedef signed char ssign;

/* bits in word */
#define WORD_BITS 32
/* binary logarithm of bits in word */
#define LB_WORD_BITS 5
/* bytes in word */
#define WORD_BYTES 4
/* binary logarithm of bytes in word */
#define LB_WORD_BYTES 2

/* signed word minimum */
#define SWORD_MIN -0x80000000ul
/* signed word maximum */
#define SWORD_MAX  0x7FFFFFFFul
/* unsigned word most significant bit */
#define UWORD_MSB  0x80000000ul
/* unsigned word maximum */
#define UWORD_MAX  0xFFFFFFFFul

/* pari type code none */
#define PARI_TYPE_NONE 0x0001
/* pari type code integer */
#define PARI_TYPE_INT  0x0101
/* pari type code integer */
#define PARI_TYPE_REAL 0x0201

/* maximum pari object length */
#define PARI_LEN_MAX 0x0000FFFFl

/* pari object sign shift */
#define PARI_SIGN_SHIFT 24
/* pari sign code negative infinity */
#define PARI_SIGN_NINF 0xFEul
/* pari sign code negative */
#define PARI_SIGN_NEG  0xFFul
/* pari sign code zero */
#define PARI_SIGN_ZERO 0x00ul
/* pari sign code positive */
#define PARI_SIGN_POS  0x01ul
/* pari sign code positive infinity */
#define PARI_SIGN_PINF 0x02ul
/* pari sign code not a number */
#define PARI_SIGN_NAN  0x80ul

/* pari exponent bias */
#define PARI_EXP_BIAS 0x00800000l
/* pari biased exponent minimum */
#define PARI_EXP_MIN  0x00000000l
/* pari biased exponent maximum */
#define PARI_EXP_MAX  0x00FFFFFFl

/* decimal logarithm of maximum sword value rounded down */
#define PARI_LOG10_SWORD_MAX 9
/* decimal digits needed for one data word */
#define PARI_UWORD_DEC_DIG 10
/* smallest decimal exponent for 1 <= mantissa < 10 */
#define PARI_DEC_EXP_MIN -2525223l
/* greatest decimal exponent for 1 <= mantissa < 10 */
#define PARI_DEC_EXP_MAX  2525222l
/* greatest decimal exponent string length */
#define PARI_DEC_EXP_LEN 8

/* log10(2) rounded down encoded in two uwords */
#define LOG10_2_MUL  301029995ul
#define LOG10_2_DIV 1000000000ul

#ifdef PARI_LITTLE_ENDIAN

/* pari object length */
#define parilen(arg)    ((unsigned short*) arg)[0]
/* pari object type and reference count */
#define parityp(arg)    ((unsigned short*) arg)[1]
/* pari object sign */
#define parisgn(arg)    ((unsigned char*)  arg)[7]
/* pari integer effective length */
#define parieff(arg)    ((unsigned short*) arg)[2]
/* pari real sign and biased exponent */
#define parisgnexp(arg) ((unsigned long*)  arg)[1]
/* pari object mantissa most significant word */
#define parimanmsw(arg) ((unsigned long*)  arg)[2]
/* pari real get biased exponent */
#define getpariexp(arg) (((unsigned long*) arg)[1] & 0x00FFFFFFul)

#else  /* #ifdef PARI_LITTLE_ENDIAN */

/* pari object length */
#define parilen(arg)    ((unsigned short*) arg)[1]
/* pari object type and reference count */
#define parityp(arg)    ((unsigned short*) arg)[0]
/* pari object sign */
#define parisgn(arg)    ((unsigned char*)  arg)[4]
/* pari integer effective length */
#define parieff(arg)    ((unsigned short*) arg)[3]
/* pari real sign and biased exponent */
#define parisgnexp(arg) ((unsigned long*)  arg)[1]
/* pari object mantissa most significant word */
#define parimanmsw(arg) ((unsigned long*)  arg)[2]
/* pari real get biased exponent */
#define getpariexp(arg) (((unsigned long*) arg)[1] & 0x00FFFFFFul)

#endif /* #ifdef PARI_LITTLE_ENDIAN */

/* long has 64 bits */
#else  /* #if ULONG_MAX == 0xFFFFFFFFul */

/* define logical types for pari words */
typedef unsigned long uword;
typedef signed long sword;
typedef unsigned int uhalf;
typedef signed int shalf;
typedef unsigned short usign;
typedef signed short ssign;

/* bits in word */
#define WORD_BITS 64
/* binary logarithm of bits in word */
#define LB_WORD_BITS 6
/* bytes in word */
#define WORD_BYTES 8
/* binary logarithm of bytes in word */
#define LB_WORD_BYTES 3

/* signed word minimum */
#define SWORD_MIN -0x8000000000000000ul
/* signed word maximum */
#define SWORD_MAX  0x7FFFFFFFFFFFFFFFul
/* unsigned word most significant bit */
#define UWORD_MSB  0x8000000000000000ul
/* unsigned word maximum */
#define UWORD_MAX  0xFFFFFFFFFFFFFFFFul

/* pari type code none */
#define PARI_TYPE_NONE 0x00000001ul
/* pari type code integer */
#define PARI_TYPE_INT  0x00010001ul
/* pari type code integer */
#define PARI_TYPE_REAL 0x00020001ul

/* maximum pari object length */
#define PARI_LEN_MAX 0x00000000FFFFFFFFl

/* pari object sign shift */
#define PARI_SIGN_SHIFT 48
/* pari sign code negative infinity */
#define PARI_SIGN_NINF 0xFFFEul
/* pari sign code negative */
#define PARI_SIGN_NEG  0xFFFFul
/* pari sign code zero */
#define PARI_SIGN_ZERO 0x0000ul
/* pari sign code positive */
#define PARI_SIGN_POS  0x0001ul
/* pari sign code positive infinity */
#define PARI_SIGN_PINF 0x0002ul
/* pari sign code not a number */
#define PARI_SIGN_NAN  0x8000ul

/* pari exponent bias */
#define PARI_EXP_BIAS 0x0000800000000000l
/* pari biased exponent minimum */
#define PARI_EXP_MIN  0x0000000000000000l
/* pari biased exponent maximum */
#define PARI_EXP_MAX  0x0000FFFFFFFFFFFFl

/* decimal logarithm of maximum sword value rounded down */
#define PARI_LOG10_SWORD_MAX 18
/* decimal digits needed for one data word */
#define PARI_UWORD_DEC_DIG 20
/* smallest decimal exponent for 1 <= mantissa < 10 */
#define PARI_DEC_EXP_MIN -42366205509364l
/* greatest decimal exponent for 1 <= mantissa < 10 */
#define PARI_DEC_EXP_MAX  42366205509363l
/* greatest decimal exponent string length */
#define PARI_DEC_EXP_LEN 15

/* log10(2) rounded down encoded in two uwords */
#define LOG10_2_MUL  3010299956639811952ul
#define LOG10_2_DIV 10000000000000000000ul

#ifdef PARI_LITTLE_ENDIAN

/* pari object length */
#define parilen(arg)    ((unsigned int*)   arg)[0]
/* pari object type and reference count */
#define parityp(arg)    ((unsigned int*)   arg)[1]
/* pari object sign */
#define parisgn(arg)    ((unsigned short*) arg)[7]
/* pari integer effective length */
#define parieff(arg)    ((unsigned int*)   arg)[2]
/* pari real sign and biased exponent */
#define parisgnexp(arg) ((unsigned long*)  arg)[1]
/* pari object mantissa most significant word */
#define parimanmsw(arg) ((unsigned long*)  arg)[2]
/* pari real get biased exponent */
#define getpariexp(arg) (((unsigned long*) arg)[1] & 0x0000FFFFFFFFFFFFl)

#else  /* #ifdef PARI_LITTLE_ENDIAN */

/* pari object length */
#define parilen(arg)    ((unsigned int*)   arg)[1]
/* pari object type and reference count */
#define parityp(arg)    ((unsigned int*)   arg)[0]
/* pari object sign */
#define parisgn(arg)    ((unsigned short*) arg)[4]
/* pari integer effective length */
#define parieff(arg)    ((unsigned int*)   arg)[3]
/* pari real sign and biased exponent */
#define parisgnexp(arg) ((unsigned long*)  arg)[1]
/* pari object mantissa most significant word */
#define parimanmsw(arg) ((unsigned long*)  arg)[2]
/* pari real get biased exponent */
#define getpariexp(arg) (((unsigned long*) arg)[1] & 0x0000FFFFFFFFFFFFl)

#endif /* #ifdef PARI_LITTLE_ENDIAN */

#endif /* #if ULONG_MAX == 0xFFFFFFFFul */

/* choose word types (#endif) */
#endif /* #ifdef PARI_WORD_LONG_LONG */

/* pari integer default length */
extern sword pariintdeflen;

/* default memory allocation error handler */
void defaulterrormalloc(void);
/* pointer to memory allocation error handler */
extern void (*errormalloc)(void);

/* default pari object memory allocation error handler */
void defaulterrorparimalloc(void);
/* pointer to pari object memory allocation error handler */
extern void (*errorparimalloc)(void);

/* copy double to pari real with directed rounding, word size dependent */
void rdcopyrealdouble(sword *z, double x, ssign r);
/* copy pari real to double with directed rounding, word size dependent */
double rdcopydoublereal(sword *x, ssign r);

#ifdef INTEL80BIT

/* copy long double to pari real with directed rounding, word size dependent */
void rdcopyreallongdouble(sword *z, long double x, ssign r);
/* copy pari real to long double with directed rounding, word size dependent */
long double rdcopylongdoublereal(sword *x, ssign r);

#endif /* #ifdef INTEL80BIT */

/* uword hexadecimal output for debugging */
void uwordhexout(uword arg);
/* uword array hexadecimal output for debugging */
void uwordarrayhexout(sword *x, sword lx);
/* pari object hexadecimal output for debugging */
void parihexout(sword *x);

/* dynamic expansion of pari integer */
sword *dynexpandint(sword *z, sword fz);
/* dynamic copying of sword to pari integer */
sword *dyncopyis(sword *z, sword x);
/* dynamic copying of pari integer to pari integer */
sword *dyncopyii(sword *z, sword *x);
/* copying of pari integer to sword with error return */
int copysi(sword *a, sword *x);

/* dynamic pari integer and sword addition */
sword *dynaddis(sword *z, sword *x, sword y);
/* dynamic pari integer and sword multiplication */
sword *dynmulis(sword *z, sword *x, sword y);
/* dynamic pari integer and sword shift */
sword *dynshiftis(sword *z, sword *x, sword y);
/* pari integer and sword division */
sword moddivis(sword *x, sword y);

/* pari integer comparison */
usign compareii(sword *x, sword *y);
/* dynamic pari integer subtraction and addition */
sword *dynaddsubii(sword *z, sword *x, sword *y, int sub);
/* dynamic pari integer multiplication */
sword *dynmulii(sword *z, sword *x, sword *y);
/* dynamic pari integer division */
void dyndivmodii(sword **div, sword **mod, sword *x, sword *y);

/* dynamic decimal string to pari integer conversion */
sword *dyncopyit(sword *z, const char *s);
/* dynamic pari integer to decimal string conversion */
char *mallocedcopyti(sword *x);

/* dynamic pari real constructor */
sword *dynconstructreal(sword lz);

/* minimum finite number */
void setminfin(sword *z);
/* maximum negative number */
void setmaxneg(sword *z);
/* minimum positive number */
void setminpos(sword *z);
/* maximum finite number */
void setmaxfin(sword *z);

/* pari real decrement or increment with directed rounding */
void rddecinc(sword *z, ssign d, ssign r);
/* pari real to real conversion with directed rounding */
void rdcopyrr(sword *z, sword *x, ssign r);
/* pari real absolute value */
void rdcopyabsrr(sword *z, sword *x, ssign r);
/* pari integer to real absolute value */
void rdcopyabsri(sword *z, sword *x, ssign r);
/* pari real negation */
void rdcopynegrr(sword *z, sword *x, ssign r);
/* pari integer to real negation */
void rdcopynegri(sword *z, sword *x, ssign r);
/* pari integer to real conversion with directed rounding */
void rdcopyri(sword *z, sword *x, ssign r);
/* sword to pari real conversion */
void copyrs(sword *z, sword x);

/* pari real comparison */
usign comparerr(sword *x, sword *y);
/* pari real and pari integer comparison */
usign compareri(sword *x, sword *y);
/* pari real subtraction and addition with directed rounding */
void rdaddsubrr(sword *z, sword *x, sword *y, int sub, ssign r);
/* pari real multiplication with directed rounding */
void rdmulrr(sword *z, sword *x, sword *y, ssign r);
/* pari real division with directed rounding */
void rddivrr(sword *z, sword *x, sword *y, ssign r);

/* pari real and integer addition with directed rounding */
void rdaddri(sword *z, sword *x, sword *y, ssign r);
/* pari real and integer subtraction with directed rounding */
void rdsubri(sword *z, sword *x, sword *y, ssign r);
/* pari real and integer multiplication with directed rounding */
void rdmulri(sword *z, sword *x, sword *y, ssign r);
/* pari real and integer division with directed rounding */
void rddivri(sword *z, sword *x, sword *y, ssign r);

/* pari integer and real subtraction with directed rounding */
void rdsubir(sword *z, sword *x, sword *y, ssign r);
/* pari integer and real division with directed rounding */
void rddivir(sword *z, sword *x, sword *y, ssign r);

/* pari integer and integer to real addition with directed rounding */
void rdaddii(sword *z, sword *x, sword *y, ssign r);
/* pari integer and integer to real subtraction with directed rounding */
void rdsubii(sword *z, sword *x, sword *y, ssign r);
/* pari integer and integer to real multiplication with directed rounding */
void rdmulii(sword *z, sword *x, sword *y, ssign r);
/* pari integer and integer to real division with directed rounding */
void rddivii(sword *z, sword *x, sword *y, ssign r);

/* pari real fission of integer and fractional part */
void rdmodreal(sword *ip, sword *fp, sword *x, ssign r);
/* pari real fission of mantissa and exponent part */
sword rdfission(sword *mant, sword *x, ssign r);
/* pari real fusion of mantissa and exponent part */
void rdfusion(sword *z, sword *mant, sword expo, ssign r);
/* pari real power of 10 with directed rounding */
void rdpowerof10(sword *z, sword expo, ssign r);

/* decimal string to pari real conversion with directed rounding */
int rdcopyrt(sword *z, const char *s, ssign r);
/* convert pari real length to decimal digits */
sword lentodecdigits(sword len);
/* convert decimal digits to pari real length */
sword decdigitstolen(sword dig);
/* dynnamic rounding of pari real to pari integer */
sword *dynroundir(sword *z, sword *x, ssign r);
/* dynamic pari real to decimal string conversion with directed rounding */
char *mallocedroundtr(sword *x, sword digits, int format, ssign r);

#if defined(__cplusplus) && !defined(PARI_CPLUSPLUS_COMPILED)
}
#endif

#endif /* #ifndef PARI_ARITHMETIC_H */
